# Ornstein–Uhlenbeck process

|  | Auxiliary | Description |
| --- | --- | --- |
| 1 | __"Ornstein–Uhlenbeck"__ | __Randomise the electrical network's dynamics__ |


## Summary

- In practicality an energy network is always in flux. 
- The stochastic nature of the the loads and renewable generation can be simulated with ElectricGrid.jl. 
- This notebooks will show you how to add stochastic reference signals via Ornstein–Uhlenbeck processes.

The Ornstein-Uhlenbeck process is described by the following stochastic differential equation (SDE):
```math

  dX_t = κ(γ-X_t)dt + σdW_t
```

Here, $γ$ is the long term mean, $κ$ is the mean reversion parameter, $W_t$ denotes a Wiener process, and $σ$ scales the random impact. 



```julia
using ElectricGrid;
```

_______________________________________________________________________________
### Network Configuration 

The following grid example is configured with
- Two inverters, a load, and two cables.
- The one inverter is placed in the familiar PQ mode, while the other is in VSG mode.
- The active and reactive power set points are confired as Ornstein–Uhlenbeck processes.
- The simulation is run for multiple episodes.



```julia
t_end    = 0.2     # total run time, seconds
num_eps  = 4       # number of episodes to run   

# Connectivity Matrix
CM = [ 0. 0. 1.
        0. 0. 2.
        -1. -2. 0.]  

# Nominal 3-phase power rating [VA], power factor, nominal rms voltage
R_load, L_load, _, _ = ParallelLoadImpedance(10e3, 0.95, 230) 

parameters = Dict{Any, Any}(
        "source" => Any[
                        Dict{Any, Any}("pwr"       => 200e3,
                                        "mode"     => "VSG")
                        Dict{Any, Any}("pwr"       => 100e3,   
                                        "mode"     => "PQ",    
                                        "p_set"    => -30e3, # Asymptotoic Mean
                                        "q_set"    => 10e3,
                                        "std_asy"  => 2.5e3, # Asymptotic Standard Deviation
                                        "σ"        => 100e3, # Random impact scale
                                        "Δt"       => 0.01,  # Process time Step, seconds
                                        "X₀"       => 0.0,   # Initial Process Value, Watt
                                        "k"        => 0)     # Interpolation degree, {0, 1, 2}
                        ],
        "load"   => Any[
                        Dict{Any, Any}("impedance" => "RL", 
                                        "R"        => R_load, 
                                        "L"        => L_load),
                        ],
        "cable"   => Any[
                        Dict{Any, Any}("R"  => 0.1, 
                                        "L" => 0.25e-3, 
                                        "C" => 0.1e-4),
                        Dict{Any, Any}("R"  => 0.1, 
                                        "L" => 0.25e-3, 
                                        "C" => 0.1e-4),
                        ],
    );

env = ElectricGridEnv(CM = CM, parameters = parameters, t_end = t_end, verbosity = 2);

agents = SetupAgents(env);

hook = Simulate(agents, env, num_episodes = num_eps, );
```

_______________________________________________________________________________
### Rendering


```julia
RenderHookResults(hook = hook, 
                    episode = 2,
                    states_to_plot  = [], 
                    actions_to_plot = [],  
                    power_p_inv     = [2],   # Real power [Watts]
                    power_q_inv     = [2],   # Imaginary power [VAi]
                    )
```


```@raw html
<div
id = UMJsNmNNVpU5 > </div>
<script>
gd = 'UMJsNmNNVpU5'
require(['plotly'], function(plotly) {
























plotly.newPlot(gd,[{"mode":"lines","y":[0.0,0.0,0.0,-6.664812424768094e-10,-3.5680862729876443e-7,-1.0146432200273062e-5,-9.794547921441502e-5,-0.00052435245008619,-0.0019290640331530364,-0.005482526041520897,-0.012954516634331824,-0.026707375382488727,-0.0496389618194846,-0.08506997662106683,-0.13656197872920725,-0.20769154371421183,-0.3018154453842133,-0.4218744271795016,-0.57025399540252,-0.7486987717091969,-0.9582732615396958,-1.1993550282925038,-1.4716580059126425,-1.7742817670306994,-2.105782106272365,-2.4642575690385273,-2.8474420770432713,-3.2527966473562024,-3.677594261366509,-4.118995471920385,-4.574115162543411,-5.040080897246499,-5.514083424074751,-5.993418812697188,-6.475521738746966,-6.957990146475804,-7.438602153803639,-7.915326625361709,-8.386328669983845,-8.84997100893454,-9.30481190481258,-9.749600227926624,-10.183268294051699,-10.604923095141912,-11.0138364705838,-11.40943463613117,-11.791287355533457,-12.159096952127674,-12.512687291404418,-12.851992821370798,-13.17704772974442,-13.487975269019167,-13.784977307594213,-14.068324168881968,-14.338344811574613,-14.595417387213015,-14.839960201586756,-15.072423114104646,-15.293279427771342,-15.503018335410895,-15.702137982885363,-15.891139189383612,-16.070519841313875,-16.240769961892976,-16.402367454784322,-16.555774519176,-16.701434726735577,-16.839770736002897,-16.971182602677377,-17.096046632423537,-17.214714719378627,-17.327514039768463,-17.43473320335756,-17.536628566892468,-17.633445847601728,-17.72541979399058,-17.812767529514964,-17.895685805875846,-17.974352494696866,-18.04892972042881,-18.119566498439085,-18.186400245829606,-18.249557608587033,-18.30915521301226,-18.36530057893545,-18.41809314873596,-18.467625292339655,-18.51398319383434,-18.55724760698886,-18.59749450772012,-18.63479568140134,-18.669219268415237,-18.700830272191354,-18.729691022651252,-18.755861585680535,-18.779400115638683,-18.8003631558111,-18.818805896103587,-18.834782396802904,-18.848345783516784,-18.859548414809552,-18.868442022460474,-18.875077824675145,-18.879506613790408,-18.8817788208577,-18.88194455951451,-18.88005365106449,-18.87615563217158,-18.870299746374748,-18.862534920788605,-18.852909729656467,-18.84147234673249,-18.828270488589325,-18.813351350904547,-18.79676153968976,-18.77854699925193,-18.758752938586824,-18.737423757803203,-18.714602976060167,-18.69033316236321,-18.664655870359535,-18.63761157806651,-18.60923963324271,-18.579578204928254,-18.548664241540834,-18.516533435733386,-18.483220196134404,-18.448757625934107,-18.413177508181676,-18.376510297568167,-18.338785118396295,-18.300029768400165,-18.260270728055794,-18.219533174972824,-18.17784100296653,-18.13521684539291,-18.09168210233296,-18.04725697120508,-18.00196048043111,-17.95581052576346,-17.908823908924354,-17.86101637820198,-17.812402670692734,-17.76299655587915,-17.712810880255553,-17.661857612738103,-17.610147890622386,-17.55769206584872,-17.504499751358146,-17.450579867361228,-17.39594068731988,-17.34058988349395,-17.28453457188732,-17.227781356469684,-17.170336372544455,-17.112205329142057,-17.0533935503765,-16.99390601563009,-16.933747398535687,-16.87292210466223,-16.81143430787322,-16.749287985290497,-16.68648695086298,-16.623034887481534,-16.558935377633134,-16.49419193260612,-16.428808020212948,-16.362787091042108,-16.296132603267957,-16.22884804600905,-16.160936961255615,-16.09240296440352,-16.023249763402163,-15.953481176569184,-15.883101149066647,-15.812113768110208,-15.740523276915084,-15.668334087431873,-15.595550791906774,-15.522178173288353,-15.448221214549918,-15.373685106926814,-15.298575257144194,-15.222897293639946,-15.14665707184156,-15.069860678518424,-14.99251443526434,-14.914624901138211,-14.836198874485092,-14.757243393994997,-14.67776573901682,-14.597773429163695,-14.517274223244726,-14.436276117557682,-14.354787343578323,-14.272816365052662,-14.190371874540489,-14.107462789449734,-14.024098247564972,-13.94028760210464,-13.856040416348115,-13.77136645784428,-13.686275692229568,-13.600778276678362,-13.514884553007521,-13.428605040473203,-13.341950428283232,-13.254931567807592,-13.16755946454076,-13.079845269860161,-12.99180027255175,-12.903435890150327,-12.814763660108824,-12.725795230812857,-12.636542352460594,-12.547016867833548,-12.457230702964097,-12.36719585770686,-12.276924396270786,-12.18642843767526,-12.095720146195049,-12.004811721765208,-11.913715390394785,-11.822443394577476,-11.731007983741137,-11.639421404722754,-11.547695892300197,-11.455843659759068,-11.363876889585441,-11.271807724197572,-11.1796482567959,-11.08741052230144,-10.995106488451341,-10.902748046971936,-10.810347004934926,-10.717915076229957,-10.625463873211402,-10.533004898503492,-10.440549536959395,-10.348109047843252,-10.255694557151656,-10.163317050175621,-10.070987364227236,-9.978716181589405,-9.886514022668905,-9.794391239383074,-9.70235800877279,-9.610424326812463,-9.518600002483183,-9.426894652081138,-9.335317693740201,-9.24387834223004,-9.152585603983443,-9.061448272375852,-8.970474923251428,-8.879673910722637,-8.789053363198292,-8.69862117968114,-8.60838502632415,-8.518352333249794,-8.428530291620739,-8.338925850963165,-8.249545716771792,-8.160396348337457,-8.07148395687615,-7.982814503882082,-7.8943936997463595,-7.806227002641691,-7.7183196176482625,-7.630676496122362,-7.543302335324682,-7.456201578310186,-7.369378414048555,-7.282836777750958,-7.196580351513603,-7.110612565111865,-7.024936597104329,-6.939555376101936,-6.854471582292916,-6.76968764918638,-6.6852057655707995,-6.601027877697675,-6.517155691636589,-6.4335906758902865,-6.350334064151411,-6.267386858312534,-6.184749831605785,-6.102423531985949,-6.02040828566075,-5.938704200810207,-5.857311171459742,-5.776228881548917,-5.695456809124494,-5.614994230723681,-5.534840225883812,-5.454993681799678,-5.3754532981261605,-5.296217591917738,-5.217284902683872,-5.138653397573123,-5.060321076676942,-4.982285778421094,-4.904545185090761,-4.827096828464711,-4.749938095502268,-4.67306623414743,-4.596478359207307,-4.520171458291017,-4.4441423978682195,-4.368387929337601,-4.292904695181065,-4.217689235169502,-4.142737992602065,-4.06804732062227,-3.993613488552967,-3.9194326882403345,-3.845501040433504,-3.7718146012203135,-3.6983693684542516,-3.6251612881434596,-3.5521862609221557,-3.4794401484818707,-3.4069187799914573,-3.334617958522614,-3.2625334674206954,-3.190661076702156,-3.118996549397991,-3.0475356478515607,-2.9762741400070922,-2.905207805660382,-2.834332442591659,-2.763643872731649,-2.693137948221505,-2.622810557428153,-2.5526576308749327,-2.482675147113696,-2.4128591385251923,-2.3432056970191013,-2.273710979653771,-2.2043712141817764,-2.13518270447355,-2.066141835877346,-1.9972450804218056,-1.9284890019609406,-1.8598702612071918,-1.7913856206037337,-1.7230319491003456,-1.6548062268270343,-1.5867055496104712,-1.5187271333734884,-1.450868318430416,-1.3831265735118308,-1.315499499839241,-1.247984834937423,-1.1805804562884457,-1.1132843848785825,-1.0460947885818541,-0.9790099853925511,-0.9120284464660244,-0.8451487990142965,-0.7783698290109129,-0.711690483775854,-0.645109874357388,-0.5786272777578176,-0.5122421389427814,-0.4459540726766136,-0.3797628652324079,-0.313668475874465,-0.24767103816162717,-0.18177086110443952,-0.11596843002439527,-0.05026440732552828,0.015340366902307778,0.08084487454959799,0.1462479193406807,0.21154812684875424,0.27674394469858044,0.3418336429592159,0.40681531476082,0.47168687710961876,0.5364460718649298,0.6010904669367392,0.6656174576385752,0.7300242682661633,0.7943079538781603,0.8584654022445761,0.922493335995398,0.9863883149534445,1.0501467386411036,1.113764848990428,1.1772387332784149,1.240564327128169,1.30373741780528,1.366753647649631,1.4296085176727615,1.4922973913251951,1.5548154984344649,1.6171579393378153,1.6793196891429574,1.7412956021704247,1.803080416515698,1.8646687588417947,1.9261680617318433,2.0021459628986893,2.1987494384074466,2.7585452278282787,3.977218060511106,6.106756850419288,9.290162762873138,13.555758037898784,18.855602900774855,25.100681670467964,32.1833455287761,39.982858577104906,48.36492225517412,57.184717330647544,66.29301495942059,75.54470272330937,84.80579858799683,93.95755291556986,102.89805997063166,111.54194653627485,119.81943086748157,127.67519437136303,135.06727617619293,141.96599013461955,148.35268950114684,154.21839550790094,159.56232541376025,164.39043920229497,168.71411796320564,172.54901159970674,175.9140578453511,178.83063480889388,181.32181806530014,183.41173247059228,185.1249992013583,186.48628302455103,187.5199357969388,188.24972539590038,188.69863624730434,188.88872816706348,188.8410432794731,188.57555247834458,188.1111339433886,187.46557669219533,186.655602581741,185.69690089516234,184.60417038436603,183.39116456628284,182.07073718527036,180.6548859674464,179.15479384430594,177.58086744253376,175.94277283762096,174.24946858572372,172.50923615464527,170.72970814886486,168.91789502291095,167.08021109940785,165.2225005751906,163.3500639151735,161.4676847755075,159.57965747613753,157.68981503448214,155.80155777493883,153.91788246560216,152.04141180796302,150.17442398518173,148.3188819219243,146.47646193392595,144.64858150903692,142.8364260150139,141.04097415276576,139.2630219813238,137.50320535920247,135.76202068912633,134.03984391103197,132.3369477424193,130.6535172025126,128.9896634765875,127.34543619016185,125.72083417701592,124.1158148421298,122.53030223554468,120.96419396036735,119.41736703644322,117.88968283473304,116.38099118957467,114.89113378961461,113.4199469428199,111.96726380434797,110.5329161474024,109.11673574740306,107.71855543965033,106.33820990256591,104.97553621148033,96.47121965292425,-4815.841575445237,-11042.50736959115,-13502.097049209657,-12250.654768868064,-10951.165314793496,-11455.73891572211,-13761.0876115278,-17032.40438950214,-20336.186107149275,-23312.822683342918,-26074.89183623891,-28789.23096560348,-31462.164606932492,-33945.141291789696,-36091.2947182027,-37840.23548249444,-39214.95322466837,-40274.173248045605,-41062.598244647845,-41601.657474918204,-41899.161131242065,-41964.94611199297,-41819.680568741154,-41492.55186638963,-41015.98455949507,-40420.358632703355,-39732.122974301994,-38974.59509254311,-38169.180123070095,-37336.14132112007,-36494.25352123013,-35660.015561663844,-34847.049010324256,-34065.94465738349,-33324.57602605168,-32628.5697406795,-31981.73922871113,-31386.397267708093,-30843.570550732176,-30353.194138060368,-29914.31576193852,-29525.310801115884,-29184.08273291311,-28888.226328086723,-28635.148908799867,-28422.155133392564,-28246.507240712457,-28105.470211370368,-27996.347105174496,-27916.506777177223,-27863.40410032965,-27834.592803564443,-27827.731616617642,-27840.585384578902,-27871.023424846335,-27917.017122458958,-27976.637824470818,-28048.05500553845,-28129.534153017637,-28219.43404039692,-28316.20367251918,-28418.379613129993,-28524.58430815469,-28633.525536012436,-28743.99664667455,-28854.877114831386,-28965.133124917895,-29073.81818049991,-29180.07384063032,-29283.130572983137,-29382.308515648376,-29477.017832087222,-29566.75839833457,-29651.11871235538,-29729.77404240741,-29802.483866137965,-29869.088615657347,-29929.50570682156,-29983.72484511703,-30031.80266229125,-30073.85680486694,-30110.05962862992,-30140.63164521679,-30165.834840059877,-30185.965961615315,-30201.34987988843,-30212.333118599585,-30219.277664704598,-30222.555144158723,-30222.54142846622,-30219.611712898994,-30214.1360911054,-30206.475641715133,-30196.979035494944,-30185.97966244936,-30173.79326676092,-30160.71606645908,-30147.0233270653,-30132.968354857072,-30107.07349526007,-30099.789102321123,-30093.999805085685,-30083.335816573275,-30069.637182004597,-30056.788760246814,-30046.656471041384,-30038.871640670877,-30032.350982559354,-30026.3664445901,-30020.833312321814,-30015.964421434895,-30011.931538538636,-30008.76077859749,-30006.375388801585,-30004.699503609707,-30003.697952897208,-30003.366629937867,-30003.705326636165,-30004.69505518323,-30006.29739613054,-30008.46206095743,-30011.13620669934,-30014.269731627683,-30017.81509038847,-30021.725750677975,-30025.954548288693,-30030.453562940682,-30035.175194883042,-30040.073387005046,-30045.10458266004,-30050.228024249758,-30055.405684039244,-30060.602136470636,-30065.78454001077,-30070.922789104,-30075.98970515037,-30080.961173481905,-30085.81617708145,-30090.53673659751,-30095.10780448326,-30099.517141269625,-30103.75518669339,-30107.81492013165,-30111.69170159945,-30115.383091755204,-30118.88865470155,-30122.209752039977,-30125.349336304218,-30128.31174971833,-30131.10253157912,-30133.728235109174,-30136.19625359833,-30138.51465566994,-30140.69203030755,-30142.737342934906,-30144.659803739763,-30146.468748619747,-30148.173532150384,-30149.783431498727,-30151.30756040427,-30152.754792912172,-30154.133696947792,-30155.452477787505,-30156.718931121897,-30157.940405056037,-30159.123770296108,-30160.27539792665,-30161.401144386924,-30162.50634332249,-30163.59580388214,-30164.673814860453,-30165.74415398525,-30166.81010167075,-30167.874458648123,-30168.939566962024,-30170.00733383505,-30171.079257871832,-30172.156457051642,-30173.2396979794,-30174.329425925294,-30175.425795253148,-30176.528699888382,-30177.637803504545,-30178.75256912485,-30179.872287862163,-30180.996106562132,-30182.123054159252,-30183.252066598798,-30184.38201020867,-30185.51170342795,-30186.639936821626,-30187.765491334823,-30188.887154766024,-30190.003736463856,-30191.11408027104,-30192.2170757551,-30193.31166777707,-30194.3968644609,-30195.47174363812,-31709.188535803416,-30792.98305692804,-29738.140168600916,-29376.69058161361,-29472.838392990503,-29544.520628492548,-29369.81689566971,-29012.239248582904,-28621.15884753181,-28292.74656992471,-28037.822188645365,-27829.248053297113,-27644.94737387392,-27480.21206791547,-27341.631974615833,-27234.250777735266,-27157.084744906708,-27104.723826390124,-27070.89406744128,-27051.280283239976,-27043.620473196817,-27046.830204581383,-27059.960469896243,-27081.668599812547,-27110.34188436308,-27144.362506279645,-27182.363696820314,-27223.290616352075,-27266.313020785277,-27310.718385766013,-27355.833109794192,-27401.01684789826,-27445.693531634846,-27489.382286898915,-27531.709026753975,-27572.392451319352,-27611.220900379893,-27648.031556580114,-27682.69817959254,-27715.12652298125,-27745.25211071035,-27773.03752124702,-27798.468158096868,-27821.547639244716,-27842.29425347427,-27860.739019283003,-27876.925211962433,-27890.90864108338,-27902.75798063768,-27912.55469015751,-27920.392357907596,-27926.375568356147,-27930.61849598696,-27933.24340080077,-27934.379078665588,-27934.159216259675,-27932.720591503272,-27930.20113940993,-27926.738002636674,-27922.465723332392,-27917.514688247553,-27912.009851705585,-27906.069696090726,-27899.80538099999,-27893.320064766038,-27886.708414991863,-27880.056326982096,-27873.440842775723,-27866.930233357663,-27860.58419547422,-27854.45412623715,-27848.58345921617,-27843.008058473042,-27837.756665793495,-27832.85138752142,-27828.308201477197,-27824.137466856948,-27820.344428239834,-27816.929712307072,-27813.889818140877,-27811.21759953844,-27808.902734412815,-27806.93217538465,-27805.290577563697,-27803.960702367778,-27802.92379787662,-27802.159955997096,-27801.648445551822,-27801.368019656125,-27801.29719606345,-27801.4145101778,-27801.698741350825,-27802.129113402894,-27802.68547012504,-27803.348426274628,-27804.099494624417,-27804.921189972178,-27805.797111429893,-27806.712004548175,-27807.651804841575,-27794.511091818917,-27804.385691316133,-27815.626322551016,-27820.10924432732,-27820.0358273348,-27820.155178960547,-27822.66187632295,-27826.957006108063,-27831.58254975255,-27835.571226233053,-27838.794194208174,-27841.51132251889,-27843.939622427286,-27846.12323181202,-27847.991066733914,-27849.4878769194,-27850.6216998592,-27851.448129430366,-27852.033326998193,-27852.424188285415,-27852.646912598917,-27852.715864481434,-27852.644277818465,-27852.449657649922,-27852.15238306348,-27851.772852494705,-27851.328754260867,-27850.8343986978,-27850.30160809248,-27849.740811651925,-27849.16184645315,-27848.57400111699,-27847.985686053413,-27847.404113410197,-27846.83518256315,-27846.283631478684,-27845.753280833505,-27845.247251993773,-27844.768096774376,-27844.317849521256,-27843.898057353188,-27843.50981771372,-27843.153833014105,-27842.8304700241,-27842.539808754784,-27842.28167529843,-27842.055659990525,-27841.861128099954,-27841.697229839912,-27841.562913981626,-27841.45694645132,-27841.377932692198,-27841.324341722324,-27841.294530171366,-27841.28676582213,-27841.29925121518,-27841.330147959823,-27841.377601618926,-27841.439766057556,-27841.514825778555,-27841.601015205448,-27841.69663473008,-27841.800063974817,-27841.909772791067,-27842.024330181022,-27842.142410994027,-27842.262800220728,-27842.384394968183,-27842.50620449907,-27842.627348830527,-27842.74705627261,-27842.864660082665,-27842.979594285396,-27843.091388724493,-27843.19966350234,-27843.304123026937,-27843.404549861727,-27843.50079849275,-27843.59278905225,-27843.680501015086,-27843.76396690739,-27843.843266101154,-27843.918518775525,-27843.989880105935,-27844.05753471049,-27844.121691365137,-27844.182577998865,-27844.240436989996,-27844.29552079046,-27844.348087899427,-27844.398399195776,-27844.446714627527,-27844.49329025215,-27844.53837562101,-27844.582211503304,-27844.625027942744,-27844.66704263602,-27844.70845961656,-27844.74946822445,-27844.79024234252,-26307.679312990284,-27222.236538385736,-28315.448246113232,-28692.047426932473,-28569.225231508826,-28466.7780059161,-28631.587717999373,-29006.894389265406,-29437.166771575052,-29813.233417657295,-30116.262937530708,-30372.389037984725,-30604.889037800825,-30818.344780367013,-31003.934766193797,-31154.120967026487,-31268.453331999786,-31352.15799628802,-31412.03338818207,-31452.9073871229,-31477.327141422662,-31486.495312055013,-31481.49750383789,-31463.960052753406,-31435.909993862268,-31399.443904125947,-31356.39369136927,-31308.229056608143,-31256.14900896605,-31201.20513843657,-31144.39532048272,-31086.669626464583,-31028.89121520547,-30971.79726371941,-30915.98363694729,-30861.921121933785,-30809.983158326897,-30760.470891189,-30713.62798972868,-30669.646255723586,-30628.668540787265,-30590.792476762723,-30556.076257189234,-30524.545081086046,-30496.1964688085,-30471.003772035914,-30448.918004568703,-30429.86880936419,-30413.765335973345,-30400.497511025784,-30389.93785115549,-30381.94367196314,-30376.35945474269,-30373.019177318398,-30371.748560453587,-30372.3673022448,-30374.69138615952,-30378.535462835778,-30383.71519461351,-30390.04940713439,-30397.361938083082,-30405.483167139857,-30414.251281398076,-30423.51333779013,-30433.126144664093,-30442.956944062975,-30452.88387031814,-30462.796187890555,-30472.594344788056,-30482.189890969894,-30491.505298571334,-30500.473697999252,-30509.038529765672,-30517.153113778015,-30524.780148837617,-30531.891163170818,-30538.465935401226,-30544.49189701664,-30549.963519496767,-30554.881687096156,-30559.253059413248,-30563.089432088214,-30566.40710539052,-30569.2262684351,-30571.5704034525,-30573.465712449935,-30574.940568592094,-30576.024995712818,-30576.750179980812,-30577.14801712283,-30577.25069715026,-30577.09032716426,-30576.69859214081,-30576.106453532953,-30575.343885575232,-30574.439648922893,-30573.42110067757,-30572.31403921255,-30571.142581790526,-30569.92907283489,-31403.500394729042,-30914.8078988852,-30342.471568545974,-30145.67198648031,-30200.85968307802,-30243.720875695686,-30150.940696490223,-29955.418334415393,-29738.361808857757,-29553.473242853666,-29407.894897369682,-29287.332804660033,-29179.805116588966,-29082.774906251896,-29000.06097273995,-28934.720946694983,-28886.458535591963,-28852.434481818353,-28829.234350358587,-28814.479721616102,-28806.91631256628,-28805.94381856743,-28811.03461031269,-28821.430561841345,-28836.208694145706,-28854.430775978828,-28875.291715208245,-28898.15866026257,-28922.524848596753,-28947.95062878137,-28974.01913466394,-29000.332985706264,-29026.531966194336,-29052.310889396096,-29077.42663015209,-29101.690587627898,-29124.95580808811,-29147.105312355194,-29168.045186315067,-29187.70201054463,-29206.021627049216,-29222.967612856526,-29238.51886328672,-29252.666914167905,-29265.413825402382,-29276.770937352776,-29286.758433616607,-29295.40531297559,-29302.749383893388,-29308.837028535952,-29313.722647107686,-29317.46784419816,-29320.14047308454,-29321.813638053896,-29322.564686110753,-29322.47416095533,-29321.624686038464,-29320.099786284885,-29317.982712799716,-29315.355355841763,-29312.297307088014,-29308.88508403596,-29305.191493274422,-29301.285104448067,-29297.22982484965,-29293.084583228014,-29288.903133167423,-29284.73397223612,-29280.620356462292,-29276.6003833811,-29272.707123283293,-29268.96878965479,-29265.408946932894,-29262.046753074363,-29258.89722944096,-29255.971547051595,-29253.277319404697,-29250.818896516877,-29248.597658923158,-29246.61231163017,-29244.859176645485,-29243.332480803678,-29242.02463503486,-29240.92650226217,-29240.027652723013,-29239.31660647741,-29238.781062804628,-29238.40811559407,-29238.18445346808,-29238.096543594856,-29238.130798768318,-29238.273727898388,-29238.512070290133,-29238.83291403528,-29239.22379874291,-29239.672802893827,-29240.16861632495,-29240.70059860644,-29241.25882423018,-29241.834115553727,-25615.593292271722,-27487.93086116414,-29875.561842810628,-30685.10135847286,-30336.14291021081,-30016.91596432102,-30315.490785805734,-31130.31055977879,-32117.989762864545,-33017.8683969589,-33770.81818233453,-34428.08513516481,-35041.03556962802,-35619.88098911489,-36141.75442148639,-36584.81111018929,-36943.44648937232,-37226.43746186683,-37447.72971504976,-37617.50992675492,-37740.89971193523,-37819.88020656472,-37856.254662650914,-37853.4096364108,-37816.07478066726,-37749.50795990912,-37658.58356988721,-37547.44584960888,-37419.662671001395,-37278.49424076194,-37127.11105153068,-36968.596794793506,-36805.83378007354,-36641.38505993592,-36477.43695677603,-36315.82549640587,-36158.09724335188,-36005.56747746972,-35859.35496333205,-35720.394578173786,-35589.443821803994,-35467.09223287336,-35353.77706754277,-35249.801816057385,-35155.35284893122,-35070.51216426143,-34995.26624801833,-34929.512855168636,-34873.06748842869,-34825.67065590366,-34786.99618417884,-34756.660155297446,-34734.229838788546,-34719.23212760635,-34711.16136529818,-34709.486769850046,-34713.65971329314,-34723.120924084076,-34737.307413229275,-34755.658808175016,-34777.622874541965,-34802.660220627564,-34830.24834245072,-34859.885180739606,-34891.09225994588,-34923.417370799405,-34956.43673107452,-34989.756615311584,-35023.01452033415,-35055.8799651107,-35088.054994950595,-35119.2744061477,-35149.30567315459,-35177.94856559688,-35205.03447172178,-35230.42546892431,-35254.01318384795,-35275.717468991876,-35295.48490710047,-35313.287151851306,-35329.11912317145,-35342.99708798073,-35354.95666273762,-35365.050770849324,-35373.34758073012,-35379.92844521529,-35384.88586254642,-35388.321481104285,-35390.34417080351,-35391.06818166484,-35390.61140542466,-35389.093751386135,-35386.635644639486,-35383.3566532056,-35379.374249450484,-35374.80270927896,-35369.75215004923,-35364.32770549297,-35358.62883386214,-35352.7487542669,-40834.54085680533,-37768.11283756506,-34244.70101630354,-33047.73680808738,-33338.11942397686,-33540.81270489245,-32936.049366238956,-31750.3865950258,-30483.818301167244,-29441.440910391328,-28646.708288087528,-28007.099127298923,-27450.342130619872,-26959.510413280164,-26551.986374432865,-26240.217567533447,-26019.04877870811,-25870.975583826916,-25776.66862592095,-25723.038496517525,-25703.297052250946,-25714.276349567954,-25753.312569622893,-25816.65341753519,-25899.78096913595,-25998.155674906455,-26107.951436071708,-26226.23039699172,-26350.68815822062,-26479.341207268604,-26610.294295668657,-26741.723310856854,-26871.968872521742,-26999.632677539015,-27123.61803401858,-27243.093507895934,-27357.427083624105,-27466.12512171468,-27568.79560131751,-27665.13424209946,-27754.918225184858,-27837.999003238278,-27914.290886180315,-27983.758566998855,-28046.407863422468,-28102.28135839935,-28151.45859918749,-28194.058664102788,-28230.242851581515,-28260.215906685127,-28284.225127720652,-28302.557602046643,-28315.536207759607,-28323.51497756262,-28326.87404922588,-28326.014099417614,-28321.35012552161,-28313.304691593992,-28302.301070668076,-28288.756831364164,-28273.07826755455,-28255.655781314268,-28236.860111352573,-28217.03926097624,-28196.516073549436,-28175.586500803067,-28154.51861195486,-28133.552306555728,-28112.899598107128,-28092.745300514493,-28073.247987251743,-28054.541158963144,-28036.73459717549,-28019.91588033883,-28004.1520130952,-27989.49110335247,-27975.964031108746,-27963.586080094756,-27952.358527751607,-27942.270197027327,-27933.29896661976,-27925.41322633899,-27918.57326152813,-27912.732556449122,-27907.839015325942,-27903.83610472612,-27900.663920119223,-27898.26017565195,-27896.561113653363,-27895.502331121064,-27895.01952334118,-27895.04914740411,-27895.529009158912,-27896.3987763967,-27897.600420134946,-27899.078585868152,-27900.780897592147,-27902.658198567828,-27904.66473343696,-27906.758276262684,-27032.629595233146,-27567.89422729269,-28198.47179107765,-28417.357795923366,-28353.524492853947,-28302.20803838106,-28402.22270073349,-28619.404237414146,-28864.523929637548,-29076.384338898326,-29245.42638484045,-29387.051339035814,-29514.593797555328,-29630.704045522554,-29730.598809966203,-29810.324813241827,-29869.9116068816,-29912.494028297253,-29941.993046943702,-29961.170609032648,-29971.49196125329,-29973.66714583839,-29968.33527041731,-29956.42206649558,-29939.057452607623,-29917.39277168471,-29892.42083065685,-29864.927698354873,-29835.54710905393,-29804.830313776183,-29773.298314145864,-29741.445367197837,-29709.71790743994,-29678.493689330688,-29648.07400426626,-29618.693103062706,-29590.53371232163,-29563.740909469998,-29538.43028132096,-29514.690995408415,-29492.587394504986,-29472.16102747116,-29453.43377840383,-29436.41131714982,-29421.08588903786,-29407.43808360405,-29395.437665521622,-29385.043927710758,-29376.205999920072,-29368.863383675485,-29362.946799849946,-29358.37927089318,-29355.07730786892,-29352.952095817436,-29351.910650180755,-29351.85698295641,-29352.693323545085,-29354.321391045072,-29356.643652899125,-29359.56448051668,-29362.99113872749,-29366.834598862886,-29371.010204645732,-29375.438224305006,-29380.044300685913,-29384.759789038446,-29389.52196945246,-29394.27413655853,-29398.9655878542,-29403.55153920541,-29407.99298889967,-29412.256538829213,-29416.314173436876,-29420.142998030635,-29423.72494406849,-29427.046453274786,-29430.0981513993,-29432.87451756437,-29435.373550615303,-29437.596432594706,-29439.547191141628,-29441.232364886455,-29442.660676635012,-29443.84271796411,-29444.790647005153,-29445.517900060167,-29446.038917741203,-29446.368886978875,-29446.52350063553,-29446.51873616557,-29446.370654010316,-29446.095215717174,-29445.708121463213,-29445.22466668899,-29444.65961763002,-29444.027105437835,-29443.340538302422,-29442.612530671206,-29441.85484846175,-29441.07836912481,-31553.026832318254,-30202.067641630678,-28679.864810744795,-28155.331601508096,-28277.092631182924,-28361.194018486975,-28097.967057028945,-27583.304515839096,-27031.06310476577,-26574.737916588972,-26226.021859459932,-25944.571693210575,-25698.648218010414,-25481.468806012577,-25301.836787280314,-25166.065062779708,-25071.9973202313,-25011.50450481961,-24975.53441937031,-24957.950019524724,-24955.54956721213,-24966.781918432323,-24990.290032037068,-25024.201117694825,-25066.315580761944,-25114.474260319672,-25166.898163513393,-25222.254349497467,-25279.522936828493,-25337.84437124079,-25396.410250831843,-25454.456015011325,-25511.302976530744,-25566.399034954673,-25619.3324842563,-25669.811226333,-25717.630372825657,-25762.643787240464,-25804.74763969533,-25843.87445752306,-25879.99024116169,-25913.090807975695,-25943.196084155083,-25970.343986139465,-25994.58588562494,-26015.98434791708,-26034.612911861946,-26050.556899425683,-26063.914297577703,-26074.79609032167,-26083.325816084624,-26089.63848771261,-26093.87914001459,-26096.201230789047,-26096.764956848532,-26095.735410376838,-26093.280494167117,-26089.568625206273,-26084.766392167723,-26079.036381915,-26072.535327510956,-26065.412610523206,-26057.80906270544,-26049.856002011336,-26041.674483152274,-26033.374787407807,-26025.056178182487,-26016.806911903266,-26008.7044523853,-26000.81582202972,-25993.198039965944,-25985.898625635444,-25978.956163649767,-25972.40092386927,-25966.25551851862,-25960.535570380118,-25955.250369914003,-25950.403510828954,-25945.993504022368,-25942.01437287117,-25938.45622943587,-25935.305826517677,-25932.547079231234,-25930.161552385245,-25928.128913805034,-25926.427355828626,-25925.03398669223,-25923.92519172199,-25923.07696306421,-25922.465196982226,-25922.065959006934,-25921.855718332346,-25921.811553126492,-25921.911328027156,-25922.13384463835,-25922.45896581357,-25922.867714914555,-25923.342351727693,-25923.866426979024,-25924.42481734105,-22098.536119372813,-24310.690663819027,-27065.651520324118,-28014.340369617603,-27636.046499934928,-27294.644209224844,-27657.372544056947,-28604.969250144364,-29744.32995141993,-30776.935005977386,-31636.22276667565,-32382.912573899102,-33076.73177880928,-33728.69038688198,-34311.4557916046,-34799.6757778442,-35187.681741037006,-35486.844208031354,-35714.37668938498,-35882.81527610583,-35998.62264978347,-36064.58133725397,-36083.29099554012,-36059.159786701595,-35998.04448109757,-35906.26243896428,-35789.54327471323,-35652.66632021653,-35499.67626403828,-35334.21509066618,-35159.780496647574,-34979.730095752704,-34797.15308735683,-34614.743186322565,-34434.74380023561,-34258.98897660005,-34088.98078326427,-33925.96066816817,-33770.95223118577,-33624.77818562224,-33488.070544905444,-33361.28421009571,-33244.71738281878,-33138.53453555991,-33042.786487289035,-32957.42543060624,-32882.3151196233,-32817.23841673197,-32761.904255646437,-32715.955232387863,-32678.97613220937,-32650.502931381317,-32630.03162560755,-32617.026402335254,-32610.927091038015,-32611.156165477776,-32617.12561340455,-32628.24376147129,-32643.921838597773,-32663.57992888403,-32686.652073934616,-32712.590524872845,-32740.86932376563,-32770.98740452199,-32802.471285837644,-32834.877306414695,-32867.79332520187,-32900.8398761027,-32933.67085450678,-32965.973848888796,-32997.47019756559,-33027.914789284696,-33057.095588257245,-33084.832871009676,-33110.97819594413,-33135.41315308158,-33158.04794214212,-33178.81980794377,-33197.691343234044,-33214.648665308094,-33229.699483454635,-33242.87108786426,-33254.208296234225,-33263.77138982823,-33271.63406201416,-33277.88139640657,-33282.60789130507,-33285.915549570665,-33287.91205414651,-33288.70904693433,-33288.420523747314,-33287.1613531859,-33285.0459243031,-33282.18692669909,-33278.694265858096,-33274.67411496266,-33270.22810199102,-33265.45262837044,-33260.43831360532,-33255.269559346154,-38864.68823540141,-35441.603742928455,-31626.8142087085,-30324.403769113578,-30593.143486729117,-30759.811076691043,-30077.55880664452,-28804.31800266197,-27475.296919834393,-26404.77045924009,-25606.368207935397,-24976.03781450295,-24435.662657724693,-23967.160631651357,-23587.821104729574,-23308.793450865003,-23122.601033378178,-23009.387053966268,-22948.175907849254,-22925.046872147323,-22932.87350887629,-22968.31997891573,-23028.556739275962,-23109.6736021957,-23207.076509153045,-23316.273466639344,-23433.603434350054,-23556.367672596527,-23682.530453504034,-23810.38080043832,-23938.29215566974,-24064.708521952358,-24188.240919295393,-24307.762112799515,-24422.44178719621,-24531.70383663596,-24635.156194759842,-24732.527947279697,-24823.632461519654,-24908.3538511512,-24986.640618390862,-25058.49795236671,-25123.97569295214,-25183.15551893362,-25236.141793807845,-25283.057675638403,-25324.046011597067,-25359.272708202887,-25388.930299622774,-25413.240148175853,-25432.45264546571,-25446.845671699604,-25456.72192763007,-25462.405692495737,-25464.2391811629,-25462.578351852513,-25457.787999923734,-25450.23624135268,-25440.288815512773,-25428.303752698488,-25414.62679738299,-25399.5876866368,-25383.49716852371,-25366.644615489597,-25349.296189372217,-25331.69361456099,-25314.05361559893,-25296.56798581733,-25279.404155174067,-25262.70609134668,-25246.59540856305,-25231.172626274893,-25216.518561357858,-25202.695834283528,-25189.75044290982,-25177.713341174123,-25166.601970538402,-25156.421720510065,-25147.167319574335,-25138.82416547115,-25131.369596147986,-25124.77409227602,-25119.002399458725,-25114.014564408364,-25109.76688813135,-25106.212803743885,-25103.303685020077,-25100.98958730325,-25099.21991938944,-25097.944045411256,-25097.11181837904,-25096.674049309717,-25096.582916252795,-25096.79231637722,-25097.258163059945,-25097.938629711556,-25098.794342886027,-25099.788528262383,-25100.887113599623,-25102.058792595846,-25089.321952815546,-25100.493962639062,-25112.83516903211,-25117.919816136677,-25118.18121687917,-25118.65116790456,-25121.59995660718,-25126.344344280573,-25131.324284310278,-25135.530987588565,-25138.851750560774,-25141.576888507672,-25143.94048999365,-25145.988643371777,-25147.64778721662,-25148.86531975658,-25149.65785522861,-25150.091823651892,-25150.242826028134,-25150.164464544774,-25149.887800600594,-25149.431410522964,-25148.812772976773,-25148.0536405611,-25147.178276041195,-25146.21036315398,-25145.17025170728,-25144.074441185934,-25142.93664663378,-25141.768998655432,-25140.58287895075,-25139.388951608118,-25138.19682151852,-25137.014721214997,-25135.849419054124,-25134.706398686085,-25133.590123672628,-25132.50426626561,-25131.451841279373,-25130.435260781145,-25129.45636932199,-25128.516488933266,-25127.61648232662,-25126.756820223152,-25125.937636780676,-25125.15876787494,-25124.419774154296,-25123.719956541172,-25123.058371042123,-25122.433846977823,-25121.845009825316,-25121.290307349496,-25120.7680369803,-25120.276372806646,-25119.813391811902,-25119.377099975653,-25118.965458902614,-25118.576412825274,-25118.207914835653,-25117.857950848018,-25117.524560259935,-25117.205853156567,-25116.900024536215,-25116.60536608956,-25116.320275714515,-25116.04326461014,-25115.77296177841,-25115.50811604052,-25115.247595985584,-25114.990388380626,-25114.735595445272,-25114.482431183624,-25114.230216838638,-25113.978375553812,-25113.7264264205,-25113.4739781497,-25113.22072257926,-25112.96642813553,-25112.710933291783,-25112.454140039015,-25112.19600740999,-25111.93654512974,-25111.675807472187,-25111.413887379167,-25111.15091086565,-25110.88703171558,-25110.622426474096,-25110.35728975231,-25110.09182986746,-25109.826264835665,-25109.560818722937,-25109.295718348796,-25109.031190333066,-25108.767458477756,-25108.504741477635,-25108.243250951826,-25107.983189784332,-25107.724750757137,-25107.46811545585,-25107.21345342786,-20377.240544279208,-23069.892954205534,-26492.278315170995,-27668.239405867516,-27153.143160823296,-26674.59249083987,-27088.8632921406,-28262.139367730448,-29705.82508794812,-31035.63178652183,-32157.738175853043,-33144.57701214936,-34071.081333651266,-34950.78704315695,-35746.77384272777,-36423.622851214335,-36971.27124116981,-37402.39476166814,-37738.09005657794,-37993.859509319205,-38177.48269773282,-38291.992906145824,-38340.28046906767,-38327.799569229646,-38262.136404230056,-38151.69082870273,-38004.237103763066,-37826.38847190185,-37623.84568446149,-37401.815356335384,-37165.33873915517,-36919.2835832177,-36668.159061336286,-36415.9336147198,-36165.951795132525,-35920.98195110338,-35683.3152658971,-35454.85877852892,-35237.19182967721,-35031.58933736014,-34839.03707662897,-34660.25234169295,-34495.714311505006,-34345.698209374175,-34210.30578978939,-34089.489122919986,-33983.067825360355,-33890.74253961892,-33812.10728403282,-33746.66218434673,-33693.826940468476,-33652.95441525491,-33623.343530298065,-33604.250894639015,-33594.90114239526,-33594.49639775073,-33602.22533938263,-33617.27204305836,-33638.82438530605,-33666.08161291813,-33698.260812435285,-33734.60231278558,-33774.374276873365,-33816.87674403381,-33861.445226504205,-33907.453797239024,-33954.31756312038,-34001.49449776321,-34048.486718411776,-34094.841336803016,-34140.150971047886,-34184.053926300294,-34226.234002957266,-34266.41990073334,-34304.3842322481,-34339.94219703309,-34372.94997090823,-34403.30284374136,-34430.93311657177,-34455.80776628076,-34477.925901809336,-34497.316055448435,-34514.03336169038,-34528.15667161497,-34539.78564024975,-34549.037817059754,-34556.04576923688,-34560.954270536895,-34563.91758953753,-34565.096907523824,-34564.65788911379,-34562.76842165663,-34559.596534738805,-34555.308508710565,-34550.06717927051,-34544.030442332616,-34537.34995951383,-34530.17006059903,-34522.62683629908,-34514.84741282353,-35974.84501825495,-35212.65144486056,-34289.73460326094,-33973.19986774775,-34068.700782848624,-34146.34420396992,-34001.23164555959,-33683.77207817754,-33325.69644484713,-33015.92427471757,-32768.026488259165,-32560.05331868121,-32372.90319616927,-32202.307281982776,-32054.407738142905,-31934.428801095943,-31842.36285784444,-31774.066544675068,-31724.30579937697,-31689.377109279674,-31667.326799816725,-31657.253016679126,-31658.379144723673,-31669.539749703094,-31689.262746659697,-31716.00342087709,-31748.393846974926,-31785.322283938753,-31825.870581153715,-31869.22411913161,-31914.60042478134,-31961.24324384823,-32008.45218118965,-32055.61429745955,-32102.218747097337,-32147.847095956997,-32192.153808005944,-32234.847791012573,-32275.681318736104,-32314.44608943311,-32350.971662630196,-32385.123516536656,-32416.799590835588,-32445.926236195515,-32472.454907132163,-32496.360156088667,-32517.638868405887,-32536.310107366437,-32552.414924774068,-32566.015701210574,-32577.194858100615,-32586.05304846064,-32592.70703744041,-32597.287463625653,-32599.936556908768,-32600.805786341425,-32600.0533952017,-32597.841847781474,-32594.33530205385,-32589.69725744059,-32584.088486517707,-32577.665277794425,-32570.577953726555,-32562.969616894483,-32554.9751054623,-32546.7201697507,-32538.32088542708,-32529.883296032087,-32521.503249987614,-32513.266386195595,-32505.24823233555,-32497.514398635718,-32490.120862013537,-32483.114334836275,-32476.53270443676,-32470.405523573274,-32464.754533541203,-32459.59420868762,-32454.932317787887,-32450.77049989935,-32447.104850167438,-32443.926507954802,-32441.22223869025,-32438.975002550447,-32437.164505835826,-32435.767732689266,-32434.759454935505,-32434.11271705175,-32433.799292846543,-32433.7901109417,-32434.055647327546,-32434.56628437307,-32435.292636236733,-32436.20584069776,-32437.277817403592,-32438.481492735566,-32439.79099195754,-32441.18179983063,-32442.63089122702,-32444.11683340512,-35894.3104124003,-33867.498653245144,-31547.780811111108,-30757.100804924314],"type":"scatter","name":"source2_p_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1,0.1001,0.1002,0.1003,0.1004,0.1005,0.1006,0.1007,0.1008,0.1009,0.101,0.1011,0.1012,0.1013,0.1014,0.1015,0.1016,0.1017,0.1018,0.1019,0.102,0.1021,0.1022,0.1023,0.1024,0.1025,0.1026,0.1027,0.1028,0.1029,0.103,0.1031,0.1032,0.1033,0.1034,0.1035,0.1036,0.1037,0.1038,0.1039,0.104,0.1041,0.1042,0.1043,0.1044,0.1045,0.1046,0.1047,0.1048,0.1049,0.105,0.1051,0.1052,0.1053,0.1054,0.1055,0.1056,0.1057,0.1058,0.1059,0.106,0.1061,0.1062,0.1063,0.1064,0.1065,0.1066,0.1067,0.1068,0.1069,0.107,0.1071,0.1072,0.1073,0.1074,0.1075,0.1076,0.1077,0.1078,0.1079,0.108,0.1081,0.1082,0.1083,0.1084,0.1085,0.1086,0.1087,0.1088,0.1089,0.109,0.1091,0.1092,0.1093,0.1094,0.1095,0.1096,0.1097,0.1098,0.1099,0.11,0.1101,0.1102,0.1103,0.1104,0.1105,0.1106,0.1107,0.1108,0.1109,0.111,0.1111,0.1112,0.1113,0.1114,0.1115,0.1116,0.1117,0.1118,0.1119,0.112,0.1121,0.1122,0.1123,0.1124,0.1125,0.1126,0.1127,0.1128,0.1129,0.113,0.1131,0.1132,0.1133,0.1134,0.1135,0.1136,0.1137,0.1138,0.1139,0.114,0.1141,0.1142,0.1143,0.1144,0.1145,0.1146,0.1147,0.1148,0.1149,0.115,0.1151,0.1152,0.1153,0.1154,0.1155,0.1156,0.1157,0.1158,0.1159,0.116,0.1161,0.1162,0.1163,0.1164,0.1165,0.1166,0.1167,0.1168,0.1169,0.117,0.1171,0.1172,0.1173,0.1174,0.1175,0.1176,0.1177,0.1178,0.1179,0.118,0.1181,0.1182,0.1183,0.1184,0.1185,0.1186,0.1187,0.1188,0.1189,0.119,0.1191,0.1192,0.1193,0.1194,0.1195,0.1196,0.1197,0.1198,0.1199,0.12,0.1201,0.1202,0.1203,0.1204,0.1205,0.1206,0.1207,0.1208,0.1209,0.121,0.1211,0.1212,0.1213,0.1214,0.1215,0.1216,0.1217,0.1218,0.1219,0.122,0.1221,0.1222,0.1223,0.1224,0.1225,0.1226,0.1227,0.1228,0.1229,0.123,0.1231,0.1232,0.1233,0.1234,0.1235,0.1236,0.1237,0.1238,0.1239,0.124,0.1241,0.1242,0.1243,0.1244,0.1245,0.1246,0.1247,0.1248,0.1249,0.125,0.1251,0.1252,0.1253,0.1254,0.1255,0.1256,0.1257,0.1258,0.1259,0.126,0.1261,0.1262,0.1263,0.1264,0.1265,0.1266,0.1267,0.1268,0.1269,0.127,0.1271,0.1272,0.1273,0.1274,0.1275,0.1276,0.1277,0.1278,0.1279,0.128,0.1281,0.1282,0.1283,0.1284,0.1285,0.1286,0.1287,0.1288,0.1289,0.129,0.1291,0.1292,0.1293,0.1294,0.1295,0.1296,0.1297,0.1298,0.1299,0.13,0.1301,0.1302,0.1303,0.1304,0.1305,0.1306,0.1307,0.1308,0.1309,0.131,0.1311,0.1312,0.1313,0.1314,0.1315,0.1316,0.1317,0.1318,0.1319,0.132,0.1321,0.1322,0.1323,0.1324,0.1325,0.1326,0.1327,0.1328,0.1329,0.133,0.1331,0.1332,0.1333,0.1334,0.1335,0.1336,0.1337,0.1338,0.1339,0.134,0.1341,0.1342,0.1343,0.1344,0.1345,0.1346,0.1347,0.1348,0.1349,0.135,0.1351,0.1352,0.1353,0.1354,0.1355,0.1356,0.1357,0.1358,0.1359,0.136,0.1361,0.1362,0.1363,0.1364,0.1365,0.1366,0.1367,0.1368,0.1369,0.137,0.1371,0.1372,0.1373,0.1374,0.1375,0.1376,0.1377,0.1378,0.1379,0.138,0.1381,0.1382,0.1383,0.1384,0.1385,0.1386,0.1387,0.1388,0.1389,0.139,0.1391,0.1392,0.1393,0.1394,0.1395,0.1396,0.1397,0.1398,0.1399,0.14,0.1401,0.1402,0.1403,0.1404,0.1405,0.1406,0.1407,0.1408,0.1409,0.141,0.1411,0.1412,0.1413,0.1414,0.1415,0.1416,0.1417,0.1418,0.1419,0.142,0.1421,0.1422,0.1423,0.1424,0.1425,0.1426,0.1427,0.1428,0.1429,0.143,0.1431,0.1432,0.1433,0.1434,0.1435,0.1436,0.1437,0.1438,0.1439,0.144,0.1441,0.1442,0.1443,0.1444,0.1445,0.1446,0.1447,0.1448,0.1449,0.145,0.1451,0.1452,0.1453,0.1454,0.1455,0.1456,0.1457,0.1458,0.1459,0.146,0.1461,0.1462,0.1463,0.1464,0.1465,0.1466,0.1467,0.1468,0.1469,0.147,0.1471,0.1472,0.1473,0.1474,0.1475,0.1476,0.1477,0.1478,0.1479,0.148,0.1481,0.1482,0.1483,0.1484,0.1485,0.1486,0.1487,0.1488,0.1489,0.149,0.1491,0.1492,0.1493,0.1494,0.1495,0.1496,0.1497,0.1498,0.1499,0.15,0.1501,0.1502,0.1503,0.1504,0.1505,0.1506,0.1507,0.1508,0.1509,0.151,0.1511,0.1512,0.1513,0.1514,0.1515,0.1516,0.1517,0.1518,0.1519,0.152,0.1521,0.1522,0.1523,0.1524,0.1525,0.1526,0.1527,0.1528,0.1529,0.153,0.1531,0.1532,0.1533,0.1534,0.1535,0.1536,0.1537,0.1538,0.1539,0.154,0.1541,0.1542,0.1543,0.1544,0.1545,0.1546,0.1547,0.1548,0.1549,0.155,0.1551,0.1552,0.1553,0.1554,0.1555,0.1556,0.1557,0.1558,0.1559,0.156,0.1561,0.1562,0.1563,0.1564,0.1565,0.1566,0.1567,0.1568,0.1569,0.157,0.1571,0.1572,0.1573,0.1574,0.1575,0.1576,0.1577,0.1578,0.1579,0.158,0.1581,0.1582,0.1583,0.1584,0.1585,0.1586,0.1587,0.1588,0.1589,0.159,0.1591,0.1592,0.1593,0.1594,0.1595,0.1596,0.1597,0.1598,0.1599,0.16,0.1601,0.1602,0.1603,0.1604,0.1605,0.1606,0.1607,0.1608,0.1609,0.161,0.1611,0.1612,0.1613,0.1614,0.1615,0.1616,0.1617,0.1618,0.1619,0.162,0.1621,0.1622,0.1623,0.1624,0.1625,0.1626,0.1627,0.1628,0.1629,0.163,0.1631,0.1632,0.1633,0.1634,0.1635,0.1636,0.1637,0.1638,0.1639,0.164,0.1641,0.1642,0.1643,0.1644,0.1645,0.1646,0.1647,0.1648,0.1649,0.165,0.1651,0.1652,0.1653,0.1654,0.1655,0.1656,0.1657,0.1658,0.1659,0.166,0.1661,0.1662,0.1663,0.1664,0.1665,0.1666,0.1667,0.1668,0.1669,0.167,0.1671,0.1672,0.1673,0.1674,0.1675,0.1676,0.1677,0.1678,0.1679,0.168,0.1681,0.1682,0.1683,0.1684,0.1685,0.1686,0.1687,0.1688,0.1689,0.169,0.1691,0.1692,0.1693,0.1694,0.1695,0.1696,0.1697,0.1698,0.1699,0.17,0.1701,0.1702,0.1703,0.1704,0.1705,0.1706,0.1707,0.1708,0.1709,0.171,0.1711,0.1712,0.1713,0.1714,0.1715,0.1716,0.1717,0.1718,0.1719,0.172,0.1721,0.1722,0.1723,0.1724,0.1725,0.1726,0.1727,0.1728,0.1729,0.173,0.1731,0.1732,0.1733,0.1734,0.1735,0.1736,0.1737,0.1738,0.1739,0.174,0.1741,0.1742,0.1743,0.1744,0.1745,0.1746,0.1747,0.1748,0.1749,0.175,0.1751,0.1752,0.1753,0.1754,0.1755,0.1756,0.1757,0.1758,0.1759,0.176,0.1761,0.1762,0.1763,0.1764,0.1765,0.1766,0.1767,0.1768,0.1769,0.177,0.1771,0.1772,0.1773,0.1774,0.1775,0.1776,0.1777,0.1778,0.1779,0.178,0.1781,0.1782,0.1783,0.1784,0.1785,0.1786,0.1787,0.1788,0.1789,0.179,0.1791,0.1792,0.1793,0.1794,0.1795,0.1796,0.1797,0.1798,0.1799,0.18,0.1801,0.1802,0.1803,0.1804,0.1805,0.1806,0.1807,0.1808,0.1809,0.181,0.1811,0.1812,0.1813,0.1814,0.1815,0.1816,0.1817,0.1818,0.1819,0.182,0.1821,0.1822,0.1823,0.1824,0.1825,0.1826,0.1827,0.1828,0.1829,0.183,0.1831,0.1832,0.1833,0.1834,0.1835,0.1836,0.1837,0.1838,0.1839,0.184,0.1841,0.1842,0.1843,0.1844,0.1845,0.1846,0.1847,0.1848,0.1849,0.185,0.1851,0.1852,0.1853,0.1854,0.1855,0.1856,0.1857,0.1858,0.1859,0.186,0.1861,0.1862,0.1863,0.1864,0.1865,0.1866,0.1867,0.1868,0.1869,0.187,0.1871,0.1872,0.1873,0.1874,0.1875,0.1876,0.1877,0.1878,0.1879,0.188,0.1881,0.1882,0.1883,0.1884,0.1885,0.1886,0.1887,0.1888,0.1889,0.189,0.1891,0.1892,0.1893,0.1894,0.1895,0.1896,0.1897,0.1898,0.1899,0.19,0.1901,0.1902,0.1903,0.1904,0.1905,0.1906,0.1907,0.1908,0.1909,0.191,0.1911,0.1912,0.1913,0.1914,0.1915,0.1916,0.1917,0.1918,0.1919,0.192,0.1921,0.1922,0.1923,0.1924,0.1925,0.1926,0.1927,0.1928,0.1929,0.193,0.1931,0.1932,0.1933,0.1934,0.1935,0.1936,0.1937,0.1938,0.1939,0.194,0.1941,0.1942,0.1943,0.1944,0.1945,0.1946,0.1947,0.1948,0.1949,0.195,0.1951,0.1952,0.1953,0.1954,0.1955,0.1956,0.1957,0.1958,0.1959,0.196,0.1961,0.1962,0.1963,0.1964,0.1965,0.1966,0.1967,0.1968,0.1969,0.197,0.1971,0.1972,0.1973,0.1974,0.1975,0.1976,0.1977,0.1978,0.1979,0.198,0.1981,0.1982,0.1983,0.1984,0.1985,0.1986,0.1987,0.1988,0.1989,0.199,0.1991,0.1992,0.1993,0.1994,0.1995,0.1996,0.1997,0.1998,0.1999,0.2]},{"mode":"lines","y":[0.0,0.0,0.0,2.6619422065608593e-12,5.063645377902115e-9,2.1414919793416375e-7,2.7323911576943433e-6,1.832913768857619e-5,8.213495234765829e-5,0.00027883932340484914,0.0007734073314687459,0.00184305761675015,0.003909153338266516,0.007569942201014489,0.013631911743045708,0.023133237381498105,0.03735089809083825,0.05778726537755939,0.08613985638132043,0.12426095534557002,0.17411295473014848,0.23772174340885094,0.31712804053606136,0.414338336392882,0.5312787064843566,0.6697557913067713,0.8314283619830549,1.0177901805641645,1.230163158367555,1.469698359592683,1.7373821874183089,2.034045774231486,2.360376160049656,2.7169286159821704,3.1041395103011675,3.5223388454410625,3.9717614476691536,4.452555896720899,4.964790977059978,5.508460127622606,6.083484733573155,6.689717011047185,7.326942812538421,7.994884391421147,8.693203141970791,9.421502517017286,10.179331481747422,10.966188791103988,11.78152814022496,12.624763987147364,13.495277742412977,14.392424080336287,15.315537243186675,16.26393727673852,17.236936113369957,18.23384335744569,19.253971600011308,20.29664112546042,21.3611839486279,22.44694718704184,23.55329580086421,24.679614729417914,25.82531044495559,26.989811952282682,28.17257128598262,29.37306357973231,30.59078679054482,31.825261153053408,33.07602842439282,34.34265096911441,35.624710729545285,36.92180988427716,38.23357549109366,39.55965064298823,40.89968269055012,42.253325489254465,43.620244442083845,45.000118534453556,46.392638787719164,47.79750539212007,49.21442579241916,50.64311432902433,52.083292672337535,53.53469013060941,54.99704353195,56.47009687469611,57.953601042374174,59.447313691342096,60.950999219762124,62.464428687321956,63.987379620697936,65.51963572707781,67.06098657411063,68.61122727765559,70.17015820853926,71.73758470944796,73.31331681359043,74.897168966851,76.4889597627812,78.08851170113564,79.69565097610375,81.31020729507028,82.93201372536215,84.56090656551152,86.19672523830863,87.83931220367664,89.48851288962656,91.1441756391893,92.80615167065268,94.4742950481444,96.14846265971525,97.82851420050359,99.5143121590827,101.20572180550133,102.90261117981663,104.60485108008513,106.31231504895543,108.0248793582197,109.74242299091303,111.46482762076909,113.1919775890039,114.92375987849263,116.66006408546912,118.40078238890949,120.14580951780539,121.8950427165554,123.64838170871496,125.40572865934604,127.16698813618297,128.9320670697997,130.70087471294562,132.47332259919557,134.24932450102153,136.02879638739898,137.8116563810214,139.59782471519668,141.38722369046457,143.1797776309983,144.97541284080205,146.77405755974928,148.57564191950166,150.38009789932192,152.18735928182494,153.99736160870083,155.8100421364362,157.62533979208683,159.44319512913177,161.2635502834469,163.0863489294398,164.91153623640787,166.73905882512418,168.56886472472382,170.400903329903,172.23512535848096,174.0714828093424,175.90992892079686,177.75041812938375,179.5929060291256,181.43734933127163,183.2837058245202,185.1319343357562,186.98199469128545,188.8338476785993,190.68745500865572,192.54277927867554,194.39978393547773,196.25843323932753,198.11869222831325,199.9805266832431,201.8439030930648,203.7087886208035,205.57515107001876,207.44295885178116,209.31218095216454,211.1827869002604,213.05474673670454,214.9280309827368,216.80261060977898,218.6784570095436,220.55554196467207,222.43383761991302,224.3133164538488,226.19395125116603,228.0757150754806,229.95858124273593,231.8425232951824,233.72751497591562,235.61353020402368,237.50054305034283,239.38852771380445,241.27745849841978,243.16730979089422,245.05805603888177,246.9496717299032,248.84213137093263,250.7354094686466,252.62948051037597,254.5243189457566,256.41989916909904,258.3161955024647,260.2131821794894,262.1108333299596,264.00912296512615,265.90802496380013,267.80751305921603,269.70756082668476,271.60814167204353,273.5092288209075,275.41079530873014,277.31281397168493,279.215257438379,281.11809812238863,283.0213082156423,284.92485968263895,286.82872425551903,288.7328734299856,290.6372784620749,292.54191036578413,294.4467399115591,296.3517376256333,298.25687379022276,300.16211844458087,302.067441386902,303.97281217707337,305.87820014027045,307.7835743713952,309.6889037403492,311.5941568981311,313.49930228375695,315.40430813198145,317.3091424818357,319.21377318594546,321.1181679206294,323.022294196771,324.92611937143823,326.8296106602432,328.73273515043206,330.6354598146805,332.5377515255835,334.43957707081745,336.340903168963,338.24169648596785,340.14192365221646,342.04155128021114,343.94054598281605,345.83887439207376,347.73650317852474,349.63339907105427,351.5295288772163,353.424859504023,355.31935797914304,357.21299147253455,359.10572731843746,360.9975330377279,362.8883763605859,364.77822524947203,366.66704792235197,368.5548128761782,370.44148891056886,372.32704515166336,374.2114510761249,376.09467653525877,377.9766917792208,379.8574674812697,381.73697476204956,383.6151852138558,385.4920709248672,387.3676045032994,389.24175910144606,391.11450843959165,392.9858268297503,394.8556891992003,396.72407111378845,398.59094880096217,400.45629917250335,402.3200998469471,404.18232917160844,406.0429662442365,407.9019909342578,409.75938390353633,411.61512662666894,413.4692014107412,415.32159141457885,417.1722806674022,419.0212540868624,420.8684974964815,422.71399764243523,424.5577422095992,426.39971983696614,428.23992013224665,430.0783336857915,431.91495208364796,433.749767919874,435.58277480799416,437.4139673915903,439.24334135405576,441.07089342743177,442.89662140035387,444.72052412503047,446.54260152332444,448.36285459184194,450.18128540604505,451.9978971233899,453.8126939854519,455.625681319038,457.43686553626287,459.2462541336078,461.0538556899048,462.859679863317,464.66373738721376,466.4660400650099,468.26660076393756,470.06543340772964,471.86255296826846,473.6579754561262,475.4517179100641,477.24379838545497,479.0342359416446,480.8230506282515,482.61026347042065,484.39589645303215,486.17997250387384,487.9625154757851,489.74355012778443,491.5231021052201,493.30119791888956,495.07786492321077,496.8531312934457,498.62702600196667,500.39957879358246,502.1708201600111,503.94078131345367,505.709494159259,507.47699126781635,509.24330584559027,511.00847170542704,512.7725232360124,514.5354953706493,516.2974235553597,518.0583437162237,519.8182922261411,521.5773058709183,523.3354218148652,525.0926775657183,526.8491109391282,528.604760022668,530.3596631393118,532.1138588105845,533.8673857192844,535.6202826718992,537.3725885606962,539.1243423255208,540.8755829154453,542.6263492501528,544.3766801812191,546.1266144532592,547.8761906650243,549.6254472304583,551.3744223398048,553.1231539206963,554.8716795993956,556.6200366621604,558.3682620167796,560.1163921543259,561.864463111186,563.6125104313702,565.360569129175,567.1086736522302,568.8568578449687,570.6051549125516,572.3535973853232,574.1022170837819,575.8510450841447,577.6001116845379,579.3494463718375,581.0990777892596,582.8490337046162,584.5993409793514,586.3500255384548,588.1011123411606,589.8526253525329,591.6045875160036,593.357020726804,595.1099458063893,596.8633824779001,598.6173493426528,600.3718638577086,602.1269423145192,603.88259981867,605.638850270864,607.3957063490122,609.1531794915959,610.9112798821656,612.6700164350897,614.4293967826644,616.189427263478,617.9501129120088,619.7114574496169,621.4734632768074,623.2361314669398,624.9994617612253,626.7634525651151,628.5280937711658,630.2928905988863,632.0567036061874,633.8231115895547,635.6000106559211,637.3870602407067,639.1657540206486,640.9008921362312,642.5512598418086,644.0783096350598,645.4475486110902,646.624287876512,647.5697010104855,648.2415911255374,648.598797094843,648.6065389988241,648.2397804533703,647.4838979471064,646.333697044248,644.7918542133178,642.8676363681678,640.5759000753197,637.936128335487,634.9713478181534,631.7068715290916,628.1690284040482,624.3840239737309,620.3770474802421,616.1716773035572,611.7895500807467,607.250236654209,602.571249305819,597.768127834513,592.8545827309385,587.8426870442463,582.743111701386,577.5653842060475,572.3181418071204,567.0093534019203,561.6464965931763,556.2366914816856,550.7867995859075,545.3034953374732,539.7933129284122,534.2626682670635,528.7178573239901,523.1650356961712,517.6101868383722,512.0590861542742,506.5172656302136,500.98998108061886,495.48218272822373,489.99848990592716,484.5431711665289,479.12013108241115,473.73290433011186,468.38465670509555,463.07819207185463,457.815964130292,452.6000920796821,447.43237946620116,442.31433552599077,437.2471982342915,432.23195820020896,427.26938261107915,422.3600386122943,417.50431571200113,412.70244694530874,407.9545286048592,403.26053838332155,398.62035182161026,394.0337570282009,389.5004677126901,385.0201346360477,380.5923556101753,376.2166841861494,371.8926371685833,367.61970109168726,363.39733779129045,359.2249892029586,355.102081506227,351.0280287195599,347.0022358341152,343.0241015587504,339.0930207362831,335.20838647957606,331.3695920662803,327.5760326222629,323.827106615651,320.1222171775357,316.4607732614467,312.842190650474,309.2658928198909,305.73131166182117,302.23788807865884,131.6239277352696,3250.1322787495574,5941.4684178506595,7230.268778038242,7342.25982016381,7168.17747307404,7357.256163437312,7958.4816586934485,8688.163331208458,9287.57107788878,9708.00899502727,10038.6268719354,10370.023506818341,10731.16275244269,11091.450056397598,11409.993093943216,11668.941595699245,11875.343092944899,12046.346479424135,12192.830921702294,12315.61370569093,12409.465648644606,12469.251484094422,12493.863325696759,12485.777845852846,12449.084189940902,12387.556794636643,12303.8894975905,12200.127532026321,12078.357452572609,11941.229833803178,11791.996515007038,11634.193088542464,11471.271425964713,11306.335747482844,11142.070746981604,10980.7857813797,10824.477833212344,10674.860534484664,10533.344604708996,10401.007314196408,10278.58368295417,10166.493222289544,10064.89486497729,9973.747858550294,9892.863211530666,9821.940543134537,9760.59491249437,9708.380867543521,9664.816368914973,9629.404891344951,9601.651851824376,9581.073244858137,9567.197585805035,9559.56409904431,9557.720044185093,9561.218537925297,9569.616839268143,9582.474651220633,9599.352365852383,9619.809755267843,9643.40575771849,9669.699693391729,9698.25376938868,9728.636422443338,9760.426035640252,9793.214726651915,9826.612063202745,9860.248614046304,9893.779208715427,9926.885741586408,9959.27937267869,9990.702049559506,10020.927359416877,10049.76077600583,10077.0393810023,10102.631128499423,10126.433708855626,10148.373067304812,10168.40164164175,10186.49639170951,10202.656692940973,10216.902155985763,10229.27041974216,10239.814952261399,10248.60288628577,10255.71291285221,10261.233254543475,10265.259737169792,10267.893974367278,10269.241674866418,10269.411078397814,10268.511523967929,10266.65215312166,10263.940749879548,10260.482717611007,10256.380191133496,10251.731280265605,10246.629439422011,10241.162956817861,10235.004464237503,10230.876872207375,10226.578431574406,10221.256722699993,10215.116785159647,10208.799135265608,10202.795930772214,10197.243188889062,10192.039109038884,10187.034240100946,10182.166985599868,10177.464459347426,10172.978474396657,10168.736651984535,10164.72935274081,10160.932622834274,10157.330951245058,10153.924469472033,10150.72329174623,10147.736207900363,10144.96513215871,10142.405416047923,10140.049117485838,10137.887972253664,10135.913769724631,10134.117428145566,10132.487909320535,10131.011937942098,10129.67467266011,10128.460652609312,10127.354633752799,10126.342039935438,10125.409101556785,10124.542892547086,10123.731368737232,10122.963466360025,10122.22920806256,10121.519760225312,10120.827421459491,10120.145547728393,10119.46845260238,10118.791311777957,10118.110085934419,10117.421460805917,10116.722793341622,10116.012057500167,10115.287788549544,10114.549029605674,10113.795284474425,10113.026477169358,10112.242915727136,10111.445256765954,10110.634468574284,10109.811792665063,10108.978704954136,10108.136877841744,10107.288143705831,10106.434459659025,10105.577873328637,10104.720489757692,10103.864439952184,10103.011851731779,10102.164823365027,10101.325400166113,10100.495553997669,10099.677165560772,10098.872009392366,10098.081741532966,10097.307889809563,10096.551846607583,10095.814863935098,10095.09805054904,10094.40237092418,10093.7286458751,10093.077554664216,10092.44963843309,10091.845304787168,10091.264833359035,10090.708382180528,10090.175994709774,10089.667607376185,10089.1830575211,10088.722091619351,10088.284373672195,10087.869493667276,10087.476976009057,10087.106287834391,10086.756847138005,10086.428030644132,10086.119181368116,10085.829615819885,10085.558630808468,10085.305509814712,10085.069528907485,10084.849962186596,10084.646086742658,10084.457187129872,10084.282559353218,10084.121514375922,10083.973381157823,10132.53433762742,9902.417064038546,9678.922628759457,9573.495671220953,9565.467306129105,9584.336545982453,9577.864527642192,9534.208811182885,9468.270108977322,9400.55356426903,9341.583410832047,9291.181406402824,9245.322810337315,9201.850786301413,9162.427678350488,9130.206414893033,9107.191699763549,9093.241990153772,9086.618226786872,9085.329918752905,9087.904901785674,9093.463325852445,9101.395625774147,9111.033493316632,9121.61503464456,9132.416430849977,9142.913780644452,9152.852636435317,9162.197776525352,9171.039833303772,9179.507606321004,9187.72205991009,9195.786552519185,9203.789012975258,9211.803269582586,9219.881981795052,9228.046951243497,9236.283747465652,9244.543473854888,9252.751424614236,9260.818267319733,9268.650193550575,9276.156202796068,9283.252515121156,9289.865428367735,9295.933505861036,9301.40937326032,9306.260804844378,9310.470687461335,9314.03586826888,9316.96523883162,9319.277578359395,9320.999540692417,9322.163905838985,9322.808044987221,9322.972507339768,9322.699712862284,9322.032809632994,9321.014764002108,9319.687705033928,9318.092481612422,9316.268360658174,9314.252805406228,9312.081303649398,9309.787241574362,9307.401825002824,9304.954042414223,9302.470655602847,9299.976203099548,9297.493008557438,9295.041194779938,9292.6387084521,9290.301359625595,9288.042876311254,9285.874971713454,9283.80742131894,9281.848148528677,9280.003319003468,9278.277444142252,9276.673493155702,9275.193011908232,9273.836245961584,9272.602265348094,9271.489089218476,9270.493809129863,9269.612710036772,9268.84138803171,9268.174863777516,9267.60769059402,9267.134056365921,9266.747878738808,9266.442893329859,9266.212734823397,9266.05101086555,9265.951368689715,9265.907554456591,9265.913465392983,9265.963194927168,9266.051071106898,9266.17168862713,9265.874292582941,9268.412067065223,9270.869996694237,9272.144827155944,9272.444225451887,9272.481158415923,9272.783063565961,9273.46220079971,9274.358413392129,9275.259529849227,9276.060419357447,9276.766485961547,9277.420158992467,9278.042299269458,9278.614308410193,9279.103539207888,9279.490176502011,9279.776501221237,9279.980580650797,9280.122458644457,9280.216647377718,9280.271680536449,9280.293519384402,9280.288892166589,9280.265609119533,9280.231217845361,9280.191372632566,9280.149138359691,9280.105491501574,9280.060240992048,9280.012882414398,9279.963027685237,9279.9104700722,9279.855132998202,9279.797033633828,9279.736338260436,9279.673450744445,9279.609062916521,9279.544136989573,9279.479821332548,9279.41734273949,9279.357910436916,9279.302649720357,9279.252564671044,9279.20851613347,9279.171205627945,9279.141162239948,9279.11873576371,9279.104100421315,9279.097269269334,9279.098115868772,9279.106398043888,9279.121779943347,9279.14385128776,9279.17214441437,9279.20615010836,9279.245332403607,9279.289141749809,9279.337025832838,9279.388437807247,9279.442842341508,9279.49972017866,9279.558571806567,9279.618920517763,9279.680314882477,9279.742330603345,9279.804571802855,9279.866671886817,9279.928294136487,9279.989132112303,9280.048909867908,9280.107381930773,9280.164333017881,9280.219577493928,9280.272958607671,9280.324347545431,9280.373642324688,9280.420766535814,9280.465667936885,9280.508316916512,9280.54870485187,9280.586842396378,9280.622757729556,9280.656494794957,9280.688111545118,9280.717678208803,9280.745275595837,9280.770993454917,9280.794928899184,9280.817184911466,9280.837868937724,9280.857091573735,9280.874965348336,9280.89160360573,9280.907119488233,9280.921625020426,9280.935230294019,9280.94804275116,9280.960166563062,9280.971702099832,9232.574194734503,9491.352605675052,9741.934226781363,9859.2175612356,9862.90527154349,9835.835419478673,9841.439665338246,9892.899657283087,9970.086617567998,10047.331881441649,10112.869512170635,10168.472026133604,10219.765422316514,10269.129896981605,10314.052324712557,10350.605576234862,10376.670015382311,10392.789573070295,10401.20702457879,10404.181752575845,10403.202704668078,10399.04249512057,10392.207958937337,10383.325937396978,10373.1530042787,10362.408292137146,10351.591316376931,10340.918288941699,10330.393853218913,10319.920722232691,10309.39396214444,10298.742292384855,10287.928009357882,10276.935925885,10265.765876716292,10254.43693315689,10242.99553329734,10231.519280842993,10220.113304754135,10208.899692125842,10198.005198944109,10187.551080138286,10177.646741225204,10168.386775152216,10159.8495669837,10152.096367122698,10145.170532787259,10139.097357547515,10133.884968713986,10129.526253356224,10126.00139159141,10123.280420419786,10121.325455814476,10120.09251545271,10119.533061952923,10119.595404021544,10120.225984618251,10121.37049648102,10122.974759481633,10124.985350137882,10127.350040542882,10130.018127518186,10132.940712132555,10136.07095142754,10139.364276922737,10142.778571010986,10146.274302746428,10149.81463418666,10153.365508628156,10156.895724335063,10160.37698932005,10163.78394983321,10167.094187972927,10170.288188761862,10173.349280349732,10176.263551384938,10179.019748111832,10181.60915239227,10184.025441806405,10186.264534090766,10188.324419452918,10190.204984891237,10191.90783433053,10193.436107546258,10194.794300046991,10195.988085651994,10197.024143407516,10197.909990493195,10198.653822650947,10199.264363362927,10199.75072260974,10200.12226570034,10200.388492462458,10200.558926995483,10200.6430181342,10200.650050679571,10200.589067308501,10200.46880091574,10200.297617016422,10200.083465769967,10226.156279419318,10100.348854204762,9977.367638874835,9919.479349396537,9916.28602051163,9928.110066912308,9925.005346222617,9900.331113151127,9862.970845970567,9824.840302512473,9791.834123447727,9763.615105239198,9737.777266880177,9713.146087561046,9690.77160876714,9672.482617739875,9659.37438943878,9651.301607386311,9647.246262716886,9646.09688422426,9647.06205615976,9649.686432589366,9653.656111458986,9658.614313664402,9664.149196862923,9669.871521124654,9675.504607483917,9680.919522756525,9686.103483539318,9691.106617257861,9695.993843745282,9700.82120042752,9705.632328875827,9710.46103289048,9715.332882104714,9720.261827376113,9725.245355119696,9730.262080076493,9735.273260003174,9740.22797997596,9745.069478879555,9749.740672185966,9754.187933444558,9758.363230517698,9762.225414948685,9765.741168197117,9768.885751730244,9771.643373439607,9774.00694888888,9775.977280172014,9777.56186230347,9778.773609288944,9779.629701589856,9780.150606030436,9780.35922795587,9780.280139178278,9779.938871397932,9779.361305720322,9778.573192048774,9777.599805270014,9776.465711461073,9775.194603523545,9773.80917372449,9772.33100860229,9770.780505564635,9769.176813236005,9767.537793235268,9765.879996539406,9764.218647470014,9762.567632242459,9760.939493350183,9759.345432993307,9757.7953278231,9756.297755056736,9754.860028423052,9753.488242202973,9752.187322381504,9750.961084655286,9749.812299113899,9748.742760880903,9747.753365335966,9746.84418619632,9746.014554844749,9745.2631396724,9744.588024577683,9743.986785957395,9743.456567550033,9742.99415246414,9742.596031761686,9742.258469096683,9741.977561088599,9741.74929326132,9741.569591465504,9741.434368731949,9741.339567517241,9741.281197336779,9741.255367838527,9741.258317438644,9741.286437691422,9741.3362935921,9627.643653797513,10204.987923305374,10777.244705118741,11039.81924780121,11023.73425255962,10932.75319774694,10930.62374064273,11050.315837951577,11236.600814315838,11421.705632357895,11576.852628480297,11709.240135358326,11834.462403608086,11957.847770117689,12071.599322881026,12165.115963926315,12233.38705970236,12278.453110562115,12306.23977762988,12322.077809657443,12329.004099857044,12328.247411071552,12320.538704599412,12307.062651826098,12289.381093525417,12268.9698787899,12246.783632484752,12223.153009361931,12198.006971529483,12171.149999953253,12142.473226122807,12112.020291944325,12079.95436248981,12046.507796609394,12011.948334924542,11976.579150128155,11940.749709242862,11904.856956698257,11869.330290831176,11834.602782982682,11801.082147599005,11769.129965907567,11739.051928617351,11711.096577823228,11685.457200631623,11662.274042847586,11641.63623260118,11623.584599070204,11608.11651197257,11595.192474995376,11584.74335926247,11576.67694590261,11570.883106895184,11567.237747823267,11565.605989954576,11565.84501257489,11567.806649899003,11571.339621844858,11576.291286192298,11582.508941522017,11589.840849769776,11598.137164694057,11607.250873606006,11617.038759813546,11627.362337002127,11638.088711369845,11649.091358250747,11660.250822302894,11671.455348696678,11682.601436151883,11693.59428990517,11704.34815333723,11714.78650889807,11724.842152694047,11734.457154948002,11743.582719276708,11752.178950790949,11760.214540817342,11767.666376557776,11774.51908641052,11780.764533914587,11786.401273797937,11791.433982285042,11795.872871591926,11799.733096542512,11803.034160092167,11805.799324142306,11808.05503185456,11809.83034725018,11811.156417063636,11812.065958799109,11812.5927780208,11812.771317272578,11812.6362386314,11812.22204156544,11811.562717330004,11810.691440558036,11809.640298083617,11808.440054504983,11807.119953610421,11979.279561188763,11289.215811751794,10600.36622220225,10275.813525184503,10253.653486678686,10316.556995968112,10303.14912767513,10174.005725206121,9971.640813982713,9758.294710242044,9568.367848383283,9404.140587051737,9254.751737734601,9113.505921323745,8984.99266678676,8878.919688474474,8802.025165150682,8754.513287247515,8731.340570235152,8726.294265240322,8734.538804377773,8753.011057447313,8779.504939126022,8811.665569187722,8846.874712943694,8882.66543237744,8917.237695527734,8949.684905772827,8979.840063255555,9007.97988623784,9034.541957842297,9059.969322241805,9084.667066573935,9108.996813450582,9133.270736610782,9157.722465435327,9182.473177586195,9207.513913545974,9232.712089744435,9257.840656302564,9282.615701370798,9306.731115158109,9329.88435798595,9351.793104904762,9372.206641414235,9390.914573584956,9407.75364602915,9422.611716050922,9435.427736539763,9446.187921918068,9454.919319247176,9461.682500168823,9466.564635485625,9469.6733883835,9471.131523848968,9471.071991808789,9469.633454874074,9466.956441382694,9463.180321001939,9458.44115160132,9452.87025203804,9446.593264747182,9439.729505171097,9432.391491217651,9424.684627016608,9416.70703732427,9408.549529398126,9400.295635615585,9392.02168952918,9383.796910649131,9375.683499851864,9367.736761262015,9360.005263742487,9352.531043938248,9345.349844073437,9338.491376459267,9331.979610950522,9325.833085983553,9320.065244554904,9314.684793586513,9309.696081196033,9305.09948413254,9300.891797939781,9297.066624331166,9293.614752216816,9290.524529792663,9287.782225096904,9285.372372130405,9283.278099685667,9281.481440626754,9279.96362024658,9278.705323089427,9277.686938028575,9276.88878148295,9276.291298664151,9275.875242862043,9275.621833046755,9275.512890389402,9275.530954553746,9275.65938072584,9248.582281757013,9395.282470504557,9536.845712233096,9603.563901976057,9607.351504235234,9594.060355711117,9598.357455004538,9627.564624380097,9670.93764394917,9714.552554372738,9751.848451941292,9783.591706385676,9812.76840575576,9840.723690108167,9866.158610840548,9886.940106546743,9901.860990511017,9911.161452162392,9916.046796748898,9917.775140243766,9917.204521854334,9914.800355821517,9910.872971934645,9905.791580047175,9899.995564825591,9893.904521619179,9887.816084818218,9881.867157331806,9876.070810992,9870.376462647235,9864.723182973843,9859.064496472243,9853.37010248254,9847.62099156796,9841.806127424594,9835.925355138404,9829.994427883863,9824.047589292135,9818.1359235343,9812.321686012732,9806.671456311851,9801.250293013021,9796.117912751948,9791.326732957517,9786.920820250685,9782.935142371356,9779.394950067517,9776.315513620964,9773.702484030226,9771.552867720324,9769.856386299964,9768.596897040461,9767.753653944055,9767.302362509068,9767.216084365848,9767.466063251492,9768.022487001746,9768.855150675188,9769.933981315817,9771.229415611013,9772.71266033311,9774.355880877547,9776.132353386129,9778.016595053527,9779.98447152461,9782.013277697562,9784.081793785776,9786.170323985421,9788.260725338412,9790.33642995269,9792.382458880335,9794.38542389966,9796.33351468088,9798.216471445548,9800.025545063872,9801.753446727218,9803.394288440744,9804.943514744842,9806.397826003435,9807.755094219407,9809.014273097255,9810.17530445012,9811.239022906326,9812.20706041975,9813.081751654749,9813.866041090563,9814.563392658289,9815.177702745368,9815.71321734997,9816.174454004024,9816.56612886958,9816.893089228899,9817.160251481664,9817.372544721013,9817.53485993481,9817.652004834119,9817.728664235656,9817.76936584352,9817.778451212533,9817.760051643847,9884.186771504854,9552.695930327209,9234.929189950119,9084.274948275666,9067.657681972581,9088.0992334991,9075.860652928508,9014.908989199741,8924.248878275335,8830.76448979626,8748.828287686123,8678.796522086484,8615.501443889418,8555.825967620121,8501.680806263792,8457.21496795638,8425.280621547226,8405.90018950881,8396.839279105567,8395.381685603661,8399.438959603405,8407.72479506215,8419.350441609458,8433.368520829214,8448.70070558343,8464.306510479548,8479.408858488216,8493.602735299542,8506.798476276223,8519.098020882175,8530.673090187083,8541.697306285565,8552.32844901035,8562.708805565366,8572.965448099403,8583.199412958626,8593.47084108425,8603.789658991995,8614.116091894013,8624.371102903617,8634.45108007851,8644.241776613191,8653.628735636743,8662.50391553657,8670.770159992777,8678.344726935607,8685.162293219206,8691.177014640707,8696.3630310908,8700.71334883311,8704.237533737558,8706.958928001655,8708.91195512059,8710.139723617347,8710.691887940087,8710.622650773661,8709.988886679399,8708.848475124993,8707.258952182532,8705.276527889002,8702.955423620875,8700.347433129757,8697.501618735965,8694.464095153122,8691.27789054364,8687.982884779078,8684.61581468385,8681.210323073712,8677.797026389453,8674.403585727254,8671.05477913617,8667.772580678104,8664.576251404143,8661.48244268389,8658.505308521953,8655.656623228708,8652.945903220196,8650.380534168737,8647.96590529384,8645.705551245928,8643.60130008883,8641.653424680224,8639.860794753235,8638.221027775044,8636.730637438151,8635.385178950219,8634.179390152267,8633.107327250686,8632.162493905224,8631.337962654787,8630.626488046797,8630.020611169302,8629.512755461254,8629.095313721886,8628.760726255216,8628.501550152756,8628.310519849965,8628.18059924195,8628.105025756071,8628.077346827034,8508.374028600136,9210.882943074539,9890.615043166938,10205.80927731029,10201.262582626348,10111.38628839362,10120.25558389997,10263.505775432997,10478.548105472757,10689.6202483147,10865.060261105502,11013.542186340957,11153.077209324174,11290.061518797736,11416.043794727197,11519.264153888911,11594.099859792372,11642.793866249764,11671.98417162221,11687.700097677964,11693.414541289618,11690.546147088151,11679.919523762397,11662.858984647723,11641.137513097177,11616.468630812471,11590.006219908048,11562.204521722271,11533.047600380913,11502.353648127653,11470.014376339628,11436.07535642721,11400.70343583405,11364.131841826857,11326.622821928002,11288.468927311513,11250.008303501521,11211.630233876798,11173.762419973891,11136.841724538644,11101.283058627821,11067.456324290157,11035.67506095569,11006.194465472061,10979.212977627842,10954.874174901885,10933.268206635716,10914.434096787514,10898.364314267534,10885.011476854635,10874.29602396402,10866.113349333227,10860.33955449368,10856.835879829707,10855.452308992946,10856.030814373771,10858.408352699156,10862.419461106188,10867.89829751966,10874.680128739783,10882.60244187766,10891.505893629475,10901.235233581712,10911.640225273273,10922.576519803364,10933.906437537225,10945.499648121513,10957.233766358204,10968.994880903949,10980.678012619406,10992.18748164591,11003.437159812493,11014.350596725342,11024.861023193595,11034.911244850828,11044.453439680166,11053.448869378455,11061.867511470198,11069.687619493796,11076.89522158593,11083.483570808814,11089.452561545431,11094.808124939387,11099.561613805789,11103.729185113094,11107.331186818532,11110.39155540788,11112.937230309974,11114.997590851917,11116.603920439393,11117.788901444415,11118.586143230907,11119.029745059051,11119.153895223611,11118.992507476418,11118.578895371467,11117.945484613962,11117.123562903978,11116.14306627954,11115.032400648644,11291.549118830364,10529.682140394383,9786.855197067296,9434.17508927149,9391.69404555819,9436.666243675498,9411.211452257823,9275.997878255492,9069.472215579359,8850.752204133176,8654.551564867286,8485.083867797159,8332.371912870323,8189.0337767533465,8058.5229503615055,7950.148522261764,7871.091589525421,7822.238117005451,7798.9154578442185,7794.826449958849,7804.823400761613,7825.546729977471,7854.6088370377265,7889.5470596290625,7927.620227938092,7966.187912764595,8003.245432140708,8037.698796215582,8069.2496191121645,8098.109260279378,8124.706600276322,8149.514880057646,8172.994548634161,8195.580762272035,8217.67389356996,8239.60734079717,8261.608078191282,8283.771716111765,8306.061875572945,8328.334062715263,8350.370856092604,8371.916488968545,8392.703922671953,8412.473232228642,8430.984721705756,8448.029413496744,8463.437850786968,8477.086285951083,8488.898876960116,8498.845743163845,8506.937902803495,8513.220791525382,8517.767759020868,8520.67413308572,8522.051837732884,8522.024347969282,8520.72195731724,8518.277573921809,8514.823310446405,8510.487989277208,8505.395464598734,8499.663534783065,8493.40322751946,8486.718331461943,8479.705136144188,8472.452370075733,8465.041306127792,8457.545974008852,8450.033414788666,8442.563936095201,8435.191358608685,8427.963264408256,8420.921258788445,8414.101247180162,8407.533720234878,8401.24403954105,8395.252722185945,8389.575728359905,8384.224757793994,8379.2075577989,8374.528241024913,8370.187608022374,8366.183469455227,8362.5109644369,8359.162873141151,8356.129922446376,8353.401082932642,8350.963854849877,8348.804540447709,8346.908500504593,8345.26039370289,8343.84439821025,8342.644415191873,8341.644254048788,8340.827799181208,8340.179158204597,8339.682791831914,8339.323625970623,8339.087146837286,8338.959480004387,8338.493194927747,8341.177670436049,8343.758860782573,8345.11277174364,8345.500894854746,8345.653990155613,8346.088236281092,8346.908913393181,8347.958940247301,8349.028774606668,8350.009714595026,8350.900762343468,8351.740656575128,8352.550239014117,8353.31195196116,8353.992552460375,8354.569808618826,8355.043289021178,8355.429432814479,8355.748050936598,8356.014158597112,8356.236839213854,8356.422381953558,8356.577759205333,8356.71120035571,8356.830931307399,8356.943453700283,8357.052674320836,8357.160272397507,8357.266582897837,8357.371475783788,8357.474834507544,8357.576654372706,8357.677003015619,8357.775986048022,8357.873805475658,8357.970860597346,8358.06781687839,8358.165606656712,8358.265357510416,8358.368288747843,8358.475613388015,8358.58846670467,8358.707863690528,8358.834672771714,8358.969595921331,8359.113151565,8359.265663374988,8359.42725992684,8359.59788624168,8359.777324288087,8359.965217190465,8360.1610928427,8360.364385296842,8360.57445426541,8360.790603672185,8361.01209943573,8361.238185778417,8361.468099136195,8361.70107920417,8361.936377381633,8362.173263300952,8362.411030087293,8362.648998689747,8362.886521341108,8363.122984128036,8363.357808743222,8363.590453603922,8363.820414547517,8364.047225242126,8364.2704573497,8364.489720419546,8364.704661491913,8364.91496442904,8365.120349019919,8365.32056990659,8365.515415359598,8365.7047059084,8365.888292827813,8366.066056489934,8366.23790460615,8366.403770392473,8366.563610690082,8366.717404066541,8366.865148916117,8367.006861574459,8367.142574463644,8367.272334284224,8367.3962002712,8367.514242527408,8367.626540444558,8367.733181218275,8367.834258462379,8367.929870925169,8368.020121310836,8368.105115207334,8368.184960120494,8368.259764612656,8368.329637542553,8368.394687402355,8219.826635222327,9124.89504465676,10000.549872958354,10404.675741962694,10390.068804387018,10264.274461788353,10271.137857462054,10456.444827753252,10735.186006931966,11006.530892301202,11229.995902901024,11419.0120414212,11598.379804360884,11776.339068784855,11941.086283061628,12076.811565139764,12176.284051839626,12242.760535306585,12285.03464614949,12310.741384223616,12323.958273981483,12326.06250054084,12317.753418147096,12300.489026485053,12276.373781155515,12247.470701535627,12215.152536342846,12179.939322582451,12141.812272559202,12100.60726555516,12056.311215103715,12009.145178493764,11959.502246690881,11907.861516519362,11854.729582608272,11800.636075938797,11746.149496152755,11691.881903868582,11638.471364144109,11586.544458869335,11536.678003381687,11489.37239900492,11445.040754667796,11404.010126414421,11366.526899341228,11332.761957938048,11302.81467721727,11276.717558416933,11254.443377207603,11235.914670447404,11221.014060569454,11209.593513510172,11201.481551061579,11196.488615593184,11194.411326851352,11195.036291492584,11198.143626659923,11203.510014395317,11210.911099606594,11220.123253770449,11230.924942571502,11243.097970994702,11256.428765373183,11270.709701759792,11285.740403875654,11301.328940630912,11317.292901569934,11333.460364481165,11349.670768095415,11365.775677444106,11381.639409087162,11397.139483509216,11412.166889699944,11426.62616782082,11440.43532785565,11453.52562314377,11465.841193010398,11477.33858522988,11487.98616992592,11497.763460413405,11506.660360134458,11514.676355785383,11521.819674774884,11528.106421789913,11533.559706252012,11538.208770752452,11542.088129971573,11545.236729301547,11547.697131690456,11549.514739915054,11550.737059890738,11551.41300920583,11551.592274087792,11551.3247174006,11550.659839750955,11549.646295111994,11548.331461511269,11546.76106642483,11544.978865756884,11543.026374752379,11587.459624305484,11394.116709795659,11198.995874660188,11107.459038713738,11106.609498853184,11130.712790364298,11128.25058208172,11087.662924750428,11024.60905454645,10959.917948302957,10903.70071796351,10855.20710581006,10810.3064907535,10767.183886345974,10727.861966204931,10695.554003462219,10672.07543066471,10657.086154610064,10648.815326645567,10645.413778923734,10645.613175463044,10648.705060264656,10654.190693456565,10661.478912505012,10669.88736082941,10678.783810221586,10687.734857011714,10696.553825809466,10705.237359587216,10713.872488261737,10722.557425798355,10731.366106573325,10740.345831698358,10749.523383555446,10758.908581121706,10768.489292299942,10778.22469017746,10788.043142877557,10797.846725590327,10807.521425923413,10816.948469549856,10826.013623706094,10834.613162694568,10842.65693270891,10850.070001308644,10856.793723545656,10862.786433607991,10868.023436844043,10872.495990339516,10876.209398373536,10879.180629961922,10881.435952943217,10883.008883428909,10883.938490171582,10884.267958016311,10884.043306130552,10883.312241572552,10882.123190738153,10880.524548365827,10878.564137669719,10876.288825817654,10873.744226172596,10870.97443904545,10868.021814218107,10864.926739643462,10861.727463153944,10858.45994642076,10855.157743833017,10851.851899490235,10848.57086143362,10845.340417783726,10842.183660867884,10839.12098267975,10836.170100907048,10833.346112075673,10830.661567944406,10828.126572277533,10825.748896045638,10823.534109117747,10821.48572576018,10819.605360439049,10817.892890143476,10816.346619835258,10814.963448387341,10813.739033091211,10812.667951235468,10811.743857404688,10810.959635190165,10810.307542120043,10809.7793468566,10809.36645800958,10809.060044172098,10808.85114493327,10808.730772679679,10808.6900050281,10808.720067795162,10808.812408518821,10808.958760673899,10809.15119882942,10809.382185058641,10917.678126712744,10438.413058615979,9968.765559391151,9747.594710560295],"type":"scatter","name":"source2_q_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1,0.1001,0.1002,0.1003,0.1004,0.1005,0.1006,0.1007,0.1008,0.1009,0.101,0.1011,0.1012,0.1013,0.1014,0.1015,0.1016,0.1017,0.1018,0.1019,0.102,0.1021,0.1022,0.1023,0.1024,0.1025,0.1026,0.1027,0.1028,0.1029,0.103,0.1031,0.1032,0.1033,0.1034,0.1035,0.1036,0.1037,0.1038,0.1039,0.104,0.1041,0.1042,0.1043,0.1044,0.1045,0.1046,0.1047,0.1048,0.1049,0.105,0.1051,0.1052,0.1053,0.1054,0.1055,0.1056,0.1057,0.1058,0.1059,0.106,0.1061,0.1062,0.1063,0.1064,0.1065,0.1066,0.1067,0.1068,0.1069,0.107,0.1071,0.1072,0.1073,0.1074,0.1075,0.1076,0.1077,0.1078,0.1079,0.108,0.1081,0.1082,0.1083,0.1084,0.1085,0.1086,0.1087,0.1088,0.1089,0.109,0.1091,0.1092,0.1093,0.1094,0.1095,0.1096,0.1097,0.1098,0.1099,0.11,0.1101,0.1102,0.1103,0.1104,0.1105,0.1106,0.1107,0.1108,0.1109,0.111,0.1111,0.1112,0.1113,0.1114,0.1115,0.1116,0.1117,0.1118,0.1119,0.112,0.1121,0.1122,0.1123,0.1124,0.1125,0.1126,0.1127,0.1128,0.1129,0.113,0.1131,0.1132,0.1133,0.1134,0.1135,0.1136,0.1137,0.1138,0.1139,0.114,0.1141,0.1142,0.1143,0.1144,0.1145,0.1146,0.1147,0.1148,0.1149,0.115,0.1151,0.1152,0.1153,0.1154,0.1155,0.1156,0.1157,0.1158,0.1159,0.116,0.1161,0.1162,0.1163,0.1164,0.1165,0.1166,0.1167,0.1168,0.1169,0.117,0.1171,0.1172,0.1173,0.1174,0.1175,0.1176,0.1177,0.1178,0.1179,0.118,0.1181,0.1182,0.1183,0.1184,0.1185,0.1186,0.1187,0.1188,0.1189,0.119,0.1191,0.1192,0.1193,0.1194,0.1195,0.1196,0.1197,0.1198,0.1199,0.12,0.1201,0.1202,0.1203,0.1204,0.1205,0.1206,0.1207,0.1208,0.1209,0.121,0.1211,0.1212,0.1213,0.1214,0.1215,0.1216,0.1217,0.1218,0.1219,0.122,0.1221,0.1222,0.1223,0.1224,0.1225,0.1226,0.1227,0.1228,0.1229,0.123,0.1231,0.1232,0.1233,0.1234,0.1235,0.1236,0.1237,0.1238,0.1239,0.124,0.1241,0.1242,0.1243,0.1244,0.1245,0.1246,0.1247,0.1248,0.1249,0.125,0.1251,0.1252,0.1253,0.1254,0.1255,0.1256,0.1257,0.1258,0.1259,0.126,0.1261,0.1262,0.1263,0.1264,0.1265,0.1266,0.1267,0.1268,0.1269,0.127,0.1271,0.1272,0.1273,0.1274,0.1275,0.1276,0.1277,0.1278,0.1279,0.128,0.1281,0.1282,0.1283,0.1284,0.1285,0.1286,0.1287,0.1288,0.1289,0.129,0.1291,0.1292,0.1293,0.1294,0.1295,0.1296,0.1297,0.1298,0.1299,0.13,0.1301,0.1302,0.1303,0.1304,0.1305,0.1306,0.1307,0.1308,0.1309,0.131,0.1311,0.1312,0.1313,0.1314,0.1315,0.1316,0.1317,0.1318,0.1319,0.132,0.1321,0.1322,0.1323,0.1324,0.1325,0.1326,0.1327,0.1328,0.1329,0.133,0.1331,0.1332,0.1333,0.1334,0.1335,0.1336,0.1337,0.1338,0.1339,0.134,0.1341,0.1342,0.1343,0.1344,0.1345,0.1346,0.1347,0.1348,0.1349,0.135,0.1351,0.1352,0.1353,0.1354,0.1355,0.1356,0.1357,0.1358,0.1359,0.136,0.1361,0.1362,0.1363,0.1364,0.1365,0.1366,0.1367,0.1368,0.1369,0.137,0.1371,0.1372,0.1373,0.1374,0.1375,0.1376,0.1377,0.1378,0.1379,0.138,0.1381,0.1382,0.1383,0.1384,0.1385,0.1386,0.1387,0.1388,0.1389,0.139,0.1391,0.1392,0.1393,0.1394,0.1395,0.1396,0.1397,0.1398,0.1399,0.14,0.1401,0.1402,0.1403,0.1404,0.1405,0.1406,0.1407,0.1408,0.1409,0.141,0.1411,0.1412,0.1413,0.1414,0.1415,0.1416,0.1417,0.1418,0.1419,0.142,0.1421,0.1422,0.1423,0.1424,0.1425,0.1426,0.1427,0.1428,0.1429,0.143,0.1431,0.1432,0.1433,0.1434,0.1435,0.1436,0.1437,0.1438,0.1439,0.144,0.1441,0.1442,0.1443,0.1444,0.1445,0.1446,0.1447,0.1448,0.1449,0.145,0.1451,0.1452,0.1453,0.1454,0.1455,0.1456,0.1457,0.1458,0.1459,0.146,0.1461,0.1462,0.1463,0.1464,0.1465,0.1466,0.1467,0.1468,0.1469,0.147,0.1471,0.1472,0.1473,0.1474,0.1475,0.1476,0.1477,0.1478,0.1479,0.148,0.1481,0.1482,0.1483,0.1484,0.1485,0.1486,0.1487,0.1488,0.1489,0.149,0.1491,0.1492,0.1493,0.1494,0.1495,0.1496,0.1497,0.1498,0.1499,0.15,0.1501,0.1502,0.1503,0.1504,0.1505,0.1506,0.1507,0.1508,0.1509,0.151,0.1511,0.1512,0.1513,0.1514,0.1515,0.1516,0.1517,0.1518,0.1519,0.152,0.1521,0.1522,0.1523,0.1524,0.1525,0.1526,0.1527,0.1528,0.1529,0.153,0.1531,0.1532,0.1533,0.1534,0.1535,0.1536,0.1537,0.1538,0.1539,0.154,0.1541,0.1542,0.1543,0.1544,0.1545,0.1546,0.1547,0.1548,0.1549,0.155,0.1551,0.1552,0.1553,0.1554,0.1555,0.1556,0.1557,0.1558,0.1559,0.156,0.1561,0.1562,0.1563,0.1564,0.1565,0.1566,0.1567,0.1568,0.1569,0.157,0.1571,0.1572,0.1573,0.1574,0.1575,0.1576,0.1577,0.1578,0.1579,0.158,0.1581,0.1582,0.1583,0.1584,0.1585,0.1586,0.1587,0.1588,0.1589,0.159,0.1591,0.1592,0.1593,0.1594,0.1595,0.1596,0.1597,0.1598,0.1599,0.16,0.1601,0.1602,0.1603,0.1604,0.1605,0.1606,0.1607,0.1608,0.1609,0.161,0.1611,0.1612,0.1613,0.1614,0.1615,0.1616,0.1617,0.1618,0.1619,0.162,0.1621,0.1622,0.1623,0.1624,0.1625,0.1626,0.1627,0.1628,0.1629,0.163,0.1631,0.1632,0.1633,0.1634,0.1635,0.1636,0.1637,0.1638,0.1639,0.164,0.1641,0.1642,0.1643,0.1644,0.1645,0.1646,0.1647,0.1648,0.1649,0.165,0.1651,0.1652,0.1653,0.1654,0.1655,0.1656,0.1657,0.1658,0.1659,0.166,0.1661,0.1662,0.1663,0.1664,0.1665,0.1666,0.1667,0.1668,0.1669,0.167,0.1671,0.1672,0.1673,0.1674,0.1675,0.1676,0.1677,0.1678,0.1679,0.168,0.1681,0.1682,0.1683,0.1684,0.1685,0.1686,0.1687,0.1688,0.1689,0.169,0.1691,0.1692,0.1693,0.1694,0.1695,0.1696,0.1697,0.1698,0.1699,0.17,0.1701,0.1702,0.1703,0.1704,0.1705,0.1706,0.1707,0.1708,0.1709,0.171,0.1711,0.1712,0.1713,0.1714,0.1715,0.1716,0.1717,0.1718,0.1719,0.172,0.1721,0.1722,0.1723,0.1724,0.1725,0.1726,0.1727,0.1728,0.1729,0.173,0.1731,0.1732,0.1733,0.1734,0.1735,0.1736,0.1737,0.1738,0.1739,0.174,0.1741,0.1742,0.1743,0.1744,0.1745,0.1746,0.1747,0.1748,0.1749,0.175,0.1751,0.1752,0.1753,0.1754,0.1755,0.1756,0.1757,0.1758,0.1759,0.176,0.1761,0.1762,0.1763,0.1764,0.1765,0.1766,0.1767,0.1768,0.1769,0.177,0.1771,0.1772,0.1773,0.1774,0.1775,0.1776,0.1777,0.1778,0.1779,0.178,0.1781,0.1782,0.1783,0.1784,0.1785,0.1786,0.1787,0.1788,0.1789,0.179,0.1791,0.1792,0.1793,0.1794,0.1795,0.1796,0.1797,0.1798,0.1799,0.18,0.1801,0.1802,0.1803,0.1804,0.1805,0.1806,0.1807,0.1808,0.1809,0.181,0.1811,0.1812,0.1813,0.1814,0.1815,0.1816,0.1817,0.1818,0.1819,0.182,0.1821,0.1822,0.1823,0.1824,0.1825,0.1826,0.1827,0.1828,0.1829,0.183,0.1831,0.1832,0.1833,0.1834,0.1835,0.1836,0.1837,0.1838,0.1839,0.184,0.1841,0.1842,0.1843,0.1844,0.1845,0.1846,0.1847,0.1848,0.1849,0.185,0.1851,0.1852,0.1853,0.1854,0.1855,0.1856,0.1857,0.1858,0.1859,0.186,0.1861,0.1862,0.1863,0.1864,0.1865,0.1866,0.1867,0.1868,0.1869,0.187,0.1871,0.1872,0.1873,0.1874,0.1875,0.1876,0.1877,0.1878,0.1879,0.188,0.1881,0.1882,0.1883,0.1884,0.1885,0.1886,0.1887,0.1888,0.1889,0.189,0.1891,0.1892,0.1893,0.1894,0.1895,0.1896,0.1897,0.1898,0.1899,0.19,0.1901,0.1902,0.1903,0.1904,0.1905,0.1906,0.1907,0.1908,0.1909,0.191,0.1911,0.1912,0.1913,0.1914,0.1915,0.1916,0.1917,0.1918,0.1919,0.192,0.1921,0.1922,0.1923,0.1924,0.1925,0.1926,0.1927,0.1928,0.1929,0.193,0.1931,0.1932,0.1933,0.1934,0.1935,0.1936,0.1937,0.1938,0.1939,0.194,0.1941,0.1942,0.1943,0.1944,0.1945,0.1946,0.1947,0.1948,0.1949,0.195,0.1951,0.1952,0.1953,0.1954,0.1955,0.1956,0.1957,0.1958,0.1959,0.196,0.1961,0.1962,0.1963,0.1964,0.1965,0.1966,0.1967,0.1968,0.1969,0.197,0.1971,0.1972,0.1973,0.1974,0.1975,0.1976,0.1977,0.1978,0.1979,0.198,0.1981,0.1982,0.1983,0.1984,0.1985,0.1986,0.1987,0.1988,0.1989,0.199,0.1991,0.1992,0.1993,0.1994,0.1995,0.1996,0.1997,0.1998,0.1999,0.2]}],{"xaxis":{"title":{"text":"Time in Seconds"}},"template":{"layout":{"coloraxis":{"colorbar":{"ticks":"","outlinewidth":0}},"xaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"hovermode":"closest","paper_bgcolor":"white","geo":{"showlakes":true,"showland":true,"landcolor":"#E5ECF6","bgcolor":"white","subunitcolor":"white","lakecolor":"white"},"colorscale":{"sequential":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]],"diverging":[[0,"#8e0152"],[0.1,"#c51b7d"],[0.2,"#de77ae"],[0.3,"#f1b6da"],[0.4,"#fde0ef"],[0.5,"#f7f7f7"],[0.6,"#e6f5d0"],[0.7,"#b8e186"],[0.8,"#7fbc41"],[0.9,"#4d9221"],[1,"#276419"]],"sequentialminus":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]},"yaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"shapedefaults":{"line":{"color":"#2a3f5f"}},"hoverlabel":{"align":"left"},"mapbox":{"style":"light"},"polar":{"angularaxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","radialaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"autotypenumbers":"strict","font":{"color":"#2a3f5f"},"ternary":{"baxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","caxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"aaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"annotationdefaults":{"arrowhead":0,"arrowwidth":1,"arrowcolor":"#2a3f5f"},"plot_bgcolor":"#E5ECF6","title":{"x":0.05},"scene":{"xaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"zaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"yaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"}},"colorway":["#636efa","#EF553B","#00cc96","#ab63fa","#FFA15A","#19d3f3","#FF6692","#B6E880","#FF97FF","#FECB52"]},"data":{"barpolar":[{"type":"barpolar","marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"carpet":[{"aaxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"},"type":"carpet","baxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"}}],"scatterpolar":[{"type":"scatterpolar","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"parcoords":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"parcoords"}],"scatter":[{"type":"scatter","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2dcontour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2dcontour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattercarpet":[{"type":"scattercarpet","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"mesh3d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"mesh3d"}],"surface":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"surface","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattermapbox":[{"type":"scattermapbox","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergeo":[{"type":"scattergeo","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram":[{"type":"histogram","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"pie":[{"type":"pie","automargin":true}],"choropleth":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"choropleth"}],"heatmapgl":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmapgl","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"bar":[{"type":"bar","error_y":{"color":"#2a3f5f"},"error_x":{"color":"#2a3f5f"},"marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"heatmap":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmap","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contourcarpet":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contourcarpet"}],"table":[{"type":"table","header":{"line":{"color":"white"},"fill":{"color":"#C8D4E3"}},"cells":{"line":{"color":"white"},"fill":{"color":"#EBF0F8"}}}],"scatter3d":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"scatter3d","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergl":[{"type":"scattergl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2d","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scatterternary":[{"type":"scatterternary","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scatterpolargl":[{"type":"scatterpolargl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}]}},"legend":{"yanchor":"bottom","xanchor":"right","y":1.02,"orientation":"h","x":1},"margin":{"l":100,"b":80,"pad":10,"r":80,"t":100},"plot_bgcolor":"#f1f3f7","yaxis":{"title":{"text":"State values"}},"yaxis2":{"titlefont":{"color":"orange"},"overlaying":"y","title":"Action values","side":"right"},"height":550,"width":800},{"showLink":false,"editable":false,"responsive":true,"staticPlot":false,"scrollZoom":true});
});





</script>
```




_______________________________________________________________________________
### Analysis

- The active and reactive powers of the PQ load fluctuate around their mean set points. 
- Increasing the interpolation degree to either 1 or 2 will remove the "steps" and smooth the output powers.
- The user can render all 4 episodes, which will show different dynamics.
