# Droop Mode - Simple Grid Forming

|  | Mode | Description |
| --- | --- | --- |
| 1 | "Swing" | Ideal voltage source without dynamics (i.e., an infinite bus) |
| 2 | "PQ" | Grid following controllable source/load (real and imaginary power) |
| 3 | __"Droop"__ | __Simple grid forming with power balancing through a droop mechanism__ |
| 4 | "Synchronverter" or "VSG"|  Grid forming control mimicking a generator, i.e., virtual synchronous generator |

- The control structure of power electronic inverters can be divided into cascading levels. 
- The second level above the current control loop is a voltage control loop.
- A third level on top of the voltage control is called "Droop".

## Summary

- The following example is intended to introduce you to a decentralised and self-organising control mode.
- The mode "communicates" with other inverters in the network through the voltage and frequency.
- Often battery energy storage systems (BESSs) operate in this mode, especially in remote islanded grid areas.
- This mode is autonomous and provides a measure of "inertia" to the network.

### Theory

The droop control method can be referred to as an independent, autonomous concept since intercommunication links between the converters other than the electrical cables can be eliminated. That is, in a network with loads and inverters in droop mode, the power required by the loads is partitioned amongst the inverters in droop modes by only varying electrical quantities are their respective control points. 

This mode works quite well when high frequency harmonics and switching ripples are neglected (as they presently are in the set-up). An inverter in "Droop" mode is capable of other negative oscillatory behaviour, similar to the hunting effect in synchronous generators. In fact, the droop mode is a simplified version of the Virtual Synchronous Generator (VSG) mode with a low-pass filtering characteristic on the measured real and imaginary powers. Unlike the PQ mode, a source in droop mode has to have some energy storage available. 

The voltage control loop has a very similar structure to the inner current control loop, also operating in the DQ0 frame and with PI controllers. 
![](assets/voltage_control.jpg)

The "droop" control sitting on top of the voltage control loop, provides the reference voltage signals. The droop mode for every inverter has two coefficients, $D_p$ and $D_q$, that determine its behaviour. $D_p$ is associated with the real power and frequency, while $D_q$ corresponds to the voltage magnitude and the imaginary power. These coefficients can by tuned either heuristically or via more advanced optimisation algorithms.

![](assets/Droop_control.jpg)

If the user does not specify $D_p$ and $D_q$, then they are automatically tuned based on the converter power rating and the network's maximum allowable frequency and voltage deviations, i.e., $Δfmax$ and $ΔEmax$. The formulas that are implemented to calculate the droop coefficients are as follows,

```math

  D_p = \frac{S_{rated}}{{ω_{set}}^2 Δfmax}, \quad D_q = \frac{S_{rated}}{V_{set}*\sqrt{2}*ΔEmax}.
```


![](assets/Droop_mode.jpg)


```julia
using ElectricGrid;
```

_______________________________________________________________________________
### Network Configuration 

- Two inverters are connected to each other through a cable.
- The one inverter is placed in the familiar PQ mode, while the other is in droop mode.
- The droop inverter will regulate the frequency and the PQ inverter will follow along.
- If not provided by the user, the proportional and integral gains for the PI controllers are tuned automatically via a loop-shaping method.

With this configuration, a time domain simulation is conducted below.


```julia
# total run time, seconds
t_end = 0.08     

# Connectivity Matrix
CM = [ 0. 1.
        -1. 0.]     

parameters = Dict{Any, Any}(
        "source" => Any[
                        Dict{Any, Any}("pwr"    => 400e3,   # Rated appared power [VA] 
                                        "mode"  => "Droop", # Controller mode
                                        "V_kp"  => 0.6,     # Voltage proportional gain [A/V] (**optional**)
                                        "V_ki"  => 12,      # Current integral gain [A/V⋅s] (**optional**)
                                        "Dp"    => 810,     # Frequency droop coefficient [N⋅m⋅s/rad] (**optional**)
                                        "Dq"    => 24e3)    # Voltage droop coefficient [VAi/V] (**optional**)
                        Dict{Any, Any}("pwr"    => 100e3,   
                                        "mode"  => "PQ",    
                                        "p_set" => 50e3,    # Real power set point (generating) [W] 
                                        "q_set" => -40e3)   # Imaginary power set point (inductive) [VAi]  
                        ],
        "cable"   => Any[
                        Dict{Any, Any}("R"  => 0.1, 
                                        "L" => 0.25e-3, 
                                        "C" => 0.1e-4),
                        ],
        "grid"   => Dict{Any, Any}("Δfmax"  => 0.5, # The % drop (increase) in frequency that causes a 100% increase (decrease) in active power (from nominal).
                                    "ΔEmax" => 5)   # The % drop (increase) in voltage that causes a 100% increase (decrease) in reactive power (from nominal).

    );

env = ElectricGridEnv(CM = CM, parameters = parameters, t_end = t_end, verbosity = 2);

agents = SetupAgents(env);

hook = Simulate(agents, env);
```

    ┌ Info: Normalization is done based on the defined parameter limits.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\electric_grid_env.jl:329
    ┌ Info: Time simulation run time: 0.08 [s] ~> 801 steps
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\electric_grid_env.jl:330
    ┌ Info: 2 'classically' controlled sources have been initialised.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2686
    ┌ Info: 1 source has been set up in PQ mode.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2695
    ┌ Info: 1 source has been set up in Droop mode.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2695
    ┌ Info: 2 sources have automatically calculated proportional and integral gains for their current control loops.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2720
    

_______________________________________________________________________________
### Low-Level Rendering


```julia
RenderHookResults(hook = hook, 
                    states_to_plot  = ["source1_i_L1_a"], # Inductor current [A]
                    actions_to_plot = ["source1_u_a"],    # Inverter voltage [V]
                    )
```


```@raw html
<div
id = oxfhEZUl2jZ1 > </div>
<script>
gd = 'oxfhEZUl2jZ1'
require(['plotly'], function(plotly) {
























plotly.newPlot(gd,[{"mode":"lines","y":[0.0,0.0,0.17629019093995515,0.488311420592334,0.8323979832410104,1.1441031575082794,1.418740874396133,1.6791378010705293,1.9407165784307827,2.1981766557549047,2.432777739295826,2.626357320286936,2.7696048556621955,2.862832435813763,2.9116288300264475,2.9228636726736816,2.902523312536116,2.8555968716839875,2.7865347054051997,2.6994164083033336,2.597809218920944,2.48440570449784,2.360963247020636,2.228377127864885,2.086975217907924,1.936759800161717,1.777610728835654,1.6094113310067852,1.4321146827770153,1.2458026275086853,1.0506946531811474,0.8471595108036862,0.6356852150000625,0.41685415178884705,0.19130560710055877,-0.04029367104137753,-0.27727280180833697,-0.5189774077145277,-0.7647790575838616,-1.0140845962146763,-1.266334394590995,-1.5210003950617743,-1.777578171149347,-2.035579837469935,-2.294526943329333,-2.553944421665771,-2.8133569839192094,-3.072285702521216,-3.3302470710438374,-3.58675205932698,-3.8413070518203267,-4.093415043646906,-4.342577685721182,-4.58829748781775,-4.830079814788256,-5.067434826681523,-5.299878796033818,-5.5269353643298755,-5.748136308685837,-5.963022277121401,-6.171143265637162,-6.372059023376735,-6.565339380423224,-6.750564506619643,-6.927325238085045,-7.095223395543205,-7.2538722284354336,-7.402896881697821,-7.541934946206126,-7.670637021195896,-7.788667294155005,-7.8957041217073165,-7.991440588648923,-8.075585054459252,-8.147861647113734,-8.208010725022234,-8.255789307607218,-8.290971513604163,-8.313349000215881,-8.322731383851176,-8.31894661443743,-8.301841287020373,-8.271280904492853,-8.227150104816808,-8.169352871991759,-8.097812729619854,-8.012472915222014,-7.9132965288745405,-7.800266655766308,-7.673386465492546,-7.532679289712878,-7.378188679765931,-7.2099784413834485,-7.02813264521549,-6.832755610674325,-6.6239718629132875,-6.401926062596557,-6.166782908506947,-5.9187270130482235,-5.65796275035783,-5.384714077417037,-5.099224328275483,-4.801755982124479,-4.492590405615128,-4.172027570002245,-3.8403857434927904,-3.498001159168503,-3.1452276588903274,-2.7824363135112584,-2.4100150198401913,-2.0283680746912234,-1.6379157264547974,-1.239093704574916,-0.8323527273963942,-0.4181579888692586,0.003011375355574291,0.4306628418488861,0.8642910727374634,1.30337850778761,1.7473959745685927,2.1958033159825128,2.6480500344354327,3.1035759518941837,3.561811885056841,4.022180334847475,4.4840961894265865,4.946967439897644,5.410195907870339,5.873177984029872,6.33530537684269,6.795965870516589,7.254544091316896,7.710422281327546,8.16298107873354,8.61160030368873,9.055659748823757,9.494539973438139,9.92762310041333,10.354293614875903,10.773939163635752,11.185951354418787,11.589726553911715,11.984666683634575,12.370180012656203,12.745681946169325,13.110595808942533,13.464353622670968,13.80639687625132,14.136177288011561,14.453157558933745,14.756812115915157,15.046627844121792,15.322104807499203,15.582756956516395,15.828112822230617,16.057716195774997,16.27112679238547,16.467920899099227,16.647692005273512,16.810051415092047,16.954628841245217,17.081072978989887,17.189052059816277,17.27825438397085,17.348388831107442,17.399185348362664,17.430395415176584,17.441792484204996,17.433172397696627,17.404353778735405,17.355178396776374,17.285511506932398,17.195242162499024,17.084283500234243,16.952572997941413,16.800072703934752,16.6267694379992,16.432674963488942,16.21782613024193,15.982284988021247,15.726138870228667,15.449500447669536,15.152507752183057,14.835324169987475,14.498138404624363,14.141164409421927,13.764641289433515,13.368833172842226,12.95402905185982,12.52054259318284,12.068711918106121,11.598899352428452,11.111491146322603,10.606897164376596,10.085550546049223,9.547907336818493,8.994446090336092,8.425667441937176,7.842093653888525,7.244268132792677,6.632754919599163,6.0081381527085425,5.371021504686857,4.722027593140636,4.061797366335289,3.390989464170501,2.7102795551568715,2.0203596500686007,1.3219373929756506,0.6157353303885694,-0.09750984072455937,-0.8170480452609761,-1.5421166218324198,-2.2719411333016324,-3.0057361951429913,-3.742706320739706,-4.4820467827083,-5.222944489318964,-5.964578875059979,-6.706122804374584,-7.446743487580447,-8.185603407964308,-8.921861259028015,-9.654672890846205,-10.38319226448405,-11.106572413407301,-11.823966410807836,-12.534528341756147,-13.237414279083,-13.931783261884329,-14.61679827553684,-15.291627232105501,-15.95544395002051,-16.60742913189793,-17.246771339375396,-17.87266796383514,-18.484326191886467,-19.080963964481743,-19.661810928543144,-20.2261093799824,-20.773115197001253,-21.302098762566978,-21.812345874966542,-22.303158645351747,-22.77385638119906,-23.223776454619763,-23.65227515446909,-24.058728521218107,-24.442533163567617,-24.803107055800105,-25.13989031488407,-25.452345956364734,-25.739960628094646,-26.002245320880483,-26.238736055143832,-26.448994542718406,-26.632608822930628,-26.789193872136067,-26.918392185911767,-27.019874333131543,-27.093339481180788,-27.1385158915967,-27.1551613854503,-27.1430637778189,-27.102041280728695,-27.031942873981247,-26.932648643311012,-26.804070085356305,-26.646150378960023,-26.45886462235376,-26.242220035814903,-25.996256129422918,-25.72104483557967,-25.416690605996095,-25.08333047288611,-24.721134074148125,-24.33030364235303,-23.91107395739853,-23.463712262728333,-22.988518145055878,-22.485823377571894,-21.955991726656872,-21.3994187221588,-20.816531391338337,-20.20778795662339,-19.573677497357426,-18.914719575764344,-18.23146382739535,-17.524489516362234,-16.79440505570227,-16.04184749325996,-15.267481963510027,-14.472001105786426,-13.656124449419261,-12.820597766322342,-11.966192391609518,-11.093704512857917,-10.203954428671556,-9.29778577723583,-8.376064735588477,-7.439679190368846,-6.48953788084018,-5.526569515014578,-4.551721859741143,-3.5659608056525847,-2.570269407892925,-1.5656469035832656,-0.5531077070070266,0.46631961647116665,1.4915933967147215,2.5216599242869258,3.5554545339952255,4.591902705393753,5.62992117910616,6.6684190878114045,7.706299100712774,8.742458580292483,9.775790750134508,10.80518587258493,11.829532435000141,12.847718343321393,13.858632121700467,14.861164116891326,15.85420770611077,16.83666050706349,17.80742558881972,18.76541268222838,19.70953938854309,20.638732384935945,21.55192862557282,22.448076536924823,23.326137205988903,24.18508556009706,25.02391153699476,25.84162124387805,26.637238104083625,27.409803990134723,28.15838034185834,28.88204926829772,29.5799146321593,30.251103115546695,30.89476526574952,31.51007651987335,32.09623820711445,32.652478527504485,33.178053505969146,33.672247920569205,34.1343762038163,34.5637833159802,34.95984558933023,35.32197154228283,35.64960266245515,35.942214157653936,36.19931567386254,36.42045197931726,36.60520361380234,36.75318750232357,36.86405753235778,36.937505093912144,36.97325958166344,36.97108885848709,36.93079967972428,36.85223807757493,36.73528970504694,36.57988013893206,36.385975141321914,36.15358087921959,35.88274410184813,35.573552275298496,35.226133674206494,34.840657430192266,34.41733353684201,33.95641281105716,33.45818681064353,32.922987708058066,32.35118812027959,31.743200894814866,31.09947885190072,30.42051448300881,29.70683960580697,28.959024975779634,28.177679854755024,27.363451536636667,26.5170248306813,25.63912150271372,24.730499674713958,23.79195318326036,22.82431089735626,21.828435996214512,20.805225207618232,19.75560800752161,18.680545781596713,17.581030949477334,16.458086052492384,15.3127628057236,14.146141115263474,12.959328061589302,11.753456850008623,10.529685729170499,9.289196878674703,8.033195266845986,6.762907479778152,5.256565488973273,2.517807717854592,-5.096511958692663,-23.50889260228605,-54.02883933914173,-89.63852468651095,-118.6257802617894,-132.04164352682093,-128.46201271999405,-113.00764795650795,-92.82491970138138,-73.17362777602318,-56.27652354264524,-42.41972838017615,-31.445304532625343,-23.446237634542307,-18.59142048063684,-16.70692071615169,-17.12818537716938,-18.895567919439,-21.083280140567314,-23.017371731313347,-24.320784929090724,-24.8380989081656,-24.54409987925788,-23.484986058557467,-21.753894934086755,-19.47843992279259,-16.802511598085285,-13.864957553485196,-10.78037181328664,-7.630766542863502,-4.466161869410088,-1.3114876251299157,1.8253503433558624,4.946588453426198,8.059339283995376,11.17217242967721,14.292572341930592,17.425676978935645,20.57360862590062,23.735590184008775,26.908336600577606,30.08662485181319,33.2638969335251,36.43275775511073,39.58545529896305,42.714188608203344,45.811366953975984,48.86972709926418,51.88241778610917,54.843040130589486,57.74568295325582,60.58496031599238,63.356021280389555,66.05454911477884,68.67671850502283,71.21914571987347,73.67882318453167,76.05306386328994,78.3394527302299,80.5358071866767,82.64014629979906,84.6506619770011,86.56569829718146,88.38373348038786,90.10337031677112,91.72332978468734,93.24244827435274,94.65967474214656,95.97406632622962,97.18478290046562,98.29108007681442,99.29230314574215,100.18788142953088,100.97732445978279,101.66021906781836,102.23622762722923,102.7050871210728,103.06660897436088,103.32067988791114,103.46726332965247,103.50640179389899,103.43821925842406,103.26292377417235,102.98080984302362,102.59226059045267,102.09774972379165,101.49784328915106,100.79320129425288,99.98457911083692,99.07282868136552,98.05889943785336,96.94383896178282,95.7287933631919,94.41500740463427,93.00382437622247,91.49668571109812,89.89513034560512,88.20079380477613,86.41540703193684,84.54079496673339,82.57887490192792,80.53165463680759,78.40123044854425,76.18978489770544,73.89958447953097,71.53297713782698,69.09238965336148,66.58032492509,63.999359156484566,61.352138960943165,58.641378395601905,55.8698559322421,53.04041137323006,50.15594271943365,47.219402998064055,44.23379705650339,41.20217832846897,38.12764557674005,35.01333961642967,31.862440021721312,28.67816181875858,25.46375216721122,22.222487032586688,18.957667851325915,15.672618190113711,12.370680400839596,9.055212272322644,5.729583680053375,2.397173235223047,-0.9386350655645279,-4.274455188382575,-7.60690240631897,-10.93259663835456,-14.24816578101566,-17.550249030829924,-20.835500195484297,-24.100590991386888,-27.34221432519437,-30.557087556728444,-33.741955740579314,-36.893594843593505,-40.00881493532857,-43.08446334847094,-46.11742780612161,-49.104639512796915,-52.043076205942306,-54.929765164734036,-57.76178617293065,-60.53627443253946,-63.250423425077855,-65.90148771723474,-68.48678570777682,-71.00370231259525,-73.44969158485358,-75.82227926726637,-78.11906527362083,-80.3377260967418,-82.47601714018997,-84.53177497108338,-86.50291949153473,-88.38745602630527,-90.1834773243847,-91.88916547231959,-93.50279371722814,-95.02272819755437,-96.44742957972926,-97.77545459902574,-99.00545750300991,-100.13619139610871,-101.16650948392976,-102.0953662160878,-102.92181832640507,-103.64502576946985,-104.2642525526501,-104.77886746277412,-105.18834468680132,-105.49226432591938,-105.69031280261429,-105.78228316037011,-105.76807525576581,-105.64769584284427,-105.42125854973749,-105.08898374763885,-104.65119831232083,-104.10833527850214,-103.46093338747427,-102.70963652850126,-101.85519307461222,-100.89845511350696,-99.8403775744006,-98.68201725173259,-97.42453172676709,-96.06917818821077,-94.61731215307546,-93.07038608910676,-91.42994794020032,-89.69763955632018,-87.8751950295276,-85.96443893782259,-83.967284498591,-81.88573163353591,-79.72186494706578,-77.47785162018991,-75.15593922206257,-72.75845344139253,-70.28779574001805,-67.74644093102236,-65.13693468384008,-62.46189095887834,-59.72398937424327,-56.925972507233006,-54.070643133320395,-51.16086140541145,-48.19954197622415,-45.18965106668822,-42.134203483319915,-39.03625958757528,-35.898922220232286,-32.725333583897694,-29.5186720867726,-26.282149150850554,-23.019005987755854,-19.732510345459115,-16.425953229138052,-13.102645599473238,-9.765915051690452,-6.419102478680141,-3.065558721538411,0.29135878911577806,3.6482893976833326,7.001872590620545,10.348751362598335,13.685575578419217,17.009005327331188,20.315714266387115,23.602392949510214,26.865752138942042,30.102526095768763,33.30947584624298,36.48339242064453,39.62110006145248,42.719459397631645,45.775370581872934,48.78577638766421,51.74766526311055,54.65807433846645,57.51409238438972,60.312862717977495,63.051586053697775,65.72752329638564,68.33799827353232,70.8804004041564,73.35218730161107,75.7508873077476,78.07410195592497,80.31950836042641,82.48486152991877,84.56799660266624,86.56683100128878,88.47936650493672,90.30369123683602,92.03798156524218,93.68050391592968,95.22961649443089,96.68377091632892,98.04151374400232,99.30148792831011,100.46243415380499,101.52319208615391,102.48270152054747,103.34000342997545,104.09424091234716,104.7446600355354,105.29061057952498,105.73154667494983,106.0670273374049,106.29671689702366,106.42038532291627,106.43790844216781,106.34926805320113,106.15455193341388,105.85395374110462,105.44777281180774,104.93641384926228,104.32038651134397,103.60030489139501,102.77688689548876,101.85095351627044,100.82342800411877,99.69533493647265,98.46779918627124,97.14204479055384,95.7193937203648,94.20126455320819,92.58917104939175,90.88472063369649,89.08961278390038,87.20563732777889,85.23467265029349,83.17868381276999,81.03972058595481,78.81991539892225,76.52148120589003,74.14670927308082,71.69796688784565,69.17769499234385,66.58840574414776,63.93268000621197,61.21316476871711,58.43257050536576,55.59366846677223,52.699287913649236,49.75231329255676,46.755681357030426,43.71237823696412,40.62543645917075,37.497931922093926,34.33298082768709,31.133736573520395,27.90338660821253,24.645149253322103,21.362270494866525,18.058020747663043,14.735691595716021,11.39859251189651,8.050047560180548,4.6933920837292,1.3319693821054328,-2.03087261906286,-5.391784702662342,-8.747419695597344,-12.094435801899804,-15.429499930130955,-18.749291011766385,-22.050503307264204,-25.32984969653198,-28.58406495052032,-31.809908980693574,-35.00417006314743,-38.16366803417203,-41.28525745408569,-44.3658307361981,-47.402321237795576,-50.39170631007861,-53.331010304026755,-56.217307529205016,-59.04772516257777,-61.81944610444336,-64.52971177865605,-67.17582487435854,-69.7551520265063,-72.26512643252637,-74.70325040251642,-77.06709784045667,-79.35431665397672,-81.56263109028949,-83.68984399597903,-85.73383899840505,-87.69258260656393,-89.56412622932761,-91.34660810906311,-93.03825516872092,-94.63738477056543,-96.14240638480929,-97.55182316650287,-98.86423343912188,-100.07833208338879,-101.19291182995771,-102.2068644546884,-103.11918187533198,-103.9289571485485,-104.63538536627692,-105.23776445057588,-105.73549584615708,-106.12808510993324,-106.41514239700483,-106.5963828426142,-106.67162683969727,-106.64080021176765,-106.50393428097195,-106.26116583125938,-105.91273696671216,-105.45899486518796,-104.90039142752913,-104.237482822698,-103.47092892929939,-102.60149267405566,-101.63003926790155,-100.55753534046679,-99.38504797381663,-98.11374363641907,-96.74488701840721,-95.27983976930352,-93.72005913946909,-92.0670965266344,-90.32259592896685,-88.48829230621836],"type":"scatter","name":"source1_i_L1_a","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.32120476446538326,0.644144520343221,0.8850969611558327,1.1017467695028802,1.373336194423464,1.7301506560226105,2.1550070854495,2.6151565227232743,3.0896206282384866,3.576451595023004,4.083214513183778,4.616758823524379,5.176818165905979,5.7575459282128385,6.350562570811913,6.9480374818891395,7.543481838132241,8.131284291821746,8.706289276158392,9.263421873882287,9.79808127405656,10.306196411043441,10.784440416962992,11.230037961399821,11.64063571682009,12.014144376719058,12.348622344877937,12.642263312077795,12.893319666446784,13.100122943261457,13.26103312293195,13.374464559536692,13.438880351809773,13.452816005843617,13.414897498872763,13.323849159322068,13.178513419023755,12.977848936138106,12.720945468086104,12.407019906231959,12.035422049962598,11.60563006560537,11.11724809108973,10.57000318333004,9.963740395965434,9.29842230374441,8.57412505103697,7.791038917844127,6.949465954190287,6.049820300689492,5.0926274108176495,4.078524055974472,3.0082586252114596,1.8826907691765111,0.7027917019276787,-0.5303567179223395,-1.815562420214006,-3.151524526679006,-4.536834580617496,-5.969978246368051,-7.449337159384608,-8.973190901577379,-10.539719319991885,-12.14700481698867,-13.793034975943991,-15.475705135373433,-17.1928211932999,-18.942102430423883,-20.721184476967608,-22.527622391663957,-24.358893837938776,-26.212402418586528,-28.085481080352537,-29.975395674712768,-31.879348624709092,-33.794482757245845,-35.71788523163234,-37.64659154921995,-39.577589628594964,-41.50782395300328,-43.43419982583349,-45.353587721215014,-47.2628277336476,-49.158734089654274,-51.038099716625005,-52.89770085572935,-54.734301721421005,-56.54465920592223,-58.32552762025692,-60.07366346649087,-61.78583022835865,-63.458803177151715,-65.08937418458687,-66.67435653989432,-68.2105897643165,-69.69494441706517,-71.12432688625668,-72.4956841576007,-73.80600855529906,-75.05234244799131,-76.231782914137,-77.34148635973007,-78.37867308219624,-79.3406317737842,-80.2247239581322,-81.02838835386882,-81.74914515897655,-82.384600250084,-82.93244929057295,-83.39048174183608,-83.75658477187014,-84.0287470557053,-84.20506246219705,-84.28373362186234,-84.26307537059527,-84.14151806416804,-83.91761075864153,-83.59002425188335,-83.15755398161602,-82.61912277553554,-81.97378344924581,-81.22072124791653,-80.35925612776074,-79.38884487362581,-78.30908304917446,-77.11970677635148,-75.82059434102145,-74.41176762188753,-72.8933933400051,-71.26578412642701,-69.52939940573613,-67.68484609344274,-65.73287910545532,-63.67440167805648,-61.51046549705488,-59.24227063501192,-56.871165295683014,-54.398645365046505,-51.826353768537146,-49.156079634338276,-46.389757262829676,-43.52946490253047,-40.577423333116954,-37.53599425634108,-34.407678495912954,-31.195114007654684,-27.901073701474104,-24.528463076942604,-21.080317674502357,-17.559800344563367,-13.9701983369826,-10.314920213651416,-6.597492587145679,-2.821556688617707,1.0091352316667688,4.890723667524863,8.819245779947586,12.790639476037512,16.800747639536013,20.845322508480375,24.92003019533385,29.020455344737485,33.14210592384549,37.28041814002332,41.43076148051111,45.588443868483765,49.74871692977576,53.90678136437859,58.05779241666896,62.19686543817961,66.31908153658823,70.41949330446683,74.49313062121276,78.5350065214647,82.54012312319995,86.50347760860767,90.42006825074027,94.28490047886015,98.09299297532368,101.83938379677369,105.51913651235422,109.12734635160824,112.65914635467678,116.10971351738289,119.47427492375992,122.74811385856424,125.92657589230689,129.0050749313372,131.97909922552336,134.84421732608996,137.59608398620398,140.23044599693284,142.7431479512459,145.1301379287828,147.38747309417525,149.51132520178066,151.49798599976444,153.3438725265573,155.04553229281052,156.5996483420784,158.0030441835691,159.25268859042797,160.34570025714862,161.27935230984306,162.05107666324906,162.65846821850653,163.09928889589543,163.37147149689685,163.47312339011268,163.40253001576536,163.15815820368468,162.7386592998883,162.14287209706174,161.36982556445645,160.41874137293487,159.28903621111553,157.9803238887948,156.4924172240548,154.82532971070177,152.97927696292197,150.95467793428563,148.75215590848273,146.3725392594223,143.81686197859173,141.0863639678276,138.18249109591522,135.1068950177028,131.86143275468248,128.44816603626327,124.86936040123598,121.12748405920144,117.22520651201256,113.16539693555904,108.95112232249717,104.58564538680957,100.07242223135644,95.4150997798585,90.61751297502805,85.68368174484242,80.61780773922979,75.42427083970922,70.10762544480028,64.67259653428722,59.12407551568868,53.46711585655047,47.706928506437414,41.848877112760675,35.89847303482994,29.861370160769802,23.743359532185487,17.55036378170541,11.288431388761696,4.963730759204069,-1.4174558654349152,-7.848738662983782,-14.323626643731801,-20.835534215125264,-27.37778789139758,-33.94363314127887,-40.526241366741885,-47.11871700556125,-53.71410475029247,-60.305396876108986,-66.88554066978216,-73.44744595193853,-79.98399268458964,-86.48803865579487,-92.95242723319934,-99.3699951780704,-105.73358051135206,-112.03603042316263,-118.27020921707077,-124.42900628040935,-130.50534407181925,-136.4921861171545,-142.38254500483492,-148.16949037169098,-153.8461568703176,-159.40575210893525,-164.84156455474823,-170.14697139178926,-175.3154463242538,-180.34056731634521,-185.21602425968953,-189.9356265594119,-194.49331063002768,-198.88314729235555,-203.09934906273693,-207.13627732592644,-210.98844938310913,-214.65054536660486,-218.11741501292778,-221.38408428599536,-224.44576184240748,-227.29784533086047,-229.93592751790942,-232.35580223245037,-234.5534701214636,-236.52514420973668,-238.26725525647046,-239.77645690186762,-241.04963059700282,-242.08389031049222,-242.87658700568832,-243.4253128823654,-243.72790537708315,-243.7824509166696,-243.58728841950472,-243.1410125395459,-242.44247664829865,-241.49079555020364,-240.28534792718753,-238.82577850840386,-237.11199996147602,-235.1441945018486,-232.92281521714736,-230.44858710374422,-227.7225078130395,-224.74584810527043,-221.5201520089757,-218.04723668455574,-214.32919199069124,-210.3683797527017,-206.1674327322475,-201.72925329810622,-197.05701179808113,-192.15414463242593,-187.02435202949889,-181.67159552468786,-176.10009514398178,-170.31432629388524,-164.31901635970905,-158.1191410145899,-151.71992024192434,-145.12681407422525,-138.34551805172612,-131.3819584043864,-124.24228696126198,-116.9328757915209,-109.46031158169775,-101.83138975407986,-94.05310833142885,-86.13266155353054,-78.07743325136865,-69.89498998499721,-61.59307395147131,-53.17959566947501,-44.662626447552775,-36.05039064311276,-27.351257719630297,-18.573734109728832,-9.726454892053843,-0.8181752900988788,8.142237998644287,17.145815634606613,26.183494620330166,35.24612742791524,44.324491260524326,53.4092974385619,62.491200900973794,71.56080981194563,80.60869526313046,89.62540106138842,98.6014535918898,107.52737174631035,116.39367690573872,125.19090296780871,133.9096064074865,142.5403763608583,151.0738447212009,159.50069623655745,167.81167859799982,175.99761250772212,184.04940171609206,191.95804301677268,199.71463618903385,207.31039387638185,214.7366513906628,221.98487643083422,229.04667870564407,235.9138194495225,242.57822082105912,249.03197517352646,255.26735418700244,261.2768178517578,267.053023292686,272.58883342469153,277.8773254290867,282.91179904120634,287.68578463960904,292.1930511274112,296.4276135964864,300.3837407654563,304.055962182609,307.4390751850957,310.5281516059856,313.31854422099434,315.80589292694987,317.9861306443132,319.8554889363381,321.0892984210148,322.00387294390293,321.11599884104226,315.1944638130409,308.5366335465428,309.5907882323645,322.3538247405403,342.09721319879503,359.14305536991634,366.1423345935678,362.1338960315451,351.1129753091666,338.0570654846713,325.9356564089173,315.38056675026957,305.919723205636,297.2270352862928,289.4710386438722,282.96919528279994,277.77169230662463,273.5331017535401,269.7021248557178,265.7617906038958,261.3783376772972,256.3977825763131,250.77948716071822,244.53325372215477,237.6911925455793,230.3073992494681,222.45522363042875,214.22089981228908,205.68730245290777,196.9234854517279,187.9788909896893,178.88511287359304,169.66048136543395,160.31440073106972,150.85149279816898,141.27363382776224,131.5823081587747,121.7795361077654,111.86894071748654,101.85589781723469,91.74749756435224,81.55230842974238,71.27995430345874,60.94091298304595,50.54611751380115,40.10681511122704,29.634300666670782,19.13984111684767,8.6345880448355,-1.8704420123272054,-12.36436031165503,-22.836436455909432,-33.27608725697232,-43.67291486631447,-54.01669006066342,-64.29735747977836,-74.50500760876957,-84.62985908773237,-94.66224293412839,-104.59258953382853,-114.41143075102653,-124.10939481759573,-133.67721363853298,-143.10572229246935,-152.38586753056433,-161.50871472757638,-170.4654592592026,-179.24743916810445,-187.84614624153414,-196.25323793683546,-204.460545361863,-212.46008239859103,-220.24405256678347,-227.80485732122534,-235.13510387951953,-242.22761348767992,-249.07542976026548,-255.67182648841325,-262.01031578696796,-268.0846556366921,-273.888857843437,-279.41719552453577,-284.6642105968767,-289.624720744278,-294.29382592072756,-298.6669143862041,-302.7396681728673,-306.5080682390013,-309.9683991232897,-313.1172533098339,-315.9515351235187,-318.4684642560024,-320.66557885571774,-322.5407382133103,-324.0921250732592,-325.318247547555,-326.2179406807719,-326.79036761483815,-327.03502039130933,-326.95172035947394,-326.5406182187567,-325.8021936961547,-324.73725487171333,-323.3469371657919,-321.632701985064,-319.5963350417648,-317.23994434043095,-314.5659578460744,-311.5771208339651,-308.27649293120714,-304.66744485407236,-300.7536548453744,-296.53910481711176,-292.0280762001294,-287.2251455073632,-282.1351796134354,-276.76333075745083,-271.11503127298323,-265.1959880509128,-259.01217673982075,-252.56983568878556,-245.87545963821339,-238.93579316383838,-231.75782388029415,-224.3487754099728,-216.71610012380876,-208.86747166024756,-200.81077722920213,-192.55410970795413,-184.10575953623868,-175.4742064181834,-166.66811083884642,-157.69630540353876,-148.5677860081339,-139.29170284893223,-129.87735128075332,-120.33416253220987,-110.6716942872952,-100.89962114259066,-91.02772494958604,-81.06588505171062,-71.024068425832,-60.91231973806461,-50.74075132386182,-40.51953310243988,-30.25888243567114,-19.96905394163752,-9.660329273082944,0.6569931289642882,10.972608296112597,21.27621499777339,31.557526000279772,41.8062783096744,52.012243387919376,62.16523733231109,72.25513100793371,82.27186012305141,92.20543523740177,102.04595169343445,111.7835994606245,121.40867288309347,130.9115803208717,140.28285367525535,149.51315778883145,158.59329971088655,167.51423781904313,176.26709078812718,184.8431463974225,193.23387016763627,201.43091381906797,209.42612354266151,217.21154807580342,224.7794465749295,232.1222962772013,239.23279994372882,246.10389307702815,252.72875090563122,259.1007951289926,265.2137004160761,271.0614006512482,276.63809492135584,281.93825323811785,286.9566219902271,291.6882291198198,296.1283890182468,300.2727071363554,304.1170843047735,307.65772075997324,310.89111987218246,313.8140915715075,316.42375546892833,318.71754366912927,320.69320327243116,322.3487985634027,323.6827128840341,324.69365018966954,325.3806362862135,325.743019747433,325.7804725115001,325.49299015624126,324.88089185286185,323.9448199982559,322.685739526306,321.10493689891496,319.2040187778167,316.9849103785344,314.44985350816773,311.60140428900445,308.44243057026324,304.97610903058177,301.20592197417204,297.1356538238672,292.7693873145845,288.1114993910258,283.1666568137277,277.93981147786633,272.43619544950184,266.6613157242269,260.6209487134635,254.3211344639146,247.7681706159452,240.96860610692633,233.92923462582175,226.65708782555117,219.15942829989632,211.44374233195208,203.51773242134854,195.38930959769118,187.0665855278743,178.55786442512823,169.87163476785466,161.01656083649448,152.00147407684958,142.83536429845248,133.52737071673977,124.0867728479384,114.52298126572182,104.84552822882371,95.06405818892952,85.1883181882808,75.2281481565345,65.19347111652155,55.09428330863507,44.94064424366107,34.74266669393163,24.51050663274254,14.25435313202673,3.984418228316206,-6.289073232946526,-16.55589376466338,-26.805823406810138,-37.02865988654549,-47.214228759616915,-57.35239352299777,-67.43306568872221,-77.44621480892116,-87.38187844211447,-97.23017205087218,-106.98129882102722,-116.62555939270041,-126.1533614934889,-135.5552294642635,-144.82181366813185,-153.9438997732384,-162.91241790019905,-171.7184516251053,-180.3532468291716,-188.8082203862615,-197.07496867967876,-205.14527593979088,-213.01112239422122,-220.6646922225414,-228.09838130758297,-235.3048047756944,-242.2768043184787,-249.00745528876277,-255.49007356377643,-261.7182221687499,-267.68571765437514,-273.3866362218258,-278.815319589276,-283.9663805941182,-288.8347085253451,-293.41547418082314,-297.70413464446466,-301.69643777858084,-305.388426426981,-308.7764423246707,-311.8571297102963,-314.62743863777246,-317.08462798383374,-319.22626814855187,-321.0502434461623,-322.5547541838557,-323.73831842649696,-324.5997734455469,-325.13827685077496,-325.3533074036687,-325.2446655117568,-324.8124734033849,-324.0571749827952,-322.97953536568383,-321.5806400957201,-319.8618940428375,-317.8250199844117,-315.4720568707629,-312.80535777673157,-309.82758754138536,-306.54172009822867,-302.95103549858726,-299.05911663115285,-294.8698456409641,-290.3874000514091,-285.6162485931192,-280.5611467439221,-275.22713198430256,-269.6195187731071,-263.7438932485005,-257.6061076594593,-251.21227453335104,-244.56876058540882,-237.68218037617385,-230.5593897232173,-223.20747887370405,-215.63376544459558,-207.84578713751847,-199.85129423555207,-191.65824188939484,-183.27478220059245,-174.70925610969664,-165.9701850974276,-157.0662627070921,-148.0063458966884,-138.79944622929384,-129.45472091050055,-119.98146368180218,-110.3890955789889,-100.68715556473266,-90.88529104467581,-80.99324827644537,-71.02086268112545,-60.97804906681262,-50.87479177396563,-40.721134752346,-30.527171579397944,-20.303035429989645,-10.058889007473772,0.1950855539324356,10.44869680541872,20.6917541084601,30.914077758248503,41.10550909562799,51.2559205974933,61.35522593563522,71.39338999404846,81.36043883475853,91.24646960227653,101.04166035685395,110.73627982677729,120.32069707002711,129.78539103571381,139.12096001580522,148.31813097776828,157.36776876886805,166.26088518299326,174.98864788101602,183.54238915583954,191.91361453344112,200.09401120138318,208.0754562564337,215.85002476312133,223.40999761523602,230.74786919248336,237.85635480470276,244.72839791627626,251.35717714356437,257.7361130184414,263.8588745112278,269.7193853065589,275.3118298259757,280.6306589912758,285.6705957229159,290.42664016802706,294.89407465286797,299.0684683548162,302.94568168927833,306.5218704071795,309.79348939898455,312.7572962014903,315.4103542039281,317.7500355502099,319.77402373445716,321.480315887252,322.86722475036555,323.93338033801183,324.6777312830083,325.09954586651753],"type":"scatter","name":"source1_u_a","yaxis":"y2","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]}],{"xaxis":{"title":{"text":"Time in Seconds"}},"template":{"layout":{"coloraxis":{"colorbar":{"ticks":"","outlinewidth":0}},"xaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"hovermode":"closest","paper_bgcolor":"white","geo":{"showlakes":true,"showland":true,"landcolor":"#E5ECF6","bgcolor":"white","subunitcolor":"white","lakecolor":"white"},"colorscale":{"sequential":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]],"diverging":[[0,"#8e0152"],[0.1,"#c51b7d"],[0.2,"#de77ae"],[0.3,"#f1b6da"],[0.4,"#fde0ef"],[0.5,"#f7f7f7"],[0.6,"#e6f5d0"],[0.7,"#b8e186"],[0.8,"#7fbc41"],[0.9,"#4d9221"],[1,"#276419"]],"sequentialminus":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]},"yaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"shapedefaults":{"line":{"color":"#2a3f5f"}},"hoverlabel":{"align":"left"},"mapbox":{"style":"light"},"polar":{"angularaxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","radialaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"autotypenumbers":"strict","font":{"color":"#2a3f5f"},"ternary":{"baxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","caxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"aaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"annotationdefaults":{"arrowhead":0,"arrowwidth":1,"arrowcolor":"#2a3f5f"},"plot_bgcolor":"#E5ECF6","title":{"x":0.05},"scene":{"xaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"zaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"yaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"}},"colorway":["#636efa","#EF553B","#00cc96","#ab63fa","#FFA15A","#19d3f3","#FF6692","#B6E880","#FF97FF","#FECB52"]},"data":{"barpolar":[{"type":"barpolar","marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"carpet":[{"aaxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"},"type":"carpet","baxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"}}],"scatterpolar":[{"type":"scatterpolar","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"parcoords":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"parcoords"}],"scatter":[{"type":"scatter","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2dcontour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2dcontour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattercarpet":[{"type":"scattercarpet","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"mesh3d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"mesh3d"}],"surface":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"surface","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattermapbox":[{"type":"scattermapbox","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergeo":[{"type":"scattergeo","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram":[{"type":"histogram","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"pie":[{"type":"pie","automargin":true}],"choropleth":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"choropleth"}],"heatmapgl":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmapgl","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"bar":[{"type":"bar","error_y":{"color":"#2a3f5f"},"error_x":{"color":"#2a3f5f"},"marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"heatmap":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmap","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contourcarpet":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contourcarpet"}],"table":[{"type":"table","header":{"line":{"color":"white"},"fill":{"color":"#C8D4E3"}},"cells":{"line":{"color":"white"},"fill":{"color":"#EBF0F8"}}}],"scatter3d":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"scatter3d","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergl":[{"type":"scattergl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2d","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scatterternary":[{"type":"scatterternary","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scatterpolargl":[{"type":"scatterpolargl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}]}},"legend":{"yanchor":"bottom","xanchor":"right","y":1.02,"orientation":"h","x":1},"margin":{"l":100,"b":80,"pad":10,"r":80,"t":100},"plot_bgcolor":"#f1f3f7","yaxis":{"title":{"text":"State values"}},"yaxis2":{"titlefont":{"color":"orange"},"overlaying":"y","title":"Action values","side":"right"},"height":550,"width":800},{"showLink":false,"editable":false,"responsive":true,"staticPlot":false,"scrollZoom":true});
});





</script>
```




_______________________________________________________________________________
### High-Level Rendering


```julia
RenderHookResults(hook = hook, 
                    states_to_plot  = [], 
                    actions_to_plot = [],  
                    v_mag_poc       = [1],   # Scaled L₂ norm in αβγ coordinates [V]
                    power_p_inv     = [2],   # Real power [Watts]
                    power_q_inv     = [2],   # Imaginary power [VAi]
                    power_p_poc     = [1],   # Real power [Watts]
                    power_q_poc     = [1],   # Imaginary power [VAi]
                    angles          = [1 2], # Relative angle [degrees]
                    freq            = [1 2], # Angular velocity [Hz]
                    )
```


```@raw html
<div
id = ZvGpqSu4vIRy > </div>
<script>
gd = 'ZvGpqSu4vIRy'
require(['plotly'], function(plotly) {
























plotly.newPlot(gd,[{"mode":"lines","y":[0.0,0.0,0.0,-5.208458348445778e-6,-0.0003849424402872666,-0.005486919743497671,-0.03412654982680085,-0.12593759700708904,-0.3248795727912727,-0.6505151335188943,-1.0885326791414724,-1.6064557244914708,-2.1725108759071645,-2.7619195361813627,-3.3542879314404757,-3.931701474192759,-4.480879693009969,-4.996224809268838,-5.480108868467992,-5.939995203756103,-6.384560370753386,-6.820953590763217,-7.253822182502464,-7.685548915256506,-8.116848750634142,-8.547254271034491,-8.975422156283905,-9.39939165457161,-9.816881797199887,-10.225610368791145,-10.623567355307589,-11.009186408845729,-11.381411655590304,-11.739684481556646,-12.083887661167346,-12.414269393986125,-12.731359441045557,-13.035882152718532,-13.328670981813795,-13.610591545836039,-13.882478585003,-14.145090465722156,-14.39908010831521,-14.644980027353988,-14.88319850320401,-15.114025182872506,-15.337644882198326,-15.554157918651434,-15.763604467328625,-15.965989631803689,-16.16130624494494,-16.349553253905157,-16.530748797500948,-16.704938004901273,-16.872196110376382,-17.032627705542033,-17.18636305522069,-17.333552443287598,-17.47435953767751,-17.608954655554193,-17.737508656966966,-17.86018795042809,-17.97715088671188,-18.088545621922293,-18.19450936650042,-18.29516881185765,-18.390641407345502,-18.481037133070004,-18.566460418692902,-18.64701193960608,-18.72278632842192,-18.793844512534072,-18.860263058039344,-18.922139919201143,-18.979580871308237,-19.032691908172353,-19.0815795841811,-19.126354281696926,-19.16713175683463,-19.20403183883728,-19.23717587022513,-19.266684618703596,-19.292677216241895,-19.315270769739477,-19.334580131824307,-19.350717607871644,-19.363792649563653,-19.373911663584863,-19.38117799262784,-19.38569203189774,-19.387551400388787,-19.386851097942305,-19.383683618364575,-19.378139022344325,-19.370304988817324,-19.360266858622907,-19.3481076727794,-19.33390819911357,-19.317746940983827,-19.299700127844794,-19.279841693577726,-19.258243251147356,-19.23497407028928,-19.21010106157271,-19.183688767404036,-19.15579935986904,-19.126492645750748,-19.09582607956572,-19.063854785276384,-19.030631586379634,-18.996207043138426,-18.960629495118493,-18.923945107240108,-18.88619791790392,-18.84742988825724,-18.807680951988814,-18.766989065283102,-18.725390256637063,-18.682918676358437,-18.63960664572211,-18.59548470584373,-18.550581666503646,-18.504924655137117,-18.45853916621114,-18.411449111112802,-18.36367686859309,-18.315243335741258,-18.266167979404514,-18.216468887949787,-18.16616282324513,-18.115265272719114,-18.063790501345537,-18.01175160341353,-17.95916055393363,-17.906028259550453,-17.85236460886493,-17.798178522075816,-17.743477999874273,-17.688270171542854,-17.63256134221524,-17.57635703926347,-17.519662057788835,-17.462480505184587,-17.404815844754864,-17.346670938371645,-17.2880480881417,-17.228949077079974,-17.16937520874489,-17.109327345853956,-17.048805947816177,-16.98781110719611,-16.926342585091948,-16.864399845403106,-16.80198208797349,-16.739088280629122,-16.675717190084182,-16.611867411706736,-16.54753739816242,-16.482725486931244,-16.417429926703605,-16.35164890266594,-16.285380560668294,-16.218623030308226,-16.15137444692803,-16.08363297254313,-16.015396815703582,-15.94666425032586,-15.877433633498114,-15.807703422267053,-15.737472189445825,-15.666738638440705,-15.595501617131447,-15.523760130805286,-15.451513354188563,-15.378760642585874,-15.305501542140291,-15.231735799237796,-15.157463369113813,-15.082684423614552,-15.007399358205546,-14.931608798209481,-14.85531360430872,-14.778514877340442,-14.701213962391137,-14.623412452241709,-14.545112190159301,-14.46631527206344,-14.387024048117325,-14.307241123731032,-14.226969360018245,-14.146211873730593,-14.06497203668151,-13.983253474697136,-13.901060066095411,-13.81839593974324,-13.735265472687203,-13.651673287390636,-13.567624248601732,-13.48312345985606,-13.39817625966581,-13.312788217377467,-13.226965128742854,-13.140713011222882,-13.054038099026808,-12.96694683788044,-12.879445879654838,-12.791542076721296,-12.70324247612625,-12.614554313636347,-12.525485007592529,-12.436042152634485,-12.346233513305378,-12.256067017539856,-12.165550750064341,-12.074692945708733,-11.983501982656293,-11.891986375634389,-11.800154769053222,-11.708015930116574,-11.615578741926512,-11.522852196552293,-11.429845388102933,-11.336567505839184,-11.243027827277558,-11.14923571132121,-11.055200591437455,-10.960931968887593,-10.866439405990747,-10.771732519462773,-10.676820973821464,-10.581714474863645,-10.486422763217064,-10.390955607992026,-10.295322800509297,-10.199534148137275,-10.103599468224047,-10.007528582151764,-9.911331309477617,-9.815017462202235,-9.71859683915153,-9.622079220488757,-9.525474362337718,-9.42879199155459,-9.332041800600166,-9.235233442566951,-9.138376526324379,-9.041480611814663,-8.944555205461327,-8.847609755741914,-8.750653648887806,-8.653696204708517,-8.556746672577255,-8.45981422755538,-8.362907966629589,-8.266036905127272,-8.169209973236903,-8.072436012715382,-7.975723773662612,-7.879081911493351,-7.7825189840194895,-7.686043448693539,-7.589663659950233,-7.4933878667032445,-7.3972242099745955,-7.301180720651942,-7.205265317369012,-7.109485804532966,-7.013849870450628,-6.918365085595923,-6.8230389010016665,-6.727878646742596,-6.632891530598471,-6.538084636763756,-6.443464924687589,-6.349039228081864,-6.254814253952617,-6.160796581812519,-6.066992662939668,-5.973408819778626,-5.88005124542488,-5.7869260032020975,-5.6940390263401355,-5.601396117756707,-5.509002949914162,-5.416865064742808,-5.324987873715656,-5.23337665792306,-5.14203656828127,-5.05097262582035,-4.960189722006675,-4.869692619160247,-4.779485950964499,-4.689574222997834,-4.599961813372346,-4.510652973400653,-4.421651828342775,-4.332962378225744,-4.244588498670467,-4.156533941807069,-4.068802337236832,-3.9813971930634793,-3.8943218969092044,-3.8075797170323575,-3.7211738034437243,-3.6351071891276034,-3.549382791238493,-3.464003412341551,-3.3789717417129026,-3.2942903566370276,-3.209961723768103,-3.1259882005188757,-3.0423720364286737,-2.9591153746174967,-2.8762202532142567,-2.7936886068214903,-2.711522268006661,-2.6297229688195216,-2.5482923422705426,-2.4672319238920295,-2.3865431532881587,-2.306227375655908,-2.226285843397932,-2.1467197176364152,-2.067530069818673,-1.988717883321101,-1.9102840550021938,-1.8322293968183203,-1.7545546373931984,-1.6772604236208508,-1.6003473222569369,-1.5238158215189515,-1.4476663326389598,-1.371899191475336,-1.2965146600753883,-1.2215129282656108,-1.146894115192481,-1.0726582708877856,-0.9988053778239419,-0.9253353524417491,-0.8522480467016749,-0.7795432495653642,-0.7072206885335959,-0.6352800311523197,-0.563720886459663,-0.4925428064677533,-0.4217452876287382,-0.3513277722764201,-0.28128965003679696,-0.21163025927619117,-0.142348888442271,-0.07344477749775535,-0.004917119242655588,0.06323493931516566,0.13101229563477101,0.1984158903033606,0.26544670580723206,0.3321057651957702,0.39839413084844466,0.4643129032294553,0.5298632196566615,0.5950462530966263,0.6598632110232074,0.7243153341802895,0.7884038954984192,0.8521301989334518,0.9154955783723153,0.97850139653886,1.041149043914686,1.1034399376898705,1.1653755207603922,1.2269572607015675,1.2881866487659295,1.349065198888752,1.4095944467817958,1.4697759489733926,1.5296112818903111,1.5891020409847556,1.6482498398264056,1.7070563092433986,1.7655230964965085,1.8236518644702073,1.8814442908256979,1.9389020672556967,1.9960268986910137,2.0528205025634634,2.109284608073949,2.165420955476975,2.221231295361008,2.2767173880091036,2.331881002697884,2.3867239170807864,2.4412479165380034,2.4954547935546145,2.549346347156046,2.6029243822980277,2.656190709288609,2.7091471433061827,2.761795503768326,2.8141376138761416,2.8661753000980923,2.9179103916486255,-407.5240988385623,35879.36053655167,58799.06109537624,68571.5982687166,73599.28436988016,76905.90537967451,76361.76704581938,70358.68089198307,61267.587102454054,52815.820495018,46603.84377400979,42369.13911405766,39522.202323411344,37737.008025536445,36864.38815415894,36781.762447468034,37344.720960719686,38379.530063881575,39682.748146736296,41049.27448076126,42318.59790841718,43399.98979537912,44264.42269773051,44921.88248581595,45402.42368983129,45745.26695969655,45992.16813948542,46181.86516450977,46345.4348089584,46503.68267542235,46667.1331066307,46838.15092646556,47014.09151877427,47190.29227384684,47362.14435769033,47526.088174619115,47679.82963770525,47822.19012199514,47952.86803958018,48072.20507070256,48180.95301733166,48280.039908611805,48370.36788721002,48452.68393558925,48527.53957177838,48595.321839193384,48656.32024809279,48710.797592363684,48759.04601419493,48801.422457078996,48838.363353383975,48870.380829082635,48898.04357227404,48921.947747394435,48942.68477229405,48960.812602097365,48976.83507590517,48991.1906259862,49004.24933050588,49016.315859238035,49027.636075893664,49038.40550076953,49048.77849769189,49058.8772933158,49068.80003315152,49078.62726120128,49088.42644026386,49098.25460673974,49108.15954439008,49118.18009960304,49128.346185319395,49138.678861640634,49149.19065830282,49159.88614110229,49170.76267201034,49181.81129356259,49193.01770883115,49204.3633189413,49215.82628270425,49227.382537374644,49239.00672011979,49250.67294516522,49262.35541877287,49274.028907514046,49285.66908788194,49297.25281211231,49308.75831262989,49320.16535928156,49331.45537409725,49342.611506769994,49353.618675363134,49364.46357765676,49375.13467950691,49385.622183755804,49395.91798203701,49406.015589732626,49415.91006505113,49425.59791407022,49435.07698498833,49444.34635558683,49453.40621756417,49462.25776085556,49470.903059906545,49479.34496334438,49487.58698794475,49495.63321771399,49503.488208761206,49511.156900493814,49518.6445334819,49525.95657407286,49533.09864570424,49540.07646672345,49546.895794532,49553.56237586058,49560.081903006794,49566.45997584825,49572.702069406194,49578.81350668329,49584.79943644266,49590.66481556262,49596.41439557884,49602.052713029974,49607.584083227266,49613.01259708388,49618.34212064782,49623.57629699202,49628.71855012583,49633.77209060845,49638.73992256687,49643.624851847286,49648.429495057455,49653.15628928254,49657.807502277836,49662.385242962424,49666.8914720534,49671.328012699385,49675.69656099071,49679.998696243136,49684.235890969685,49688.409520472036,49692.520871995715,49696.571153405246,49700.56150134779,49704.492988881786,49708.36663255903,49712.183398956506,49715.944210661866,49719.64995172327,49723.30147257827,49726.8995944809,49730.44511344821,49733.938803752266,49737.38142098296,49740.773704710846,49744.11638077952,49747.41016325612,49750.65575606999,49753.85385436847,49757.00514561744,49760.11031047392,49763.170023457074,49766.18495344155,49769.15576399802,49772.0831136014,49774.96765572719,49777.81003885472,49780.61090639435,49783.37089655353,49786.090642155716,49788.77077042466,49791.4119027444,49794.01465440545,49796.57963434421,49799.10744488316,49801.59868147789,49804.05393247549,49806.47377888794,49808.85879418456,49811.20954410462,49813.526586492604,49815.810471156474,49818.06173975008,49820.28092567928,49822.46855403172,49824.625141529,49826.75119650159,49828.847218883886,49830.91370022907,49832.951123742,49834.95996432889,49836.9406886619,49838.89375525727,49840.81961456545,49842.718709071945,49844.591473406865,49846.43833446183,49848.259711514016,49850.05601635419,49851.82765341898,49853.57501992626,49855.29850601158,49856.99849486594,49858.675362873546,49860.329479748834,49861.96120867232,49863.57090642484,49865.158923519244,49866.72560432967,49868.27128721799,49869.79630465714,49871.3009833509,49872.78564435046,49874.25060316756,49875.696169883886,49877.122649257115,49878.53034082328,49879.91953899599,49881.29053316213,49882.64360777436,49883.979042440544,49885.29711201035,49886.59808665887,49887.88223196762,49889.14980900306,49890.401074392794,49891.63628039963,49892.85567499321,49894.059501920376,49895.24800077331,49896.421407056,49897.5799522496,49898.72386387601,49899.853365560426,49900.968677092744,49902.070014487625,49903.15759004359,49904.231612401796,49905.292286603115,49906.33981414496,49907.37439303715,49908.396217857524,49909.40547980603,49910.40236675935,49911.38706332391,49912.35975088891,49913.32060767874,49914.269808804806,49915.207526316866,49916.133929253665,49917.04918369359,49917.95345280424,49918.846896891795,49919.72967344994,49920.60193720806,49921.463840179276,49922.315531707405,49923.15715851403,49923.988864744664,49924.81079201453,49925.62307945349,49926.425863751,49927.21927920005,49928.003457740706,49928.77852900298,49929.544620349276,49930.30185691628,49931.05036165602,49931.79025537672,49932.5216567825,49933.24468251327,49933.959447183275,49934.66606341953,49935.36464189942,49936.05529138787,49936.738118773894,49937.41322910635,49938.08072562943,49938.74070981722,49939.39328140805,49940.03853843793,49940.6765772738,49941.307492645625,49941.93137767851,49942.54832392385,49943.15842139036,49943.761758573884,49944.358422487436,49944.94849869008,49945.53207131559,49946.10922310058,49946.68003541206,49947.24458827423,49947.80296039532,49948.35522919349,49948.90147082229,49949.44176019588,49949.97617101334,49950.504775783105,49951.0276458463,49951.54485139999,49952.05646152003,49952.562544183136,49953.063166288826,49953.558393680854,49954.04829116809,49954.53292254507,49955.01235061206,49955.48663719522,49955.95584316534,49956.42002845713,49956.87925208772,49957.333572174655,49957.783045954,49958.22772979758,49958.667679229955,49959.102948945365,49959.53359282388,49959.959663947426,49960.381214615656,49960.79829636108,49961.210959964264,49961.6192554682,49962.02323219314,49962.42293875036,49962.818423055985,49963.20973234448,49963.59691318182,49963.98001147845,49964.359072501815,49964.73414088868,49965.10526065718,49965.472475218776,49965.83582738965,49966.19535940196,49966.551112915025,49966.90312902594,49967.25144828025,49967.596110682076,49967.93715570439,49968.27462229901,49968.60854890578,49968.93897346225,49969.26593341299,49969.58946571847,49969.90960686369,49970.226392867,49970.53985928854,49970.8500412382,49971.1569733839,49971.46068995923,49971.7612247714,49972.05861120816,49972.35288224591,49972.64407045627,49972.932208013175,49973.2173266999,49973.49945791556,49973.77863268176,49974.05488164889,49974.32823510229,49974.59872296858,49974.86637482142,49975.131219887306,49975.39328705143,49975.65260486314,49975.9092015414,49976.16310497992,49976.414342752774,49976.66294211905,49976.90893002812,49977.15233312416,49977.393177751284,49977.63148995788,49977.867295501375,49978.1006198525,49978.331488199714,49978.55992545331,49978.78595624985,49979.009604955965,49979.230895672394,49979.44985223802,49979.66649823336,49979.880856984586,49980.09295156705,49980.3028048088,49980.5104392941,49980.71587736684,49980.91914113401,49981.120252468594,49981.319233013324,49981.5161041833,49981.71088716939,49981.90360294107,49982.09427224934,49982.28291562982,49982.46955340537,49982.6542056891,49982.83689238662,49983.01763319928,49983.19644762625],"type":"scatter","name":"source2_p_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.0,0.0005350364853775977,0.009470565083281353,0.05897817451788768,0.20650179167543556,0.5096510252868371,1.0018861957663594,1.6913222695239232,2.5720235999076064,3.6242075721775704,4.81984125061532,6.1179645143057595,7.477460715725596,8.859709837227914,10.238587229422857,11.600582871944939,12.942157807560628,14.268217586220864,15.584624089698455,16.89996498766208,18.219639067664684,19.54871204208681,20.88866872418975,22.23900313302417,23.597082615880254,24.9584415627473,26.318659431518153,27.672594255717875,29.01637691731225,30.346276252157793,31.65992687287994,32.955623636789994,34.23267842512673,35.49126074617923,36.73207146218659,37.956461715608796,39.16585492719359,40.361941051452995,41.54623307576963,42.72020517630721,43.8850907051291,45.04190461131435,46.19142616931763,47.334157903996996,48.47041400531269,49.60029211244897,50.72378675210739,51.84079302469303,52.95118593483977,54.05483978665435,55.15166167063178,56.241610424516956,57.32469978210751,58.40100602303776,59.47065625477886,60.533822329279964,61.59070496385996,62.641521254792174,63.686491468982226,64.7258290152113,65.75973218246422,66.78838058929071,67.81193172599536,68.83052333799513,69.8442739767579,70.85328812717768,71.85765916265822,72.8574737246358,73.8528153187372,74.8437660946151,75.83040635535612,76.812806605061,77.79103216303696,78.76515287047576,79.73525267755792,80.7014314391917,81.66380187535319,82.6224847247203,83.57760444566047,84.52928732789415,85.47765965197695,86.42284696337268,87.36497248927381,88.30415629905288,89.24051448653826,90.17415898474958,91.10519779599102,92.03373520944348,92.9598723054275,93.88370710066195,94.80533491331326,95.72484845789468,96.64233808168704,97.55789188576139,98.4715958320119,99.38353382809244,100.29378769901784,101.20243720089483,102.10955992845757,103.01523128861135,103.91952441745502,104.82251014075723,105.72425692405584,106.62483083434687,107.52429552514711,108.42271221827694,109.32013972317439,110.21663445089256,111.11225045835548,112.00703948917766,112.90105102805926,113.79433235523183,114.68692860138401,115.57888280578963,116.47023596894834,117.36102710880489,118.25129331064639,119.14106977884904,120.03038988454607,120.91928521368337,121.80778561429597,122.6959192436969,123.5837126171192,124.4711906562317,125.35837673963763,126.24529275337889,127.13195914290927,128.01839496532696,128.90461794222924,129.7906445128706,130.67648988716513,131.56216809869954,132.44769205707695,133.33307359988723,134.21832354378998,135.10345173493448,135.98846709843437,136.8733776870087,137.75819072866895,138.64291267348054,139.52754923937135,140.41210545697064,141.29658571350498,142.1809937956917,143.0653329316757,143.94960583193708,144.83381472917046,145.7179614171451,146.60204728845713,147.48607337123997,148.37004036471217,149.25394867367442,150.13779844179382,151.02158958382128,151.9053218165931,152.78899468892092,153.67260761033904,154.55615987864417,155.43965070639834,156.32307924617479,157.20644461480026,158.08974591639705,158.9729822643714,159.85615280233006,160.73925672387452,161.62229329142255,162.5052618539095,163.38816186353978,164.27099289148632,165.15375464265617,166.0364469694419,166.91906988459164,167.8016235730941,168.68410840322042,169.56652493664217,170.44887393771853,171.33115638195864,172.21337346362895,173.0955266026216,173.97761745055107,174.85964789607624,175.7416200695476,176.6235363469322,177.50539935310024,178.38721196441378,179.26897731079293,180.15069877703382,181.03238000374068,181.91402488751112,182.7956375807671,183.67722249094868,184.5587842793219,185.4403278592726,186.32185839416258,187.2033812948182,188.08490221654444,188.9664270558834,189.84796194686788,190.72951325708968,191.61108758336994,192.49269174713845,193.37433278961524,194.25601796659708,195.13775474320784,196.01955078819896,196.90141396824166,197.7833523419588,198.66537415371315,199.54748782744775,200.4297019601076,201.31202531524855,202.19446681624507,203.07703553967755,203.95974070837843,204.8425916846782,205.72559796333582,206.60876916464093,207.4921150273732,208.37564540175237,209.25937024243706,210.14329960145596,211.0274436212144,211.9118125274996,212.79641662244148,213.681266277722,214.56637192755068,215.45174406195747,216.33739321996416,217.22332998292097,218.10956496793142,218.9961088212466,219.88297221195228,220.7701658254722,221.65770035749938,222.54558650770957,223.43383497385298,224.32245644577438,225.2114615996237,226.10086109225927,226.99066555555225,227.8808855911717,228.77153176506965,229.66261460250905,230.5541445829249,231.446132135085,232.33858763235918,233.23152138801817,234.12494365091186,235.0188646009541,235.91329434511505,236.8082429132452,237.70372025420733,238.5997362321673,239.49630062290544,240.39342311037083,241.29111328334324,242.18938063222157,243.0882345460292,243.98768430935337,244.88773909979523,245.7884079850509,246.68969992070527,247.59162374757528,248.49418818969383,249.39740185206182,250.30127321874488,251.2058106509662,252.11102238546255,253.01691653278442,253.92350107589039,254.83078386874593,255.73877263508746,256.6474749673148,257.5568983254436,258.46705003621196,259.37793729230566,260.2895671516287,261.201946536715,262.1150822343018,263.0289808948478,263.94364903232463,264.8590930239236,265.7753191100562,266.69233339421476,267.6101418431617,268.5287502869512,269.44816441925286,270.36838979757,271.2894318437195,272.2112958441752,273.13398695072385,274.05751018090416,274.9818704187944,275.907072415662,276.8331207907338,277.7600200321325,278.68777449758704,279.61638841564536,280.5458658863526,281.47621088261764,282.40742725103996,283.3395187132652,284.2724888669836,285.20634118729174,286.14107902786776,287.07670562226633,288.01322408532707,288.9506374144163,289.8889484909381,290.8281600816086,291.7682748400663,292.7092953081631,293.6512239176312,294.5940629914222,295.5378147453979,296.48248128972557,297.4280646305373,298.3745666714616,299.32198921522263,300.2703339652453,301.2196025272417,302.16979641083117,303.1209170312061,304.07296571064853,305.02594368029554,305.97985208166006,306.9346919683503,307.8904643076112,308.8471699821275,309.8048097913204,310.76338445345664,311.72289460677484,312.6833408114794,313.644723551178,314.60704323446,315.57030019670833,316.5344947013043,317.4996269416945,318.46569704247827,319.4327050613779,320.40065099047183,321.3695347579037,322.3393562294377,323.310115209699,324.28181144413554,325.2544446199329,326.2280143680133,327.20252026407115,328.177961830261,329.1543385365828,330.1316498020888,331.10989499661525,332.089073441756,333.069184412528,334.05022713851963,335.0322008053072,336.0151045556891,336.9989374909725,337.983698672255,338.96938712162415,339.9560018234672,340.9435417254748,341.9320057401429,342.9213927455165,343.911701586813,344.90293107703997,345.895079998573,346.8881471038253,347.88213111663845,348.8770307331258,349.872844622834,350.86957142956976,351.86720977258733,352.86575824739134,353.8652154268325,354.8655798619134,355.86685008274003,356.869024599416,357.87210190288147,358.87608046591413,359.8809587436717,360.88673517488223,361.89340818222604,362.90097617355445,363.90943754221917,364.91879066819865,365.92903391849313,366.9401656481248,367.95218420063134,368.9650879087892,369.97887509530347,370.99354407347846,372.0090931476591,373.02552061418623,374.04282476154617,375.06100387141214,376.08005621875975,377.09998007279864,378.1207736972344,379.14243535085893,380.16496328825417,-935.1901817161013,-7963.260377787479,-21822.033843288125,-40213.43231594956,-57919.55154127289,-69580.5206052429,-73178.14122455003,-70365.02794396188,-64061.63400461749,-56537.87048777833,-49006.763190044476,-42158.2458470509,-36481.10323750993,-32337.30143036707,-29846.55361167751,-28876.47651797935,-29126.966897209648,-30230.594381777988,-31840.182392050912,-33652.34888649788,-35432.31354517337,-37009.62331396888,-38286.14991824308,-39227.97095232381,-39854.10966132287,-40219.454026877655,-40393.720273064755,-40448.56596185775,-40444.56163628649,-40428.338417950785,-40429.5269165178,-40463.32399847417,-40533.00545769784,-40633.76035669651,-40756.652427520654,-40891.16907564152,-41027.849482600155,-41158.96200118665,-41279.37543311078,-41386.21427212336,-41478.665919604566,-41557.3823217289,-41623.889646997064,-41680.13129264709,-41727.97749238085,-41769.07352723989,-41804.655578048885,-41835.62362156309,-41862.5718398517,-41885.906053001316,-41905.92852868538,-41922.902805894104,-41937.11210968685,-41948.867516741346,-41958.52449377112,-41966.45982936145,-41973.05921531128,-41978.68999998057,-41983.68363151065,-41988.32241289247,-41992.832502999736,-41997.38612316172,-42002.10527105568,-42007.07107753991,-42012.33203590966,-42017.91348621428,-42023.82527676533,-42030.06795984756,-42036.6370296395,-42043.52513864327,-42050.7229117045,-42058.218871648845,-42065.99890350469,-42074.04593902959,-42082.33965297922,-42090.8565752816,-42099.570312629156,-42108.451983153085,-42117.47088393836,-42126.59517524368,-42135.79280158123,-42145.032211646576,-42154.283118740655,-42163.51695641747,-42172.70723306754,-42181.82970410406,-42190.86243893553,-42199.785869153195,-42208.58273697861,-42217.2380754109,-42225.73908877378,-42234.07505905299,-42242.237199527925,-42250.21852961879,-42258.01375168585,-42265.619128798324,-42273.032379112716,-42280.25254601089,-42287.27988100475,-42294.11570385509,-42300.76227672322,-42307.222674190416,-42313.50066498363,-42319.600600881604,-42325.52730947851,-42331.285997571045,-42336.88215604498,-42342.321477838865,-42347.60978081163,-42352.75294307519,-42357.756845201584,-42362.62732068836,-42367.37011315238,-42371.990838585145,-42376.4949544618,-42380.88773322361,-42385.17424230962,-42389.35932828072,-42393.4476059507,-42397.443450999686,-42401.35099599416,-42405.17412949733,-42408.916497670645,-42412.58150856692,-42416.17233823667,-42419.69193875917,-42423.14304742055,-42426.52819698983,-42429.84972667387,-42433.10979359197,-42436.31038461669,-42439.4533283279,-42442.540307017356,-42445.572868485004,-42448.55243759357,-42451.48032741218,-42454.357749923765,-42457.18582621892,-42459.965596137,-42462.69802732862,-42465.38402369636,-42468.02443321691,-42470.620055123145,-42473.17164646588,-42475.67992805989,-42478.145589838285,-42480.56929563687,-42482.95168743255,-42485.29338906783,-42487.59500948866,-42489.85714553289,-42492.08038430044,-42494.26530514199,-42496.41248129883,-42498.52248122794,-42500.59586964395,-42502.6332083101,-42504.635056607105,-42506.601971908975,-42508.53450979207,-42510.4332241019,-42512.298666901406,-42514.131388320304,-42515.931936325906,-42517.70085643213,-42519.4386913621,-42521.145980678666,-42522.823260393845,-42524.47106256943,-42526.089914915276,-42527.68034039591,-42529.24285684871,-42530.77797662208,-42532.28620623461,-42533.76804606106,-42535.22399004555,-42536.654525444654,-42538.06013260072,-42539.441284745866,-42540.79844783655,-42542.132080417876,-42543.442633516905,-42544.7305505635,-42545.99626733711,-42547.2402119386,-42548.46280478392,-42549.6644586194,-42550.84557855547,-42552.00656211774,-42553.14779931299,-42554.26967270862,-42555.37255752363,-42556.456821729385,-42557.522826158536,-42558.570924620915,-42559.60146402449,-42560.61478450046,-42561.61121953134,-42562.59109608048,-42563.55473472316,-42564.502449776526,-42565.43454943032,-42566.35133587466,-42567.253105427495,-42568.140148658356,-42569.01275051045,-42569.87119041906,-42570.71574242714,-42571.54667529745,-42572.36425262116,-42573.168732922946,-42573.96036976283,-42574.73941183404,-42575.50610305804,-42576.26068267571,-42577.00338533563,-42577.734441179324,-42578.45407592302,-42579.16251093781,-42579.8599633253,-42580.54664599322,-42581.22276772624,-42581.88853325683,-42582.544143332576,-42583.18979478239,-42583.82568058132,-42584.45198991289,-42585.06890823117,-42585.676617320685,-42586.27529535579,-42586.86511695845,-42587.446253255795,-42588.018871936045,-42588.58313730419,-42589.13921033676,-42589.68724873596,-42590.227406983366,-42590.75983639266,-42591.28468516255,-42591.80209842836,-42592.31221831393,-42592.81518398275,-42593.31113168834,-42593.80019482516,-42594.28250397787,-42594.75818697152,-42595.22736891994,-42595.690172275135,-42596.14671687511,-42596.597119991915,-42597.041496379126,-42597.47995831875,-42597.912615667905,-42598.33957590477,-42598.76094417434,-42599.176823333546,-42599.5873139958,-42599.99251457532,-42600.39252133042,-42600.787428407,-42601.17732788071,-42601.56230979928,-42601.94246222366,-42602.31787126922,-42602.688621145644,-42603.05479419709,-42603.41647094099,-42603.7737301068,-42604.12664867393,-42604.4753019091,-42604.81976340313,-42605.160105107054,-42605.496397367955,-42605.8287089636,-42606.15710713724,-42606.48165763104,-42606.802424719695,-42607.11947124274,-42607.43285863684,-42607.7426469671,-42608.04889495816,-42608.35166002434,-42608.65099829956,-42608.94696466653,-42609.23961278534,-42609.52899512167,-42609.81516297422,-42610.09816650198,-42610.378054750516,-42610.65487567808,-42610.92867618104,-42611.19950211875,-42611.467398338216,-42611.732408697855,-42611.99457609116,-42612.253942469324,-42612.510548864244,-42612.76443541013,-42613.0156413654,-42613.26420513359,-42613.51016428409,-42613.753555572446,-42613.994414959976,-42614.23277763341,-42614.46867802359,-42614.70214982409,-42614.933226009365,-42615.161938852376,-42615.38831994214,-42615.61240020029,-42615.834209898094,-42616.053778672125,-42616.27113554049,-42616.48630891804,-42616.699326631584,-42616.91021593455,-42617.11900352135,-42617.32571554156,-42617.5303776135,-42617.733014837686,-42617.93365181001,-42618.13231263425,-42618.3290209348,-42618.523799868744,-42618.716672137474,-42618.90765999863,-42619.096785277165,-42619.284069376285,-42619.469533288415,-42619.65319760536,-42619.835082528785,-42620.01520787986,-42620.19359310931,-42620.37025730644,-42620.54521920876,-42620.7184972105,-42620.89010937167,-42621.060073426524,-42621.22840679166,-42621.39512657424,-42621.56024957984,-42621.72379231991,-42621.88577101956,-42622.046201624435,-42622.20509980805,-42622.36248097851,-42622.51836028515,-42622.67275262526,-42622.82567264998,-42622.97713477102,-42623.12715316613,-42623.275741785204,-42623.42291435586,-42623.56868438892,-42623.71306518384,-42623.85606983379,-42623.997711230804,-42624.138002070686,-42624.27695485779,-42624.41458190969,-42624.55089536153,-42624.685907170715,-42624.81962912081,-42624.95207282608,-42625.083249735166,-42625.21317113521,-42625.34184815559,-42625.469291771646,-42625.59551280808,-42625.720521942785,-42625.844329709864,-42625.96694650326,-42626.08838257964,-42626.208648061634,-42626.32775294099,-42626.445707081104,-42626.562520220374,-42626.678201974515,-42626.79276183946,-42626.906209193854,-42627.01855330171,-42627.12980331467,-42627.239968274516,-42627.34905711554,-42627.45707866662,-42627.56404165347,-42627.669954700774,-42627.77482633418,-42627.87866498242,-42627.98147897901,-42628.083276564466,-42628.18406588776,-42628.283855008325,-42628.38265189781,-42628.48046444146,-42628.57730044009,-42628.67316761139,-42628.768073591666,-42628.862025937124],"type":"scatter","name":"source1_p_poc","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.0,0.0,3.8953395805405693e-8,6.441932040978138e-6,0.00013067056800565937,0.0010728409602299475,0.005115913001436587,0.01690401968251211,0.043033599426926916,0.09041015221605814,0.1645942913675844,0.2693051275414906,0.4070797281290319,0.5799889076894675,0.7893781349420783,1.0348212885277182,1.3134780910576638,1.6207636978421822,1.9520427261233089,2.3042363019240173,2.6764631875933524,3.0696954580048135,3.4860235266851727,3.9280866910643795,4.398796487311774,4.901154483246829,5.437967631682957,6.011465264601993,6.622974604412665,7.272795529120992,7.960295248671416,8.684136341672735,9.442533771858196,10.23347900997218,11.054922413556612,11.904924210976422,12.781776081418604,13.68408200234452,14.610785595487329,15.561145268317649,16.534672045908206,17.5310528949527,18.550077491107,19.591578014418282,20.65538320566989,21.74128463683024,22.849013858446686,23.978230290179354,25.12852057677069,26.299408238364787,27.490370838998476,28.70086020318341,29.93032163402306,31.178209429120876,32.443997834860006,33.72718801042883,35.027311898135736,36.34393383039807,37.676650118200236,39.02508679896164,40.388895763304106,41.767749857964915,43.16133775959451,44.56935936357379,45.991522176252154,47.42753879367055,48.87712532710046,50.34000050460595,51.81588567927741,53.30450932617188,54.80561142152069,56.31892141227179,57.84416123521731,59.38105428655185,60.929330369317555,62.48872529058598,64.05897812444864,65.63982997657094,67.23102525153156,68.83231386314445,70.44345266063098,72.06420563235952,73.69434349279582,75.3336433307107,76.98188849682472,78.63886850759617,80.3043787354223,81.97821987809519,83.66019737474524,85.35012093650109,87.04780425062874,88.75306481848722,90.46572386002373,92.18560624527116,93.91254044706277,95.64635852094553,97.38689610916042,99.13399245327129,100.88749039683373,102.6472363659207,104.41308032490817,106.18487571140213,107.96247935615169,109.74575139283898,111.53455516134207,113.32875710742651,115.12822668175251,116.93283624094002,118.74246095286763,120.55697870753114,122.37627003391113,124.20021802266484,126.02870825405323,127.86162873033102,129.69886981176043,131.54032415545103,133.38588665632855,135.23545438968767,137.0889265549486,138.94620442037842,140.80719126864687,142.6717923431737,144.53991479529776,146.41146763234954,148.28636166677572,150.16450946647032,152.04582530645627,153.9302251220394,155.8176264634961,157.7079484523344,159.60111173912117,161.49703846286724,163.39565221194147,165.29687798646287,167.2006421621191,169.10687245534427,171.01549788978616,172.92644876397833,174.83965662018562,176.75505421433593,178.67257548703424,180.59215553558968,182.51373058703882,184.43723797213667,186.3626161002719,188.28980443530202,190.21874347226367,192.14937471495216,194.0816406543238,196.01548474772454,197.95085139890372,199.8876859387783,201.8259346069215,203.76554453376633,205.70646372347008,207.64864103742644,209.59202617839173,211.5365696752038,213.48222286806137,215.42893789434476,217.37666767494662,219.32536590109518,221.2749870216483,223.22548623083478,225.17681945641752,227.12894334827345,229.08181526735402,231.03539327502472,232.9896361227521,234.94450324212835,236.89995473521907,238.85595136521277,240.81245454736134,242.76942634018843,244.72682943696606,246.6846271574424,248.64278343978503,250.60126283276668,252.56003048814168,254.51905215323114,256.47829416369564,258.4377234364724,260.39730746290246,262.3570143019802,264.3168125737875,266.2766714530451,268.23656066279267,270.19645046820733,272.1563116705312,274.116115601105,276.07583411551656,278.0354395878336,279.9949049049156,281.954203460841,283.91330915139446,285.87219636863864,287.8308399955417,289.78921540067404,291.7472984329781,293.7050654165916,295.66249314571724,297.6195588795501,299.576240337247,301.5325156929367,303.48836357078824,305.443763040103,307.39869361044435,309.3531352268131,311.3070682648371,313.26047352601347,315.2133322329573,317.1656260246943,319.117336951969,321.0684474725848,323.01894044676266,324.9687991325242,326.91800718109425,328.8665486323234,330.8144079101348,332.7615698179907,334.70801953436785,336.6537426082583,338.59872495469176,340.54295285026666,342.48641292870377,344.4290921764152,346.3709779280954,348.3120578623231,350.2523199971766,352.1917526858724,354.1303446124252,356.06808478732046,358.0049625431959,359.940967530552,361.8760897134804,363.8103193654032,365.74364706483016,367.6760636911459,369.60756042041055,371.5381287211718,373.4677603503071,375.3964473488811,377.3241820380235,379.25095701484133,381.17676514832687,383.10159957531175,385.0254536964278,386.9483211720959,388.87019591855284,390.79107210387025,392.7109441440273,394.62980669899565,396.54765466884834,398.46448318989724,400.38028763085515,402.295063589036,404.2088068865558,406.12151356658575,408.03317988962175,409.94380232979114,411.85337757117657,413.7619025041697,415.66937422186345,417.57579001647014,419.4811473757717,421.385443979592,423.28867769630244,425.19084657937384,427.0919488639362,428.99198296339273,430.89094746604326,432.78884113175724,434.6856628886794,436.581411829952,438.47608721048766,440.36968844376304,442.2622150986612,444.15366689633095,446.04404370707374,447.9333455473113,449.8215725765174,451.7087250942398,453.59480353712866,455.4798084760173,457.36374061301046,459.2466007786402,461.12838992902095,463.00910914307997,464.88875961979323,466.7673426754553,468.64485974099165,470.5213123593262,472.3967021827396,474.27103097028265,476.1443005852467,478.0165129926467,479.88767025673155,481.75777453853493,483.6268280934731,485.4948332689655,487.361792502088,489.22770831725495,491.0925833239597,492.9564202145212,494.81922176187805,496.6809908174033,498.5417303087721,500.4014432378485,502.26013267860594,504.1178017750859,505.97445373937387,507.83009184963544,509.68471944815235,511.53833993941544,513.3909567882275,515.2425735178615,517.0931937082296,518.9428209940813,520.791459063255,522.6391116549446,524.4857825579853,526.331475609182,528.1761946916724,530.0199437333109,531.8627267050779,533.7045476195183,535.5454105292185,537.3853195253037,539.2242787359594,541.0622923249847,542.8993644903937,544.7354994629814,546.570701504987,548.4049749087342,550.2383239953326,552.0707531133928,553.9022666377508,555.7328689681901,557.5625645283051,559.3913577642381,561.219253143568,563.0462551541328,564.8723683029181,566.6975971149418,568.5219461322044,570.3454199126542,572.1680230291004,573.9897600682515,575.8106356297144,577.6306543250298,579.4498207767049,581.268139617318,583.0856154886005,584.902253040575,586.7180569306405,588.5330318227749,590.3471823866821,592.1605132969854,593.9730292324533,595.7847348752261,597.5956349100363,599.4057340235036,601.2150369034017,603.0235482379535,604.831272715147,606.6382150220775,608.4443798442788,610.2497718650985,612.0543957650783,613.8582562213325,615.661357906981,617.4637054905464,619.2653036354034,621.0661569992428,622.8662702335255,624.6656479829758,626.4642948850642,628.2622155695205,630.0594146578283,631.855896762818,633.651666488151,635.4467284279127,637.2410871661702,639.0347472765554,640.8277133218845,642.6199898537328,644.411581412055,646.2024925248136,647.992727707631,649.7822914634244,651.5711882820648,653.3594226400871,655.146999000295,656.9339218115192,658.7201955082941,660.5058245105292,662.2908132232919,664.075166036494,665.858887324631,667.6419814465318,1175.3724701863491,-10471.774011339567,-18815.277645291448,-21317.664874709848,-23332.967600375778,-26517.414720084293,-28198.371183295672,-27239.77746285067,-25346.616895507028,-24564.732385626598,-25482.014982799934,-27545.85191199069,-30132.728314651067,-33015.65024083949,-36144.81859582159,-39328.334508707245,-42196.23164983953,-44388.1680307018,-45729.12565869278,-46259.77974460956,-46148.04579595037,-45583.411049780196,-44724.507928618135,-43699.42140427167,-42621.08820973176,-41590.053464048106,-40683.497197835604,-39944.32015526741,-39380.62086443647,-38975.181836691176,-38698.15002354078,-38516.8748976943,-38401.184150828216,-38325.48387701359,-38269.6099619207,-38219.28955890308,-38166.1477975202,-38107.01822470267,-38042.637646281844,-37976.105003154036,-37911.48756588022,-37852.794041244604,-37803.33748847979,-37765.42543997598,-37740.291989675075,-37728.20064566313,-37728.651760186265,-37740.629254897234,-37762.833276636244,-37793.863094203974,-37832.3413979531,-37876.98613573024,-37926.64471079405,-37980.30283961189,-38037.07698242603,-38096.196636712644,-38156.98190858979,-38218.8223673478,-38281.16150433736,-38343.48921517288,-38405.34125318255,-38466.302827364954,-38526.01284165548,-38584.166313503665,-38640.51409459979,-38694.860185799975,-38747.05757591167,-38797.003295909795,-38844.633137713965,-38889.91616999817,-38932.84923540728,-38973.45170484613,-39011.760835526075,-39047.82800419026,-39081.71586582139,-39113.49630025056,-39143.24887084405,-39171.05954185225,-39197.01947702863,-39221.2238531717,-39243.770687905904,-39264.75970257199,-39284.29123638314,-39302.46521490026,-39319.38018040756,-39335.13239939658,-39349.81507379247,-39363.51768070229,-39376.32545530568,-39388.31901595391,-39399.574117709344,-39410.16151486724,-39420.146912742166,-39429.59099241149,-39438.54949496976,-39447.07335383648,-39455.20886464351,-39462.99788338206,-39470.478045193224,-39477.682998068994,-39484.64264726497,-39491.38340702414,-39497.92845638058,-39504.297995843044,-39510.50950189042,-39516.5779766363,-39522.51619048348,-39528.33491596572,-39534.043151192425,-39539.64833144296,-39545.15652770483,-39550.5726312987,-39555.90052420936,-39561.14323513483,-39566.303081544815,-39571.38179814555,-39576.3806521714,-39581.300545955026,-39586.142107298154,-39590.905768292854,-39595.59183334135,-39600.20053718375,-39604.73209373037,-39609.186736453084,-39613.564751048536,-39617.866501068725,-39622.092447223105,-39626.243161064216,-39630.31933376149,-39634.32178063075,-39638.25144203451,-39642.10938120882,-39645.89677952602,-39649.61492966259,-39653.26522711068,-39656.849160435566,-39660.368300641756,-39663.82428996348,-39667.21883034946,-39670.55367187279,-39673.83060126186,-39677.051430720894,-39680.21798718231,-39683.332102107466,-39686.39560192697,-39689.410299186784,-39692.377984446,-39695.30041895236,-39698.17932810935,-39701.016395735394,-39703.81325910405,-39706.57150474667,-39709.29266498769,-39711.97821517727,-39714.629571580874,-39717.24808988079,-39719.83506424282,-39722.39172690094,-39724.919248210485,-39727.4187371221,-39729.89124202755,-39732.33775193055,-39734.75919789752,-39737.15645474574,-39739.530342926904,-39741.881630569915,-39744.211035646666,-39746.5192282282,-39748.80683280243,-39751.07443062631,-39753.32256208859,-39755.55172906152,-39757.76239722433,-39759.95499834217,-39762.129932485346,-39764.2875701809,-39766.42825448395,-39768.552302964556,-39770.660009602194,-39772.751646585704,-39774.82746601563,-39776.887701507905,-39778.9325696989,-39780.96227165259,-39782.9769941726,-39784.9769110199,-39786.962184041244,-39788.93296421019,-39790.88939258574,-39792.831601192636,-39794.75971382715,-39796.67384679357,-39798.57410957576,-39800.46060544754,-39802.33343202781,-39804.192681783934,-39806.03844248759,-39807.870797627445,-39809.6898267824,-39811.49560595924,-39813.28820789826,-39815.06770234983,-39816.834156324956,-39818.58763432289,-39820.328198538125,-39822.055909049326,-39823.77082399209,-39825.47299971747,-39827.1624909382,-39828.83935086398,-39830.50363132716,-39832.1553829001,-39833.79465500523,-39835.42149601862,-39837.035953367864,-39838.63807362488,-39840.22790259456,-39841.80548539892,-39843.370866557925,-39844.924090066976,-39846.46519947102,-39847.99423793608,-39849.51124831759,-39851.01627322647,-39852.5093550923,-39853.99053622383,-39855.45985886732,-39856.91736526189,-39858.3630976926,-39859.79709854094,-39861.21941033272,-39862.63007578319,-39864.02913783992,-39865.41663972297,-39866.79262496221,-39868.15713743234,-39869.510221385324,-39870.85192147993,-39872.182282809255,-39873.501350925115,-39874.80917186053,-39876.105792149465,-39877.39125884421,-39878.66561953062,-39879.92892234088,-39881.181215964345,-39882.42254965605,-39883.65297324327,-39884.87253713031,-39886.0812923011,-39887.279290320424,-39888.466583333044,-39889.64322406167,-39890.80926580304,-39891.96476242297,-39893.10976834971,-39894.24433856641,-39895.36852860213,-39896.48239452206,-39897.58599291662,-39898.6793808896,-39899.76261604601,-39900.83575647803,-39901.898860751724,-39902.951987892055,-39903.995197368,-39905.02854907674,-39906.05210332799,-39907.065920827365,-39908.070062659724,-39909.06459027259,-39910.04956545882,-39911.02505033939,-39911.99110734614,-39912.947799204216,-39913.89518891457,-39914.83333973659,-39915.762315170636,-39916.68217894075,-39917.59299497692,-39918.49482739812,-39919.387740495484,-39920.27179871496,-39921.147066640595,-39922.0136089779,-39922.871490537815,-39923.720776220005,-39924.56153099717,-39925.39381989915,-39926.217707997675,-39927.033260390745,-39927.840542188096,-39928.63961849599,-39929.43055440314,-39930.21341496638,-39930.98826519685,-39931.75517004635,-39932.514194393814,-39933.265403032754,-39934.00886065837,-39934.74463185477,-39935.47278108331,-39936.19337267048,-39936.90647079641,-39937.61213948382,-39938.310442586764,-39939.00144378036,-39939.68520655004,-39940.36179418185,-39941.031269752435,-39941.69369611969,-39942.34913591361,-39942.99765152737,-39943.639305108605,-39944.27415855127,-39944.9022734875,-39945.52371127986,-39946.13853301377,-39946.74679949026,-39947.34857121935,-39947.943908413,-39948.532870978706,-39949.115518513536,-39949.69191029798,-39950.26210529044,-39950.82616212172,-39951.384139089816,-39951.93609415499,-39952.48208493523,-39953.022168701544,-39953.556402373666,-39954.08484251632,-39954.607545335035,-39955.12456667273,-39955.63596200642,-39956.141786443884,-39956.64209472065,-39957.1369411972,-39957.62637985644,-39958.11046430132,-39958.58924775241,-39959.06278304606,-39959.53112263248,-39959.994318573925,-39960.45242254305,-39960.9054858219,-39961.353559300536,-39961.796693475684,-39962.23493845023,-39962.66834393207,-39963.09695923388,-39963.52083327227,-39963.940014567524,-39964.35455124356,-39964.76449102756,-39965.169881250185,-39965.570768845675,-39965.96720035185,-39966.35922191096,-39966.74687926962,-39967.13021777997,-39967.50928239969,-39967.88411769336,-39968.25476783285,-39968.621276598846,-39968.98368738143,-39969.34204318105,-39969.69638661037,-39970.046759895136,-39970.393204875596,-39970.73576300799,-39971.07447536601,-39971.40938264218,-39971.74052515021,-39972.067942825975,-39972.391675229264,-39972.71176154597,-39973.028240589905,-39973.3411508046,-39973.6505302653,-39973.956416680594,-39974.258847395504,-39974.55785939248,-39974.85348929398,-39975.145773364915,-39975.43474751458,-39975.720447298685,-39976.002907921866,-39976.282164240394,-39976.5582507637,-39976.83120165743,-39977.1010507453,-39977.36783151181,-39977.63157710477,-39977.89232033758,-39978.15009369152,-39978.404929318494,-39978.65685904346,-39978.90591436706,-39979.15212646789],"type":"scatter","name":"source2_q_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.0,-8.470329472543003e-20,1.4960650399831546e-5,0.00038944607638244236,0.0024041354911216676,0.007417986378806932,0.013102811507343173,0.009851953318442963,-0.018314338443364053,-0.09074210507752245,-0.22568927823133134,-0.4402772704661573,-0.7477742045977163,-1.1572238527558358,-1.6706188633204881,-2.283436270926994,-2.9871098118878994,-3.771287298824169,-4.628116130553903,-5.551728751136262,-6.540726039356907,-7.5961799135995545,-8.72271225326297,-9.926872970430166,-11.216345724139497,-12.599007688667939,-14.081043233362053,-15.667322627303502,-17.359958967349453,-19.159576282549622,-21.06467433720404,-23.07270489156082,-25.180155062363553,-27.383043585336857,-29.677555583448694,-32.06001999224324,-34.52762880146081,-37.07805248481869,-39.70985208794419,-42.422034921674936,-45.21412481201233,-48.08592160141514,-51.03735920708712,-54.06849280527183,-57.17926909418193,-60.36962493749363,-63.63928869172163,-66.98789676019248,-70.41490828752924,-73.91968969110421,-77.5015356474821,-81.15969536506903,-84.89343861460691,-88.70203448029774,-92.58481492692556,-96.54114228391337,-100.57044958505335,-104.67221981732807,-108.84599555126634,-113.09137069978713,-117.40797586095684,-121.79547806517361,-126.253558997662,-130.78191926051443,-135.3802624680111,-140.04829916034237,-144.78573950422307,-149.59229380440095,-154.46767196000184,-159.41158204472788,-164.42373548473427,-169.50384917172192,-174.65164726537395,-179.8668538282938,-185.1491901102565,-190.49837329710508,-195.9141197226333,-201.3961487342703,-206.94418480497757,-212.55795934435298,-218.2372106848287,-223.9816852041302,-229.7911374827412,-235.6653307821423,-241.60403654199058,-247.6070335546969,-253.67410698269902,-259.80504730502753,-265.9996496771702,-272.25771326914514,-278.5790409579902,-284.96343899892156,-291.4107169227693,-297.92068749535855,-304.493166780775,-311.1279742633227,-317.8249329489571,-324.5838694760827,-331.4046141589285,-338.2870010153873,-345.23086773879885,-352.23605564598677,-359.3024095923136,-366.42977786671497,-373.6180120755088,-380.8669670192216,-388.1765005726873,-395.5464735694241,-402.976749694774,-410.4671953882936,-418.0176797557004,-425.62807449003304,-433.298253800978,-441.0280943500848,-448.8174751914527,-456.666277714563,-464.5743855900981,-472.5416847159837,-480.5680631647641,-488.6534111308487,-496.79762087796286,-505.0005866868039,-513.2622048024771,-521.5823733828361,-529.960992447022,-538.3979638256037,-546.8931911116392,-555.4465796135933,-564.0580363097387,-572.7274698042899,-581.4547902853383,-590.2399094843818,-599.0827406377028,-607.9831984492316,-616.9411990550554,-625.9566599892655,-635.0295001511532,-644.1596397736149,-653.34700039266,-662.5915048180616,-671.8930771049684,-681.2516425265906,-690.6671275477572,-700.1394597995082,-709.6685680544713,-719.2543822032524,-728.8968332315736,-738.5958531983856,-748.3513752147242,-758.163333423427,-768.0316629796272,-777.9563000319691,-787.9371817046188,-797.9742460798766,-808.0674321815584,-818.2166799589096,-828.4219302712203,-838.6831248729362,-849.0002063993777,-859.3731183529388,-869.8018050898102,-880.2862118071644,-890.82628453078,-901.4219701031237,-912.0732161718109,-922.7799711784855,-933.5421843480299,-944.3598056781827,-955.2327859294311,-966.1610766152718,-977.1446299927528,-988.1833990532784,-999.2773375137485,-1010.4263998078716,-1021.6305410777966,-1032.8897171659205,-1044.2038846069327,-1055.5730006200579,-1066.997023101498,-1078.475910617029,-1090.0096223948058,-1101.5981183182735,-1113.2413589192893,-1124.9393053713206,-1136.6919194828322,-1148.4991636907343,-1160.3610010540128,-1172.2773952473945,-1184.2483105551814,-1196.2737118651248,-1208.3535646624146,-1220.487835023749,-1232.676489611447,-1244.9194956676993,-1257.216821008782,-1269.5684340194468,-1281.9743036472635,-1294.4343993971102,-1306.9486913255998,-1319.5171500356869,-1332.1397466711683,-1344.816452911356,-1357.5472409656873,-1370.3320835684294,-1383.1709539733836,-1396.0638259486236,-1409.0106737712604,-1422.0114722222409,-1435.066196581146,-1448.174822621049,-1461.3373266033416,-1474.553685272636,-1487.8238758516288,-1501.147876036038,-1514.5256639894853,-1527.9572183384792,-1541.4425181673334,-1554.981543013155,-1568.57427286082,-1582.2206881379427,-1595.9207697099437,-1609.6744988749797,-1623.481857359076,-1637.3428273110849,-1651.257391297801,-1665.2255322990236,-1679.247233702595,-1693.3224792995923,-1707.451253279321,-1721.6335402245695,-1735.8693251066313,-1750.1585932805465,-1764.5013304802312,-1778.8975228136733,-1793.3471567581573,-1807.850219155427,-1822.406697207,-1837.016578469359,-1851.6798508492627,-1866.3965025990256,-1881.166522311832,-1895.9898989170706,-1910.8666216756856,-1925.7966801755606,-1940.7800643269006,-1955.8167643576637,-1970.9067708090058,-1986.0500745307372,-2001.2466666768319,-2016.496538700923,-2031.7996823518606,-2047.156089669296,-2062.5657529792225,-2078.0286648896877,-2093.544818286328,-2109.114206328195,-2124.7368224433244,-2140.4126603245754,-2156.1417139253476,-2171.9239774553985,-2187.759445376695,-2203.64811239922,-2219.5899734769478,-2235.585023803715,-2251.6332588091977,-2267.7346741549172,-2283.8892657302517,-2300.0970296485325,-2316.357962243097,-2332.6720600634644,-2349.039319871474,-2365.459738637515,-2381.933313536744,-2398.4600419453886,-2415.039921437046,-2431.6729497790034,-2448.3591249287197,-2465.098445030093,-2481.890908410119,-2498.7365135752057,-2515.6352592078624,-2532.5871441631343,-2549.592167465353,-2566.6503283046586,-2583.7616260338023,-2600.9260601648084,-2618.143630365742,-2635.414336457554,-2652.738178410871,-2670.1151563429285,-2687.545270514426,-2705.0285213265843,-2722.5649093180073,-2740.1544351618495,-2757.797099662803,-2775.492903754244,-2793.241848495353,-2811.0439350683528,-2828.899164775687,-2846.807539037281,-2864.7690593879092,-2882.7837274744134,-2900.851545053212,-2918.9725139875904,-2937.146636245248,-2955.3739138957176,-2973.6543491079206,-2991.987944147735,-3010.3747013755246,-3028.814623243877,-3047.307712295154,-3065.8539711593044,-3084.4534025514968,-3103.106009269987,-3121.811794193841,-3140.570760280835,-3159.3829105653035,-3178.248248156051,-3197.1667762343313,-3216.138498051742,-3235.1634169283134,-3254.24153625051,-3273.3728594693257,-3292.557390098356,-3311.7951317119796,-3331.086087943504,-3350.430262483348,-3369.8276590773257,-3389.2782815248256,-3408.782133677232,-3428.339219436082,-3447.949542751574,-3467.6131076208476,-3487.3299180864265,-3507.099978234648,-3526.923292194134,-3546.799864134285,-3566.729698263819,-3586.7127988293028,-3606.7491701136887,-3626.8388164350004,-3646.9817421448797,-3667.1779516273,-3687.42744929722,-3707.7302395992842,-3728.0863270065543,-3748.4957160192753,-3768.958411163642,-3789.4744169906025,-3810.0437380746903,-3830.6663790128755,-3851.342344423415,-3872.0716389447384,-3892.8542672344124,-3913.69023396803,-3934.5795438382042,-3955.5222015535023,-3976.518211837519,-3997.567579427827,-4018.6703090750493,-4039.8264055419463,-4061.0358736024286,-4082.298718040727,-4103.614943650461,-4124.9845552338265,-4146.407557600703,-4167.883955567874,-4189.413753958177,-4210.996957599752,-4232.633571325221,-4254.323599970991,-4276.067048376485,-4297.863921383401,-4319.714223835054,-4341.617960575623,-4363.575136449517,-4385.585756300738,-4407.649824972185,-4429.76734730506,-4451.938328138197,-4474.162772307567,-4496.4406846455695,-4518.772069980544,-4541.156933136215,-4563.595278931043,-4586.087112177851,-4608.632437683131,-4631.231260246673,-4653.88358466103,-4676.589415710962,-4699.348758173146,-4722.161616815465,-4745.027996396824,-4767.947901666441,-4790.921337363726,-4813.948308217578,-4837.0288189461735,-4860.162874256498,-4883.350478843958,-4906.591637392036,-4929.886354571947,-3925.4872768290033,1145.771326336205,10134.956064117223,20524.809587069016,29761.28989851829,36646.72777284868,41199.97376400125,43951.22365606805,45481.09259353852,46254.03412247688,46542.9046336285,46486.6016027774,46176.59016689026,45728.50925167854,45270.554682429974,44898.044475135444,44640.78935384449,44466.513902801395,44314.6186607414,44131.918947479644,43894.17045125693,43606.25255549032,43289.406342064074,42967.72933782137,42660.575464307214,42381.26582387221,42138.76475004346,41938.21439611671,41781.36421233007,41665.84061293142,41586.20473207672,41535.05034074566,41504.841446492566,41489.14385531405,41483.11069419652,41483.73829387494,41489.372546504965,41499.632673750195,41514.826789634964,41535.77631734258,41563.371522920315,41598.34468836192,41641.1041937044,41691.6174254086,41749.51313526044,41814.0749875139,41884.41889743389,41959.51742790446,42038.32427202955,42119.81042831655,42203.02036703561,42287.11273347169,42371.35935356591,42455.16779961127,42538.04559199029,42619.60182178362,42699.51186328838,42777.5124155027,42853.3858944221,42926.954340204786,42998.07774563947,43066.64540573372,43132.57829209858,43195.820828904165,43256.34369177534,43314.139234859176,43369.22246455321,43421.62869059856,43471.41106159979,43518.63852280175,43563.39153336629,43605.76053432877,43645.842411360936,43683.73955311027,43719.55763620241,43753.40457045488,43785.389096672865,43815.61951372669,43844.20289801424,43871.24405715491,43896.84520153596,43921.10523808685,43944.119525329654,43965.97937374595,43986.7717480852,44006.57898056859,44025.478552806846,44043.54304786729,44060.8400850129,44077.43241641691,44093.377977150274,44108.73003881657,44123.537351385385,44137.84434090522,44151.69133077808,44165.11477532407,44178.14751839003,44190.81903580642,44203.15569183298,44215.1809768454,44226.915750363325,44238.378472571174,44249.58543160428,44260.55096174663,44271.28764891902,44281.80652637208,44292.117255110265,44302.22829409196,44312.14705627223,44321.880053451125,44331.43302801311,44340.81107237017,44350.018736183585,44359.060121549046,44367.93896728965,44376.65872257688,44385.222611009885,44393.63368552974,44401.89487501267,44410.00902314302,44417.97892029538,44425.807329127885,44433.49700457298,44441.05070884747,44448.471222120315,44455.76134935392,44462.92392390348,44469.96180833179,44476.87789293129,44483.67509236784,44490.35634081852,44496.92458597361,44503.382782176,44509.733883013534,44515.980833569025,44522.12656256709,44528.17397457265,44534.125942397965,44539.98529983753,44545.754834821986,44551.43728308077,44557.03532235789,44562.55156723602,44567.988564580744,44573.34878962447,44578.63464268304,44583.84844649721,44588.99244418102,44594.06879775095,44599.079587208864,44604.026810140735,44608.91238179626,44613.738135606036,44618.50582409723,44623.217120164714,44627.87361865684,44632.47683823606,44637.02822347318,44641.52914713942,44645.98091265782,44650.38475668169,44654.74185176738,44659.053309112554,44663.32018133358,44667.54346525743,44671.72410470686,44675.862993259056,44679.96097696167,44684.01885699028,44688.037392236845,44692.01730181696,44695.959267489074,44699.863935978174,44703.7319211992,44707.56380637709,44711.360146060586,44715.12146802988,44718.848275097,44722.54104680156,44726.20024100222,44729.82629536736,44733.419628768155,44736.980642576666,44740.50972187397,44744.00723657214,44747.47354245402,44750.90898213623,44754.31388595912,44757.688572808926,44761.03335087614,44764.348518355015,44767.63436408833,44770.891168161295,44774.11920244937,44777.31873112299,44780.490011113514,44783.63329254322,44786.748819122986,44789.83682852039,44792.897552701026,44795.931218245576,44798.938046644886,44801.91825457556,44804.8720541571,44807.799653193455,44810.701255399545,44813.57706061509,44816.42726500576,44819.25206125424,44822.05163874093,44824.82618371552,44827.57587946085,44830.30090644832,44833.00144248682,44835.677662864706,44838.32974048559,44840.95784599833,44843.5621479213,44846.142812761565,44848.70000512847,44851.23388784294,44853.74462204123,44856.2323672748,44858.697281604924,44861.13952169329,44863.5592428883,44865.956599306664,44868.33174391145,44870.68482858525,44873.01600420024,44875.32542068346,44877.61322707883,44879.87957160502,44882.12460170983,44884.348464120994,44886.551304893284,44888.73326945248,44890.89450263582,44893.035148729425,44895.15535150241,44897.25525423847,44899.33499976392,44901.394730473774,44903.434588354474,44905.45471500469,44907.45525165332,44909.436339175416,44911.39811810588,44913.340728651194,44915.26431069902,44917.16900382634,44919.05494730543,44920.92228010866,44922.771140911485,44924.60166809417,44926.413999742326,44928.20827364594,44929.98462729766,44931.74319788979,44933.48412231039,44935.20753713872,44936.9135786396,44938.60238275726,44940.27408510861,44941.92882097555,44943.566725297474,44945.18793266248,44946.79257729893,44948.38079306612,44949.95271344503,44951.5084715286,44953.048200011785,44954.57203118176,44956.080096907484,44957.57252862975,44959.049457350484,44960.51101362284,44961.957327540345,44963.388528726966,44964.80474632663,44966.20610899299,44967.5927448793,44968.96478162826,44970.3223463622,44971.665565673175,44972.99456561335,44974.30947168541,44975.610408833374,44976.89750143325,44978.17087328415,44979.43064759954,44980.67694699857,44981.90989349778,44983.129608502946,44984.33621280108,44985.52982655284,44986.71056928499,44987.878559883124,44989.033916584696,44990.17675697233,44991.30719796708,44992.42535582242,44993.531346117925,44994.62528375372,44995.707282944655,44996.777457215234,44997.83591939432,44998.882781610446,44999.91815528721,45000.9421511388,45001.95487916605,45002.95644865233,45003.94696816016,45004.926545527574,45005.895287864994,45006.85330155241,45007.80069223636,45008.737564827774,45009.66402349937,45010.58017168374,45011.48611207139,45012.38194660923,45013.26777649891,45014.14370219587,45015.00982340797,45015.8662390949,45016.71304746739,45017.550345986805,45018.37823136474,45019.19679956317,45020.00614579427,45020.80636452076,45021.59754945647,45022.37979356665,45023.153189069,45023.91782743433,45024.67379938777,45025.42119491003,45026.16010323858,45026.89061286929,45027.612811558036,45028.32678632248,45029.03262344394,45029.73040846956,45030.42022621416,45031.102160762835,45031.776295473195,45032.44271297783,45033.10149518709,45033.75272329151,45034.396477764996,45035.03283836738,45035.66188414768,45036.283693447076,45036.898343902154,45037.50591244807,45038.10647532204,45038.70010806659,45039.28688553329,45039.86688188615,45040.4401706052,45041.006824490454,45041.566915665244,45042.120515580595,45042.667695018594,45043.20852409661,45043.743072271216,45044.27140834215,45044.793600456454,45045.30971611247,45045.81982216431,45046.323984825634,45046.82226967419,45047.3147416558,45047.801465088996,45048.28250366902,45048.757920472235,45049.227777960725,45049.692137986305,45050.151061795266,45050.60461003263,45051.052842746605,45051.495819393145,45051.93359884022,45052.36623937254,45052.793798695806,45053.21633394135,45053.63390167065,45054.04655787974,45054.454358003764,45054.85735692151,45055.255608959924,45055.649167898504,45056.03808697393,45056.42241888456,45056.802215794814,45057.17752933978,45057.54841062961,45057.91491025405],"type":"scatter","name":"source1_q_poc","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.0,0.030226167260987116,0.12064360429477503,0.2771261746009928,0.4876812412821561,0.737901947477795,1.0209720431989755,1.3376326152392635,1.6910896698924245,2.0824827098688825,2.5094106050346663,2.9671078529262376,3.4501521636662598,3.9537012549781205,4.473750457154271,5.007000467630782,5.5506380322245015,6.102272211318022,6.659972361476658,7.222274977333242,7.7881561969661846,8.356921661970507,8.928110767215744,9.501398278219487,10.076538431374956,10.65332612351899,11.231575859329817,11.811112828561601,12.391765518149963,12.97336810917795,13.555759834498563,14.138789990733532,14.722319580475796,15.306224522899488,15.890397159177816,16.47474693161252,17.05920070147627,17.643701341597637,18.228206581409935,18.812686576813082,19.397122028114754,19.98150191537179,20.565821691514383,21.150081597024016,21.734285158945077,22.318437984555995,22.90254664983131,23.48661794597915,24.07065826841945,24.654673305213905,25.23866786452332,25.82264586181857,26.406610390505122,26.99056384219555,27.574508058230712,28.15844447624613,28.7423742689563,29.32629845517783,29.91021798243968,30.494133778650514,31.078046775994135,31.66195791109628,32.24586810948017,32.82977825702106,33.41368917030628,33.99760156585713,34.581516038635314,35.16543304703235,35.749352908087985,36.33327580020568,36.91720177732672,37.50113080926635,38.08506281041409,38.66899762888458,39.25293500751618,39.83687455821886,40.420815772009966,41.004758060033815,41.58870080161501,42.172643379389726,42.756585196354386,43.34052567939806,43.92446427846769,44.50840046536662,45.0923337337735,45.67626359872668,46.26018959491456,46.844111273775304,47.42802820015812,48.01193994959475,48.59584610633212,49.17974626235182,49.763640016847106,50.34752697607289,50.93140675329181,51.515278968846694,52.09914325035606,52.68299923298266,53.2668465597744,53.85068488193705,54.434513859052146,55.01833315916426,55.602142458789096,56.18594144283714,56.769729804479454,57.35350724496595,57.937273473391826,58.52102820643607,59.104771168070144,59.688502089269186,60.27222070772915,60.85592676760845,61.439620019294004,62.023300219192556,62.60696712954715,63.19062051827306,63.774260158814485,64.35788583001626,64.94149731601,65.52509440610933,66.10867689471192,66.69224458120519,67.27579726987356,67.8593347698072,68.44285689481129,69.02636346331698,69.60985429829407,70.1933292271654,70.77678808172432,71.36023069805474,71.9436569164547,72.5270665813633,73.11045954129143,73.69383564875619,74.27719476021878,74.86053673602552,75.44386144035177,76.02716874114863,76.61045851009155,77.19373062253138,77.77698495744659,78.3602213973975,78.94343982848143,79.52664014028923,80.10982222586269,80.69298598165285,81.27613130747946,81.85925810649081,82.44236628512454,83.02545575306914,83.60852642322608,84.19157821167258,84.77461103762484,85.3576248234019,85.94061949439006,86.52359497900734,87.10655120866892,87.68948811775246,88.2724056435639,88.85530372630336,89.43818230903173,90.02104133763682,90.60388076080012,91.18670052996356,91.76950059929659,92.35228092566301,92.93504146858808,93.51778219022594,94.10050305532671,94.68320403120394,95.26588508770195,95.84854619716353,96.43118733439708,97.01380847664456,97.59640960354892,98.17899069712182,98.76155174171123,99.34409272396944,99.92661363282058,100.50911445942867,101.09159519716545,101.67405584157825,102.25649639035807,102.83891684330767,103.42131720230961,104.00369747129454,104.58605765620929,105.16839776498547,105.75071780750757,106.33301779558181,106.91529774290446,107.49755766503077,108.07979757934366,108.66201750502267,109.24421746301307,109.826397475995,110.40855756835279,110.99069776614438,111.57281809707109,112.1549185904472,112.7369992771699,113.31906018968944,113.90110136197937,114.48312282950694,115.06512462920385,115.64710679943686,116.229069379979,116.81101241198067,117.39293593794112,117.97484000168004,118.55672464830947,119.13858992420583,119.72043587698221,120.30226255546103,120.88407000964663,121.46585829069856,122.04762745090454,122.6293775436541,123.21110862341241,123.79282074569419,124.37451396703817,124.95618834498134,125.53784393803406,126.11948080565503,126.7010990082265,127.28269860703008,127.86427966422258,128.44584224281212,129.02738640663475,129.60891222033098,130.19041974932296,130.77190905979154,131.35338021865422,131.93483329354268,132.51626835278122,133.0976854653649,133.67908470093846,134.26046612977544,134.84182982275712,135.4231758513525,136.00450428759794,136.58581520407736,137.1671086739027,137.74838477069486,138.32964356856428,138.9108851420927,139.49210956631444,140.07331691669845,140.65450726913036,141.23568069989497,141.81683728565895,142.39797710345377,142.97910023065901,143.56020674498606,144.14129672446163,144.7223702474121,145.30342739244756,145.88446823844683,146.46549286454203,147.04650135010397,147.6274937747274,148.20847021821655,148.78943076057138,149.37037548197364,149.9513044627731,150.53221778347452,151.11311552472438,151.69399776729824,152.27486459208797,152.85571608008945,153.43655231239063,154.0173733701594,154.59817933463214,155.17897028710232,155.7597463089091,156.34050748142673,156.92125388605356,157.5019856042015,158.08270271728583,158.66340530671525,159.24409345388162,159.82476724015072,160.40542674685247,160.98607205527193,161.56670324664003,162.14732040212485,162.72792360282284,163.30851292975044,163.88908846383583,164.46965028591075,165.05019847670263,165.63073311682683,166.21125428677917,166.79176206692847,167.37225653750934,167.95273777861524,168.53320587019144,169.11366089202878,169.69410292375636,170.27453204483604,170.8549483345554,171.43535187202244,172.01574273615913,172.59612100569584,173.1764867591659,173.7568400748997,174.33718103101984,174.91750970543578,175.4978261758386,176.0781305196964,176.65842281424938,177.23870313650528,177.81897156323473,178.39922817096704,178.97947303598565,179.55970623432464,180.1399278417637,180.72013793382538,181.3003365857704,181.88052387259438,182.46069986902438,183.0408646495151,183.62101828824598,184.2011608591176,184.78129243574867,185.36141309147328,185.94152289933766,186.52162193209716,187.1017102622141,187.68178796185484,188.2618551028871,188.84191175687764,189.4219579950901,190.0019938884824,190.5820195077046,191.16203492309708,191.74204020468832,192.32203542219293,192.90202064500983,193.48199594222055,194.06196138258733,194.64191703455188,195.22186296623335,195.80179924542725,196.3817259396036,196.96164311590618,197.54155084115067,198.12144918182378,198.70133820408216,199.28121797375096,199.86108855632332,200.44095001695894,201.02080242048336,201.60064583138728,202.18048031382563,202.76030593161687,203.3401227482422,203.9199308268453,204.4997302302314,205.0795210208669,205.65930326087891,206.23907701205508,206.8188423358428,207.39859929334906,207.97834794534052,208.55808835224275,209.13782057414053,209.7175446707772,210.29726070155522,210.87696872553556,211.45666880143798,212.036360987641,212.61604534218196,213.19572192275712,213.77539078672189,214.3550519910909,214.93470559253825,215.51435164739792,216.09399021166377,216.6736213409902,217.25324509069222,217.83286151574595,218.4124706707891,218.99207261012128,219.5716673877047,220.15125505716435,220.73083567178887,221.31040928453092,221.88997594800784,222.4695357145021,223.04908863596225,223.62863476400344,224.20817414990793,224.78770684462629,225.3672328987775,225.94675236265024,226.52626528620334,227.1057717190669,227.68527171054262,228.2647653096051,228.844252564903,229.6833153932153,232.90623664133489,240.60444572852057,252.2427763793624,264.1256796142898,271.71204909013653,272.6494436366804,267.79449830143403,259.91347945840084,251.69529910927042,244.6254658598867,239.0631300776364,234.87988608754085,231.9658519538887,230.32356793612584,229.92137029107658,230.56046473465082,231.8846016381439,233.4923538830408,235.04973279143658,236.34449871086744,237.2780886843332,237.83232688789371,238.03647814022412,237.94716422171186,237.63614645435132,237.1805832963115,236.65365046882616,236.11614881897196,235.6118910947208,235.1666052898179,234.79039467881057,234.48165663483664,234.2315533341915,234.02789284303978,233.85803882349012,233.71082599142505,233.57746990729748,233.45187852066923,233.33042638271658,233.21153143022647,233.09506532748418,232.98177305234054,232.87275350366858,232.76906249948038,232.67148247863688,232.5804217466961,232.49593901992654,232.4178172800572,232.34566782599697,232.2790196801983,232.21738973688437,232.16032747819855,232.10743623562462,232.0583793279924,232.01287159945008,231.97066602455527,231.9315362590213,231.89526219724166,231.86161952756257,231.83037530184535,231.80128882287875,231.77411607208472,231.74861627897857,231.7245585091434,231.70172755787823,231.67992817351586,231.65898752812078,231.63875586658486,231.6191056240254,231.59992945262852,231.5811376595727,231.5626555192616,231.5444208078437,231.52638169011598,231.5084950322953,231.49072506901538,231.47304240317706,231.45542327642715,231.43784905240565,231.42030585595816,231.40278425521083,231.38527894057964,231.36778832228723,231.35031406975276,231.332860600591,231.31543457135473,231.2980443979376,231.28069981605964,231.2634114904142,231.2461906565747,231.2290488073704,231.21199741720437,231.19504771882447,231.17821052782713,231.16149611231955,231.14491409717309,231.1284733905952,231.11218212904484,231.0960476345852,231.08007638887082,231.06427402176482,231.04864531617025,231.0331942252035,231.01792389895027,231.00283671819625,230.98793433305207,230.97321770683573,230.95868716399661,230.94434244218502,230.9301827463827,230.9162068038808,230.90241291836128,230.8887990222137,230.87536272676977,230.8621013701909,230.8490120631432,230.83609173183086,230.82333715826016,230.81074501730382,230.79831191051437,230.78603439667077,230.77390901919668,230.7619323306345,230.75010091424986,230.73841140289107,230.72686049512706,230.71544496880313,230.70416169214408,230.6930076326134,230.68197986372957,230.67107557003644,230.6602920504092,230.64962671985214,230.63907710995773,230.62864086818084,230.61831575609997,230.60809964682045,230.59799052166736,230.58798646630052,230.5780856663656,230.56828640278945,230.55858704681404,230.54898605485988,230.53948196329839,230.5300733832057,230.5207589951582,230.51153754412064,230.50240783447066,230.49336872519405,230.4844191252808,230.47555798934474,230.4667843134843,230.45809713139678,230.4494955107527,230.4409785498338,230.4325453744336,230.4241951350173,230.41592700413455,230.40774017407736,230.3996338547731,230.3916072719009,230.38365966521957,230.37579028709277,230.36799840119957,230.36028328141435,230.35264421084383,230.34508048100682,230.33759139114224,230.33017624763448,230.32283436354115,230.3155650582127,230.30836765699232,230.30124149098492,230.29418589688638,230.28720021686286,230.28028379847257,230.27343599462193,230.2666561635499,230.25994366883327,230.25329787940905,230.24671816960762,230.24020391919325,230.23375451340863,230.22736934302003,230.22104780436047,230.214789299369,230.20859323562485,230.20245902637458,230.19638609055127,230.1903738527858,230.18442174340933,230.17852919844705,230.17269565960223,230.1669205742334,230.1612033953213,230.15554358142987,230.1499405966585,230.14439391058826,230.13890299822228,230.13346733992037,230.12808642132885,230.12275973330685,230.11748677184872,230.11226703800418,230.10710003779616,230.10198528213718,230.09692228674479,230.0919105720567,230.0869496631458,230.08203908963603,230.07717838561842,230.07236708956938,230.06760474426932,230.06289089672393,230.05822509808712,230.05360690358614,230.04903587244897,230.04451156783395,230.04003355676235,230.03560141005303,230.03121470225983,230.0268730116115,230.02257591995425,230.0183230126962,230.01411387875535,230.00994811050901,230.00582530374584,230.0017450576197,229.99770697460676,229.9937106604629,229.98975572418476,229.9858417779709,229.98196843718625,229.9781353203271,229.97434204898772,229.9705882478289,229.96687354454698,229.96319756984508,229.95955995740402,229.95596034385593,229.95239836875734,229.94887367456406,229.94538590660628,229.94193471306482,229.93851974494768,229.9351406560671,229.93179710301797,229.92848874515556,229.92521524457453,229.92197626608788,229.918771477207,229.91560054812084,229.91246315167632,229.90935896335893,229.90628766127338,229.90324892612438,229.9002424411977,229.89726789234194,229.89432496794967,229.8914133589392,229.88853275873674,229.88568286325804,229.88286337089093,229.88007398247746,229.87731440129642,229.8745843330461,229.8718834858268,229.86921157012392,229.86656829879092,229.86395338703238,229.86136655238727,229.85880751471257,229.85627599616637,229.85377172119195,229.85129441650096,229.84884381105826,229.84641963606458,229.8440216249419,229.8416495133166,229.83930303900468,229.8369819419957,229.83468596443748,229.83241485062058,229.8301683469635,229.8279462019972,229.8257481663502,229.82357399273388,229.82142343592713,229.8192962527627,229.81719220211127,229.8151110448682,229.81305254393814,229.81101646422158,229.80900257260018,229.80701063792262,229.805040430991,229.80309172454648,229.80116429325568,229.79925791369686,229.79737236434656,229.79550742556543,229.79366287958536,229.79183851049578,229.79003410423044,229.7882494485541,229.78648433304934,229.7847385491036,229.78301188989622,229.7813041503854,229.77961512729544,229.77794461910395,229.7762924260289,229.7746583500165,229.77304219472825,229.7714437655287,229.76986286947286,229.76829931529414,229.76675291339168,229.76522347581866,229.76371081626962,229.76221475006884,229.76073509415824,229.7592716670855,229.75782428899203,229.75639278160133,229.75497696820747,229.7535766736631,229.75219172436798,229.7508219482577,229.74946717479193,229.74812723494315,229.74680196118564,229.74549118748365,229.74419474928072,229.74291248348837,229.74164422847508,229.74038982405528,229.7391491114786,229.73792193341873,229.7367081339629,229.73550755860094,229.7343200542149,229.7331454690682,229.73198365279532,229.73083445639077,229.7296977321999,229.7285733339072,229.72746111652683,229.7263609363921,229.72527265114567,229.72419611972896,229.72313120237237,229.72207776058565,229.72103565714744,229.72000475609576,229.7189849227181,229.71797602354167,229.7169779263241,229.71599050004346,229.71501361488885,229.71404714225105,229.71309095471264,229.71214492603931,229.71120893117035,229.71028284620905,229.70936654841418,229.70845991619018,229.70756282907857,229.70667516774884,229.70579681398956,229.70492765069935,229.7040675618783,229.70321643261897,229.70237414909772,229.70154059856634,229.70071566934303,229.69989925080432,229.6990912333763,229.69829150852604,229.69749996875387,229.69671650758443,229.6959410195585,229.6951734002254,229.69441354613406,229.69366135482542,229.69291672482422,229.69217955563116,229.69144974771476,229.6907272025038,229.69001182237923,229.6893035106666,229.6886021716281,229.68790771045525,229.68722003326113,229.68653904707267,229.6858646598235,229.6851967803462,229.68453531836508],"type":"scatter","name":"source1_v_mag_poc","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[50.0,50.0,49.99999999966536,49.99999999407674,49.99999996311272,49.99999987084565,49.99999968124417,49.999999373380895,49.999998942180405,49.99999839135509,49.99999773327777,49.99999698548135,49.99999617358392,49.99999532330141,49.99999445878835,49.9999935963841,49.999992744538346,49.99999190546452,49.99999107609448,49.99999025276198,49.99998943009595,49.99998860471978,49.99998777346517,49.999986935403456,49.99998609085111,49.99998524145469,49.99998439000716,49.9999835392733,49.99998269246911,49.999981852014486,49.999981020243034,49.99998019863417,49.99997938825435,49.99997858953397,49.99997780236699,49.99997702631508,49.99997626053317,49.99997550413099,49.9999747560516,49.9999740153487,49.99997328110028,49.99997255253497,49.99997182901797,49.99997111006189,49.999970395352435,49.9999696846931,49.99996897802281,49.99996827534498,49.99996757672525,49.99996688224177,49.99996619197316,49.999965505977535,49.99996482428063,49.999964146873864,49.999963473709535,49.999962804708126,49.99996213976217,49.99996147874613,49.999960821524205,49.999960167958655,49.999959517916004,49.99995887127222,49.999958227914966,49.99995758774576,49.99995695067843,49.99995631663883,49.99995568556162,49.9999550573884,49.999954432064975,49.999953809539115,49.99995318975945,49.99995257267575,49.99995195824393,49.99995134642312,49.99995073716964,49.99995013043099,49.999949526144725,49.999948924240336,49.9999483246423,49.99994772727279,49.99994713205282,49.999946538903394,49.999945947746006,49.99994535850357,49.99994477110101,49.99994418546571,49.99994360152765,49.99994301921929,49.999942438475394,49.99994185923276,49.99994128143008,49.999940705007724,49.99994012990771,49.999939556073556,49.99993898345013,49.99993841198368,49.99993784162172,49.999937272313076,49.99993670400787,49.999936136657574,49.99993557021502,49.999935004634466,49.99993443987161,49.999933875883634,49.99993331262923,49.99993275006857,49.99993218816336,49.999931626876844,49.99993106617371,49.99993050602017,49.99992994638385,49.9999293872338,49.999928828540455,49.999928270275596,49.99992771241229,49.9999271549249,49.99992659778902,49.99992604098145,49.99992548448018,49.99992492826434,49.99992437231418,49.99992381661102,49.99992326113725,49.99992270587629,49.99992215081256,49.99992159593142,49.999921041219196,49.99992048666309,49.9999199322512,49.99991937797243,49.9999188238165,49.99991826977389,49.99991771583584,49.99991716199426,49.999916608241755,49.99991605457156,49.999915500977515,49.99991494745402,49.99991439399608,49.99991384059913,49.99991328725916,49.9999127339726,49.99991218073628,49.999911627547476,49.99991107440382,49.99991052130326,49.99990996824415,49.99990941522506,49.99990886224488,49.99990830930274,49.999907756397995,49.99990720353021,49.999906650699124,49.99990609790464,49.99990554514684,49.999904992425876,49.999904439742046,49.9999038870957,49.9999033344873,49.999902781917314,49.9999022293863,49.999901676894794,49.99990112444335,49.99990057203253,49.99990001966288,49.999899467334885,49.99989891504902,49.99989836280568,49.999897810605226,49.999897258447916,49.99989670633393,49.99989615426337,49.999895602236215,49.99989505025236,49.99989449831157,49.99989394641351,49.99989339455767,49.99989284274346,49.99989229097014,49.9998917392368,49.99989118754243,49.999890635885826,49.99989008426568,49.999889532680484,49.999888981128606,49.999888429608255,49.99988787811746,49.99988732665412,49.999886775215934,49.99988622380048,49.99988567240516,49.99988512102721,49.999884569663735,49.99988401831164,49.99988346696771,49.99988291562855,49.999882364290634,49.99988181295026,49.999881261603605,49.99988071024668,49.99988015887535,49.99987960748537,49.999879056072324,49.999878504631674,49.99987795315874,49.99987740164875,49.99987685009676,49.99987629849775,49.999875746846556,49.9998751951379,49.999874643366425,49.99987409152663,49.99987353961294,49.99987298761969,49.99987243554109,49.9998718833713,49.99987133110438,49.999870778734305,49.999870226255,49.999869673660285,49.99986912094395,49.99986856809969,49.99986801512119,49.999867462002015,49.99986690873576,49.999866355315916,49.99986580173597,49.99986524798934,49.999864694069466,49.9998641399697,49.99986358568342,49.99986303120396,49.99986247652465,49.999861921638804,49.99986136653974,49.99986081122075,49.99986025567515,49.99985969989627,49.999859143877416,49.999858587611925,49.999858031093154,49.99985747431446,49.99985691726925,49.99985635995093,49.99985580235295,49.99985524446878,49.99985468629195,49.99985412781599,49.99985356903451,49.99985300994112,49.99985245052951,49.9998518907934,49.9998513307266,49.99985077032289,49.999850209576195,49.999849648480435,49.99984908702963,49.999848525217836,49.99984796303917,49.999847400487845,49.99984683755812,49.99984627424431,49.99984571054081,49.99984514644211,49.999844581942746,49.99984401703733,49.99984345172056,49.999842885987206,49.99984231983211,49.999841753250216,49.99984118623652,49.999840618786116,49.999840050894164,49.99983948255591,49.99983891376671,49.99983834452196,49.99983777481718,49.99983720464794,49.999836634009924,49.99983606289887,49.99983549131064,49.999834919241145,49.9998343466864,49.9998337736425,49.999833200105634,49.99983262607206,49.99983205153815,49.99983147650031,49.999830900955104,49.999830324899115,49.99982974832904,49.99982917124166,49.99982859363384,49.99982801550252,49.99982743684474,49.9998268576576,49.99982627793828,49.99982569768408,49.999825116892346,49.99982453556051,49.99982395368609,49.99982337126668,49.99982278829995,49.999822204783655,49.99982162071561,49.99982103609373,49.99982045091599,49.99981986518044,49.999819278885205,49.99981869202849,49.99981810460854,49.99981751662373,49.99981692807245,49.999816338953174,49.99981574926447,49.99981515900494,49.99981456817325,49.999813976768166,49.999813384788496,49.99981279223311,49.99981219910092,49.99981160539096,49.99981101110227,49.999810416233956,49.999809820785195,49.99980922475521,49.99980862814331,49.99980803094883,49.999807433171135,49.99980683480971,49.999806235864035,49.99980563633367,49.99980503621821,49.99980443551731,49.99980383423066,49.99980323235801,49.99980262989915,49.999802026853914,49.99980142322217,49.999800819003866,49.99980021419895,49.99979960880742,49.999799002829334,49.999798396264765,49.99979778911385,49.999797181376735,49.99979657305362,49.99979596414474,49.999795354650345,49.99979474457076,49.999794133906285,49.99979352265731,49.99979291082422,49.999792298407435,49.99979168540741,49.99979107182463,49.999790457659614,49.999789842912875,49.999789227584984,49.99978861167654,49.999787995188136,49.99978737812042,49.99978676047406,49.99978614224971,49.99978552344809,49.99978490406993,49.99978428411594,49.9997836635869,49.99978304248361,49.99978242080684,49.999781798557414,49.99978117573617,49.99978055234395,49.999779928381614,49.99977930385005,49.99977867875015,49.999778053082814,49.99977742684896,49.999776800049524,49.99977617268545,49.99977554475769,49.99977491626722,49.99977428721501,49.99977365760205,49.99977302742932,49.99977239669784,49.99977176540862,49.99977113356268,49.99977050116105,49.99976986820478,49.99976923469488,49.99976860063243,49.99976796601847,49.999767330854056,49.999766695140266,49.999766058878166,49.999765422068826,49.999764784713335,49.999764146812765,49.999763508368204,49.99976286938075,49.999762229851484,50.000584904791005,50.00498053683414,50.01364835986719,50.02515106518872,50.0362251698654,50.043518399420826,50.045768493121095,50.04400906122413,50.04006667026796,50.03536101208913,50.03065076082034,50.02636742820685,50.0228167195099,50.020225022572376,50.018667210738144,50.01806048629767,50.01821715285146,50.01890740496897,50.019914104736735,50.02104750507114,50.02216076510854,50.02314727679216,50.023945666832255,50.02453471881961,50.02492633064134,50.02515483140402,50.02526382438139,50.025298127036365,50.02529562257337,50.02528547593334,50.025286219267166,50.025307357294004,50.02535093882442,50.025413954908515,50.0254908164522,50.02557494846459,50.02566043377702,50.02574243671266,50.02581774801789,50.025884569237824,50.02594239214123,50.02599162448096,50.026033220782914,50.026068396524316,50.026098321423014,50.02612402450262,50.026146278919306,50.02616564755402,50.026182502031645,50.02619709616804,50.02620961901332,50.026220235395165,50.02622912245367,50.026236474748465,50.026242514602444,50.02624747767401,50.02625160519003,50.026255126904644,50.02625825011858,50.02626115139525,50.02626397218332,50.02626682019681,50.02626977173786,50.02627287754891,50.02627616795959,50.02627965881855,50.02628335628532,50.02628726070531,50.02629136926041,50.02629567735521,50.02630017912612,50.02630486739484,50.02630973333334,50.026314766266445,50.02631995348197,50.02632528030079,50.02633073021549,50.02633628516222,50.02634192593816,50.026347632629495,50.026353385187434,50.026359163878574,50.026364949760406,50.02637072496635,50.02637647292751,50.02638217848037,50.02638782790867,50.026393408973554,50.02639891088133,50.026404324271056,50.0264096411398,50.026414854784086,50.02641995970838,50.02642495154672,50.02642982698577,50.02643458368831,50.026439220227175,50.02644373600407,50.02644813117629,50.02645240656926,50.02645656359785,50.02646060418499,50.02646453068782,50.0264683458281,50.0264720526252,50.026475654335655,50.026479154393684,50.02648255635987,50.02648586387294,50.02648908060929,50.02649221024686,50.026495256434146,50.02649822276344,50.02650111274721,50.02650392979877,50.0265066772157,50.026509358167395,50.026511975685075,50.02651453265503,50.02651703181398,50.026519475746596,50.02652186688494,50.02652420750947,50.0265264997517,50.02652874559807,50.02653094689493,50.026533105354275,50.02653522256021,50.026537299975764,50.02653933895008,50.02654134072583,50.02654330644666,50.02654523716478,50.026547133848226,50.026548997388154,50.02655082860582,50.026552628259246,50.0265543970496,50.026556135627246,50.026557844597384,50.026559524525354,50.02656117594152,50.02656279934573,50.02656439521151,50.02656596398966,50.0265675061117,50.02656902199281,50.02657051203451,50.02657197662699,50.0265734161511,50.02657483098012,50.02657622148125,50.02657758801685,50.026578930945426,50.02658025062252,50.02658154740134,50.02658282163322,50.02658407366803,50.026585303854375,50.026586512539716,50.02658770007046,50.0265888667919,50.02659001304809,50.02659113918176,50.026592245534104,50.02659333244455,50.02659440025056,50.02659544928739,50.02659647988781,50.02659749238185,50.026598487096614,50.02659946435598,50.02660042448038,50.026601367786604,50.02660229458763,50.026603205192366,50.02660409990558,50.026604979027674,50.02660584285466,50.026606691677934,50.0266075257843,50.02660834545585,50.026609150969925,50.02660994259907,50.026610720611046,50.02661148526878,50.026612236830445,50.02661297554939,50.026613701674265,50.026614415448975,50.02661511711281,50.02661580690042,50.02661648504199,50.02661715176316,50.02661780728524,50.02661845182519,50.02661908559575,50.02661970880548,50.02662032165889,50.02662092435647,50.0266215170948,50.02662210006666,50.02662267346103,50.02662323746324,50.02662379225503,50.026624338014614,50.02662487491676,50.02662540313287,50.02662592283103,50.02662643417611,50.02662693732978,50.02662743245066,50.02662791969427,50.02662839921317,50.026628871156994,50.02662933567251,50.02662979290364,50.026630242991565,50.02663068607474,50.02663112228895,50.026631551767345,50.026631974640516,50.026632391036514,50.02663280108089,50.026633204896754,50.02663360260479,50.02663399432332,50.02663438016835,50.02663476025356,50.026635134690395,50.026635503588075,50.026635867053606,50.02663622519189,50.02663657810569,50.02663692589567,50.026637268660494,50.02663760649674,50.026637939499075,50.026638267760184,50.026638591370826,50.026638910419905,50.02663922499445,50.02663953517969,50.02663984105905,50.02664014271419,50.02664044022509,50.026640733669964,50.02664102312542,50.0266413086664,50.02664159036626,50.02664186829676,50.02664214252814,50.02664241312909,50.02664268016685,50.02664294370717,50.02664320381439,50.02664346055142,50.02664371397982,50.02664396415981,50.02664421115026,50.02664445500876,50.02664469579162,50.02664493355392,50.02664516834954,50.02664540023111,50.026645629250176,50.02664585545706,50.02664607890102,50.026646299630194,50.02664651769167,50.02664673313146,50.02664694599458,50.026647156325005,50.026647364165775,50.02664756955893,50.026647772545594,50.02664797316598,50.02664817145937,50.02664836746419,50.026648561218025,50.02664875275758,50.02664894211878,50.02664912933673,50.02664931444576,50.02664949747943,50.02664967847055,50.026649857451204,50.02665003445276,50.026650209505895,50.02665038264061,50.026650553886235,50.02665072327145,50.0266508908243,50.02665105657223,50.02665122054204,50.026651382759994,50.026651543251745,50.0266517020424,50.0266518591565,50.02665201461808,50.02665216845062,50.02665232067713,50.02665247132008,50.02665262040148,50.026652767942885,50.02665291396535,50.0266530584895,50.026653201535524,50.02665334312317,50.02665348327181,50.026653622000325,50.02665375932729,50.02665389527084,50.02665402984875,50.02665416307842,50.0266542949769,50.026654425560885,50.026654554846736,50.026654682850456,50.026654809587754,50.02665493507403,50.026655059324334,50.02665518235346,50.02665530417588,50.02665542480579,50.026655544257125,50.02665566254353,50.026655779678386,50.02665589567482,50.02665601054573,50.02665612430372,50.026656236961216,50.02665634853037,50.02665645902312,50.02665656845119,50.02665667682607,50.02665678415909,50.0266568904613,50.02665699574362,50.02665710001675,50.02665720329118,50.02665730557725,50.026657406885114,50.02665750722474,50.02665760660592,50.02665770503831,50.02665780253137,50.026657899094424,50.02665799473663,50.026658089467006,50.02665818329442,50.02665827622761,50.02665836827513,50.02665845944545,50.02665854974689,50.02665863918764,50.02665872777575,50.02665881551916,50.026658902425716,50.026658988503094,50.02665907375889,50.0266591582006,50.02665924183558,50.026659324671094,50.0266594067143,50.02665948797226,50.02665956845195,50.0266596481602,50.02665972710382,50.02665980528945,50.026659882723706,50.02665995941309,50.02666003536399,50.02666011058276,50.02666018507564,50.026660258848814,50.02666033190835,50.026660404260284,50.02666047591053,50.026660546864974,50.026660617129416,50.02666068670956,50.02666075561106,50.02666082383953,50.026660891400454,50.026660958299324,50.02666102454152,50.026661090132365,50.02666115507715,50.02666121938107,50.02666128304929,50.02666134608692,50.02666140849896,50.02666147029043,50.026661531466246,50.02666159203129,50.0266616519904,50.02666171134834,50.02666177010984],"type":"scatter","name":"source1_freq","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[50.0,50.0,46.15163652864002,44.979220404995026,43.58097178159506,42.355444825226904,41.552295297893096,41.023022637645745,40.833250566191964,40.769279966332505,40.76133305485534,40.8021384234005,40.89481965269243,41.04663070506293,41.25675204502468,41.51208139310457,41.79256653619174,42.07847118063732,42.35508565171135,42.61383913214342,42.85101732407632,43.06598139943558,43.25995039275595,43.43546510760143,43.596081682002904,43.74597698366099,43.889414486157015,44.03022942735471,44.17148637397899,44.31534703269106,44.46310612622606,44.61530998630755,44.77190815914731,44.93240641760337,45.096015266493275,45.26178516164627,45.428720221028115,45.59586398606381,45.76235412373842,45.92745129251783,46.0905473515834,46.25116085668287,46.408924073259215,46.56356570241334,46.71489176993551,46.86276691890251,47.007098055703544,47.147821179145566,47.28489199312593,47.41827966743631,47.54796324846799,47.67392979296657,47.796173680962745,47.914696624358356,48.029508039655646,48.14062557709587,48.248075546805445,48.35189314101684,48.452122317553176,48.54881537394875,48.64203224221146,48.73183960393199,48.81830991090596,48.90152037675556,48.98155199516868,49.05848860530133,49.13241603583488,49.20342133685145,49.271718466943305,49.33852265009148,49.40385688834342,49.46763621565984,49.529859486001975,49.59051325644286,49.649581784458455,49.70705633083988,49.762932283012134,49.81720830636192,49.86988637179372,49.920971506916345,49.97047152941389,50.01839680253285,50.06476000149411,50.10957591086605,50.15286124969277,50.19463450587447,50.23491576825745,50.273726556343696,50.31108965392043,50.34702895231132,50.38156930517233,50.41473639414519,50.44655660424684,50.477056908810184,50.50626476433961,50.53420801546627,50.56091480949359,50.586413519478256,50.61073267463992,50.633900897049216,50.65594684380637,50.6768991541021,50.696786400662276,50.71563704513222,50.73347939702726,50.750341575955886,50.76625147689995,50.78123673839389,50.79532471347289,50.80854244326965,50.82091663313641,50.83247363116538,50.84323940897778,50.85323954465,50.86249920764512,50.871043145618934,50.87889567297153,50.886080661018894,50.89262152966357,50.89854124045066,50.90386229090193,50.90860671002895,50.912796054934894,50.916451408420684,50.91959337752012,50.92224209289313,50.92441720901278,50.92613790508605,50.92742288665266,50.92829038781066,50.928758174019826,50.92884354543798,50.928563340746834,50.92793394142764,50.926971276447304,50.92569082731932,50.92410763350438,50.92223629811896,50.9200909939201,50.91768546953895,50.91503305593479,50.91214667304439,50.909038836602825,50.90572166511371,50.902206886947376,50.89850584754792,50.89462951673076,50.8905884960536,50.88639302624555,50.88205299467881,50.87757794286957,50.872977073995635,50.86825926041796,50.86343305119558,50.85850667958378,50.85348807050501,50.84838484798455,50.84320434254216,50.83795359853213,50.83263938142489,50.827268185023165,50.821846238607556,50.81637951400538,50.81087373257788,50.80533437212186,50.79976667368073,50.794175648262105,50.78856608345814,50.78294254996581,50.777309408004434,50.7716708136282,50.76603072493177,50.76039290814699,50.754760943629314,50.749138231732914,50.743527998573484,50.73793330167714,50.732357035516486,50.72680193693174,50.721270590437875,50.71576543341711,50.71028876119747,50.70484273201681,50.69942937187375,50.69405057926504,50.688708129811104,50.6834036807692,50.678138775436246,50.67291484744154,50.667733224930906,50.66259513464273,50.65750170587788,50.65245397436408,50.647452886016495,50.64249930059538,50.63759399526322,50.63273766804138,50.627930941169076,50.62317436436546,50.61846841799676,50.61381351614974,50.60921000961327,50.60465818876974,50.60015828639788,50.595710480388625,50.591314896375536,50.58697161028198,50.58268065078616,50.5784420017061,50.574255604306074,50.57012135952615,50.566039130136765,50.56200874281973,50.558029990177474,50.55410263267224,50.55022640049675,50.546400995378,50.54262609231598,50.53890134125872,50.53522636871539,50.53160077930881,50.528024157269336,50.524496067871254,50.52101605881351,50.5175836615459,50.51419839254273,50.51085975452489,50.50756723763205,50.504320320546455,50.50111847156937,50.4979611496519,50.49484780538152,50.491777881925174,50.488750815930786,50.48576603838828,50.48282297545118,50.479921049220174,50.477059678489915,50.474238279459925,50.47145626641117,50.4687130523491,50.46600804961433,50.46334067046227,50.46071032761225,50.45811643476782,50.45555840710859,50.453035661755194,50.45054761820764,50.44809369875882,50.44567332888328,50.443285937602724,50.44093095782872,50.43860782668378,50.4363159858014,50.434054881605924,50.43182396557307,50.429622694471846,50.427450530588466,50.42530694193329,50.42319140243107,50.42110339209555,50.419042397188825,50.417007910366294,50.414999430807626,50.413016464334504,50.411058523515614,50.40912512775956,50.407215803396134,50.405330083746506,50.40346750918297,50.40162762717844,50.399809992346704,50.39801416647309,50.39623971853705,50.394486224725945,50.39275326844153,50.391040440298646,50.38934733811734,50.38767356690785,50.3860187388498,50.38438247326517,50.382764396585785,50.38116414231583,50.37958135098897,50.37801567012114,50.37646675415913,50.374934264424745,50.37341786905566,50.371917242942224,50.37043206766154,50.368962031408124,50.36750682892194,50.36606616141394,50.364639736488925,50.36322726806663,50.36182847630073,50.36044308749586,50.35907083402321,50.35771145423482,50.35636469237622,50.355030298498434,50.353708028368665,50.352397643380236,50.35109891046195,50.3498116019868,50.34853549568016,50.34727037452794,50.34601602668417,50.34477224537874,50.34353882882493,50.34231558012724,50.34110230718915,50.3398988226214,50.338704943650356,50.337520492026854,50.336345293935715,50.335179179905516,50.33402198471915,50.332873547324915,50.33173371074852,50.33060232200576,50.32947923201578,50.328364295515584,50.327257370975175,50.3261583205138,50.32506700981693,50.32398330805458,50.32290708780033,50.32183822495155,50.32077659865077,50.31972209120796,50.318674588024166,50.317633977516074,50.316600151041925,50.315573002828394,50.31455242989873,50.31353833200217,50.312530611544275,50.31152917351892,50.31053392544104,50.309544777280806,50.30856164139902,50.30758443248341,50.30661306748676,50.3056474655657,50.30468754802068,50.303733238237676,50.30278446163034,50.30184114558406,50.30090321940079,50.2999706142449,50.299043263090724,50.29812110067074,50.29720406342522,50.29629208945282,50.29538511846234,50.294483091725695,50.293585952031734,50.29269364364153,50.29180611224427,50.29092330491464,50.290045170071046,50.28917165743462,50.28830271798969,50.287438303945116,50.286578368696325,50.285722866788674,50.284871753881525,50.28402498671344,50.28318252306806,50.28234432174117,50.28151034250849,50.28068054609429,50.27985489414089,50.2790333491794,50.278215874600406,50.27740243462628,50.27659299428406,50.27578751937892,50.27498597646856,50.274188332838165,50.273394556476546,50.272604616052405,50.27181848089161,50.27103612095529,50.2702575068183,50.269482609648506,50.26871140118667,50.267943853726926,50.26717994009804,50.26641963364488,50.265662908211006,50.26490973812141,50.26416009816573,50.26341396358264,50.26267131004398,50.26193211363997,50.26119635086447,50.26046399860113,50.25973503410994,50.25900943501392,50.25828717928665,56.109037186504764,63.73524151638972,67.96779962876734,69.11795763193987,68.70635146229525,67.57659239424,66.08957558499516,64.32980174025369,62.29559768793956,60.030013242083385,57.64333991739886,55.30074381981405,53.18986508533154,51.456241131953185,50.141437368526454,49.20340491669126,48.565864582293436,48.1450196930749,47.875105496704386,47.7169610237657,47.65077515664948,47.66382108890973,47.742074833442075,47.86872853351898,48.0270611338314,48.203333515823125,48.387397306011465,48.571592568552326,48.749683128680545,48.91680022019353,49.07002998380006,49.208785645619045,49.33448354296477,49.449700093778766,49.55726916964646,49.65965927598569,49.758695789049746,49.85551253248007,49.95060888012185,50.04394240128366,50.13504876783222,50.22318373142938,50.30747468970712,50.38705755368228,50.461180508754886,50.52926934863662,50.590956648403086,50.64608353094133,50.694679982206296,50.736931843664046,50.77314082886711,50.80368501217975,50.82898499338286,50.849478201119716,50.865601410023764,50.877779288216566,50.88641733729633,50.89189752966886,50.894576279039114,50.89478422279139,50.892827423837296,50.88898925764399,50.883532194135924,50.87669906422822,50.86871364145085,50.859780872823436,50.85008703725179,50.839800137767945,50.829070557080506,50.818031907709916,50.80680195610825,50.7954835338739,50.78416545138286,50.77292343036854,50.76182110082337,50.75091103726865,50.740235802049234,50.729828941725955,50.7197159093787,50.70991491334944,50.70043770794562,50.691290351219685,50.68247393680944,50.67398530104395,50.66581769268413,50.65796139703922,50.65040430877659,50.6431324539491,50.63613046439702,50.62938200590609,50.62287016093621,50.61657776415668,50.61048769091738,50.60458309965666,50.59884763184453,50.5932655738398,50.58782198518624,50.582502797375014,50.57729488620352,50.572186120668405,50.567165390973685,50.56222261842352,50.557348749779926,50.55253573856147,50.54777651545337,50.54306494973157,50.53839580341149,50.53376467965641,50.52916796689447,50.52460277995077,50.5200668993938,50.51555871014959,50.51107714031232,50.50662160095192,50.502191927596435,50.49778832394399,50.493411308240454,50.48906166265322,50.48474038588296,50.480448649184005,50.476187755903794,50.4719591045952,50.46776415570079,50.46360440175445,50.45948134100309,50.45539645431676,50.45135118523475,50.44734692298074,50.44338498826924,50.439466621714274,50.43559297464097,50.43176510209468,50.427983957839366,50.42425039114079,50.42056514513681,50.41692885660677,50.413342056960076,50.40980517427333,50.406318536213355,50.40288237369458,50.399496825130086,50.396161941148115,50.39287768965814,50.38964396116329,50.38646057422625,50.383327281006814,50.38024377279803,50.377209685498805,50.37422460496968,50.37128807222716,50.368399588441285,50.365558619706746,50.362764601566475,50.36001694327058,50.35731503176025,50.35465823537053,50.35204590724927,50.349477388495174,50.34695201101929,50.34446910013756,50.3420279769048,50.33962796020016,50.33726836857804,50.33494852189713,50.33266774274293,50.33042535765768,50.328220698193284,50.32605310180198,50.32392191257921,50.321826481873806,50.319766168777846,50.317740340511115,50.31574837271126,50.31378964964226,50.31186356433218,50.30996951864965,50.30810692332974,50.30627519795651,50.304473770910825,50.30270207929007,50.300959568805894,50.29924569366579,50.29755991644282,50.2959017079383,50.29427054704031,50.29266592058124,50.291087323197225,50.28953425719048,50.288006232397414,50.28650276606259,50.28502338272003,50.2835676140822,50.28213499893685,50.280725083052005,50.27933741908923,50.277971566524194,50.2766270915748,50.275303567136326,50.27400057272238,50.27271769441217,50.27145452480241,50.27021066296376,50.268985714400955,50.2677792910159,50.26659101107329,50.265420499168,50.264267386193616,50.263131309311625,50.26201191192072,50.26090884362585,50.2598217602065,50.258750323583854,50.25769420178657,50.25665306891487,50.255626605102506,50.25461449647695,50.253616435117145,50.25263211900882,50.251661251997476,50.25070354373904,50.24975870964779,50.24882647084202,50.2479065540872,50.246998691736906,50.24610262167146,50.24521808723445,50.244344837167105,50.24348262554098,50.242631211688725,50.24179036013331,50.24095984051571,50.24013942752123,50.23932890080487,50.23852804491534,50.23773664921845,50.23695450781956,50.23618141948564,50.23541718756664,50.23466161991672,50.233914528815106,50.23317573088701,50.232445047024356,50.231722302306935,50.23100732592342,50.23029995109306,50.229600014987504,50.228907358653174,50.228221826934295,50.227543268396396,50.22687153525059,50.22620648327857,50.22554797175827,50.22489586339055,50.22425002422658,50.223610323596105,50.222976634036876,50.222348831224664,50.22172679390453,50.22111040382297,50.220499545661035,50.219894106968574,50.21929397809939,50.21869905214748,50.21810922488434,50.21752439469718,50.21694446252829,50.2163693318155,50.21579890843351,50.215233100636354,50.214671819000834,50.21411497637109,50.21356248780416,50.21301427051633,50.212470243830836,50.21193032912619,50.2113944497859,50.210862531148756,50.21033450046049,50.209810286826,50.209289821162756,50.208773036155094,50.20825986620953,50.20775024741068,50.20724411747847,50.206741415726064,50.206242083018616,50.20574606173301,50.20525329571836,50.204763730257525,50.20427731202923,50.203793989071244,50.20331371074426,50.20283642769648,50.20236209182927,50.20189065626334,50.20142207530589,50.20095630441821,50.20049330018452,50.2000330202811,50.19957542344628,50.19912046945137,50.19866811907189,50.198218334059796,50.19777107711628,50.1973263118653,50.196884002827645,50.196444115395686,50.19600661580899,50.19557147113009,50.19513864922125,50.19470811872181,50.19427984902588,50.193853810260784,50.19342997326608,50.19300830957301,50.19258879138469,50.192171391556634,50.19175608357785,50.19134284155253,50.19093164018225,50.19052245474831,50.19011526109536,50.18971003561437,50.18930675522705,50.188905397370185,50.18850593998061,50.18810836148045,50.187712640762925,50.187318757178616,50.1869266905218,50.186536421017635,50.18614792930936,50.1857611964459,50.185376203870156,50.18499293340725,50.18461136725336,50.1842314879647,50.18385327844711,50.1834767219456,50.18310180203447,50.18272850260775,50.18235680786975,50.18198670232601,50.18161817077438,50.18125119829665,50.18088577025032,50.180521872260464,50.18015949021202,50.179798610242365,50.17943921873394,50.179081302307274,50.17872484781407,50.17836984233071,50.17801627315188,50.17766412778426,50.17731339394075,50.17696405953438,50.176616112672896,50.17626954165343,50.175924334957124,50.175580481243856,50.17523796934792,50.174896788272605,50.1745569271861,50.17421837541689,50.17388112244943,50.17354515791998,50.17321047161274,50.172877053455935,50.172544893518065,50.17221398200422,50.1718843092526,50.171555865731335,50.171228642035054,50.17090262888174,50.170577817109695,50.170254197674694,50.16993176164709,50.16961050020919,50.16929040465226,50.168971466374465,50.168653676878165,50.16833702776731,50.168021510745646,50.167707117614015,50.16739384026861,50.16708167069853,50.16677060098412,50.16646062329491,50.1661517298877,50.16584391310485,50.16553716537248,50.165231479198866,50.16492684717286,50.164623261962106,50.164320716311785,50.164019203043026,50.163718715051616,50.163419245306535],"type":"scatter","name":"source2_freq","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.0,3.975693351829396e-16,3.975693351829396e-16,3.975693351829396e-16,3.975693351829396e-16,3.975693351829396e-16,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-4.373262687012336e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,-1.1927080055488188e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-7.553817368475852e-15,-2.027603609432992e-14,-2.027603609432992e-14,-2.027603609432992e-14,-2.027603609432992e-14,-2.027603609432992e-14,-2.027603609432992e-14,-2.027603609432992e-14,-2.027603609432992e-14,-2.027603609432992e-14,-2.027603609432992e-14,-2.027603609432992e-14,-2.027603609432992e-14,-2.027603609432992e-14,-7.553817368475852e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.6118042881220995e-14,-4.6515612216403935e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.4130196205306295e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-5.208158290896509e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.6118042881220995e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.7310750886769815e-14,-4.890102822750157e-14,-4.890102822750157e-14,-4.890102822750157e-14,-4.890102822750157e-14,-4.890102822750157e-14,-4.890102822750157e-14,-4.890102822750157e-14,-4.890102822750157e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-5.684341886080802e-14,-5.367186024969685e-14,-5.367186024969685e-14,-5.367186024969685e-14,-5.367186024969685e-14,-5.367186024969685e-14,-5.367186024969685e-14,-5.367186024969685e-14,-5.367186024969685e-14,-5.52621375904286e-14,-5.52621375904286e-14,-5.52621375904286e-14,-5.52621375904286e-14,-5.52621375904286e-14,-5.52621375904286e-14,-5.52621375904286e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-5.208158290896509e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-4.5720473546038055e-14,-5.208158290896509e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-5.844269227189212e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-7.116491099774618e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-9.660934844945432e-14,-1.1368683772161603e-13,-8.269442171805144e-14],"type":"scatter","name":"source1_θ","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[-0.9000000000000001,0.7500000000000003,2.309042941808995e-11,-0.2655370791081055,-0.4272494222722437,-0.6868903936224358,-0.9815630237753629,-1.2938013932339527,-1.6223912854636904,-1.950717583914601,-2.2822878587940147,-2.6141824345712825,-2.9438395171680685,-3.269179425390673,-3.5878809584612465,-3.8979808003853647,-4.198271545474924,-4.488312774356316,-4.768259899689272,-5.0386367427189125,-5.300148490859065,-5.553565863937503,-5.799653936743942,-6.0391187214428514,-6.272559010850447,-6.500431897396344,-6.723038852329286,-6.940534314881474,-7.152950086719979,-7.360226653775684,-7.56224491494778,-7.75885422671198,-7.949895969784005,-8.135221809273355,-8.314706895192687,-8.488257886607535,-8.655816478730742,-8.817359346915602,-8.972895602939118,-9.122462887747355,-9.266122811712071,-9.40395639349992,-9.53605977077411,-9.662540467655694,-9.783514298286997,-9.899102954696236,-10.009432227261257,-10.114630742300106,-10.214829104077118,-10.310159297636913,-10.400754269116097,-10.486747595043223,-10.568273201711479,-10.645465096218905,-10.71845709193011,-10.787382518938541,-10.852373915780266,-10.913562711551979,-10.971078905213826,-11.025050757206564,-11.075604501528451,-11.122864087779295,-11.16695095679871,-11.207983851985373,-11.246078666457484,-11.281348324243217,-11.313902693959452,-11.343848531416212,-11.37128944815529,-11.396317183547461,-11.418948590768146,-11.439254057221968,-11.457292708707326,-11.473119326697326,-11.486790839706027,-11.498364634409555,-11.50789814648981,-11.515448946549517,-11.521074606961498,-11.524832569230526,-11.526780042576156,-11.52697390855657,-11.525470633497834,-11.522326189303616,-11.517595980701566,-11.511334778333763,-11.503596657920268,-11.494434945681874,-11.483902169937599,-11.472050018526257,-11.458929301612443,-11.444589919483066,-11.429080834995368,-11.41245005035232,-11.39474458785124,-11.376010474233349,-11.35629272827076,-11.335635351264525,-11.314081320174058,-11.291672583138162,-11.268450057177708,-11.244453627893156,-11.219722150989002,-11.194293455474758,-11.168204348407968,-11.14149062105628,-11.114187056365086,-11.086327437623153,-11.057944558224172,-11.02907023242614,-10.99973530701545,-10.9699696737869,-10.939802282755798,-10.909261156022806,-10.878373402216745,-10.84716523144593,-10.81566197069267,-10.783888079590394,-10.751867166527296,-10.719622005024766,-10.687174550342432,-10.654545956265927,-10.621756592036302,-10.58882605938369,-10.555773209630159,-10.522616160829852,-10.489372314916505,-10.45605837483088,-10.422690361602971,-10.389283631364835,-10.355852892273393,-10.322412221321937,-10.288975081023317,-10.255554335946618,-10.222162269092859,-10.188810598095063,-10.15551049122973,-10.122272583228545,-10.089106990879324,-10.056023328407289,-10.023030722627576,-9.990137827861787,-9.957352840612511,-9.924683513989043,-9.892137171880094,-9.859720722868891,-9.827440673887173,-9.795303143605173,-9.763313875555097,-9.731478250986264,-9.699801301450993,-9.668287721119912,-9.636941878826491,-9.605767829841207,-9.574769327375016,-9.543949833813736,-9.513312531683944,-9.482860334351845,-9.452595896457275,-9.422521624084153,-9.392639684670169,-9.362952016657957,-9.333460338890205,-9.304166159752047,-9.275070786063123,-9.246175331723146,-9.217480726113603,-9.188987722259617,-9.16069690475523,-9.132608697455613,-9.10472337094037,-9.077041049751632,-9.049561719410082,-9.022285233214244,-8.99521131882584,-8.968339584645612,-8.94166952598406,-8.915200531030603,-8.888931886625492,-8.862862783838768,-8.836992323360144,-8.811319520704032,-8.78584331123385,-8.76056255500938,-8.735476041462103,-8.71058249390149,-8.685880573857007,-8.66136888525949,-8.637045978466414,-8.612910354134026,-8.588960466941183,-8.565194729168319,-8.541611514135196,-8.518209159501604,-8.494985970434527,-8.471940222645356,-8.449070165300924,-8.426374023811814,-8.403850002501585,-8.381496287160166,-8.359311047485107,-8.337292439413204,-8.315438607347971,-8.293747686282472,-8.272217803824503,-8.25084708212485,-8.229633639712542,-8.208575593239711,-8.187671059139342,-8.166918155198472,-8.146315002049816,-8.125859724584426,-8.105550453288165,-8.085385325504513,-8.065362486626347,-8.045480091219128,-8.02573630407795,-8.00612930122085,-7.986657270820653,-7.967318414077633,-7.948110946035219,-7.929033096340844,-7.910083109954081,-7.891259247804074,-7.872559787398211,-7.853983023384032,-7.83552726806624,-7.817190851880562,-7.798972123826347,-7.780869451859523,-7.762881223247638,-7.745005844888612,-7.727241743594702,-7.709587366343349,-7.6920411804962665,-7.674601673988183,-7.657267355486816,-7.640036754525195,-7.622908421607843,-7.6058809282918665,-7.588952867244438,-7.572122852277592,-7.55538951836167,-7.53875152161843,-7.522207539294913,-7.505756269719064,-7.489396432238187,-7.473126767141095,-7.456946035564975,-7.4408530193877755,-7.424846521107022,-7.4089253637058965,-7.393088390507387,-7.377334465017259,-7.361662470756508,-7.3460713110842475,-7.330559909011389,-7.3151272070060624,-7.299772166791202,-7.28449376913502,-7.269291013634921,-7.254162918495399,-7.239108520300452,-7.224126873781023,-7.209217051577974,-7.194378144001129,-7.179609258784572,-7.164909520838975,-7.150278072001207,-7.135714070781447,-7.121216692108503,-7.106785127073409,-7.092418582671822,-7.078116281545391,-7.063877461722472,-7.0497013763586,-7.035587293476748,-7.021534495707742,-7.007542280031191,-6.993609957516967,-6.979736853067581,-6.96592230516168,-6.952165665598705,-6.938466299245043,-6.924823583781915,-6.911236909454828,-6.8977056788251065,-6.884229306523392,-6.870807219005525,-6.857438854310521,-6.844123661821136,-6.830861102026976,-6.817650646290109,-6.804491776613545,-6.791383985412508,-6.778326775288551,-6.765319658806505,-6.7523621582747655,-6.739453805528253,-6.726594141714832,-6.71378271708463,-6.701019090782643,-6.688302830644694,-6.6756335129963675,-6.663010722455471,-6.650434051737727,-6.637903101465667,-6.625417479980978,-6.612976803160083,-6.6005806942330745,-6.588228783605988,-6.575920708686377,-6.563656113712257,-6.5514346495841895,-6.5392559737008895,-6.527119749797882,-6.515025647789545,-6.502973343614327,-6.4909625190832525,-6.4789928617315296,-6.467064064673316,-6.455175826459608,-6.443327850939367,-6.431519847123568,-6.419751529052255,-6.408022615664798,-6.396332830672875,-6.38468190243653,-6.37306956384309,-6.361495552188886,-6.349959609063955,-6.338461480239233,-6.327000915556765,-6.315577668822408,-6.304191497701255,-6.292842163615775,-6.2815294316462325,-6.270253070433938,-6.259012852086858,-6.247808552087549,-6.236639949203754,-6.225506825400967,-6.214408965757711,-6.203346158382886,-6.192318194335069,-6.181324867544646,-6.170365974737379,-6.159441315360518,-6.148550691510842,-6.137693907864584,-6.126870771609396,-6.116081092378393,-6.105324682185781,-6.094601355364569,-6.083910928505958,-6.0732532204005185,-6.0626280519810365,-6.052035246267099,-6.041474628311252,-6.030946025146893,-6.020449265737682,-6.009984180928134,-5.999550603396189,-5.9891483676071555,-5.978777309768728,-5.968437267787573,-5.9581280812274064,-5.94784959126822,-5.93760164066675,-5.9273840737182555,-5.917196736219572,-5.907039475433187,-5.89691214005238,-5.886814580168041,-5.8767466472357475,-5.866708194044366,-5.856699074685875,-5.846719144525643,-5.836768260173986,-5.826846279458736,-5.816953061398474,-5.807088466176828,-5.797252355117546,-5.787444590660466,-5.777665036338266,-5.767913556753919,-5.758190017559065,-5.748494285432945,-5.73882622806214,-5.729185714121042,-5.719572613252909,-5.709986796051609,-5.700428134043851,-5.690896499672447,-5.681391766279617,-5.671913808091269,-5.662462500201422,-5.653037718557706,-5.64363933994696,-5.634267241981438,-5.6249213030855,-5.61560140248281,-5.60636422891946,-5.193665321190553,-4.536004578457254,-3.8650374493383506,-3.203821143159147,-2.5628690915467067,-1.962832242456359,-1.4172252036267496,-0.9384583184219603,-0.5377257732910287,-0.22191509647491203,0.007590156316468214,0.15614288397518625,0.23576213482885058,0.26198091888481045,0.2490349764816653,0.20845445065785487,0.14915613646236475,0.0773122940037205,-0.0025584151539554695,-0.0867353970799357,-0.17196861591082524,-0.2555213461407745,-0.3353106802936855,-0.40993212246584526,-0.47853999161714544,-0.5406876523810125,-0.5962215596845248,-0.64523678563115,-0.6880441133139626,-0.72510628039223,-0.7569471264165657,-0.7840641077061364,-0.8068723326588128,-0.8256857121480684,-0.8407248123773466,-0.8521370757217884,-0.8600198255188235,-0.8644426636699838,-0.8654677355881082,-0.8631670801127868,-0.8576352729971269,-0.8489966305062442,-0.8374069636373874,-0.8230510210245836,-0.8061370723618813,-0.7868899033116104,-0.7655433843356656,-0.7423333353358932,-0.7174914005741742,-0.6912402301694729,-0.6637901327125788,-0.6353370314383737,-0.6060614574573655,-0.5761282856362315,-0.5456869494774161,-0.5148719928036044,-0.48380382008660033,-0.4525895806510324,-0.4213240891061405,-0.39009072874784395,-0.35896229221117426,-0.32800175289925926,-0.29726298632363374,-0.26679146138867793,-0.23662492585912653,-0.20679408494030246,-0.17732327141294818,-0.14823109291210712,-0.11953105045122098,-0.09123212387815276,-0.06333932453307545,-0.03585421662524917,-0.00877540544745235,0.017901008580245334,0.04418101358249064,0.07007230753737943,0.09558395677185294,0.12072608865642816,0.14550961574759896,0.16994598914363193,0.19404697948373362,0.2178244849916265,0.24129036549975344,0.2644563014932263,0.2873336764013391,0.30993348041131,0.3322662338925087,0.3543419287105915,0.37616998580908,0.3977592273979293,0.4191178621155277,0.44025348139864096,0.4611730653939691,0.48188299678518803,0.5023890811103077,0.5226965722881675,0.5428102022306727,0.5627342135418483,0.5824723943967937,0.602028114805197,0.6214043635487448,0.640603785192703,0.6596287166559083,0.6784812229118591,0.6971631314633077,0.7156760652970295,0.7340214740841497,0.7522006634447822,0.7702148221466323,0.7880650471512545,0.8057523664639709,0.8232777597771219,0.8406421769289624,0.8578465542261916,0.8748918287005736,0.8917789503881682,0.9085088927329253,0.9250826612267763,0.9415013004059636,0.9577658993278034,0.9738775956561206,0.9898375784836989,1.0056470900193737,1.0213074262649056,1.0368199368021114,1.0521860238056462,1.0674071403916083,1.0824847884049926,1.0974205157439025,1.112215913309831,1.1268726116676193,1.141392277489242,1.1557766098509274,1.170027336442243,1.1841462097426607,1.1981350032114557,1.2119955075325672,1.2257295269486126,1.2393388757137818,1.2528253746885558,1.2661908480966624,1.279437120457493,1.2925660137056412,1.3055793445042225,1.318478921756513,1.3312665443166087,1.343943998898302,1.3565130581790117,1.3689754790931237,1.3813330013089884,1.3935873458816377,1.4057402140728246,1.4177932863289713,1.4297482214072545,1.4416066556402607,1.4533702023286286,1.4650404512516997,1.4766189682861086,1.4881072951221777,1.4995069490694197,1.510819422941174,1.5220461850102813,1.5331886790269968,1.5442483242923213,1.5552265157784035,1.5661246242906222,1.5769439966640675,1.5876859559897418,1.5983518018650864,1.6089428106639463,1.6194602358227186,1.6299053081382557,1.6402792360748786,1.6505832060776988,1.6608183828895553,1.6709859098699844,1.681086909314244,1.6911224827709608,1.701093711357692,1.7110016560729262,1.720847358104316,1.730631839132129,1.7403561016282196,1.7500211291497607,1.7596278866279347,1.76917732065164,1.7786703597458524,1.7881079146456629,1.7974908785650714,1.806820127461813,1.8160965202980113,1.8253208992967083,1.8344940901950368,1.8436169024942384,1.852690129706269,1.8617145495981189,1.8706909244335832,1.8796200012125255,1.8885025119083967,1.8973391737038927,1.9061306892249132,1.9148777467730465,1.923581020556653,1.932241170920683,1.9408588445752766,1.9494346748232065,1.9579692817862209,1.966463272630236,1.9749172417894556,1.9833317711893512,1.991707430468461,2.0000447771990038,2.008344357106209,2.0166067042863256,2.024832341423217,2.0330217800035033,2.0411755205301025,2.0492940527341243,2.057377855785056,2.065427398499107,2.073443139545601,2.0814255276514024,2.0893750018032784,2.097291991448023,2.105176916690391,2.113030188488662,2.1208522088478454,2.1286433710104027,2.1364040596444624,2.1441346510294776,2.151835513239165,2.1595070063218946,2.167149482478259,2.174763286235992,2.1823487546220357,2.189906217331824,2.1974359968957757,2.2049384088429176,2.212413761861666,2.219862357957825,2.227284492609632,2.2346804549200034,2.2420505277659464,2.2493949879451556,2.25671410631971,2.2640081479571013,2.271277372268292,2.2785220331432106,2.2857423790833384,2.292938653331723,2.3001110940001697,2.3072599341938598,2.3143854021333543,2.321487721273891,2.328567110422251,2.335623783851017,2.342657951410336,2.349669818637245,2.3566595868625853,2.363627453315478,2.3705736112254736,2.3774982499224038,2.3844015549339264,2.391283708080877,2.3981448875703455,2.4049852680865875,2.41180502087998,2.4186043138535824,2.4253833116479173,2.4321421757236137,2.4388810644420413,2.445600133144061,2.4522995342268916,2.458979417219019,2.4656399288532977,2.472281213138298,2.4789034114278174,2.4855066624887714,2.492091102567241,2.498656865452954,2.505204082542131,2.511732882898678,2.518243393313906,2.524735738364595,2.5312100404697353,2.5376664199456136,2.544104995059527,2.5505258820822507,2.556929195338822,2.563315047258261,2.569683548421839,2.576034807610073,2.5823689318485,2.5886860264521308,2.5949861950689574,2.601269539721836,2.6075361608497065,2.6137861573474126,2.6200196266044555,2.6262366645427506,2.632437365653321,2.6386218230319685,2.6447901284139057,2.650942372207544,2.6570786435271914,2.663199030224908,2.669303618921523,2.675392495036634,2.6814657428178412,2.6875234453691714,2.693565684678658,2.699592541645134,2.7056040961042367,2.7116004268537255,2.717581611678008,2.723547727372061,2.7294988497644557,2.735435053739927,2.7413564132612307,2.747263001390288,2.7531548903088203,2.759032151338273,2.76489485495928,2.7707430708304157,2.7765768678065923,2.782396313956668,2.7882014765806953,2.79399242222682,2.799769216707283,2.8055319251142476,2.811280611835168,2.8170153405674965,2.822736174333135,2.828443175492431,2.8341364057577625,2.839815926206619,2.8454817972945263,2.85113407886731,2.8567728301732074,2.8623981098744715,2.8680099760588185,2.873608486250473,2.8791936974206505,2.884765665998192,2.8903244478794377,2.895870098438201,2.9014026725350295,2.906922224526672,2.912428808274829,2.917922477155061,2.9234032840650808,2.92887128143305,2.9343265212254663,2.9397690549550006,2.9451989336879776,2.9506162080518044,2.9560209282418937,2.9614131440286844,2.9667929047644623,2.972160259389618,2.9775152564393648,2.982857944049642,2.9881883699631717,2.9935065815355113,2.998812625740553,3.0041065491761207,3.0093883980693117,3.0146582182818924,3.019916055315336,3.0251619543158097,3.030395960079009,3.0356181170548906,3.0408284693523537,3.0460270607437687,3.051213934669254,3.0563891342409484,3.061552702247337,3.066704681157323,3.0718451131239437,3.07697403998854,3.0820915032845253,3.087197544240845,3.092292203785996,3.097375522551285,3.1024475408743375,3.1075082988026126,3.112557836096504,3.117596192232852,3.1226234064078975,3.127639517540485,3.132644564275219,3.1376385849852113,3.1426216177752377,3.1475937004844843,3.152554870689449,3.1575051657065387,3.1624446225950065,3.1673732781594657],"type":"scatter","name":"source2_θ","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]}],{"xaxis":{"title":{"text":"Time in Seconds"}},"template":{"layout":{"coloraxis":{"colorbar":{"ticks":"","outlinewidth":0}},"xaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"hovermode":"closest","paper_bgcolor":"white","geo":{"showlakes":true,"showland":true,"landcolor":"#E5ECF6","bgcolor":"white","subunitcolor":"white","lakecolor":"white"},"colorscale":{"sequential":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]],"diverging":[[0,"#8e0152"],[0.1,"#c51b7d"],[0.2,"#de77ae"],[0.3,"#f1b6da"],[0.4,"#fde0ef"],[0.5,"#f7f7f7"],[0.6,"#e6f5d0"],[0.7,"#b8e186"],[0.8,"#7fbc41"],[0.9,"#4d9221"],[1,"#276419"]],"sequentialminus":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]},"yaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"shapedefaults":{"line":{"color":"#2a3f5f"}},"hoverlabel":{"align":"left"},"mapbox":{"style":"light"},"polar":{"angularaxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","radialaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"autotypenumbers":"strict","font":{"color":"#2a3f5f"},"ternary":{"baxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","caxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"aaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"annotationdefaults":{"arrowhead":0,"arrowwidth":1,"arrowcolor":"#2a3f5f"},"plot_bgcolor":"#E5ECF6","title":{"x":0.05},"scene":{"xaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"zaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"yaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"}},"colorway":["#636efa","#EF553B","#00cc96","#ab63fa","#FFA15A","#19d3f3","#FF6692","#B6E880","#FF97FF","#FECB52"]},"data":{"barpolar":[{"type":"barpolar","marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"carpet":[{"aaxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"},"type":"carpet","baxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"}}],"scatterpolar":[{"type":"scatterpolar","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"parcoords":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"parcoords"}],"scatter":[{"type":"scatter","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2dcontour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2dcontour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattercarpet":[{"type":"scattercarpet","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"mesh3d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"mesh3d"}],"surface":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"surface","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattermapbox":[{"type":"scattermapbox","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergeo":[{"type":"scattergeo","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram":[{"type":"histogram","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"pie":[{"type":"pie","automargin":true}],"choropleth":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"choropleth"}],"heatmapgl":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmapgl","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"bar":[{"type":"bar","error_y":{"color":"#2a3f5f"},"error_x":{"color":"#2a3f5f"},"marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"heatmap":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmap","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contourcarpet":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contourcarpet"}],"table":[{"type":"table","header":{"line":{"color":"white"},"fill":{"color":"#C8D4E3"}},"cells":{"line":{"color":"white"},"fill":{"color":"#EBF0F8"}}}],"scatter3d":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"scatter3d","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergl":[{"type":"scattergl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2d","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scatterternary":[{"type":"scatterternary","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scatterpolargl":[{"type":"scatterpolargl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}]}},"legend":{"yanchor":"bottom","xanchor":"right","y":1.02,"orientation":"h","x":1},"margin":{"l":100,"b":80,"pad":10,"r":80,"t":100},"plot_bgcolor":"#f1f3f7","yaxis":{"title":{"text":"State values"}},"yaxis2":{"titlefont":{"color":"orange"},"overlaying":"y","title":"Action values","side":"right"},"height":550,"width":800},{"showLink":false,"editable":false,"responsive":true,"staticPlot":false,"scrollZoom":true});
});





</script>
```




_______________________________________________________________________________
### Analysis

- As expected, the inverter in PQ mode behaves in a grid-following manner.
- Notice that the frequency of the network is not exactly 50 Hz. This is due to the frequency droop mechanism.
- Neither does the voltage magnitude in steady state equal 230 V. This is due to the voltage droop mechanism.

_______________________________________________________________________________
### References

- N. Pogaku, M. Prodanovic and T. C. Green, "Modeling, Analysis and Testing of Autonomous Operation of an Inverter-Based Microgrid," in IEEE Transactions on Power Electronics, vol. 22, no. 2, pp. 613-625, March 2007, doi: 10.1109/TPEL.2006.890003.
- J. M. Guerrero, J. C. Vasquez, J. Matas, L. G. de Vicuna and M. Castilla, "Hierarchical Control of Droop-Controlled AC and DC Microgrids—A General Approach Toward Standardization," in IEEE Transactions on Industrial Electronics, vol. 58, no. 1, pp. 158-172, Jan. 2011, doi: 10.1109/TIE.2010.2066534.
