# PQ Mode - Controllable Source and Load

|  | Mode | Description |
| --- | --- | --- |
| 1 | "Swing" | Ideal voltage source without dynamics (i.e., an infinite bus) |
| 2 | __"PQ"__ | __Grid following controllable source/load (real and imaginary power)__ |
| 3 | "Droop" | Simple grid forming with power balancing through a droop mechanism |
| 4 | "Synchronverter" or "VSG"|  Grid forming control mimicking a generator, i.e., virtual synchronous generator |

- The control structure of power electronic inverters can be divided into cascading levels. 
- The lowest of these employed in the package is an "inner" current control loop, on top of which all other levels are built.

## Summary

- The following example is intended to introduce you to the control mode which will enable the inverter to act like a controllable source or load.
- The mode takes as input the active power (P, Watts) and the reactive power (Q, VAr) as set points.
- Most solar photovoltaic resources, wind power plants and variable loads can be represented by this mode.
- An inverter in this control mode must be placed in a network with other "grid-forming" sources (e.g., swing, droop, or VSG) as it is "grid following".

### Theory

An inverter in the PQ mode is effectively controlled as a current supply, only ever regulating the current exchanged with the grid. This current control loop operates in the direct-quadrature-zero (DQ0) frame, and employs over voltage and current limitations to simulate the protection of the switches (not depicted). This limitations cause non-linear saturation effects when the inverter is pushed to extreme. 


![](assets/current_control.jpg)

The main disadvantage of an inverter in this mode is that the output voltage is maintained by the external network, which means that the inverter needs to synchronize to the network by making use of a phase-locked-loop (PLL). The PLL extracts the angle (and frequency) of the positive phase sequence voltage measured after the first filter inductor. 

A possible disadvantage to the PQ mode is that the inverter may continue injecting currents into the grid when there is a network fault, leading to excessively high voltages. By the same token, when the network voltage is low, the inverter will continue drawing the same amount of power, further deteriorating the voltage profile. Other modes, such as the "Droop" and in particular the "VSG" mode can alleviate some of these issues, providing control structures which regulate both the grid frequency and voltage.  

![](assets/PQ_mode.jpg)



```julia
using ElectricGrid;
```

_______________________________________________________________________________
### Network and Control Configuration 

The following example network structure is considered:
- The PQ-controlled inverter is connected to an external network represented by a source in "Swing" mode.
- The "strength" of this external network is quantifiable by its _fault_ _level_.
- The effective impedance of the external network can be computed from the short-circuit power, the X/R ratio, and the nominal rms voltage.
- Setting I_kp and I_ki (the PI gains) is optional.
- If not provided they will be automatically determined via a loop-shaping method.
- The kp's and ki's for all PI controllers are equal.

With this configuration, a time domain simulation is conducted below.


```julia
# total run time, seconds
t_end = 0.16     

# Connectivity Matrix
CM = [ 0. 1.
        -1. 0.]     

R, L = FaultLevel(300e3, 0.5, 230) # short-circuit power [VA], X/R ratio, nominal rms voltage [V]

parameters = Dict{Any, Any}(
        "source" => Any[
                        Dict{Any, Any}("pwr"    => 300e3,    
                                        "mode"  => "Swing", 
                                        "R1"    => R,       # Equivalent external network resistance [Ω]
                                        "L1"    => L)       # Equivalent external network inductance [H] 
                        Dict{Any, Any}("pwr"    => 100e3,   # Rated apparent power [VA]
                                        "mode"  => "PQ",    # Controller mode
                                        "p_set" => 50e3,    # Real power set point (generating) [W] 
                                        "q_set" => -40e3,   # Imaginary power set point (inductive) [VAi]  
                                        "I_kp"  => 0.01,    # Current proportional gain [V/A] (**optional**)
                                        "I_ki"  => 5.0)     # Current integral gain [V/A⋅s] (**optional**)
                        ],
        "cable"   => Any[
                        Dict{Any, Any}("R"  => 0.1, 
                                        "L" => 0.25e-3, 
                                        "C" => 0.1e-4),
                        ],
        "grid"   => Dict{Any, Any}("process_start" => 0.06) # Time to wait before accepting p_set and q_set [s]
    );

env = ElectricGridEnv(CM = CM, parameters = parameters, t_end = t_end, verbosity = 2);

agents = SetupAgents(env);

hook = Simulate(agents, env);
```

    ┌ Info: Normalization is done based on the defined parameter limits.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\electric_grid_env.jl:329
    ┌ Info: Time simulation run time: 0.16 [s] ~> 1601 steps
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\electric_grid_env.jl:330
    ┌ Info: 2 'classically' controlled sources have been initialised.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2686
    ┌ Info: 1 source has been set up in Swing mode.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2695
    ┌ Info: 1 source has been set up in PQ mode.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2695
    

_______________________________________________________________________________
### Low-Level Rendering


```julia
RenderHookResults(hook = hook, 
                    states_to_plot  = ["source2_i_L1_a"], # Inductor current [A]
                    actions_to_plot = ["source2_u_a"],    # Inverter voltage [V]
                    )
```


```@raw html
<div
id = 61SPR3ru6GKu > </div>
<script>
gd = '61SPR3ru6GKu'
require(['plotly'], function(plotly) {
























plotly.newPlot(gd,[{"mode":"lines","y":[0.0,0.0,-0.000954637140269403,-0.018095894918262304,-0.06598731579052096,-0.13267619390597757,-0.20538755399766692,-0.2665512493507514,-0.31584856753386165,-0.3518736402182066,-0.37709937540433847,-0.387811761995197,-0.38075593043832145,-0.35542890102970437,-0.31302104264589464,-0.2593467892907285,-0.19699680243698892,-0.13145402483233404,-0.06429033245155,-0.00044300289370588497,0.05773726057203177,0.1062850938218658,0.1435143399082796,0.16830093960081827,0.18047397922660138,0.18125684247386586,0.17145579816688686,0.15374300287171888,0.1298091992576103,0.10307114479732463,0.07565888976384165,0.05062097365611319,0.02976769393687205,0.014987552066989554,0.007343141449552323,0.007317740990221422,0.015030272329784766,0.029640180528350976,0.05034737899349803,0.07539887876189626,0.1033876098125508,0.13225658691269598,0.16043941632799968,0.1861535764564905,0.2081101384271608,0.22522614592055196,0.23681874929046964,0.24269487637072343,0.24290614210172629,0.23807544197317265,0.22889603114307838,0.216527671180836,0.20204615934640446,0.18676828804681941,0.1718171228037797,0.15829843063216684,0.1470729135947803,0.1387752222166778,0.13380316092014907,0.13222071689230344,0.13391167095213075,0.13843208782183214,0.14524490628909698,0.15358450179027933,0.16268844090246287,0.17171463935196174,0.17990850624059063,0.1865844326631496,0.19119942042842297,0.1933860329812206,0.1929355877346976,0.18985598335562814,0.18429061415774256,0.17657338923263483,0.16712008288144137,0.15646137234062318,0.14514168655652748,0.13372381188184257,0.12271991919437064,0.11257297607410777,0.1036304370072273,0.0961147683030028,0.09013688008519716,0.08566927792293788,0.08258690418126968,0.08065272164131088,0.07956956852557995,0.07898164710756617,0.07852194154426617,0.07782676873229244,0.07656780632318738,0.07447266743810177,0.07133672061887608,0.06704157906444164,0.0615478406200195,0.05490523599711172,0.04723182791435454,0.038713229834127394,0.02957598576653564,0.020077251526095297,0.010479991644627651,0.0010374715715809613,-0.008023962903450865,-0.016520085283764585,-0.024315535560607282,-0.03133396226831753,-0.03755443522409574,-0.043013914554266286,-0.04779593317957875,-0.05202544248192842,-0.05585380409231788,-0.05944837100142508,-0.06297807451935966,-0.06660140378311707,-0.07045606225186263,-0.07464905420625523,-0.07925236649003176,-0.08429781515192902,-0.0897789443139907,-0.09565158851892308,-0.10184062480471279,-0.10824544623541243,-0.11474919697266839,-0.12122704450842836,-0.1275552370315729,-0.13361951939917538,-0.13932175332122498,-0.14458609902440245,-0.14936184643520123,-0.15362589718041658,-0.15738158413582687,-0.16065719921726856,-0.1635016229359173,-0.16597977514531737,-0.16816658419646846,-0.17014098426082702,-0.1719799861144522,-0.1737530473204179,-0.17551776072498207,-0.17731609657282132,-0.1791726203197783,-0.18109340764469564,-0.18306693807272356,-0.1850656545914034,-0.18704896958327688,-0.18896672373771023,-0.19076326194704463,-0.19238162188918484,-0.19376748589082798,-0.19487287558503003,-0.1956589621297254,-0.1960983275108169,-0.19617603604021763,-0.19589002596333982,-0.1952503720553424,-0.19427792992648002,-0.1930022026465421,-0.19145881662063918,-0.1896867227111254,-0.18772532625257043,-0.18561184078463533,-0.18337888653948575,-0.18105267648326606,-0.17865167122283376,-0.17618597576714645,-0.17365728249319168,-0.17105949171524792,-0.1683798000254727,-0.16560023247269795,-0.16269944379437373,-0.15965464595018874,-0.1564435515210112,-0.15304613712398762,-0.1494461897745023,-0.14563245679514303,-0.14159942788847135,-0.1373476383436647,-0.13288356877931906,-0.12821912094783094,-0.12337076753560061,-0.11835843737259899,-0.1132042322488418,-0.10793108718297262,-0.1025614486309403,-0.09711609184651904,-0.09161311748033918,-0.08606722089505858,-0.08048923592177881,-0.07488599570159482,-0.06926047854838746,-0.06361222585164739,-0.0579379775005604,-0.0522324681273334,-0.046489322288991275,-0.04070197072019899,-0.03486453364469955,-0.028972597283875245,-0.023023849334455226,-0.017018523880751868,-0.010959647399952699,-0.00485307118807099,0.0012926929428083156,0.007466814216171347,0.013656626419731088,0.019848209873695843,0.026027028823920606,0.03217857199511309,0.03828895713654967,0.04434545417014468,0.050336901122411515,0.05625398589187158,0.06208938656238905,0.06783776670014441,0.07349563679806814,0.07906109958959129,0.08453350402741733,0.08991303932455286,0.09520029962340693,0.10039585405829698,0.10549984999185957,0.11051167769985781,0.11542971455884365,0.12025116370529987,0.12497199182810603,0.1295869654140993,0.13408977695260327,0.13847324693956775,0.14272958370304445,0.14685067892287418,0.15082841715760242,0.15465497622926616,0.15832309919692378,0.16182632008547126,0.1651591313351637,0.16831708456951744,0.1712968222515293,0.174096042333435,0.1767134026355936,0.17914837570041225,0.18140106720081856,0.1834720133482618,0.18536197253353964,0.18707172663746935,0.18860190510945563,0.18995284312018013,0.19112448151613265,0.19211631323859368,0.19292737712677827,0.19355629669146723,0.19400135850938882,0.19426062225489227,0.19433205289620226,0.19421366430972492,0.1939036637147567,0.1934005866204847,0.1927034135147243,0.19181166112093295,0.19072544340433062,0.1894454998226005,0.18797319077859687,0.18631046254471473,0.18445978584845035,0.18242407398352956,0.18020658725835137,0.17781083124537433,0.17524045614319483,0.17249916413228683,0.1695906305275996,0.1665184432612996,0.16328606361859643,0.15989680948410934,0.1563538606984638,0.15266028456281,0.148819078263559,0.14483322394280806,0.14070575155242304,0.1364398043211945,0.13203870182838973,0.1275059961039311,0.12284551695901635,0.11806140371098417,0.1131581215899955,0.10814046228669882,0.103013529227441,0.09778270920591736,0.0924536328328105,0.0870321269130252,0.08152416221217829,0.07593580019942614,0.07027314218887115,0.06454228393231043,0.058749278138019,0.052900106683593684,0.04700066350124832,0.04105674829909578,0.035074070515477285,0.029058262211099528,0.023014898061378908,0.0169495202132266,0.010867665569732399,0.004774893034467098,-0.0013231915982889753,-0.007420915121143956,-0.013512522344807376,-0.019592168556955336,-0.025653919526557267,-0.0316917590094592,-0.03769960319009158,-0.043671321037516364,-0.049600759202867106,-0.0554817698441874,-0.06130823965648213,-0.06707411839879995,-0.0727734453454405,-0.07840037232362351,-0.08394918231499338,-0.08941430296557762,-0.09479031473862487,-0.10007195383110234,-0.10525411032687848,-0.11033182235896288,-0.11530026727859055,-0.12015475097215944,-0.12489069652014395,-0.1295036333591062,-0.13398918799451898,-0.13834307713264438,-0.14256110386969345,-0.1466391573150252,-0.15057321575369498,-0.1543593531896026,-0.15799374887461615,-0.16147269923355245,-0.1647926314529471,-0.16795011791871972,-0.17094189066697707,-0.1737648550510812,-0.17641610191953538,-0.17889291773493643,-0.18119279223135065,-0.18331342339189688,-0.1852527197176374,-0.1870087999378951,-0.18857999046995233,-0.18996482106267887,-0.19116201914710784,-0.19217050346318526,-0.19298937753583145,-0.19361792353596152,-0.19405559699073321,-0.19430202270628105,-0.19435699214665714,-0.19422046238314272,-0.19389255659779314,-0.19337356600483047,-0.19266395294911243,-0.1917643548599539,-0.19067558868554146,-0.18939865540918432,-0.18793474425413725,-0.18628523621686588,-0.18445170662465893,-0.1824359264875346,-0.18023986250095092,-0.17786567564631262,-0.17531571842579696,-0.1725925308495604,-0.1696988353619555,-0.16663753094448097,-0.16341168666569272,-0.16002453495887026,-0.156479464899709,-0.15278001572860767,-0.14892987081987474,-0.1449328522460238,-0.14079291602440724,-0.1365141480703409,-0.13210076082025335,-0.12755709043405705,-0.12288759444161501,-0.11809684966632439,-0.11318955024067812,-0.10817050552537477,-0.10304463775338935,-0.0978169792430128,-0.09249266905605857,-0.08707694901706908,-0.0815751590524938,-0.07599273185265076,-0.07033518690109464,-0.0646081239519729,-0.05881721606482948,-0.05296820232622146,-0.04706688039719653,-0.041119099025769046,-0.035130750653979526,-0.02910776423169814,-0.02305609832550637,-0.016981734582198982,-0.010890671576233135,-0.004788919040790596,0.0013175075471680842,0.008377229220196014,0.03161621762006073,0.08559197595451495,0.15834511619222402,0.2370869018024253,0.3042205661401294,0.3593879280426604,0.4011495234861311,0.43196994293831814,0.44816079752349225,0.4464976070580011,0.4264942218577654,0.3893438413017365,0.34087279765471745,0.28370371147843665,0.22336565907028838,0.16146316434003571,0.10293967670949167,0.050126432073470006,0.006959234804956083,-0.024907226045191244,-0.04438514188406284,-0.051345562556445845,-0.047060925021428224,-0.032385020369176454,-0.010030977328699286,0.0182874909921249,0.04914082079162929,0.08039486051857292,0.10899679421615094,0.13313413205554103,0.15091909653024027,0.16129684774203634,0.16380139278314176,0.15833924817091594,0.14578474013312992,0.1269744241978915,0.1036996495099014,0.07740094288551591,0.05016268165090498,0.023567214800369496,-0.00059847465691798,-0.021053720833728695,-0.036734564105361134,-0.046985260450576716,-0.05164534747847627,-0.05080055802053374,-0.04510769521952562,-0.03529028930479483,-0.022534280249513483,-0.007935420045492414,0.007175507773523604,0.021668499967535282,0.03443761073053042,0.044629044662329864,0.051621803503759445,0.05503782019654979,0.05483805014295906,0.0511670526533865,0.04449891344597279,0.03540014687485149,0.024663299717905485,0.01307291823199952,0.0014870870125349711,-0.009330883753404517,-0.01869415050110912,-0.026065686795228693,-0.031090317642342047,-0.03357650202335982,-0.03355257073930318,-0.031183756151073272,-0.026825522799318602,-0.020913173209034408,-0.013993540836647611,-0.0066226326057483575,0.0006305096579546365,0.007253256583945156,0.012808014229409359,0.016958240563122562,0.019497270873094052,0.020333939338548657,0.019518267509752718,0.017199292122289656,0.013638126967828512,0.00915476841224283,0.004125436879971911,-0.0010658152514069208,-0.00604232075570188,-0.010467186614768753,-0.014063880039427268,-0.01662774915366061,-0.018043923366874104,-0.01827924275672199,-0.017391438564328787,-0.015507285922208958,-0.01282077778623325,-0.00956553033631824,-0.0060036314436962,-0.002400189441728262,0.0009925936317851658,0.003952823727957799,0.006304048462034917,0.007921716361059428,0.008742833602182642,0.008761743319147808,0.008031860857784956,0.006653580195637776,0.004768356070949498,0.0025429955701135043,0.00015864820981947858,-0.002204057233897494,-0.004377316969496725,-0.006216605695360311,-0.007610525851577066,-0.008484903196662952,-0.008807581481226734,-0.00858606166332665,-0.007866412120055788,-0.006725963660438428,-0.005267299211721754,-0.003608534398197305,-0.0018746467710291093,-0.000188128722508063,0.0013395933196340692,0.0026148144694002617,0.0035676433313729383,0.004154631493106686,0.004360963060290768,0.004198915913055767,0.0037059953012645863,0.0029401616947355436,0.0019749015785084367,0.0008928623387201018,-0.00022041804063277493,-0.001282627177048079,-0.0022204060918951107,-0.0029736679222064355,-0.0034992771380256693,-0.0037726768293779855,-0.0037887581248812197,-0.0035607079476692727,-0.0031181689221997377,-0.002503951438170837,-0.0017703118538475726,-0.0009746501321586942,-0.00017514567727966446,0.0005733095123501286,0.001222855019856972,0.0017354561716590444,0.0020850887646014504,0.002258681986688993,0.0022563418659364842,0.002090419942697319,0.0017839533792108897,0.00136833185107188,0.0008805921888530577,0.0003604683716806889,-0.0001525904197847705,-0.0006221330931113211,-0.0010171279325597027,-0.0013137500333391117,-0.0014966578670430053,-0.0015594927243691714,-0.0015048057684787786,-0.001343291479704689,-0.00109254831221417,-0.0007754010677758494,-0.0004179695430810426,-4.7634273832919405e-5,0.0003089841427885709,0.0006278326768542897,0.0008889823091355886,0.0010777552563414616,0.0011854552680090613,0.0012095954274609078,0.0011537057663550954,0.0010267078566252863,0.0008419622404298721,0.0006160576304412579,0.0003674447866677255,0.0001150323104512907,-0.00012317732099820233,-0.0003312882073829293,-0.0004965076229193066,-0.0006098467460406987,-0.0006665268010248582,-0.0006660521922530194,-0.000611987812854764,-0.0005114590768067195,-0.0003744348373957429,-0.00021285517988234304,-3.96705804264208e-5,0.00013212607805002108,0.0002904181617591206,0.000424759583184553,0.0005270106323571611,0.0005917671428807836,0.0006165770424826356,0.0006019328295896464,32.164325857439955,17.949689879145264,40.28483986388068,50.085310585777385,57.663971046294925,45.827949481379285,40.72026049162465,35.95998609574124,38.24371778533244,34.488002604710665,33.67945360600693,30.040794460043173,21.957897035002485,11.118498880334084,0.7867814626577485,-8.024517729740174,-16.448235167754664,-25.78705255362948,-36.46164020550136,-47.24181666287199,-57.24518869343202,-65.71932139802911,-72.93702671680856,-78.8211331484866,-83.36880489586301,-86.22414912007659,-87.29419118160176,-86.73368397237117,-84.84395072341921,-82.08224451342562,-78.64899087603055,-74.85408760794948,-70.76410212531847,-66.6607725821095,-62.679193440638244,-59.119357293302365,-56.15262779413646,-53.93790787466707,-52.59343833285601,-52.221274293052595,-52.96471701538521,-54.85808867623982,-57.94712600203302,-62.110069753509244,-67.23245379617661,-73.07833917248661,-79.44091225893726,-86.05679447985818,-92.69215933005489,-99.13194638428217,-105.19363072332916,-110.74690127771223,-115.68686858965113,-120.00283366675758,-123.67841000931416,-126.74270384019394,-129.2079111240542,-131.1310653943165,-132.57159067790667,-133.61014113658135,-134.31827909476675,-134.7496985054588,-134.9458239418077,-134.9223088409138,-134.70012765165032,-134.28519505032662,-133.70052194556774,-132.96160511477277,-132.09824747080518,-131.1336268260026,-130.0938147174829,-128.99836525003258,-127.85951014652176,-126.69025766891781,-125.50163697251094,-124.3085611305093,-123.1171042192729,-121.93624441131819,-120.76587951716755,-119.60185739675191,-118.42210742396492,-117.18783502072289,-115.84058581380214,-114.30753249555211,-112.51012434361787,-110.37340832282351,-107.84143264812154,-104.88109880840048,-101.49051976351666,-97.69325093226097,-93.54352388613233,-89.11971705717671,-84.50199600354611,-79.7617383816069,-74.97679190011783,-70.23618634658109,-65.63293356575004,-61.25598499060623,-57.180562391228364,-53.465354692741585,-50.151380845517586,-47.26357144794117,-44.804869592282515,-42.75732513341819,-41.08502807700226,-39.73751510285528,-38.65144972661136,-37.751400553732374,-36.95262416350604,-36.16403028738094,-35.28039499617369,-34.19697452868238,-32.83775088723214,-31.138023597534353,-29.03257108513593,-26.458328103108787,-23.363719814530405,-19.715511205280148,-15.500050205091256,-10.722136754027835,-5.402955708516597,0.4186537138802999,6.680966074168306,13.297211533302232,20.161136114388718,27.155831251489445,34.16274114752279,41.06875471542128,47.77152225380906,54.17402494972108,60.19053783305715,65.77025574020988,70.89094103565965,75.53797908799172,79.69822323090142,83.36508189241107,86.54442236909226,89.2546600492921,91.52194943806849,93.37787498343303,94.86029229800573,96.01366917089406,96.88884816068744,97.54205460639417,98.03338432033043,98.42484140385733,98.77825451959303,99.1530161292949,99.60429075764989,100.18751317846055,100.94924917998449,101.91444032235486,103.09703368733099,104.50548402807516,106.1407290792213,107.99125794609068,110.02958539767239,112.21308866566999,114.48747140845414,116.78953711729987,119.05003649230557,121.19779284484575,123.16481824265978,124.89150492498536,126.3308459159995,127.45088447205578,128.23536870215506,128.68302377991205,128.80599969881675,128.6288826901982,128.18516806382118,127.51262417839303,126.65086843719975,125.6391742803445,124.51442856196331,123.31009296676572,122.05517510358426,120.77228315055332,119.47682546024105,118.17754185861921,116.87749348897151,115.57477077675259,114.26273908962827,112.93010264598531,111.56119119749853,110.13658584310411,108.63395989274926,107.02891649693174,105.29445385309593,103.4003076612851,101.31884159889273,99.02875823577622,96.50973935779865,93.74132550818526,90.70557731439052,87.38946493616824,83.78694949161029,79.90125760931441,75.74515047804111,71.3404645064989,66.71763665472642,61.91500498787742,56.977525073156684,51.95489532434725,46.899328783771296,41.86329271990896,36.89743064947342,32.04877758556378,27.360217795497878,22.872145304773362,18.614834445521737,14.602056839107538,10.84035477177831,7.3337368801364065,4.080936509059434,1.069994302580603,-1.7224777149934467,-4.328426288457765,-6.785057918350437,-9.13304219794361,-11.415961064771428,-13.679787813642529,-15.971541980398548,-18.337309290854616,-20.820230697174026,-23.4589116143908,-26.2862423981585,-29.328399788411836,-32.60387281348374,-36.12252422320688,-39.884844092197866,-43.88151548784171,-48.093390800577495,-52.49190164443874,-57.03994655565956,-61.693263722995404,-66.40227127126305,-71.11427278202274,-75.7758560618671,-80.3352732857867,-84.7445883408091,-88.9614342437122,-92.95027731821773,-96.68316880405072,-100.14000784180556,-103.30839259869789,-106.18314710357308,-108.76562490594522,-111.06287926890823,-113.0867777799197,-114.85312180335573,-116.38080838672212,-117.69105572912427,-118.80668806841877,-119.75146664045764,-120.54944073998718,-121.22430184135987,-121.79873125862945,-122.29375074847917,-122.72809627136303,-123.11764258757387,-123.47490550149556,-123.80864161775195,-124.1235602194702,-124.42015534224257,-124.6946687006658,-124.93919580902785,-125.14195519965318,-125.28774124000928,-125.35857817924207,-125.33457833425456,-125.19498475631488,-124.91934969150661,-124.48877104982547,-123.88708878193918,-123.10193530053205,-122.12554520956117,-120.95525409487605,-119.59365215062545,-118.04839458693787,-116.33170317069458,-114.45961508588063,-112.45104736143186,-110.32674830231505,-108.1082047756838,-105.81656883119567,-103.4716592023545,-101.09108417542515,-98.68952036482095,-96.27816908957422,-93.86439820733088,-91.45156573299239,-89.039012994995,-86.62221074274501,-84.19304034216846,-81.74019252152657,-79.2496667262574,-76.70535385251227,-74.0896845547108,-71.38432442064718,-68.57089729403177,-65.63171848602029,-62.5505207933596,-59.313157020253634,-55.908262589693024,-52.327860069668624,-48.56788406583882,-44.628600785788294,-40.51489310060274,-36.236381374623356,-31.80735433409116,-27.246494011491105,-22.57639356004804,-17.82288459676494,-13.014208319175593,-8.180078677751972,-3.3506937089867384,1.4442482774319962,6.176476906798256,10.819982084421637,15.351722545449753,19.752194525659867,24.005854133855077,28.101393980245298,32.031879012613615,35.79474903310063,39.3916966401395,42.828429930826346,46.11432958435785,49.26201025602955,52.28679690275024,55.206127985758265,58.03889957895327,60.80476699085575,63.52342312931909,66.2138748736206,68.89373976024605,71.57858520493123,74.28133152515804,77.01173862311362,79.77599473620627,82.5764243153763,85.41133061896697,88.27498646585424,91.15778301205637,94.04654073857049,96.92497873016464,99.77432808030177,102.57406385374954,105.30271906831288,107.93873548232,110.46130127868292,112.85112607306144,115.09110916500744,117.16686677866154,119.06709665544082,120.783771915868,122.31216884800254,123.65074388914432,124.8008827832493,125.76654949878589,126.55386416169844,127.17063843263469,127.62589401393984,127.92938595714514,128.09114783648138,128.12107130814988,128.0285286370063,127.8220437875582,127.5090157394553,127.09549664240649,126.58602695216264,125.98352942440286,125.28926349815171,124.50284102360138,123.62230346038145,122.64425964356732,121.56408202737465,120.37615794636605,119.07419077598465,117.65154379073373,116.10161692023341,114.41824354729866,112.59609124440222,110.63104738758945,108.52056854591204,106.26397204937376,103.86264965322529,101.32018688937654,98.64237728604118,95.83712754212164,92.91425712840054,89.88520276986763,86.76264409579923,83.5600709432239,80.29131518676385,76.97007062835677,73.60942366970686,70.22141553504208,66.81665403437434,63.403989542432015,59.99026625862344,56.58015612026496,53.17607914819026,49.778210658378875,46.38457277346837,42.99120505608952,39.59240685668843,36.18104209134327,32.74889560913967,29.28706905784403,25.786403209979134,22.23791307122874,18.633221751785076,14.964979016517445,11.227250614069867,7.415864905451285,3.5287039944680103,-0.4340724173461312,-4.469880744589578,-8.573652999797648,-12.737867418290081,-16.952672011003884,-21.20609857190271,-25.484359996496202,-29.772219086625295,-34.05341281114702,-38.31111274824609,-42.52840054008954,-46.6887368541546,-50.77640356066672,-54.77690138794149,-58.67728883894275,-62.46645220004567,-66.13530061109874,-69.67688403157896,-73.08643527916736,-76.36134000322464,-79.50104046320423,-82.50688037054195,-85.38189893478163,-88.13058276780825,-90.75858457971661,-93.27241775843468,-95.67913603610289,-97.98600753549033,-100.20019254331858,-102.32843432754693,-104.37677214313398,-106.35028520098612,-108.25287576890537,-110.08709871003126,-111.85404363369302,-113.5532744285467,-115.1828292583401,-116.73928211661361,-118.21786476067756,-119.61264531198054,-120.9167571063273,-122.12266865480451,-123.22248304860022,-124.20825306372103,-125.07229685472967,-125.80749868717665,-126.40757977564341,-126.86732597801038,-127.18276172761644,-127.35126293110369,-127.37160531196005,-127.24394850042776,-126.96975974128145,-126.55168415158705,-125.99337083150917,-125.29926572120155,-124.47438289656823,-123.52406606472833,-122.45375146418255,-121.26874233270772,-119.97400372688251,-118.57398490422422,-117.07247483606722,-115.47249480147647,-113.77623048182741,-111.98500456354529,-110.09928957007114,-108.11875947542585,-106.04237758640686,-103.86851720424986,-101.59511067914771,-99.2198216473949,-96.74023449244608,-94.15405440697356,-91.45931087258818,-88.65455695007181,-85.73905653336011,-82.71295172548834,-79.57740281041224,-76.33469398090226,-72.98829907786251,-69.54290310199457,-66.00437712871299,-62.37970639531935,-58.6768735962059,-54.90470165115179,-51.07266223324194,-47.190658005608014,-43.26878770802434,-39.31710389395268,-35.34537323727133,-31.362848943425146,-27.37806398410701,-23.39865271964445,-19.431207077081087,-15.481171908240269,-11.552782543384078,-7.649045950440215,-3.771765360012771,0.07839323850131263,3.9017956568174865,7.699699241550422,11.474082934680816,15.227453681113666,18.962635893335616,22.682551184243348,26.389995881648154,30.08742393271688,33.77674268624488,37.45912869952745,41.13487015080447,44.803241645903334,48.462416190122184,52.109417860816265,55.740117279902194,59.34927038091349,62.930599244067686,66.47691200833813,69.9802571550466,73.43210590012364,76.82355514241634,80.14554249376532,83.3890644397318,86.54538868918952,89.60625226576883,92.56403783091412,95.41192202669245,98.14399117986713,100.7553213979338,103.24202179497338,105.60124120584528,107.8311401988723,109.93083142352845,111.90029230123685,113.74025477860829,115.45207732646284,117.0376046104634,118.4990203119918,119.83869847338735,121.05905850854248,122.16242868064913,123.1509224192873,124.02633133864583,124.79003823275958,125.44295266543166,125.98547104522522,126.41746228469535,126.73827929641327,126.94679568950441,127.04146611836778,127.02040782578477,126.88150004902802,126.62249716040542,126.2411507393185,125.73533527051437,125.1031718799399,124.34314449471447,123.45420307233711,122.4358490919042,121.28819931949084,120.01202491021212,118.60876412944333,117.0805082888495,115.42996181547917,113.66037862144225,111.77547804362854,109.77934451847403,107.67631580564324,105.47086495646076,103.16748133699421,100.77055587743666,98.28427535731544,95.7125299872207,93.0588378539402,90.32628899979437,87.51751104931877,84.63465741401517,81.67941823041282,78.65305334455618,75.55644586871861,72.39017412089171,69.15459912843806,65.84996434590514,62.47650381354787,59.034554676801676,55.52466980637977,51.94772621181404,48.30502503512227,44.59837915110037,40.83018478871436,37.0034741217508,33.12194644738334,29.189976361667146,25.212598225140148,21.19546715454631,17.144797735218212,13.067282574651042,8.969993661559858,4.860270209419088,0.745597209601252,-3.3665207317978085,-7.468685562657718,-11.553727324673279,-15.614816706778178,-19.645564707796787,-23.640106101232565,-27.593164099350577,-31.500094386484292,-35.35690749580872,-39.16026930626333,-42.90748020828749,-46.59643420523068,-50.225559863492364,-53.79374558488555,-57.30025213931399,-60.74461575734111,-64.12654533546775,-67.44581744843418,-70.70217289056717,-73.8952183815102,-77.02433687168453,-80.08860957263231,-83.08675242254505,-86.01706918627205,-88.87742279384362,-91.66522585781411,-94.37745059750364,-97.01065766173093,-99.56104260907706,-102.0244981074111,-104.39668928532058,-106.67313913984553,-108.84932050727177,-110.9207508609386,-112.88308612824126,-114.73220982443827,-116.46431407858411,-118.07596956089088,-119.56418188513807,-120.92643272069424,-122.16070456771239,-123.26548888589171,-124.23977798316409,-125.08304173131215,-125.79519075270184,-126.37652819535944,-126.82769256975135,-127.14959435454254,-127.34334919157261,-127.41021048879546,-127.3515041442755,-127.16856790720863,-126.86269761722798,-126.435102225126,-125.8868691104413,-125.21894078738583,-124.43210364298355,-123.52698889223194,-122.50408547653156,-121.3637641855158,-120.10631186094002,-118.73197415693727,-117.24100499644402,-115.63372059161848,-113.91055569878174,-112.07211966673148,-110.11924982000427,-108.0530598014753,-105.87498068313572,-103.5867929365861,-101.1906477268667,-98.68907644030851,-96.08498785965259,-93.38165293418761,-90.58267763293432,-87.69196488796852,-84.7136671069239,-81.65213113553236,-78.51183786412729,-75.29733888319313,-72.0131926951478,-68.66390298147165,-65.2538613105387,-61.787296461351964,-58.268232244937835,-54.700455344187944,-51.08749428184753,-47.432610183125576,-43.73879954184732,-40.008808744135834,-36.24515966696512,-32.45018526466775,-28.626073696963235,-24.774919247859977,-20.89877804459504,-16.999726416492067,-13.079919640237812,-9.141648803657203,-5.18739358560936,-1.2198688939388882,2.7579364770689954,6.742730719668598,10.730900617073909,14.718504682542761,18.7012782037879,22.674650112376074,26.633771146405728,30.57355234203527,34.48871249435138,38.37383288844579,42.22341733292515,46.03195534254915,49.79398622218685,53.50416180394308,57.1573056813425,60.74846696237735,64.27296681642069,67.72643640463039,71.10484514347303,74.4045186392408,77.6221460307303,80.75477687127594,83.79980805533438,86.75496163609766,89.61825467878482,92.38796254131951,95.0625771643414,97.64076208235312,100.12130593562715,102.50307626835732,104.78497534402862,106.96589959709031,109.04470417504015,111.02017381251635,112.89100102572894,114.65577232953885,116.3129628698717,117.86093954116383,119.2979723333233,120.62225333703704,121.8319225422528,122.9250993042169,123.8999181357252,124.75456732306353,125.48732876428262,126.09661739719378,126.58101862291426,126.9393222376632,127.1705515560547,127.27398663525511,127.24918078006846,127.09596981101949,126.81447389609568,126.4050920667356,125.86848984503358,125.20558068833311,124.41750219771967,123.50558822918424,122.47133818399332,121.316384834547,120.04246206301337,118.6513738541438,117.1449657949588,115.52510019830278,113.79363579182132,111.95241270687623,110.00324327201957,107.94790887186765,105.7881628835101,103.5257394576796,101.16236767911204,98.69979042762562,96.13978707552991,93.48419900433385,90.73495680968728,87.89410799228759,84.96384390705073,81.94652476467014,78.84470154862959,75.6611338249085,72.39880257730681,69.06091739298851,65.65091754326254,62.17246674506222,58.629441639197545,55.0259142716804,51.36612910354386,47.65447529226151,43.89545517474712,40.09365003005882,36.25368430330047,32.38018952691945,28.47776918006636,24.550965681626863,20.60423062084842,16.64189919595452,12.668169662138302,8.687088393405022,4.70254094622313,0.718249285562381,-3.262224895617624,-7.235471055573118,-11.198215382844305,-15.147301764092703,-19.079668989908093,-22.99232517554272,-26.882320464816633,-30.746719137539692,-34.58257225227523,-38.38689192697935,-42.15662829115664,-45.888650037142455,-49.57972935868876,-53.22653189712966,-56.825612125063394,-60.373414391683056,-63.866279640322226,-67.30045759562854,-70.67212401337703,-73.97740239842355,-77.21238943325028,-80.37318322759289,-83.45591340409248,-86.45677197952176,-89.3720439878229,-92.19813681998659,-94.93160732479359,-97.56918581994307,-100.1077962998369,-102.54457228775749,-104.87686795895779,-107.1022643493941,-109.21857065451623,-111.22382080603688,-113.11626568491623,-114.8943614797986,-116.55675482679332,-118.10226546496618,-119.52986720962032,-120.8386680810488,-122.02789042978475,-123.09685187140067,-124.04494778656556,-124.87163605818225,-125.57642461058924,-126.15886219025315,-126.61853268784691,-126.95505315324468,-127.16807550316733,-127.2572917714897,-127.22244261005099,-127.0633286184549,-126.7798239696869,-126.37189170871649,-125.83960003714338,-125.18313886099459,-124.40283587251598,-123.4991714605528,-122.472791796941,-121.32451952602997,-120.05536158762098,-118.66651382575668,-117.15936217159172,-115.5354803320506,-113.79662406085545,-111.944722228496,-109.9818650368267,-107.91028983682072,-105.73236510003345,-103.45057316200538,-101.06749239681139,-98.58577949509827,-96.00815250332698,-93.33737524078242,-90.57624364550338,-87.72757451378845,-84.794196994222,-81.77894708065689,-78.68466522402622,-75.51419705514463,-72.27039708466988,-68.9561351268333,-65.57430508477576,-62.127835641263154,-58.61970232254998,-55.05294034789579,-51.430657644665246,-47.75604740019592,-44.03239953699357,-40.26311053673349,-36.4516910995642,-32.60177120604638,-28.71710224665564,-24.801555994403202,-20.859120315495723,-16.893891636393786,-12.910064308265005,-8.911917126786237,-4.903797371880465,-0.890102824043647,3.1247377121836784,7.136284794144181,11.140109022955468,15.131811401809511,19.10704276878719,23.06152174268678,26.991050681513524,30.89152923241622,34.75896514556821,38.589482128877236,42.3793246312691,46.12485955530379,49.82257501077774,53.4690763256804,57.061079625691,60.595403375068926,64.06895833760964,67.47873646423196,70.82179924232658,74.09526605052284,77.29630305101489,80.4221131207191,83.46992727366339,86.43699796210343,89.32059456544718,92.11800128714543,94.82651758364942,97.44346114998888,99.96617338726672,102.39202718217771,104.7184367411768,106.94286914548883,109.06285723067668,111.07601334832042,112.98004353918388,114.77276163799259,116.45210283975206,118.01613628573595,119.46307627242913,120.79129174668597,121.99931382242562,123.0858411351355,124.04974293679071,124.89005992187467,125.60600286138394,126.19694920257987,126.6624378646996,127.00216252218893,127.2159637151417,127.30382015998883,127.26583965118223,127.10224994639611,126.8133900139895,126.39970199306586,125.86172417488926,125.20008526156678,124.41550009605322,123.50876698921279,122.48076669760566,121.33246303266601,120.06490501080229,118.67923038738326,117.17667035808279,115.55855516087617,113.82632027299036,111.98151287076064,110.02579820760938,107.96096556668516,105.7889334599823,103.51175377434693,101.13161460547676,98.65084157215932,96.07189746245567,93.39738012886836,90.63001861804054,87.77266758939152,84.82830014348235,81.79999924211135,78.6909479557,75.50441881732127,72.24376259504639,68.9123968139541],"type":"scatter","name":"source2_i_L1_a","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1,0.1001,0.1002,0.1003,0.1004,0.1005,0.1006,0.1007,0.1008,0.1009,0.101,0.1011,0.1012,0.1013,0.1014,0.1015,0.1016,0.1017,0.1018,0.1019,0.102,0.1021,0.1022,0.1023,0.1024,0.1025,0.1026,0.1027,0.1028,0.1029,0.103,0.1031,0.1032,0.1033,0.1034,0.1035,0.1036,0.1037,0.1038,0.1039,0.104,0.1041,0.1042,0.1043,0.1044,0.1045,0.1046,0.1047,0.1048,0.1049,0.105,0.1051,0.1052,0.1053,0.1054,0.1055,0.1056,0.1057,0.1058,0.1059,0.106,0.1061,0.1062,0.1063,0.1064,0.1065,0.1066,0.1067,0.1068,0.1069,0.107,0.1071,0.1072,0.1073,0.1074,0.1075,0.1076,0.1077,0.1078,0.1079,0.108,0.1081,0.1082,0.1083,0.1084,0.1085,0.1086,0.1087,0.1088,0.1089,0.109,0.1091,0.1092,0.1093,0.1094,0.1095,0.1096,0.1097,0.1098,0.1099,0.11,0.1101,0.1102,0.1103,0.1104,0.1105,0.1106,0.1107,0.1108,0.1109,0.111,0.1111,0.1112,0.1113,0.1114,0.1115,0.1116,0.1117,0.1118,0.1119,0.112,0.1121,0.1122,0.1123,0.1124,0.1125,0.1126,0.1127,0.1128,0.1129,0.113,0.1131,0.1132,0.1133,0.1134,0.1135,0.1136,0.1137,0.1138,0.1139,0.114,0.1141,0.1142,0.1143,0.1144,0.1145,0.1146,0.1147,0.1148,0.1149,0.115,0.1151,0.1152,0.1153,0.1154,0.1155,0.1156,0.1157,0.1158,0.1159,0.116,0.1161,0.1162,0.1163,0.1164,0.1165,0.1166,0.1167,0.1168,0.1169,0.117,0.1171,0.1172,0.1173,0.1174,0.1175,0.1176,0.1177,0.1178,0.1179,0.118,0.1181,0.1182,0.1183,0.1184,0.1185,0.1186,0.1187,0.1188,0.1189,0.119,0.1191,0.1192,0.1193,0.1194,0.1195,0.1196,0.1197,0.1198,0.1199,0.12,0.1201,0.1202,0.1203,0.1204,0.1205,0.1206,0.1207,0.1208,0.1209,0.121,0.1211,0.1212,0.1213,0.1214,0.1215,0.1216,0.1217,0.1218,0.1219,0.122,0.1221,0.1222,0.1223,0.1224,0.1225,0.1226,0.1227,0.1228,0.1229,0.123,0.1231,0.1232,0.1233,0.1234,0.1235,0.1236,0.1237,0.1238,0.1239,0.124,0.1241,0.1242,0.1243,0.1244,0.1245,0.1246,0.1247,0.1248,0.1249,0.125,0.1251,0.1252,0.1253,0.1254,0.1255,0.1256,0.1257,0.1258,0.1259,0.126,0.1261,0.1262,0.1263,0.1264,0.1265,0.1266,0.1267,0.1268,0.1269,0.127,0.1271,0.1272,0.1273,0.1274,0.1275,0.1276,0.1277,0.1278,0.1279,0.128,0.1281,0.1282,0.1283,0.1284,0.1285,0.1286,0.1287,0.1288,0.1289,0.129,0.1291,0.1292,0.1293,0.1294,0.1295,0.1296,0.1297,0.1298,0.1299,0.13,0.1301,0.1302,0.1303,0.1304,0.1305,0.1306,0.1307,0.1308,0.1309,0.131,0.1311,0.1312,0.1313,0.1314,0.1315,0.1316,0.1317,0.1318,0.1319,0.132,0.1321,0.1322,0.1323,0.1324,0.1325,0.1326,0.1327,0.1328,0.1329,0.133,0.1331,0.1332,0.1333,0.1334,0.1335,0.1336,0.1337,0.1338,0.1339,0.134,0.1341,0.1342,0.1343,0.1344,0.1345,0.1346,0.1347,0.1348,0.1349,0.135,0.1351,0.1352,0.1353,0.1354,0.1355,0.1356,0.1357,0.1358,0.1359,0.136,0.1361,0.1362,0.1363,0.1364,0.1365,0.1366,0.1367,0.1368,0.1369,0.137,0.1371,0.1372,0.1373,0.1374,0.1375,0.1376,0.1377,0.1378,0.1379,0.138,0.1381,0.1382,0.1383,0.1384,0.1385,0.1386,0.1387,0.1388,0.1389,0.139,0.1391,0.1392,0.1393,0.1394,0.1395,0.1396,0.1397,0.1398,0.1399,0.14,0.1401,0.1402,0.1403,0.1404,0.1405,0.1406,0.1407,0.1408,0.1409,0.141,0.1411,0.1412,0.1413,0.1414,0.1415,0.1416,0.1417,0.1418,0.1419,0.142,0.1421,0.1422,0.1423,0.1424,0.1425,0.1426,0.1427,0.1428,0.1429,0.143,0.1431,0.1432,0.1433,0.1434,0.1435,0.1436,0.1437,0.1438,0.1439,0.144,0.1441,0.1442,0.1443,0.1444,0.1445,0.1446,0.1447,0.1448,0.1449,0.145,0.1451,0.1452,0.1453,0.1454,0.1455,0.1456,0.1457,0.1458,0.1459,0.146,0.1461,0.1462,0.1463,0.1464,0.1465,0.1466,0.1467,0.1468,0.1469,0.147,0.1471,0.1472,0.1473,0.1474,0.1475,0.1476,0.1477,0.1478,0.1479,0.148,0.1481,0.1482,0.1483,0.1484,0.1485,0.1486,0.1487,0.1488,0.1489,0.149,0.1491,0.1492,0.1493,0.1494,0.1495,0.1496,0.1497,0.1498,0.1499,0.15,0.1501,0.1502,0.1503,0.1504,0.1505,0.1506,0.1507,0.1508,0.1509,0.151,0.1511,0.1512,0.1513,0.1514,0.1515,0.1516,0.1517,0.1518,0.1519,0.152,0.1521,0.1522,0.1523,0.1524,0.1525,0.1526,0.1527,0.1528,0.1529,0.153,0.1531,0.1532,0.1533,0.1534,0.1535,0.1536,0.1537,0.1538,0.1539,0.154,0.1541,0.1542,0.1543,0.1544,0.1545,0.1546,0.1547,0.1548,0.1549,0.155,0.1551,0.1552,0.1553,0.1554,0.1555,0.1556,0.1557,0.1558,0.1559,0.156,0.1561,0.1562,0.1563,0.1564,0.1565,0.1566,0.1567,0.1568,0.1569,0.157,0.1571,0.1572,0.1573,0.1574,0.1575,0.1576,0.1577,0.1578,0.1579,0.158,0.1581,0.1582,0.1583,0.1584,0.1585,0.1586,0.1587,0.1588,0.1589,0.159,0.1591,0.1592,0.1593,0.1594,0.1595,0.1596,0.1597,0.1598,0.1599,0.16]},{"mode":"lines","y":[0.0,0.0,0.0,0.03724632028123725,0.30269339703108966,0.8253324370701486,1.6233435149185165,2.60573803383909,3.73585891997069,4.9818917072665325,6.318539718017509,7.717299064143145,9.101338675708808,10.428050152364955,11.625805083411308,12.68361952443858,13.563638467208872,14.26985463530269,14.786155891971706,15.124208036605028,15.297133299191245,15.329224010679507,15.255789502923466,15.102703192432882,14.909561406490226,14.696429817776277,14.49687661251511,14.322795266166793,14.193461992979515,14.109245490284037,14.070729529723408,14.06732875534516,14.083846257611068,14.102746769132773,14.099528449545556,14.0543822629457,13.94119064740947,13.742906064486565,13.439143177681109,13.01972129452963,12.474673011197346,11.802516542877445,11.004847263230925,10.088374529233757,9.064263131631893,7.94493595863851,6.746904331044849,5.484606795961315,4.174888873666693,2.8304253011862337,1.4640841822754962,0.08392657465061834,-1.3037455293836941,-2.6967720001490494,-4.095847446329056,-5.504252773152267,-6.928131919076675,-8.374307351314911,-9.851538404860909,-11.36751676762341,-12.930426478373311,-14.546176704129135,-16.219720479151622,-17.95325802309788,-19.747025487591653,-21.59875048247468,-23.50387732965922,-25.45613600678615,-27.44735013637967,-29.46870120563884,-31.51033817028702,-33.562822181351756,-35.61673434247043,-37.66385496289424,-39.696888404959054,-41.710106877146565,-43.69922281145282,-45.66143647880485,-47.59542075265726,-49.500889671535944,-51.37863346444252,-53.22983094935855,-55.05608134305837,-56.8587051203571,-58.63874821362294,-60.396466902261515,-62.13131105971444,-63.84169769998065,-65.52499872680276,-67.1776082367109,-68.79496105700098,-70.37181891816566,-71.90233191692482,-73.38042507306017,-74.79989431652992,-76.15477148742441,-77.43942933262097,-78.64883229871687,-79.77861841120941,-80.8251910001894,-81.78575133631618,-82.65823668535714,-83.44129329145453,-84.13410548431881,-84.73631783832948,-85.24782049679644,-85.66864518883195,-85.99877183390018,-86.23803040703348,-86.38597705220606,-86.44183068304656,-86.40444054543606,-86.2722746137735,-86.04347311767623,-85.71588956660331,-85.28720332136514,-84.75499876089442,-84.11689894309137,-83.37065854825208,-82.51428111032341,-81.54609937993341,-80.46484855782064,-79.2697134705467,-77.96034569078192,-76.53686800265538,-74.99983949756907,-73.3502205160682,-71.58930255478421,-69.71864742609165,-67.7400053887116,-65.6552479078943,-63.46629705823277,-61.17507093298143,-58.78344114339339,-56.29320435388122,-53.70607502030311,-51.02368783946341,-48.24762267273042,-45.37943364309776,-42.42069515518874,-39.37304648781315,-36.2382437168589,-33.018205582238686,-29.715056483618735,-26.331160554239514,-22.86914505800065,-19.331914180648248,-15.722648551691062,-12.044796624784059,-8.302052742736434,-4.498330188161128,-0.6377254322593394,3.2755186731012222,7.237052244703765,11.2424549692706,15.287269243607323,19.367027809565805,23.47727380178021,27.61357416787587,31.771524105615395,35.94674552890502,40.13487786321757,44.331564990554234,48.53243778867937,52.73309573960518,56.92908817553295,61.11589749290467,65.28892561200679,69.44348453462165,73.5747923991673,77.6779745110003,81.74807034156295,85.78004501455416,89.76880556608852,93.7092200961007,97.59613937866591,101.42441918635743,105.18894240941874,108.884639759721,112.50650801576505,116.04962532537077,119.50916275275335,122.88039229611319,126.15869102763442,129.33954210915232,132.4185328578434,135.39135086723533,138.25377876682938,141.0016885897691,143.6310365147935,146.1378586842428,148.5182687929485,150.7684577512374,152.88469584339526,154.8633372712127,156.70082713428377,158.393710409969,159.93864263949357,161.33240170987094,162.57190021535993,163.65419778901568,164.57651283753043,165.3362332208035,165.93092542493972,166.35834201375985,166.61642713608683,166.70332013779222,166.61735731912896,166.35707210221236,165.92119386889348,165.30864585384802,164.51854247318386,163.55018647933065,162.40306632170797,161.07685400852841,159.57140374262235,157.8867514666023,156.02311542028372,153.98089766976184,151.76068653231795,149.36325971182737,146.7895879315059,144.0408387987266,141.11838062542094,138.0237859367055,134.75883441031985,131.32551504676852,127.72602739915403,123.96278177374893,120.03839835256316,115.95570527011232,111.71773571649595,107.32772419754974,102.78910211129306,98.10549282442163,93.28070644217746,88.31873445432369,83.22374442901892,78.00007488896073,72.65223047704973,67.18487746856272,61.60283965219095,55.91109455437909,50.114769949434134,44.21914056458991,38.22962486857465,32.151781820259664,25.991307449447632,19.75403115137445,13.44591158799158,7.073032114655783,0.6415956738127304,-5.842080869985595,-12.371572952262392,-18.940355056096063,-25.541807929449632,-32.16922619641426,-38.81582627246999,-45.474754487337094,-52.13909531739884,-58.801879636945664,-65.45609290737777,-72.09468324083045,-78.7105692921268,-85.2966479544593,-91.84580185362773,-98.35090665510428,-104.80483821410012,-111.20047961130413,-117.53072812539943,-123.78850219656353,-129.96674843463774,-136.05844871969336,-142.05662743435695,-147.95435885491992,-153.7447747151699,-159.42107194220313,-164.97652054971297,-170.4044716613963,-175.69836562646825,-180.85174018141933,-185.8582386071305,-190.71161782908231,-195.40575640972494,-199.93466238676564,-204.29248091773678,-208.47350170000936,-212.47216614493857,-216.28307429492847,-219.90099148177183,-223.32085473321536,-226.53777894176642,-229.54706281477564,-232.34419462782375,-234.92485780400764,-237.28493634035672,-239.42052009914158,-241.3279099770278,-243.00362295897457,-244.44439705725446,-245.64719612929662,-246.6092145618199,-247.32788180332227,-247.8008667227399,-248.02608176930102,-248.00168690729544,-247.72609329980156,-247.19796671711944,-246.4162306487051,-245.38006910137838,-244.08892907128057,-242.5425226820563,-240.74082898671415,-238.68409543525874,-236.3728390141679,-233.80784706693643,-230.99017780702812,-227.92116053565158,-224.60239557676346,-221.0357539407353,-217.22337672631306,-213.1676742680693,-208.8713250337152,-204.33727427264924,-199.56873241421724,-194.56917321155566,-189.3423316247831,-183.89220143581719,-178.22303258633752,-172.33932823039314,-166.24584149387408,-159.94757193442985,-153.4497616973318,-146.75789136508695,-139.8776755011522,-132.81505789071178,-125.57620648397858,-118.16750804974585,-110.59556254879976,-102.86717723822456,-94.98936051855529,-86.9693155361142,-78.81443355275974,-70.53228709471958,-62.130622891266874,-53.617354612827405,-45.00055541680111,-36.288450308044126,-27.489408319728586,-18.611934519236655,-9.664661842968005,-0.6563427634773842,8.404159207755603,17.507878178373502,26.64575468989099,35.808644684179356,44.98732861564282,54.172520701831786,63.35487830396169,72.52501142754653,81.67349233220011,90.79086523865895,99.86765612029131,108.89438256582515,117.86156369974603,126.75973014679893,135.5794340272465,144.31125896996025,152.9458301309917,161.47382420595827,169.88597942529475,178.17310552213914,186.32609366326813,194.33592633404314,202.19368716873745,209.8905707178624,217.41789214419813,224.7670968391628,231.92976995094122,238.8976458154696,245.66261728096546,252.21674491624756,258.5522660926426,264.6616039288645,270.53737608791647,276.1724034148209,281.55971840386746,286.6925734840785,291.56444911173037,296.1690616590427,300.50037108852626,304.5525884029513,308.32018286143926,311.79788895275624,314.9807131174761,317.86394021125477,320.4431397019903,322.7141715941275,324.6731920737727,326.3166588686146,327.64133631691965,328.6443001400434,329.3229419130442,329.63772664356514,329.39573053012083,328.5662725632224,327.12975971325596,325.17607358918974,322.74119838431614,319.8567397763678,316.54821998944607,312.84457673579453,308.823351821683,304.528236982323,300.0324579885334,295.34903025270154,290.51816650325384,285.53839843245527,280.4285910464682,275.1801103898228,269.7832166794581,264.21729770242223,258.45095739604227,252.46249806296882,246.21677675320268,239.69855075615806,232.87945009840837,225.75323368691284,218.3066309478104,210.54568864954,202.47659229388634,194.11710391694322,185.48996535134427,176.6206144174621,167.54177130506326,158.28172240186876,148.8753046309741,139.34846043006638,129.73065465058352,120.04126988306407,110.29961873382344,100.51662965895345,90.70028370753485,80.85356847483004,70.97516371172152,61.062655279977044,51.10972108392928,41.112321356145344,31.064221287380246,20.963595069490452,10.808648690941766,0.6026306199445438,-9.64924353014547,-19.937419055765098,-30.249326610838335,-40.56968005996698,-50.88024071351208,-61.162041478475174,-71.39416530628274,-81.55678852097864,-91.62964657759473,-101.59482998234131,-111.43547722404061,-121.13758551992318,-130.68922099543488,-140.08106340387704,-149.30617181511886,-158.3594072703919,-167.2376142442299,-175.93834669677477,-184.46025262976588,-192.80162513190717,-200.96079778685953,-208.9349710453467,-216.72050108712767,-224.31227243239334,-231.7038483922915,-238.8874482032238,-245.85399031580653,-252.59355215553495,-259.09536321976884,-265.34851911933174,-271.3419731236832,-277.06525542832503,-282.5084844127098,-287.66289335803094,-292.52085406487316,-297.07610677377494,-301.3237723323591,-305.2602827219269,-308.8833587062781,-312.19171946866095,-315.18501729813397,-317.8634483850947,-320.22765604276367,-322.278367141766,-324.01629112841346,-325.44187571353467,-326.55523391102616,-327.35606316606925,-327.8436251431439,-328.0168209448381,-327.87423169359437,-327.414302539031,-326.6354325945517,-325.5362001119233,-324.1154749501837,-322.3726181543335,-320.3075838084521,-317.9210442101224,-315.21445223851845,-312.1900707461272,-308.8509794311026,-305.20101535130135,-301.24472515037104,-296.98724541764847,-292.43421655549236,-287.59164261440446,-282.46579260537857,-277.0630784525699,-271.389970850048,-265.4529232414019,-259.25832275654517,-252.81247188665566,-246.121583243937,-239.19181380739923,-232.02929910841857,-224.64022192082896,-217.0308708551182,-209.20771890363423,-201.17748606645884,-192.9472039455205,-184.52426288146088,-175.91644559926831,-167.13194550966693,-158.17936267099336,-149.06768911298727,-139.80627099877282,-130.40476618631243,-120.8730846788975,-111.22133055359242,-101.45973678816449,-91.59860652915165,-81.6482576951811,-71.6189770087313,-61.52098518193978,-51.364412117016435,-41.15928666846771,-30.91553469282991,-20.642990399131204,-10.351412525910664,-0.0505089777346428,10.250037970137056,20.54054451850453,30.811301610294795,41.05255535629418,51.254493657272604,61.40724105117804,71.50086073982719,81.52536603507394,91.47073807599321,101.32695137220536,111.08400320479979,120.73194730498895,130.26092821310036,139.661215655444,148.92323653863426,158.0376032479416,166.99513738936216,175.786887597993,184.40414198050684,192.83843426207832,201.08154519608053,209.12549902992112,216.96255698909587,224.58520828105998,231.98616042022343,239.1583298301112,246.0948339496571,252.78898588150423,259.23429203915987,265.42445356024314,271.3533711980644,277.0151539665499,282.4041307054813,287.51486431071595,292.3421675462825,296.8811197854921,301.12708365324,305.0757207593476,308.72300579262753,312.06523826850434,315.09905162964293,317.8214192944448,320.22965779073166,322.3214269575363,324.0947276908161,325.54789756585234,326.67960498315296,327.48884238481446,561.5675743833262,275.1284621451785,549.87404319882,501.11895263316535,524.114584434861,414.66890817217825,456.79465387374415,435.25213779841596,445.0197045507331,360.863719138502,331.66998203512895,266.33232537943337,189.6170813101339,127.32124684948151,88.10785169675131,63.79569619088987,40.57467851515425,19.624112386345388,4.343611496614662,3.7112673759381756,14.196072003507357,32.82745098062725,53.15678485863621,74.82849623711508,97.84179286539346,122.2346571980109,147.18627129929573,169.82501637131008,189.69782757709066,204.94795816096905,216.8317178907854,224.20726969636053,227.8319884267798,226.47814214582317,220.30379070969593,208.7735782820986,192.12466752164096,171.1070270184808,145.8042129329561,116.99985962302681,84.6557214972461,50.00295930857622,13.503075119967054,-23.326917449667864,-59.67774286651073,-94.2370406184447,-126.09448292685212,-154.25029527166285,-177.82751467511363,-196.40133992726857,-209.51952289427402,-217.34124406604823,-220.05016159743516,-218.63809405707931,-213.82360098083274,-206.77552708468093,-198.57580404376054,-190.54926728023958,-183.8905520466633,-179.5769886172891,-178.33200491997005,-180.4019778617304,-185.84055378964052,-194.237052675618,-205.1085001251819,-217.64399248472358,-231.0702657551783,-244.50363241165007,-257.2024434183884,-268.5325686538249,-278.05829617141035,-285.59232307469193,-291.12024730355404,-294.9264612429519,-297.4157726410137,-299.1412820130006,-300.61809816549606,-302.3911795947021,-304.86774946440744,-308.3202476967329,-312.79201656371504,-318.1070570390909,-323.90057425862585,-329.63847402790486,-334.72061324787524,-338.52116339399794,-340.52120555033247,-340.31743380254824,-337.7206254343934,-332.74029207608635,-325.6674580006531,-317.0283341686236,-307.413709674036,-297.50884409555,-288.15934559386505,-280.2411191855315,-274.50111870812714,-271.49519400550537,-271.5240136141775,-274.6380388873803,-280.64219325021025,-289.1151421413608,-299.4011867139872,-310.69155022322735,-322.1055204900871,-332.75060533701776,-341.78384509392754,-348.4586202187961,-352.18509386797336,-352.56002783238114,-349.2992921014143,-342.3823147655539,-332.1357265155944,-318.9198182726316,-303.12419988837615,-285.2163139816257,-265.7333795671401,-245.23183611599708,-224.22352203477044,-203.14630380631576,-182.3508930691152,-162.12538351132756,-142.7371695428592,-124.43587958172438,-107.43003041018366,-91.88274788581109,-77.91929859462834,-65.633590647726,-55.087098417390145,-46.36564341503623,-39.48962522947116,-34.26941226236814,-30.476280539583794,-27.93100375630883,-26.419277127065328,-25.64124367842462,-25.227557068625316,-24.798121646152456,-24.01395661809308,-22.581200158956516,-20.24593926002935,-16.807286269147365,-12.123332235920248,-6.112413913326497,1.2515396818396227,9.946853703713977,19.914414415468997,31.067150164307744,43.30131667668415,56.54821151943852,70.68738076634868,85.51461828785703,100.88785827978595,116.68968939944551,132.77809462200142,148.96512505309158,165.02078311351852,180.69573796305207,195.736562784886,209.8840854804981,222.88195048743958,234.49782637800985,244.54706516245426,252.91367096015952,259.5650545898152,264.5590855517151,268.044739534397,270.2569419403562,271.5051495773545,272.1625279748569,272.6263361056984,273.2810422814843,274.48038907074476,276.51749444602933,279.5993286348665,283.8345560018406,289.2243545549844,295.6562441807054,302.92043753546244,310.7354973801385,318.77503697105794,326.6935130290364,334.1513964111983,340.83943860305385,346.5003587793373,350.94467425803447,354.0584361013299,355.80245206372217,356.19513954268575,355.2973038608069,353.2404773990845,350.1997264870152,346.31942302137105,341.72605004482097,336.5404502722537,330.8692621083492,324.8028119269274,318.4226855179043,311.7988247835613,304.99666421768524,298.088533321816,291.1616610184547,284.3205526579395,277.68482704138034,271.3837598070552,265.5482413783248,260.300627443581,255.7432504270903,251.95358978506994,248.98929936859932,246.81653417009218,245.30625207443637,244.33332310467807,243.75242951859212,243.36163573151146,242.89645596501265,242.07100740805808,240.6125668189338,238.27724936883513,234.85360086573257,230.16595057517765,224.0822696573232,216.52367330600475,207.46958848759718,196.95626519183085,185.07029929735722,171.9402576922913,157.72855391237462,142.62370699285225,126.83280932364008,110.57389306604162,94.06879946851389,77.5368085904297,61.18924512177224,45.22463516762558,29.824040995921642,15.146317176752666,1.3233220877335727,-11.544577293526554,-23.39240026431956,-34.19337680650596,-43.960439158152184,-52.745982346232864,-60.63972241565696,-67.76446064040181,-74.26984270993997,-80.32434705081664,-86.10598772279795,-91.7923975909326,-97.55104344899945,-103.53039460181621,-109.85272028074678,-116.60908411939863,-123.85678805077308,-131.61931925366483,-139.8885521249462,-148.6288077936684,-157.78221002389878,-167.27472142575107,-177.02223101216276,-186.93610525953378,-196.92774289824663,-206.91180942151766,-216.80805103576745,-226.54176399025167,-236.04321144416312,-245.2464016295961,-254.0877423885663,-262.5050792922651,-270.437557749489,-277.82660093296664,-284.6180927989985,-290.76563993794286,-296.2345665341313,-301.00614599790873,-305.08147139239793,-308.48437213271194,-311.2628475992429,-313.48863452837656,-315.25470298965706,-316.670689424884,-317.8564922273253,-318.93446154247374,-320.0207930944197,-321.2168540550584,-322.60122246153344,-324.2231800528818,-326.09827991891325,-328.20641015274146,-330.4925350981721,-332.87003607283674,-335.226335852435,-337.43029531763324,-339.34073506494195,-340.81536795950547,-341.71942791057086,-341.9333479207401,-341.3589615388675,-339.9238705884365,-337.5838125523927,-334.3230582708367,-330.15304704174173,-325.1096068042361,-319.24919534860607,-312.6446283985489,-305.38073262993515,-297.5502827615159,-289.250468828688,-280.5800092611733,-271.6369014807219,-262.5167000170729,-253.31114828986506,-244.10696723795814,-234.98462137257005,-226.01693148215998,-217.2674724179457,-208.7887722913933,-200.6204043487458,-192.78712608003153,-185.2972627890675,-178.14155029654142,-171.29263917238063,-164.7054221421073,-158.3182803107952,-152.05526147040288,-145.8291156524557,-139.54503153293138,-133.10485295395213,-126.41151470138749,-119.37342397032036,-111.90852697556082,-103.94783532494635,-95.43823799335084,-86.34448638669069,-76.65030545290897,-66.35864709218392,-55.49115694442385,-44.086966804935905,-32.20094886640462,-19.90157351542396,-7.268501303085351,5.609984090575187,18.639625444596895,31.722878063246966,44.76160994258535,57.65980570041012,70.32626144421674,82.67724114304559,94.63903777751482,106.15035057177364,117.16436024554622,127.65036440884678,137.59483044596618,147.00173644835695,155.89210246649873,164.30266250471416,172.28368835790155,179.89604392898076,187.20761630478248,194.28933010722218,201.21099708017226,208.0372771932888,214.82402652402772,221.61527968090152,228.44106281378282,235.3161628789348,242.23989765608573,249.19684809524978,256.15843867112613,263.0851900224982,269.92942665810796,276.638203944144,283.156223508995,289.4285327807114,295.40284865601103,301.0314015493227,306.2722573989253,311.09013428186205,315.45678024648396,319.3510142226186,322.7585488586304,325.67171178838163,328.0891617295223,330.0156619700181,331.4619320734908,332.4445558675339,332.98588673852225,333.1138656385983,332.86165682721946,332.26701273132846,331.3713015650117,330.2181666065105,328.8518298494289,327.3150997537178,325.64718731522623,323.8814711881199,322.0433764674498,320.1485396123874,318.2014222612544,316.1945097786359,314.1081887649267,311.9113457096372,309.56267196533213,307.01260421372183,304.2057802332589,301.08385171854053,297.58847236149256,293.664271934179,289.2616355832529,284.3391302866228,278.86545447061224,272.82082823164535,266.1977860210955,259.00137657872085,251.24881235563365,242.96863959129053,234.19951878911226,224.98871316685504,215.39038070613708,205.46375579979608,195.2712920664053,184.8768218120624,174.34377277898912,163.7334714552581,153.10355557585254,142.50651667970646,131.98839586169981,121.58766058779504,111.33429563486017,101.24914483718375,91.34354060643179,81.61925390279828,72.06878792464822,62.676024474370315,53.41721373071708,44.26227764960506,35.17637652250814,26.121669626300122,17.05918655034054,7.950717391869104,-1.2393714152584057,-10.542483142747153,-19.984373940361223,-29.584030991972167,-39.35288279294241,-49.29435675861731,-59.40377831800239,-69.66858730828142,-80.06883328573758,-90.57790217268703,-101.16342270651164,-111.78830205139272,-122.41184471643058,-132.99091621070116,-143.48112104597283,-153.83797217130177,-164.0180343412563,-173.98002639088622,-183.6858666274405,-193.1016419160235,-202.19847549875985,-210.95326255737768,-219.34923763947484,-227.37633590415385,-235.03131198211014,-242.31758685194157,-249.24480462555542,-255.82809694421113,-262.0870716135605,-268.0445624479401,-273.7251970311942,-279.1538561228478,-284.3541107865778,-289.346729424526,-294.14834579105343,-298.7703704922778,-303.21821300130983,-307.49086012564686,-311.58083204665627,-315.47451078783627,-319.15281065348995,-322.5921380564322,-325.7655710757757,-328.64417833465865,-331.1983929627574,-333.39936039793577,-335.2201878084611,-336.6370366533314,-337.6300166314429,-338.18385711027656,-338.28834922175173,-337.93856656674717,-337.1348837029752,-335.8828186649355,-334.19272862436947,-332.0793869401319,-329.5614662078603,-326.66094672580857,-323.4024643570717,-319.8126072970591,-315.9191686786723,-311.7503617997043,-307.3340071175692,-302.6967046631856,-297.86301144393235,-292.85464974391033,-287.68977788331983,-282.3823588998155,-276.9416638740851,-271.37194464628396,-265.67230524175613,-259.83679263621394,-253.85471612845384,-247.71119146746184,-241.38789213728126,-234.86397706335507,-228.11715264217628,-221.12481840342485,-213.86524049641818,-206.31869590726367,-198.46853284609514,-190.3020987331844,-181.8114960056077,-172.99413670356182,-163.853078513716,-154.39713668430102,-144.64077712276148,-134.60380536530297,-124.31087353125892,-113.7908326677316,-103.07596112355704,-92.20110104151823,-81.20273513228081,-70.11803505124647,-58.983911365433286,-47.8360935929733,-36.7082672941275,-25.63129369526139,-14.632535691325371,-3.7353120468492618,7.041501095393228,17.68370595525716,28.181845463876993,38.53108755019165,48.73093370128716,58.784761857666865,68.69922167789981,78.48350784604591,88.14854376893673,97.70611311319345,107.16797965800757,116.54503654856863,125.84652409811312,135.07935090926884,144.24754659997492,153.35186635058338,162.38955849666624,171.35429719977702,180.2362735450939,189.02243086285543,197.6968241252705,206.24107921960228,214.63492582269436,222.8567773832745,230.88433306509103,238.69517900101073,246.26736937620123,253.57997121936796,260.6135599206949,267.3506551150415,273.77608851948105,279.8772966059567,285.64453178072944,291.07098630377106,296.1528238469613,300.88911469617886,305.2816724281375,309.33479260831547,313.05489768761106,316.4500966889533,319.5296731853889,322.30352006846164,324.7815442019548,326.9730677416519,328.8862552073242,330.5275959480421,331.90147022309105,333.0098236855228,333.85196975376294,334.42453252230797,334.7215349819998,334.7346289853803,334.45345524073394,333.8661142690347,332.9597232518895,331.7210294471886,330.13704859964,328.1956965917445,325.8863843551568,323.20054952431565,320.1321030749899,316.67777478810854,312.8373473302047,308.6137745909779,304.0131852824682,299.04477740298427,293.7206128417872,288.0553241073528,282.0657469680604,275.77049384342735,269.18948326469257,262.34344083123284,255.25338699307497,247.94012680628748,240.4237565863976,232.72320210421765,224.8558025470285,216.83695378327798,208.6798233829389,200.3951482339136,191.99112337279362,183.47338779275137,174.8451095533649,166.10716862491083,157.2584317507086,148.2961094553653,139.2161814421144,130.01387328599895,120.6841647888494,111.22230880855197,101.62433892540814,91.88754499241071,82.01089737609014,71.99540339224583,61.844382874353634,51.56365373747887,41.16162255536492,30.649279298778055,20.040099264391863,9.349858677257785,-1.4036266454175617,-12.200827888335306,-23.020881735548667,-33.842006327202185,-44.6419396618885,-55.39838530579135,-66.08945051786294,-76.69406246724826,-87.1923490749175,-97.5659721514117,-107.79840194254022,-117.87512396110567,-127.78377109837542,-137.51417647982842,-147.05834533100006,-156.41034718687072,-165.5661330048556,-174.52328497919933,-183.2807099242216,-191.83828979932974,-200.19650509556968,-208.35604821913057,-216.31744456015372,-224.08069855124103,-231.64498069295877,-239.00836931911437,-246.16765792671526,-253.11823539598558,-259.85404260715575,-266.36760507544506,-272.65013752354344,-278.69171301719695,-284.48148658626263,-290.0079612659435,-295.2592832819952,-300.22355266624544,-304.8891358634562,-309.24496776983494,-313.28083198936224,-316.9876097557387,-320.3574897988988,-323.3841333088668,-326.0627899697954,-328.39036274283364,-330.36542064350067,-331.9881601939103,-333.2603175602065,-334.1850346482108,-334.7666836595333,-335.01065582564047,-334.92312123479763,-334.51076781511,-333.7805285777164,-332.7393070772348,-331.39371161883554,-329.74980893897174,-327.8129078284088,-325.58738239527554,-323.0765433613197,-320.28256396745536,-317.2064648006446,-313.8481592505204,-310.20655850289927,-306.27973214352943,-302.0651177536307,-297.55977049757945,-292.76064178093566,-287.6648747094057,-282.27010338000514,-276.5747430114767,-270.57825855167954,-264.2814006218575,-257.68639937272707,-250.79710891172851,-243.61909727813747,-236.1596793557025,-228.42789249355255,-220.4344168461728,-212.1914444569854,-203.71250283838438,-195.0122402100335,-186.10618063468573,-177.01045804097052,-167.74153856012182,-158.31594074756345,-148.74996312948352,-139.05942812468595,-129.25945075476088,-119.36423967979233,-109.38693699132583,-99.33950187202973,-89.232641713388,-79.07579260160942,-68.87714928437516,-58.64374287813213,-48.3815627410997,-38.095717203593026,-27.790626300655635,-17.470238375069354,-7.138261483964975,3.2015999950352243,13.54541326866337,23.88879892245058,34.226735232381145,44.55339249611304,54.862003453293305,65.14477420718998,75.39283827077263,85.59625453440333,95.74404817967837,105.82429192398041,115.82422353695445,125.7303943693352,135.52884270369378,145.20528508573716,154.74531841922342,164.13462549208586,173.35917672407828,182.4054212626962,191.2604610802305,199.91220241901598,208.34947977564755,216.562148591773,224.54114391299734,232.27850347014245,239.76735490535165,247.0018681789484,253.97717551247982,260.68926250150236,267.13483521374644,273.3111691154569,279.21594648222106,284.8470894966082,290.20259646991013,295.28038852075156,300.0781735897479,304.59333387863086,308.8228417080267,312.76320744455217,316.4104616263495,319.7601718000477,322.80749296077624,325.54724894916234,327.97404078682973,330.08237679364083,331.86681847719325,333.32213564921085,334.44346401466737,335.22645858778844,335.66743668697734,335.76350490681506,335.51266530909146,334.9138970613946,333.96721082628545,332.6736743156426,331.03540852899147,329.05555525530934,326.7382174068455,324.08837465042484,321.1117775917877,317.8148244409952,314.20442463294,310.2878542880513,306.07260866530413,301.5662568728941,296.77630405173846,291.7100660251809,286.3745610102343,280.77642241239255,274.9218359860185,268.81650375324284,262.4656360634871,255.87397208000056,249.04582784399793,241.9851699418338,234.6957117395414,227.1810282046007,219.4446845550425,211.49037340101086,203.3220547028533,194.94409277891916,186.36138475777267,177.57947527189816,168.60465280890884,159.44402393689708,150.1055625601718,140.5981323927239,130.9314819104023,121.11621211138726,111.16371843451344,101.08610911849433,90.89610310124196,80.60691123377846,70.23210510140153,59.785478096146484,49.28090356530067,38.73219487165258,28.152972047282216,17.556539412352627,6.955778075054832,-3.636943356851232,-14.209837220706323,-24.751742557072873,-35.252147194208746,-45.70118606504959,-56.08961656198655,-66.40877260568338,-76.65049991189478,-86.80707565033731,-96.87111627034919,-106.83547769223875,-116.69315231031369,-126.43716731132855,-136.06048867748544,-145.55593492309114,-154.91610412501208,-164.13331717440812,-173.19957943319253,-182.1065621606422,-190.84560422353755,-199.40773375725186,-207.7837086431072,-215.96407394248925,-223.93923380809096,-231.6995348976294,-239.23535795845936,-246.53721403842778,-253.59584170819605,-260.4023017467159,-266.9480659337778,-273.2250968976795,-279.22591636623997,-284.94365964869996,-290.3721147171255,-295.50574484130766,-300.33969434331937,-304.86977765912286,-309.092452506947,-313.00477854708004,-316.60436345641216,-319.88929881451367,-322.858088587761,-325.5095732867579,-327.84285304526884,-329.85721291466126,-331.5520535798918,-332.9268304803793,-333.9810039671434,-334.7140026585138,-335.12520158915595,-335.2139161055599,-334.9794117745488,-334.420929871826,-333.5377273383075,-332.32912946524107,-330.79459302442706,-328.93377712204017,-326.7466187425288,-324.23340977467274,-321.3948722796905,-318.23222886842666,-314.74726529162996,-310.94238269885534,-306.8206374678158,-302.38576702428605,-297.6422006385044,-292.5950547728138,-287.2501131433185,-281.61379222378224,-275.69309344349807,-269.4955437958821,-263.0291269676308,-256.3022074088172,-249.32344998462875,-242.10173797454982,-234.64609221219717,-226.9655940888602,-219.06931497865662,-210.96625438833328,-202.6652887980788,-194.17513275194406,-185.50431329107252,-176.66115831595457,-167.65379893358804,-158.4901853116851,-149.17811504575357,-139.72527256684847,-130.13927769784982,-120.42774112203554,-110.59832427414563,-100.65880101181253,-90.61711838031206,-80.48145384726288,-70.26026655233076,-59.962340381568694,-49.59681702389434,-39.17321758204781,-28.701451773371595,-18.19181424661639,-7.654968039045825,2.8980843167053614,13.456041072798397,24.007351564236547,34.540274811228265,45.04294444004165,55.50343774396704,65.90984659751523,76.25034791394455,86.51327139722741,96.68716248061398,106.76083855868865,116.72343689979292,126.564452960525,136.27376820169619,145.84166691213994,155.25884196852942,164.51638988044562,173.60579587466282,182.5189101454347,191.24791672376173,199.7852966849869,208.1237876094733,216.25634132725122,224.17608200931295,231.87626661433964,239.3502495625393,246.5914532943405,253.59334609081472,260.3494281979718,266.85322692388274,273.0983009830814,279.0782539645222,284.7867564150762,290.2175756764002,295.364612303417,300.22194163936376,304.7838589341411,309.04492627504345,313.0000195540165,316.6443737224223,319.97362467906333,322.9838462936572,325.6715812778579,328.0338648695484,330.06824058284525,331.7727675847407,333.1460195782044,334.18707538950605,334.89550176360973,335.27132915543103,335.31502155701946,335.0274416127917,334.4098124395201,333.4636776789003,332.19086136395083,330.5934291737545,328.6736525837945,326.43397729323465,323.8769971300236,321.0054344061762,317.8221274275336],"type":"scatter","name":"source2_u_a","yaxis":"y2","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1,0.1001,0.1002,0.1003,0.1004,0.1005,0.1006,0.1007,0.1008,0.1009,0.101,0.1011,0.1012,0.1013,0.1014,0.1015,0.1016,0.1017,0.1018,0.1019,0.102,0.1021,0.1022,0.1023,0.1024,0.1025,0.1026,0.1027,0.1028,0.1029,0.103,0.1031,0.1032,0.1033,0.1034,0.1035,0.1036,0.1037,0.1038,0.1039,0.104,0.1041,0.1042,0.1043,0.1044,0.1045,0.1046,0.1047,0.1048,0.1049,0.105,0.1051,0.1052,0.1053,0.1054,0.1055,0.1056,0.1057,0.1058,0.1059,0.106,0.1061,0.1062,0.1063,0.1064,0.1065,0.1066,0.1067,0.1068,0.1069,0.107,0.1071,0.1072,0.1073,0.1074,0.1075,0.1076,0.1077,0.1078,0.1079,0.108,0.1081,0.1082,0.1083,0.1084,0.1085,0.1086,0.1087,0.1088,0.1089,0.109,0.1091,0.1092,0.1093,0.1094,0.1095,0.1096,0.1097,0.1098,0.1099,0.11,0.1101,0.1102,0.1103,0.1104,0.1105,0.1106,0.1107,0.1108,0.1109,0.111,0.1111,0.1112,0.1113,0.1114,0.1115,0.1116,0.1117,0.1118,0.1119,0.112,0.1121,0.1122,0.1123,0.1124,0.1125,0.1126,0.1127,0.1128,0.1129,0.113,0.1131,0.1132,0.1133,0.1134,0.1135,0.1136,0.1137,0.1138,0.1139,0.114,0.1141,0.1142,0.1143,0.1144,0.1145,0.1146,0.1147,0.1148,0.1149,0.115,0.1151,0.1152,0.1153,0.1154,0.1155,0.1156,0.1157,0.1158,0.1159,0.116,0.1161,0.1162,0.1163,0.1164,0.1165,0.1166,0.1167,0.1168,0.1169,0.117,0.1171,0.1172,0.1173,0.1174,0.1175,0.1176,0.1177,0.1178,0.1179,0.118,0.1181,0.1182,0.1183,0.1184,0.1185,0.1186,0.1187,0.1188,0.1189,0.119,0.1191,0.1192,0.1193,0.1194,0.1195,0.1196,0.1197,0.1198,0.1199,0.12,0.1201,0.1202,0.1203,0.1204,0.1205,0.1206,0.1207,0.1208,0.1209,0.121,0.1211,0.1212,0.1213,0.1214,0.1215,0.1216,0.1217,0.1218,0.1219,0.122,0.1221,0.1222,0.1223,0.1224,0.1225,0.1226,0.1227,0.1228,0.1229,0.123,0.1231,0.1232,0.1233,0.1234,0.1235,0.1236,0.1237,0.1238,0.1239,0.124,0.1241,0.1242,0.1243,0.1244,0.1245,0.1246,0.1247,0.1248,0.1249,0.125,0.1251,0.1252,0.1253,0.1254,0.1255,0.1256,0.1257,0.1258,0.1259,0.126,0.1261,0.1262,0.1263,0.1264,0.1265,0.1266,0.1267,0.1268,0.1269,0.127,0.1271,0.1272,0.1273,0.1274,0.1275,0.1276,0.1277,0.1278,0.1279,0.128,0.1281,0.1282,0.1283,0.1284,0.1285,0.1286,0.1287,0.1288,0.1289,0.129,0.1291,0.1292,0.1293,0.1294,0.1295,0.1296,0.1297,0.1298,0.1299,0.13,0.1301,0.1302,0.1303,0.1304,0.1305,0.1306,0.1307,0.1308,0.1309,0.131,0.1311,0.1312,0.1313,0.1314,0.1315,0.1316,0.1317,0.1318,0.1319,0.132,0.1321,0.1322,0.1323,0.1324,0.1325,0.1326,0.1327,0.1328,0.1329,0.133,0.1331,0.1332,0.1333,0.1334,0.1335,0.1336,0.1337,0.1338,0.1339,0.134,0.1341,0.1342,0.1343,0.1344,0.1345,0.1346,0.1347,0.1348,0.1349,0.135,0.1351,0.1352,0.1353,0.1354,0.1355,0.1356,0.1357,0.1358,0.1359,0.136,0.1361,0.1362,0.1363,0.1364,0.1365,0.1366,0.1367,0.1368,0.1369,0.137,0.1371,0.1372,0.1373,0.1374,0.1375,0.1376,0.1377,0.1378,0.1379,0.138,0.1381,0.1382,0.1383,0.1384,0.1385,0.1386,0.1387,0.1388,0.1389,0.139,0.1391,0.1392,0.1393,0.1394,0.1395,0.1396,0.1397,0.1398,0.1399,0.14,0.1401,0.1402,0.1403,0.1404,0.1405,0.1406,0.1407,0.1408,0.1409,0.141,0.1411,0.1412,0.1413,0.1414,0.1415,0.1416,0.1417,0.1418,0.1419,0.142,0.1421,0.1422,0.1423,0.1424,0.1425,0.1426,0.1427,0.1428,0.1429,0.143,0.1431,0.1432,0.1433,0.1434,0.1435,0.1436,0.1437,0.1438,0.1439,0.144,0.1441,0.1442,0.1443,0.1444,0.1445,0.1446,0.1447,0.1448,0.1449,0.145,0.1451,0.1452,0.1453,0.1454,0.1455,0.1456,0.1457,0.1458,0.1459,0.146,0.1461,0.1462,0.1463,0.1464,0.1465,0.1466,0.1467,0.1468,0.1469,0.147,0.1471,0.1472,0.1473,0.1474,0.1475,0.1476,0.1477,0.1478,0.1479,0.148,0.1481,0.1482,0.1483,0.1484,0.1485,0.1486,0.1487,0.1488,0.1489,0.149,0.1491,0.1492,0.1493,0.1494,0.1495,0.1496,0.1497,0.1498,0.1499,0.15,0.1501,0.1502,0.1503,0.1504,0.1505,0.1506,0.1507,0.1508,0.1509,0.151,0.1511,0.1512,0.1513,0.1514,0.1515,0.1516,0.1517,0.1518,0.1519,0.152,0.1521,0.1522,0.1523,0.1524,0.1525,0.1526,0.1527,0.1528,0.1529,0.153,0.1531,0.1532,0.1533,0.1534,0.1535,0.1536,0.1537,0.1538,0.1539,0.154,0.1541,0.1542,0.1543,0.1544,0.1545,0.1546,0.1547,0.1548,0.1549,0.155,0.1551,0.1552,0.1553,0.1554,0.1555,0.1556,0.1557,0.1558,0.1559,0.156,0.1561,0.1562,0.1563,0.1564,0.1565,0.1566,0.1567,0.1568,0.1569,0.157,0.1571,0.1572,0.1573,0.1574,0.1575,0.1576,0.1577,0.1578,0.1579,0.158,0.1581,0.1582,0.1583,0.1584,0.1585,0.1586,0.1587,0.1588,0.1589,0.159,0.1591,0.1592,0.1593,0.1594,0.1595,0.1596,0.1597,0.1598,0.1599,0.16]}],{"xaxis":{"title":{"text":"Time in Seconds"}},"template":{"layout":{"coloraxis":{"colorbar":{"ticks":"","outlinewidth":0}},"xaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"hovermode":"closest","paper_bgcolor":"white","geo":{"showlakes":true,"showland":true,"landcolor":"#E5ECF6","bgcolor":"white","subunitcolor":"white","lakecolor":"white"},"colorscale":{"sequential":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]],"diverging":[[0,"#8e0152"],[0.1,"#c51b7d"],[0.2,"#de77ae"],[0.3,"#f1b6da"],[0.4,"#fde0ef"],[0.5,"#f7f7f7"],[0.6,"#e6f5d0"],[0.7,"#b8e186"],[0.8,"#7fbc41"],[0.9,"#4d9221"],[1,"#276419"]],"sequentialminus":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]},"yaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"shapedefaults":{"line":{"color":"#2a3f5f"}},"hoverlabel":{"align":"left"},"mapbox":{"style":"light"},"polar":{"angularaxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","radialaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"autotypenumbers":"strict","font":{"color":"#2a3f5f"},"ternary":{"baxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","caxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"aaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"annotationdefaults":{"arrowhead":0,"arrowwidth":1,"arrowcolor":"#2a3f5f"},"plot_bgcolor":"#E5ECF6","title":{"x":0.05},"scene":{"xaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"zaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"yaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"}},"colorway":["#636efa","#EF553B","#00cc96","#ab63fa","#FFA15A","#19d3f3","#FF6692","#B6E880","#FF97FF","#FECB52"]},"data":{"barpolar":[{"type":"barpolar","marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"carpet":[{"aaxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"},"type":"carpet","baxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"}}],"scatterpolar":[{"type":"scatterpolar","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"parcoords":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"parcoords"}],"scatter":[{"type":"scatter","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2dcontour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2dcontour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattercarpet":[{"type":"scattercarpet","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"mesh3d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"mesh3d"}],"surface":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"surface","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattermapbox":[{"type":"scattermapbox","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergeo":[{"type":"scattergeo","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram":[{"type":"histogram","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"pie":[{"type":"pie","automargin":true}],"choropleth":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"choropleth"}],"heatmapgl":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmapgl","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"bar":[{"type":"bar","error_y":{"color":"#2a3f5f"},"error_x":{"color":"#2a3f5f"},"marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"heatmap":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmap","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contourcarpet":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contourcarpet"}],"table":[{"type":"table","header":{"line":{"color":"white"},"fill":{"color":"#C8D4E3"}},"cells":{"line":{"color":"white"},"fill":{"color":"#EBF0F8"}}}],"scatter3d":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"scatter3d","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergl":[{"type":"scattergl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2d","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scatterternary":[{"type":"scatterternary","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scatterpolargl":[{"type":"scatterpolargl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}]}},"legend":{"yanchor":"bottom","xanchor":"right","y":1.02,"orientation":"h","x":1},"margin":{"l":100,"b":80,"pad":10,"r":80,"t":100},"plot_bgcolor":"#f1f3f7","yaxis":{"title":{"text":"State values"}},"yaxis2":{"titlefont":{"color":"orange"},"overlaying":"y","title":"Action values","side":"right"},"height":550,"width":800},{"showLink":false,"editable":false,"responsive":true,"staticPlot":false,"scrollZoom":true});
});





</script>
```




_______________________________________________________________________________
### High-Level Rendering


```julia
RenderHookResults(hook = hook, 
                    states_to_plot  = [], 
                    actions_to_plot = [],  
                    v_mag_inv       = [2], # Scaled L₂ norm in αβγ coordinates [V]
                    power_p_inv     = [2], # Real power [Watts]
                    power_q_inv     = [2], # Imaginary power [VAi]
                    angles          = [2], # Relative angle [degrees]
                    freq            = [2], # Angular velocity [Hz]
                    )
```


```@raw html
<div
id = msoFhyc252Iq > </div>
<script>
gd = 'msoFhyc252Iq'
require(['plotly'], function(plotly) {
























plotly.newPlot(gd,[{"mode":"lines","y":[0.0,0.0,0.0,-0.0010112127245424983,-0.029973019042705065,-0.1645126021306731,-0.5022914285374576,-1.0517906918062812,-1.8006804510377072,-2.7024002487681953,-3.7194831619012905,-4.7478143692388475,-5.621175273191752,-6.205753475690972,-6.384842878513018,-6.186603648118499,-5.616220481795531,-4.78848216484905,-3.757913304063934,-2.6684717489804313,-1.6076567216964834,-0.6933915047809021,0.018799910148457055,0.492668338427916,0.7226097402754807,0.7319038369117807,0.5365545337852824,0.18238939031752777,-0.3102882667785285,-0.8919074842359027,-1.5381712291924716,-2.200211045594966,-2.8457690934809974,-3.4309638406965783,-3.920984050717933,-4.287092778196489,-4.505617635571571,-4.573340506429096,-4.48981272061853,-4.27863883578343,-3.961727798502931,-3.5764740724649973,-3.152360381593826,-2.7214151571267675,-2.304565333286309,-1.9171594872009572,-1.5656520226512085,-1.2505330365104224,-0.9689580258929125,-0.7165125918221784,-0.49045293135572143,-0.29057031397667565,-0.1198556724236362,0.015567456276129432,0.10815185168635288,0.14967123065657395,0.13308836588180073,0.05392930665294915,-0.0896649140325183,-0.29462016357293574,-0.5553431723195241,-0.8603506434583141,-1.1965507069725376,-1.545462626941429,-1.8881317607222599,-2.2023091659826655,-2.466695290431647,-2.660421130799951,-2.7654851544411265,-2.76890876644463,-2.66256759757702,-2.447108675691781,-2.1290160737291854,-1.7246150165565854,-1.2550937269494629,-0.7490663036036587,-0.2370389329614646,0.24817385596947972,0.6758217778235149,1.0187125860680393,1.255174305105764,1.3716065821870238,1.361705019114872,1.2290467854322182,0.9841615652457962,0.6462934835183267,0.2394392146881259,-0.20721269624950445,-0.6630868083309442,-1.0971465673697605,-1.4806185527479876,-1.7888673565604138,-2.0025304991568698,-2.109605475746979,-2.104926612681382,-1.9916820301124725,-1.7795630154535838,-1.4851931388540058,-1.1295778342086114,-0.7373828476435544,-0.33443221419285274,0.053793359939088514,0.4039613396874465,0.6965666959842245,0.9166730358918476,1.0551948083954983,1.1085602025948669,1.0792421729357988,0.9746207885786866,0.8067003371436607,0.5906006922262401,0.34365107480701784,0.08394385236021762,-0.1708512726703102,-0.404539789044982,-0.6034908426500696,-0.7571194879332541,-0.8585786254453645,-0.904658288637096,-0.8959376900213556,-0.836229469746085,-0.73222906329819,-0.5927219819083973,-0.42791234077567,-0.2486324942422744,-0.06559644297254508,0.11119610617068432,0.27306219675585197,0.4129374097719811,0.52570497372243,0.6082038506337248,0.6592634294394539,0.6794928057768672,0.6710856089389274,0.6374978496715844,0.5831299151855625,0.512994025459216,0.43238032126895476,0.34658461079765956,0.2606363976109307,0.1791198369490452,0.10600091666419686,0.04454193374485804,-0.0027719479787915446,-0.03424474048227388,-0.04896673778791705,-0.046779111522635475,-0.028223810509103053,0.005519080186447756,0.052707095411463456,0.11111629682930424,0.1781338089266218,0.25085679779549763,0.3262013418715273,0.4010199384081292,0.47222512687917195,0.5369180039175347,0.5925154987699983,0.6368705976159914,0.6683807227998084,0.6860721004033348,0.6896610598632815,0.6795760519957241,0.6569487271046892,0.6235590342751145,0.5817484697456976,0.5342932171835635,0.48425318794554073,0.43479893036396433,0.3890298237983103,0.3497955355790374,0.31952813098246935,0.3001029841570215,0.29272838504562415,0.297882144031572,0.3152882176680052,0.343945839696552,0.38219955750711065,0.42785303490067506,0.4783134971730263,0.5307596436631705,0.5823214964023258,0.6302578864742507,0.6721240585463342,0.7059129710181544,0.7301680339098375,0.7440539809334474,0.7473888080199416,0.7406303546648196,0.7248244281974436,0.7015161753052404,0.6726336088886828,0.6403517664648444,0.606946240817706,0.5746481073564578,0.5455069270332995,0.5212735143786693,0.5033058475782584,0.49250617354497095,0.48928899569077267,0.4935825369464626,0.5048601474504257,0.5221987768187495,0.5443589126697432,0.5698791805973169,0.5971794477320307,0.6246641321664006,0.6508204776548219,0.6743044821099424,0.6940112490412407,0.7091252470864013,0.7191497385782348,0.7239144010385505,0.7235627212076636,0.7185213794130796,0.709454823146551,0.6972093056703663,0.6827502289078744,0.6670976590420494,0.6512635176999986,0.6361945754206602,0.6227236767013302,0.6115317075527535,0.6031213022351558,0.5978029037101287,0.5956927806634802,0.5967219747111407,0.6006547387625645,0.6071144006835283,0.6156146879065574,0.6255941161466154,0.6364514737443017,0.6475802812082279,0.6584006540172922,0.6683871062071454,0.6770913369696316,0.6841593237084034,0.6893424147232068,0.6925024549583938,0.693611192471856,0.6927445175636855,0.6900721717045997,0.6858437755222635,0.680372025764326,0.6740140144921476,0.6671515782508326,0.6601715909886625,0.6534470429990051,0.6473196807955697,0.6420848884742263,0.6379793743657973,0.6351721058538236,0.6337587884607985,0.6337600381943482,0.6351232364103945,0.6377278961715271,0.641394214675449,0.6458943392409715,0.6509657483410596,0.6563260479716604,0.661688410941391,0.6667768604623809,0.6713405999589774,0.6751666535940437,0.6780901625951579,0.6800018254218685,0.6808521183366167,0.6806521257610356,0.6794709949368496,0.6774302265802525,0.674695191463357,0.6714644163484103,0.6679573094200393,0.6644010679777779,0.661017554223406,0.6580109000156611,0.6555565505865815,0.6537923438759989,0.6528120915664388,0.6526619575443817,0.6533397549325173,0.6547971001069506,0.6569441894450598,0.6596568153203535,0.662785110779673,0.6661634311442484,0.6696207287601759,0.6729907790162457,0.676121646955643,0.6788838616802444,0.681176865872537,0.6829334361245145,0.6841219080538252,0.6847461827604455,0.6848436297513718,0.6844811220353044,0.6837495436392818,0.6827571813040443,0.6816224601914653,0.6804664942931744,0.6794059081869865,0.6785463413363668,0.6779769807943143,0.677766383884685,0.6779597583412889,0.6785777690328061,0.6796168438064978,0.6810508639694177,0.6828340489588705,0.6849047880957713,0.6871901321791256,0.6896106399147399,0.6920852744046648,0.6945360643754555,0.6968922790432295,0.6990939124074167,0.7010943277201562,0.7028619720937144,0.7043811314513064,0.705651752428885,0.7066884086676808,0.7075185307939762,0.708180051361829,0.7087186365628426,0.7091846860844058,0.7096302808488772,0.7101062474386808,0.7106594886173667,0.7113307035649115,0.7121525914240294,0.7131485989699797,0.7143322403513181,0.7157069849267188,0.7172666803209964,0.7189964526494634,0.7208740057832017,0.7228712268333197,0.724955995940018,0.7270940950552287,0.729251112310827,0.7313942455547604,0.7334939193121279,0.7355251441299728,0.7374685637853879,0.7393111546239055,0.7410465601320642,0.742675062978357,0.7442032142943855,0.7456431564474002,0.7470116887100833,0.748329135845264,0.7496180870416038,0.7509020761001226,0.7522042739251527,0.7535462606553764,0.7549469380254976,0.7564216332202971,0.7579814332659378,0.759632775492463,0.7613773054558166,0.7632119988236248,0.7651295299432519,0.7671188573367651,0.769165985505083,0.7712548545408104,0.7733683038692973,0.7754890541915174,0.7776006529046455,0.7796883324729578,0.7817397375540551,0.7837454861526965,0.7856995403820619,0.787599374299667,0.789445938163837,0.7912434298479027,0.7929988949230449,0.7947216855773007,0.7964228153974133,0.7981142513813069,0.7998081862759534,0.8015163336321152,0.8032492847011312,0.8050159608593077,0.8068231882990275,0.8086754131601239,0.8105745665326012,0.8125200794471965,0.8145090393210239,0.8165364716066179,0.8185957238991648,0.8206789250392745,0.8227774889452064,0.8248826317829234,0.8269858721184313,0.8290794863018149,0.8311568955928053,0.8332129666563901,0.8352442131437794,0.8372488925787884,0.8392269992285923,0.841180159018375,0.8431114388313871,1.3168188117114035,9.804282560522998,33.54820976698008,66.70393684519352,103.05755781881939,134.09471765506584,159.6972844470139,179.1453051715978,193.64011454486067,201.55326136492317,201.60403022690906,193.7118073781497,178.45369624795177,158.36240505383296,134.53115441066845,109.33436982721614,83.42507419130631,58.79620137405927,36.264908222953295,17.29620126732942,2.4719492721792182,-7.78992806328691,-13.422207775501484,-14.885521305839873,-12.491543684844778,-7.194992792911799,0.43564755263128596,9.293956581930965,18.724234754256187,27.798607082031825,35.984460232205116,42.74516746802571,47.79892162397431,51.023515628283256,52.39015632783368,52.07791778537006,50.24097666891984,47.208350187953734,43.22430281236065,38.63017376787101,33.66971212270177,28.606645525914374,23.6281945218149,18.894323689724402,14.519992985876508,10.577043204502477,7.117001307130009,4.156870452976044,1.7085210520772849,-0.23083420162476379,-1.663507944796576,-2.5918138095590915,-3.01988868405128,-2.950989058154617,-2.395999145953673,-1.3764948592909598,0.07329452849253038,1.8928672228965757,4.008851588026336,6.313341926498362,8.692716027517452,11.00517661265833,13.11651602073982,14.884568217925116,16.189641704295155,16.930567543509618,17.03947023926922,16.489404909256535,15.289169884299668,13.498393220962214,11.205226469031706,8.542643716647605,5.6571169597622655,2.7202499127901585,-0.10165615732350775,-2.6440677374909476,-4.765404736753512,-6.351575481539442,-7.323326815113331,-7.645758063599034,-7.3217353252287865,-6.401807429132338,-4.967419803017966,-3.137171245386043,-1.0458077902626637,1.155188102786382,3.3165796026798433,5.29593419222414,6.9698614523515,8.24128497534673,9.042831039845645,9.34370645201929,9.14489652659888,8.483122585227989,7.420577609232117,6.044819639150163,4.4566874091153235,2.7665446805959695,1.0837706261305389,-0.48903981292371135,-1.8626903785517024,-2.9671248366613616,-3.7532571676875475,-4.196808788394412,-4.295820972523187,-4.07157213441077,-3.5632117692909597,-2.8258014970300804,-1.9239431447824327,-0.9278931121244991,0.09205700406587525,1.0699750263207357,1.947934090319919,2.679558723498785,3.23122104323518,3.583859181756784,3.7319607201435443,3.6834364291643755,3.4571234525772914,3.081095714786957,2.589613515857984,2.020616283376306,1.4129596377105784,0.8039098068657364,0.2272365945813547,-0.2886442171281661,-0.7214967048566404,-1.0561427846105698,-1.284292619977297,-1.4044464291874283,-1.4210258154849464,-1.3435576660853221,-1.1854851586042732,-0.9630111672392665,-0.6939327582236157,-0.3965031860180505,-0.08852155415542828,0.2135409248277378,0.4951003394047918,0.7439692444538606,0.9505816013624346,1.1081579660668797,1.2126745944698658,1.26277135020225,1.2595551536831224,1.2063485588347818,1.108397121458931,0.9725339317728908,0.806832321822266,0.6202317013777949,0.4221600136074678,0.22214840855700438,0.029448652149593593,-0.14733538323672368,-0.30059653161183353,-0.42403552653774224,-0.5129226711916302,-0.5642996617380941,-0.5770979212932996,-0.5521812938493103,-0.49228156026500075,-0.4018550194395829,-0.28683154866023086,-0.15429845732452396,-0.012105407674983415,0.1315648526896852,0.2686519835981424,0.39166489334573534,0.4940918958261221,0.5707579454928076,0.6180902841874486,0.6342977401504514,0.6194276937911921,0.5753241660527165,0.5054656957122117,0.41471701221798274,0.3089941685649534,0.1948783235149048,0.07919772823579163,-0.031394090193980456,-0.13081207689364505,-0.21388999983155169,-0.2766541509318805,-0.3165194628509611,-0.33238380724321004,-0.32463214833369375,-0.29504190489336773,-0.24660946480141846,-0.1833058600180793,-0.10978426721012167,-0.03105985505100471,0.04781801272805135,0.12207562061081562,0.18749349503556995,0.24063110260678322,0.27899309193858324,0.30112349344037276,0.3066318551907455,0.5065824557925173,19749.030332243463,31305.745629288765,51872.985521495684,70656.58040951514,72295.8885106453,79206.86850747015,79333.40201605365,84173.23848987541,77937.23138136603,78344.98627683117,73600.44705228297,61457.0262342621,47737.805715103415,37915.724592199054,32281.136912914793,28628.625281599056,26158.578453724596,24998.210439364815,25801.84427067013,28024.35308083447,30893.182053496916,33710.40739362252,36088.51006663798,37820.386936071554,38873.98056786103,39416.101170284775,39721.91147265838,39882.57196905989,39980.54916659257,39894.74166654742,39742.80048345835,39535.02397190203,39458.55662213586,39551.72912432518,39927.92192869162,40607.75771500086,41648.18403924651,43080.791482131826,44907.68079098486,47099.89668647341,49594.72450039535,52325.048084144626,55182.487657572296,58038.90851376441,60720.565424202876,63038.957938149615,64800.13903847117,65825.40395221145,65994.61525726118,65232.84739836243,63574.81418340749,61108.71340752498,58030.0628249701,54538.627783687385,50886.2302427852,47297.81046280193,43998.590631631356,41170.064146909914,38934.748457472204,37367.119138421054,36462.79306382305,36204.563776043506,36515.11623936739,37332.540370158116,38553.74040162657,40091.99825046261,41832.52146204505,43667.53981715819,45480.81730026677,47161.22575105638,48618.826392934716,49782.01010466209,50621.66772163305,51138.343464734025,51399.280649349916,51490.79438530187,51521.86973432818,51596.12091766109,51811.76062693439,52241.75842964444,52923.400048247364,53852.88767664582,54977.054068270634,56203.38747175763,57399.33016539381,58417.81413129488,59108.24780183894,59345.794116200996,59042.67091887031,58163.01589846431,56728.36046027606,54818.48338296148,52557.436616799765,50093.46772238554,47589.83144696722,45209.86422642059,43107.89988036195,41410.4034923758,40206.36849418236,39539.24963059071,39412.31151970016,39794.69462721956,40626.64502130347,41824.37562672096,43285.632369289466,44898.65155322941,46549.86619723009,48133.799060990044,49560.11211372075,50761.85011659384,51695.56406695889,52349.113981055634,52750.490537017904,52954.600425109296,53031.00584434575,53054.62932076602,53097.088231390546,53216.714197112015,53450.34557646201,53807.13426871046,54266.04421195151,54776.988645797246,55267.35086473455,55652.01520591194,55845.7190674786,55775.327843277395,55389.85814422826,54667.477970014384,53618.28818446879,52287.64407215761,50752.43091334565,49107.01804681296,47451.65968301022,45886.750562323024,44507.67023086428,43397.60925874459,42618.895195450015,42206.68341374163,42168.04396916357,42484.53620514888,43115.89647730674,44004.62282548861,45080.830519182215,46267.762065511706,47487.950075426124,48669.70302443329,49752.96708070315,50693.907216875654,51467.60776490214,52065.08932762127,52496.462478410846,52793.28814444184,52996.65363742072,53148.51519777582,53286.05523032327,53436.951335312915,53615.667446150444,53820.91423699785,54035.1261378755,54226.66282913354,54354.01781613601,54371.7146347215,54237.24199320376,53917.76368072274,53395.34365088802,52670.15228159554,51761.53264203391,50706.53838529728,49558.76787716757,48384.08504120809,47250.19742795437,46219.205211334534,45345.827715937194,44675.73202368075,44242.397019287564,44063.937028816996,44141.084013163374,44456.86830511202,44979.47740845735,45666.79226763125,46470.72124266815,47340.853719298524,48227.854100128556,49087.01513422909,49881.73263717835,50586.19706112149,51186.685435216816,51681.31933301125,52077.613185136055,52388.99039591248,52634.78804428198,52837.488945768535,53015.916768778545,53181.332863355594,53335.732801188715,53471.31609551664,53571.959875125045,53616.140929878246,53579.97564502388,53440.317008472135,53177.957997789366,52780.91339282226,52247.31008034459,51587.13668629626,50822.274642748554,49984.8531230573,49114.468320617045,48254.806588982574,47450.322438159164,46744.05404753534,46172.572668931556,45761.115163129696,45524.22613938421,45467.11166740076,45585.87074860721,45866.711249664884,46286.44871880484,46814.78499094947,47417.71793250175,48060.93465117393,48712.460173048006,49344.51551913266,49934.91381824717,50468.11514405963,50935.76100168141,51336.39916290296,51674.29619189582,51957.533217699965,52195.74753781895,52397.9167472218,52570.45808097801,52715.842795802775,52831.84915315639,52911.56712531837,52944.191004373686,52916.52975747887,52815.03378260905,52628.03590219283,52347.88889821233,51972.69924887184,51507.445261392284,50964.33437490113,50362.36758568841,49726.16010603956,49084.16910863721,48466.53629961543,47902.787573302056,47419.62859407205,47039.03849805118,46776.822785779965,46641.71831196641,46635.09550562961,46751.23676592256,46978.13520187506,47298.71462611762,47692.35482857059,48136.58615429231,48608.81045085802,49087.90395854561,49555.56348449855,49997.281472940296,50402.86545713422,50766.46885276464,51086.1487667176,51363.02498512865,51600.15762554549,51801.29718155461,51969.67406083088,52106.98972384074,52212.74540450882,52283.99934761081,52315.588292055705,52300.784771461964,52232.30412208053,52103.52454827996,51909.75515842632,51649.378461857494,51324.712106192834,50942.47240321202,50513.77551886559,50053.6714614941,49580.26209618258,49113.501243652165,48673.805220061375,48280.61628387673,47951.05695855766,47698.795353763904,47533.21208711787,47458.92485912188,47475.6903684964,47578.66964974778,47759.01423311731,48004.70780570215,48301.58198720193,48634.414611679436,48988.015434014436,49348.20710832078,49702.62004608386,50041.23794364131,50356.65615729678,50644.04536354318,50900.84509457054,51126.24190321448,51320.51111064056,51484.31618670921,51618.06340530842,51721.400971192066,51792.931912781634,51830.181151538025,51829.82269930019,51788.13742170235,51701.63994688947,51567.78954120817,51385.68752264202,51156.66473279738,50884.676485780365,50576.44694547747,50241.3362603321,49890.93736854292,49538.44056995636,49197.82891336195,48882.98354728903,48606.7843337039,48380.287480499304,48212.050188475485,48107.654543138684,48069.46169640496,48096.605289804735,48185.212164703196,48328.820260678855,48518.94916780249,48745.7686685686,48998.80506594814,49267.62441117099,49542.43603503298,49814.5690114476,50076.78787540741,50323.43108087139,50550.37474850697,50754.84318442825,50935.104262438086,51090.0999637494,51219.068571364245,51321.21429417277,51395.47242258697,51440.40432821738,51454.238392675004,51435.05257785422,51381.07447975945,51291.057991285285,51164.68439191535,51002.931322488475,50808.35621652633,50585.250814132196,50339.6388420321,50079.1075785664,49812.483269707984,49549.37775771464,49299.64720198906,49072.81207527944,48877.490163154864,48720.891289893196,48608.41478252842,48543.37952046914,48526.9032603919,48557.93420774505,48633.42479582009,48748.62633193341,48897.47431021877,49073.02826156946,49267.927312276544,49474.82328782095,49686.75717820301,49897.45180429322,50101.50303746211,50294.46308258822,50472.821032205575,50633.89688261639,50775.67421270924,50896.602677157716,50995.40360064781,51070.90997167172,51121.966227100405,51147.40409896199,51146.099599876194,51117.10438212959,51059.83375753828,50974.2850504939,50861.25481114159,50722.52240185611,50560.970679082515,50380.62143427813,50186.57294751644,49984.83812952174,49782.09284313358,49585.353753502655,49401.61237526305,49237.45617021764,49098.70834237125,48990.11553752552,48915.10748416157,48875.64544847819,48872.168072518296,48903.63457963208,48967.6572469158,49060.70810966446,49178.37957538293,49315.67532616635,49467.30676870192,49627.9713892765,49792.59256541471,49956.505383599855,50115.57934864051,50266.275925453636,50405.64590167911,50531.27783827102,50641.21367865493,50733.85035633942,50807.846644296354,50862.05247576323,50895.47378256542,50907.28006755744,50896.855193180505,50863.88509431668,50808.47018884189,50731.24593560881,50633.4928358985,50517.217458316816,50385.1887296638,50240.918403745396,50088.58067286227,49932.872543884194,49778.82304241057,49631.564785733855,49496.085392876295,49376.97821639271,49278.21187118575,49202.936119714126,49153.33816343232,49130.55873451501,49134.67209733081,49164.72868264989,49218.85407657542,49294.39388397631,49388.09090110462,49496.279276316534,49615.08001169241,49740.58324680172,49869.00513714806,49996.81054909322,50120.79689467219,50238.13880185306,50346.39749802112,50443.50232557015,50527.714319036866,50597.58297505221,50651.90711130909,50689.70908187554,50710.22880784382,50712.94045728738,50697.59063478399,50664.253132694,50613.39214789644,50545.923786817046,50463.26493595211,50367.359256774114,50260.672079448435,50146.14905679554,50027.13720731904,49907.270960843554,49790.32955110474,49680.07515752981,49580.08326532728,49493.57760244957,49423.28167997659,49371.297499155255,49339.019594778685,49327.089539911176,49335.3926617933,49363.09534026131,49408.71818196729,49470.23784349519,49545.2085089816,49630.893135090824,49724.39460166606,49822.77781337795,49923.17547182952,50022.87249452844,50119.36665671601,50210.40569896235,50294.003594020964,50368.44063874166,50432.25332088297,50484.22037435889,50523.35104353302,50548.88039385189,50560.274695021246,50557.247704423746,50539.786381296355,50508.18246706812,50463.06474833909,50405.42588825546,50336.63759208841,50258.448581787874,50172.96130524413,50082.58532072751,49989.967631673266,49897.90262563684,49809.22642162444,49726.70211672183,49652.903473445826,49590.10490968809,49540.185233222015,49504.55147029044,49484.08750487611,49479.1302491762,49489.473907631465,49514.40078045271,49552.73516595019,49602.91542118398,49663.078242510484,49731.14879713705,49804.93048650875,49882.18881204241,49960.72495413909,50038.43613599169,50113.361464385045,50183.71354610433,50247.89759978963,50304.52086997962,50352.39579232143,50390.5404981426,50418.179884038065,50434.74966924699,50439.90473509806,50433.53174096624,50415.76471480101,50387.00119672424,50347.91572142142,50299.467066663085,50242.89582315728,50179.70944530189,50111.65295424535,50040.6647657292,49968.818554155485,49898.253476535756,49831.096307873726,49769.37995099006,49714.963286100356,49669.45737355505,49634.16262160352,49610.02073026524,49597.584110217125,49597.0041657774,49608.0384503028,49630.075377430505,49662.174017842335,49703.11562269596,49751.4629572656,49805.62333452912,49863.911406641695,49924.60826702497,49986.014171869094,50046.49311838792,50104.50851458979,50158.65013367262,50207.65336433832,50250.41236304425,50285.98902923885,50313.619737292516,50332.72148271942,50342.89858369949,50343.95039904241,50335.87977721685,50318.90124195762,50293.4473467738,50260.17127402533,50219.94366686862,50173.84188280799,50123.13032827188,50069.23122369066,50013.68598077528,49958.108254558916,49904.130564132916,49853.34706681578,49807.2555470041,49767.201894295154,49734.33027586662,49709.54186546621,49693.464412665235,49686.43417993368,49688.490914868766,49699.38564045061,49718.600216062645,49745.376916160036,49778.75574729506,49817.61691414403,49860.725766159034,49906.77770128742,49954.44084382002,50002.394804110474,50049.364409978356,50094.147907512845,50135.639697085215,50172.848139851194,50204.909295216305,50231.097603094386,50250.83449983224,50263.6957685952,50269.418108680715,50267.9050138424,50259.2316365479,50243.64794411702,50221.579198968066,50193.622661088775,50160.53944078849,50123.24062747221,50082.76716829636,50040.26343323868,49996.94493052761,49954.061170992776,49912.8551634021,49874.52140286441,49840.16445019886,49810.760266849655,49787.122360364585,49769.87452038916,49759.43151124271,49755.98857362224,49759.52002203689,49769.78665649269,49786.35118513103,49808.60042105369,49835.77270354647,49866.9888209316,49901.28468506568,49937.64411791396,49975.030337887874,50012.415046710135,50048.804378566354,50083.26134145536,50114.92471663749,50143.02465219026,50166.895366361525,50185.98545284193,50199.86625341361,50208.238646540376,50210.93841718529,50207.94015523632,50199.359413048194,50185.452671891624,50166.61455271473,50143.37168008926,50116.37267978152,50086.373957821925,50054.22115791585,50020.826499944,49987.14253332572,49954.133159963516,49922.74305786928,49893.86683852114,49868.319376441366,49846.80874634808,49829.913090403505,49818.06252518207,49811.52690399556,49810.40990124198,49814.64951186623,49824.02469239147,49838.16753998695,49856.58013777082,49878.65500637065,49903.698003846526,49930.95250980855,49959.62380774473,49988.90272748633,50017.98780685675,50046.10545381694,50072.527812860375,50096.58823876109,50117.69443795934,50135.3394399406,50149.110602401524,50158.69683710425,50163.894177764945,50164.609712748315,50160.863793363904,50152.79032422318,50140.63486524451,50124.75024160524,50105.58937862763,50083.69515664814,50059.68721233915,50034.24578680228,50008.09292051915,49981.97150079541,49956.62285726557,49932.76375510995,49911.06373704637,49892.12380224846,49876.4573778053,49864.4744375975,49856.4694623945,49852.61372690918,49852.95216180253,49857.40479072175,49865.77250412933,49877.746721074756,49892.922322035105,49910.81312088791,49930.869087048406,49952.494529386735,49975.066506062634,49997.95281879683,50020.5290730193,50042.194421752174,50062.385745865315,50080.59014287296,50096.355690037526,50109.300508522174,50119.12018159117,50125.59357395561,50128.58706802461,50128.057186002494,50124.05151627789,50116.70782053658,50106.25117541897,50092.98900761727,50077.303918675534,50059.644265733354,50040.51256300173,50020.451887950534,50000.03060510959,49979.82584641777,49960.40629728703,49942.31492023249,49926.05229375402,49912.061247248785,49900.71343120119,49892.29837799901,49887.01548855828,49884.96923277216,49886.16768909058,49890.5243827972,49897.86322611744,49907.92622721253,49920.38352817109,49934.845260090995,49950.87466868299,49968.00196550631,49985.73839380941,50003.59005725008,50021.07113601551,50037.71619875048,50053.09140114938,50066.804435171485,50078.51315065942,50087.93281028823,50094.84195863496,50099.08688885051,50100.58468038099,50099.32476453671,50095.36895838634,50088.84989829225,50079.967808207264,50068.985558662935,50056.22201187257,50042.04370566643,50026.85500065529,50011.08689550687,49995.18479741911,49979.59561093375,49964.754570448094,49951.07228342735,49938.92246737978,49928.63085136344,49920.46567190009,49914.63012596395,49911.25705476804,49910.40602777041,49912.06288423695,49916.14167783351,49922.488865798165,49930.88949498646,49941.07506766373,49952.732723564455,49965.51535252953,49979.05225293687,49992.95997223163,50006.853002736956,50020.35405322219,50033.10366873117,50044.76902251258,50055.05174997756,50063.694732146556,50070.48776333757,50075.27205483902,50077.94353452926,50078.454904644575,50076.81641979233,50073.09534880525,50067.41409084262,50059.94693120882,50050.91544744065,50040.58261163213,50029.24567949882,50017.22800767562,50004.869994373046,49992.51939021536,49980.5212710268,49969.20799796931,49958.88950899312,49949.84428644598,49942.31132776727,49936.4834099271,49932.501885653975,49930.45318388205,49930.36711262003,49932.21698458316,49935.92150958196,49941.34832765243,49948.31899735592,49956.615207622846,49965.98595076059,49976.1553792326,49986.83106868239,49997.712422445235,50008.49897561575,50018.898386291265,50028.63393441464,50037.45138148494,50045.125074638454,50051.463204408195,50056.31214596885,50059.55982898389,50061.138092277324,50061.02398813022,50059.24000912542,50055.85322028362,50050.973292663984,50044.74945300505,50037.366387975875,50029.03917092106,50020.007312450565,50010.52807188112,50000.86920181089,49991.30133009399,49982.09020924574,49973.48908021218,49965.73140341822,49959.024203810055,49953.54225792954,49949.42332058246,49946.764548020554,49945.620226166495,49946.000859264605,49947.87361973089,49951.16410721476,49955.75931700777,49961.5116774653,49968.24398486676,49975.755043129415,49983.82580517115,49992.22581184039,50000.719731891804,50009.07382068046,50017.06213400539,50024.47235475759,50031.111111800674,50036.80869129341,50041.42305942072,50044.84313176879,50046.991238459574,50047.824746239035,50047.336809932334,50045.556237155615,50042.54646300861,50038.40364656954,50033.25391890393,50027.24983305012,50020.566089588196,50013.3946359584,50005.9392622731,49998.40983924688,49991.01636323871,49983.96298749235,49977.44222598044,49971.62951572466,49966.67831454619,49962.71589397753,49959.83996223886,49958.11622102003,49957.57692404256,49958.220467050065,49960.01200017073,49962.88501665702,49966.74383875714,49971.466893476,49976.91064934033,49982.914070562736,49989.303437244824,49995.89737902311,50002.51197398993,50008.96577368816,50015.0846272034,50020.70619158671,50025.684030888515,50029.89122105429,50033.22339222305,50035.601153304364,50036.97185612961,50037.31066829058,50036.62093546965,50034.933826183136,50032.30726490869,50028.82417389988,50024.59005972678,50019.72999754614,50014.385083818954,50008.70844589905,50002.86091363183,49997.00647271004,49991.307630911026,49985.92083544275,49980.992081637334,49976.65284961564,49973.016496157994,49970.17521407789,49968.197651540875,49967.12725995872,49966.9814125549,49967.75130784017,49969.40264450951,49971.87702807525,49975.09404611349,49978.95392930895,49983.340700211425,49988.12570110179,49993.17138660997,49998.335265431466,50003.47387811195,50008.446703704874,50013.119896366734,50017.369762860784,50021.08590279301,50024.17394465054,50026.557821974726,50028.18154509116,50029.01043472791,50029.03179469773,50028.25501181578,50026.71108258935,50024.45157813186,50021.54707129154,50018.08506307217,50014.16745880642,50009.90765780915,50005.42733281663,50000.852986734186,49996.31238335748,49991.930955106305,49987.82829382313,49984.114829907216,49980.88880025669,49978.23359670027,49976.2155740937,49974.8823815513,49974.261862106345,49974.36154631401,49975.168744903196,49976.65122551101,49978.75843972886,49981.42324994349,49984.5640914056,49988.0874940246,49991.890880777435,49995.86555533018,49999.899790308475,50003.88192927729,50007.70341945911,50011.26169804791,50014.46286217706,50017.22406074145,50019.4755550034,50021.16240397931,50022.245739860526,50022.7036081171,50022.53135647683,50021.74156672032,50020.36353322637,50018.44230245961,50016.03729804244,50013.22056656114,50010.074689558984,50006.690416941885,50003.164085838725,49999.5948963835,49996.08212144582,49992.722330630255,49989.6067095623],"type":"scatter","name":"source2_p_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1,0.1001,0.1002,0.1003,0.1004,0.1005,0.1006,0.1007,0.1008,0.1009,0.101,0.1011,0.1012,0.1013,0.1014,0.1015,0.1016,0.1017,0.1018,0.1019,0.102,0.1021,0.1022,0.1023,0.1024,0.1025,0.1026,0.1027,0.1028,0.1029,0.103,0.1031,0.1032,0.1033,0.1034,0.1035,0.1036,0.1037,0.1038,0.1039,0.104,0.1041,0.1042,0.1043,0.1044,0.1045,0.1046,0.1047,0.1048,0.1049,0.105,0.1051,0.1052,0.1053,0.1054,0.1055,0.1056,0.1057,0.1058,0.1059,0.106,0.1061,0.1062,0.1063,0.1064,0.1065,0.1066,0.1067,0.1068,0.1069,0.107,0.1071,0.1072,0.1073,0.1074,0.1075,0.1076,0.1077,0.1078,0.1079,0.108,0.1081,0.1082,0.1083,0.1084,0.1085,0.1086,0.1087,0.1088,0.1089,0.109,0.1091,0.1092,0.1093,0.1094,0.1095,0.1096,0.1097,0.1098,0.1099,0.11,0.1101,0.1102,0.1103,0.1104,0.1105,0.1106,0.1107,0.1108,0.1109,0.111,0.1111,0.1112,0.1113,0.1114,0.1115,0.1116,0.1117,0.1118,0.1119,0.112,0.1121,0.1122,0.1123,0.1124,0.1125,0.1126,0.1127,0.1128,0.1129,0.113,0.1131,0.1132,0.1133,0.1134,0.1135,0.1136,0.1137,0.1138,0.1139,0.114,0.1141,0.1142,0.1143,0.1144,0.1145,0.1146,0.1147,0.1148,0.1149,0.115,0.1151,0.1152,0.1153,0.1154,0.1155,0.1156,0.1157,0.1158,0.1159,0.116,0.1161,0.1162,0.1163,0.1164,0.1165,0.1166,0.1167,0.1168,0.1169,0.117,0.1171,0.1172,0.1173,0.1174,0.1175,0.1176,0.1177,0.1178,0.1179,0.118,0.1181,0.1182,0.1183,0.1184,0.1185,0.1186,0.1187,0.1188,0.1189,0.119,0.1191,0.1192,0.1193,0.1194,0.1195,0.1196,0.1197,0.1198,0.1199,0.12,0.1201,0.1202,0.1203,0.1204,0.1205,0.1206,0.1207,0.1208,0.1209,0.121,0.1211,0.1212,0.1213,0.1214,0.1215,0.1216,0.1217,0.1218,0.1219,0.122,0.1221,0.1222,0.1223,0.1224,0.1225,0.1226,0.1227,0.1228,0.1229,0.123,0.1231,0.1232,0.1233,0.1234,0.1235,0.1236,0.1237,0.1238,0.1239,0.124,0.1241,0.1242,0.1243,0.1244,0.1245,0.1246,0.1247,0.1248,0.1249,0.125,0.1251,0.1252,0.1253,0.1254,0.1255,0.1256,0.1257,0.1258,0.1259,0.126,0.1261,0.1262,0.1263,0.1264,0.1265,0.1266,0.1267,0.1268,0.1269,0.127,0.1271,0.1272,0.1273,0.1274,0.1275,0.1276,0.1277,0.1278,0.1279,0.128,0.1281,0.1282,0.1283,0.1284,0.1285,0.1286,0.1287,0.1288,0.1289,0.129,0.1291,0.1292,0.1293,0.1294,0.1295,0.1296,0.1297,0.1298,0.1299,0.13,0.1301,0.1302,0.1303,0.1304,0.1305,0.1306,0.1307,0.1308,0.1309,0.131,0.1311,0.1312,0.1313,0.1314,0.1315,0.1316,0.1317,0.1318,0.1319,0.132,0.1321,0.1322,0.1323,0.1324,0.1325,0.1326,0.1327,0.1328,0.1329,0.133,0.1331,0.1332,0.1333,0.1334,0.1335,0.1336,0.1337,0.1338,0.1339,0.134,0.1341,0.1342,0.1343,0.1344,0.1345,0.1346,0.1347,0.1348,0.1349,0.135,0.1351,0.1352,0.1353,0.1354,0.1355,0.1356,0.1357,0.1358,0.1359,0.136,0.1361,0.1362,0.1363,0.1364,0.1365,0.1366,0.1367,0.1368,0.1369,0.137,0.1371,0.1372,0.1373,0.1374,0.1375,0.1376,0.1377,0.1378,0.1379,0.138,0.1381,0.1382,0.1383,0.1384,0.1385,0.1386,0.1387,0.1388,0.1389,0.139,0.1391,0.1392,0.1393,0.1394,0.1395,0.1396,0.1397,0.1398,0.1399,0.14,0.1401,0.1402,0.1403,0.1404,0.1405,0.1406,0.1407,0.1408,0.1409,0.141,0.1411,0.1412,0.1413,0.1414,0.1415,0.1416,0.1417,0.1418,0.1419,0.142,0.1421,0.1422,0.1423,0.1424,0.1425,0.1426,0.1427,0.1428,0.1429,0.143,0.1431,0.1432,0.1433,0.1434,0.1435,0.1436,0.1437,0.1438,0.1439,0.144,0.1441,0.1442,0.1443,0.1444,0.1445,0.1446,0.1447,0.1448,0.1449,0.145,0.1451,0.1452,0.1453,0.1454,0.1455,0.1456,0.1457,0.1458,0.1459,0.146,0.1461,0.1462,0.1463,0.1464,0.1465,0.1466,0.1467,0.1468,0.1469,0.147,0.1471,0.1472,0.1473,0.1474,0.1475,0.1476,0.1477,0.1478,0.1479,0.148,0.1481,0.1482,0.1483,0.1484,0.1485,0.1486,0.1487,0.1488,0.1489,0.149,0.1491,0.1492,0.1493,0.1494,0.1495,0.1496,0.1497,0.1498,0.1499,0.15,0.1501,0.1502,0.1503,0.1504,0.1505,0.1506,0.1507,0.1508,0.1509,0.151,0.1511,0.1512,0.1513,0.1514,0.1515,0.1516,0.1517,0.1518,0.1519,0.152,0.1521,0.1522,0.1523,0.1524,0.1525,0.1526,0.1527,0.1528,0.1529,0.153,0.1531,0.1532,0.1533,0.1534,0.1535,0.1536,0.1537,0.1538,0.1539,0.154,0.1541,0.1542,0.1543,0.1544,0.1545,0.1546,0.1547,0.1548,0.1549,0.155,0.1551,0.1552,0.1553,0.1554,0.1555,0.1556,0.1557,0.1558,0.1559,0.156,0.1561,0.1562,0.1563,0.1564,0.1565,0.1566,0.1567,0.1568,0.1569,0.157,0.1571,0.1572,0.1573,0.1574,0.1575,0.1576,0.1577,0.1578,0.1579,0.158,0.1581,0.1582,0.1583,0.1584,0.1585,0.1586,0.1587,0.1588,0.1589,0.159,0.1591,0.1592,0.1593,0.1594,0.1595,0.1596,0.1597,0.1598,0.1599,0.16]},{"mode":"lines","y":[0.0,0.0,0.0,8.485300021614665e-6,0.0006717598270588444,0.005478597106249517,0.022908531119390996,0.0641237693973595,0.13853444387724848,0.25407907884021524,0.41696441575232457,0.645323889100627,0.9557102682971723,1.3641388759481867,1.867468070474165,2.4499044388616387,3.0981558305457915,3.789869968210491,4.516581569898385,5.236522269930225,5.917614964716336,6.500632612632381,6.950173802008566,7.2317368325175675,7.325171006467222,7.227857682050807,6.932921322158929,6.472743023433404,5.868242480619173,5.18810813894714,4.479801568497706,3.8236911703400045,3.2749924860262936,2.8997036105748943,2.7449036793824915,2.844296154464855,3.220841340440953,3.8639199672701134,4.76155310965277,5.861106350284521,7.118130629045316,8.454039629979897,9.80447327921413,11.087378900468668,12.237128092440614,13.191568980739566,13.903790479515044,14.348734401114625,14.51233891666108,14.412035530770112,14.070271831562398,13.539778092065598,12.871952937827206,12.137924351131437,11.401873358029976,10.73229623703597,10.186961329285417,9.813283577162538,9.646766410864078,9.703104014277532,9.987585402016634,10.483491826268901,11.167351608256794,11.99822473526188,12.93321915342209,13.920810105625954,14.912278960131387,15.86012795295855,16.723302948493686,17.46984441102941,18.0760232907622,18.531186925860496,18.832598308508608,18.99023033255603,19.019834586470544,18.94606604483232,18.796135943125805,18.600684360185824,18.389567422795317,18.190852842129555,18.029254032225438,17.924169045530263,17.890374240812413,17.93607606623084,18.06495114054724,18.2748679754192,18.560267004756753,18.911765590119533,19.318112609543444,19.766517068355988,20.243808579721374,20.73713768044497,21.234350894326838,21.724709932369503,22.198733289924473,22.648710943666316,23.068337604073825,23.45296830053292,23.799300917407262,24.105444318048697,24.370765436121886,24.595881727186864,24.782643476683365,24.93412232174862,25.05460876670778,25.149592449665775,25.2256436647724,25.29032355516029,25.351878414673813,25.419014968256235,25.500433153808924,25.604454572622174,25.738509044424067,25.908683827055803,26.119324581628447,26.372636849790787,26.66854611469807,27.004497392384224,27.37564836342518,27.77496675025513,28.193730367071865,28.62193026191416,29.048956014322076,29.464209155541646,29.857792302747246,30.221160723731945,30.547628685068382,30.832865902551347,31.0750849479657,31.275232152914842,31.436809963801274,31.565704244778615,31.669697321640697,31.757984173107204,31.84051509977013,31.927345336906644,32.02798715851357,32.15078201444835,32.302430511074505,32.48756140978224,32.708550416293015,32.96539430267742,33.25584504285647,33.57560362342247,33.918710474047835,34.27798255405294,34.64553825701022,35.01333989280008,35.373706771465216,35.719808562118,36.04603734225982,36.34832590249515,36.62429946781266,36.87335383038062,37.096573182879645,37.29657500915317,37.47724236977127,37.64340396427473,37.800470881552705,37.954055140327185,38.109613545978554,38.27211008607179,38.44575352239611,38.63378236043877,38.838346490485804,39.06044979333841,39.29998232200519,39.55580968620801,39.82592384984314,40.107633179331174,40.3977762307859,40.69294940625151,40.98972277001316,41.28484475087508,41.57541001830407,41.85899809777927,42.13376455680917,42.3984950455188,42.652614847109724,42.896163654107504,43.12973821544247,43.35441009271511,43.57162761756172,43.78310625103866,43.99071850125595,44.19638497166372,44.401976133879266,44.60922464638959,44.81965425428827,45.03452424494677,45.25479164091489,45.48108990685532,45.71372338863593,45.95267635374667,46.19763422468418,46.44801595004186,46.70301463428012,46.96164526688464,47.222796858744154,47.48528755096686,47.74792036488696,48.00953785624916,48.26907362460802,48.52559881910964,48.7783618168322,49.026819452710534,49.270658336398945,49.50980530627402,49.744426237462456,49.97491328370122,50.20186081035635,50.42603127015431,50.64831250424807,50.869668732703126,51.09108779258287,51.313527444666995,51.537863867500974,51.764845065837775,51.99505211165755,52.228870175671915,52.46647128723352,52.70780950283054,52.952628892810374,53.20048355840366,53.450768427101146,53.70275876157045,53.955655801064694,54.208635720444796,54.46089869819539,54.71171524771467,54.96046683693247,55.20667863335782,55.45004240217088,55.69042859220595,55.92788710204535,56.162637142215864,56.39504719203559,56.62560666492038,56.85489142304782,57.08352547102252,57.31214147174909,57.541342495933065,57.77166743149997,58.00356193894872,58.23735656534836,58.47325295535638,58.711318632510284,58.951490199241924,59.19358429277393,59.437315203591965,59.68231765238593,59.928173060760265,60.17443744443336,60.42066916499661,60.66645480906741,60.91143177199283,61.1553063599318,61.397866628456676,61.63898952906756,61.87864232457223,62.11687859663886,62.353829454949754,62.58969083405287,62.82470789799268,63.05915770569127,63.29333126786771,63.52751610404753,63.76198026309007,63.99695862386902,64.23264208006681,64.46916999861683,64.70662612113257,64.945037858881,65.1843787537939,65.4245737069841,65.66550646999373,65.90702880088139,66.14897066036312,66.39115081335589,66.63338724252242,66.87550683891948,67.11735392152062,67.35879723751252,67.59973520099965,67.84009924035388,68.07985522575382,68.31900304906348,68.55757450831048,68.79562972160805,69.03325234556999,69.27054391103425,69.507617606846,69.74459184613995,69.98158393816666,70.21870416384282,70.45605051811732,70.69370433691411,70.93172697556064,71.17015764871508,71.40901248362354,71.64828477970491,71.8879464111718,72.12795025742409,72.36823350066807,72.60872159319803,72.84933267007996,73.08998216704495,73.33058739996983,73.57107187041464,73.8113690819819,74.05142568293418,74.29120379048173,74.53068239897046,74.76985782566054,75.00874320084591,75.24736706082611,75.48577115000693,75.72400757908952,75.962135518363,76.20021762586794,76.43831641954745,76.67649079922552,76.91479290957578,77.15326550942983,77.39193997834828,77.63083505006571,77.86995631715558,78.10929650484888,78.34883646701738,78.58854681703964,78.82839007222819,79.0683231654353,79.30830016173732,79.54827501295907,79.78820418783559,80.02804903022206,80.26777772069332,80.50736674607845,80.74680181533563,80.98607819607747,81.2252004814651,81.46418183077411,81.70304275569279,81.9418095476241,82.18051245730908,82.41918374669122,82.65785573376415,82.89655894478739,83.13532047503801,83.37416264143602,83.61310198793416,83.85214868028282,84.09130630179621,84.33057203739259,84.56993721144214,84.8093881265652,85.04890713628622,85.28847387539933,85.5280665677349,85.76766333184482,86.00724341060035,86.24678825999666,86.48628244487111,86.7257143038158,86.96507636120651,87.20436548006917,87.44358276445863,87.6827332335217,87.9218253005329,88.16087009865873,88.39988070083206,88.63887128357504,88.87785628411618,89.11684959727775,89.35586385289221,89.59490980759034,89.83399587622017,90.07312781887181,90.31230859021832,90.55153834855486,90.79081461408792,91.03013255832778,91.26948540124047,91.5088648883496,91.74826181767179,91.9876665858381,92.22706972348051,92.46646239245601,92.70583682120795,92.94518665888802,93.18450723439952,93.42379571222588,93.66305114254517,93.90227440888751,94.14146808156265,94.38063618947501,94.6197839266577,94.85891731233467,95.0980428247686,95.33716702945422,95.57629622165547,95.8154361013731,96.06105157905706,96.65762364225016,98.34586963297102,100.74133084273976,102.89531804007615,103.50786137264947,102.8770842617745,101.29644958106918,99.1080589550174,95.45261319770133,89.35920286934102,80.49807483196864,69.07306468881625,56.22728434545256,42.35810475755831,28.3961346831341,14.45056778089587,1.603819477339254,-9.552364980190333,-17.942084455354316,-23.116865095412223,-24.803873327164375,-22.975694237806238,-17.998669540294383,-10.065006487592914,-0.004010401015767773,11.674986673920046,23.814383979278404,35.71241439884001,46.300163589917396,54.96407399253628,61.02874343882092,64.11966547228919,64.07767270615275,60.87386058916205,54.87931009522365,46.430076618430355,36.27649968289212,25.00160020432362,13.482583340976065,2.3689551718708106,-7.5837224297630055,-15.844172108399492,-21.966960938366082,-25.681615826002417,-26.935501329460635,-25.77312555427739,-22.497145141335476,-17.424663218675565,-11.085438103650159,-3.9550960739338863,3.3798561620036405,10.431029057193408,16.71612205818405,21.866204519082522,25.61213844317681,27.786825696443227,28.369737496100075,27.411726924484807,25.105309444291812,21.674093037414416,17.438311983393938,12.709374656951855,7.831424793258231,3.1087754980600337,-1.1816818451908686,-4.8195857535036275,-7.651771164916671,-9.58266071884858,-10.59630755975239,-10.722265435180448,-10.057453478133661,-8.724285725578817,-6.884832048732912,-4.704984762169682,-2.358646544206123,-0.0056836706141738524,2.2118427373011214,4.17652847338524,5.803183445472089,7.03378247654073,7.844697357880925,8.235254199688402,8.232054862796438,7.8765226642314525,7.225692087322601,6.342463429930436,5.29376699161204,4.145060437045253,2.957424557388252,1.7859702703304148,0.6772365130005076,-0.3298942114659231,-1.205785550935592,-1.9282189173989628,-2.4829527436094665,-2.8621324012355736,-3.0642129116179118,-3.093050451752274,-2.957625858047071,-2.6716422593930242,-2.253270119790833,-1.7247629463072074,-1.1121696116781603,-0.4445177856546866,0.24671540646439505,0.9290135397376922,1.569756385097728,2.138017389753557,2.6058645447203794,2.950063205354956,3.1534547061240903,3.2060081374899747,3.105835627156336,2.8591847198331157,2.4806764569758126,1.9921386098561942,1.4218640414976553,0.8025403133743398,0.16964362447223524,-0.44100027525291474,-0.9952695202696406,-1.4629053941539456,-1.8194259395765944,-2.047440132197529,-2.1378594937078796,-2.090048934732941,-1.9119847451331453,-1.619253112391717,-1.234123975862503,-0.7837186462289825,-0.2982913807315556,0.19090925490502608,0.6536391193404326,1.0625332188528098,1.3948183433343448,1.6334668874141447,1.7681740847799312,1.795561483111077,1.719209618252382,1.54895074827445,1.3000286808605614,0.9917531336864052,0.6460930684028579,0.28610149780158245,-0.06562702520900437,-0.38833911347811834,-0.6643803865613669,-0.8800633509979816,-1.0263460529921469,-1.0990349873748775,-1.0987816305870726,-1.0306594323713834,-0.9035697500065683,-0.7293852366481448,-0.5219991986721004,-0.2963091774606405,-0.06720236002765162,0.15134823359346702,0.34709819823290355,0.510121689999848,0.6332555441361472,0.712285081679123,0.7459705229883823,0.7358534215374566,0.6859315499403906,0.602199051732853,0.492111520645406,0.36401645521466885,0.22657491037749938,0.08823479680564775,-0.04324711846384527,-0.16117752890767517,-0.2602144921665762,-0.3365334591328907,-0.38789976877923815,-0.4136353786474995,-0.4145063578617521,-0.3925406864287442,-0.35079595638614036,-0.29309939607973673,-0.22377141050278915,-0.1473579501348256,-0.06837561690167,0.008909985145750893,0.0806736769443924,0.14367601718561482,0.19536803031150313,0.233952160663891,0.25840371460962336,0.26845598148082006,0.2645531320071714,0.24777737348930445,0.2197541325882683,0.18254246062409496,0.13851431742657155,0.09022917810524238,0.04030771429392027,0.20343455077674136,-6867.705761931216,-29325.202158180302,-33722.664358982445,-31744.914937102803,-20814.747140848325,-27903.133376183552,-36096.11357697417,-43346.00786040732,-37993.76430838608,-35865.05572718616,-30169.321977852258,-22880.24119725519,-16814.538708900705,-13402.508976606985,-11627.73433641517,-9372.32848995969,-7183.310220887955,-6112.772661352043,-8545.387359512762,-14112.24556815122,-22310.51848766765,-31406.837967363626,-41086.5072055941,-50972.218938915,-60774.40328400977,-69784.70547730612,-76936.53016340877,-81948.8100271376,-84483.19384378789,-85179.0534605689,-84075.1406531218,-81738.8113713655,-78146.81784404194,-73674.22730480041,-68426.96430125192,-62715.656289899096,-56850.89439242665,-50954.22246049334,-45242.06698004193,-39734.51632590011,-34717.114645434245,-30252.548144656368,-26601.350095612284,-23763.143475350334,-21819.388928229364,-20670.56842280077,-20244.294498795814,-20393.37925200736,-20917.997515091673,-21668.148661003033,-22480.044817133312,-23295.507622729685,-23987.959679380037,-24646.770533123527,-25355.478242678124,-26277.5255779516,-27526.237734613023,-29229.197552930542,-31478.16754332585,-34321.99684686603,-37757.18627481803,-41690.60301016127,-45981.61742405414,-50406.68772605708,-54713.2424387347,-58613.2744278406,-61852.6427208715,-64210.40031310078,-65546.38634575965,-65796.20917892904,-64977.261572609365,-63149.37631271464,-60445.32926563537,-57077.45390482748,-53218.69749030567,-49035.722218907315,-44694.37370874976,-40376.93771260277,-36256.12247196426,-32489.929879958603,-29207.023893153877,-26510.366955597823,-24471.110183264427,-23125.64252917657,-22477.273954822165,-22481.37334270901,-23045.04475900919,-24039.275967920723,-25323.08934195817,-26779.176852440978,-28316.61956168312,-29839.7230793848,-31298.623548238815,-32722.93911978409,-34179.18101867872,-35733.58099713386,-37440.25879211694,-39334.69767792392,-41430.89622682352,-43715.43902630333,-46137.57969983639,-48599.67897079749,-50972.34753177908,-53113.17055675911,-54881.3004694643,-56151.852262896005,-56827.04164468918,-56848.04883695661,-56199.45580664459,-54891.21643091364,-52988.491669317686,-50619.14148654804,-47897.46072820762,-44930.84129404144,-41837.71539531417,-38748.79755041588,-35796.03543785116,-33097.953603698064,-30752.161088692592,-28831.300668437605,-27385.40188724372,-26445.16384644716,-26016.094422329385,-26071.056038845767,-26552.915907823306,-27384.31711055254,-28479.67112125522,-29756.0194901962,-31154.971349509822,-32630.512062923313,-34124.75458703286,-35612.204092555694,-37111.85799250706,-38652.82440243546,-40251.549198902285,-41906.854985260856,-43604.226966372335,-45318.09714894701,-47004.96947347075,-48600.991470124034,-50030.23063858287,-51213.288965529784,-52075.52757544247,-52554.30905765145,-52605.500240427784,-52208.14625827798,-51366.81551974779,-50110.09889734475,-48474.19775954215,-46530.47749803332,-44386.51670442755,-42127.04769031406,-39828.134710611186,-37572.00001015008,-35446.51877070219,-33535.28057039274,-31905.18806585599,-30601.950485655423,-29655.18250609273,-29079.457615879335,-28870.925168968384,-29006.671204034843,-29448.0438846035,-30146.48945322205,-31049.933840254085,-32108.241399160943,-33277.39863495478,-34530.719891972716,-35850.00572077634,-37204.251890778,-38573.387494506096,-39961.10709840418,-41370.70579144442,-42790.294269597725,-44194.71857809913,-45548.88639815312,-46807.93991282531,-47923.45776208053,-48847.02893355689,-49532.13275089854,-49937.29006721302,-50030.657629846726,-49794.457699815735,-49227.64758157566,-48346.20450931303,-47181.577233460484,-45778.161439700176,-44186.97709642335,-42463.39511734519,-40680.94269379867,-38918.27572536628,-37234.217724297625,-35678.812727561955,-34300.28143130409,-33141.4102101775,-32235.730814659873,-31604.801117949566,-31253.72673795659,-31172.753126561292,-31341.702272414845,-31733.797203411574,-32318.147424884344,-33061.893717594925,-33932.70261395171,-34901.32818658769,-35943.35106281852,-37039.504108798275,-38176.02544397005,-39345.4341481902,-40530.392893811084,-41704.61037282633,-42852.799899907055,-43960.09415618253,-45001.06113639896,-45938.21770061936,-46731.611865478386,-47345.933742288085,-47752.7367550736,-47930.182316945065,-47863.468809766106,-47546.77162014021,-46985.450174034646,-46196.801947406864,-45208.96895035442,-44058.719309513996,-42789.03139525816,-41446.90331327771,-40081.24678741556,-38740.722018247005,-37471.49165080626,-36315.1408364131,-35306.916307210144,-34474.40932459496,-33836.679818747056,-33403.87267044062,-33177.37775489784,-33150.572290741766,-33310.11534959398,-33637.65371935533,-34111.751535400916,-34709.786340043414,-35409.58477201472,-36190.57926485796,-37034.37107941645,-37924.66814334702,-38846.689267310976,-39786.20472987337,-40728.439609237554,-41657.08015776223,-42553.59501302144,-43397.03427735101,-44164.383189222484,-44831.467050373314,-45374.31077550818,-45770.79496456647,-46002.40200789094,-46055.83945018821,-45924.34662992906,-45608.54138858637,-45116.72968568597,-44464.66990581242,-43674.84449059528,-42775.330501786,-41798.382467557734,-40778.84132464839,-39752.47621125145,-38754.35086396984,-37817.29391922165,-36970.54075910012,-36238.60741733573,-35640.450879101074,-35188.96270414853,-34890.82950741891,-34746.77135425237,-34752.136889819056,-34897.79559271268,-35171.22964748104,-35557.69970060539,-36041.34759570073,-36606.109804481894,-37236.34715329552,-37917.14426041994,-38634.287765905334,-39373.98500958296,-40122.42638208651,-40865.31697653051,-41587.50514600164,-42272.81616019945,-42904.16401724735,-43463.96845087225,-43934.855665616524,-44300.576944367735,-44547.04556392015,-44663.37418883234,-44642.79374789138,-44483.35086277364,-44188.309543097304,-43766.21937138848,-43230.64899456226,-42599.615704213444,-41894.764736153695,-41140.36534482297,-40362.19539995601,-39586.384771438636,-38838.28269803216,-38141.40752519732,-37516.530206536096,-36980.93544644302,-36547.896131395595,-36226.385901001224,-36021.04071594981,-35932.36227140555,-35957.13551310706,-36089.01166303124,-36319.19084527863,-36637.12839796163,-37031.189025800035,-37489.18425502794,-37998.74972487108,-38547.546647172894,-39123.301370302914,-39713.72363449641,-40306.36320950955,-40888.47339985848,-41446.94719294259,-41968.37852655431,-42439.279715579,-42846.460058536264,-43177.544176417956,-43421.5855743194,-43569.71461704996,-43615.75258248413,-43556.72546920732,-43393.22182939888,-43129.55593476741,-42773.71795241286,-42337.113321507066,-41834.111469790485,-41281.437667822895,-40697.45059922977,-40101.35241314925,-39512.37852782796,-38949.012261822216,-38428.265379633696,-37965.060329021246,-37571.74346078653,-37257.75066630354,-37029.43741562666,-36890.07410306134,-36839.99536443192,-36876.87968688233,-36996.12475991064,-37191.27642392444,-37454.46632449743,-37776.81639323318,-38148.77696647568,-38560.3785957393,-39001.39337747537,-39461.41742055835,-39929.89933915357,-40396.1483218448,-40849.358109274515,-41278.67982708165,-41673.3677748605,-42023.00948709165,-42317.836718693834,-42549.0997146131,-42709.47529601436,-42793.47156746252,-42797.78936414309,-42721.60305189983,-42566.73031670665,-42337.670862041596,-42041.505856057745,-41687.661900459316,-41287.553868062394,-40854.129271896236,-40401.342484779845,-39943.590149548916,-39495.13980550073,-39069.58251061012,-38679.33743489151,-38335.23230407607,-38046.17833743655,-37818.95204867248,-37658.08911042678,-37565.88772347482,-37542.511108719824,-37586.17160626202,-37693.373303135486,-37859.18696405443,-38077.53091077915,-38341.43457190923,-38643.26737860074,-38974.92368577881,-39327.9632524487,-39693.71515224887,-40063.35951761161,-40428.005259344936,-40778.782323005995,-41106.96415393596,-41404.130368388855,-41662.37210191909,-41874.53430810312,-42034.4816463499,-42137.368629331366,-42179.891208498644,-42160.496370778535,-42079.52857416914,-41939.29653622488,-41744.05028117445,-41499.86557126884,-41214.44002550919,-40896.81163277643,-40557.015485762335,-40205.69812723621,-39853.71085104616,-39511.70372373363,-39189.74115225991,-38896.957703200875,-38641.269753473505,-38429.154588229685,-38265.50394030455,-38153.55392313638,-38094.88816433801,-38089.50610894356,-38135.944399505825,-38231.43642079267,-38372.09389458755,-38553.0950193959,-38768.86601972654,-39013.246782834496,-39279.63596677595,-39561.115860006874,-39850.56161997635,-40140.742662611476,-40424.42545627231,-40694.48659828983,-40944.04291591037,-41166.6017997583,-41356.23061198857,-41507.7394863725,-41616.86782919394,-41680.46191526363,-41696.629551194135,-41664.85802012926,-41586.08336372614,-41462.702225946225,-41298.52155865422,-41098.64597986579,-40869.306997816675,-40617.642255152,-40351.436105226414,-40078.835026860994,-39808.05254616341,-39547.078481433884,-39303.40653434354,-39083.79262625582,-38894.05405312196,-38738.916656647576,-38621.91396136008,-38545.338819895,-38510.24479194899,-38516.491531109976,-38562.8261427138,-38646.99104039098,-38765.84842231215,-38915.512144382286,-39091.47938129215,-39288.75679588554,-39501.97863439872,-39725.516825236366,-39953.58538410475,-40180.34289013843,-40399.99728886185,-40606.91674321008,-40795.74879596448,-40961.54796657145,-41099.9094208562,-41207.103905919226,-41280.207102636814,-41317.215215617834,-41317.13818907762,-41280.062472411075,-41207.17669809469,-41100.75580408258,-40964.101784237355,-40801.442099552616,-40617.789551048656,-40418.76986084723,-40210.4251478709,-39999.00280178025,-39790.739899362285,-39591.65327165613,-39407.34465989699,-39242.82916823744,-39102.39353108418,-38989.48868513256,-38906.65891419532,-38855.50758047055,-38836.69734245227,-38849.980959870925,-38894.25745365026,-38967.64763762924,-39067.58292249648,-39190.901790885466,-39333.94935829416,-39492.67680144857,-39662.73894087653,-39839.589678035125,-40018.57609720114,-40195.03268607558,-40364.37721835982,-40522.2093746648,-40664.41223854071,-40787.25554657995,-40887.49819603999,-40962.486236252975,-41010.24159095266,-41029.53623323665,-41019.946555157476,-40981.88325893547,-40916.593194665154,-40826.13106929591,-40713.300702696826,-40581.567339914895,-40434.94427889084,-40277.85859541539,-40115.00193003853,-39951.17307106971,-39791.119388064595,-39639.38404058158,-39500.16533690902,-39377.19370139734,-39273.630502220025,-39191.99158603655,-39134.0968672956,-39101.04584062373,-39093.21753807032,-39110.292342824476,-39151.29228096383,-39214.635995073586,-39298.20457051623,-39399.4147049492,-39515.29631044765,-39642.572407703345,-39777.73998617731,-39917.151234098594,-40057.09507613198,-40193.87921596589,-40323.91283205859,-40443.78973228083,-40550.37119795664,-40640.86703650729,-40712.912632238746,-40764.639158036676,-40794.733694316135,-40802.48587481688,-40787.817883064956,-40751.29515817118,-40694.115994189306,-40618.0792613619,-40525.53064666813,-40419.289002409314,-40302.555505058845,-40178.80927467956,-40051.693819593725,-39924.89910654134,-39802.044192182235,-39686.56518943906,-39581.612904537425,-39489.96380816979,-39413.94715209748,-39355.3900767965,-39315.58154944898,-39295.25499968222,-39294.58865412295,-39313.221871071844,-39350.285287823586,-39404.442337861445,-39473.939670560096,-39556.66418377607,-39650.20470883454,-39751.91680089323,-39858.989511279986,-39968.51338111702,-40077.549139518975,-40183.196678316315,-40282.6637983689,-40373.333997183974,-40452.83223517274,-40519.087237804844,-40570.38853143337,-40605.436139047495,-40623.38073568886,-40623.85212088761,-40606.97412374255,-40573.36450760988,-40524.11905595483,-40460.779751131406,-40385.28774459153,-40299.92259667996,-40207.229975434704,-40109.9405933095,-40010.88358721441,-39912.8977840679,-39818.74433002221,-39731.02400067454,-39652.10216981793,-39584.04392543045,-39528.56122293118,-39486.973303051316,-39460.180923970205,-39448.65431306486,-39452.43417655386,-39471.14465079042,-39504.01676027147,-39549.92077340977,-39607.40581114427,-39674.745144301865,-39749.98578019681,-39831.00114670335,-39915.54589047197,-40001.31197649647,-40085.985379517646,-40167.30267766473,-40243.106794452426,-40311.40100107601,-40370.40011394283,-40418.577638252136,-40454.70745625624,-40477.89857544496,-40487.621466788834,-40483.72465456034,-40466.44047179049,-40436.37925977332,-40394.51174426968,-40342.13983340219,-40280.856614823824,-40212.496842743705,-40139.079660950294,-40062.74567358157,-39985.69072643633,-39910.098882930106,-39838.0770653168,-39771.59368838192,-39712.42335339149,-39662.09931624997,-39621.87502298673,-39592.69554889498,-39575.17931708064,-39569.61003860862,-39575.9384365779,-39593.79301069651,-39622.4988794629,-39661.103606887824,-39708.408873581626,-39763.00687351923,-39823.32038730229,-39887.64557637209,-39954.19663652082,-40021.15152263354,-40086.69799538207,-40149.079238113845,-40206.63825070837,-40257.860157226125,-40301.41148281972,-40336.17538364722,-40361.281773336756,-40376.131300523404,-40380.41220868371,-40374.10925941087,-40357.50412273641,-40331.166924305595,-40295.93897317814,-40252.907054391646,-40203.37003269915,-40148.798852934735,-40090.79131481004,-40031.02322589641,-39971.19768161799,-39912.994277031656,-39858.02002034537,-39807.76359730886,-39763.554439405314,-39726.527792561064,-39697.596685715944,-39677.4313807897,-39666.44656826641,-39664.7962752036,-39672.37619125062,-39688.832905084324,-39713.57938503444,-39745.81593439147,-39784.55579929243,-39828.654595969405,-39876.842742149034,-39927.760110807736,-39979.992160179136,-40032.10682116498,-40082.69143522611,-40130.38903028377,-40173.93320189253,-40212.180839120716,-40244.14190958423,-40269.0055080063,-40286.16138943093,-40295.21626181481,-40296.004209751016,-40288.590763690605,-40273.27031448693,-40250.55679389693,-40221.167786306585,-40186.0024909626,-40146.11420144121,-40102.67819401655,-40056.95610447302,-40010.25801183532,-39963.90352913077,-39919.183221416715,-39877.32162997971,-39839.44308306817,-39806.541325652535,-39779.45381434656,-39758.84131159861,-39745.17318932196,-39738.71862982749,-39739.54370351577,-39747.514118294945,-39762.30328215826,-39783.40520140342,-39810.15165275806,-39841.733015146776,-39877.22212043847,-39915.6004750903,-39955.786208495745,-39996.66311200618,-40037.110139591925,-40076.0307438408,-40112.3814190717,-40145.19881908807,-40173.62481508147,-40196.928865531445,-40214.52709130014,-40225.997491672926,-40231.09080555178,-40229.73661881365,-40222.04444366905,-40208.29964542476,-40188.95426035791,-40164.61292716238,-40136.01433385237,-40104.00875165958,-40069.53237719964,-40033.57932500148,-39997.17219727492,-39961.33220182445,-39927.04979033666,-39895.256748709355,-39866.8005921344,-39842.422006096895,-39822.73593785991,-39808.216790079794,-39799.188008039884,-39795.81619340717,-39798.10972822575,-39805.921759337005,-39818.95727987236,-39836.78395310365,-39858.846254821205,-39884.48246178484,-39912.9439824712,-39943.416508352595,-39975.04245513992,-40006.94416013705,-40038.247301428164,-40068.104005758374,-40095.71511489936,-40120.35108664461,-40141.37101914086,-40158.23930940816,-40170.53949214031,-40177.98485611357,-40180.42550465477,-40177.85161399045,-40170.39274757085,-40158.3132025427,-40142.00349174966,-40121.96819506913,-40098.81054088537,-40073.21419521323,-40045.922835955294,-40017.71816742388,-39989.3970813739,-39961.748692792906,-39935.53197082847,-39911.45464856828,-39890.15403269689,-39872.180249529985,-39857.98236284586,-39847.89768719293,-39842.144504014126,-39840.818272769975,-39843.89132031939,-39851.21589316128,-39862.53037152813,-39877.468373206895,-39895.570418523654,-39916.29778526679,-39939.04815175961,-39963.17260571174,-39987.99358374241,-40012.82329972676,-40036.98221813786,-40059.81713085557,-40080.71840282009,-40099.13596446201,-40114.59364858795,-40126.70149810101,-40135.16571021057,-40139.7959337943,-40140.50969971303,-40137.33383857474,-40130.40282506496,-40119.95407984572,-40106.32035561038,-40089.91942897425,-40071.24140994578,-40050.83406125703,-40029.28658670426,-40007.212397429554,-39985.2313952846,-39963.95232171052,-39943.95570880429,-39925.77793748544,-39909.896858052096,-39896.71936396826,-39886.57123411755,-39879.689476026215,-39876.21731676904,-39876.20190326733,-39879.594692820436,-39886.25444063857,-39895.9526257557,-39908.38110104325,-39923.1617074103,-39939.8575563119,-39957.9856575976,-39977.0305504765,-39996.45858290334,-40015.732478129474,-40034.32582596538,-40051.73714031711,-40067.503134042185,-40081.21087769955,-40092.508531111125,-40101.114366585716,-40106.823840775156,-40109.51451865011,-40109.1487076958,-40105.77372220543,-40099.51976491827,-40090.595484033656,-40079.28133516753,-40065.92094712349,-40050.910754342796,-40034.68821459279,-40017.71897522326,-40000.48338305179,-39983.46275019958,-39967.12579033577,-39951.91562695506,-39938.237748439446,-39926.449245372285,-39916.849616066516,-39909.67336911436,-39905.08458973801,-39903.173572591026,-39903.955560037066,-39907.3715640756,-39913.29119383142,-39921.51736020031,-39931.79268567252,-39943.807410884605,-39957.208560050596,-39971.61010478859,-39986.60384954063,-40001.77075131309,-40016.69238145899,-40030.96223745003,-40044.19661801199,-40056.04478575085,-40066.198157745224,-40074.39828680361,-40080.443424418256,-40084.193490916834,-40085.57331868912,-40084.574080040744,-40081.25286122008,-40075.73039708811,-40068.187035043244,-40058.85705019193,-40048.0214842652,-40035.999726305214,-40023.140091724315,-40009.809686277586,-39996.383861502996,-39983.235577460466,-39970.72498687179,-39959.18954228447,-39948.93490545354,-39940.22690696832,-39933.28476584028,-39928.27573513605,-39925.31129270985,-39924.44494759977,-39925.67168448733,-39928.92902236618,-39934.09962052064,-39941.01532605231,-39949.46252317846,-39959.1886157284,-39969.90945080934,-39981.317473462965,-39993.09038914025,-40004.90010281424,-40016.421700372404,-40027.34223949369,-40037.36912349364,-40046.23784264669,-40053.718883319714,-40059.62362589096,-40063.80907780782,-40066.181318022835,-40066.69756300862,-40065.366801912794,-40062.24898826167,-40057.45281680677,-40051.1321553023,-40043.481240744884,-40034.72878641241,-40025.13117845329,-40014.964967498214,-40004.518880721145,-39994.08559216888,-39983.953493579655,-39974.398704239444,-39965.67754698079,-39958.0196988472,-39951.6222001504,-39946.64447580008,-39943.204489183496,-39941.37611288594,-39941.18776353049,-39942.62231124629,-39945.618238873474,-39950.07199292988,-39955.841438335854,-39962.75030246656,-39970.59347161607,-39979.14298462891,-39988.15455434254,-39997.374437607934,-40006.546468966815,-40015.41907151456,-40023.752061026506,-40031.323066026314,-40037.93339709257,-40043.413213250846,-40047.62585166917,-40050.47120882663,-40051.88808653151,-40051.85544412888,-40050.392528326796,-40047.557883485686,-40043.44727704153,-40038.19060595049,-40031.94787958186,-40024.90440129397,-40017.26529400746,-40009.24953357284,-40001.08366693785,-39992.99539955973,-39985.20723798121,-39977.9303689943,-39971.35894666472,-39965.66494316594,-39960.99369959176,-39957.46028952009,-39955.14678204789,-39954.10046326634,-39954.33304670393,-39955.82087501543,-39958.5060879892,-39962.29870646584,-39967.079558591686,-39972.703954410485,-39979.0059974357,-39985.80340776374,-39992.90272060166],"type":"scatter","name":"source2_q_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1,0.1001,0.1002,0.1003,0.1004,0.1005,0.1006,0.1007,0.1008,0.1009,0.101,0.1011,0.1012,0.1013,0.1014,0.1015,0.1016,0.1017,0.1018,0.1019,0.102,0.1021,0.1022,0.1023,0.1024,0.1025,0.1026,0.1027,0.1028,0.1029,0.103,0.1031,0.1032,0.1033,0.1034,0.1035,0.1036,0.1037,0.1038,0.1039,0.104,0.1041,0.1042,0.1043,0.1044,0.1045,0.1046,0.1047,0.1048,0.1049,0.105,0.1051,0.1052,0.1053,0.1054,0.1055,0.1056,0.1057,0.1058,0.1059,0.106,0.1061,0.1062,0.1063,0.1064,0.1065,0.1066,0.1067,0.1068,0.1069,0.107,0.1071,0.1072,0.1073,0.1074,0.1075,0.1076,0.1077,0.1078,0.1079,0.108,0.1081,0.1082,0.1083,0.1084,0.1085,0.1086,0.1087,0.1088,0.1089,0.109,0.1091,0.1092,0.1093,0.1094,0.1095,0.1096,0.1097,0.1098,0.1099,0.11,0.1101,0.1102,0.1103,0.1104,0.1105,0.1106,0.1107,0.1108,0.1109,0.111,0.1111,0.1112,0.1113,0.1114,0.1115,0.1116,0.1117,0.1118,0.1119,0.112,0.1121,0.1122,0.1123,0.1124,0.1125,0.1126,0.1127,0.1128,0.1129,0.113,0.1131,0.1132,0.1133,0.1134,0.1135,0.1136,0.1137,0.1138,0.1139,0.114,0.1141,0.1142,0.1143,0.1144,0.1145,0.1146,0.1147,0.1148,0.1149,0.115,0.1151,0.1152,0.1153,0.1154,0.1155,0.1156,0.1157,0.1158,0.1159,0.116,0.1161,0.1162,0.1163,0.1164,0.1165,0.1166,0.1167,0.1168,0.1169,0.117,0.1171,0.1172,0.1173,0.1174,0.1175,0.1176,0.1177,0.1178,0.1179,0.118,0.1181,0.1182,0.1183,0.1184,0.1185,0.1186,0.1187,0.1188,0.1189,0.119,0.1191,0.1192,0.1193,0.1194,0.1195,0.1196,0.1197,0.1198,0.1199,0.12,0.1201,0.1202,0.1203,0.1204,0.1205,0.1206,0.1207,0.1208,0.1209,0.121,0.1211,0.1212,0.1213,0.1214,0.1215,0.1216,0.1217,0.1218,0.1219,0.122,0.1221,0.1222,0.1223,0.1224,0.1225,0.1226,0.1227,0.1228,0.1229,0.123,0.1231,0.1232,0.1233,0.1234,0.1235,0.1236,0.1237,0.1238,0.1239,0.124,0.1241,0.1242,0.1243,0.1244,0.1245,0.1246,0.1247,0.1248,0.1249,0.125,0.1251,0.1252,0.1253,0.1254,0.1255,0.1256,0.1257,0.1258,0.1259,0.126,0.1261,0.1262,0.1263,0.1264,0.1265,0.1266,0.1267,0.1268,0.1269,0.127,0.1271,0.1272,0.1273,0.1274,0.1275,0.1276,0.1277,0.1278,0.1279,0.128,0.1281,0.1282,0.1283,0.1284,0.1285,0.1286,0.1287,0.1288,0.1289,0.129,0.1291,0.1292,0.1293,0.1294,0.1295,0.1296,0.1297,0.1298,0.1299,0.13,0.1301,0.1302,0.1303,0.1304,0.1305,0.1306,0.1307,0.1308,0.1309,0.131,0.1311,0.1312,0.1313,0.1314,0.1315,0.1316,0.1317,0.1318,0.1319,0.132,0.1321,0.1322,0.1323,0.1324,0.1325,0.1326,0.1327,0.1328,0.1329,0.133,0.1331,0.1332,0.1333,0.1334,0.1335,0.1336,0.1337,0.1338,0.1339,0.134,0.1341,0.1342,0.1343,0.1344,0.1345,0.1346,0.1347,0.1348,0.1349,0.135,0.1351,0.1352,0.1353,0.1354,0.1355,0.1356,0.1357,0.1358,0.1359,0.136,0.1361,0.1362,0.1363,0.1364,0.1365,0.1366,0.1367,0.1368,0.1369,0.137,0.1371,0.1372,0.1373,0.1374,0.1375,0.1376,0.1377,0.1378,0.1379,0.138,0.1381,0.1382,0.1383,0.1384,0.1385,0.1386,0.1387,0.1388,0.1389,0.139,0.1391,0.1392,0.1393,0.1394,0.1395,0.1396,0.1397,0.1398,0.1399,0.14,0.1401,0.1402,0.1403,0.1404,0.1405,0.1406,0.1407,0.1408,0.1409,0.141,0.1411,0.1412,0.1413,0.1414,0.1415,0.1416,0.1417,0.1418,0.1419,0.142,0.1421,0.1422,0.1423,0.1424,0.1425,0.1426,0.1427,0.1428,0.1429,0.143,0.1431,0.1432,0.1433,0.1434,0.1435,0.1436,0.1437,0.1438,0.1439,0.144,0.1441,0.1442,0.1443,0.1444,0.1445,0.1446,0.1447,0.1448,0.1449,0.145,0.1451,0.1452,0.1453,0.1454,0.1455,0.1456,0.1457,0.1458,0.1459,0.146,0.1461,0.1462,0.1463,0.1464,0.1465,0.1466,0.1467,0.1468,0.1469,0.147,0.1471,0.1472,0.1473,0.1474,0.1475,0.1476,0.1477,0.1478,0.1479,0.148,0.1481,0.1482,0.1483,0.1484,0.1485,0.1486,0.1487,0.1488,0.1489,0.149,0.1491,0.1492,0.1493,0.1494,0.1495,0.1496,0.1497,0.1498,0.1499,0.15,0.1501,0.1502,0.1503,0.1504,0.1505,0.1506,0.1507,0.1508,0.1509,0.151,0.1511,0.1512,0.1513,0.1514,0.1515,0.1516,0.1517,0.1518,0.1519,0.152,0.1521,0.1522,0.1523,0.1524,0.1525,0.1526,0.1527,0.1528,0.1529,0.153,0.1531,0.1532,0.1533,0.1534,0.1535,0.1536,0.1537,0.1538,0.1539,0.154,0.1541,0.1542,0.1543,0.1544,0.1545,0.1546,0.1547,0.1548,0.1549,0.155,0.1551,0.1552,0.1553,0.1554,0.1555,0.1556,0.1557,0.1558,0.1559,0.156,0.1561,0.1562,0.1563,0.1564,0.1565,0.1566,0.1567,0.1568,0.1569,0.157,0.1571,0.1572,0.1573,0.1574,0.1575,0.1576,0.1577,0.1578,0.1579,0.158,0.1581,0.1582,0.1583,0.1584,0.1585,0.1586,0.1587,0.1588,0.1589,0.159,0.1591,0.1592,0.1593,0.1594,0.1595,0.1596,0.1597,0.1598,0.1599,0.16]},{"mode":"lines","y":[0.0,0.0,0.026338613966138794,0.21405151549283044,0.5838020874535711,1.1491351934227116,1.8472607277600537,2.6543894152312033,3.550172803803888,4.518956358968883,5.543334605206068,6.573113563511909,7.582486910649405,8.52577335142319,9.398801705872888,10.17913218393484,10.872931117337734,11.473226760092654,11.9912515952953,12.438213457863109,12.831293357255902,13.192807982191187,13.538179695062885,13.8890969865528,14.254960811073024,14.651080630610302,15.080932058081471,15.551790696185124,16.06257864506548,16.61339377383315,17.20022799455188,17.817975027460015,18.460990700098176,19.121328702115097,19.79294079667391,20.46757451166576,21.140142534962546,21.804428730733942,22.45755351736609,23.09672252436686,23.72163434910029,24.33266937754244,24.93129774259234,25.51956761265123,26.09936456449672,26.67273089671381,27.240926103869107,27.8050569408449,28.36552114530381,28.922579278243617,29.47625056787618,30.026658033566626,30.574113255017348,31.119258593477,31.663057165333047,32.206866419615245,32.75220289789965,33.30086282536856,33.85449597019104,34.414798020443705,34.983030167714006,35.56022466005459,36.14681140066386,36.7427512105251,37.34739689146317,37.95950469849321,38.577385090848715,39.1988039084041,39.82138272212987,40.44244322299284,41.05952928505481,41.670254937483904,42.27277850274121,42.865695315052434,43.44831800170489,44.02062721167506,44.58328853657538,45.137651466868974,45.6855269389558,46.22921266365419,46.771128265902966,47.313844657605166,47.85970393917268,48.41083887735846,48.968890078561095,49.53501376152215,50.109762358376315,50.69308866703288,51.284397993596606,51.8825600103969,52.486090049837095,53.09317570957493,53.701915250644646,54.31035709725375,54.91671941503023,55.519432092858565,56.11728010535894,56.709433662454785,57.29548954606705,57.8754776168387,58.44980744015018,59.01924698540052,59.58480669008822,60.14769563487584,60.70918621530569,61.270560536751915,61.83299565647388,62.397515744312514,62.96492421708953,63.535774693160256,64.11035828209431,64.68869985026393,65.27059223898526,65.85561640111527,66.44320361956073,67.03267138129402,67.62329141341228,68.21432969095899,68.80510154251706,69.3950046275966,69.98355007755794,70.57038097257467,71.15527741177151,71.73815833600182,72.31906580181543,72.89815299054732,73.47565705639897,74.05187931007883,74.62715643748945,75.20183953633182,75.7762710861199,76.35076810932594,76.92560852309666,77.50102117654238,78.0771816444316,78.65420934720413,79.23217046400971,79.8110805493198,80.39091136193578,80.97159665135645,81.5530400641868,82.1351219607163,82.71770661065345,83.30064863319731,83.88379883190532,84.46700967482766,85.0501398962437,85.63305898381961,86.21565090005339,86.79781766516096,87.37948230288826,87.96059138646152,88.54111687256002,89.12105717094353,89.70043723477147,90.27930760668718,90.85774221793103,91.43583511530397,92.01369591278477,92.59144447376212,93.16920467584207,93.74709799543403,94.32523689310986,94.90371874018292,95.48262043213427,96.06199418610107,96.641864798966,97.22222845702746,97.80305340917096,98.38428216479953,98.9658354456615,99.54761723703746,100.1295209942453,100.71143623631608,101.29325538514989,101.87488018716485,102.45622742929828,103.03723355936692,103.61785788172921,104.19808428509492,104.77792125054589,105.3574004139186,105.93657359404179,106.51550876718052,107.09428508782226,107.6729874914358,108.25170112725912,108.83050606784174,109.4094726036519,109.9886573805838,110.56810064782482,111.14782464874835,111.7278333015778,112.30811300407183,112.88863455347888,113.46935589021133,114.05022551762592,114.63118627282276,115.21217921902138,115.79314738601671,116.37403912241331,116.95481089545125,117.5354293630526,118.11587268385124,118.69613099201848,119.2762061176155,119.85611058776905,120.43586606404448,121.01550133456261,121.59505003997013,122.17454829058914,122.7540323307469,123.33353639857586,123.91309088215694,124.49272087417039,125.07244515848672,125.65227566673477,126.23221737714427,126.81226863194816,127.39242180388645,127.97266424434103,128.5529794267118,129.13334819926445,129.71375006712037,130.29416442378766,130.87457167400763,131.45495419173673,132.03529708414231,132.61558873617443,133.19582113439768,133.77598997350572,134.35609456507032,134.93613757278433,135.51612460537189,136.09606370193774,136.67596474386784,137.25583882917897,137.83569763997647,138.41555283345966,138.9954154800204,139.5752955700096,140.15520160419968,140.73514027974952,141.31511627816184,141.89513215784237,142.47518834982,143.05528325123979,143.63541340815976,144.2155737757188,144.79575804144963,145.37595899502912,145.95616892637648,146.5363800329352,147.11658481686428,147.69677645345544,148.27694911360075,148.85709822539272,149.4372206631012,150.01731485535544,150.5973808087435,151.17742004727506,151.75743547287428,152.3374311560521,152.9174120698577,153.49738378301996,154.07735213037296,154.65732287960583,155.23730141321806,155.81729244339863,156.3972997750821,156.9773261294423,157.5573730359741,158.13744079724674,158.71752852573212,159.29763424800203,159.8777550675381,160.45788737425983,161.03802708643283,161.61816990919442,162.19831159359623,162.77844818061126,163.3585762162531,163.93869292620906,164.51879634153542,165.09888537024062,165.678959813262,166.25902032674853,166.8390683358115,167.41910590758292,167.99913559349125,168.57916025201644,169.15918286369723,169.73920634999897,170.3192334066532,170.89926636061145,171.47930705767448,172.0593567855934,172.6394162349229,173.21948549749072,173.79956410005988,174.37965106879727,174.9597450186178,175.53984426038105,176.1199469183665,176.70005105037333,177.2801547632212,177.86025631722228,178.44035421435962,179.02044726625522,179.6005346395186,180.1806158775752,180.76069089951775,181.34075997781514,181.92082369778475,182.50088290254564,183.08093862769033,183.66099203015145,184.24104431568952,184.82109666913746,185.40115019103484,185.98120584362647,186.5612644084326,187.14132645677975,187.72139233386147,188.3014621561219,188.88153582106202,189.46161302798274,190.0416933077295,190.6217760591887,191.20186059012335,191.7819461599022,192.3620320217759,192.94211746255894,193.52220183786878,194.10228460143998,194.68236532742785,195.26244372504743,195.84251964531472,196.42259308006228,197.0026641537689,197.58273310906176,198.16280028700328,198.74286610346752,199.32293102302336,199.9029955317855,200.48306011066632,201.0631252103671,201.64319122929052,202.2232584953516,202.8033272524262,203.38339765190392,203.96346974953406,204.54354350747283,205.12361880117857,205.70369543056785,206.28377313464483,206.8638516086675,207.4439305228243,208.0240095413503,208.60408834103634,209.1841666281607,209.7642441529935,210.34432072119628,210.9243962016334,211.50447053033,212.08454371053563,212.66461580907185,213.24468694933563,213.82475730151015,214.40482707065877,214.98489648347328,215.56496577448618,216.1450351725479,216.72510488832265,217.3051751034587,217.88524596196248,218.4653175641554,219.04538996342228,219.62546316579267,220.20553713222762,220.78561178333777,221.36568700612935,221.9457626622824,222.52583859740136,223.10591465065303,223.68599066421697,224.26606649201793,224.84614200727933,225.42621710853436,226.0062917238414,226.5863658130681,227.1664393682301,227.74651241198313,228.3265849944691,228.90665718879745,229.48672908550537,230.06680078637785,230.64687239802194,231.226944025572,231.8070157668836,232.3870877075048,232.96715991666824,233.52092749267717,233.9139660361771,234.12642268101158,234.14524783754047,234.03269264571338,233.81236046815596,233.50495002895173,233.12652869485592,232.69438902800633,232.25664950878124,231.83745013479427,231.48009582795876,231.18867578972743,230.98482943562027,230.86307843859151,230.8304938141509,230.87737022327514,230.99449046028764,231.16705312330936,231.37489296458133,231.60360392484984,231.83175128082624,232.04904597396094,232.23872892834925,232.39562115740156,232.51085974125942,232.58457478961927,232.61633965078516,232.61048933693883,232.57281617435274,232.50968766193765,232.42956209363368,232.3386373331663,232.2448518550438,232.15279796917946,232.06797277466532,231.9928074000532,231.92980794284168,231.8794683609977,231.84190036573688,231.8164058295927,231.8018240150009,231.7970410129603,231.80061701620292,231.81153137589374,231.82863504196882,231.85117695212378,231.87838289212604,231.90963735359344,231.94425897282335,231.9814790430674,232.020338290711,232.05969144690098,232.09810911478502,232.13407412925943,232.16582294805528,232.1917560558175,232.21022678894826,232.2200432490691,232.22026047211685,232.21058220864083,232.1912222800578,232.16304583020326,232.12754018381978,232.08663740180526,232.04279252749643,231.9985500940878,231.95669130869902,231.91969485952413,231.8898983027101,231.86902812995595,231.85832761251825,231.8582927312581,231.8687421615664,231.88881662237412,231.91699343876155,231.95131900319106,231.98938846676324,232.02871398132834,232.06669410309846,232.10099261973562,232.12951425927602,232.15068323511377,232.16342884723912,232.16729865119873,232.16244714422234,232.14957670199433,232.12991936128088,232.10504957837904,232.07685336867692,232.0472864117328,232.01833316134318,231.9917868347661,231.969208809477,231.95178747123165,231.94031137914754,231.9351316742533,231.93615965805176,231.94292418523233,231.9545967387096,231.97011066027005,231.98820769737128,232.0075753441832,232.02690177908264,232.04499121263368,232.0608116814416,232.07356225624773,232.08270094087786,232.0879558733547,232.08932727600703,232.08705139624462,232.0815784847556,232.07350890423658,232.06355571347783,232.0524754395442,232.04102707546087,232.02991630521785,232.01976202910356,232.01106492844224,232.00418882789768,231.99935556285038,231.99664324463674,231.99600215494473,231.99726712122597,232.00018517434017,232.00443612495582,232.00966223902822,232.01549028818758,232.02155572813876,232.02752071865729,232.03308887877182,232.03801620512004,232.04211603619584,232.04526241716556,232.04738742173944,232.04847847345016,232.0485712268156,232.04774324590386,232.04610539711177,232.04379417263115,232.04096354196557,232.03777749771012,232.03440319498603,232.03100449159618,232.02773645071238,232.02474011517637,232.0221382294146,232.02003134476374,232.01849480836768,232.01757642454004,232.01729507109738,232.01764030082003,232.01857308644344,232.02002771422718,232.02191496239985,232.0241263172836,232.02653937477183,232.02902386182737,232.03144839578547,232.03368721328792,232.0356268865229,232.03717228691,232.03825167335373,232.0388204343804,232.0388632456264,232.03839459417225,232.03745740129702,232.0361201292226,232.03447218222456,232.03261829640226,232.0306718893897,232.02874816168324,232.02695710071023,232.02539705337728,232.02414914584023,232.02327292384626,232.0228035184938,232.02275035115642,232.02309759820704,232.02380610857628,232.0248168330445,232.0260552576174,232.02743672006287,232.02887205818698,232.03027333356312,232.03155918165976,232.03265948432744,232.0335191161646,232.03410051096023,232.03438503228625,232.03437301975183,232.03408269556047,232.033547958006,232.0328153698042,232.03194050244218,232.03098397227913,232.0300074049117,427.5245380799796,216.31277392518592,427.5245380799797,389.2492096121112,427.52453807997966,368.0328210692149,427.5245380799796,410.94037395439693,427.52453807997966,357.54790116802144,350.94417853589954,304.71485336290516,240.25931816328642,184.61984232809607,150.12665132632378,130.80503947633255,115.47982244496767,103.38811477175241,97.25837872837755,101.84954320304661,116.11805465593655,138.48064117491268,163.7696316644227,190.16358772792793,216.66081394925524,242.9746668587347,267.7704502741587,288.4629699537807,304.1379448653182,313.6613552595551,318.42683454295644,318.4406428199549,315.1149658969386,308.4244667524315,299.42510526966254,288.57710927753044,276.9589584052681,265.7964309241549,255.81844442460346,248.0255722549595,242.7796145980815,240.8450500866302,242.02400080385146,246.01553568592772,251.6668229631415,257.8252856498048,263.2072693984292,266.8214256133434,267.89063896769596,265.97517480639453,260.9422755086358,253.05582538908212,242.81122287344942,230.91935894795398,218.2447007833213,205.8062807697924,194.60248395418705,185.60213241686833,179.63091195107833,177.2738024028224,178.8296802635644,184.21647096300305,193.0593663068386,204.68936228481593,218.2860397821382,232.87201959950323,247.45921053004025,261.0850780285897,272.90776883587506,282.25115251811513,288.65159524104166,291.8904759649314,291.90810795175344,288.9570150512162,283.5875451788306,276.42476750999106,268.14348620701514,259.517566899612,251.34370927686524,244.362694985985,239.1309670404393,235.94012604058926,234.7854113139114,235.35482601583539,237.11528217919016,239.38675175652264,241.48327961079164,242.77578586073693,242.8064481026951,241.30016347935887,238.22107680101607,233.7468013271657,228.19342196922673,222.01522643577363,215.80941394644412,210.23775506784494,205.92340825785854,203.39618051101263,203.01491343322738,204.9434620414698,209.13531437802914,215.35247586797257,223.17133786006508,232.0447389660646,241.36130695571723,250.49133182089722,258.83516925258505,265.86160856891905,271.1572930998985,274.45248372771715,275.59501181546176,274.636609894636,271.87098826212184,267.6564379100511,262.4116332961695,256.625980595441,250.8229333002019,245.4921580495492,241.01643324295716,237.6282661479008,235.38726839292607,234.1906251844393,233.80460940434847,233.9060709753846,234.1324277797658,234.14235390058954,233.6657327835382,232.53467606627166,230.69794722421196,228.2453496232425,225.38204881800976,222.34993403240463,219.45134234696548,217.04684408991727,215.48372486506062,215.03812573684718,215.88264095674364,218.07559754541248,221.56234545804918,226.1719992501021,231.63003192454786,237.59457004085277,243.6892891956875,249.5319543113711,254.76144031367608,259.0658041910976,262.20796207605423,264.0448370965643,264.53426353435617,263.70004754442186,261.6960685003901,258.8107926291827,255.32638227520928,251.5269402261233,247.7046440202936,244.13202791576944,241.02055230701376,238.48720924989988,236.54822123632528,235.139025398774,234.13451414097236,233.3684937339809,232.66170280471485,231.85308824009735,230.82599430099162,229.52435742664656,227.95891735029318,226.20543446182506,224.4122610218463,222.7740179525128,221.46607698899297,220.65667660365068,220.52053430358757,221.19134999247228,222.72901519087182,225.11839647362785,228.27256715830754,232.03335759570763,236.19281968331975,240.51456095949527,244.75095226121525,248.6598596244827,252.02243345472453,254.66052469542936,256.45112932531663,257.3352334759569,257.3201756228911,256.4758010266964,254.91715202016528,252.7915146490434,250.29909970592868,247.65809027275924,245.04066273471798,242.58174195859215,240.381760824838,238.49416943000014,236.92129191909294,235.6222980905279,234.52220842875235,233.52983824050497,232.5568234828795,231.5328240461357,230.4167618010059,229.2043797682844,227.931069277561,226.6688919263308,225.5182380903144,224.5958771087288,224.02423224720914,223.92452385624688,224.37534536531058,225.3999669144058,227.00433014659686,229.1612523405579,231.7922893783929,234.76657268880976,237.92351878844275,241.092870409754,244.10699076158315,246.80935595475694,249.06393896109662,250.76569337520775,251.8493147214674,252.2934425962652,252.11995903106876,251.38948402732805,250.1940760234934,248.64763354202023,246.874405929969,244.99675698749334,243.1236664183578,241.3415040008484,239.70805444583726,238.2503741120471,236.96661458978238,235.83153287467428,234.8048396091765,233.84104634925407,232.89927950116933,231.95169422701304,230.9895087742684,230.0261022059508,229.096961194254,228.25658346073752,227.57273286603066,227.1187444873156,226.96479326954636,227.16916208548633,227.77052925335093,228.78215224499746,230.18859044237288,231.9452804425295,233.98095377814678,236.20255956566592,238.50213800990534,240.76493243691908,242.87798254382858,244.73843733692726,246.26088720770068,247.38311289733872,248.06978893479112,248.31386453150964,248.13553888686764,247.57897017981327,246.7070499294812,245.59477018142968,244.32184870138082,242.96538137425335,241.59330478161846,240.2593832912297,239.0002610968191,237.83485676363154,236.76606701997184,235.78442568438936,234.8731129390903,234.01354825026067,233.19077576956093,232.39792730433882,231.639221260244,230.93116846990833,230.30188946373102,229.78866838958638,229.4340625031592,229.2810422060632,229.3677359683687,229.7223963602876,230.35916831864367,231.275143392928,232.44902404641851,233.84153503375,235.39752339470908,237.04951778559695,238.72238517180725,240.33863758350446,241.82390418628978,243.11208769108484,244.14976847045025,244.89949477527335,245.3417027634298,245.47513293218682,245.31574556225627,244.89427287162675,244.25267203318293,243.43984834488282,242.50709133268535,241.503697895136,240.4732362673592,239.45083035862285,238.46171818882115,237.52117637049895,236.63572549829127,235.80536797070565,235.02648477242502,234.29495036574406,233.6090210855609,232.97160750767907,232.3916414968417,231.88437612194932,231.4705941228719,231.17483223228948,231.02284236359046,231.0385958495082,231.24118627057322,231.6419947424372,232.24244837723998,233.03263220079276,233.99091673105949,235.08465103629837,236.27185887929255,237.50377655631019,238.72799444610422,239.89191532093022,240.94622249176393,241.8480589606795,242.56365225550763,243.07017484000315,243.35670206735313,243.42421274721463,243.28466484145235,242.95926337306315,242.4761119436773,241.86749607969838,241.1670794533161,240.40729780012055,239.61720758695682,238.82098837588828,238.03721505468363,237.2789187411378,236.55435643441066,235.8683234365963,235.2237813119573,234.62354544764187,234.07178233125964,233.5751042431535,233.14311093139435,232.78830448652644,232.52538506777063,232.37001197996582,232.3371788483303,232.43939696066133,232.68490297345687,233.07610427716028,233.6084483219372,234.26985491668844,235.0407891219789,235.89498457398736,236.80076083698916,237.7228207163811,238.62436962187346,239.46937225245685,240.22475331339623,240.86235827904522,241.3605156955992,241.7050814391348,241.88989412658646,241.91662525381986,241.7940628973387,241.53691904386707,241.16429301117753,240.6979526640202,240.16060781177032,239.57434432485232,238.9593630938208,238.33312724731945,237.70996864774867,237.10114711876307,236.51530065221968,235.9591793091312,235.4385255010263,234.95895221880326,234.52667910493352,234.14901189191278,233.83448942670597,233.59266880390658,233.43356716816868,233.36682292553107,233.40067462241063,233.5408789737927,233.7896982659973,234.14508108197612,234.60014015736957,235.1429998856854,235.75704745323551,236.42158033694022,237.1128034909607,237.8050959773004,238.4724421050231,239.08990823534134,239.6350440334357,240.08909579663992,240.43793834719636,240.67265888069574,240.78975855274277,240.79097249548943,240.6827431907649,240.4754124973773,240.18222119672052,239.8182192668671,239.39919361415517,238.94071215063894,238.45736464017466,237.96225369611756,237.4667569170386,236.9805474051177,236.51182907598957,236.0677190903306,235.65469528243244,235.27902310418108,234.94708428082924,234.6655466025798,234.44133846639997,234.2814196772541,234.19236813638878,234.17982709056932,234.24787681430254,234.3984059468569,234.6305601555955,234.94033932336063,235.32040003722906,235.76009962150434,236.24579374734466,236.76137448817852,237.28901224632284,237.81004557763436,238.30594935186798,238.75930499124007,239.15469710224818,239.47946835362723,239.72427805482297,239.88342815766586,239.954941569722,239.94039972883598,239.8445672646568,239.67484927563126,239.44063955878215,239.15262476937426,238.82210928705365,238.46041857442063,238.07842585263447,237.6862295268974,237.29298907654393,236.90690749011384,236.53533116837744,236.18492558068894,235.861878253716,235.57208049945478,235.32124539944465,235.11493092902847,234.9584521120177,234.85668277262354,234.81376373290183,234.83274827113294,234.9152257404796,235.06096938228796,235.26765408573277,235.53068430015458,235.8431622346509,236.19601309119855,236.57826889714096,236.97749718116157,237.38034686224208,237.7731726546884,238.14269204577974,238.47662606593957,238.76427678874586,238.99700048451209,239.1685449403605,239.27523167615973,239.31597743430876,239.29216309103398,239.20737071332795,239.0670196692816,238.8779395205058,238.6479202473431,238.3852789575955,238.09847684030052,237.79581142884007,237.4851982974661,237.17404446071706,236.869204404381,236.57700019310596,236.303280556964,236.05349094317728,235.8327274405758,235.6457519703175,235.4969534879049,235.3902491421661,235.32892919302208,235.31545878099922,235.3512572622906,235.43648090300564,235.56983671575543,235.7484539399701,235.9678353063759,236.22190331378548,236.50314807314217,236.80287379323406,237.11153169965732,237.41911904016834,237.7156176290256,237.9914416897064,238.2378638722018,238.44739026982919,238.61405979384864,238.73364989896788,238.80377872074934,238.82390239724282,238.79521486344188,238.72046491856892,238.60371116822571,238.45003900712425,238.26526483684336,238.05565117862426,237.8276524928374,237.58770585822293,237.34207389887953,237.09674029612157,236.85735172370335,236.62919484926553,236.41719372293934,236.2259117488782,236.05954354948253,235.92188515192979,235.81627558357658,235.74550851996534,235.71171836538142,235.7162503444249,235.75952821442343,235.84093559878326,235.95872741933746,236.10998642527997,236.2906365523828,236.49552016697209,236.71854066742534,236.95286602754376,237.19118328261973,237.4259892428644,237.64989932798974,237.85595466357012,238.0379076068863,238.1904676435986,238.30949291657487,238.39211717185353,238.43680718468747,238.44335124871458,238.41278454776972,238.34726169375094,238.24989000073916,238.1245388960195,237.97564111792016,237.80800006224723,237.62661502186742,237.43653246769694,237.24272739548547,237.0500146107543,236.8629861314156,236.68596807062397,236.5229887178528,236.3777492076986,236.25358913125396,236.15344152466338,236.07977455466448,236.03452051529894,236.0189960204653,236.03382010626277,236.07883899131113,236.15306723049954,236.25465481285883,236.3808884087746,236.52823260731634,236.6924138529311,236.86854622282107,237.05129455309043,237.2350670939336,237.4142271835636,237.58331163763336,237.73724282113477,237.87152176077416,237.98239111533923,238.0669591958436,238.12327927839652,238.1503818836707,238.14826118060634,238.11781988413335,238.06077966683594,237.97956596498076,237.87717698790328,237.75704669362844,237.6229105332986,237.47868104893004,237.3283381670697,237.17583654888958,237.02502993531036,236.87961034488487,236.7430584714017,236.618600838504,236.50916925725613,236.4173588592009,236.3453823122966,236.2950195699842,236.26756441016522,236.26377083329436,236.2838038666586,236.3272002696099,236.39284492245102,236.4789682589899,236.58316900087675,236.702464778331,236.83337114666128,236.97200724296258,237.1142241025233,237.2557496931322,237.39234322137185,237.51995036336595,237.63485085685235,237.7337903772619,237.81408975515015,237.87372626241492,237.91138373820564,237.9264705486335,237.91910657136637,237.89008236926122,237.84079529624637,237.77316833798497,237.6895579599794,237.59265711000828,237.48539885556247,237.37086504072786,237.25220297281686,237.1325516725917,237.0149778215865,236.90242037641835,236.79764201740886,236.70318523193686,236.6213309193429,236.5540579044488,236.50300257097652,236.4694188466963,236.4541398399776,236.45754339076618,236.47952452121422,236.51947814645501,236.57629537087388,236.6483762347859,236.7336609249438,236.8296802982976,236.933625205074,237.0424326693814,237.15288563480198,237.26172184226238,237.3657465895364,237.46194370499103,237.54757909180987,237.62029165920973,237.67816730899258,237.71979280789913,237.74428774004375,237.75131417444953,237.74106507199994,237.71423367472073,237.67196706974877,237.61580773229807,237.54762709453752,237.4695550674155,237.3839090038859,237.29312490970653,237.1996928787889,237.1060978612529,237.0147660688988,236.92801667560775,236.8480180457614,236.77674755742467,236.71595417978781,236.6671232840617,236.63144365347387,236.60977723108962,236.60263271420482,236.61014458279593,236.63205946072543,236.667731796603,236.71613068777432,236.7758592569721,236.84518735698532,236.9220975799155,237.00434365879917,237.0895194556969,237.17513591842848,237.25870273687605,237.33781100253273,237.4102130132334,237.4738954847666,237.5271428216317,237.5685877248275,237.59724721839325,237.61254308659605,237.61430665053794,237.60276869715446,237.57853613371657,237.54255752050688,237.49607999685983,237.44060024785756,237.3778120709534,237.30955282565907,237.23775063452388,237.1643737106718,237.09138268081705,237.02068631420485,236.954100708203,236.89331175524322,236.839840639155,236.79501217684432,236.75992600991722,236.7354309209927,236.72210285166418,236.720227480201,236.72978842672325,236.7504622502264,236.78162135770432,236.8223457498062,236.87144418711094,236.92748490045494,236.98883542642776,237.0537105737509,237.1202269711857,237.18646216538156,237.25051587381404,237.31057078886315,237.3649502949442,237.41217060593527,237.4509851430292,237.48041942661746,237.49979531083858,237.50874399885197,237.50720789027562,237.49543188071195,237.47394521435996,237.44353535216385,237.4052155405558,237.36018784496204,237.30980335741071,237.25552112099595,237.1988670674586,237.14139397514714,237.08464316182287,237.03010836549126,236.9792020648549,236.93322436713723,236.89333455108294,236.86052539108226,236.83560048800743,236.81915496848544,236.8115600571395,236.81295214480335,236.82322704176656,236.84204009691695,236.86881276829874,236.9027460457401,236.9428408602535,236.98792528624628,237.0366879774986,237.08771690781754,237.1395421454136,237.1906811079526,237.2396845494692,237.2851814399577,237.32592092307493,237.3608096763544,237.38894324088545,237.4096302139138,237.42240858177632,237.4270538812868,237.4235792833022,237.4122280627598,237.39345922970264,237.36792732738283,237.33645754646486,237.30001735770603,237.2596858367847,237.2166217590729,237.17203139918314,237.12713680290508,237.083145130522,237.04121952071495,237.00245180874253,236.96783736040277,236.9382522563793,236.91443307492608,236.89695956337442,236.886240545231,236.8825034619785,236.88578797954258,236.8959440835104,236.91263503375112,236.9353454429234,236.96339458580044,236.99595484514427,237.0320749683971,237.07070756543644,237.1107400411728,237.15102794833473,237.1904295841093,237.2278405544573,237.26222700184147,237.29265623956474,237.31832365630578,237.3385749389705,237.35292289662297,237.36105843524197,237.3628555124032,237.3583701726934,237.3478340102956,237.33164260972092,237.31033966862367,237.28459760291364,237.25519547432327,237.22299506970523,237.188915909241,237.15390987995434,237.11893609518165,237.08493648335366,237.052812522066,237.02340346434204,236.99746635768415,236.97565813317917,236.9585200378413,236.94646469146002,236.93976606021414,236.9385526430386,236.94280415354467,236.9523519423599,236.96688333714374,236.98594997894526,237.00898010662962,237.03529459215005,237.06412636803657,237.09464272634432,237.12596981823444,237.15721855786634,237.18751104434546,237.21600656944216,237.24192628156055,237.26457562891892,237.2833638040376,237.29781955061065,237.30760286282106,237.312512294227,237.3124877856104,237.3076091059311,237.2980901664892,237.28426960652297,237.26659815274286,237.24562332306655,237.22197207679184,237.19633201332147,237.16943169548216,237.14202062909197,237.11484937592203,237.08865022035172,237.06411875732002,237.04189672513488,237.02255637348026,237.00658663424196,236.99438134799777,236.98622978791786,236.98230970997074,236.98268313816615,236.9872950609977,236.99597516657647,237.0084426773486,237.02431426120114,237.0431148969616,237.06429146371627,237.08722871165673,237.11126716504143,237.13572241336632,237.15990517256975,237.18314145037328,237.204792133186,237.22427132851075,237.24106284617062,237.25473428101662,237.264948264028,237.27147057100908,237.2741749104527,237.27304434621612,237.26816943839404,237.25974330014844,237.24805386381047,237.23347372276436,237.216447965062,237.19748044099543,237.17711891242817,237.15593952032174,237.134530983214,237.1134789082514,237.09335056223458,237.07468041666272,237.0579567503885,237.04360956727356,237.03200006379407,237.02341186130863,237.0180441971773,237.0160072451417,237.01731970543705,237.02190876655678,237.029612491941,237.040184625813,237.05330174406973,237.0685726009826,237.08554944418958,237.10374099359984,237.12262670944295,237.1416719157914,237.16034330300167,237.17812430925517,237.19452988001794,237.20912012553322,237.22151243958126,237.23139170525423,237.23851829169263,237.24273363483263,237.24396328992546,237.24221743842085,237.23758892161504,237.23024895386817,237.2204407357821,237.20847124042155,237.19470148261047,237.17953560301873,237.1634091066671,237.14677659190238,237.13009929361255,237.11383274626542,237.0984148508693,237.08425460722196,237.0717217501913,237.0611375067589,237.05276666896103,237.04681115574184,237.0434052127892,237.04261237213032,237.0444242612735,237.04876131397228,237.05547539093183,237.0643542693536,237.0751279063716,237.08747632519666,237.1010389168098,237.1154248974608,237.130224616234,237.1450213707467,237.15940336520916,237.17297543567202,237.18537017333017,237.19625809835455,237.20535657290208,237.21243719080042,237.2173314402059,237.21993450099836,237.2202071072229,237.21817547291667,237.21392934384568,237.20761829521888,237.19944644418837,237.18966578460729,237.17856837954525,237.16647766476652,237.15373912462434,237.14071060199598,237.12775249758607,237.11521810282738,237.10344429617558,237.0927428160382,237.08339230564098,237.07563130615313,237.06965235425506,237.0655973186025,237.06355408574086,237.06355467922728,237.0655748655814,237.06953526694366,237.0753039631507],"type":"scatter","name":"source2_v_mag_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1,0.1001,0.1002,0.1003,0.1004,0.1005,0.1006,0.1007,0.1008,0.1009,0.101,0.1011,0.1012,0.1013,0.1014,0.1015,0.1016,0.1017,0.1018,0.1019,0.102,0.1021,0.1022,0.1023,0.1024,0.1025,0.1026,0.1027,0.1028,0.1029,0.103,0.1031,0.1032,0.1033,0.1034,0.1035,0.1036,0.1037,0.1038,0.1039,0.104,0.1041,0.1042,0.1043,0.1044,0.1045,0.1046,0.1047,0.1048,0.1049,0.105,0.1051,0.1052,0.1053,0.1054,0.1055,0.1056,0.1057,0.1058,0.1059,0.106,0.1061,0.1062,0.1063,0.1064,0.1065,0.1066,0.1067,0.1068,0.1069,0.107,0.1071,0.1072,0.1073,0.1074,0.1075,0.1076,0.1077,0.1078,0.1079,0.108,0.1081,0.1082,0.1083,0.1084,0.1085,0.1086,0.1087,0.1088,0.1089,0.109,0.1091,0.1092,0.1093,0.1094,0.1095,0.1096,0.1097,0.1098,0.1099,0.11,0.1101,0.1102,0.1103,0.1104,0.1105,0.1106,0.1107,0.1108,0.1109,0.111,0.1111,0.1112,0.1113,0.1114,0.1115,0.1116,0.1117,0.1118,0.1119,0.112,0.1121,0.1122,0.1123,0.1124,0.1125,0.1126,0.1127,0.1128,0.1129,0.113,0.1131,0.1132,0.1133,0.1134,0.1135,0.1136,0.1137,0.1138,0.1139,0.114,0.1141,0.1142,0.1143,0.1144,0.1145,0.1146,0.1147,0.1148,0.1149,0.115,0.1151,0.1152,0.1153,0.1154,0.1155,0.1156,0.1157,0.1158,0.1159,0.116,0.1161,0.1162,0.1163,0.1164,0.1165,0.1166,0.1167,0.1168,0.1169,0.117,0.1171,0.1172,0.1173,0.1174,0.1175,0.1176,0.1177,0.1178,0.1179,0.118,0.1181,0.1182,0.1183,0.1184,0.1185,0.1186,0.1187,0.1188,0.1189,0.119,0.1191,0.1192,0.1193,0.1194,0.1195,0.1196,0.1197,0.1198,0.1199,0.12,0.1201,0.1202,0.1203,0.1204,0.1205,0.1206,0.1207,0.1208,0.1209,0.121,0.1211,0.1212,0.1213,0.1214,0.1215,0.1216,0.1217,0.1218,0.1219,0.122,0.1221,0.1222,0.1223,0.1224,0.1225,0.1226,0.1227,0.1228,0.1229,0.123,0.1231,0.1232,0.1233,0.1234,0.1235,0.1236,0.1237,0.1238,0.1239,0.124,0.1241,0.1242,0.1243,0.1244,0.1245,0.1246,0.1247,0.1248,0.1249,0.125,0.1251,0.1252,0.1253,0.1254,0.1255,0.1256,0.1257,0.1258,0.1259,0.126,0.1261,0.1262,0.1263,0.1264,0.1265,0.1266,0.1267,0.1268,0.1269,0.127,0.1271,0.1272,0.1273,0.1274,0.1275,0.1276,0.1277,0.1278,0.1279,0.128,0.1281,0.1282,0.1283,0.1284,0.1285,0.1286,0.1287,0.1288,0.1289,0.129,0.1291,0.1292,0.1293,0.1294,0.1295,0.1296,0.1297,0.1298,0.1299,0.13,0.1301,0.1302,0.1303,0.1304,0.1305,0.1306,0.1307,0.1308,0.1309,0.131,0.1311,0.1312,0.1313,0.1314,0.1315,0.1316,0.1317,0.1318,0.1319,0.132,0.1321,0.1322,0.1323,0.1324,0.1325,0.1326,0.1327,0.1328,0.1329,0.133,0.1331,0.1332,0.1333,0.1334,0.1335,0.1336,0.1337,0.1338,0.1339,0.134,0.1341,0.1342,0.1343,0.1344,0.1345,0.1346,0.1347,0.1348,0.1349,0.135,0.1351,0.1352,0.1353,0.1354,0.1355,0.1356,0.1357,0.1358,0.1359,0.136,0.1361,0.1362,0.1363,0.1364,0.1365,0.1366,0.1367,0.1368,0.1369,0.137,0.1371,0.1372,0.1373,0.1374,0.1375,0.1376,0.1377,0.1378,0.1379,0.138,0.1381,0.1382,0.1383,0.1384,0.1385,0.1386,0.1387,0.1388,0.1389,0.139,0.1391,0.1392,0.1393,0.1394,0.1395,0.1396,0.1397,0.1398,0.1399,0.14,0.1401,0.1402,0.1403,0.1404,0.1405,0.1406,0.1407,0.1408,0.1409,0.141,0.1411,0.1412,0.1413,0.1414,0.1415,0.1416,0.1417,0.1418,0.1419,0.142,0.1421,0.1422,0.1423,0.1424,0.1425,0.1426,0.1427,0.1428,0.1429,0.143,0.1431,0.1432,0.1433,0.1434,0.1435,0.1436,0.1437,0.1438,0.1439,0.144,0.1441,0.1442,0.1443,0.1444,0.1445,0.1446,0.1447,0.1448,0.1449,0.145,0.1451,0.1452,0.1453,0.1454,0.1455,0.1456,0.1457,0.1458,0.1459,0.146,0.1461,0.1462,0.1463,0.1464,0.1465,0.1466,0.1467,0.1468,0.1469,0.147,0.1471,0.1472,0.1473,0.1474,0.1475,0.1476,0.1477,0.1478,0.1479,0.148,0.1481,0.1482,0.1483,0.1484,0.1485,0.1486,0.1487,0.1488,0.1489,0.149,0.1491,0.1492,0.1493,0.1494,0.1495,0.1496,0.1497,0.1498,0.1499,0.15,0.1501,0.1502,0.1503,0.1504,0.1505,0.1506,0.1507,0.1508,0.1509,0.151,0.1511,0.1512,0.1513,0.1514,0.1515,0.1516,0.1517,0.1518,0.1519,0.152,0.1521,0.1522,0.1523,0.1524,0.1525,0.1526,0.1527,0.1528,0.1529,0.153,0.1531,0.1532,0.1533,0.1534,0.1535,0.1536,0.1537,0.1538,0.1539,0.154,0.1541,0.1542,0.1543,0.1544,0.1545,0.1546,0.1547,0.1548,0.1549,0.155,0.1551,0.1552,0.1553,0.1554,0.1555,0.1556,0.1557,0.1558,0.1559,0.156,0.1561,0.1562,0.1563,0.1564,0.1565,0.1566,0.1567,0.1568,0.1569,0.157,0.1571,0.1572,0.1573,0.1574,0.1575,0.1576,0.1577,0.1578,0.1579,0.158,0.1581,0.1582,0.1583,0.1584,0.1585,0.1586,0.1587,0.1588,0.1589,0.159,0.1591,0.1592,0.1593,0.1594,0.1595,0.1596,0.1597,0.1598,0.1599,0.16]},{"mode":"lines","y":[50.0,50.0,46.15163652863997,45.29976735927375,44.69744842368503,43.616084641166346,42.739613406452115,41.84270894723497,41.10836848664022,40.62317217595678,40.316659487893354,40.22001182221333,40.282017558411674,40.542137197688426,40.94471097214307,41.54505026484136,42.30146214235069,43.246598186478096,44.337148214203,45.56470293856971,46.88710840734127,48.256472733961736,49.638403156375176,50.969257108472476,52.20595711382762,53.25369542641133,54.084986416039065,54.63608057950765,54.910853974411005,54.89826907693789,54.63373659456535,54.155691455603126,53.517707228806955,52.78596353030546,52.013610886713586,51.267790704432294,50.58667145945171,50.01907025969619,49.58239500578335,49.29905704958775,49.166758104788364,49.18214216270821,49.32786792254286,49.58107682089055,49.91582814211102,50.29918338128166,50.70375150651898,51.096730916369026,51.45589947235953,51.75705784489622,51.98768405292342,52.13680317736648,52.20297888348172,52.18861765079484,52.101722198806314,51.95469944900829,51.76132127686008,51.53876731468616,51.302169296721374,51.0681265141276,50.84917068837505,50.65721800133098,50.4995997683092,50.38151950452578,50.304330064633085,50.26670893571131,50.26497899400963,50.293186630090524,50.34470599983126,50.41163438188345,50.48677329853189,50.56277949245594,50.63380630520338,50.69474396885584,50.7421582683345,50.77379588797565,50.78879800237083,50.78749606562279,50.77109891218664,50.74170380009403,50.70173488345812,50.65405954185647,50.60142628736605,50.54659528440931,50.49193475485224,50.43951773326369,50.390932932011545,50.34733873446344,50.30946018747573,50.27761177351071,50.25180543660507,50.231757062457106,50.21702446138239,50.20700759269016,50.201059343442644,50.19848414369235,50.19859959725877,50.200735367251134,50.204253065712294,50.208546750065175,50.21304277684592,50.21720337280349,50.22052722089987,50.22255702366959,50.22289111542536,50.22119420171138,50.21721774662808,50.21081068247894,50.20193782325535,50.19068564893171,50.177266595964106,50.16201536908864,50.14537218974102,50.12786780862288,50.11008744391798,50.092646749324146,50.07614605138813,50.06114319629418,50.04811156947119,50.03741745949525,50.02929412090971,50.023830425742034,50.02096792781186,50.02050447742411,50.02211410760838,50.02536499631183,50.029755456776606,50.03474146763126,50.03977798793135,50.044348507360986,50.04800025769326,50.050367731272175,50.05119300865192,50.05033711980682,50.047781601459945,50.043625290046236,50.03806873376428,50.0313982685841,50.02395937100067,50.016133304906,50.00830773577954,50.00085284456373,49.99409717046093,49.98830955148587,49.98368567666295,49.98033992461119,49.97830449489906,49.97753205504295,49.977905755245,49.97925058292165,49.98135001199215,49.98396193833994,49.98683669991721,49.98973285550978,49.99243184894198,49.994749655089784,49.99654503754005,49.997724793033576,49.99824471818828,49.998108212304146,49.99736106333975,49.996084912053085,49.994388299636206,49.99239753319833,49.99024688765147,49.98806959410585,49.985989713822704,49.984115410121554,49.982534076757965,49.98130904572219,49.98047841657882,49.98005525797656,49.98002958813221,49.9803712585577,49.981033908442114,49.9819592572194,49.98308167431628,49.98433257939499,49.98564446997697,49.98695443267965,49.988206901298206,49.98935575000835,49.99036553804254,49.99121211791014,49.991882519461086,49.992374344400154,49.992694678356216,49.992858710130704,49.99288812988112,49.992809424039166,49.992652163996745,49.99244734020506,49.99222583144215,49.9920170174772,49.99184759967052,49.99174061889142,49.991714706302915,49.99178355387514,49.99195561604502,49.99223403173243,49.99261676101301,49.99309692402742,49.993663324257916,49.99430113501211,49.994992721197654,49.99571856202152,49.99645823785943,49.99719143543346,49.99789892866559,49.99856348600372,49.99917066194391,49.999709432129094,50.00017263895351,50.000557227051175,50.00086425363733,50.0010986798111,50.001268951255824,50.00138640066913,50.00146450326342,50.00151803603701,50.00156218792663,50.00161167697461,50.001679925448954,50.001778340270995,50.00191574026072,50.002097957343445,50.00232763320798,50.00260421274533,50.00292413090293,50.00328116989659,50.003666960705374,50.00407158901376,50.004484265002,50.00489401166909,50.00529032812055,50.00566378872206,50.006006543039184,50.0063126926799,50.00657852664945,50.006802610773164,50.00698573248702,50.00713071485191,50.007242118295174,50.00732585663428,50.007388756394896,50.00743809058659,50.007481117849856,50.00752465455467,50.007574704678476,50.00763616527501,50.00771262093862,50.007806232419576,50.007917719708665,50.008046432751875,50.008190498924485,50.00834703193074,50.00851238447751,50.008682425903494,50.0088528257059,50.00901932568009,50.00917798501395,50.0093253863675,50.00945879380273,50.00957625774354,50.00967666532298,50.00975973824932,50.00982598302785,50.009876600831895,50.00991336599015,50.009938482913775,50.00995443172171,50.009963812244216,50.00996919541196,50.00997298949123,50.00997732721192,50.009983977894755,50.009994286994925,50.01000914366017,50.010028975364506,50.010053767331755,50.010083103371024,50.010116224026426,50.010152097420125,50.010189498054544,50.010227088857505,50.01026350211312,50.01029741535285,50.01032761892592,50.01035307264366,50.010372949645706,50.010386666385735,50.01039389835639,50.010394581861725,50.01038890274084,50.01037727348154,50.01036030057413,50.010338744285086,50.0103134732313,50.01028541623857,50.01025551395426,50.01022467257009,50.01019372179822,50.01016337894532,50.01013422056043,50.0101066627055,50.01008095043714,50.01005715661119,50.010035189652136,50.01001480949132,50.00999565048769,50.009977249828474,50.00995907967308,50.009940581168976,50.00992119843621,50.00990041068696,50.0098777608173,50.00985287906186,50.00982550063074,50.00979547662199,50.009762777910645,50.009727492119744,50.00968981416935,50.00965003124195,50.009608503286884,50.00956564039264,50.009521878474985,50.00947765476077,50.0094333844854,50.00938944008495,50.00934613395106,50.009303705555546,50.009262313450996,50.00922203234073,50.00918285510235,50.00914469936121,50.00910741796727,50.00907081253158,50.00903464905011,50.00899867457495,50.00896263389564,50.008926285255825,50.00888941424921,50.00885184520131,50.00881344953781,50.00877415085026,50.00873392658548,50.00869280648935,50.00865086811778,50.00860822987916,50.00856504218505,50.008521477356474,50.008477718958595,50.00843395122192,50.00839034915477,50.00834706986591,50.008304245507105,50.008261978121176,50.008220336548,50.008179355411016,50.008139036086206,50.00809934944739,50.0080602400996,50.008021631746786,50.007983433307295,50.0079455453768,50.00790786665183,50.00787029996038,50.007832757596866,50.00779516572024,50.00775746764772,50.007719625948596,50.007681623318,50.007643462277514,50.007605163809494,50.007566765081435,50.00752831645307,50.00748987798135,50.00745151564899,50.007413297538235,50.00737529015814,50.00733755510829,50.00730014623156,50.007263107370086,50.007226470799345,50.00719025637482,50.007154471385775,50.00711911107467,50.00708415974998,50.00704959239398,50.00701537664864,50.00698147505203,50.006947847392055,50.006914453047635,50.00688125319699,50.0068482127855,50.00681530216661,50.00678249834967,50.00674978581369,50.00671715687085,50.00668461158699,50.006652157290105,50.006619807716845,50.00656826563563,50.00772067814216,50.010813806801615,50.01844878484299,50.03118254178497,50.051878523611705,50.08067702934152,50.11730862236937,50.16000180105235,50.2049422482728,50.250429455109426,50.291192742728995,50.32602121476859,50.34902362022926,50.35940202711684,50.35307909150626,50.33089508889408,50.29254203308802,50.24035172338658,50.17750476149089,50.10707460832756,50.03470042660104,49.96362110120333,49.90014723729897,49.846722440906305,49.80821197371865,49.78551661137763,49.78071287256562,49.79296531512791,49.821174544497886,49.86295419005884,49.91455499752467,49.97264846375737,50.03214378574229,50.089676070199125,50.14040432429502,50.18181757503129,50.21067892795602,50.22596576114655,50.22680152113136,50.213852710839475,50.18858784853133,50.15307382766098,50.11043048797964,50.063466125305105,50.01587833888754,49.97040883097047,49.930241622407124,49.897335309190545,49.873564850965806,49.85967431850073,49.85591291189347,49.861755365211614,49.87600367637432,49.89715732047178,49.923117088103496,49.951925596396826,49.981283700793554,50.00934611656756,50.03424887203313,50.05471366726676,50.069717174402626,50.07878207493736,50.081833667040485,50.079181826230055,50.071539168693114,50.05978747403364,50.04509234695647,50.02858367834289,50.01149639038714,49.99488190300784,49.97972337406316,49.966751653018925,49.95651172935262,49.949304075284594,49.94520382516667,49.94410508788205,49.94570863227153,49.94962322852748,49.955340686531294,49.962346092598075,49.9700945162014,49.97809542995796,49.98589803824309,49.99313503931895,49.99951710354153,50.004838923260316,50.00897973824727,50.011885175942915,50.01356997496373,50.01409140597522,50.013551459668506,50.01207374832343,50.00980406779866,50.00689610636058,50.00351047195715,49.999808562438766,49.99595054645733,49.992093416946716,49.98838828210085,49.9849783880493,49.98199589230498,49.979557644253276,49.977761704276254,49.976681394909704,49.976362195560675,49.97681658498064,49.978022319724055,49.97992105691833,49.982419083528875,49.98539160781639,49.98868655223478,49.99213434233114,49.9955544686945,49.99876848990835,50.001608016258935,50.003927312353625,50.00561059099572,50.00658053112901,50.00680266326626,50.00628730569863,50.005089501163056,50.003303872373905,50.00105984616375,49.998511037713065,49.995826939206424,49.99317996903849,49.99073590315382,49.98864222525251,49.98701996504941,49.985956382804105,49.98550043527081,49.985661399356665,49.98640908224992,49.9876782317772,49.989373215126804,49.991376515510915,49.99355636402739,49.995776699070085,49.99790551088045,49.999823531322036,50.00143101925618,50.002653091659845,50.00344329445866,50.00378461583952,50.00368921485631,50.00319540492216,50.00236401412884,50.00127263605279,50.000009947043694,49.998669050284406,49.99734146215981,49.996111347794546,49.99505075843822,49.99421606065599,49.993645463703075,49.99335818413853,49.99335455554565,49.99361769076271,49.99411575343356,49.99480529245519,49.99563476715321,49.996548456137,49.99749016178084,49.99840665386201,49.99925062344783,49.999982932951085,50.0005742482793,50.00100580007702,50.00126955858256,50.001367628282,50.00131121119609,50.00111905459628,50.0008156866804,50.000429464989836,49.99999063332711,49.99952948402031,49.99907470021665,49.99865199665875,49.99828303581443,49.997984716146775,49.99776875251335,49.99764159990481,49.99760462454678,49.997654525389166,49.99778392329641,49.997982085586116,49.99823572999383,49.99852985936265,49.99884859780976,49.99917598036858,49.99949668564872,49.99979667472321,50.00006373516052,50.00028790808791,50.00046180011123,50.00058077133112,50.00064300233265,50.00064944168229,57.088009864216225,66.26933304083045,74.64048448435834,78.71672813125552,79.73775925162153,79.51045645038509,79.23841231163746,78.97043001168161,78.37673781248571,77.35987079968375,76.0705334723876,74.91955037665416,74.03172854928732,73.32473796291508,72.99681817305408,72.3749759017737,71.79986578118796,70.2109792137366,68.77640242333744,66.76985254169654,64.97731532981328,62.760425121604534,60.41548732021488,57.847037599364015,55.062351280145876,52.16407138968043,48.946343366469875,45.599258666528186,41.995962108225605,38.444015653418994,34.92158263540697,31.719428884300118,28.96422615669843,27.171078307205313,26.12703382888541,25.937129854758453,26.77356405415903,28.646982898447497,31.623641116408564,35.48464090016189,40.090740474228646,44.99809682778243,49.95910381273408,54.64692944094076,58.88944706575047,62.51283568696245,65.48432071806016,67.8314317273981,69.50493488345245,70.32762818484375,70.59617189093214,70.42046873709755,69.72308019778333,68.55751126571637,66.87949720923324,64.74188611684077,62.170822998945575,59.272053917025545,56.20319555782715,53.14153284541301,50.29842860793952,47.74525246972625,45.583796725363364,43.77483170810394,42.28013743863584,40.99509989284236,39.85909072823048,38.811086434530395,37.83141723403552,36.93420219282673,36.37607277529554,35.98143556917338,35.78149769819909,35.874965643202465,36.32642280425159,37.21080220184027,38.5732808099839,40.442632284929104,42.7974043265454,45.56769856249445,48.62412427259152,51.8382728333485,55.02842136915663,58.00192082312433,60.61377076884993,62.76458478245441,64.40625003078134,65.52016877900199,65.96492431667575,65.76066850627025,65.09453388585843,63.99518292658985,62.48171442321298,60.59944817270477,58.413577328124404,56.014112663917444,53.512519794871956,51.03601897744009,48.67208376674079,46.52787046977067,44.68293473167594,43.16791832790272,41.969991494339354,41.04814124490208,40.35031402799129,39.827922992634214,39.445454227361864,39.18425092801054,39.21957995611831,39.4028700315935,39.725633247732254,40.23453708865932,40.96612114155746,41.95021553087711,43.20607639581432,44.73567153119154,46.51862844215201,48.50769069377084,50.65378772617212,52.87384112012244,55.054325788012456,57.083789087804185,58.86673593657093,60.329695616792,61.42310933709388,62.119225390912895,62.40816219136411,62.16390167134561,61.45644480269036,60.417350548468704,59.08023166725486,57.48518998410492,55.69463218048061,53.784818799644334,51.84059919038099,49.9325807527321,48.11986410217726,46.47001965765675,45.030913564827685,43.825371800541525,42.85503244119189,42.10598877419158,41.556504247518525,41.18408852193934,40.9708403412664,40.90627027908614,40.99559624404708,41.38045846145397,41.916948255608084,42.60295755478309,43.46255027356426,44.501520597063276,45.71534243323202,47.08711491533931,48.58575205536259,50.18392694186867,51.83856020337997,53.48200355505601,55.04284195531049,56.4538520697669,57.65677384315668,58.60580103473376,59.26886928548033,59.627297526588784,59.67474596188323,59.41596358509087,58.78771016320374,57.834109030676416,56.6757181087721,55.355085808714186,53.91598180760893,52.41580495269682,50.91513516221618,49.447662435670395,48.06170892418687,46.80003636718071,45.69062148467627,44.74884722473328,43.979826406866664,43.381079185068835,42.945788933141976,42.6659826634896,42.53510444097894,42.54957664067529,42.70915519378929,43.01622477437431,43.50823904390845,44.21675409767026,45.05931124537988,46.02552318559487,47.108485837013845,48.28582181678137,49.53064343746217,50.82250297891417,52.115750343740345,53.360900686212666,54.5109341731561,55.5223077329243,56.35739418452496,56.986584563886375,57.38957648122262,57.55568769254221,57.48343253687376,57.17977648737444,56.65949423808884,55.94469889393496,55.05167716841614,53.97840594058912,52.83682282885252,51.672460625930086,50.51574138272689,49.38770192292586,48.323631138914145,47.3531326379098,46.495975413421654,45.76390800401537,45.16252771716064,44.6934980114071,44.35633235515206,44.14962803845146,44.07191756286599,44.122280912426135,44.30050737819583,44.60667782332399,45.04019408869584,45.59846529277245,46.27556622701556,47.06106508213419,47.93923683679783,48.88874271744314,49.882939805550095,50.89083499840235,51.87866693521658,52.81189819660258,53.657346224314885,54.38514962749977,54.97033834601659,55.39391358830241,55.643425292082135,55.713158994165155,55.60402428946858,55.32327044238934,54.8840608026922,54.30492090282096,53.60900106790496,52.82310296181229,51.9764365600348,51.099138002323755,50.22066617642446,49.368250955185935,48.565611881352694,47.83211673381594,47.182487600507436,46.62703401893667,46.172306443680945,45.82198448605986,45.577804520613654,45.44035184248152,45.40959575345925,45.48510977954349,45.66597197204219,45.95039228245497,46.335140954483144,46.814881797326876,47.381521931508416,48.02369937846742,48.72651744432473,49.4716131662797,50.23760314864392,51.00089423499876,51.736787126840184,52.42075084649636,53.02972333059793,53.54329828701998,53.9446947675304,54.22145356779168,54.365854596659,54.37508265316263,54.25118395148964,54.000850731320604,53.635055581640124,53.1685400723421,52.61915298577322,52.0070387985575,51.35369666782089,50.68096046034357,50.00998011048124,49.36030334315407,48.74915437366788,48.1909806303564,47.69729471923598,47.27678858933776,46.935653305371844,46.67801111781209,46.50636071383746,46.42194950756025,46.425012518539084,46.51484867975824,46.6897362343839,46.94671551187091,47.28128781008543,47.687092727953925,48.155632489601615,48.6761098635862,49.23543546791018,49.81844038293554,50.40830289357541,50.987167444665246,51.536905533238944,52.03994803947127,52.48011115500802,52.84334435694545,53.118346213896274,53.29701643651152,53.37473425343724,53.35046907155284,53.226737327788875,53.00942059610444,52.70745731732305,52.33241820786178,51.897976606060396,51.419291587504446,50.912332647981444,50.3931872226228,49.87740174130545,49.379409068895725,48.912087540118925,48.4864797209114,48.11167594847351,47.7948437291565,47.54136467720254,47.35502955572002,47.23824065694085,47.192178164958705,47.2169007967714,47.31136771683395,47.47338559801103,47.699499541101446,47.98485785661396,48.32308754821166,48.70621907663173,49.124695387226666,49.567491338731195,50.022356447494985,50.47617789491078,50.915444538317075,51.32677901862516,51.69749646204966,52.01614611955815,52.272996430823746,52.46043284639838,52.573248761964514,52.608820558765466,52.567166121919655,52.4508916448424,52.265034468030954,52.01681140647485,51.71528392728617,51.370954697936895,50.995314644902166,50.60036497478732,50.198143081352804,49.800283079562185,49.417639463873314,49.05999562611252,48.73586848399671,48.45240812412648,48.21537957046316,48.02920481216269,47.89703858929436,47.82085162457235,47.801499489121184,47.838762884295264,47.93135427492087,48.07689501442272,48.27187508004087,48.51161332723264,48.79023913187695,49.1007160848564,49.434925028249246,49.783817543688514,50.1376427992906,50.486241587374174,50.81939282620912,51.12719110919225,51.40043007990248,51.63096591552655,51.812037753909955,51.93852666322256,52.00714059030767,52.01651853916439,51.967252261884205,51.86182770330012,51.70449148329759,51.50105024918496,51.25861322534193,50.985290955987374,50.689865953966006,50.3814532543938,50.06917003482207,49.76183284034747,49.467698170427674,49.19425734836511,48.948090324697624,48.73477637057888,48.55885364283301,48.42381534726413,48.33212831513809,48.28526038411427,48.28370573000521,48.32700158980511,48.413734842351445,48.54154184925519,48.70710908140412,48.90618480685274,49.133613136752466,49.38340088097848,49.648825052959204,49.92258481949475,50.19699676538914,50.464227233556024,50.71655095046097,50.946621803027874,51.147739923672056,51.31409928815766,51.44100164953478,51.52502539711781,51.564141312436696,51.557770700345635,51.50678465400677,51.41344610865338,51.28129885719257,51.11500993375522,50.920173799799365,50.70308857749459,50.47051602270333,50.22943776469463,49.98682028301439,49.74939994189738,49.52349713095803,49.31486536477373,49.1285774821383,48.9689474005713,48.83948278101793,48.74286189319879,48.680927198779365,48.65468870769393,48.66433181448477,48.70922672687756,48.78793931330205,48.898245765467856,49.0371554951948,49.20094785829899,49.385228446213404,49.58500977441723,49.794819337180975,50.00883543538065,50.22104826594465,50.42544088931877,50.61618225682036,50.7878227986441,50.935482332407084,51.055020283957056,51.143179304498645,51.1976951054492,51.21736745612047,51.20208955726471,51.15283523270342,51.07160546240117,50.96133766482451,50.825782804309085,50.669356823088656,50.49697401739778,50.313870695391195,50.125427656970146,49.936999627732575,49.75375872986434,49.580557445575145,49.42181449130311,49.28142482533991,49.162692950112394,49.06828701935482,49.00021022647781,48.959785633512666,48.94765098009747,48.96376096004482,49.007395760239945,49.077176073045905,49.17108606806077,49.286506725005744,49.42026233062782,49.56868275105032,49.72768331529779,49.89286287761034,50.05961902802592,50.22327770227142,50.379232829159456,50.523090353834576,50.650810145576344,50.75883902457155,50.84422842893122,50.90473102650628,50.93887174198541,50.94599008322831,50.92625218122178,50.88063249685497,50.81086661409314,50.719377875147075,50.6091817725768,50.48377294988424,50.347000324458065,50.20293618068711,50.05574504183066,49.909557697798114,49.76835496470166,49.635864652987756,49.5154739429813,49.41015806032665,49.32242496718937,49.25427487647626,49.20717285015185,49.182032594200656,49.179209784165444,49.19850376327761,49.2391671274964,49.299923407486254,49.37899363937373,49.47413296971889,49.582678489149444,49.70160920433236,49.82761846012573,49.957198279675765,50.08673410278234,50.212607395812775,50.33130270627777,50.43951505273422,50.53425315594805,50.61293397009645,50.67346426120857,50.714305565698915,50.73451968095011,50.733792816167664,50.71243758815586,50.67137311348005,50.612084466375414,50.53656369108594,50.447235335232925,50.34687006887891,50.23849033817568,50.12527214660611,50.01044694821873,49.897207283484825,49.78861922187065,49.68754395359558,49.59657007436587,49.51795731959263,49.45359181448494,49.404952383106775,49.37308714559236,49.35859953763443,49.36164298597338,49.38192371542626,49.41871147556793,49.47085827980758,49.536825472397275,49.61471952141972,49.702336842953926,49.7972176861203,49.896708671491254,49.99803302214049,50.098366921091866,50.19491984354081,50.285016217888405,50.36617542613032,50.43618700269568,50.49317795000559,50.535669354464105,50.56261993420523,50.573454741316745,50.568077930133285,50.54686924101224,50.510664589231794,50.460721848978245,50.398673545581126,50.326468683124666,50.24630631275519,50.160563669569875,50.0717217619456,49.982291187056006,49.89474068422525,49.81143055198598,49.73455258608403,49.66607769344657,49.60771185281426,49.56086067302441,49.526602480060305,49.50566966279738,49.49843792504914,49.504923110288324,49.52478535309589,49.55734042691894,49.60157825781627,49.65618861968096,49.719593989228805,49.7899894042844,49.86538893810415,49.94367809283852,50.02267105702037,50.10017140416393,50.17403447445538,50.24222941897636,50.30289872830411,50.35441303583198,50.395419088827,50.424879012144295,50.44209933444591,50.44674868010724,50.4388635219832,50.41884190902735,50.38742559697237,50.34567149058135,50.294913726333306,50.236718062947226,50.172830487066264,50.10512207201908,50.03553214540595,49.966011731077415,49.89846904617298,49.83471857544545,49.7764349403681,49.725112459908686,49.68203099324747,49.648228387693266,49.62447964557854,49.61128278018063,49.60885125042476,49.617112835426134,49.63571481370548,49.66403532511792,49.70120079293864,49.74610924939709,49.79745932696545,49.8537845450653,49.91349234189337,49.97490708667873,50.036316078522766,50.09601731777303,50.15236764929322,50.203829746407635,50.24901634715386,50.28673018132586,50.315998140428384,50.336098438249124,50.34657977580283,50.347271844340696,50.33828685411724,50.32001214335044,50.29309428006492,50.25841539941602,50.217062803192825,50.170293072315395,50.119492097143784,50.06613250890082,50.0117299983178,49.95779994000767,49.90581561349456,49.85716913901765,49.81313604579392,49.77484418093159,49.743247466015056,49.71910482970295,49.702964497820766,49.69515371096097,49.69577386126958,49.704700987564365,49.721591530299136,49.74589321263456,49.77686086899025,49.813576978409436,49.85497657138406,49.89987606493782,49.94700544636642,49.99504308029486,50.04265226924051,50.08851856904466,50.13138676167598,50.17009633150518,50.2036142863925,50.23106421672779,50.25175059409506,50.265177472101335,50.27106095669422,50.26933505050334,50.26015073186187,50.243868389780545,50.22104398696592,50.192409550745104,50.15884878505456,50.12136874652229,50.08106862818768,50.03910674300733,49.9966667969968,49.95492449312928,49.91501541948858,49.87800505852025,49.84486161972786,49.816432257124205,49.79342309536832,49.77638336280282,49.765693820578036,49.76155958639789,49.764007377618995,49.77288713729268,49.7878779523645,49.80849811916362,49.83411915167876,49.86398345875119,49.897225335509724,49.93289482326664,49.96998389466712,50.00745432351367,50.044266509382446,50.079408454571166,50.111924043312264,50.14093975739739,50.165688983024026,50.185533122708094,50.199978822498124,50.20869075465788,50.211499553201634,50.20840467604801,50.199572153792865,50.18532737138843,50.16614320576272,50.142624000859634,50.11548599433486,50.08553491163589,50.053641509994584,50.02071588586045,49.98768135572662,49.95544868555268,49.92489138326721,49.896822688601645,49.87197480182785,49.850980794760204,49.83435954972696,49.82250397977834,49.81567269907293,49.813985237139015,49.81742082377348,49.825820710565225,49.83889393748222,49.856226395558735,49.87729297691955,49.90147253970596,49.92806534791373,49.95631257631412,49.985417401619074,50.01456713720007,50.0429558150084,50.069806580047285,50.09439324447431,50.11606035380034,50.134241148824614,50.148472864293424,50.1584088873309,50.16382740227842,50.16463626901592,50.16087401333913,50.15270694405901,50.14042254550258,50.12441941958136,50.105194162789,50.08332565562877,50.059457311595544,50.0342778779666,50.008501400864496,49.98284696343504,49.95801878096898,49.93468719388931,49.913471042926794,49.894921844978576,49.879510117378466,49.8676141263283,49.85951126490402,49.85537219915522,49.855257858108395,49.85911928474676,49.866800309374305,49.87804295292816,49.89249541455041,49.90972244422588,49.92921784735,49.95041881436799,49.97272171665941,49.99549896194448,50.01811646154146,50.03995123093849,50.06040862739288,50.07893872620427,50.095051352727864,50.10832932088478,50.11843948053978,50.12514124413271,50.128292344775694,50.12785167016851,50.12387911499115,50.11653249439263,50.106061658238424,50.09280003557842,50.077153917536904,50.05958985143501,50.040620567270956,50.02078988855049,50.00065709275782,49.9807811833385,49.96170551662061,49.943943195962845,49.927963604314535,49.91418039815765,49.90294123319199,49.89451943748946,49.88910779301807,49.886814532677136,49.88766160793623,49.89158523199631,49.898438654895806,49.907997079838026,49.91996458394374,49.933982861617,49.94964156511838,49.96648997560476,49.984049700101785,50.0018280572878,50.0193317894709,50.03608072164921,50.0516209827956,50.06553741084303,50.07746478205311,50.08709753761419,50.094197724787755,50.09860092530209,50.10022000791379,50.09904661251355,50.09515034685555,50.088675750779224,50.07983715352106,50.068911614461285,50.056230193879415,50.04216784600364,50.0271322604656,50.01155199952838,49.995864287140215,49.98050280260477,49.96588581761177,49.95240499208503,49.94041511357421,49.930225028592155,49.92208997416283,49.916205475453694,49.91270293199223,49.911646971539334,49.9130346078296,49.91679619643966,49.92279814219933,49.93084727194424,49.940696748206236,49.95205336298925,49.964586016632204,49.97793515570776,49.9917229169551,50.005563702552436,50.019074896815546,50.03188742680787,50.043655870314716,50.054067824813075,50.06285227068445,50.06978669073734,50.07470274538966,50.077490347410134,50.07810003030133,50.07654355828666,50.07289278128131,50.067276792953265,50.05987750184012,50.05092377248751,50.04068433401769,50.029459686105035,50.017573256148786,50.005362076077354,49.99316725272736,49.981324502565286,49.97015501044646,49.9599568541784,49.95099721306465,49.94350555059171,49.93766793018467,49.93362258957784,49.9314568647479,49.93120551927533,49.93285050000837,49.93632210543863,49.941501519611386,49.94822463202413,49.9562870331812,49.965450046724115,49.97544763290514,49.9859939752834,49.99679154365413,50.0075394121584,50.01794160303073,50.02771522418354,50.03659817328823,50.04435619243332,50.05078907576934,50.05573585741422,50.05907883759148,50.060746340502114,50.06071413653379,50.05900550263181,50.055689936439066,50.05088058057166,50.04473045161741,50.03742760275467,50.02918937813963,50.0202559405227,50.01088327034973,50.001335844625274,49.99187920710516,49.98277263826198,49.9742621244701,49.966573811704045,49.95990811053743,49.9544345972346,49.950287831076295,49.9475641815368,49.9463197312363,49.94656929232615,49.94828654564729,49.95140528407907,49.95582171437622,49.96139774587148,49.967965170112535,49.975330613262145,49.983281123403884,49.991590238301754,50.00002436619909,50.008349303442046,50.016336708539264,50.02377035308696,50.030451976005885,50.036206578787265,50.04088701573839,50.04437775415693,50.046597704321954,50.04750204735832,50.047083019434496,50.04536964231283,50.042426421849235,50.03835106650302,50.03327130620546,50.02734091708822,50.02073507881204,50.01364520795478,50.006273422743824,49.998826801174374,49.99151159629299,49.98452756936449,49.97806259416211,49.97228767422351,49.96735250016453,49.96338165665,49.960471568993924,49.95868825818514,49.95806595095616,49.95860656883045,49.96028009735263,49.96302581434365,49.96675433442778,49.97135040663697,49.97667638300083,49.982576259078705,49.98888017278398,49.99540923600365,50.00198056480091,50.0084123687596,50.0145289585526,50.020165533275986,50.025172615540534,50.02942001267765,50.03280019645782,50.035231011064624,50.03665763917541,50.0370537782375,50.03642200263537,50.03479331161978,50.03222588678248,50.028803105692226,50.024630879307765,50.0198343992793],"type":"scatter","name":"source2_freq","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1,0.1001,0.1002,0.1003,0.1004,0.1005,0.1006,0.1007,0.1008,0.1009,0.101,0.1011,0.1012,0.1013,0.1014,0.1015,0.1016,0.1017,0.1018,0.1019,0.102,0.1021,0.1022,0.1023,0.1024,0.1025,0.1026,0.1027,0.1028,0.1029,0.103,0.1031,0.1032,0.1033,0.1034,0.1035,0.1036,0.1037,0.1038,0.1039,0.104,0.1041,0.1042,0.1043,0.1044,0.1045,0.1046,0.1047,0.1048,0.1049,0.105,0.1051,0.1052,0.1053,0.1054,0.1055,0.1056,0.1057,0.1058,0.1059,0.106,0.1061,0.1062,0.1063,0.1064,0.1065,0.1066,0.1067,0.1068,0.1069,0.107,0.1071,0.1072,0.1073,0.1074,0.1075,0.1076,0.1077,0.1078,0.1079,0.108,0.1081,0.1082,0.1083,0.1084,0.1085,0.1086,0.1087,0.1088,0.1089,0.109,0.1091,0.1092,0.1093,0.1094,0.1095,0.1096,0.1097,0.1098,0.1099,0.11,0.1101,0.1102,0.1103,0.1104,0.1105,0.1106,0.1107,0.1108,0.1109,0.111,0.1111,0.1112,0.1113,0.1114,0.1115,0.1116,0.1117,0.1118,0.1119,0.112,0.1121,0.1122,0.1123,0.1124,0.1125,0.1126,0.1127,0.1128,0.1129,0.113,0.1131,0.1132,0.1133,0.1134,0.1135,0.1136,0.1137,0.1138,0.1139,0.114,0.1141,0.1142,0.1143,0.1144,0.1145,0.1146,0.1147,0.1148,0.1149,0.115,0.1151,0.1152,0.1153,0.1154,0.1155,0.1156,0.1157,0.1158,0.1159,0.116,0.1161,0.1162,0.1163,0.1164,0.1165,0.1166,0.1167,0.1168,0.1169,0.117,0.1171,0.1172,0.1173,0.1174,0.1175,0.1176,0.1177,0.1178,0.1179,0.118,0.1181,0.1182,0.1183,0.1184,0.1185,0.1186,0.1187,0.1188,0.1189,0.119,0.1191,0.1192,0.1193,0.1194,0.1195,0.1196,0.1197,0.1198,0.1199,0.12,0.1201,0.1202,0.1203,0.1204,0.1205,0.1206,0.1207,0.1208,0.1209,0.121,0.1211,0.1212,0.1213,0.1214,0.1215,0.1216,0.1217,0.1218,0.1219,0.122,0.1221,0.1222,0.1223,0.1224,0.1225,0.1226,0.1227,0.1228,0.1229,0.123,0.1231,0.1232,0.1233,0.1234,0.1235,0.1236,0.1237,0.1238,0.1239,0.124,0.1241,0.1242,0.1243,0.1244,0.1245,0.1246,0.1247,0.1248,0.1249,0.125,0.1251,0.1252,0.1253,0.1254,0.1255,0.1256,0.1257,0.1258,0.1259,0.126,0.1261,0.1262,0.1263,0.1264,0.1265,0.1266,0.1267,0.1268,0.1269,0.127,0.1271,0.1272,0.1273,0.1274,0.1275,0.1276,0.1277,0.1278,0.1279,0.128,0.1281,0.1282,0.1283,0.1284,0.1285,0.1286,0.1287,0.1288,0.1289,0.129,0.1291,0.1292,0.1293,0.1294,0.1295,0.1296,0.1297,0.1298,0.1299,0.13,0.1301,0.1302,0.1303,0.1304,0.1305,0.1306,0.1307,0.1308,0.1309,0.131,0.1311,0.1312,0.1313,0.1314,0.1315,0.1316,0.1317,0.1318,0.1319,0.132,0.1321,0.1322,0.1323,0.1324,0.1325,0.1326,0.1327,0.1328,0.1329,0.133,0.1331,0.1332,0.1333,0.1334,0.1335,0.1336,0.1337,0.1338,0.1339,0.134,0.1341,0.1342,0.1343,0.1344,0.1345,0.1346,0.1347,0.1348,0.1349,0.135,0.1351,0.1352,0.1353,0.1354,0.1355,0.1356,0.1357,0.1358,0.1359,0.136,0.1361,0.1362,0.1363,0.1364,0.1365,0.1366,0.1367,0.1368,0.1369,0.137,0.1371,0.1372,0.1373,0.1374,0.1375,0.1376,0.1377,0.1378,0.1379,0.138,0.1381,0.1382,0.1383,0.1384,0.1385,0.1386,0.1387,0.1388,0.1389,0.139,0.1391,0.1392,0.1393,0.1394,0.1395,0.1396,0.1397,0.1398,0.1399,0.14,0.1401,0.1402,0.1403,0.1404,0.1405,0.1406,0.1407,0.1408,0.1409,0.141,0.1411,0.1412,0.1413,0.1414,0.1415,0.1416,0.1417,0.1418,0.1419,0.142,0.1421,0.1422,0.1423,0.1424,0.1425,0.1426,0.1427,0.1428,0.1429,0.143,0.1431,0.1432,0.1433,0.1434,0.1435,0.1436,0.1437,0.1438,0.1439,0.144,0.1441,0.1442,0.1443,0.1444,0.1445,0.1446,0.1447,0.1448,0.1449,0.145,0.1451,0.1452,0.1453,0.1454,0.1455,0.1456,0.1457,0.1458,0.1459,0.146,0.1461,0.1462,0.1463,0.1464,0.1465,0.1466,0.1467,0.1468,0.1469,0.147,0.1471,0.1472,0.1473,0.1474,0.1475,0.1476,0.1477,0.1478,0.1479,0.148,0.1481,0.1482,0.1483,0.1484,0.1485,0.1486,0.1487,0.1488,0.1489,0.149,0.1491,0.1492,0.1493,0.1494,0.1495,0.1496,0.1497,0.1498,0.1499,0.15,0.1501,0.1502,0.1503,0.1504,0.1505,0.1506,0.1507,0.1508,0.1509,0.151,0.1511,0.1512,0.1513,0.1514,0.1515,0.1516,0.1517,0.1518,0.1519,0.152,0.1521,0.1522,0.1523,0.1524,0.1525,0.1526,0.1527,0.1528,0.1529,0.153,0.1531,0.1532,0.1533,0.1534,0.1535,0.1536,0.1537,0.1538,0.1539,0.154,0.1541,0.1542,0.1543,0.1544,0.1545,0.1546,0.1547,0.1548,0.1549,0.155,0.1551,0.1552,0.1553,0.1554,0.1555,0.1556,0.1557,0.1558,0.1559,0.156,0.1561,0.1562,0.1563,0.1564,0.1565,0.1566,0.1567,0.1568,0.1569,0.157,0.1571,0.1572,0.1573,0.1574,0.1575,0.1576,0.1577,0.1578,0.1579,0.158,0.1581,0.1582,0.1583,0.1584,0.1585,0.1586,0.1587,0.1588,0.1589,0.159,0.1591,0.1592,0.1593,0.1594,0.1595,0.1596,0.1597,0.1598,0.1599,0.16]},{"mode":"lines","y":[-0.9000000000000001,0.7500000000000003,3.975693351829396e-16,-0.26553707952384137,-0.40513168510867026,-0.6031220291899424,-0.8595932028972396,-1.1336702142727515,-1.4437834708057453,-1.7746618735980666,-2.117224046607253,-2.4686612790889324,-2.8193325361357395,-3.1656839997531727,-3.498513198583231,-3.8151204636182316,-4.105726064041721,-4.366916924349764,-4.590196078342992,-4.772247632378678,-4.907767271101946,-4.9946053088337035,-5.032019349664184,-5.019973596994347,-4.96189111700633,-4.861628370013275,-4.728470470427536,-4.569694431481168,-4.39707878806885,-4.219486935410275,-4.047686151180678,-3.8894124322325196,-3.751615042180977,-3.6388603843437686,-3.553143475901385,-3.4951649657407033,-3.46255127674255,-3.4527207365524344,-3.4605482581205997,-3.4814783032951873,-3.5095122732557877,-3.539624777318855,-3.5665755013780176,-3.586194066960729,-3.59510529416072,-3.590886820919667,-3.57248676611921,-3.5395512923392554,-3.4931691807034873,-3.434798928498616,-3.3669941481285055,-3.2923442329464203,-3.2139577805750212,-3.1347035293345247,-3.057379850176312,-2.9843699824446004,-2.9175491212438076,-2.858317693710763,-2.807365678551583,-2.764957009029845,-2.730678896077377,-2.7038236518068026,-2.6832139050450534,-2.667550424769987,-2.6553480978165696,-2.6451682630494653,-2.6356503970198717,-2.6256039243778404,-2.6140924045783693,-2.6004059639242034,-2.5841512801147726,-2.565151782849007,-2.543510600470812,-2.519489781571677,-2.493533457983532,-2.466165153395476,-2.4379756744723737,-2.4095584409065864,-2.3814765781727654,-2.354238594346198,-2.32826133894031,-2.303876930703414,-2.2813045397198977,-2.2606749606488585,-2.2420174546903353,-2.2252891359466846,-2.210376351318239,-2.1971198088833157,-2.1853254509429347,-2.174780963281169,-2.1652697588908127,-2.1565786460814347,-2.148509893126264,-2.140882462739731,-2.1335397570536028,-2.1263476598844355,-2.119197988564557,-2.112005965099274,-2.1047107432719807,-2.097273585406989,-2.0896769762979086,-2.0819234727129814,-2.0740342920271475,-2.0660480340269127,-2.0580188554048324,-2.0500141972631147,-2.04211221553039,-2.0343981459638036,-2.0269607476852127,-2.0198876844401594,-2.0132612299429242,-2.0071536786212754,-2.0016231300264473,-1.9967101877110869,-1.9924349434873276,-1.9887959818247543,-1.9857695363001038,-1.9833111910631074,-1.9813576197455534,-1.9798306041033236,-1.978641006788341,-1.977694076959261,-1.9768946334942887,-1.976152295097232,-1.9753864903038,-1.9745303128779814,-1.9735340381363051,-1.972366719827554,-1.9710173455415678,-1.9694939229689592,-1.9678220973672753,-1.966042138120801,-1.9642055494218766,-1.962370879060565,-1.9605993662400851,-1.9589507223602982,-1.957479037420072,-1.956229584690629,-1.9552360240283195,-1.9545189133148289,-1.9540847910555932,-1.9539265653572615,-1.954024490826159,-1.954348106566716,-1.9548586190278452,-1.955511708252492,-1.9562605226618186,-1.9570585437451813,-1.957862348833181,-1.9586338229398628,-1.959342028144366,-1.9599643189684506,-1.9604869870547783,-1.960905177620917,-1.9612223431116464,-1.9614491721003526,-1.9616021818157203,-1.9617020799356213,-1.961771986191873,-1.961835715139404,-1.961916123067944,-1.962033733115257,-1.9622055820393354,-1.9624444491928623,-1.9627583841039251,-1.9631506059549144,-1.9636196915709143,-1.964160038519455,-1.9647625375729507,-1.9654153802551457,-1.9661049606328425,-1.9668167729321953,-1.9675362884417582,-1.9682497229948326,-1.9689446975150529,-1.9696107344213314,-1.9702396044000205,-1.9708255045921044,-1.9713650881227407,-1.9718573583905135,-1.972303448403552,-1.9727063179329147,-1.9730703859544598,-1.973401136310444,-1.9737047097505624,-1.9739875144967935,-1.9742558638986292,-1.974515661831334,-1.9747721402274128,-1.9750296571765256,-1.9752915565001308,-1.9755600875900077,-1.9758363836271036,-1.9761204911274635,-1.976411446727689,-1.9767073917176803,-1.9770057185361698,-1.9773032395129897,-1.9775963709147446,-1.9778813233990715,-1.9781542914715828,-1.9784116342441653,-1.9786500405389114,-1.9788666721032369,-1.9790592794913693,-1.9792262863454426,-1.9793668391582628,-1.9794808209050283,-1.9795688288470497,-1.9796321181020362,-1.9796725146122267,-1.9796923023362916,-1.9796940910026355,-1.9796806717278812,-1.9796548682491824,-1.9796193920623604,-1.9795767089242222,-1.979528924113944,-1.9794776919308905,-1.9794241541689965,-1.9793689098290521,-1.97931201704294,-1.9792530258116166,-1.9791910386315257,-1.979124794419759,-1.9790527697930436,-1.978973291164797,-1.9788846503220245,-1.9787852166764301,-1.9786735393577883,-1.9785484337270849,-1.9784090476298997,-1.9782549045533615,-1.9780859221303024,-1.9779024062100228,-1.977705022154798,-1.9774947463077297,-1.977272801774873,-1.9770405832150164,-1.9767995759792558,-1.9765512747248888,-1.9762971066006545,-1.97603836327365,-1.9757761454368692,-1.975511322284496,-1.9752445074372682,-1.9749760516442374,-1.974706051514828,-1.9744343726485574,-1.974160684676618,-1.9738845052788971,-1.9736052497971566,-1.97332228308614,-1.973034970268305,-1.9727427234681552,-1.9724450420188278,-1.9721415442727046,-1.9718319897925611,-1.9715162913811883,-1.9711945170836906,-1.9708668828570977,-1.9705337371381977,-1.9701955388743135,-1.9698528308723418,-1.9695062103950463,-1.9691562989523907,-1.9688037130825218,-1.9684490377097241,-1.9680928033639025,-1.9677354682050785,-1.967377405439055,-1.967018896340246,-1.9666601287743224,-1.966301200802772,-1.9659421287239784,-1.9655828587107433,-1.9652232810998056,-1.9648632463279223,-1.9645025815226846,-1.9641411068132932,-1.9637786505299963,-1.9634150625988298,-1.9630502255928244,-1.9626840630735314,-1.9623165450235374,-1.9619476903376778,-1.9615775664862043,-1.9612062865954165,-1.9608340042947567,-1.9604609067604784,-1.9600872064379098,-1.9597131319506138,-1.9593389187058439,-1.958964799682691,-1.9585909968475883,-1.9582177135809773,-1.9578451284273408,-1.957473390396989,-1.957102615961096,-1.9567328877914385,-1.9563642552095646,-1.9559967362284314,-1.9556303209983952,-1.955264976408915,-1.9549006515526433,-1.9545372837286847,-1.9541748046499705,-1.9538131465241992,-1.9534522476989642,-1.9530920575981792,-1.9527325407253273,-1.952373679568198,-1.9520154763051918,-1.9516579532814093,-1.9513011522903068,-1.9509451327602934,-1.9505899690003223,-1.9502357467039764,-1.9498825589431659,-1.9495305019005869,-1.9491796705931477,-1.9488301548272446,-1.9484820356026493,-1.948135382145279,-1.9477902497055415,-1.9474466782075914,-1.9471046917823478,-1.9467642991647842,-1.9464254948866875,-1.946088261154712,-1.9457525702688974,-1.945418387413926,-1.9450856736432183,-1.9447543888739993,-1.9444244947225213,-1.9440959570263696,-1.9437687479294325,-1.9434428474367584,-1.943118244384295,-1.9427949368050348,-1.9424729317100904,-1.942152244335795,-1.9418328969364131,-1.9415149172226163,-1.9411983365604244,-1.9408831880515238,-1.940569504614859,-1.9402573171804338,-1.9399466530929779,-1.9396375348035944,-1.9393299789052623,-1.9390239955443582,-1.9387195882165407,-1.938416753931885,-1.9381154837144585,-1.9378157633847888,-1.937517574561051,-1.9372208958070185,-1.9369257038524308,-1.9366319748115255,-1.936339685333077,-1.9360488136221439,-1.9357593402874185,-1.9354712489799337,-1.9351845268040926,-1.934899164495626,-1.9346151563751322,-1.9343325000964813,-1.9340511962206215,-1.93377124765117,-1.9334926589740795,-1.933215435744889,-1.9329395837667616,-1.9326651083996194,-1.932392013936295,-1.9321203030743566,-1.9318499765058423,-1.9315810326386427,-1.9313134674551828,-1.9310472745064682,-1.9307824450326407,-1.9305189681946755,-1.9302568313975246,-1.929996020681454,-1.9297365211565825,-1.9294783174545205,-1.929221394172885,-1.9289657362896366,-1.9287113295276976,-1.9284581606543831,-1.9282062177042938,-1.927955490118904,-1.9277059688016236,-1.9274576460903694,-1.9272105156551762,-1.926964572330721,-1.9267198118969424,-1.9264762308230172,-1.9262338259901703,-1.9259925944099014,-1.9257525329530267,-1.925513638103114,-1.9252759057467719,-1.9250406638289708,-1.9247239166719192,-1.9242498325688964,-1.923380118969075,-1.9219518581563497,-1.9195922702601744,-1.916047986242195,-1.9110480108529266,-1.9044285450139218,-1.8962079869980708,-1.8863655554968277,-1.875219736369688,-1.8629450823750056,-1.8501435797471468,-1.8372076554255614,-1.824861141109799,-1.813586145961656,-1.8039875235729048,-1.7964818459140453,-1.791382769597415,-1.7889095743235826,-1.7889922546654746,-1.7915619000343879,-1.7961850471195056,-1.8025139495692728,-1.809887791986704,-1.8177804839265501,-1.8254929134598727,-1.8324697754285313,-1.8381603738958334,-1.84213843319076,-1.8441383213169193,-1.8439799043479677,-1.8417307844292374,-1.8374963103447652,-1.831630706551829,-1.824479560387842,-1.816563893095957,-1.808357580493265,-1.8003944481508953,-1.7931355977000796,-1.786985943454739,-1.7822682754130437,-1.7791672977422068,-1.7777926911042619,-1.7780870024159792,-1.77993896146604,-1.783093738323216,-1.7872730674000543,-1.792115563188453,-1.7972691184204095,-1.8023720220233057,-1.807099606817088,-1.8111809169990212,-1.814398907874266,-1.8166273250321598,-1.8177967193026785,-1.8179338162534224,-1.8171116679023973,-1.8154778538154512,-1.8132063648831076,-1.8105084027976281,-1.807598158989265,-1.8046874179444201,-1.8019701568284323,-1.7996071768420292,-1.7977279938375186,-1.7964133331207377,-1.7957086798584847,-1.7956110602778814,-1.7960872805337604,-1.7970682532119675,-1.7984658825635658,-1.8001748019721924,-1.8020841089911779,-1.8040839647280313,-1.8060709361428795,-1.8079560273569621,-1.8096644626193668,-1.8111422407339406,-1.8123523648702529,-1.8132783453990726,-1.813919106120667,-1.8142893043768407,-1.8144146610501313,-1.814329992219428,-1.8140759018946655,-1.8136958117189672,-1.8132341781658676,-1.8127336422649045,-1.8122344164117181,-1.8117722035569432,-1.8113782138970735,-1.8110780012434415,-1.8108916589340327,-1.810833388457296,-1.8109116587075544,-1.811129224919688,-1.8114832769437157,-1.8119657112953238,-1.812563438806541,-1.8132588606323383,-1.8140305101887468,-1.814853781133251,-1.8157019620259442,-1.8165472519237662,-1.8173620720233659,-1.8181202951080513,-1.8187985647527127,-1.8193775239254328,-1.8198428031417042,-1.8201859419597435,-1.8204047527289031,-1.8205036445373328,-1.8204932180960278,-1.8203898354593457,-1.820214508338732,-1.819991768309113,-1.8197481103437008,-1.8195104332075829,-1.819304428984213,-1.8191530041284802,-1.8190750234050304,-1.8190841373762228,-1.8191881903042952,-1.8193887636168353,-1.8196813732693575,-1.820055830377509,-1.8204971406508856,-1.8209865513272865,-1.821502885857371,-1.8220239627225219,-1.8225280013178233,-1.8229950152826357,-1.8234079322476446,-1.8237535992954195,-1.8240233605745866,-1.8242134459742998,-1.8243249114791187,-1.8243633673232869,-1.8243383576982444,-1.8242625642098327,-1.8241508368397688,-1.8240191326329187,-1.8238834758991138,-1.823758932217456,-1.823658762353304,-1.8235936865918019,-1.8235714163084833,-1.8235963562810578,-1.823669579728722,-1.8237889840476902,-1.8239496554792425,-1.8241443759087226,-1.824364233911566,-1.8245993064509478,-1.8248393330932915,-1.825074379643725,-1.8252954048852008,-1.8254947487216835,-1.825666474345631,-1.8258065931484002,-1.8259131391114878,-1.8259861223358773,-1.8260273641427978,-1.8260402380835492,-1.8260293458274925,-1.8260001441261418,-1.8259585638439813,-1.8259106289812967,-1.8258621144406395,-1.825818243646912,-1.825783452592957,-1.825761216664667,-1.8257539507220508,-1.8257629760417697,-1.8257885500692173,-1.8258299514873129,-1.825885607067965,-1.825953252933113,-1.8260301132881618,-1.826113090202574,-1.8261989491875517,-1.8262844956015587,-1.82636673132938,-1.826442988472387,-1.8265110350043234,-1.8265691508934314,-1.8266161743533122,-1.8266515184039482,-1.8266751605629403,-1.8266876073436968,-1.8266898379934082,-1.8266832303693281,-1.826669473878086,-1.8266504732311877,-1.826628247793334,-1.8266048306546168,-1.8265821717205537,-1.337531019255395,-0.5551617712952388,0.4704238201288683,1.5131748015487572,2.5562845068757536,3.5958444798433242,4.642859408502148,5.681032135105359,6.687022588280716,7.647326708634805,8.558570787032012,9.437032218341807,10.29014107224905,11.119818276974215,11.947487236934334,12.737384371293844,13.512528539506485,14.196317186283888,14.848759937952632,15.407777135215541,15.917905007321282,16.331010993006398,16.66183894221141,16.89474752202128,17.02362426538509,17.047657893816492,16.947015428600434,16.726400865845534,16.3695529858168,15.890372764705807,15.284588646808157,14.573653508126608,13.773476266023495,12.93378924696363,12.06680621476134,11.209936725561246,10.404235919703094,9.684803612917719,9.093383131636815,8.653593323637084,8.390946269897219,8.303729021753655,8.392359644212172,8.639932245042578,9.029638036605217,9.536034181463712,10.137177904021925,10.811991834049937,11.534188428863953,12.268029375123419,13.006013105977019,13.73132362084466,14.420976233475006,15.061043692372303,15.63081466202176,16.11615160702632,16.5015203184837,16.77822082656169,16.943745077018704,17.00183826533189,16.964684196067296,16.841905035978513,16.64989132060082,16.398511252688447,16.098405764845797,15.75424353601896,15.370958062989983,14.950613170410175,14.495935171624641,14.010653392213953,13.515231966964194,13.005912560915375,12.493368086400226,11.990950359805248,11.51219764789788,11.074199189868827,10.694533402133162,10.39171958394167,10.182093344946487,10.078628622358378,10.088404731065646,10.214803069919725,10.452888912479699,10.791231316055997,11.214915620134033,11.70623978556546,12.24677752966432,12.817637945674774,13.390343372595028,13.944315664012281,14.468800278585686,14.946340301593247,15.362227824606505,15.704395200107774,15.963384239806725,16.133498024457637,16.213388142357633,16.206484191605906,16.117816857516992,15.957520183789214,15.737386154226995,15.469109708778419,15.163223063123654,14.828503992212157,14.47191475280336,14.098948484625067,13.714199233086472,13.321949589095354,12.938938374933443,12.562960333139884,12.200984965058717,11.861380678758985,11.552669755986798,11.284378869152466,11.066479612104924,10.908602513721215,10.819166788669731,10.804338354284504,10.868859980721679,11.013388567560435,11.233999489060006,11.523280915095725,11.870878705324706,12.264281214244999,12.689651407947025,13.132514145991397,13.57830115848811,14.003806969489183,14.396556813520696,14.747903175906082,15.046253006660544,15.283140253762586,15.45298422998754,15.553072232261496,15.583821756722072,15.547593349516985,15.448709084291796,15.294375675056592,15.09274572899486,14.852262826985372,14.581105922474068,14.286758167724223,13.975894986259677,13.654474722005506,13.327974020212572,13.00166766091785,12.681425433479978,12.382482501792337,12.1034238689402,11.851721300872883,11.634449530953354,11.457896402341369,11.328120295678826,11.250317596998077,11.228283109380317,11.265164693441966,11.361983135095958,11.516749394760621,11.725147722085053,11.980637154369898,12.27481227952857,12.597895187500194,12.939220326178487,13.28768514533107,13.632165374706958,13.96188851880667,14.261395457411627,14.521378346470872,14.737781314951764,14.904359401992144,15.017653799530477,15.076703501632046,15.082628917209677,15.036828211775035,14.943325358065,14.807280775574577,14.634457546255383,14.430978719005223,14.20300139656196,13.95649690117311,13.697161932779606,13.430433055545604,13.161574928122766,12.895800443114727,12.638379039348457,12.394702749088527,12.172629781855097,11.980433712102357,11.819745577004035,11.695911048496924,11.61306112702302,11.57395835998818,11.580480597522751,11.634049745321363,11.733516027613518,11.876199703146426,12.057867183312112,12.272795086865585,12.514048527014781,12.773802557059609,13.0436881879653,13.315129736110318,13.579657219130924,13.829192350377951,14.056319669465218,14.254544817029082,14.41765740831225,14.538857397537951,14.619409845106745,14.658318221619835,14.656178609416136,14.614261365116079,14.535718342141616,14.424365728333404,14.284492132323537,14.120691954325007,13.937678413817677,13.740187066241283,13.532924009956645,13.3205468617336,13.10767201305296,12.898901773569564,12.698846062311612,12.51211669165396,12.343277158927357,12.19674211522111,12.076632762162541,11.986596053324638,11.929605764526478,11.907761620095778,11.922113368793429,11.972531013778434,12.057643049468311,12.17485053711952,12.320416317188569,12.489617495667975,12.676943852787858,12.876324894184423,13.081368462289827,13.285600722691816,13.482698146326456,13.666708025869301,13.832251604362403,13.974705284763667,14.090351067154085,14.176486933802012,14.231487130295989,14.254805242001842,14.246919132464111,14.20922354193842,14.143884708549194,14.053675157205351,13.941809376637488,13.811797070126998,13.667325895839639,13.51217762636513,13.350175479611814,13.185154907044485,13.02094719340326,12.861364449662933,12.710175182617636,12.571062242142155,12.447557718024292,12.342953680461479,12.260191481781135,12.201736913142902,12.169451875607496,12.164475837430222,12.18713078437011,12.236861732944336,12.312221168646042,12.410900608490968,12.52980728457302,12.665179409206146,12.812730880347768,12.967815301989237,13.12559991941768,13.28124140536316,13.43005694961404,13.567685060201114,13.690230819503258,13.79439017754596,13.877547643816506,13.937842078685009,13.974196581220493,13.98631099490716,13.974618970451193,13.940215262730879,13.884762055699788,13.810384859405646,13.719568440380861,13.6150614159772,13.499795062548188,13.376818299859437,13.249247455039841,13.12022682356449,12.992894521688374,12.87034772231523,12.755601973637415,12.651540734902465,12.560853292744335,12.485961608767122,12.428939116759127,12.39142672876501,12.374552989942716,12.378866121859383,12.404285415153852,12.450078035688186,12.514865023541386,12.596657484339419,12.692921231138952,12.800665876919872,12.916552903870457,13.037016585076916,13.158391652819992,13.277042021139824,13.38948539512429,13.492509060594939,13.583272486794083,13.659392700846604,13.719008872447255,13.76082338801211,13.784117993613533,13.788745318683983,13.7750980418679,13.744059792378366,13.696943223459172,13.6354212583098,13.561457185252513,13.47723815285125,13.385114955805845,13.287549146582142,13.187066803394526,13.086216988579348,12.987532201492693,12.893488008183612,12.806459475989946,12.728672951373946,12.662152949477596,12.608665306303266,12.569659088125931,12.546210867389437,12.538975677324494,12.548149118750942,12.573444654100985,12.614089135001565,12.668838217871167,12.736011748933358,12.813547686283883,12.899071873202697,12.989980102718008,13.083528437128782,13.176927607805307,13.267437414826478,13.352457272849975,13.429609344020488,13.496811051274424,13.55233421176687,13.594848628511913,13.623448772418485,13.637663173192568,13.637447246689652,13.62316138589632,13.595537067304873,13.55563431743418,13.504794044734684,13.444588452459255,13.376772080123105,13.303235114641984,13.22595963987643,13.14697862186822,13.068336787477545,12.992052225384674,12.920077544244853,12.854259738308254,12.79629847226259,12.747703213216017,12.709750399808962,12.683442531242452,12.669471573096718,12.66818932352813,12.679587310005868,12.70328838832405,12.738551539127908,12.784290493814742,12.83910588847993,12.901329759453818,12.969080451726892,13.040325468034494,13.112949455239699,13.184824386055617,13.253879011100366,13.318164792145126,13.37591575866253,13.42560005020755,13.465961323044723,13.496048717929227,13.51523470124556,13.523220774348227,13.520031739849538,13.505999842977959,13.48174059092312,13.448122328384995,13.406231680624023,13.357336774833955,13.302849765272112,13.244289694936038,13.183246216654382,13.121344254240599,13.060209375900547,13.001433516318562,12.94654073008025,12.896952869093065,12.853955409580971,12.818664053085238,12.791993125013288,12.774627127339473,12.766997010757974,12.769262773975743,12.781303855246234,12.802718461915342,12.83283252176729,12.870718389057421,12.915222861502366,12.965003523094397,13.018571969935813,13.074342132481634,13.130681688721133,13.18596446500478,13.238621732201207,13.287190410811142,13.330356388552168,13.366991420465826,13.396182418689621,13.417252337152698,13.42977229799647,13.433565063333058,13.42870039049626,13.415483179372902,13.39443558844314,13.366274435454995,13.331885201177768,13.292293833146159,13.24863733204104,13.202133840191467,13.154052687748965,13.10568463143641,13.058312376648242,13.013181424322932,12.971471331267914,12.934267602949854,12.902534625075365,12.877090250010655,12.85858284802895,12.84747177491998,12.844012266706628,12.848245730179455,12.859996249689079,12.878873885817445,12.90428502283992,12.935449660958762,12.971425181412249,13.011135771029425,13.053406404396016,13.097000064315445,13.140656743293697,13.183132711432926,13.223238555582576,13.25987458467748,13.292062349845674,13.318971237367462,13.339939348526126,13.354488170282954,13.362330847776203,13.363374172989348,13.357714680199207,13.345629465283249,13.327562504410444,13.304107327870735,13.275986906920282,13.244031546231266,13.209155461341455,13.172332584291086,13.134572007422845,13.096893368038147,13.06030241173927,13.025766957815147,12.994193524764475,12.966404948358067,12.943119422212252,12.924931491246209,12.91229561007365,12.905512921738211,12.904721902424791,12.909893446962766,12.920830838238407,12.93717485916177,12.958414083310789,12.98390013880891,13.012867499824218,13.044457140424694,13.077743202033501,13.111761689392168,13.145540126917236,13.178127079556381,13.208620468389153,13.236193687630891,13.260118650992398,13.279785054504538,13.29471533109008,13.30457497847191,13.309178153613615,13.30848862969384,13.302616391544351,13.291810290241925,13.276447278459383,13.257018801652226,13.234114928346488,13.208406773331596,13.18062771228826,13.151553819385727,13.121983894725458,13.092719398063535,13.06454457700278,13.038207074925209,13.014399324371109,12.993741068841013,12.976763400173992,12.963894737969438,12.955449200135012,12.951617809412971,12.952462942689568,12.957916355259018,12.967781002831677,12.98173674610755,12.999349865782104,13.020086151605433,13.043327169847581,13.068389170645586,13.094543979692244,13.121041134442375,13.147130477557845,13.172084411178798,13.195219044019945,13.215913526575171,13.2336269635763,13.247912411530885,13.258427605558639,13.26494220561507,13.26734149853163,13.265626629994298,13.25991156123933,13.250417042164425,13.237461961564993,13.2214524752517,13.202869326037968,13.182253761012467,13.160192428145454,13.137301604019486,13.114211074874419,13.09154797028402,13.069920836386236,13.049904234649555,13.032024160860418,13.016744593158599,13.004455491658634,12.995462578880272,12.989979223582107,12.988120725625542,12.989901253213477,12.995233615778455,13.003931968102483,13.015717438463223,13.030226562060689,13.04702228806079,13.065607221873728,13.085438670771941,13.105944986467911,13.126542647093894,13.146653495702957,13.165721553806172,13.183228855846593,13.19870980175815,13.211763596490764,13.222064433173141,13.22936917514701,13.233522395581952,13.234458735708165,13.232202638085223,13.226865594747279,13.218641117797356,13.207797690005236,13.194669985000814,13.179648662350758,13.163169045256097,13.145698981867326,13.127726179652882,13.109745289867874,13.092245009017379,13.075695457962896,13.060536097126022,13.047164436719674,13.035925801506266,13.027104406885615,13.020915993593471,13.017502248755438,13.016927209196565,13.019175797746284,13.024154585289198,13.031694802482058,13.041557548680029,13.05344106593898,13.066989867737439,13.08180544007113,13.09745817118287,13.113500119009112,13.129478195065268,13.144947331453645,13.159483204322946,13.172694111710472,13.18423164454132,13.193799844064998,13.20116260390308,13.206149146384181,13.208657476918523,13.20865579280061,13.206181890345755,13.20134067362106,13.194299917076654,13.185284472033274,13.17456913326098,13.16247039782496,13.149337355798538,13.13554195363289,13.121468868292974,13.107505225807282,13.094030393134,13.081406067935598,13.069966886842254,13.060011768154945,13.051796198179062,13.0455256596627,13.041350384318815,13.039361587650797,13.039589312396727,13.042001966782614,13.046507596240037,13.052956873966,13.061147739068646,13.070831553934088,13.081720597978299,13.093496666171527,13.10582050039004,13.118341751990904,13.13070915658561,13.142580597628374,13.153632744187036,13.163569969442458,13.172132288709312,13.179102097158614,13.184309535586365,13.187636364866048,13.189018283402037,13.188445674315895,13.18596281789069,13.181665648076228,13.175698168320844,13.16824767101021,13.15953892638167,13.149827521538914,13.139392539133675,13.128528769647806,13.117538652254275,13.106724138023072,13.0963786664885,13.086779442595116,13.078180195626006,13.070804594334382,13.064840482322628,13.060435083819343,13.05769131154977,13.05666528480282,13.057365136891653,13.059751157310606,13.063737275855562,13.069193855135683,13.075951715984766,13.083807279258751,13.092528669438723,13.101862592324254,13.111541772633615,13.12129271889609,13.130843573541032,13.13993180596577,13.148311515488823,13.155760128857938,13.162084302331694,13.167124869869188,13.170760714957806,13.172911482245507,13.173539084557845,13.172647999309094,13.17028438422429,13.166534074453768,13.161519550753171,13.155395990997746,13.148346534870193,13.140576904404579,13.132309531717166,13.12377735035306,13.115217408868823,13.106864465086119,13.098944717229587,13.091669824016977,13.085231359585169,13.079795840616617,13.075500451777332,13.07244958118265,13.070712259764854,13.070320577012744,13.071269120740151,13.073515460778875,13.076981666546631,13.081556817356523,13.087100433356543,13.093446725479783,13.10040953612246,13.107787819730191,13.115371495162284,13.122947490447105,13.130305795860645,13.137245343320089,13.143579538704572,13.14914128838246,13.153787381135983,13.15740211079028,13.15990005197801,13.161227930324145,13.161365557643322,13.160325831317087,13.158153823833278,13.154925012708082,13.150742722108081,13.145734865130342,13.140050089810108,13.133853442620184,13.127321670759411,13.12063828921199,13.113988540715768,13.107554376659555,13.101509584704514,13.096015184640255,13.091215207593125,13.087232965076014,13.084167903356986,13.082093125110973,13.081053644272362,13.081065421539694,13.082115207373485,13.084161197047775,13.087134479008656,13.090941234260255,13.095465621638406,13.100573262591993,13.106115220391661,13.111932353363809,13.117859910493532,13.123732231056255,13.129387408098088,13.134671778633143,13.139444111158088,13.143579373085966,13.146971976361037,13.149538418079288,13.151219253551034,13.151980361008572,13.151813479199793,13.150736020604814,13.148790183259745,13.146041402603156,13.14257620097295,13.138499506124358,13.133931521308432,13.12900423806562,13.123857689040925,13.118636041962155,13.113483637585757,13.108541074031416,13.103941437582025,13.099806775755301,13.096244902259357,13.093346615306473,13.091183400669024,13.08980567885095,13.089241641921092,13.089496710111177,13.09055362153407,13.092373150744038,13.09489543387222,13.098041860329532,13.101717474220843,13.10581381332402,13.110212100369715,13.114786690962235,13.119408675241484,13.123949526595872,13.128284690533807,13.13229701017985,13.13587989159232,13.138940121867675,13.141400265343536,13.143200577584457,13.144300392620327,13.14467895547283,13.144335688731756,13.143289898257994,13.141579938495935,13.139261871969657,13.13640767001045,13.133103012415672,13.129444752459065,13.125538120424984,13.121493743641613,13.11742456389381,13.113442734168181,13.109656575985143,13.106167676156176,13.103068197709938,13.100438473993268,13.098344947613121,13.096838507007828,13.095953263123123,13.095705797077887,13.096094897059931,13.09710178928513,13.09869085404364,13.10081080405944,13.103396289058157,13.106369878041289,13.109644359760766,13.113125292696443,13.116713728816684,13.12030903082005,13.123811700583156,13.127126137234242,13.130163246576647,13.132842829335035,13.135095683625531,13.136865366811879,13.138109573096838,13.138801095368978,13.138928352524145,13.13849547626644,13.137521963856367,13.136041915053642,13.134102882301212,13.13176437278193,13.129096049184508,13.126175682732923,13.123086917208287,13.119916906306411,13.11675388873882,13.11368476602506,13.11079274697668,13.108155120476775,13.105841214361979,13.103910593067873,13.102411540279395,13.101379865226633,13.100838062616964,13.100794846662398,13.101245069460301,13.102170023372405,13.103538116313292,13.105305898330483,13.107419407869601,13.109815797010526,13.11242518705076,13.115172699387534,13.1179806019412,13.120770508537152,13.123465567815638,13.125992579380814,13.128283977967394,13.130279631251824,13.131928403359437,13.13318944385487,13.134033170750593,13.134441925502081,13.134410287747873,13.133945047376578,13.13306484106385,13.131799469452433,13.130188919427809,13.128282123285402,13.12613549286182,13.123811271812146,13.12137575310303,13.118897411423672,13.116445001589454,13.114085674136042,13.111883158200046,13.1098960594923,13.108176317742215,13.106767863488733,13.105705508600378,13.105014098525805,13.104707947141797,13.104790567333923,13.105254702306718,13.106082654287052,13.107246898992614,13.10871096623416,13.110430559556995,13.11235488114761,13.114428122558763,13.116591077333238,13.118782828484722,13.12094246212532,13.123010758360843,13.124931811896845,13.126654537546695,13.128134019887451,13.129332671508456,13.13022117044212,13.130779154228765,13.130995655409926,13.130869270820112,13.130408064623616,13.129629212392032,13.12855840045046,13.127229001057431,13.125681049589783,13.123960054668736,13.122115676003789,13.120200307596717,13.11826760580442,13.116371002598619,13.114562244185205,13.112889993980488,13.111398536821381,13.110126618257636,13.109106448909218,13.108362899246968,13.10791290487696,13.107765096595324,13.107919663275666,13.108368449210158,13.109095281027157,13.11007651293192,13.111281772956266,13.112674887340603,13.114214955284867,13.115857542248069,13.117555956882832,13.119262574655481,13.120930170288084,13.122513221385223,13.123969146953328,13.125259446927991,13.12635071219176,13.127215478770532,13.127832904781428,13.128189254103665,13.12827817647337,13.128100779582013,13.127665494611016,13.126987742292734,13.126089411904816,13.124998170446913,13.123746623507957,13.122371352922276,13.120911859166155,13.119409438518984,13.117906026276998,13.116443037753195,13.115060238431033,13.113794673480259,13.112679684927603,13.111744042148182,13.111011208073114,13.110498759665091,13.110217976896452,13.110173609773248,13.110363828008088,13.110780352882118,13.111408765793827,13.112228983110734,13.113215882364104,13.114340060692072,13.115568702872238,13.116866533395921,13.118196824914776,13.11952243409877,13.120806835512838,13.122015124561642,13.123114961831448,13.124077433232912,13.12487780313369],"type":"scatter","name":"source2_θ","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1,0.1001,0.1002,0.1003,0.1004,0.1005,0.1006,0.1007,0.1008,0.1009,0.101,0.1011,0.1012,0.1013,0.1014,0.1015,0.1016,0.1017,0.1018,0.1019,0.102,0.1021,0.1022,0.1023,0.1024,0.1025,0.1026,0.1027,0.1028,0.1029,0.103,0.1031,0.1032,0.1033,0.1034,0.1035,0.1036,0.1037,0.1038,0.1039,0.104,0.1041,0.1042,0.1043,0.1044,0.1045,0.1046,0.1047,0.1048,0.1049,0.105,0.1051,0.1052,0.1053,0.1054,0.1055,0.1056,0.1057,0.1058,0.1059,0.106,0.1061,0.1062,0.1063,0.1064,0.1065,0.1066,0.1067,0.1068,0.1069,0.107,0.1071,0.1072,0.1073,0.1074,0.1075,0.1076,0.1077,0.1078,0.1079,0.108,0.1081,0.1082,0.1083,0.1084,0.1085,0.1086,0.1087,0.1088,0.1089,0.109,0.1091,0.1092,0.1093,0.1094,0.1095,0.1096,0.1097,0.1098,0.1099,0.11,0.1101,0.1102,0.1103,0.1104,0.1105,0.1106,0.1107,0.1108,0.1109,0.111,0.1111,0.1112,0.1113,0.1114,0.1115,0.1116,0.1117,0.1118,0.1119,0.112,0.1121,0.1122,0.1123,0.1124,0.1125,0.1126,0.1127,0.1128,0.1129,0.113,0.1131,0.1132,0.1133,0.1134,0.1135,0.1136,0.1137,0.1138,0.1139,0.114,0.1141,0.1142,0.1143,0.1144,0.1145,0.1146,0.1147,0.1148,0.1149,0.115,0.1151,0.1152,0.1153,0.1154,0.1155,0.1156,0.1157,0.1158,0.1159,0.116,0.1161,0.1162,0.1163,0.1164,0.1165,0.1166,0.1167,0.1168,0.1169,0.117,0.1171,0.1172,0.1173,0.1174,0.1175,0.1176,0.1177,0.1178,0.1179,0.118,0.1181,0.1182,0.1183,0.1184,0.1185,0.1186,0.1187,0.1188,0.1189,0.119,0.1191,0.1192,0.1193,0.1194,0.1195,0.1196,0.1197,0.1198,0.1199,0.12,0.1201,0.1202,0.1203,0.1204,0.1205,0.1206,0.1207,0.1208,0.1209,0.121,0.1211,0.1212,0.1213,0.1214,0.1215,0.1216,0.1217,0.1218,0.1219,0.122,0.1221,0.1222,0.1223,0.1224,0.1225,0.1226,0.1227,0.1228,0.1229,0.123,0.1231,0.1232,0.1233,0.1234,0.1235,0.1236,0.1237,0.1238,0.1239,0.124,0.1241,0.1242,0.1243,0.1244,0.1245,0.1246,0.1247,0.1248,0.1249,0.125,0.1251,0.1252,0.1253,0.1254,0.1255,0.1256,0.1257,0.1258,0.1259,0.126,0.1261,0.1262,0.1263,0.1264,0.1265,0.1266,0.1267,0.1268,0.1269,0.127,0.1271,0.1272,0.1273,0.1274,0.1275,0.1276,0.1277,0.1278,0.1279,0.128,0.1281,0.1282,0.1283,0.1284,0.1285,0.1286,0.1287,0.1288,0.1289,0.129,0.1291,0.1292,0.1293,0.1294,0.1295,0.1296,0.1297,0.1298,0.1299,0.13,0.1301,0.1302,0.1303,0.1304,0.1305,0.1306,0.1307,0.1308,0.1309,0.131,0.1311,0.1312,0.1313,0.1314,0.1315,0.1316,0.1317,0.1318,0.1319,0.132,0.1321,0.1322,0.1323,0.1324,0.1325,0.1326,0.1327,0.1328,0.1329,0.133,0.1331,0.1332,0.1333,0.1334,0.1335,0.1336,0.1337,0.1338,0.1339,0.134,0.1341,0.1342,0.1343,0.1344,0.1345,0.1346,0.1347,0.1348,0.1349,0.135,0.1351,0.1352,0.1353,0.1354,0.1355,0.1356,0.1357,0.1358,0.1359,0.136,0.1361,0.1362,0.1363,0.1364,0.1365,0.1366,0.1367,0.1368,0.1369,0.137,0.1371,0.1372,0.1373,0.1374,0.1375,0.1376,0.1377,0.1378,0.1379,0.138,0.1381,0.1382,0.1383,0.1384,0.1385,0.1386,0.1387,0.1388,0.1389,0.139,0.1391,0.1392,0.1393,0.1394,0.1395,0.1396,0.1397,0.1398,0.1399,0.14,0.1401,0.1402,0.1403,0.1404,0.1405,0.1406,0.1407,0.1408,0.1409,0.141,0.1411,0.1412,0.1413,0.1414,0.1415,0.1416,0.1417,0.1418,0.1419,0.142,0.1421,0.1422,0.1423,0.1424,0.1425,0.1426,0.1427,0.1428,0.1429,0.143,0.1431,0.1432,0.1433,0.1434,0.1435,0.1436,0.1437,0.1438,0.1439,0.144,0.1441,0.1442,0.1443,0.1444,0.1445,0.1446,0.1447,0.1448,0.1449,0.145,0.1451,0.1452,0.1453,0.1454,0.1455,0.1456,0.1457,0.1458,0.1459,0.146,0.1461,0.1462,0.1463,0.1464,0.1465,0.1466,0.1467,0.1468,0.1469,0.147,0.1471,0.1472,0.1473,0.1474,0.1475,0.1476,0.1477,0.1478,0.1479,0.148,0.1481,0.1482,0.1483,0.1484,0.1485,0.1486,0.1487,0.1488,0.1489,0.149,0.1491,0.1492,0.1493,0.1494,0.1495,0.1496,0.1497,0.1498,0.1499,0.15,0.1501,0.1502,0.1503,0.1504,0.1505,0.1506,0.1507,0.1508,0.1509,0.151,0.1511,0.1512,0.1513,0.1514,0.1515,0.1516,0.1517,0.1518,0.1519,0.152,0.1521,0.1522,0.1523,0.1524,0.1525,0.1526,0.1527,0.1528,0.1529,0.153,0.1531,0.1532,0.1533,0.1534,0.1535,0.1536,0.1537,0.1538,0.1539,0.154,0.1541,0.1542,0.1543,0.1544,0.1545,0.1546,0.1547,0.1548,0.1549,0.155,0.1551,0.1552,0.1553,0.1554,0.1555,0.1556,0.1557,0.1558,0.1559,0.156,0.1561,0.1562,0.1563,0.1564,0.1565,0.1566,0.1567,0.1568,0.1569,0.157,0.1571,0.1572,0.1573,0.1574,0.1575,0.1576,0.1577,0.1578,0.1579,0.158,0.1581,0.1582,0.1583,0.1584,0.1585,0.1586,0.1587,0.1588,0.1589,0.159,0.1591,0.1592,0.1593,0.1594,0.1595,0.1596,0.1597,0.1598,0.1599,0.16]}],{"xaxis":{"title":{"text":"Time in Seconds"}},"template":{"layout":{"coloraxis":{"colorbar":{"ticks":"","outlinewidth":0}},"xaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"hovermode":"closest","paper_bgcolor":"white","geo":{"showlakes":true,"showland":true,"landcolor":"#E5ECF6","bgcolor":"white","subunitcolor":"white","lakecolor":"white"},"colorscale":{"sequential":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]],"diverging":[[0,"#8e0152"],[0.1,"#c51b7d"],[0.2,"#de77ae"],[0.3,"#f1b6da"],[0.4,"#fde0ef"],[0.5,"#f7f7f7"],[0.6,"#e6f5d0"],[0.7,"#b8e186"],[0.8,"#7fbc41"],[0.9,"#4d9221"],[1,"#276419"]],"sequentialminus":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]},"yaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"shapedefaults":{"line":{"color":"#2a3f5f"}},"hoverlabel":{"align":"left"},"mapbox":{"style":"light"},"polar":{"angularaxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","radialaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"autotypenumbers":"strict","font":{"color":"#2a3f5f"},"ternary":{"baxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","caxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"aaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"annotationdefaults":{"arrowhead":0,"arrowwidth":1,"arrowcolor":"#2a3f5f"},"plot_bgcolor":"#E5ECF6","title":{"x":0.05},"scene":{"xaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"zaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"yaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"}},"colorway":["#636efa","#EF553B","#00cc96","#ab63fa","#FFA15A","#19d3f3","#FF6692","#B6E880","#FF97FF","#FECB52"]},"data":{"barpolar":[{"type":"barpolar","marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"carpet":[{"aaxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"},"type":"carpet","baxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"}}],"scatterpolar":[{"type":"scatterpolar","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"parcoords":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"parcoords"}],"scatter":[{"type":"scatter","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2dcontour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2dcontour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattercarpet":[{"type":"scattercarpet","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"mesh3d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"mesh3d"}],"surface":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"surface","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattermapbox":[{"type":"scattermapbox","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergeo":[{"type":"scattergeo","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram":[{"type":"histogram","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"pie":[{"type":"pie","automargin":true}],"choropleth":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"choropleth"}],"heatmapgl":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmapgl","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"bar":[{"type":"bar","error_y":{"color":"#2a3f5f"},"error_x":{"color":"#2a3f5f"},"marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"heatmap":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmap","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contourcarpet":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contourcarpet"}],"table":[{"type":"table","header":{"line":{"color":"white"},"fill":{"color":"#C8D4E3"}},"cells":{"line":{"color":"white"},"fill":{"color":"#EBF0F8"}}}],"scatter3d":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"scatter3d","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergl":[{"type":"scattergl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2d","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scatterternary":[{"type":"scatterternary","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scatterpolargl":[{"type":"scatterpolargl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}]}},"legend":{"yanchor":"bottom","xanchor":"right","y":1.02,"orientation":"h","x":1},"margin":{"l":100,"b":80,"pad":10,"r":80,"t":100},"plot_bgcolor":"#f1f3f7","yaxis":{"title":{"text":"State values"}},"yaxis2":{"titlefont":{"color":"orange"},"overlaying":"y","title":"Action values","side":"right"},"height":550,"width":800},{"showLink":false,"editable":false,"responsive":true,"staticPlot":false,"scrollZoom":true});
});





</script>
```




_______________________________________________________________________________
### Analysis

- The plot shows the instantaneous 3-phase real and imaginary power delivered by the inverter. 
- At 0.06 s the real and imaginary power set points take effect.
- After the transients have died away the inverter settles down to the specified power set points.
- The inverter's phase-locked loop frequency and angle are indicators of the control stability.
- The stability of the inverter decreases with the short-circuit power (i.e., fault level).

_______________________________________________________________________________
### References

- D. Weber, S. Heid, H. Bode, J. H. Lange, E. Hüllermeier and O. Wallscheid, "Safe Bayesian Optimization for Data-Driven Power Electronics Control Design in Microgrids: From Simulations to Real-World Experiments," in IEEE Access, vol. 9, pp. 35654-35669, 2021, doi: 10.1109/ACCESS.2021.3062144.
