# Swing Mode - Infinite Bus

|  | Mode | Description |
| --- | --- | --- |
| 1 | __"Swing"__ | __Ideal voltage source without dynamics (i.e., an infinite bus)__ |
| 2 | "PQ" | Grid following controllable source/load (real and imaginary power) |
| 3 | "Droop" | Simple grid forming with power balancing through a droop mechanism |
| 4 | "Synchronverter" or "VSG"|  Grid forming control mimicking a generator, i.e., virtual synchronous generator |

- There are a number of in-built "classically" control modes, which can be applied to the energy sources. 
- All sources of energy are modelled as voltage supplies, which are effective models (or averaged models) of power electronic converters.

## Summary

- This notebook is intended to introduce you to the simplest control mode available, i.e., "open-loop" control. 
- One source, an ideal oscillating voltage source, which is generating a 3-phase AC signal while connected to a static load through a cable.
- Plotting of the simulation results.


![](assets/Swing_mode.jpg)


```julia
using ElectricGrid;
```

_______________________________________________________________________________
### Network Configuration 

- We specify the control mode of the source as "Swing", often referred to as an infinite bus. 
- This is open-loop control of the source, where the output voltage magnitude, relative angle, and frequency at the source terminals are fixed.


```julia
# total run time, seconds
t_end = 0.1     

# Connectivity Matrix
CM = [ 0. 1.
        -1. 0.]     

parameters = Dict{Any, Any}(
        "source" => Any[
                        Dict{Any, Any}("pwr"       => 100e3,    # Power Rating [VAr]
                                        "mode"     => "Swing",  # Controller mode
                                        "v_pu_set" => 1.05,     # Voltage set point [per unit]
                                        "v_δ_set"  => 20.0),    # Angle set point [degrees]
                        ],
        "load"   => Any[
                        Dict{Any, Any}("impedance" => "RL", 
                                        "R"        => 3.73, 
                                        "L"        => 0.019),
                        ],
        "cable"   => Any[
                        Dict{Any, Any}("R"         => 0.1, 
                                        "L"        => 0.25e-3, 
                                        "C"        => 0.1e-4),
                        ],
        "grid"   => Dict{Any, Any}("f_grid"        => 60.0,     # Nominal grid frequency [Hz]
                                    "ramp_end"     => 0.04,     # Ramp up time to voltage set point [s]
                                    "v_rms"        => 230)      # Nominal grid voltage line-to-neutral [V]
    );
```

_______________________________________________________________________________
###  Defining the environment

To put all of the parameters together and construct the Linear Time Invariant (LTI) system of equations we run the following command.


```julia
env = ElectricGridEnv(CM = CM, parameters = parameters, t_end = t_end, verbosity = 2);
```

    ┌ Info: Normalization is done based on the defined parameter limits.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\electric_grid_env.jl:329
    ┌ Info: Time simulation run time: 0.1 [s] ~> 1001 steps
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\electric_grid_env.jl:330
    

_______________________________________________________________________________
### Initialising the Controller 

- All sources are controlled within the reinforcement learning framework, even if the control mode selected for the source does not entail any "learning" or "training" of the control structure. 
- If not specified by the user, for all the classically controlled "agents", the function SetupAgents computes all the necessary coefficients for control.


```julia
agent = SetupAgents(env);
```

    ┌ Info: 1 'classically' controlled source has been initialised.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2684
    ┌ Info: 1 source has been set up in Swing mode.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2695
    ┌ Info: All 'classically' controlled sources have been automatically set up with droop coeficients, and proportional and integral gains.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2707
    

_______________________________________________________________________________
### Running the Time Domain Simulation 

- The system is evolved for the specified amount of time, controller actions are computed, and the results are stored in a "hook". 
- Most relevant quantities are automatically collected, however, by passing a DataHook to the "simulate" function, signals may be specified.


```julia
hook = Simulate(agent, env);
```

_______________________________________________________________________________
### Low-Level Rendering

- The software solves the electrical network in the time domain using a set of linear time-invariant ODEs.
- For illustration reasons, the source voltage (action) and current through the filter inductor (state) of phase 'a' are shown below.


```julia
RenderHookResults(hook = hook, 
                    states_to_plot  = ["source1_i_L1_a"], # Inductor current [A]
                    actions_to_plot = ["source1_u_a"],    # Inverter voltage [V]
                    )
```


```@raw html
<div
id = J9WKeXBkZI4G > </div>
<script>
gd = 'J9WKeXBkZI4G'
require(['plotly'], function(plotly) {
























plotly.newPlot(gd,[{"mode":"lines","y":[0.0,0.0,0.08130092450894029,0.20643687604658917,0.3555111067129418,0.5196848081519435,0.6883562123180967,0.8590790841950058,1.027147526799587,1.1910383329849514,1.3487322139730842,1.4989216671587668,1.6405026112972276,1.7723873823272611,1.8937169999331822,2.0035883431848207,2.101217284034724,2.185823287835147,2.2566786839998128,2.3130870084398625,2.3543851216912857,2.3799482414361313,2.3891849016799904,2.3815430536423357,2.3565080652467367,2.3136058646103366,2.252403424200228,2.1725102646382832,2.0735796934486492,1.9553098094567012,1.817444681217141,1.6597752162808708,1.482140093300191,1.284426511803792,1.06657088710928,0.8285594327743683,0.5704286425460781,0.29226567540228887,-0.005791368254645934,-0.3235532702992511,-0.6607796264223295,-1.017178881858463,-1.3924084737878928,-1.7860750798439422,-2.1977349727909767,-2.626894481356691,-3.0730105565627137,-3.535491443140366,-4.01369745508981,-4.50694185444562,-5.014491832011373,-5.535569588671664,-6.069353515690088,-6.614979472204065,-7.171542157950088,-7.738096579059212,-8.313659604587293,-8.897211611263325,-9.48769821376704,-10.08403207767716,-10.68509481206674,-11.289738938562628,-11.896789933530991,-12.505048339902196,-13.1132919450047,-13.720278020640555,-14.324745621504068,-14.925417937920846,-15.521004698766799,-16.110204620316143,-16.691707896664056,-17.264198727273744,-17.826357877109356,-18.376865264735674,-18.914402573693035,-19.437655882391304,-19.945318307710796,-20.436092657449937,-20.90869408672033,-21.361852753358928,-21.794316467404975,-22.204853329675938,-22.59225435447192,-22.95533607144226,-23.29294310166099,-23.603950702979407,-23.887267279754965,-24.14183685209462,-24.366641479799352,-24.56070363625298,-24.723088527564208,-24.852906352344522,-24.94931449758723,-25.01151966620346,-25.038779931870064,-25.03040671695129,-24.985766689370955,-24.904283574434544,-24.785439877730774,-24.628778515379643,-24.433904348038638,-24.200485615230406,-23.92825526671315,-23.617012187779984,-23.266622315543856,-22.877019643441844,-22.448207111374373,-21.98025737908293,-21.47331348056221,-20.927589357500153,-20.343370269940827,-19.721013082571005,-19.06094642524075,-18.36367072654109,-17.6297581194778,-16.85985221849876,-16.05466776735333,-15.214990157484724,-14.341674816880897,-13.435646469534822,-12.497898265891443,-11.529490784885231,-10.53155090839913,-9.505270569202038,-8.45190537364771,-7.372773100642549,-6.269252078612761,-5.142779442422636,-3.9948492724145637,-2.827010617957677,-1.6408654081052916,-0.43806625217111117,0.7796858667597388,2.0106440021668037,3.253017754862086,4.5049758805814015,5.764648990603785,7.030132341421526,8.299488709307292,9.570751345454964,10.841927007207389,12.1109990607276,13.375930650320557,14.63466792947018,15.88514334852199,17.125278993814977,18.352989972947576,19.56618784075277,20.7627840604555,21.940693494392615,23.09783791859234,24.232149555434727,25.341574618549974,26.424076864055095,27.477641142183476,28.500276943325108,29.49002193246871,30.44494546602004,31.36315208496405,32.24278497834152,33.082029411024486,33.87911610979782,34.63232460178806,35.33998649932432,36.000488725369394,36.61227667372347,37.17385729827605,37.68380212566552,38.14075018579915,38.543410854789286,38.89056660497411,39.181075656813576,39.41387452758187,39.587980471918875,39.70249380945157,39.75660013485471,39.749572405886475,39.680772905109045,39.54965507118699,39.35576519584666,39.09874398277709,38.7783279649583,38.39435077711423,37.946744280206005,37.435539535105484,36.8608676228194,36.222960308870334,35.522150549681356,34.75887283905745,33.93366339310616,33.04716017219496,32.100102738799784,31.093331950360543,30.027789486523186,28.90451721041372,27.724656363858124,26.48944659673134,25.200224830889493,23.858423959410782,22.46557138214227,21.023287378820996,19.533283321308872,17.99735972675037,16.41740415373018,14.795388943774267,13.133368810801493,11.433478281393926,9.697928989011558,7.9290068255311725,6.129068953738784,4.3005406846503895,2.4459122237759985,0.5677352906767119,-1.3313803836062403,-3.248770676448691,-5.181720805849034,-7.127469169714393,-9.083211286798264,-11.046103833665534,-13.01326877187502,-14.981797559391143,-16.94875544006761,-18.911185804885132,-20.86611461847476,-22.810554904316557,-24.741511281872004,-26.655984548786986,-28.550976301191067,-30.42349358501828,-32.27055357118481,-34.08918824738011,-35.87644911916046,-37.62941191297748,-39.345181273729246,-41.020895449388306,-42.653730955239396,-44.24090721024926,-45.77969113809355,-47.26740172537917,-48.701414529626106,-50.079166129610904,-51.398158510722844,-52.655963378046046,-53.850226389953505,-54.9786713050845,-56.0391040356735,-57.029416600307336,-57.94759096930695,-58.79170279606182,-59.5599250277872,-60.25053138932822,-60.8618997337988,-61.39251525401872,-61.840973548897246,-62.205983539107336,-62.48637022659975,-62.68107729272159,-62.78916952992751,-62.80983510230615,-62.742387630385004,-62.58626809592848,-62.341046562701145,-62.00642370943459,-61.58223217150948,-61.0684376881435,-60.46514005216333,-59.77257385973002,-58.99110905768564,-58.12125128649194,-57.1636420170389,-56.11905847991374,-54.988413386035575,-53.77275443788054,-52.47326363084311,-51.091256344603686,-49.628180224697765,-48.08561385480946,-46.46526522063901,-44.76896996652273,-42.998689446311104,-41.156508570337394,-39.244633450635774,-37.26538884689072,-35.22121541592182,-33.11466676782653,-30.948406332219047,-28.725204038315066,-26.447932812919284,-24.11956490067539,-21.74316801123506,-19.32190129829402,-16.85901117572827,-14.357826976341897,-11.821756459008963,-9.2542811702554,-6.658951666582251,-4.039382604078208,-1.399247702107428,1.2577254119131451,3.9277604694138937,6.6070379318085415,9.291700458226508,11.977858462014876,14.661595748007297,17.3389752223927,20.006044666864113,22.658842568587385,25.293403997401498,27.905766521546674,30.49197615311395,33.048093314320276,35.57019881563704,38.05439983673712,40.4968359011762,42.89368483568911,45.241168704959726,47.53555971271588,49.77318606000678,51.95043775154102,54.06377234099765,56.10972060627172,58.084892145678054,59.98598088621404,61.809770495072854,63.55313968570359,65.21306740983232,66.78663792699169,68.27104574325092,69.66360041099819,70.96173118179937,72.16299150454188,73.26506336127142,74.26576143333837,75.16303709069517,75.954982197419,76.63983272678121,77.2159721794426,77.68193479862175,78.0364085763627,78.27823804531826,78.40642685076348,78.42014009786288,78.31870646953213,78.10162011056121,77.76854227399954,77.31930272614562,76.75390090683204,76.07250684205205,75.27546180633475,74.3632787326435,73.33664236794293,72.196409172957,70.94360696501988,69.5794343033054,68.10525961610682,66.52262007022664,64.83322018292658,63.0389301772786,61.14178408214904,59.14397757843918,57.04786559359612,54.85595964679649,52.57092494759285,50.19557725119706,47.73287947395589,45.18593807295231,42.55799919403833,39.852444592974265,37.072787334711556,34.22266727621324,31.305846338556044,28.3262035744011,25.287730037255166,22.194523459271203,19.05078274465499,15.860802286053298,12.628966111597695,9.359741870566696,6.057674665906633,2.7273807421178553,-0.6264589627323609,-3.9991053918662818,-7.385768062287756,-10.78161186545923,-14.181763966509067,-17.581320792125208,-20.975355097114594,-24.35892309944711,-27.727071673455104,-31.074845590728412,-34.397294798127675,-37.6894817222374,-40.946488589495274,-44.16342475116351,-47.33543400225532,-50.45770188349094,-53.525462955337304,-56.53400803317984,-59.47869137268642,-62.35493779445197,-65.15824973705581,-67.88421422772569,-70.44431543680307,-72.8689918841477,-75.17238196654513,-77.35822416371224,-79.43260118649226,-81.393773835394,-83.24257066091188,-84.97687490784945,-86.59528350884918,-88.09588843171142,-89.47677451322068,-90.7362093216668,-91.87242866338802,-92.88391349258758,-93.76922877159689,-94.5271423784268,-95.15657902603925,-95.65664494606874,-96.02662879766079,-96.26599921784508,-96.37441230619477,-96.35170757342779,-96.19791171635904,-95.91323690107632,-95.49808137140599,-94.95302869911978,-94.27884694935946,-93.4764877368051,-92.54708475607856,-91.49195227300902,-90.31258320240805,-89.01064701893127,-87.58798737440505,-86.04661948072845,-84.38872724875225,-82.61666018207102,-80.73293004103586,-78.74020727173243,-76.64131721219704,-74.43923607702017,-72.13708672814575,-69.73813423712676,-67.24578124521756,-64.6635631280979,-61.99514297185468,-59.24430636756185,-56.41495603172453,-53.51110626030492,-50.53687722417883,-47.496489114138804,-44.394256143765745,-41.23458041868602,-38.02194568093158,-34.76091093728863,-31.456103980691857,-28.112214813869432,-24.73398898458455,-21.32622084194446,-17.89374672336013,-14.441438081838223,-10.974194563371904,-7.496937044267886,-4.0146006383040245,-0.5321276836547426,2.9455392804498466,6.413464537350501,9.866726206890144,13.300423239682745,16.709682381780514,20.089665099842257,23.43557445696181,26.74266192938569,30.006234154433017,33.221659600028424,36.38437514637043,39.48989257038209,42.53380492372936,45.51179279534362,48.41963044954882,51.253191831070964,54.00845642839605,56.681514987144716,59.26857506534432,61.76596642270424,64.1701462362358,66.4777041348053,68.68536704546577,70.79000384468027,72.78862980782756,74.67841085066664,76.4566675567333,78.12087898494508,79.66868625200419,81.0978958845073,82.40648293599891,83.59259386453874,84.65454916669376,85.59084576421205,86.40015913998693,87.08134522027567,87.63344200049859,88.05567091230762,88.34743792998218,88.50833441457972,88.53813769464152,88.43681138262863,88.20450542663792,87.84155589732534,87.34848451033838,86.72599788493724,85.97498653985767,85.09652362784246,84.09186341064031,82.96243947663856,81.70986270366316,80.33591896984052,78.8425666157745,77.23193366164455,75.50631478318007,73.66816805080809,71.72011143660794,69.66491909403717,67.50551741571506,65.24498087486594,62.88652765633179,60.43351508336298,57.88943484668543,55.25790804262446,52.54268002733604,49.74761509445811,46.87669098374516,43.93399322848971,40.923709349763165,37.850122905726636,34.717607404467614,31.53062008901293,28.293695603349814,25.01143954845589,21.688521937495416,18.329670559482317,14.939664260840482,11.523326154408739,8.085516765540834,4.631127125040118,1.1650718187443374,-2.307717996362534,-5.782301599587074,-9.253735727553615,-12.717081598844905,-16.167411932273527,-19.59981794881098,-23.009416347224082,-26.391356243506184,-29.740826064241535,-33.05306038410739,-36.32334669779697,-39.5470321167403,-42.719529981106334,-45.8363263776905,-48.89298655442527,-51.885161222398466,-54.80859273642368,-57.65912114538003,-60.43269010372332,-63.12535263576791,-65.73327674454816,-68.25275085728876,-70.68018909974575,-73.01213639192295,-75.24527335792277,-77.37642104295402,-79.40254543079354,-81.32076175528236,-83.12833859972932,-84.82270177839659,-86.40143799455164,-87.86229826988689,-89.20320114043408,-90.42223561443149,-91.5176638879409,-92.48792381435524,-93.33163112428733,-94.04758139268566,-94.63475175038133,-95.09230233763431,-95.41957749761312,-95.61610670811223,-95.68160525018304,-95.6159746127282,-95.41930263248439,-95.09186336919478,-94.63411671614821,-94.04670774663886,-93.33046579727505,-92.48640328943965,-91.51571429057715,-90.41977281735173,-89.2001308830875,-87.85851629226595,-86.39683018521445,-84.81714433647515,-83.12169821069301,-81.3128957802075,-79.39330210886949,-77.36563970693844,-75.23278466223968,-72.99776255307948,-70.66374414872602,-68.23404090356608,-65.71210025134044,-63.101500706145245,-60.40594677716091,-57.62926370433533,-54.77539202250179,-51.8483819616568,-48.852387691354956,-45.79166141740004,-42.67054733922133,-39.493475476521795,-36.264955373971105,-32.98956969288934,-29.67196769902781,-26.316858655701477,-22.929005131661512,-19.51321623321789,-16.074340770229632,-12.617260365674092,-9.146882518587029,-5.668133630231411,-2.185952003405097,1.2947191751641949,4.768938859364219,8.23177516520165,11.678312379285599,15.103657944331815,18.502949411760838,21.871361351509957,25.204112209239867,28.49647110119151,31.743764537037226,34.94138306117271,38.08478780301207,41.169516926977785,44.19119197301964,47.14552407865246,50.0283200736708,52.835488438879686,55.5630451203737,58.20711919110197,60.763958351673715,63.2299342625874,65.60154770030609,67.87543352985197,70.0483654868536,72.1172607622505,74.07918438313989,75.93135338354018,77.6711407591441,79.29607920044208,80.80386459891025,82.19235932128115,83.45959524724383,84.60377656625684,85.62328232949949,86.51666875333531,87.28267127101364,87.9202063296946,88.42837293024323,88.80645390760478,89.0539169499418,89.17041535508453,89.15578852322004,89.01006218511948,88.73344836557962,88.32634508212992,87.78933577943452,87.12318850019118,86.32885479370712,85.4074683637014,84.36034345725616,83.18897299720577,81.89502646061796,80.48034750638169,78.9469513552732,77.29702192622311,75.53290873285432,73.65712354470072,71.67233681785127,69.58137390009209,67.38721101593853,65.09297103726362,62.70191904553222,60.21745769194776,57.643122362104705,54.98257615201836,52.23960466267098,49.41811062047174,46.52210833127503,43.55571797583818,40.52315975482525,37.428747891677006,34.276884501869326,31.072053337272315,27.818813414499544,24.521792536301994,21.185680715213138,17.815223508790492,14.415215275924652,10.990492363798953,7.54592623518147,4.086416545815844,0.6168841817482935,-2.857735733514696,-6.332500852081017,-9.802468625013754,-13.262703325957759,-16.708283065070912,-20.134306783293034,-23.535901217012984,-26.90822782323709,-30.246489655417523,-33.54593818016997,-36.80188002519348,-40.009683648803474,-43.16478592159997,-46.26269861091832,-49.29901475884752,-52.26941494475272,-55.169673423402415,-57.99566412997781,-60.74336654343086,-63.408871399858825,-65.98838624777636,-68.47824083739084,-70.87489233622222,-73.17493036365607,-75.37508183727533,-77.47221562408366,-79.46334699001088,-81.34564184137778,-83.11642075229248,-84.77316277225539,-86.3135090085622,-87.73526597841361,-89.03640872596867,-90.21508369991219,-91.26961138744667,-92.19848870096638,-93.00039111402116,-93.67417454353541,-94.21887697560693,-94.63371983257566,-94.91810907941944,-95.07163606790459,-95.09407811729172,-94.9853988307715,-94.74574814718098,-94.37546212792651,-93.87506247941648,-93.24525581168234,-92.48693263424134,-91.60116609062824,-90.58921043339392,-89.45249924173828,-88.1926433843101,-86.811428730069,-85.31081361046226,-83.69292603652298,-81.9600606748446,-80.11467558672824,-78.15938873513774,-76.09697426442531,-73.93035855811513,-71.66261608034651,-69.29696500688678,-66.8367626519217,-64.28550069712283,-61.64680022977124,-58.92440659698883,-56.122184083389584,-53.24411041971415,-50.29427113025128,-47.27685372707882,-44.19614175937429,-41.05650872625173,-37.86241186177455,-34.61838580097664,-31.329036135892316,-27.999032870752565,-24.633103785648053,-21.236027718089417,-17.81262777201224,-14.367764463876988,-10.906328815603718,-7.433235404156859,-3.9534153776569934,-0.4718094479441337,3.0066391304495887,6.476991584940485,9.934320643164842,13.373717534907927,16.79029896774047,20.179214066449816,23.53565126640443,26.85484515105584,30.132083223861745,33.36271260500679,36.54214664340476,39.6658714345861,42.7294522352085,45.72853976507518,48.65887638770542,51.516302160674215,54.2967607471235,56.99630518004402,59.61110347113682,62.13744405628353,64.57174106988738,66.91053944058979,69.15051980112099,71.28850320530796,73.32145564553603,75.2464923642451,77.06088195333321,78.76205023564209,80.34758392300918,81.81523404568719,83.16291914825865,84.38872824750297,85.49092354801353,86.46794291170518,87.31840207770323,88.04109662945902,88.63500370629708,89.09928345696115,89.43328023309408,89.63652352095517,89.7087286100511,89.64979699773038,89.45981652916606,89.13906127252818,88.68799112952304,88.10725118185283,87.3976707745244,86.56026233730985,85.59621994603339,84.5069176257295,83.2939073980833,81.95891707592786,80.5038478079328,78.93077137697303,77.24192725601712,75.43971942571868,73.5267129582334,71.50563037211634,69.37934776347905,67.15089071890472,64.82343001592922,62.40027711719769,59.88487946470005,57.28081558077228,54.591789982825475,51.82162791902887,48.97426993242798,46.05376626122225,43.06427108316005,40.01003661222974,36.895407056035985,33.724812442447984,30.5027623242925,27.233839371037497,23.922692856573086,20.574032052343874,17.192619535221574,13.783264419627576,10.350815523523243,6.9001544779791555,3.436188790115258,-0.03615513073038912,-3.511938973693905,-6.986219543990554,-10.454055787460156,-13.910515812070436,-17.350683896400596,-20.76966747514547,-24.162604091713696,-27.524668308039633,-30.85107856179001,-34.137103961220795,-37.37807100802831,-40.569370238641156,-43.70646277451528,-46.784886772123976,-49.8002637634767,-52.748304878156844,-55.624816938036496,-58.42570841600708,-61.146995250258406,-63.784806505843704,-66.33538987548522,-68.79511701180321,-71.1604886833914,-73.42813974741148,-75.59484393164024,-77.65751841917441,-79.61322822927707,-81.45919038814132,-83.19277788364337,-84.81152339846587,-86.31312281628593,-87.69543849604585,-88.95650230965313,-90.09451843879354,-91.10786592688234,-91.99510098252765,-92.75495903123212,-93.38635651241795,-93.88839241922082,-94.26034957886537,-94.50169567180211,-94.61208398815805,-94.5913539204255,-94.43953119168962,-94.15682781906966,-93.74364181242635,-93.20055660876348,-92.5283402431273,-91.7279442571819,-90.80050234701189,-89.74732875207349,-88.5699163875837,-87.26993472300143,-85.84922740961537,-84.30980966060997,-82.65386538733262,-80.8837440958315,-79.00195754807419,-77.01117619259196,-74.91422536962156,-72.71408129613764,-70.41386683648076,-68.01684706459191,-65.52642462415903,-62.946134893269594,-60.27964096044022,-57.53072841916299,-54.70329998836526,-51.801369966427984,-48.82905852664348,-45.79058586221912,-42.69026618914736,-39.53250161546415,-36.32177588560799,-33.06264800876916,-29.759745780283716,-26.41775920527849,-23.041433833912677,-19.635564017686853,-16.204986096402628,-12.754571525454606,-9.28921995322124,-5.81385225839184,-2.3334035571242087,1.1471838100299685,4.622967301465987,8.089011198152663,11.540393617563655,14.972213507946309,18.379597612988874,21.75770739698884,25.10174592068138,28.406964657957005,31.66867024378142,34.88223114372857,38.04308423564884,41.14674129411984,44.1887953684647,47.16492704527457,50.07091058653583,52.90261993463952,55.65603457573929,58.32724525312605,60.91245952250006,63.40800714124617,65.81034528405381,68.11606357747003,70.32188894623111,72.42469026448606,74.42148280530212,76.30943248212942,78.0858598761974,79.7482440441199,81.29422610029793,82.72161256902965,84.02837850156355,85.21267035366596,86.27280861961297,87.20729021886446,88.01479063202798],"type":"scatter","name":"source1_i_L1_a","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1]},{"mode":"lines","y":[0.0,0.7966923046297512,1.5691032677883026,2.313888217071141,3.0277778275868448,3.707585909720083,4.3502170230049995,4.952673900903702,5.512064671638755,6.025609860609919,6.490649160333377,6.904647954275382,7.265203581411084,7.570051328822376,7.817070140155087,8.004288028284781,8.129887181091044,8.192208749811323,8.189757310036152,8.12120498601702,7.985395229585021,7.781346245621594,7.5082540566815,7.1654952000404695,6.752629051125942,6.269399767986335,5.715737852162355,5.091761322040728,4.3977764954960055,3.6342783793576423,2.8019506639769647,1.9016653219098087,0.9344818104747715,-0.0983541213073501,-1.195412020141597,-2.3550787082583775,-3.575560652846336,-4.854886721642011,-6.190911320466462,-7.58131790777214,-9.023622880544051,-10.515179825189115,-12.053184126347581,-13.634677925871921,-15.256555423542844,-16.91556851043039,-18.608332725161123,-20.331333522722627,-22.080932844822748,-23.853375980227728,-25.64479870292759,-27.4512346754239,-29.26862310390217,-31.092816631541815,-32.9195894557302,-34.74464565448617,-36.56362770696167,-38.372125192480866,-40.165683652192236,-41.93981359705465,-43.689999645550934,-45.411709774224256,-47.10040466386536,-48.751547123939,-50.36061157763143,-51.92309358972365,-53.43451941935011,-54.8904555795891,-56.28651838574975,-57.61838347417121,-58.88179527333306,-60.0725764090917,-61.18663702590524,-62.21998400599139,-63.16873006847512,-64.02910273072884,-64.79745311428655,-65.47026457792224,-66.04416116072478,-66.51591581827547,-66.8824584353365,-67.14088359879473,-67.28845811496893,-67.32262825578235,-67.24102671872603,-67.04147928598846,-66.72201116860641,-66.28085302199678,-65.71644661976015,-65.0274501732041,-64.21274328461384,-63.27143152290201,-62.20285061089616,-61.00657021416873,-59.682397321983196,-58.23037921161613,-56.65080598802024,-54.94421269151447,-53.11138096692575,-51.1533402883572,-49.07136873452421,-46.86699331037524,-44.541989811502994,-42.09838222864662,-39.53844169039306,-36.86468494299309,-34.079872367028614,-31.187005531483496,-28.18932428659725,-25.090303397701096,-21.89364872306169,-18.603292939578367,-15.223390820998322,-11.758314074127945,-8.212645739326145,-4.591174162366067,-0.898886545542607,2.8590379133152,6.677235284498472,10.550164562554183,14.472115703974579,18.437218034785857,22.439449015671592,26.472643351574753,30.530502432049694,34.606604087983705,38.69441264967807,42.78728929067102,46.87850264110283,50.96123965386469,55.02861670624228,59.07369091925999,63.089471676455645,67.06893232336886,71.00502202860827,74.89067778697702,78.71883654478074,82.48244742711934,86.1744840466746,89.78795687324934,93.31592564309113,96.75151178684717,100.08791085484306,103.31840491826287,106.43637492472561,109.43531298670761,112.30883458125358,115.05069063944453,117.65477950415568,120.11515873473971,122.426056737405,124.58188420023583,126.57724531200894,128.4069487442093,130.0660183759284,131.54970374164787,132.85349018226287,133.97310868008827,134.90454535901227,135.64405063141803,136.18814797398193,136.5336423149788,136.67762801627762,136.61749643379252,136.35094304077072,135.87597409893925,135.1909128632044,134.29440530629526,133.1854253504678,131.86327959413427,130.3276115220566,128.57840518853834,126.61598836386689,124.44103513509552,122.0545679531119,119.45795911881376,116.65293170210316,113.64155988831601,110.42626874762287,107.009833423868,103.39537774025453,99.58637222023458,95.58663152292108,91.40031129330153,87.03190442850259,82.48623676232532,77.7684621712427,72.88405710602458,67.83881455412231,62.63883743891541,57.290531462880324,51.80059740269895,46.17602286526806,40.42407351450987,34.55228377980593,28.568447057791012,22.480605420138467,16.297038840851616,10.026253957437559,3.676972381185285,-2.7418814274069683,-9.22119270120804,-15.751669309408529,-22.32385508247486,-28.928143499996278,-35.55479172161807,-42.193934940583226,-48.8356010387623,-55.46972552144197,-62.08616670956298,-68.67472116655406,-75.22513933639705,-81.72714136908309,-88.17043310917917,-94.54472222282497,-100.83973443811036,-107.0452298734642,-113.1510194283918,-119.14698121065864,-125.02307697380468,-130.7693685387162,-136.37603417284996,-141.83338490063173,-147.13188071850362,-152.26214668810363,-157.21498888110204,-161.98141014931082,-166.55262569381128,-170.92007840701928,-175.07545396182508,-179.01069562220155,-182.7180187499793,-186.18992498282665,-189.4192160588606,-192.39900726373986,-195.1227404765569,-197.5841967913555,-199.7775086916447,-201.69717175586646,-203.33805587239766,-204.695415943328,-205.76490205695177,-206.5425691096449,-207.02488585856443,-207.2087433874093,-207.0914629683116,-206.67080330379343,-205.94496713361502,-204.91260719226378,-203.57283150378143,-201.9252080016011,-199.96976846206653,-197.707011741325,-195.13790630633167,-192.2638920517633,-189.0868813957228,-185.6092596482118,-181.83388464746295,-177.7640856603482,-173.40366154421704,-168.75687816866596,-163.82846509689583,-158.6236115274734,-153.14796149848144,-147.40760835720818,-141.40908849969574,-135.1593743856364,-128.66586683526975,-121.9363866160915,-114.97916532834402,-107.80283559939538,-100.41642059825735,-92.82932288260521,-85.05131259178238,-77.09251500035205,-68.96339744784711,-60.67475566141307,-52.23769948908172,-43.66363806242176,-34.96426440829987,-26.15153953044863,-17.237675982469675,-8.235120954805748,0.8434611009117305,9.985206285891941,19.177069481535867,28.405842964282634,37.65817538138879,46.92059106038092,56.179509624263765,65.42126588394795,74.63212997878487,83.79832773556299,92.9060612158286,101.94152942094907,110.89094912393622,119.74057579669272,128.47672460103814,137.08579141161155,145.5542738385361,153.86879221756914,162.0161105353485,169.98315725728148,177.75704602561035,185.32509619522523,192.67485317488112,199.79410854161532,206.67091989634753,213.29363042888602,219.65088816084864,225.7316648353494,231.52527442268712,237.02139121171223,242.2100674570328,247.08175055275666,251.62729970404766,255.83800206840564,259.7055883392518,263.22224774512546,266.380642438561,269.1739212495238,271.59573277913387,273.6402378102991,275.3021210128117,276.57660192143175,277.4594451664915,277.94696993759726,278.03605866208454,277.7241648809955,277.00932030648755,275.89014104576137,274.36583297778884,272.4361962703596,270.1016290262084,267.36313004826457,264.2223007153626,260.681345961063,256.7430743495695,252.41089724407416,247.6888270642219,242.5814746307581,237.09404559680297,231.2323359665863,225.0027267038634,218.41217743363407,211.46821924217804,204.17894658181575,196.55300828819378,188.59959771927964,180.32844202662494,171.74979057082646,162.87440249446388,153.7135334671387,144.27892161855658,134.58277267690548,124.63774433106238,114.45692983642594,104.05384088540698,93.4423897648217,82.63687082361103,71.65194127546455,60.50260136204205,49.20417390357374,37.77228326466634,26.22283376415404,14.571987558803912,2.836142031615712,-8.968093283656879,-20.823920209293366,-32.71437582843423,-44.62235675393146,-56.53064372650189,-68.42192650681152,-80.27882902551666,-92.08393475473987,-103.81981226396029,-115.46904092285786,-127.01423671326064,-138.43807811201538,-149.72333200632391,-160.85287960287022,-171.80974229190147,-182.57710742732635,-193.13835398384404,-203.47707805213838,-213.57711813323772,-223.4225801932815,-232.99786244011722,-242.2876797834099,-251.27708794025025,-259.9515071486165,-268.2967454514749,-276.2990215147827,-283.94498694320237,-291.2217480579339,-298.1168871017245,-304.61848283682826,-310.7151305024484,-316.395961099016,-321.65065996752696,-326.46948463308223,-330.01823629201726,-333.09801397897246,-335.70444116133274,-337.83381396345135,-339.48310643005976,-340.64997482631554,-341.33276096837795,-341.5304945797779,-341.2428946702338,-340.4703699349543,-339.2140181738598,-337.4756247315489,-335.2576599602261,-332.56327570919564,-329.396300845912,-325.76123581494886,-321.6632462426218,-317.1081555963502,-312.1024369091916,-306.6532035813076,-300.7681992714326,-294.45578689271144,-287.72493672854216,-280.58521368531217,-273.0467637001431,-265.12029932295906,-256.8170844933667,-248.14891853398103,-239.1281193829423,-229.76750608945147,-220.08038059720025,-210.08050884158163,-199.78210118754097,-189.1997922358719,-178.34862002664582,-167.2440046693358,-155.90172642999556,-144.33790330663982,-132.56896812468509,-120.61164518500777,-108.4829264977952,-96.20004763597136,-83.78046324250356,-71.24182222639996,-58.60194268264408,-45.878786571706925,-33.09043419461823,-20.25505849986943,-7.390899258659382,5.48376285481852,18.350632241214683,31.191424375071115,43.98789178813258,56.72184999999751,69.37520335925966,81.92997075841892,94.36831118601901,106.67254907970103,118.82519944414457,130.80899269820316,142.60689921592365,154.2021535265768,165.57827813930862,176.7191069585562,187.60880825695443,198.23190717308663,208.5733077021095,218.61831414800207,228.35265200695378,237.76248825221572,246.83445099158797,255.55564846961101,263.91368738745473,271.896690514475,279.4933135664076,286.6927613262177,293.48480298469315,299.8597866789848,305.80865320843145,311.3229489081798,316.3948376623048,321.01711203935895,325.18320353452714,328.8871919038311,332.12381357712013,334.8884691378924,337.1772298593174,338.9868432871716,340.3147378617541,341.1590265722137,341.5185096380943,341.39267621428877,340.78170511697783,339.68646456952234,338.10851096867054,336.050086672833,333.5141168155687,330.50420514880994,327.0246289217342,323.08033280255825,318.67692185189486,313.8206535576544,308.51842894281214,302.7777827586773,296.60687277759894,290.0144682003251,283.00993719448866,275.60323358192863,267.8048826937645,259.625966413325,251.0781074281852,242.17345271369183,232.9246562714464,223.34486114727653,213.44768075424872,203.24717952726368,192.75785293672425,181.99460688967912,170.97273654771337,159.70790459168674,148.2161189642069,136.51371012146765,124.61730782677638,112.54381751875111,100.31039628776657,87.9344284947905,75.43350106725497,62.825378507070596,50.127977646297,37.35934218634414,24.537617056885324,11.68102263091918,-1.19217116737729,-14.063670825210718,-26.915185237254025,-39.72845169830798,-52.485261855604584,-65.1674875838727,-77.757106746396,-90.23622880545543,-102.5871202457622,-114.79222977475345,-126.83421326393922,-138.69595839585818,-150.36060898161733,-161.81158891445818,-173.0326257253123,-184.00777370686922,-194.7214365733003,-205.15838962343312,-215.30380137588577,-225.14325464541292,-234.66276703051548,-243.84881078319754,-252.68833203263705,-261.1687693354505,-269.278071526191,-277.00471484271407,-284.33771930207405,-291.2666643036809,-297.7817034375436,-303.8735784765577,-309.53363253295254,-314.75382236020243,-319.5267297829203,-323.8455722384905,-327.70421241546177,-331.0971669750023,-334.0196143430246,-336.4674015619059,-338.43705019206794,-339.9257612550306,-340.93141921091274,-341.45259496473074,-341.4885478972202,-341.0392269172969,-340.1052705346595,-338.6880059524328,-336.7894471811391,-334.41229217667933,-331.5599190063901,-328.23638104862465,-324.44640123268124,-320.1953653272614,-315.48931428699854,-310.3349356679304,-304.7395541241161,-298.71112099890223,-292.2582030256284,-285.3899701538307,-278.1161825182416,-270.4471765691043,-262.3938503835115,-253.96764817864153,-245.18054404889926,-236.0450249500731,-226.57407295468727,-216.7811468037667,-206.68016278122883,-196.28547493808418,-185.61185469454244,-174.674469849018,-163.4888630238568,-152.070929578421,-140.43689502091192,-128.60329195103705,-116.5869365662789,-104.40490476516027,-92.07450788145701,-79.61326808384563,-67.0388934759429,-54.369252932121604,-41.62235070486233,-28.81630083972545,-15.9693014343013,-3.0996087777176573,9.774488592546064,22.634695879671277,35.462738025407596,48.240385679930576,60.94948110674533,73.57196398582361,86.08989707830688,98.48549171630414,110.74113308156251,122.83940523708765,134.76311587614342,146.4953207534603,158.0193477639352,169.31882063460506,180.37768219622686,191.18021720139356,201.71107465676093,211.95528963764875,221.89830455401776,231.52598983760183,240.82466402079734,249.7811131787775,258.38260970720245,266.61693040884154,274.47237386340544,281.93777705590463,289.00253123990444,295.65659701313444,301.8905185840282,307.6954372089206,313.0631037808075,317.98589055177774,322.4568019724607,326.4694846330846,330.0182362920194,333.09801397897434,335.70444116133433,337.83381396345266,339.48310643006073,340.64997482631617,341.3327609683783,341.5304945797779,341.2428946702334,340.4703699349535,339.2140181738586,337.4756247315474,335.25765996022415,332.5632757091933,329.3963008459092,325.7612358149457,321.66324624261813,317.1081555963462,312.1024369091872,306.65320358130276,300.76819927142736,294.4557868927059,287.72493672853625,280.58521368530586,273.0467637001364,265.120299322952,256.8170844933593,248.14891853397336,239.12811938293422,229.7675060894431,220.0803805971917,210.08050884157257,199.7821011875318,189.19979223586233,178.3486200266361,167.24400466932565,155.90172642998533,144.3379033066292,132.56896812467443,120.61164518499675,108.48292649778415,96.20004763596026,83.7804632424925,71.24182222638886,58.60194268263298,45.878786571695834,33.09043419460717,20.25505849985841,7.390899258648422,-5.483762854829405,-18.35063224122548,-31.1914243750818,-43.98789178814315,-56.72185000000795,-69.37520335926993,-81.92997075842905,-94.36831118602898,-106.6725490797108,-118.82519944415415,-130.8089926982125,-142.60689921593263,-154.20215352658565,-165.57827813931715,-176.71910695856454,-187.60880825696245,-198.2319071730944,-208.57330770211698,-218.61831414800932,-228.35265200696077,-237.76248825222237,-246.83445099159428,-255.55564846961693,-263.9136873874603,-271.8966905144802,-279.4933135664125,-286.69276132622224,-293.48480298469735,-299.85978667898866,-305.808653208435,-311.32294890818304,-316.3948376623077,-321.0171120393615,-325.1832035345294,-328.887191903833,-332.1238135771218,-334.88846913789376,-337.1772298593184,-338.9868432871724,-340.3147378617547,-341.15902657221403,-341.5185096380944,-341.3926762142886,-340.7817051169775,-339.6864645695218,-338.1085109686698,-336.0500866728321,-333.51411681556755,-330.5042051488087,-327.02462892173276,-323.08033280255677,-318.67692185189327,-313.82065355765263,-308.5184289428102,-302.77778275867524,-296.6068727775968,-290.01446820032277,-283.0099371944862,-275.6032335819261,-267.80488269376184,-259.6259664133222,-251.07810742818236,-242.1734527136889,-232.92465627144335,-223.3448611472735,-213.4476807542456,-203.24717952726036,-192.75785293672098,-181.9946068896757,-170.97273654771007,-159.70790459168322,-148.21611896420353,-136.51371012146407,-124.61730782677296,-112.54381751874749,-100.31039628776284,-87.93442849478659,-75.43350106725094,-62.82537850706647,-50.12797764629276,-37.359342186339816,-24.5376170568809,-11.681022630914676,1.192171167381874,14.063670825215373,26.915185237258747,39.72845169831276,52.48526185560941,65.16748758387756,77.75710674640091,90.23622880546037,102.58712024576717,114.7922297747584,126.83421326394418,138.69595839586322,150.36060898162233,161.81158891446324,173.03262572531722,184.0077737068742,194.7214365733051,205.1583896234379,215.30380137589046,225.14325464541756,234.66276703051994,243.84881078320197,252.6883320326414,261.1687693354548,269.2780715261952,277.0047148427181,284.337719302078,291.2666643036847,297.78170343754726,303.8735784765612,309.53363253295583,314.7538223602055,319.5267297829231,323.84557223849305,327.70421241546404,331.09716697500437,334.01961434302643,336.4674015619074,338.4370501920692,339.92576125503143,340.9314192109133,341.4525949647309,341.4885478972201,341.0392269172964,340.1052705346586,338.68800595243147,336.7894471811374,334.4122921766773,331.5599190063876,328.2363810486217,324.44640123267783,320.1953653272576,315.4893142869943,310.3349356679257,304.73955412411107,298.7111209988968,292.2582030256226,285.38997015382455,278.116182518235,270.4471765690974,262.39385038350423,253.96764817863388,245.1805440488913,236.04502495006483,226.57407295467866,216.78114680375776,206.68016278121974,196.2854749380747,185.61185469453275,174.67446984900798,163.48886302384656,152.07092957841041,140.43689502090126,128.60329195102605,116.5869365662678,104.40490476514918,92.07450788144588,79.61326808383446,67.03889347593172,54.36925293211041,41.622350704851144,28.816300839714305,15.9693014342902,3.09960877770662,-9.77448859255702,-22.63469587968214,-35.46273802541835,-48.240385679941205,-60.94948110675582,-73.57196398583395,-86.08989707831704,-98.48549171631413,-110.7411330815723,-122.83940523709724,-134.7631158761528,-146.4953207534694,-158.01934776394398,-169.31882063461367,-180.37768219623513,-191.18021720140163,-201.71107465676866,-211.95528963765625,-221.89830455402495,-231.52598983760873,-240.82466402080397,-249.7811131787837,-258.3826097072083,-266.61693040884705,-274.47237386341067,-281.93777705590946,-289.00253123990893,-295.6565970131386,-301.890518584032,-307.6954372089241,-313.06310378081065,-317.9858905517805,-322.45680197246315,-326.4694846330868,-330.0182362920213,-333.09801397897587,-335.7044411613356,-337.8338139634536,-339.4831064300614,-340.6499748263166,-341.3327609683785,-341.5304945797779,-341.24289467023317,-340.47036993495306,-339.214018173858,-337.47562473154653,-335.2576599602231,-332.5632757091921,-329.3963008459079,-325.7612358149442,-321.6632462426166,-317.1081555963445,-312.10243690918537,-306.6532035813008,-300.76819927142526,-294.45578689270366,-287.72493672853386,-280.58521368530336,-273.0467637001338,-265.1202993229493,-256.8170844933565,-248.14891853397043,-239.12811938293123,-229.76750608944002,-220.08038059718845,-210.0805088415694,-199.7821011875284,-189.19979223585906,-178.34862002663263,-167.24400466932232,-155.9017264299818,-144.3379033066258,-132.56896812467085,-120.6116451849933,-108.48292649778053,-96.20004763595651,-83.78046324248857,-71.24182222638483,-58.60194268262884,-45.87878657169159,-33.090434194602835,-20.255058499853988,-7.390899258643914,5.483762854833989,18.35063224123013,31.19142437508652,43.98789178814792,56.72185000001275,69.3752033592748,81.92997075843394,94.3683111860339,106.67254907971572,118.82519944415908,130.80899269821742,142.6068992159377,154.20215352659062,165.57827813932215,176.71910695856945,187.60880825696736,198.23190717309922,208.57330770212178,218.61831414801395,228.35265200696534,237.7624882522268,246.8344509915987,255.55564846962127,263.9136873874645,271.8966905144843,279.4933135664165,286.69276132622605,293.48480298470105,299.8597866789922,305.80865320843833,311.32294890818616,316.39483766231064,321.01711203936424,325.1832035345319,328.88719190383523,332.1238135771237,334.8884691378954,337.1772298593198,338.98684328717343,340.3147378617554,341.15902657221443,341.51850963809443,341.3926762142883,340.7817051169768,339.68646456952075,338.1085109686684,336.0500866728303,333.51411681556533,330.504205148806,327.02462892172974,323.08033280255324],"type":"scatter","name":"source1_u_a","yaxis":"y2","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1]}],{"xaxis":{"title":{"text":"Time in Seconds"}},"template":{"layout":{"coloraxis":{"colorbar":{"ticks":"","outlinewidth":0}},"xaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"hovermode":"closest","paper_bgcolor":"white","geo":{"showlakes":true,"showland":true,"landcolor":"#E5ECF6","bgcolor":"white","subunitcolor":"white","lakecolor":"white"},"colorscale":{"sequential":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]],"diverging":[[0,"#8e0152"],[0.1,"#c51b7d"],[0.2,"#de77ae"],[0.3,"#f1b6da"],[0.4,"#fde0ef"],[0.5,"#f7f7f7"],[0.6,"#e6f5d0"],[0.7,"#b8e186"],[0.8,"#7fbc41"],[0.9,"#4d9221"],[1,"#276419"]],"sequentialminus":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]},"yaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"shapedefaults":{"line":{"color":"#2a3f5f"}},"hoverlabel":{"align":"left"},"mapbox":{"style":"light"},"polar":{"angularaxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","radialaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"autotypenumbers":"strict","font":{"color":"#2a3f5f"},"ternary":{"baxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","caxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"aaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"annotationdefaults":{"arrowhead":0,"arrowwidth":1,"arrowcolor":"#2a3f5f"},"plot_bgcolor":"#E5ECF6","title":{"x":0.05},"scene":{"xaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"zaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"yaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"}},"colorway":["#636efa","#EF553B","#00cc96","#ab63fa","#FFA15A","#19d3f3","#FF6692","#B6E880","#FF97FF","#FECB52"]},"data":{"barpolar":[{"type":"barpolar","marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"carpet":[{"aaxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"},"type":"carpet","baxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"}}],"scatterpolar":[{"type":"scatterpolar","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"parcoords":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"parcoords"}],"scatter":[{"type":"scatter","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2dcontour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2dcontour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattercarpet":[{"type":"scattercarpet","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"mesh3d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"mesh3d"}],"surface":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"surface","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattermapbox":[{"type":"scattermapbox","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergeo":[{"type":"scattergeo","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram":[{"type":"histogram","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"pie":[{"type":"pie","automargin":true}],"choropleth":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"choropleth"}],"heatmapgl":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmapgl","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"bar":[{"type":"bar","error_y":{"color":"#2a3f5f"},"error_x":{"color":"#2a3f5f"},"marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"heatmap":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmap","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contourcarpet":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contourcarpet"}],"table":[{"type":"table","header":{"line":{"color":"white"},"fill":{"color":"#C8D4E3"}},"cells":{"line":{"color":"white"},"fill":{"color":"#EBF0F8"}}}],"scatter3d":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"scatter3d","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergl":[{"type":"scattergl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2d","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scatterternary":[{"type":"scatterternary","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scatterpolargl":[{"type":"scatterpolargl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}]}},"legend":{"yanchor":"bottom","xanchor":"right","y":1.02,"orientation":"h","x":1},"margin":{"l":100,"b":80,"pad":10,"r":80,"t":100},"plot_bgcolor":"#f1f3f7","yaxis":{"title":{"text":"State values"}},"yaxis2":{"titlefont":{"color":"orange"},"overlaying":"y","title":"Action values","side":"right"},"height":550,"width":800},{"showLink":false,"editable":false,"responsive":true,"staticPlot":false,"scrollZoom":true});
});





</script>
```




_______________________________________________________________________________
### High-Level Rendering

- When sources are classically controlled and the network is a 3 phase system, the user has the option of plotting a number of additional quantities. 
- The only quantities that can be plotted for reinforcement learning controllers are those specified through "states_to_plot" and "actions_to_plot".
- The elements of the vectors in the function "RenderHookResults" refer to the _name_ of the classically controlled source which may differ if there are RL sources in the network, e.g., power_p_inv = [1], indicates that the real power of the first _classically_ controlled source should be plotted.



```julia
RenderHookResults(hook = hook, 
                    states_to_plot  = [], 
                    actions_to_plot = [],  
                    power_p_inv     = [1], # Real power [Watts]
                    power_q_inv     = [1], # Imaginary power [VAi]
                    v_mag_inv       = [1], # Scaled L₂ norm in αβγ coordinates [V]
                    i_mag_inv       = [1], # Scaled L₂ norm in αβγ coordinates [A]
                    angles          = [1], # Relative angle [degrees]
                    freq            = [1], # Angular velocity [Hz]
                    )
```


```@raw html
<div
id = 89cyrVtYdmDx > </div>
<script>
gd = '89cyrVtYdmDx'
require(['plotly'], function(plotly) {
























plotly.newPlot(gd,[{"mode":"lines","y":[0.0,0.0,0.22302919252035303,0.8592533889083606,1.998955163014415,3.7058647749090428,5.9877026173516175,8.87658120507686,12.372654990892645,16.493354697022358,21.24613687560939,26.642214482735962,32.69313830353328,39.408870373788986,46.80133177724083,54.88064922733059,63.657777972981506,73.14270281840137,83.34524960780642,94.2746976391212,105.93977089476743,118.34869349419961,131.50904447033676,145.42782954874954,160.1113954266492,175.56544710746067,191.79501548384394,208.8044475680871,226.59739462297688,245.17679876199418,264.54488603331305,284.7031569307631,305.6523812415715,327.39259290525735,349.9230871242603,373.242418801201,397.3484025322377,422.2381142128434,447.90789412093505,474.3533515848385,501.56937115437927,529.5501203021238,558.2890586244924,587.77894853139,618.0118674037409,648.9792211970046,680.6717594654166,713.079591780059,746.1922055099885,779.9984849342261,814.4867316491649,849.6446862338502,885.4595511332068,921.9180147168379,959.0062764690099,996.710073263117,1035.0147066720028,1073.9050712634712,1113.3656838284764,1153.3807134877077,1193.934012620593,1235.0091485591613,1276.5894359877402,1318.6579699880895,1361.1976596682985,1404.1912623126332,1447.6214179884767,1491.470684545579,1535.7215729420143,1580.3565828305746,1625.3582383387165,1670.7091239747601,1716.3919205926773,1762.3894413476087,1808.6846675741422,1855.2607845194243,1902.1012168633247,1949.1896639581316,1996.5101347206667,2044.0469821102047,2091.784937126221,2139.7091422607255,2187.805184340831,2236.0591266981605,2284.457540602797,2332.9875359007037,2381.636790794827,2430.3935807115445,2479.246806195633,2528.1860197785645,2577.201451766683,2626.2840348976033,2675.425427815154,2724.618037315127,2773.855039316255,2823.130398512985,2872.4388866688814,2921.7760995118397,2971.1384721946883,3020.5232932872445,3069.9287172684026,3119.353775489477,3168.798385582635,3218.263359290957,3267.7504086994713,3317.2621508491893,3366.8021107191353,3416.3747225640564,3465.985329598566,3515.640182021199,3565.3464333749234,3615.112135243478,3664.9462302859465,3714.858543614804,3764.8597725257277,3814.961474590278,3865.1760541256067,3915.5167470580477,3965.997604200595,4016.6334729667838,4067.4399775466372,4118.433497572823,4169.631145308129,4221.050741387841,4272.710789153443,4324.630447616406,4376.829503093602,4429.328339558016,4482.147907751155,4535.309693105489,4588.835682527835,4642.748330096395,4697.070521726538,4751.825538862089,4807.037021251019,4862.72892886595,4918.925503031848,4975.651226824548,5032.930784805535,5090.789022159438,5149.250903302305,5208.3414700295,5268.085799273418,5328.508960541824,5389.6359731087105,5451.491763029926,5514.101120056686,5577.488654520195,5641.6787542612055,5706.695541678288,5772.562830968789,5839.30408563627,5906.942376338142,5975.500339146791,6045.000134297109,6115.463405492747,6186.9112398427005,6259.364128499113,6332.8419280660755,6407.3638228483915,6482.948288007806,6559.613053693236,6637.375070209821,6716.250474290368,6796.254556530873,6877.4017300504,6959.7055004333615,7043.178437010719,7127.832145534154,7213.677242295567,7300.723329741636,7388.978973631148,7478.451681780201,7569.147884438048,7661.072916333679,7754.231000430652,7848.625233425034,7944.257573018393,8041.128826995082,8139.238644130159,8238.585506951276,8339.166726374948,8440.97843823455,8544.015601714382,8648.271999700963,8753.740241059657,8860.411764841574,8968.276846422526,9077.324605572621,9187.543016451942,9298.918919524542,9411.438035379826,9525.084980447231,9639.843284586917,9755.695410536102,9872.622775187574,9990.605772673724,10109.623799226534,10229.655279780905,10350.677696285662,10472.667617683821,10595.600731520724,10719.451877135853,10844.195080391532,10969.803589888863,11096.24991461885,11223.505862993974,11351.542583203223,11480.330604831048,11609.839881678641,11740.03983572366,11870.899402152476,12002.387075397037,12134.470956106697,12267.118798983429,12400.298061407342,12533.975952777799,12668.119484494135,12802.695520498506,12937.670828302404,13073.012130417266,13208.686156108575,13344.659693392205,13480.899641191056,13617.373061569408,13754.04723196206,13890.88969731511,14027.868322055021,14164.95134180256,14302.107414748452,14439.3056726077,14576.5157710701,14713.70793966475,14850.853030957269,14987.922568999038,15124.88879694885,15261.724723788153,15398.404170052665,15534.901812503966,15671.193227666445,15807.254934156434,15943.064433731915,16078.60025099321,16213.841971666701,16348.770279405911,16483.36699104615,16617.615090251464,16751.498759494614,16885.003410313537,17018.115711790186,17150.823617200207,17283.11638878471,17414.98462059817,17546.420259389397,17677.41662347538,17807.968419570996,17938.07175754053,18067.724163040122,18196.92458802367,18325.673419087634,18453.97248363382,18581.82505383246,18709.235848371183,18836.211031979117,18962.758212718723,19088.88643704124,19214.606182605465,19339.92934886286,19464.869245415444,19589.440578156566,19713.659433208115,19837.543258671205,19961.110844210565,20084.382298497057,20207.379024535043,20330.12369290582,20452.64021296089,20574.953702002484,20697.09045249196,20819.07789732985,20940.94457325436,21062.720082408545,21184.435052128723,21306.121093010555,21427.810755310988,21549.537483747914,21671.335570761505,21793.24010830408,21915.286938227488,22037.512601340004,22159.954285205942,22282.64977076475,22405.63737784694,22528.95590966742,22652.64459637745,22776.74303775943,22901.2911451487,23026.329082669865,23151.897207874295,23278.03601186893,23404.786059025384,23532.187926361097,23660.282142683503,23789.10912759018,23918.709130417054,24049.122169228587,24180.387969942425,24312.545905682604,24445.63493645408,24579.69354923207,24714.759698558315,24850.870747737135,24988.063410722003,25126.373694783942,25265.836844051028,25406.48728400792,25548.358567042633,25691.483319126695,25835.89318771322,25981.618790935696,26128.68966818874,26277.134232170076,26426.979722460725,26578.252160718795,26730.976307559475,26885.175621191578,27040.872217878965,27198.08683429146,27356.838791808794,27517.145962836432,27679.024739190947,27842.49000260813,28007.555097425415,28174.231805485302,28342.530323305004,28512.45924155267,28684.02552686815,28857.23450606231,29032.0898527255,29208.593576272277,29386.746013445758,29566.545822301065,29747.98997868437,29931.073775219254,30115.79082280919,30302.13305466072,30490.090732828194,30679.6524572771,30870.805177459348,31063.534206389537,31257.82323720841,31453.65436221456,31651.008094343,31849.863391064493,32050.197680676254,32251.98689095086,32455.205480106684,32659.826470059168,32865.82148190931,33073.16077362164,33281.81327984105,33491.74665379418,33702.92731121783,33915.32047625401,34128.890229247314,34343.59955637846,34559.41040106387,34776.28371704883,34994.17952311902,35213.05695935257,35432.87434483227,35653.58923673509,35875.1584907141,36097.53832248561,36320.684370532595,36544.5517598328,36769.095166519524,36994.26888337999,37220.026886096006,37446.32290012944,37673.11046815407,37900.343017934814,38127.97393055367,38355.95660888134,38584.24454619342,38812.7913948285,39041.55103478608,39270.47764216219,39499.5257573195,39728.650352690594,39957.80690011178,40186.951437586315,40416.040635376216,40645.03186132266,40873.883245296005,41102.553742677184,41331.003196774196,41559.192400078064,41787.083154264714,42014.63832885049,42241.821918411035,42362.69236735722,42437.91782180458,42486.724352253754,42519.077521410814,42539.078012358965,42551.81422020613,42558.06352593835,42559.80523925322,42557.36561312908,42551.33524964432,42541.950436159124,42529.35545644734,42513.69793060397,42495.02688582831,42473.42629827313,42448.9391477865,42421.62231944785,42391.52519160555,42358.700225766006,42323.202799069644,42285.089856646126,42244.4225749073,42201.26455036301,42155.68306005673,42107.74836404034,42057.53392731187,42005.11624969757,41950.57475469018,41893.9917146697,41835.452096692105,41775.043463208596,41712.8558289201,41648.98153512349,41583.51511123634,41516.55313630997,41448.1940976818,41378.538246421,41307.68745132627,41235.74505040027,41162.8157008504,41089.00522741375,41014.42046940662,40939.16912666186,40863.35960458998,40787.100858583006,40710.502237988636,40633.67332987062,40556.723802784894,40479.76325078783,40402.901037901356,40326.24614325299,40249.90700711002,40173.99137802447,40098.60616130349,40023.8572690173,39949.84947175483,39876.68625233304,39804.46966166418,39733.30017698012,39663.27656261066,39594.49573350714,39527.052621699826,39461.040045871516,39396.54858422632,39333.66645082633,39272.47937556442,39213.07048793535,39155.52020476207,39099.90612202734,39046.30291095592,38994.782218484455,38945.412572251764,38898.259290233116,38853.38439513781,38810.84653367969,38770.70090082547,38732.999169116454,38697.78942315349,38665.11609932582,38635.0199308586,38607.53789824428,38582.703185117156,38560.545139620684,38541.08924131102,38524.35707363043,38510.36630197819,38499.1306573966,38490.659925883636,38484.95994333406,38482.03259610397,38481.87582718528,38484.48364796863,38489.84615556597,38497.94955565484,38508.77619080075,38522.30457420376,38538.50942881054,38557.36173172292,38578.82876382943,38602.874164576126,38629.457991788666,38658.53678644841,38690.063642319816,38723.988280319376,38760.25712750957,38798.813400595376,38839.59719379474,38882.54557094798,38927.592661726325,38974.66976179304,39023.70543676661,39074.62562982934,39127.35377282053,39181.81090064839,39237.915768850915,39295.58497413158,39354.73307769177,39415.272731178644,39477.11480506373,39540.168519263956,39604.341575815124,39669.54029340421,39735.66974356548,39802.6338883432,39870.33571922188,39938.677397123975,40007.56039327373,40076.88563072513,40146.553626351335,40216.46463309275,40286.51878226135,40356.616225698686,40426.657277585946,40496.542555705586,40566.17312195484,40635.45062191336,40704.27742326853,40772.556752904726,40840.192832464454,40907.091012192366,40973.157902875624,41038.30150569788,41102.431339826486,41165.45856755725,41227.296116843994,41287.85880104521,41347.06343572332,41404.82895233768,41461.07650867662,41515.72959587886,41568.71414189984,41619.9586112845,41669.39410111187,41716.95443298527,41762.57624094498,41806.19905518813,41847.76538148666,41887.220776199414,41924.51391678219,41959.596667705424,41992.42414169643,42022.95475622923,42051.15028519287,42076.97590567512,42100.40023980658,42121.395391616265,42139.93697885809,42156.00415977357,42169.57965476485,42180.64976295822,42189.20437364596,42195.23697260214,42198.74464327457,42199.72806286283,42198.191493299295,42194.14276715764,42187.593268519864,42178.55790884071,42167.055097854674,42153.10670957808,42136.73804346593,42117.97778078881,42096.85793630333,42073.41380529484,42047.68390607811,42019.709918048036,41989.53661537803,41957.21179647047,41922.786209268575,41886.31347254525,41847.84999328987,41807.454880318895,41765.18985424169,41721.11915391746,41675.309439544384,41627.82969252569,41578.75111226272,41528.14701002802,41476.09270007601,41422.66538815229,41367.944057565255,41312.00935298815,41254.94346216084,41196.829995664935,41137.75386494669,41077.80115876542,41017.05901824603,40955.61551071683,40893.55950251414,40830.980530936824,40767.9686755349,40704.61442891606,40641.00856725505,40577.242020690624,40513.405743794196,40449.59058629409,40385.88716423882,40322.3857317809,40259.17605376254,40196.34727928187,40133.98781641776,40072.18520828785,40011.02601061381,39950.595670963376,39890.97840983794,39832.25710376966,39774.51317059048,39717.82645703023,39662.27512879947,39607.935563306266,39554.882245154884,39503.18766456735,39452.922218866544,39404.154117153324,39356.949288306525,39311.37129242829,39267.48123585386,39225.33768983749,39184.99661302275,39146.51127779799,39109.932200633804,39075.30707649152,39042.68071738786,39012.09499519266,38983.58878873242,38957.19793526427,38932.95518638044,38910.89016839492,38891.029347259566,38873.39599804836,38858.01017904394,38844.88871045188,38834.04515776347,38825.48981977963,38819.22972130276,38815.26861049647,38813.60696090627,38814.24197812829,38817.16761110543,38822.37456802572,38829.85033678831,38839.579209999574,38851.54231445209,38865.71764503586,38882.08010302244,38900.60153865911,38921.25079800145,38943.993773909664,38968.79346112589,38995.61001534549,39024.4008161895,39055.12053397985,39087.721200213964,39122.15228163061,39158.360757753246,39196.291201793356,39235.88586479096,39277.08476286572,39319.825767447655,39364.04469835246,39409.675419563064,39456.64993757505,39504.89850216055,39554.34970940217,39604.93060684524,39656.566800614324,39709.182564337636,39762.700949720114,39817.04389860479,39872.13235635968,39927.88638642612,39984.22528586342,40041.06770172307,40098.33174808577,40155.93512359308,40213.79522930594,40271.82928672159,40329.95445578124,40388.08795270041,40446.1471674552,40504.049780758556,40561.713880360796,40619.05807651153,40676.00161642009,40732.46449755467,40788.36757962128,40843.632695067026,40898.18275795321,40951.94187104807,41004.83543099037,41056.790231379215,41107.73456364769,41157.59831558283,41206.31306735644,41253.812184936185,41300.0309107504,41344.90645148352,41388.37806288398,41430.38713147115,41470.87725303145,41509.79430780003,41547.086532228306,41582.70458724291,41616.601622907314,41648.733339402024,41679.05804424515,41707.53670568014,41734.133002164104,41758.8133678944,41781.54703431813,41802.306067574464,41821.06540182579,41837.802868439394,41852.499220987644,41865.13815604046,41875.706329730056,41884.19337007379,41890.59188504722,41894.89746640533,41897.108689256085,41897.227107396495,41895.25724442719,41891.20658066762,41885.08553590008,41876.90744797617,41866.688547325335,41854.447927411326,41840.207511187036,41823.99201360475,41805.8289002438,41785.748342123115,41763.783166771376,41739.9688056327,41714.34323789084,41686.94693079971,41657.82277661281,41627.01602620878,41594.57421951484,41560.547112834174,41524.98660318732,41487.946649781814,41449.483192728134,41409.65406912365,41368.5189266294,41326.13913466846,41282.5776933774,41237.899140444824,41192.169455974334,41145.45596551183,41097.82724137826,41049.35300245277,41000.104012551084,40950.15197754756,40899.56944138913,40848.429681152076,40796.806601292206,40744.77462724086,40692.408598499445,40639.783661385074,40586.975161581126,40534.05853664507,40481.1092086271,40428.202476951265,40375.41341171111,40322.816747530276,40270.48677813782,40218.49725180597,40166.92126779763,40115.831173967665,40065.298465661705,40015.39368605244,39966.18632805252,39917.7447379388,39870.136020822705,39823.425948094664,39777.67886697213,39732.95761227257,39689.323420533925,39646.835846597554,39605.55268276801,39565.529880657545,39526.82147582148,39489.47951528433,39453.55398805377,39419.09275871406,39386.14150418673,39354.74365374095,39324.9403323319,39296.77030733955,39270.26993877675,39245.473133028776,39222.41130018282,39201.113314999326,39181.60548157316,39163.91150172591,39148.05244716642,39134.04673545007,39121.91010976309,39111.655622551436,39103.29362300941,39096.83174843659,39092.27491946727,39089.62533917013,39088.88249601061,39090.043170663695,39093.101446658016,39098.04872482852,39104.87374154737,39113.56259070024,39124.09874936679,39136.463107161995,39150.63399918714,39166.587242536836,39184.29617630085,39203.73170499756,39224.86234536813,39247.65427645879,39272.07139291158,39298.075361381394,39325.625679991754,39354.67974073821,39385.19289474401,39417.11852026906,39450.408093369224,39485.01126109976,39520.875917153055,39557.94827981798,39596.17297214458,39635.49310419554,39675.850357262694,39717.185069924715,39759.436325819435,39802.54204300222,39846.43906476021,39891.0632517496,39936.34957532273,39982.23221190911,40028.64463831436,40075.51972779953,40122.78984680226,40170.38695216139,40218.242688705446,40266.28848706558,40314.45566157358,40362.67550810557,40410.879401732716,40458.99889404039,40506.96580997836,40554.712344105494,40602.17115609366,40649.27546535666,40695.959144671775,40742.15681266312,40787.80392501822,40832.83686431064,40877.19302830448,40920.810916618626,40963.630215631274,41005.591881508,41046.63822123976,41086.71297158008,41125.761375773734,41163.73025797366,41200.56809524428,41236.225087055696,41270.653222174806,41303.806342864154,41335.640206304,41366.11254315576,41395.18311319052,41422.81375791009,41448.96845009249,41473.6133401984,41496.71679957967,41518.24946043578,41538.18425246869,41556.49643619134,41573.16363285003,41588.16585092586,41601.48550918525,41613.10745625415,41623.01898669651,41631.209853581604,41637.67227753031,41642.40095223545,41645.393046456215,41646.64820249169,41646.16853114371,41643.958603183804,41640.02543734452,41634.37848485967,41627.029610583384,41617.99307072243,41607.28548722095,41594.92581884172,41580.93532899213,41565.33755034808,41548.158246333216,41529.42536951534,41509.16901698574,41487.42138279194,41464.216707497915,41439.591224949705,41413.58310632866,41386.23240157766,41357.58097828932,41327.67245814905,41296.55215102834,41264.26698682751,41230.86544516955,41196.397483049885,41160.91446054949,41124.469064720965,41087.11523176033,41048.90806757813,41009.90376688738,40970.15953092536,40929.73348393057,40888.68458849495,40847.07255991537,40804.957779667275,40762.4012081263,40719.46429666275,40676.20889923577,40632.69718361368,40588.99154234753,40545.154503624944,40501.24864213099,40457.33649004296,40413.48044828493,40369.74269816803,40326.18511354044,40282.8691735715,40239.855876291724,40197.20565301073,40154.97828373149,40113.232813680246,40072.027471066875,40031.41958619083,39991.46551200375,39952.22054623908,39913.738855214986,39876.07339941597,39839.275860954025,39803.39657300918,39768.484451343975,39734.586927986,39701.749887166305,39670.01760360119,39639.43268319831,39610.03600626717,39581.86667330811,39554.961953451864,39529.357235616044,39505.08598244299,39482.17968707695,39460.66783283687,39440.57785583442,39421.93511058483,39404.76283865181,39389.08214036499,39374.91194964247,39362.269011948156,39351.167865407246,39341.62082510038,39333.63797055108,39327.22713641757,39322.39390639463,39319.14161032707,39317.471324532,39317.381875321706,39318.86984571549,39321.92958532334,39326.55322338082,39332.73068490876,39340.449709968874,39349.69587597984,39360.45262305642,39372.70128232775,39386.42110718887],"type":"scatter","name":"source1_p_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1]},{"mode":"lines","y":[0.0,0.0,0.008411987954409075,0.0396013604772234,0.10623393559512684,0.2209415460804618,0.38847017273961937,0.6186263991417507,0.9141271371603628,1.2824472141138976,1.7284200722945915,2.258550571568815,2.8800709751328526,3.6001135296663627,4.427449143858372,5.370637865635167,6.439491583050277,7.644032420841764,8.995015134879957,10.5036705240652,12.181690383171642,14.04127398195866,16.094991139971974,18.35584325160854,20.837165601184836,23.55263499386122,26.516219124658676,29.742149875025895,33.24489203506684,37.03910645957421,41.139620802332445,45.56139276335964,50.3194782192706,55.428996039293835,60.90509424765001,66.7629155777301,73.0175629625291,79.68406527795217,86.77734292048268,94.3121737437046,102.30315908409733,110.7646901472323,119.71091471709528,129.1557042914917,139.1126217051405,149.59488929285362,160.61535766944905,172.1864751765129,184.32025806401649,197.02826146156573,210.321551198738,224.2106765300232,238.70564381859305,253.81589123185023,269.5502644991846,285.91699378099753,302.92367169561885,320.57723254887435,338.88393280873896,357.8493328652752,377.4782801137249,397.77489339617597,418.7425488347785,440.38386708691434,462.70070205015753,485.69413104220445,509.3644464782618,533.711149065643,558.7329425325624,584.4277299053184,610.7926113452122,637.8238835537065,665.5170407514628,693.8667772340216,722.8669915039894,752.5107919767087,782.7905042535183,813.697679953822,845.2231070943153,877.3568220008815,910.0881227358466,943.4055840204819,977.2970736298955,1011.7497702346986,1046.7501826611947,1082.2841705391718,1118.336966303815,1154.893198515716,1191.9369164605002,1229.4516159872041,1267.420266542178,1305.8253393530504,1344.648836715111,1383.8723223303643,1423.4769526474952,1463.4435091490618,1503.7524315303926,1544.383851712948,1585.3176286332257,1626.5333837468006,1668.0105371855884,1709.728344505208,1751.665933957906,1793.8023442256454,1836.1165625467052,1878.5875631684962,1921.1943460583161,1963.9159758033145,2006.7316206302226,2049.6205914751845,2092.5623810335687,2135.5367027195734,2178.523529465383,2221.5031322896575,2264.456118565411,2307.3634699176037,2350.2065796812294,2392.9672898512476,2435.6279274564167,2478.1713402898226,2520.5809319298864,2562.8406959866625,2604.9352495093303,2646.849865492133,2688.570504417321,2730.083844775213,2771.377312502993,2812.4391092856404,2853.2582396641365,2893.824536898007,2934.128687531213,2974.162254612579,3013.917699523956,3053.3884023717787,3092.5686808997816,3131.4538078833016,3170.040026967814,3208.3245669171833,3246.305654239455,3283.9825241609715,3321.3554299220486,3358.42565037056,3395.195495832234,3431.6683122397935,3467.848483505506,3503.7414321251504,3539.353618003883,3574.692535498038,3609.7667086692973,3644.5856847513915,3679.160025831795,3713.501298754658,3747.6220632535537,3781.5358583263533,3815.257186866853,3848.8014985714744,3882.1851711417166,3915.4254898065305,3948.540625191233,3981.5496095629132,4014.472311484694,4047.3294089143897,4080.1423607855486,4112.933377111751,4145.725387657538,4178.542009221981,4211.407511583375,4244.346782155899,4277.385289411494,4310.549045122492,4343.864565482506,4377.358831165493,4411.059246384521,4444.993597013981,4479.190007840536,4513.676899009948,4548.482941738362,4583.637013358223,4619.168151770251,4655.105509374209,4691.4783065523025,4728.31578478,4765.647159440107,4803.501572416439,4841.908044544325,4880.89542799554,4920.492358675694,4960.72720871233,5001.628039112211,5043.222552666108,5085.538047179516,5128.601369107173,5172.438867669176,5217.076349525714,5262.539034087021,5308.851509534165,5356.037689625644,5404.120771363473,5453.123193591551,5503.0665965976705,5553.971782789209,5605.858678511043,5658.746297072606,5712.652703049273,5767.594977921421,5823.589187112546,5880.65034848581,5938.79240235617,5998.028183073067,6058.36939222619,6119.826573524529,6182.409089396265,6246.125099354547,6310.981540171352,6376.984107899155,6444.137241776784,6512.444110053682,6581.90659776284,6652.525296470683,6724.299496028272,6797.227178345522,6871.305013206767,6946.528356142955,7022.891248372163,7100.386418817408,7179.00528820682,7258.737975258374,7339.573304947796,7421.498818855176,7504.500787582129,7588.564225228298,7673.672905912608,7759.809382321269,7846.955006261422,7935.0899511957505,8024.193236730467,8114.24275502576,8205.215299094549,8297.086592952339,8389.83132357792,8483.423174641704,8577.834861955434,8673.038170594036,8769.003993637856,8865.702372480648,8963.102538645826,9061.172957051234,9159.881370659961,9259.19484645256,9359.07982265349,9459.502157142542,9560.427176979869,9661.819728971363,9763.644231199074,9865.864725439666,9968.444930392305,10071.348295635893,10174.538056233952,10277.977287904394,10381.628962670178,10485.456004905804,10589.421347693813,10693.48798940449,10797.61905041165,10901.777829856552,11005.927862372006,11110.032974678099,11214.057341961314,11317.9655439487,11421.722620588604,11525.294127250676,11628.646189357278,11731.745556360089,11834.559654975408,11937.056641593605,12039.205453778162,12140.975860771905,12242.33851292849,12343.264989989055,12443.727848125534,12543.700665673507,12643.158087479651,12742.075867790407,12840.430911610794,12938.201314464439,13035.366400487803,13131.906758794696,13227.80427804881,13323.04217918526,13417.605046224504,13511.478855124795,13604.651000622289,13697.110321010961,13788.847120817216,13879.853191327637,13970.121828931118,14059.64785124017,14148.427610959294,14236.459007471973,14323.741496120829,14410.276095159625,14496.06539035849,14581.113537248246,14665.426260992295,14749.010853879252,14831.876170431895,14914.032620133097,14995.49215777138,15076.268271414387,15156.37596802004,15235.831756700958,15314.653629659375,15392.861040815294,15470.474882152374,15547.517457811326,15624.012455962617,15699.984918495247,15775.461208559993,15850.4689760113,15925.037120792414,15999.19575431476,16072.976158882737,16146.410745220935,16219.533008161561,16292.377480554618,16364.979685464818,16437.376086723278,16509.604037903544,16581.70172979518,16653.70813644954,16725.662959875855,16797.60657346688,16869.579964236575,16941.624673953436,17013.782739255417,17086.0966308342,17158.609191778047,17231.363575163796,17304.403180990597,17377.7715925483,17451.512512315363,17525.669697481426,17600.286895191057,17675.407777605316,17751.075876878702,17827.334520149343,17904.226764640283,17981.79533297004,18060.082548770388,18139.130272709335,18218.979839016378,18299.671992607226,18381.246826903993,18463.743722446467,18547.20128638856,18631.65729297358,18717.14862508017,18803.71121692967,18891.37999804385,18980.188838540802,19070.17049585411,19161.356562959387,19253.777418189657,19347.46217671878,19442.4386437902,19538.73326976553,19636.37110706495,19735.375769068865,19835.76939104718,19937.572593180077,20040.804445730675,20145.482436427286,20251.622440109517,20359.23869068923,20468.343755474132,20578.948511898117,20691.062126699202,20804.692037581983,20919.84393739827,21036.52176087552,21154.727673919213,21274.462065510947,21395.723542221334,21518.508925351256,21642.81325071275,21768.62977105559,21895.94996114265,22024.76352547208,22155.058408642035,22286.82080834741,22420.035190996234,22554.684309928372,22690.749226215135,22828.209332015,22967.04237645631,23107.224494014397,23248.730235346688,23391.532600544866,23535.60307476061,23680.911666157284,23827.426946136,23915.327772350545,24002.242906358566,24087.89896160412,24172.424656153733,24255.814679224233,24338.56424997097,24420.571346790774,24502.117148654394,24583.12632004804,24663.641039365968,24743.61060737452,24822.961511578244,24901.631596175685,24979.516299053474,25056.532105874685,25132.573428126852,25207.543648789244,25281.341085828702,25353.86580850411,25425.0195973521,25494.70476194355,25562.826632906264,25629.291977459987,25694.01047151233,25756.894147783678,25817.8578811611,25876.81942443871,25933.699513696105,25988.42203985472,26040.91409537004,26091.106118296906,26138.931957445053,26184.328974364955,26227.238119382957,26267.604008844624,26305.374996907893,26340.50323912059,26372.944753323725,26402.659472605552,26429.611294015463,26453.768120742076,26475.101898824076,26493.58864795495,26509.208486390835,26521.945650046648,26531.788505635617,26538.72955795413,26542.765451244893,26543.896964688287,26542.129002024758,26537.470575340853,26529.93478305589,26519.538782151176,26506.303754696237,26490.254868731136,26471.421233573696,26449.835849627467,26425.535552773945,26398.56095344057,26368.956370442454,26336.769759704162,26302.05263797364,26264.860001648634,26225.250240841062,26183.285048813006,26139.02932692229,26092.551085224084,26043.921338878004,25993.214000518772,25940.505768751005,25885.876012936482,25829.406654444825,25771.182044545225,25711.288839119043,25649.81587037995,25586.854015788776,25522.49606435735,25456.83658053556,25389.971765881914,25321.999318717528,25253.01829196913,25183.128949405407,25112.432620476247,25041.031553962435,24969.028770648412,24896.527915227365,24823.633107653124,24750.448794148928,24677.079598087723,24603.630170954275,24530.205043602084,24456.908478014597,24383.844319781296,24311.11585149559,24238.825647281654,24167.0754286536,24095.965921909366,24025.59671725808,23956.066129877396,23887.47106309374,23819.90687387508,23753.467240822414,23688.24403484135,23624.32719267218,23561.80459345113,23500.76193847182,23441.282634310297,23383.447679472192,23327.335554715355,23273.02211719516,23220.580498574713,23170.08100723558,23121.591034718844,23075.174966519815,23030.894097353696,22988.80655100217,22948.967204844725,22911.427619171365,22876.23597136661,22843.436995046955,22813.071924227603,22785.17844258622,22759.790637884533,22736.93896160074,22716.65019381886,22698.94741341245,22683.84997355384,22671.373482571245,22661.529790169334,22654.32697902051,22649.76936172695,22647.857483145774,22648.588128061787,22651.95433418502,22657.945410442804,22666.546960528256,22677.740911660105,22691.50554850132,22707.815552176682,22726.642044322525,22747.95263609478,22771.71148205453,22797.87933884355,22826.413628555893,22857.26850670488,22890.394934678403,22925.740756569998,22963.250780266262,23002.866862666317,23044.527998902395,23088.17041542661,23133.72766682203,23181.13073619333,23230.308138985,23281.186030073248,23333.688313971295,23387.73675798462,23443.25110814908,23500.149207780636,23558.347118462698,23617.75924329344,23678.298452212897,23739.876209227667,23802.402701347575,23865.786969047513,23929.93703806568,23994.76005234827,24060.162407948563,24126.049887688583,24192.327796389887,24258.90109648053,24325.67454378435,24392.5528232993,24459.440684772024,24526.243077876523,24592.865286805878,24659.213064086645,24725.19276342811,24790.711471419556,24855.677137891136,24919.998704755926,24983.586233153655,25046.35102871945,25108.20576480357,25169.064603471703,25228.843314119014,25287.459389534677,25344.832159257814,25400.882900069744,25455.53494347193,25508.713780003993,25560.34716026044,25610.365192469828,25658.70043650577,25705.2879942042,25750.065595866476,25792.973682834483,25833.95548602878,25872.95710034771,25909.927554830436,25944.818878494683,25977.58616176399,26008.18761340857,26036.58461292766,26062.74175831022,26086.626909115468,26108.21122482357,26127.469198411625,26144.378685118798,26158.92092636994,26171.08056883471,26180.845678606285,26188.20775048998,26193.161712400048,26195.705924868962,26195.842175681108,26193.57566964934,26188.91501355975,26181.872196317036,26172.462564329326,26160.70479217797,26146.620848624974,26130.23595801622,26111.57855714624,26090.680247655415,26067.575744037862,26042.302817343058,26014.90223466148,25985.417694489082,25953.895758072304,25920.385776839463,25884.93981603117,25847.612574646002,25808.4613018242,25767.545709795355,25724.92788352247,25680.672187176853,25634.845167585496,25587.515454792898,25538.753659887316,25488.63227024042,25437.22554231722,25384.60939221188,25330.86128407209,25276.060116573237,25220.286107609987,25163.62067737121,25106.146329969855,25047.946533797214,24989.105600776616,24929.708564688146,24869.84105874184,24809.589192572545,24749.039428834752,24688.278459571637,24627.393082536095,24566.470077638285,24505.596083696015,24444.857475661425,24384.340242498263,24324.12986588086,24264.31119988618,24204.968351846903,24146.184564532785,24088.04209982423,24030.622124040445,23974.004595080987,23918.268151537155,23863.4900039266,23809.745828200255,23757.109661668583,23705.65380148891,23655.448705852792,23606.562898007473,23559.062873241583,23513.013008960435,23468.47547797198,23425.510165099047,23384.174587229456,23344.52381690917,23306.61040957977,23270.484334554905,23236.19290982585,23203.78074077995,23173.289662910094,23144.75868858827,23118.223957968898,23093.718694083167,23071.273162178186,23050.914633349756,23032.66735251021,23016.552510727757,23002.58822196615,22990.78950424849,22981.16826526153,22973.733292411256,22968.490247333793,22965.441664859445,22964.58695642135,22965.922417893984,22969.44124184048,22975.133534141263,22982.986334970934,22992.98364408342,23005.106450360174,23019.332765569645,23035.63766228076,23053.993315867112,23074.36905053322,23096.7313892881,23121.0441077867,23147.268291953835,23175.36239930001,23205.282323834545,23236.981464474604,23270.410796846114,23305.518948366283,23342.252276494164,23380.554950030295,23420.36903334438,23461.63457340292,23504.28968946873,23548.270665337684,23593.512043976472,23639.94672442123,23687.50606079486,23736.119963296966,23785.7170010189,23836.224506433515,23887.568681407738,23939.674704583704,23992.46683997276,24045.868546605467,24099.802589079547,24154.191148846567,24208.955936077487,24264.018301946988,24319.299351176283,24374.720054673657,24430.201362112537,24485.664314287325,24541.03015508801,24596.220442935173,24651.157161518262,24705.762829681298,24759.960610302183,24813.674418012826,24866.82902560963,24919.35016900622,24971.164650582752,25022.20044078825,25072.386777855565,25121.65426549176,25169.934968409485,25217.16250556857,25263.272141000256,25308.200872090823,25351.88751520448,25394.27278853035,25435.29939204138,25474.912084458738,25513.057757118608,25549.685504643603,25584.746692325443,25618.19502013059,25649.98658324556,25680.07992908339,25708.436110678682,25735.01873640264,25759.794015936943,25782.73080244803,25803.80063091201,25822.97775254272,25840.239165284365,25855.5646403328,25868.936744657847,25880.340859502834,25889.765194844458,25897.20079980137,25902.641568985826,25906.08424479846,25907.52841567194,25906.976510275217,25904.433787695583,25899.908323621494,25893.410992554745,25884.955446086286,25874.558087275153,25862.238041175908,25848.01712156482,25831.919793920897,25813.973134722386,25794.206787125237,25772.652913094218,25749.346142062983,25724.323516203498,25697.624432390585,25669.29058095082,25639.36588129093,25607.89641450301,25574.930353050277,25540.517887638543,25504.711151385152,25467.564141397303,25429.132637878905,25389.474120884686,25348.647684846743,25306.713950998237,25263.734977825283,25219.774169676566,25174.89618366681,25129.16683500821,25082.653000909628,25035.422523181478,24987.54410968964,24939.087234798724,24890.122038950718,24840.71922752124,24790.949969101133,24740.885793346984,24690.59848854847,24640.15999905725,24589.64232272473,24539.117408493046,24488.65705428584,24438.332805341888,24388.21585313615,24338.37693502964,24288.88623478966,24239.81328411953,24191.226865335375,24143.194915325883,24095.784430928412,24049.061375852893,24003.090589281954,23957.935696273926,23913.659020091283,23870.32149657544,23827.982590684398,23786.700215307377,23746.5306524664,23707.528477011656,23669.746482913233,23633.23561224827,23598.04488697831,23564.22134360769,23531.80997080887,23500.85365009735,23471.393099633082,23443.46682122134,23417.11105058142,23392.359710946068,23369.24437005058,23347.79420056458,23328.035944015388,23309.993878246154,23293.689788447282,23279.142941793863,23266.37006571774,23255.385329836252,23246.200331555538,23238.824085360564,23233.263015798642,23229.520954158324,23227.59913884005,23227.496219409648,23229.20826432084,23232.728772287453,23238.048687280952,23245.156417123984,23254.037855645394,23264.67640835716,23277.05302160912,23291.146215171953,23306.93211819471,23324.384508478055,23343.47485500016,23364.172363627305,23386.444025937475,23410.254671080103,23435.567020591898,23462.341746084236,23490.537529713223,23520.111127341148,23551.017434292524,23583.20955360645,23616.638866682057,23651.25510621154,23687.00643129216,23723.839504605465,23761.699571549325,23800.5305412062,23840.275069028532,23880.87464111973,23922.26965998745,23964.399531644158,24007.202753928228,24050.617005917306,24094.579238304308,24139.02576460572,24183.89235307063,24229.114319158576,24274.62661845306,24320.36393987816,24366.260799085045,24412.25163187541,24458.27088752865,24504.253121900747,24550.13309016317,24595.845839050613,24641.32679848756,24686.51187246515,24731.337529040844,24775.740889334887,24819.659815399347,24863.032996837603,24905.800036053894,24947.90153201498,24989.279162408107,25029.87576408212,25069.635411661675,25108.503494226457,25146.42678995107,25183.353538604035,25219.233511807906,25254.01808096534,25287.660282760145,25320.1148821456,25351.33843273607,25381.28933452224,25409.927888833532,25437.2163504762,25463.118976979305,25487.60207488452,25510.634043021753,25532.185412714476,25552.228884866003,25570.739363879497,25587.693988372077,25603.072158645246,25616.855560881126,25629.02818803628,25639.576357411825,25648.48872488177,25655.75629576699,25661.3724323469,25665.3328580058,25667.635658015333,25668.28127695981,25667.27251281506,25664.61450769745,25660.31473530296,25654.38298506204,25646.831343039972,25637.674169617138,25626.928073988227,25614.611885523605,25600.74662204073,25585.355455037425,25568.463671943446,25550.098635450377,25530.289739984546,25509.068365390765,25486.46782789939,25462.523328452084,25437.271898466,25410.75234311858,25383.00518223999,25354.07258890159,25323.99832579362,25292.82767948677,25260.60739267659,25227.385594510444,25193.21172910169,25158.13648233501,25122.211707072536,25085.4903468689,25048.026358308467,25009.874632076746,24971.09091288238,24931.731718344585,24891.854256965045,24851.51634530159,24810.776324464154,24769.692976052167,24728.325437655334,24686.73311803714,24644.975612124268,24603.112615921185,24561.203841472605,24519.30893199327,24477.48737728622,24435.79842956829,24394.30101982213,24353.053674791892,24312.114434739582,24271.54077207668,24231.389510985144,24191.716748139468,24152.57777464005],"type":"scatter","name":"source1_q_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1]},{"mode":"lines","y":[0.6037499999999999,1.2075000000000002,1.8112499999999998,2.415,3.0187500000000007,3.6225,4.226249999999999,4.83,5.433750000000001,6.037500000000001,6.6412499999999985,7.244999999999999,7.848750000000002,8.452499999999999,9.056250000000002,9.66,10.263750000000003,10.867500000000001,11.47125,12.075000000000003,12.678750000000003,13.2825,13.88625,14.490000000000002,15.09375,15.697500000000003,16.301249999999996,16.905,17.508750000000003,18.112500000000004,18.716250000000002,19.32,19.92375000000001,20.5275,21.131249999999998,21.73500000000001,22.338750000000005,22.9425,23.546249999999997,24.150000000000002,24.753750000000007,25.357500000000005,25.96125,26.565000000000005,27.168750000000003,27.7725,28.37625,28.980000000000004,29.583749999999995,30.1875,30.79125000000001,31.394999999999996,31.998750000000005,32.6025,33.206250000000004,33.80999999999999,34.41375,35.017500000000005,35.62124999999999,36.225,36.82875,37.4325,38.036249999999995,38.63999999999999,39.243749999999984,39.84749999999999,40.45124999999999,41.055,41.658750000000005,42.262499999999996,42.86625000000001,43.47,44.07374999999999,44.677499999999995,45.28125,45.88499999999999,46.488749999999996,47.092499999999994,47.69624999999999,48.3,48.903749999999974,49.50749999999999,50.11125,50.715,51.31875000000001,51.92249999999999,52.526250000000005,53.13000000000001,53.73374999999999,54.337500000000006,54.941250000000004,55.545,56.14875000000001,56.7525,57.35625,57.96000000000001,58.563750000000006,59.16749999999999,59.771250000000016,60.37500000000001,60.978750000000005,61.58250000000001,62.186250000000015,62.79000000000001,63.393750000000004,63.9975,64.60125,65.20500000000001,65.80875,66.41250000000002,67.01625,67.62,68.22375000000001,68.82750000000001,69.43124999999998,70.03500000000003,70.63874999999999,71.24249999999999,71.84625000000001,72.45000000000002,73.05375,73.65750000000003,74.26124999999999,74.86500000000002,75.46874999999999,76.07250000000002,76.67625,77.28000000000003,77.88375,78.48750000000001,79.09125,79.69500000000001,80.29874999999998,80.90250000000002,81.50625,82.11000000000003,82.71374999999998,83.31750000000001,83.92124999999999,84.52499999999999,85.12875,85.7325,86.33624999999999,86.94,87.54374999999999,88.1475,88.75125000000003,89.355,89.95875,90.56249999999996,91.16625,91.76999999999997,92.37375000000003,92.97749999999996,93.58125000000001,94.18499999999996,94.78875000000001,95.39249999999996,95.99625000000002,96.59999999999998,97.20375,97.80749999999993,98.41125,99.01499999999999,99.61875000000002,100.22249999999997,100.82624999999997,101.42999999999996,102.03375000000001,102.63750000000002,103.24124999999998,103.845,104.44874999999999,105.05250000000004,105.65625,106.25999999999999,106.86375,107.4675,108.07124999999999,108.67500000000001,109.27875,109.88250000000001,110.48624999999998,111.09000000000002,111.69375000000004,112.29749999999999,112.90125000000002,113.50500000000002,114.10875,114.71250000000003,115.31625000000001,115.92000000000002,116.52375000000002,117.12750000000001,117.73125000000002,118.33500000000002,118.93875000000003,119.54250000000003,120.14625000000001,120.75,121.35375,121.95750000000001,122.56125000000003,123.16500000000002,123.76874999999998,124.37250000000002,124.97624999999998,125.58000000000003,126.18375000000005,126.78750000000001,127.39125,127.99500000000002,128.59875000000002,129.20250000000001,129.80625000000003,130.41,131.01375,131.6175,132.22125,132.82500000000002,133.42875,134.03250000000003,134.63624999999996,135.23999999999998,135.84375,136.4475,137.05125,137.65499999999994,138.25874999999996,138.86249999999995,139.46625,140.07000000000002,140.67374999999996,141.27749999999997,141.88125,142.48499999999996,143.08874999999995,143.6925,144.29625000000001,144.89999999999998,145.50374999999994,146.1075,146.71125,147.315,147.91875,148.5225,149.12625000000003,149.72999999999996,150.33374999999998,150.93749999999997,151.54125000000005,152.145,152.74874999999994,153.35249999999996,153.95624999999998,154.56,155.16375,155.76749999999998,156.37125000000003,156.975,157.57875,158.1825,158.78625,159.38999999999996,159.99375000000003,160.59749999999997,161.20125,161.80499999999998,162.40875,163.0125,163.61625000000004,164.21999999999997,164.82375000000002,165.42749999999998,166.03125000000003,166.63500000000002,167.23875000000004,167.8425,168.44625000000005,169.04999999999995,169.65375000000003,170.25750000000002,170.86125000000004,171.465,172.06875000000002,172.67249999999999,173.27625000000006,173.88,174.48375000000007,175.08750000000003,175.69125000000005,176.29499999999996,176.8987500000001,177.5025,178.10625000000007,178.71,179.31375000000003,179.9175,180.52125000000004,181.12499999999994,181.72875000000002,182.3325,182.93625000000003,183.53999999999996,184.14374999999998,184.7475,185.35125000000005,185.95499999999998,186.55875,187.16249999999997,187.76624999999999,188.36999999999998,188.97375,189.57750000000001,190.18124999999995,190.785,191.38874999999993,191.99249999999995,192.59624999999997,193.20000000000005,193.80374999999998,194.40750000000003,195.01124999999993,195.615,196.21874999999991,196.82250000000005,197.42624999999992,198.03000000000006,198.63374999999988,199.23750000000004,199.84124999999997,200.44500000000002,201.04875,201.6525,202.25624999999994,202.85999999999996,203.46375,204.06750000000005,204.67125,205.27500000000003,205.87874999999997,206.48249999999996,207.08625,207.69,208.29374999999996,208.89749999999998,209.50125000000003,210.10500000000008,210.70875,211.31250000000003,211.91625000000005,212.51999999999998,213.12375,213.7275,214.33125,214.93500000000003,215.53875000000008,216.14250000000004,216.74625000000006,217.35000000000005,217.95374999999999,218.55749999999998,219.16125,219.76500000000004,220.36875000000003,220.9725,221.57625000000002,222.18000000000004,222.78375000000003,223.38750000000005,223.99125,224.5950000000001,225.19874999999996,225.80250000000004,226.40625,227.01000000000008,227.61375000000007,228.21749999999997,228.82125000000002,229.42499999999998,230.02875,230.6325000000001,231.23625000000004,231.84000000000003,232.44375000000005,233.04749999999999,233.65125000000006,234.255,234.85875000000001,235.46249999999998,236.06624999999997,236.66999999999996,237.27374999999995,237.8775,238.48124999999996,239.085,239.68874999999997,240.2925,240.89625,241.49999999999997,241.49999999999997,241.49999999999997,241.5,241.49999999999997,241.49999999999997,241.5,241.49999999999997,241.4999999999999,241.5,241.5,241.49999999999997,241.49999999999991,241.49999999999997,241.5,241.5,241.49999999999997,241.5,241.5,241.49999999999997,241.49999999999997,241.5,241.50000000000003,241.5,241.5,241.5,241.5,241.5,241.5,241.5,241.50000000000003,241.5,241.50000000000003,241.50000000000003,241.5,241.49999999999997,241.5,241.5,241.50000000000003,241.50000000000003,241.50000000000003,241.5,241.50000000000003,241.5,241.50000000000006,241.5,241.50000000000009,241.5,241.50000000000009,241.5,241.50000000000006,241.5,241.50000000000006,241.49999999999997,241.50000000000006,241.49999999999997,241.50000000000006,241.49999999999997,241.50000000000006,241.5,241.50000000000009,241.5,241.50000000000003,241.5,241.50000000000003,241.5,241.50000000000006,241.5,241.5,241.49999999999997,241.50000000000003,241.5,241.50000000000003,241.5,241.50000000000003,241.5,241.5,241.5,241.5,241.50000000000003,241.49999999999997,241.5,241.5,241.5,241.49999999999997,241.5,241.49999999999997,241.50000000000006,241.49999999999991,241.5,241.4999999999999,241.50000000000003,241.4999999999999,241.5,241.4999999999999,241.50000000000006,241.4999999999999,241.50000000000006,241.49999999999997,241.5,241.49999999999997,241.49999999999997,241.49999999999997,241.5,241.5,241.50000000000003,241.5,241.50000000000003,241.49999999999997,241.5,241.5,241.50000000000003,241.5,241.50000000000003,241.5,241.5,241.5,241.50000000000006,241.5,241.5,241.50000000000003,241.50000000000003,241.50000000000006,241.50000000000003,241.50000000000003,241.5,241.50000000000006,241.50000000000006,241.50000000000003,241.50000000000003,241.50000000000006,241.50000000000003,241.5,241.50000000000006,241.50000000000006,241.5,241.50000000000009,241.50000000000006,241.50000000000006,241.5,241.50000000000006,241.50000000000003,241.5,241.5,241.50000000000003,241.5,241.49999999999997,241.50000000000006,241.5,241.50000000000003,241.50000000000006,241.49999999999997,241.50000000000003,241.50000000000003,241.49999999999997,241.5,241.50000000000003,241.5,241.5,241.49999999999997,241.49999999999997,241.5,241.5,241.49999999999997,241.49999999999997,241.5,241.5,241.5,241.49999999999997,241.49999999999991,241.5,241.5,241.49999999999997,241.5,241.50000000000006,241.49999999999997,241.49999999999997,241.49999999999997,241.5,241.50000000000003,241.4999999999999,241.5,241.5,241.5,241.49999999999997,241.5,241.50000000000006,241.5,241.49999999999997,241.50000000000003,241.50000000000003,241.50000000000003,241.49999999999997,241.50000000000003,241.50000000000003,241.50000000000003,241.49999999999997,241.5,241.50000000000006,241.5,241.50000000000003,241.5,241.5,241.5,241.50000000000003,241.5,241.5,241.50000000000006,241.50000000000003,241.50000000000006,241.50000000000003,241.5,241.5,241.50000000000009,241.50000000000003,241.50000000000006,241.5,241.50000000000006,241.5,241.50000000000006,241.5,241.50000000000006,241.50000000000003,241.50000000000006,241.5,241.50000000000006,241.5,241.50000000000003,241.49999999999997,241.50000000000003,241.5,241.50000000000006,241.5,241.50000000000006,241.49999999999997,241.50000000000003,241.5,241.50000000000003,241.5,241.50000000000003,241.5,241.50000000000003,241.5,241.5,241.5,241.50000000000003,241.49999999999997,241.49999999999997,241.49999999999997,241.49999999999997,241.5,241.4999999999999,241.5,241.49999999999991,241.50000000000006,241.49999999999991,241.50000000000003,241.4999999999999,241.50000000000009,241.4999999999999,241.50000000000003,241.49999999999991,241.50000000000006,241.49999999999991,241.50000000000003,241.49999999999991,241.5,241.49999999999997,241.5,241.49999999999997,241.49999999999997,241.49999999999997,241.50000000000003,241.5,241.5,241.5,241.5,241.5,241.50000000000003,241.5,241.5,241.50000000000003,241.50000000000003,241.49999999999997,241.5,241.5,241.50000000000003,241.5,241.5,241.5,241.50000000000003,241.5,241.50000000000003,241.50000000000003,241.50000000000003,241.5,241.50000000000003,241.50000000000003,241.50000000000003,241.50000000000006,241.5,241.5,241.49999999999997,241.50000000000003,241.50000000000006,241.5,241.49999999999997,241.5,241.50000000000003,241.50000000000003,241.49999999999997,241.5,241.50000000000006,241.5,241.50000000000006,241.50000000000003,241.5,241.50000000000003,241.5,241.5,241.5,241.49999999999997,241.5,241.5,241.5,241.5,241.50000000000003,241.5,241.5,241.49999999999997,241.50000000000003,241.5,241.5,241.5,241.49999999999997,241.49999999999997,241.5,241.5,241.5,241.49999999999997,241.49999999999997,241.5,241.49999999999997,241.5,241.5,241.50000000000003,241.49999999999997,241.4999999999999,241.5,241.50000000000003,241.5,241.49999999999997,241.5,241.5,241.5,241.5,241.5,241.50000000000003,241.5,241.5,241.50000000000003,241.5,241.49999999999997,241.49999999999997,241.50000000000003,241.5,241.50000000000003,241.5,241.50000000000003,241.50000000000006,241.50000000000006,241.50000000000003,241.5,241.5,241.50000000000003,241.50000000000003,241.50000000000006,241.50000000000003,241.50000000000006,241.5,241.50000000000006,241.5,241.50000000000006,241.5,241.50000000000006,241.50000000000003,241.50000000000006,241.49999999999997,241.50000000000006,241.5,241.50000000000006,241.49999999999997,241.50000000000006,241.49999999999997,241.50000000000006,241.5,241.50000000000006,241.5,241.50000000000003,241.49999999999997,241.50000000000003,241.50000000000003,241.50000000000003,241.5,241.5,241.5,241.50000000000003,241.5,241.5,241.50000000000003,241.5,241.5,241.49999999999997,241.50000000000003,241.5,241.5,241.49999999999991,241.49999999999997,241.4999999999999,241.49999999999997,241.4999999999999,241.50000000000006,241.49999999999991,241.50000000000003,241.4999999999999,241.50000000000006,241.4999999999999,241.50000000000003,241.49999999999991,241.50000000000003,241.49999999999991,241.50000000000006,241.49999999999991,241.5,241.49999999999997,241.49999999999997,241.5,241.5,241.49999999999997,241.5,241.5,241.5,241.5,241.5,241.50000000000003,241.50000000000003,241.5,241.5,241.5,241.5,241.50000000000003,241.5,241.50000000000003,241.5,241.50000000000003,241.5,241.5,241.50000000000003,241.5,241.50000000000003,241.50000000000006,241.50000000000006,241.5,241.5,241.50000000000006,241.50000000000006,241.50000000000003,241.49999999999997,241.50000000000003,241.50000000000006,241.5,241.50000000000003,241.50000000000009,241.50000000000003,241.50000000000006,241.50000000000006,241.50000000000003,241.50000000000006,241.50000000000003,241.50000000000003,241.50000000000003,241.50000000000006,241.49999999999997,241.50000000000003,241.5,241.49999999999997,241.50000000000003,241.5,241.5,241.50000000000003,241.50000000000003,241.50000000000003,241.5,241.5,241.50000000000003,241.5,241.49999999999997,241.49999999999997,241.49999999999997,241.49999999999997,241.5,241.49999999999997,241.4999999999999,241.5,241.5,241.5,241.49999999999991,241.5,241.5,241.49999999999997,241.49999999999991,241.49999999999997,241.5,241.5,241.4999999999999,241.49999999999997,241.50000000000003,241.5,241.49999999999991,241.50000000000003,241.50000000000003,241.5,241.5,241.5,241.50000000000003,241.5,241.49999999999997,241.5,241.5,241.5,241.5,241.50000000000003,241.50000000000003,241.5,241.49999999999997,241.50000000000003,241.50000000000003,241.50000000000006,241.5,241.50000000000003,241.50000000000003,241.50000000000003,241.50000000000006,241.5,241.5,241.50000000000006,241.50000000000003,241.50000000000006,241.49999999999997,241.50000000000009,241.49999999999997,241.50000000000003,241.5,241.50000000000006,241.5,241.50000000000006,241.5,241.50000000000003,241.5,241.50000000000006,241.5,241.50000000000003,241.49999999999997,241.50000000000003,241.49999999999997,241.50000000000006,241.5,241.50000000000006,241.5,241.50000000000003,241.5,241.50000000000003,241.5,241.50000000000003,241.49999999999997,241.5,241.49999999999997,241.5,241.5,241.49999999999997,241.49999999999997,241.5,241.49999999999997,241.5,241.49999999999997,241.5,241.5,241.49999999999991,241.5,241.49999999999991,241.50000000000006,241.49999999999991,241.5,241.4999999999999,241.5,241.4999999999999,241.50000000000006,241.49999999999991,241.50000000000003,241.49999999999991,241.5,241.49999999999997],"type":"scatter","name":"source1_v_mag_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1]},{"mode":"lines","y":[0.0,0.0,0.06161153168798866,0.15830052553948498,0.27629759276833293,0.4099318301942621,0.5521316741376042,0.7018130129995341,0.8562027740632229,1.0148386261852058,1.1768848213720107,1.3420051739707308,1.5099954768543222,1.6806484717074597,1.8539006489116703,2.0296410117635273,2.2078209776543454,2.3883749821879396,2.5712517973291753,2.7563990913555965,2.94376366259732,3.1332934754019433,3.32493361346309,3.518628996807479,3.7143223888620316,3.9119554089779554,4.111468103052815,4.312799119180292,4.515885783537209,4.720664105722121,4.927068909584743,5.1350338626551215,5.3444915746339765,5.5553736578972135,5.767610804731302,5.981132858411911,6.1958688836891636,6.411747238157499,6.628695641096257,6.8466412428394765,7.0655106926470985,7.28523020624727,7.505725632552929,7.726922519763178,7.9487461808521225,8.171121758443288,8.393974289127462,8.617228767222334,8.84081020800747,9.064643710447195,9.28865451941987,9.512768087467714,9.736910136080063,9.961006716521007,10.184984270210949,10.408769688669572,10.632290373026509,10.855474293104155,11.078250046075748,11.300546914700272,11.522294925134366,11.74342490432002,11.963868536945352,12.183558421974572,12.402428128741798,12.620412252601984,12.837446470131043,13.053467593865799,13.268413626573027,13.482223815035677,13.694838703342716,13.906200185667862,14.116251558520963,14.324937572454306,14.532204483204657,14.738000102250309,14.942273846760791,15.14497678891524,15.346061704563816,15.545483121204628,15.743197365246962,15.939162608529548,16.133338914060705,16.325688280945144,16.51617468845996,16.704764139240208,16.891424701532134,17.076126550469507,17.258842008326315,17.439545583696148,17.61821400954604,17.79482628008979,17.969363686422632,18.141809850856376,18.31215075989093,18.48037479575489,18.64647276644472,18.810437934188634,18.97226604225792,19.131955340044957,19.28950660632366,19.444923170604675,19.59821093249386,19.74937837895932,19.89843659940851,20.04539929847362,20.190282806400017,20.333106086929117,20.473890742564212,20.612661017104404,20.749443795329487,20.88426859971564,21.017167584060047,21.1481755238904,21.27732980353388,21.40467039971915,21.53023986158461,21.654083286965903,21.776248294836638,21.89678499377752,22.015745946351448,22.13318612926477,22.249162889199173,22.3637358942032,22.47696708053827,22.588920594881383,22.699662731794064,22.80926186637707,22.917788382040516,23.025314593330894,23.13191466376976,23.237664518673146,23.342641752936725,23.446925533788875,23.55059649853225,23.653736647314634,23.756429230990545,23.858758634157915,23.96081025347721,24.062670371405304,24.164426025501626,24.26616487349049,24.36797505429036,24.469945045248167,24.572163515844345,24.674719178162018,24.777700634440897,24.88119622206382,24.985293856349855,25.09008087155324,25.195643860491604,25.302068513249253,25.40943945542201,25.517840086388738,25.62735241811091,25.73805691497491,25.8500323352029,25.96335557436524,26.078101511532193,26.19434285860328,26.312150013350312,26.431590916703446,26.552730914800062,26.675632626302228,26.800355815471338,26.92695727146749,27.055490694316955,27.1860065879631,27.318552160785607,27.45317123393934,27.589904157827725,27.72878773698776,27.869855163622884,28.01313595997923,28.158655929716797,28.306437118384927,28.456497783066563,28.6088523712129,28.763511508646395,28.92048199666808,29.07976681816407,29.241365152566495,29.4052723994869,29.57148021080461,29.739976530959567,29.910745645169317,30.08376823526177,30.259021442791497,30.43647893908571,30.616111001847475,30.797884597929173,30.981763471876587,31.16770823983542,31.355676488405745,31.545622878027,31.737499250475615,31.931254740059483,32.12683588809811,32.324186760284206,32.523249066530916,32.723962282919885,32.92626377537725,33.13008892471837,33.335371252716904,33.54204254886922,33.75003299754217,33.95927130520875,34.16968482749407,34.38119969577131,34.593740943065555,34.80723262904075,35.02159796386261,35.23675943074749,35.452638907023854,35.66915778354961,35.886237082343385,36.10379757230398,36.321759882905255,36.540044615768416,36.75857245402558,36.977264269400955,37.19604122694693,37.414824887382935,37.63353730699421,37.852101135056714,38.070439708762244,38.288477145624825,38.50613843335649,38.72334951720589,38.94003738475832,39.156130148200454,39.37155712405673,39.58624891040756,39.8001374616025,40.01315616048324,40.2252398881337,40.436325091175,40.646349846624624,40.8552539243392,41.062978847060606,41.26946794808482,41.47466642657277,41.67852140052142,41.88098195741246,42.08199920255511,42.28152630513764,42.4795185420015,42.675933339149665,42.87073031099934,43.06387129738744,43.25532039833538,43.44504400657725,43.633010837854336,43.819191958976006,44.00356081364561,44.186093246047704,44.3667675221909,44.54556434899864,44.722466891138374,44.897460785577785,45.07053415385429,45.24167761204316,45.410884278407366,45.57814977871063,45.743472249174246,45.90685233705621,46.06829319883057,46.227800495943534,46.38538238812208,46.541049524209875,46.694815030505296,46.84669449657524,46.99670595851883,47.144869879654806,47.291209128607036,47.435748954762445,47.57851696107733,47.719543074208126,47.858859511944466,47.99650074792395,48.13250347360962,48.26690655751343,48.39975100165164,48.531079895220195,48.66093836548232,48.78937352586262,48.916434421247104,49.04217197049147,49.166638906145415,49.28988971140463,49.411980554308116,49.53296921920302,49.65291503550605,49.77187880379554,49.88992271927562,50.007110292660045,50.12350626853079,50.23917654123353,50.35418806837973,50.46860878203298,50.58250749766546,50.69595382097787,50.80901805268572,50.92177109138192,51.0342843345954,51.146629578173275,51.258878914123095,51.37110462705959,51.48337908940976,51.595774655537205,51.7083635549555,51.82121778480788,51.93440900179781,52.04800841376228,52.16208667108645,52.27671375816387,52.39195888511295,52.50789037996439,52.62457558153982,52.74208073324421,52.860470877998914,52.97980975454334,53.10015969533481,53.22158152627728,53.34413446850779,53.46787604246953,53.59286197449717,53.71914610613722,53.846780306421884,53.975814387309725,54.10629602249985,54.23827066982038,54.371781497382315,54.506869313682635,54.64357250182951,54.78192695805261,54.921966034649884,55.06372048751025,55.207218428338784,55.35248528169797,55.499543746964676,55.6484137652881,55.79911249161968,55.95165427187097,56.10605062524026,56.26231023173374,56.420438924892004,56.58043968971689,56.74231266577958,56.90605515547501,57.07166163737445,57.239123784612694,57.40843048823361,57.579567885404224,57.75251939239481,57.92726574221108,58.10378502675239,58.282052743360346,58.4620418456115,58.64372279819946,58.827063635743094,59.01203002535063,59.19858533276237,59.38669069188975,59.57630507756284,59.767385381295405,59.95988648987258,60.153761366564474,60.348961134767386,60.54543516387318,60.74313115716809,60.94199524156197,61.14197205895103,61.34300485901886,61.54503559328268,61.748005010195705,61.95185275111894,62.15651744698069,62.36193681544588,62.56804775842228,62.77478645973603,62.982088482813204,63.189888868211376,63.39812223084896,63.60672285678781,63.81562479942907,64.02476197498899,64.23406825712775,64.44347757060973,64.65292398388013,64.86234180044866,65.07166564897729,65.28083057197405,65.48977211300145,65.69842640231288,65.90673024083605,66.11462118242795,66.3220376143299,66.52891883575712,66.73520513456153,66.94083786190994,67.14575950492582,67.29517446050319,67.41205099181455,67.50848507143304,67.58942429022953,67.661331212234,67.72508289639845,67.78325506233355,67.83619615501644,67.88465008255966,67.92887165655317,67.96898896725799,68.00514100498727,68.03731882077648,68.06556890058468,68.0898726251932,68.1102333078573,68.12664088059343,68.13908925085603,68.1475760654247,68.15210026855844,68.15266702054615,68.14928442947813,68.14196609589766,68.13072990482371,68.11559864934179,68.09659987428685,68.0737658435702,68.04713360250447,68.01674486051641,67.98264601296351,67.9448880579601,67.90352656415367,67.8586216079923,67.81023771508157,67.75844379721887,67.70331308241626,67.64492304379895,67.58335532196914,67.51869564452589,67.4510337405228,67.38046325082779,67.30708163394836,67.230990067294,67.15229334387918,67.0710997643021,66.98752102400321,66.90167209569098,66.8136711068973,66.72363921259438,66.63170046282615,66.53798166530952,66.44261224297023,66.34572408638782,66.24745140113497,66.14793055000847,66.04729989016266,65.94569960517053,65.84327153205453,65.74015898334503,65.6365065642448,65.53245998499635,65.42816586857096,65.32377155382049,65.2194248942571,65.11527405264928,65.01146729165094,64.9081527607038,64.80547827948345,64.70359111818566,64.60263777497896,64.50276375097799,64.40411332312094,64.30682931536352,64.21105286863009,64.11692320999022,64.02457742155681,63.93415020962674,63.845773674610484,63.759577082319495,63.67568663720123,63.594225258130635,63.5153123573827,63.439063623424694,63.3655908081766,63.295001519395754,63.22739901884517,63.1628820269052,63.10154453428394,63.04347562147418,62.98875928659266,62.93747428222082,62.88969396184652,62.845486136480986,62.8049129419968,62.76803071770049,62.734889896615805,62.705534907914526,62.68000409188759,62.65832962780267,62.64053747494476,62.62664732708491,62.61667258056739,62.61062031615092,62.60849129468218,62.61027996662262,62.615974495391995,62.625556794433834,62.63900257785247,62.65628142441407,62.67735685465121,62.702186420757876,62.73072180891221,62.76290895361713,62.79868816360597,62.837994258818895,62.880756717919674,62.92689983578962,62.976342890406436,63.02900031849082,63.084781899283996,63.14359294580194,63.20533450290147,63.26990355148423,63.33719321816188,63.40709298970513,63.479488931604095,63.554263910074084,63.631297816852374,63.71046779614534,63.79164847310207,63.8747121832098,63.95952920202859,64.04596797470592,64.13389534473717,64.22317678146554,64.31367660584142,64.40525821399163,64.49778429817718,64.59111706474849,64.68511844873669,64.77965032474971,64.87457471387076,64.96975398628658,65.06505105940056,65.16032959121388,65.2554541687839,65.35029049159549,65.44470554970472,65.53856779653803,65.63174731625192,65.7241159855791,65.81554763010637,65.90591817494733,65.99510578979024,66.08299102831619,66.16945696199726,66.25438930829651,66.33767655330362,66.41921006884976,66.49888422415474,66.57659649206646,66.65224754996068,66.72574137537359,66.79698533644498,66.86589027725351,66.93237059812837,66.99634433102325,67.05773321004077,67.11646273719491,67.1724622434997,67.22566494547122,67.27600799712906,67.32343253758113,67.36788373427387,67.40931082198712,67.44766713764955,67.4829101510479,67.51500149149877,67.54390697054845,67.5695966007619,67.59204461065751,67.61122945584037,67.62713382638151,67.63974465048625,67.64905309449058,67.65505455921836,67.65774867272903,67.65713927947954,67.6532344259198,67.64604634253685,67.63559142235736,67.62189019591469,67.60496730268129,67.58485145896431,67.56157542225688,67.53517595203498,67.50569376698515,67.47317349864588,67.4376636414416,67.39921649908621,67.35788812732967,67.31373827302,67.26683030945074,67.21723116796268,67.1650112657675,67.1102444299612,67.05300781769445,66.99338183246788,66.9314500365214,66.86729905928834,66.8010185018874,66.732700837628,66.66244130850939,66.59033781769594,66.51649081795883,66.44100319607747,66.3639801532025,66.28552908118806,66.20575943491033,66.12478260059724,66.04271176020492,65.9596617518864,65.87574892660976,65.79109100099522,65.70580690645404,65.62001663472506,65.53384107992062,65.44740187720787,65.36082123826839,65.27422178369517,65.1877263725038,65.10145792895185,65.01553926687957,64.9300929118025,64.84524092100659,64.76110470191377,64.67780482900653,64.59546085961746,64.51419114890875,64.43411266438474,64.35534080029787,64.27798919232512,64.20216953290847,64.12799138766648,64.05556201329907,63.98498617741851,63.91636598075138,63.84980068216378,63.78538652697069,63.72321657899372,63.66338055683577,63.6059646748406,63.55105148920421,63.49871974970021,63.4490442574745,63.40209572935584,63.357940669115706,63.31664124609713,63.278255181614476,63.24283564350647,63.21043114920206,63.181085477634696,63.15483759031275,63.13172156182549,63.11176652003225,63.09499659614994,63.08143088491965,63.071083414997354,63.06396312967628,63.06007387801209,63.05941441638243,63.06197842047499,63.06775450765894,63.076726269656845,63.08887231539597,63.10416632388171,63.122577106899115,63.144068681315,63.16860035071883,63.196126796111216,63.22659817531751,63.259960230778724,63.29615440534596,63.33511796568267,63.37678413285899,63.42108221970532,63.46793777447711,63.517272730371246,63.56900556042519,63.623051437322445,63.679322397624915,63.737727509950105,63.79817304661247,63.860562658251084,63.92479755097132,63.99077666553524,64.05839685814517,64.12755308237551,64.19813857182093,64.27004502304295,64.34316277841226,64.41738100846078,64.49258789337463,64.56867080327766,64.6455164769734,64.72301119883325,64.80104097353733,64.87949169839489,64.95824933299016,65.03720006591959,65.11623047840514,65.19522770458808,65.27407958832548,65.3526748363301,65.43090316751194,65.50865545839608,65.58582388450812,65.66230205763323,65.73798515887054,65.8127700674165,65.88655548502618,65.95924205611153,66.03073248344803,66.10093163947103,66.16974667315345,66.23708711246388,66.30286496241362,66.36699479870713,66.42939385701739,66.48998211791337,66.54868238747089,66.60542037360358,66.6601247581533,66.71272726478293,66.76316272271615,66.81136912637156,66.857287690939,66.90086290394727,66.94204257287217,66.980777868835,67.01702336643957,67.05073707979639,67.08188049478093,67.11041859757177,67.13631989951222,67.15955645833846,67.18010389581342,67.19794141180488,67.21305179484325,67.22542142919194,67.2350402984608,67.24190198579073,67.24600367063418,67.24734612215417,67.24593368926155,67.24177428730728,67.23487938144409,67.22526396666957,67.21294654455933,67.1979490966975,67.18029705480876,67.16001926759473,67.13714796427456,67.11171871482878,67.08377038694358,67.05334509965066,67.02048817365792,66.98524807836444,66.94767637555307,66.90782765975385,66.86575949527108,66.82153234986772,66.77520952510199,66.72685708331085,66.67654377123868,66.62434094030961,66.57032246354572,66.51456464913576,66.4571461506627,66.39814787400198,66.33765288090731,66.27574628930546,66.212515170327,66.14804844210623,66.08243676038983,66.01577240600095,65.94814916921321,65.87966223109723,65.81040804191086,65.74048419661392,65.66998930759732,65.59902287472741,65.52768515281656,65.45607701664187,65.38429982364612,65.31245527446602,65.24064527144506,65.16897177530136,65.09753666013175,65.02644156694788,64.95578775595061,64.88567595776384,64.81620622385876,64.74747777641427,64.67958885786918,64.61263658043453,64.54671677584501,64.48192384563885,64.41835061226493,64.35608817132608,64.29522574527356,64.23585053887675,64.17804759679645,64.12189966359605,64.06748704652762,64.01488748143218,63.96417600209456,63.91542481339154,63.86870316857053,63.82407725099114,63.781610060657094,63.74136130585784,63.70338730023099,63.667740865545156,63.63447124049114,63.603623995754894,63.57524095563028,63.549360126412324,63.52601563179372,63.505237655466004,63.4870523911076,63.47148199991628,63.458544575822295,63.44825411849277,63.44062051421321,63.43564952470723,63.43334278392812,63.43369780283091,63.43670798210662,63.44236263283385,63.45064700497764,63.46154232363907,63.47502583293436,63.49107084735815,63.509646810461696,63.53071936065465,63.55425040391782,63.580198193194036,63.60851741420586,63.63915927743169,63.672071615955886,63.707198988895016,63.74448279008953,63.78386136173945,63.82527011265429,63.86864164077917,63.91390585965449,63.96099012846193,64.0098193853077,64.06031628339315,64.1124013297237,64.16599302600959,64.22100801141534,64.27736120682063,64.33496596026032,64.39373419322038,64.45357654747325,64.51440253214717,64.57612067073256,64.6386386477411,64.7018634547433,64.7657015355232,64.83005893010112,64.89484141738832,64.95995465624972,65.0253043247653,65.09079625749263,65.15633658054713,65.22183184432944,65.28718915374195,65.35231629574959,65.41712186415246,65.48151538144963,65.54540741768504,65.6087097061787,65.67133525605573,65.73319846149738,65.79421520764781,65.85430297311895,65.91338092904563,65.97137003465122,66.02819312929094,66.08377502094868,66.13804257116831,66.19092477640794,66.2423528458102,66.29226027538735,66.34058291862429,66.38725905350717,66.43222944598793,66.47543740989984,66.51682886334044,66.55635238154156,66.59395924624775,66.62960349162637,66.66324194673324,66.69483427455971,66.72434300768676,66.75173358057252,66.77697435850001,66.80003666321129,66.82089479525416,66.83952605306744,66.8559107488293,66.87003222109335,66.88187684423558,66.89143403473449,66.89869625430563,66.90365900991075,66.90632085066001,66.90668336162521,66.90475115458038,66.90053185568468,66.89403609012174,66.88527746370806,66.87427254148245,66.861040823287,66.84560471634938,66.8279895048758,66.80822331666258,66.78633708673462,66.76236451801798,66.73634203905405,66.70830875876302,66.67830641826379,66.64637933975885,66.61257437249303,66.5769408357956,66.53953045921727,66.50039731977449,66.45959777631518,66.41719040102315,66.3732359080796,66.32779707950375,66.28093868819735,66.23272741822156,66.18323178233808,66.13252203685109,66.08067009379029,66.02774943048121,65.97383499655322,65.91900311844161,65.86333140144632,65.80689862941546,65.74978466212906,65.69207033046465,65.63383732943412,65.5751681091883,65.51614576409341,65.45685391999177,65.39737661976702,65.33779820734233,65.27820321024873,65.21867622090897,65.15930177679093,65.10016423959293,65.04134767363246,64.98293572361698,64.925011491985,64.86765741601253,64.81095514488844,64.7549854169691,64.6998279374304,64.64556125654029,64.59226264878295,64.54000799306922,64.4888716542738,64.43892636634317,64.39024311722197,64.34289103584693,64.2969372814604,64.2524469354948,64.2094828962797,64.16810577682148,64.1283738059035,64.09034273275014,64.05406573549524,64.01959333368754,63.98697330506086,63.956250606787684,63.92746730142587,63.90066248775877,63.87587223671706,63.85312953255913,63.832464219473216,63.81390295375114,63.797469161667784,63.783183003186195,63.771061341590126],"type":"scatter","name":"source1_i_mag_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1]},{"mode":"lines","y":[60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0,60.0],"type":"scatter","name":"source1_freq","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1]},{"mode":"lines","y":[20.0,20.000000000000004,20.0,20.0,20.0,20.0,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999993,20.0,19.999999999999993,20.0,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,19.999999999999993,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.0,20.0,20.000000000000004,20.0,20.0,20.0,20.0,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,20.0,19.999999999999993,20.0,19.999999999999993,20.0,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.0,20.0,20.000000000000004,20.000000000000004,20.0,20.0,20.0,20.0,20.0,20.0,20.0,20.0,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.000000000000007,20.00000000000001,20.000000000000007,20.00000000000001,20.000000000000007,20.00000000000001,20.000000000000007,20.00000000000001,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000007,19.999999999999993,20.000000000000007,19.999999999999993,20.000000000000007,19.999999999999993,20.000000000000007,19.999999999999993,20.000000000000007,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.0,20.0,20.000000000000004,20.0,20.0,20.0,20.0,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999993,20.0,19.999999999999993,20.0,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,19.999999999999993,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.0,20.0,20.000000000000004,20.0,20.0,20.0,20.0,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,19.999999999999996,20.0,19.999999999999993,20.0,19.999999999999993,20.0,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,20.000000000000007,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.99999999999998,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.0,20.0,20.000000000000004,20.000000000000004,20.0,20.0,20.0,20.0,20.0,20.0,20.0,20.0,20.0,20.000000000000007,20.0,20.000000000000007,20.0,20.000000000000007,20.000000000000007,20.00000000000001,20.000000000000007,20.00000000000001,20.000000000000007,20.00000000000001,20.000000000000007,20.00000000000001,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000018,20.000000000000007,19.999999999999993,20.000000000000007,19.999999999999993,20.000000000000007,19.999999999999993,20.000000000000007,19.999999999999993,20.000000000000007,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,20.000000000000018,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,19.999999999999993,20.0,20.0],"type":"scatter","name":"source1_θ","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08,0.0801,0.0802,0.0803,0.0804,0.0805,0.0806,0.0807,0.0808,0.0809,0.081,0.0811,0.0812,0.0813,0.0814,0.0815,0.0816,0.0817,0.0818,0.0819,0.082,0.0821,0.0822,0.0823,0.0824,0.0825,0.0826,0.0827,0.0828,0.0829,0.083,0.0831,0.0832,0.0833,0.0834,0.0835,0.0836,0.0837,0.0838,0.0839,0.084,0.0841,0.0842,0.0843,0.0844,0.0845,0.0846,0.0847,0.0848,0.0849,0.085,0.0851,0.0852,0.0853,0.0854,0.0855,0.0856,0.0857,0.0858,0.0859,0.086,0.0861,0.0862,0.0863,0.0864,0.0865,0.0866,0.0867,0.0868,0.0869,0.087,0.0871,0.0872,0.0873,0.0874,0.0875,0.0876,0.0877,0.0878,0.0879,0.088,0.0881,0.0882,0.0883,0.0884,0.0885,0.0886,0.0887,0.0888,0.0889,0.089,0.0891,0.0892,0.0893,0.0894,0.0895,0.0896,0.0897,0.0898,0.0899,0.09,0.0901,0.0902,0.0903,0.0904,0.0905,0.0906,0.0907,0.0908,0.0909,0.091,0.0911,0.0912,0.0913,0.0914,0.0915,0.0916,0.0917,0.0918,0.0919,0.092,0.0921,0.0922,0.0923,0.0924,0.0925,0.0926,0.0927,0.0928,0.0929,0.093,0.0931,0.0932,0.0933,0.0934,0.0935,0.0936,0.0937,0.0938,0.0939,0.094,0.0941,0.0942,0.0943,0.0944,0.0945,0.0946,0.0947,0.0948,0.0949,0.095,0.0951,0.0952,0.0953,0.0954,0.0955,0.0956,0.0957,0.0958,0.0959,0.096,0.0961,0.0962,0.0963,0.0964,0.0965,0.0966,0.0967,0.0968,0.0969,0.097,0.0971,0.0972,0.0973,0.0974,0.0975,0.0976,0.0977,0.0978,0.0979,0.098,0.0981,0.0982,0.0983,0.0984,0.0985,0.0986,0.0987,0.0988,0.0989,0.099,0.0991,0.0992,0.0993,0.0994,0.0995,0.0996,0.0997,0.0998,0.0999,0.1]}],{"xaxis":{"title":{"text":"Time in Seconds"}},"template":{"layout":{"coloraxis":{"colorbar":{"ticks":"","outlinewidth":0}},"xaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"hovermode":"closest","paper_bgcolor":"white","geo":{"showlakes":true,"showland":true,"landcolor":"#E5ECF6","bgcolor":"white","subunitcolor":"white","lakecolor":"white"},"colorscale":{"sequential":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]],"diverging":[[0,"#8e0152"],[0.1,"#c51b7d"],[0.2,"#de77ae"],[0.3,"#f1b6da"],[0.4,"#fde0ef"],[0.5,"#f7f7f7"],[0.6,"#e6f5d0"],[0.7,"#b8e186"],[0.8,"#7fbc41"],[0.9,"#4d9221"],[1,"#276419"]],"sequentialminus":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]},"yaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"shapedefaults":{"line":{"color":"#2a3f5f"}},"hoverlabel":{"align":"left"},"mapbox":{"style":"light"},"polar":{"angularaxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","radialaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"autotypenumbers":"strict","font":{"color":"#2a3f5f"},"ternary":{"baxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","caxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"aaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"annotationdefaults":{"arrowhead":0,"arrowwidth":1,"arrowcolor":"#2a3f5f"},"plot_bgcolor":"#E5ECF6","title":{"x":0.05},"scene":{"xaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"zaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"yaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"}},"colorway":["#636efa","#EF553B","#00cc96","#ab63fa","#FFA15A","#19d3f3","#FF6692","#B6E880","#FF97FF","#FECB52"]},"data":{"barpolar":[{"type":"barpolar","marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"carpet":[{"aaxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"},"type":"carpet","baxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"}}],"scatterpolar":[{"type":"scatterpolar","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"parcoords":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"parcoords"}],"scatter":[{"type":"scatter","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2dcontour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2dcontour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattercarpet":[{"type":"scattercarpet","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"mesh3d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"mesh3d"}],"surface":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"surface","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattermapbox":[{"type":"scattermapbox","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergeo":[{"type":"scattergeo","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram":[{"type":"histogram","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"pie":[{"type":"pie","automargin":true}],"choropleth":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"choropleth"}],"heatmapgl":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmapgl","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"bar":[{"type":"bar","error_y":{"color":"#2a3f5f"},"error_x":{"color":"#2a3f5f"},"marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"heatmap":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmap","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contourcarpet":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contourcarpet"}],"table":[{"type":"table","header":{"line":{"color":"white"},"fill":{"color":"#C8D4E3"}},"cells":{"line":{"color":"white"},"fill":{"color":"#EBF0F8"}}}],"scatter3d":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"scatter3d","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergl":[{"type":"scattergl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2d","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scatterternary":[{"type":"scatterternary","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scatterpolargl":[{"type":"scatterpolargl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}]}},"legend":{"yanchor":"bottom","xanchor":"right","y":1.02,"orientation":"h","x":1},"margin":{"l":100,"b":80,"pad":10,"r":80,"t":100},"plot_bgcolor":"#f1f3f7","yaxis":{"title":{"text":"State values"}},"yaxis2":{"titlefont":{"color":"orange"},"overlaying":"y","title":"Action values","side":"right"},"height":550,"width":800},{"showLink":false,"editable":false,"responsive":true,"staticPlot":false,"scrollZoom":true});
});





</script>
```




_______________________________________________________________________________
### Analysis

- The plot shows the instantaneous 3-phase real [W] and imaginary [VAi] power delivered by the source. 
- The frequency is a constant 60 Hz.
- The relative positive phase sequence angle at the terminals of the source is 20 degrees.
- The voltage at the source terminals ramps up to a magnitude of 1.05*230 V over a period of 0.04 s.
