# VSG Mode - Enhanced Grid Forming

|  | Mode | Description |
| --- | --- | --- |
| 1 | "Swing" | Ideal voltage source without dynamics (i.e., an infinite bus) |
| 2 | "PQ" | Grid following controllable source/load (real and imaginary power) |
| 3 | "Droop" | Simple grid forming with power balancing through a droop mechanism |
| 4 | __"Synchronverter" or "VSG"__|  __Grid forming control mimicking a generator, i.e., virtual synchronous generator__ |

- The "VSG" mode is more enhanced form of droop allowing users to refine the behaviour of the inverter.

## Summary

- The following example is intended to introduce you to a more advanced form of decentralised control.
- The example also demonstrates a discrete Luenberger observer for when the supply's filter is "LCL".
- A supply in "VSG" mode "communicates" with other inverters in the network through the voltage and frequency.
- This mode is autonomous and provides a quantifiable measure of "inertia" to the network.

### Theory

The specific VSG mode that is available to users is the "Synchronverter" mode. The mode aims to mimic a synchronous generator, and thereby has two more coefficients, two time constants $τv$ and $τf$, which are used to calculate the virtual mass moment of inertia, $J$ and integral gain $K$. The mode provides additional flexibility over the droop mode, in that real and imaginary power set points can be provided. Unlike the PQ mode these set points are more akin to biases, because the inverter will not necessarily output these powers. For example, the real power set point will only be reached if the VSG is connected to a swing supply with a fixed frequency. Moreover, an inverter in VSG mode may present with unwanted oscillatory behaviour similar to the hunting effect in synchronous generators.

![](assets/VSG_control.jpg)

Similar to the droop mode, if the user does not specify $D_p$ and $D_q$, then they are automatically tuned based on the converter power rating and the network's maximum allowable frequency and voltage deviations, i.e. $Δfmax$ and $ΔEmax$. The formulas to compute the droop coefficients are the same as for the droop mode and given by,

```math

  D_p = \frac{S_{rated}}{{ω_{set}}^2 Δfmax}, \quad D_q = \frac{S_{rated}}{V_{set}*\sqrt{2}*ΔEmax},
```


The two new quantities that correspond to the "inertia" are computed as follows,

```math

  J = τf*D_p, \quad K = τv*ω_{set}*D_q.
```


![](assets/VSG_mode.jpg)


```julia
using ElectricGrid;
```

_______________________________________________________________________________
### Network Configuration 

- Two inverters are connected to each other through a cable.
- The one inverter is placed in the familiar PQ mode, while the other is in VSG mode.
- The VSG inverter will regulate the frequency and the PQ inverter will follow along.
- If the inverter has an LCL filter then a discrete Luenberger observer can be implemented to estimate the two sets of currents through both inductor sets.
- If not provided by the user, the proportional and integral gains for the PI controllers are tuned automatically via a loop-shaping method.


```julia
# total run time, seconds
t_end = 0.08     

# Connectivity Matrix
CM = [ 0. 1.
        -1. 0.]     

parameters = Dict{Any, Any}(
        "source" => Any[
                        Dict{Any, Any}("pwr"    => 400e3,   # Rated appared power [VA] 
                                        "mode"  => "VSG",   # Controller mode
                                        "τv"    => 0.002,   # Time constant of the voltage loop [s] 
                                        "τf"    => 0.02,    # Time constant of the frequency loop [s]
                                        "fltr"  => "LCL",
                                        "Observer" => true) # Activating the Luenberger discrete Observer   
                        Dict{Any, Any}("pwr"    => 100e3,   
                                        "mode"  => "PQ",    
                                        "p_set" => 50e3,
                                        "q_set" => -40e3) 
                        ],
        "cable"   => Any[
                        Dict{Any, Any}("R"  => 0.1, 
                                        "L" => 0.25e-3, 
                                        "C" => 0.1e-4),
                        ],
    );

env = ElectricGridEnv(CM = CM, parameters = parameters, t_end = t_end, verbosity = 2);

agents = SetupAgents(env);

hook = Simulate(agents, env);
```

    ┌ Info: Normalization is done based on the defined parameter limits.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\electric_grid_env.jl:329
    ┌ Info: Time simulation run time: 0.08 [s] ~> 801 steps
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\electric_grid_env.jl:330
    ┌ Info: 2 'classically' controlled sources have been initialised.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2686
    ┌ Info: 1 source has been set up in PQ mode.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2695
    ┌ Info: 1 source has been set up in Synchronverter mode.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2695
    ┌ Info: All 'classically' controlled sources have been automatically set up with droop coeficients, and proportional and integral gains.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2707
    ┌ Info: 1 source has been set up with a Luenberger discrete Observer.
    └ @ JEG c:\Gitlab\JEG\ElectricGrid.jl\src\classical_control.jl:2737
    

_______________________________________________________________________________
### Low-Level Rendering


```julia
RenderHookResults(hook = hook, 
                    states_to_plot  = ["source1_i_L1_a"], # Inductor current [A]
                    actions_to_plot = ["source1_u_a"],    # Inverter voltage [V]
                    )
```


```@raw html
<div
id = U3JGKw5tghl5 > </div>
<script>
gd = 'U3JGKw5tghl5'
require(['plotly'], function(plotly) {
























plotly.newPlot(gd,[{"mode":"lines","y":[0.0,0.0,0.016693190268572657,0.051314248994000194,0.10324574742559063,0.17263547462066833,0.25955250639941757,0.36326668396496054,0.4824355566555967,0.6147608557774519,0.7573889217937716,0.9067627483879863,1.0590254351455384,1.2101183040223673,1.3560574963531733,1.493133003209069,1.6180205034974224,1.727945992949045,1.820666870694975,1.8945409724754316,1.948447163275777,1.9817726719845072,1.9943326148347356,1.9863141329849108,1.9582166259041227,1.9107820817591221,1.8449508538562454,1.761794858957356,1.6624802586940732,1.5482141950275532,1.4202133431780852,1.279669806203206,1.1277282608293728,0.965469973895147,0.793899921251649,0.6139439526373525,0.426445457327346,0.23217061302474934,0.031812762063865824,-0.17399835878836423,-0.384687682248044,-0.5997247729150602,-0.8186136714439488,-1.0408839594976271,-1.2660817039114272,-1.493762425240844,-1.723484343094537,-1.9548035660907463,-2.187270082823693,-2.4204252452880817,-2.6538003351204926,-2.886916078310954,-3.1192831718671616,-3.350403312875122,-3.579770950179953,-3.8068752521479525,-4.031202440530547,-4.252238157823122,-4.469469875220586,-4.682389202975651,-4.89049401753056,-5.093290406943962,-5.290294335349761,-5.481033089182537,-5.665046441055842,-5.841887596744337,-6.011123910925756,-6.172337418472911,-6.325125207971706,-6.469099667590075,-6.6038886496599325,-6.729135568115776,-6.844499460448465,-6.949655016444436,-7.044292621255222,-7.128118453455134,-7.20085466310735,-7.262239619508539,-7.312028201620541,-7.349992113432888,-7.37592021254172,-7.389618852934119,-7.39091223717422,-7.379642773469804,-7.3556714278494315,-7.318878062943359,-7.269161755768747,-7.206441087650616,-7.130654401155603,-7.041760018486231,-6.939736417947387,-6.824582365508625,-6.696317000676174,-6.554979876736568,-6.4006309566647115,-6.23335056674189,-6.053239310374065,-5.860417945205677,-5.6550272266287385,-5.437227721051146,-5.207199591994646,-4.96514236196883,-4.71127465265348,-4.445833905558085,-4.169076084916671,-3.8812753641188116,-3.582723796601262,-3.2737309716855094,-2.9546236555354435,-2.6257454170796373,-2.287456238532539,-1.9401321099714124,-1.584164607328948,-1.2199604531259736,-0.8479410592783079,-0.4685420513831007,-0.08221277398853212,0.310584223507999,0.7093737205321441,1.113668378599255,1.5229693140234615,1.9367666897419578,2.3545403257614357,2.775760327583911,3.1998877318324506,3.6263751681850946,4.054667536630311,4.484202698986027,4.9144121835708425,5.344721901884784,5.774552876140801,6.203321976486154,6.630442666763352,7.055325757681791,7.477380166295083,7.896013680709475,8.310633728981623,8.720648151194244,9.125465973730249,9.524498184790481,9.91715851022577,10.302864188771451,10.681036745788804,11.051102764627466,11.412494654730075,11.764651415602655,12.107019395775836,12.439053045880662,12.760215664956778,13.069980139111044,13.367829671638694,13.653258503716103,13.925772624773801,14.184890471656177,14.430143615679377,14.661077436700081,14.877251783315153,15.078241618323784,15.263637648590612,15.433046938467351,15.586093505941514,15.722418900703731,15.84168276334301,15.943563364899289,16.027758126030935,16.093984115074125,16.141978524299066,16.171499123695863,16.182324691647146,16.174255421873905,16.147113306070388,16.100742491670772,16.03500961422052,15.94980410385589,15.845038465422773,15.720648531798815,15.576593690010833,15.412857079772964,15.229445764100367,15.02639087168576,14.80374771075929,14.561595854181144,14.300039195553815,14.019205976170218,13.71924878264972,13.400344515146623,13.062694326050904,12.70652352913546,12.332081479136612,11.939641421793887,11.529500314404626,11.101978616989017,10.657420054192949,10.196191348093127,9.718681922102688,9.225303576211264,8.716490133827426,8.19269706052593,7.654401055038678,7.102099612857491,6.536310562855385,5.957571577362096,5.3664396561648395,4.763490584936019,4.14931836862241,3.5245346403595765,2.889768046507317,2.2456636084315913,1.592882061686678,0.9320991732807714,0.26400503773522055,-0.4106966473243777,-1.0912893252823883,-1.7770443435624417,-2.4672217347581835,-3.161071015028797,-3.857831998895574,-4.556735629562133,-5.257004823850259,-5.957855330829735,-6.658496603200528,-7.35813268046429,-8.05596308291058,-8.751183715421353,-9.442987780088528,-10.130566696620297,-10.813111029502492,-11.489811420870149,-12.159859528030813,-12.822448964576385,-13.476776244009244,-14.122041724803692,-14.75745055581573,-15.382213620953635,-15.99554848201519,-16.596680318598004,-17.18484286398748,-17.75927933592869,-18.319243361190146,-18.863999892828506,-19.392826119072367,-19.905012362744092,-20.399862970148575,-20.876697188364627,-21.33485002988506,-21.773673123560794,-22.19253555081805,-22.590824666128718,-22.967946900729483,-23.32332854860004,-23.65641653372869,-23.966679157707958,-24.25360682672696,-24.516712757041365,-24.75553365802833,-24.969630391950865,-25.158588609583312,-25.32201936087234,-25.459559679831557,-25.570873142897245,-25.65565039999644,-25.713609677609952,-25.744497253138825,-25.74808789991495,-25.724185302226246,-25.672622439757482,-25.59326194088237,-25.485996404273173,-25.35074868832905,-25.18747216795869,-24.996150958286577,-24.776800104890285,-24.529465740209684,-24.254225205807792,-23.95118714019748,-23.620491531989604,-23.2623097381525,-22.876844467212926,-22.4643297272677,-22.02503073871235,-21.559243811634488,-21.067296187856158,-20.549545847651743,-20.006381281207045,-19.43822122492241,-18.845514362705657,-18.228738992437915,-17.58840265783596,-16.925041745975182,-16.239221050773935,-15.531533302780899,-14.802598665646126,-14.053064199694369,-13.28360329305721,-12.494915060859183,-11.687723712989085,-10.862777891026672,-10.020849974929357,-9.162735360121221,-8.289251705658895,-7.40123815418761,-6.499554524432464,-5.585080477001781,-4.6587146543151725,-3.721373795498186,-2.773991827118127,-1.817518930662762,-0.8529205876973541,0.11882339634189025,1.0967198887268141,2.079763445376306,3.066937345115446,4.057214641189226,5.0495592289472695,6.042926928585994,7.036266581824238,8.028521161358437,9.018628891931959,10.005524381833986,10.988139763625876,11.965405842881417,12.936253253710602,13.899613619825725,14.85442071990187,15.799611655965366,16.73412802354537,17.656917082310695,18.56693292591168,19.46313764974185,20.34450251533214,21.210009110088357,22.058650501083584,22.889432381620793,23.701374209279162,24.493510334166675,25.26489111610583,26.0145840294875,26.74167475453396,27.44526825372697,28.124489832162343,28.778486180613776,29.40642640009183,30.007503006712067,30.580932915692102,31.125958403325246,31.641848045789818,32.12789763368173,32.58343106117364,33.0078011887336,33.400390678356494,33.7606128002894,34.08791221025835,34.38176569623428,34.641682893805104,34.86720696924748,35.05791526943202,35.213419937717745,35.333368495037206,35.41744438539655,35.4653674850603,35.476894574725364,35.45181977401769,35.3899749376972,35.29123001298785,35.15549335748681,34.98271201715586,34.772871963933675,34.52599829255193,34.24215537617778,33.921446980553995,33.56401633634713,33.17004616945876,32.73975868910151,32.27341553348491,31.771317673002105,31.23380527085277,30.661257501084155,30.054092324080813,29.412766219570866,28.737773877275963,28.029647845365876,27.288958136931342,26.516311794735017,25.712352414540803,24.877759627376314,24.013248541119005,23.119569141850317,22.197505655459757,21.24787587003319,20.271530419593866,19.26935202981861,18.242254726388044,17.191183006672663,16.117110975503245,15.021041445808576,13.904005004949294,12.767059047614925,11.611286776187441,10.437796169519201,9.247718921101864,8.042209347646203,6.8224432691258485,5.52411097495876,3.0627835562860493,-3.4510064765147006,-18.881855939457463,-44.718228419665884,-75.98103707208993,-104.13156299006253,-122.04924893576573,-127.36382823293452,-121.87830319479184,-109.44426966944685,-93.77179455364505,-77.5860413745296,-62.56523062668647,-49.594017357316396,-39.005566502545484,-30.696649957134472,-24.327570197329806,-19.4552858960798,-15.698313105495384,-12.773540330389977,-10.49664024985264,-8.734885580513668,-7.366969746206992,-6.271697886127264,-5.320849458511923,-4.395693695919367,-3.38871656460042,-2.2156432075275463,-0.8159293935649579,0.8434923170251563,2.769628685095521,4.948163761898698,7.348716009144228,9.933786984721753,12.663694038244259,15.502293047308996,18.418162983459585,21.385728046474693,24.38463020291553,27.399213061211455,30.41820129180684,33.43394350365009,36.442085993952794,39.440431728811205,42.4281949428222,45.404870238992125,48.36956352613812,51.32043651221197,54.254428942596626,57.16720504852805,60.05312022993428,62.90538758823554,65.71620796237997,68.47707758709112,71.17908566466971,73.81328863839387,76.37104844603832,78.8443228321255,81.2258990872794,83.5095376971622,85.69007755791432,87.76346682300411,89.72676435710204,91.57808227746341,93.31649317954322,94.9418979058347,96.45486838682247,97.85648306922647,99.14816098275483,100.3315127472076,101.4082054215327,102.37985119652913,103.24791545270882,104.01364842064929,104.67803903683385,105.24179065781443,105.70531777015928,106.06875923873145,106.3320061545802,106.49473890781219,106.55647155271394,106.5165997587254,106.37445071030444,106.129332760183,105.78058304205555,105.32761170668148,104.76994124886205,104.10724046727739,103.33935237472434,102.46631627332846,101.48838399626308,100.40603069545878,99.21996048115568,97.93110728957734,96.54063148860283,95.04991268857908,93.46053940283092,91.77429609158705,89.99314820769835,88.11922575298739,86.15480586493835,84.10229490686713,81.9642105099148,79.74316400423501,77.44184361625975,75.06299878531371,72.60942587216414,70.08395549064755,67.48944162113077,64.82875261492254,62.10476414044943,59.32035406544309,56.47839921842217,53.581773916267956,50.63335010347321,47.63599890602165,44.59259337707637,41.506012189340524,38.379144018683526,35.21489235986238,32.01618051887508,28.785956538618752,25.52719783168881,22.24291531911127,18.936156901651962,15.610010122791554,12.267603915682722,8.91210936070319,5.546739413807565,2.1747475980129423,-1.2005743194226557,-4.575899614564862,-7.947870785664842,-11.313104530269634,-14.668197411910814,-18.009732096549183,-21.334284031974395,-24.638428440448703,-27.918747495794634,-31.17183756044288,-34.39431636541484,-37.58283002622515,-40.73405979988249,-43.844728501930966,-46.91160651733133,-49.93151735439081,-52.90134270645902,-55.81802700125968,-58.67858143212779,-61.48008747874175,-64.21969993686604,-66.89464948695111,-69.50224483996932,-72.03987450554759,-74.5050082322133,-76.8951981724532,-79.20807982634942,-81.44137281694255,-83.59288154833031,-85.66049579402282,-87.6421912584925,-89.536030149359,-91.34016179152867,-93.05282330805426,-94.67234038575381,-96.19712813692797,-97.62569206203376,-98.9566291120988,-100.18862884410733,-101.32047465771184,-102.35104509747447,-103.27931520149156,-104.10435787475144,-104.8253452638788,-105.44155010904464,-105.95234704871312,-106.35721385348909,-106.65573256654876,-106.84759052989847,-106.93258127791135,-106.91060528213973,-106.78167053419257,-106.5458929564027,-106.2034966329921,-105.75481385738203,-105.20028499413043,-104.54045815660015,-103.77598870386826,-102.9076385624738,-101.9362753803816,-100.86287152196836,-99.68850291390744,-98.41434775255183,-97.04168508378953,-95.57189326640756,-94.00644832975996,-92.34692223604891,-90.59498105680962,-88.7523830723088,-86.8209768015425,-84.80269896941132,-82.6995724164974,-80.51370395570792,-78.24728217892242,-75.90257521572289,-73.48192844531506,-70.98776216189817,-68.42256919302912,-65.78891246995403,-63.08942254846419,-60.3267950785767,-57.50378822122799,-54.623220010206666,-51.687965657716674,-48.70095480225954,-45.665168697898444,-42.583637344454026,-39.459436558711836,-36.295684987303744,-33.09554106253634,-29.862199903042946,-26.598890161746734,-23.308870824187853,-19.995427960806246,-16.661871437245555,-13.31153158716634,-9.947755852404622,-6.573905395586584,-3.193351690514456,0.19052690423453988,3.5743485830009982,6.954731536540004,10.328297395916545,13.691674671514825,17.041502182061755,20.37443246877617,23.68713519000471,26.976300491969106,30.238642351531137,33.470901887167074,36.66985063461983,39.8322937839716,42.9550733751314,46.03507144896625,49.06921315151463,52.05446978889992,54.987861830718444,57.86646185979733,60.68739746631236,63.44785408432572,66.14507776884233,68.77637791150408,71.33912989304143,73.83077767058276,76.24883629789628,78.59089437660064,80.85461643633761,83.03774524186196,85.13810402496078,87.15359863908307,89.08221963453649,90.92204425209304,92.67123833284799,94.32805814218854,95.89085210576187,97.35806245537296,98.72822678280993,99.99997949966675,101.17205320132905,102.2432799333917,103.21259235889248,104.07902482487954,104.84171432695994,105.49990137062693,106.05293072830601,106.50025209121809,106.84142061530531,107.07609736062454,107.20404962376098,107.22515116296645,107.13938231586918,106.94683000974737,106.6476876644854,106.24225498846855,105.73093766779029,105.11424694926198,104.39279911782711,103.56731486908443,102.63861857772177,101.6076374627609,100.4754006505915,99.24303813687166,97.91177964844172,96.48295340648185,94.9579847922232,93.33839491659516,91.62579909526521,89.8219052306066,87.92851210219995,85.947507567552,83.88086667479098,81.73064968917103,79.49900003529983,77.18814215708032,74.80037929743249,72.33809119994432,69.80373173467872,67.1998264504408,64.52897005588814,61.793823831946945,58.99711297806826,56.14162389493591,53.23020140630869,50.26574592274806,47.251210550050075,44.18959814526351,41.08395832323623,37.93738441668469,34.75301039283947,31.53400772976321,28.28358225548055,25.004970953103424,21.70143873516543,18.376275190412706,15.032791306323825,11.674316170654192,8.304193655318434,4.92577908593697,1.5424359003868036,-1.8424677003013947,-5.225562105351826,-8.603479648610662,-11.972857960531943,-15.330343314441619,-18.672593963757933,-21.9962834668534,-25.29810399626509,-28.574769628970948,-31.82301961447986,-35.03962161750045,-38.22137493198554,-41.36511366337334,-44.4677098758851,-47.526076701770975,-50.53717140942786,-53.497998427361104,-56.40561232100364,-59.257120719445496,-62.04968718918022,-64.7805340520266,-67.44694514443555,-70.0462685154459,-72.57591906061887,-75.03338108933967,-77.4162108229354,-79.7220388211342,-81.94857233445674,-84.09359758020211,-86.15498193976829,-88.13067607512646,-90.01871596234204,-91.81722484012893,-93.52441507149835,-95.1385899166546,-96.65814521538033,-98.08157097723895,-99.40745287802315,-100.63447366096563,-101.76141444132301,-102.78715591304795,-103.71067945635546,-104.5310681450936,-105.24750765292336,-105.85928705742305,-106.36579954132424,-106.7665429901962,-107.06112048599643,-107.24924069600736,-107.33071815678791,-107.30547345286607,-107.17353329001114,-106.93503046302067,-106.59020371806992,-106.13939750976819,-105.5830616531779,-104.92175087115136,-104.15612423744315,-103.28694451616515,-102.31507739824671,-101.2414906356691,-100.06725307434098,-98.793533586586,-97.42159990430918,-95.95281735400778,-94.38864749489224,-92.7306466614738,-90.98046441207772,-89.13984188482395],"type":"scatter","name":"source1_i_L1_a","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.03041537489864649,0.07024232777644975,0.12124374343385264,0.18885719340307397,0.2765592782704914,0.38619023007246367,0.519511533879434,0.6773035141964086,0.8605552876665699,1.0694700115282967,1.3042952862111867,1.5647392348801377,1.8502189686523876,2.1597250104526893,2.4917092916721417,2.844247260752211,3.214831997238453,3.600636900003357,3.998386754724327,4.404581221940208,4.815482774780555,5.227244084561615,5.635974234521787,6.037780906008034,6.42885966553571,6.805491696972366,7.164118310402752,7.501330452721231,7.813912978707623,8.0988412469136,8.35329508701571,8.574661177257948,8.760525846067575,8.908677473915608,9.017090223458903,9.083922801318383,9.107502080636076,9.086317491399203,9.019008505417151,8.90435631891161,8.741275898034884,8.528807712351243,8.266113538504742,7.952470324030367,7.587268405909548,7.1700081542781815,6.700299478338574,6.177860890572806,5.602519711700259,4.974212742232877,4.292986866514968,3.5590003866458124,2.772523733856295,1.933940669750423,1.0437488097551977,0.10256022644374008,-0.8888984534505053,-1.9297862172482254,-3.0191481666355875,-4.155916464035438,-5.338911513952216,-6.566843603052646,-7.838314738483736,-9.151820852413051,-10.50575419452891,-11.898405972732206,-13.327969172713095,-14.792541530662225,-16.29012866285111,-17.818647290283657,-19.37592858392203,-20.959721564195963,-22.56769661074916,-24.197449066118548,-25.846502921026342,-27.51231454490821,-29.192276452373246,-30.883721130280712,-32.58392493775925,-34.290112100413886,-35.99945878949541,-37.70909728711459,-39.41612022822687,-41.11758491895564,-42.81051772833924,-44.49191854884103,-46.158765321631925,-47.80801861810337,-49.436626272533516,-51.04152805716717,-52.61966039344803,-54.16796109072594,-55.68337410447485,-57.16285430553859,-58.60337225180762,-60.00191895450124,-61.35551063079774,-62.66119343560172,-63.91604816492684,-65.1171949242455,-66.26179775520968,-67.34706921469865,-68.37027490045867,-69.32873791783273,-70.21984328244095,-71.04104225371958,-71.78985659451762,-72.46388275193424,-73.06079595474235,-73.57835422272143,-74.01440228325616,-74.36687539054354,-74.63380304271058,-74.81331259213762,-74.90363274423665,-74.90309693993676,-74.8101466171158,-74.62333434624642,-74.34132683555596,-73.96290780106955,-73.48698069699238,-72.91257130199781,-72.23883015713305,-71.4650348512074,-70.59059214971562,-69.61503996354347,-68.53804915392257,-67.35942517032325,-66.07910951821461,-64.69718105386666,-63.21385710361723,-61.62949440527645,-59.944589869601415,-58.159781160022014,-56.27584708905046,-54.29370783005868,-52.214424943349016,-50.03920121569355,-47.76938031274941,-45.40644624400015,-42.95202264010178,-40.40787184274679,-37.77589380738596,-35.05812481937744,-32.25673602435488,-29.374031773832428,-26.412447787293736,-23.374549132224356,-20.263028023780123,-17.080701446002873,-13.83050859671456,-10.515508158447869,-7.138875397982964,-3.703899097289293,-0.21397831888291677,3.327380991177775,6.916569559209769,10.54987849509418,14.223503154869405,17.933547150674535,21.67602650377803,25.446873936231917,29.241943296513156,33.05701411431123,36.88779627945395,40.72993483978815,44.57901491265603,48.430566704454066,52.280070632604904,56.122962544119105,59.95463902478642,63.77046279290104,67.56576817129287,71.3358666313226,75.07605240237794,78.78160814030971,82.44781064814161,86.06993664230004,89.64326855752998,93.16310038357747,96.6247435266684,100.02353268873752,103.35483175732294,106.61403969899146,109.79659644912962,112.89798879090863,115.9137562162087,118.83949676129171,121.67087280999252,124.40361685722625,127.03353722561022,129.55652372803436,131.96855326904227,134.26569537793006,136.44411766652024,138.50009120462659,140.4299958063001,142.2303252200092,143.8976922160095,145.42883356423656,146.82061489616652,148.0700354441938,149.17423265219554,150.13048665107382,150.93622459320596,151.58902483986893,152.08662099585365,152.42690578564287,152.60793476568492,152.62792986747,152.48528276628835,152.17855807073602,151.70649632822222,151.06801684192664,150.2622202948652,149.28839117691243,148.14600001086808,146.83470537385074,145.35435571053657,143.70499093499336,141.88684381808022,139.90034115763942,137.74610472892888,135.42495201300883,132.93789670103203,130.286148972643,127.47111554695569,124.4943995048228,121.35779988138515,118.06331102814298,114.61312174406378,111.00961417549891,107.25536248495924,103.35313128906023,99.30587386622312,95.11673013498299,90.78902440402837,86.32626289536887,81.7321310422855,77.01049056400198,72.16537631926944,67.20099294133023,62.12171125698484,56.93206449275312,51.63674427137324,46.24059640214556,40.74861646887578,35.165945219425595,29.497863761121547,23.74978856652302,17.92726629427474,12.035968430020583,6.081685752566039,0.07032263071782618,-5.992108843567403,-12.099494879871225,-18.245626165777455,-24.42420418885761,-30.628847697718335,-36.853099295473996,-43.090432158816256,-49.33425687568741,-55.57792839439193,-61.81475307682158,-68.03799584832319,-74.24088743658682,-80.41663169180254,-86.5584129802047,-92.65940364300147,-98.71277151258391,-104.71168747779829,-110.64933308998141,-116.51890820136701,-122.31363862740727,-128.02678382447706,-133.65164457438112,-139.1815706670361,-144.60996857265997,-149.93030909477957,-155.13613499534304,-160.22106858322377,-165.17881925740176,-170.00319099611525,-174.68808978330807,-179.22753096371991,-183.61564651801436,-187.84669224939185,-191.91505487318886,-195.81525900104333,-199.5419740112856,-203.0900207973021,-206.45437838572053,-209.63019041637696,-212.61277147613944,-215.39761327879913,-217.98039068336888,-220.35696754328362,-222.5234023791443,-224.4759538678227,-226.21108614090977,-227.72547388566954,-229.01600724186264,-230.0797964879886,-230.91417651071257,-231.51671105144595,-231.88519672428617,-232.01766679973153,-231.91239474884276,-231.5678975427486,-230.98293870265186,-230.15653109574959,-229.0879394727306,-227.77668274280342,-226.2225359824502,-224.42553217441605,-222.3859636736969,-220.10438339759224,-217.58160573718257,-214.81870718787732,-211.81702669699288,-208.57816572661542,-205.10398803030566,-201.39661914253338,-197.4584455800122,-193.2921137544455,-188.90052859649222,-184.28685189109086,-179.45450032459,-174.4071432444596,-169.14870013267227,-163.68333779416145,-158.0154672620945,-152.1497404220019,-146.09104635713135,-139.84450741770635,-133.41547501709275,-126.80952515817248,-120.03245369355491,-113.09027132353752,-105.98919833606462,-98.73565909319495,-91.33627626892387,-83.79786484346435,-76.12742585940892,-68.33213994545133,-60.41936061364828,-52.39660733645305,-44.27155841003706,-36.05204361066565,-27.7460366511532,-19.36164744466589,-10.907114183383628,-2.3907952397733565,6.178839101570407,14.793215073311607,23.44366400760708,32.12143106257018,40.817684084188606,49.523522594717264,58.229986898382,66.92806729510123,75.60871339274155,84.26284350828429,92.88135414816469,101.45512955789269,109.97505133096573,118.43200806696589,126.81690506865688,135.12067406779218,143.3342829692842,151.44874560332178,159.4551314749683,167.34457550073574,175.10828772159775,182.7375629818921,190.2237905635586,197.55846376514546,204.73318941507142,211.73969730861538,218.56984955817774,225.21564984639747,231.66925257176337,237.92297187645536,243.96929054620594,249.8008687721008,255.4105527643186,260.79138320795374,265.93660355116924,270.8396681160995,275.49425002304946,279.8942489187087,284.03379849928257,287.907273819607,291.50929837952754,294.8347509790155,297.87877233371154,300.63677144281826,303.10443170148295,305.27771675006954,307.15287605296487,308.6960350310586,309.925033996609,311.06024887124,309.8662229083026,309.0162830512153,313.79223654728906,325.3853588724968,339.73752451681196,350.3056519383583,353.1138800592016,347.7840661588949,336.80785939644676,323.1593382182636,309.1708612213128,296.22674182007256,284.95782715338106,275.6080334854178,268.0354929432888,261.9345401219873,256.8376746907705,252.32972897262388,248.07836177867054,243.87829771254718,239.61604239721558,235.2114635387259,230.6098352331178,225.74481429425228,220.56583369180998,215.02242832591224,209.083067741743,202.7272897620457,195.949894892142,188.7610584929709,181.18177950564194,173.24577247507082,164.99010585925816,156.45542123562862,147.67836943486458,138.69207496356492,129.52399986249375,120.19692996149836,110.73012903325719,101.13945404453813,91.43892768128593,81.64016264146107,71.75365943408435,61.78860580129629,51.75388787967944,41.65835927842846,31.51126452978545,21.32249914816867,11.102475216686143,0.862250475712781,-9.386760008979216,-19.63271746956293,-29.86358504296673,-40.067168200358516,-50.231290507116256,-60.34393713116781,-70.39336640112384,-80.36824119970353,-90.25764575437101,-100.05114012899807,-109.7387221707542,-119.31082662776706,-128.7582802028442,-138.07226680346199,-147.2442797820868,-156.26606166000033,-165.12955355563824,-173.82683427672302,-182.35008368528975,-190.69154654993324,-198.84351876465607,-206.79833895205974,-214.5483920354686,-222.08611996841915,-229.40403730255073,-236.49475468691037,-243.35100366137712,-249.9656667052092,-256.3318056290483,-262.4426905642142,-268.29182574116095,-273.87297287074114,-279.18017159349995,-284.2077565892729,-288.95037225691044,-293.4029839000967,-297.5608865932919,-301.4197110165132,-304.97542732956487,-308.22434708517574,-311.1631238447555,-313.7887528470925,-316.09856989978533,-318.09024988426245,-319.7618048177842,-321.1115818196854,-322.13826093499677,-322.84085303883836,-323.21869780646153,-323.2714618015846,-322.9991366783041,-322.40203744879466,-321.48080083383593,-320.2363836358519,-318.67006115759887,-316.78342561657047,-314.57838455244683,-312.05715918496014,-309.2222826929973,-306.0765983833503,-302.6232577103683,-298.8657181219827,-294.807740690134,-290.4533874969887,-285.80701873367644,-280.87328947767173,-275.6571461112597,-270.16382234982075,-264.39883485382717,-258.3679784019933,-252.07732061174565,-245.53319619618804,-238.74220075624476,-231.71118411126582,-224.44724317998364,-216.9577144293835,-209.25016591521444,-201.33238894286242,-193.2123893807135,-184.89837866154986,-176.39876450879362,-167.72214142591673,-158.87728098695965,-149.87312196550027,-140.7187603376125,-131.4234391919913,-121.9965385775399,-112.44756531531186,-102.78614279826519,-93.02200079848579,-83.16496529792651,-73.22494835501907,-63.2119380160744,-53.135988277151824,-43.007209099214506,-32.83575647689049,-22.631822559088725,-12.405625818127382,-2.1674012628601447,8.072609309386772,18.30416702754976,28.51704563352101,38.701041145568354,48.84598150234533,58.94173619019353,68.97822585501513,78.94543189840785,88.8334060560696,98.63227995468051,108.33227464174746,117.92371008113531,127.39701460541907,136.7427343146832,145.95154241008606,155.01424844939075,163.9218075107442,172.66532925033124,181.23608683905348,189.6255257631796,197.8252724739108,205.82714287097036,213.62315060574542,221.20551519003163,228.56666989710698,235.6992694426581,242.59619743395766,249.2505735766213,255.65576062923472,261.805371097125,267.69327365750513,273.31359930916005,278.6607472407209,283.7293904124153,288.51448084690657,293.0112546255632,297.2152365870392,301.12224472561644,304.7283942871612,308.0301015609193,311.02408736566025,313.7073802289008,316.0773192581086,318.1315567029002,319.86806020734025,321.28511475149355,322.3813242814364,323.15561302695016,323.60722650616947,323.73573221649616,323.54102001115666,323.0233021608676,322.18311310019084,321.02130885830354,319.5390661740972,317.7378812957372,315.6195684650566,313.18625808745446,310.4403945882861,307.3847339570657,304.02234098118385,300.3565861712213,296.391142380355,292.1299811207401,287.57736858021383,282.7378613430375,277.61630181884175,272.2178133843343,266.5477952427158,260.6119170061513,254.41611300697681,247.96657634368634,241.26975266804612,234.33233371999643,227.16125061726103,219.76366690684333,212.14697138582332,204.31877069906795,196.2868817216652,188.05932373406512,179.6443103980551,171.05024154186154,162.28569476276826,153.35941685579118,144.28031507703696,135.05744825049183,125.70001772707096,116.2173582048635,106.6189284195972,96.91430171442609,87.11315649824475,77.22526660180593,67.26049154100835,57.22876669680335,47.14009342124293,37.004529079280445,26.83217703599715,16.633176599003253,6.417692925831012,-3.8040930938147755,-14.021994971021853,-24.225830755885248,-34.405433170974774,-44.55065973002705,-54.651402831826715,-64.69759981924183,-74.67924299337791,-84.5863895728547,-94.40917158823396,-104.13780570168076,-113.76260294200793,-123.27397834532307,-132.66246049159406,-141.9187009275449,-151.0334834664222,-159.9977333552828,-168.80252630061034,-177.439097343206,-185.89884957346672,-194.1733626783405,-202.25440131141207,-210.13392327778644,-217.80408752561064,-225.25726193630362,-232.48603090575318,-239.4832027089838,-246.24181664099933,-252.75514992675733,-259.0167243934554,-265.02031289856285,-270.75994550726193,-276.2299154132412,-281.4247845970111,-286.3393892161937,-290.968844722497,-295.308550700351,-299.35419542246177,-303.1017601178174,-306.54752294794775,-309.68806268755935,-312.52026210590543,-315.0413110455946,-317.2487091958057,-319.1402685571846,-320.7141155960037,-321.96869308547224,-322.9027616323836,-323.515400887608,-323.8060104392436,-323.7743103875553,-323.42034160115116,-322.74446565415315,-321.7473644444422,-320.43003949338254,-318.7938109277306,-316.84031614476703,-314.57150816199896,-311.98965365309823,-309.0973306720495,-305.8974260678022,-302.3931325920227,-298.5879457028504,-294.4856600678701,-290.0903657698065,-285.4064442187448,-280.4385637749755,-275.1916750868525,-269.67100614832253,-263.88205708107984,-257.8305946465655,-251.5226464932918,-244.96449514524645,-238.16267173737836,-231.12394950441995,-223.85533702954388,-216.36407125958854,-208.65761029382108,-200.74362595341833,-192.62999613908028,-184.32479698438144,-175.83629481267823,-167.17293790558497,-158.34334809119943,-149.3563121604651,-140.22077312020045,-130.94582129150777,-121.5406852624028,-112.01472270368012,-102.37741105713474,-92.63833810541078,-82.80719243284426,-72.89375378679604,-62.907883349054764,-52.859513926972234,-42.75864007409643,-32.615308150121905,-22.439606330028155,-12.241654572345821,-2.0315945565240816,8.180420400603774,18.38423543923896,28.56970425187713,38.7266991693828,48.84512122786395,58.91491020632963,68.92605462516592,78.86860169549927,88.73266720956107,98.508445362227,108.18621849398869,117.75636674566417,127.20937761527709,136.5358554076099,145.7265305670481,154.77226888446154,163.66408056897075,172.39312917560764,180.9507403800128,189.32841059145485,197.51781539563643,205.51081781890903,213.2994764057071,220.87605310117965,228.2330209312189,235.3630714722626,242.25912210346965,248.91432303408502,255.32206409902156,261.47598131594066,267.36996319731935,272.9981568112698,278.3549735850941,283.4350948458428,288.2334770923802,292.7453569937489,296.966256108888,300.8919853230283,304.5186489963902,307.8426488210745,310.8606873823432,313.56977142076755,315.96721479202745,318.05064112144663,319.81798615063997,321.26749977397384,322.3977477628253,323.20761317596623,323.6962974546747],"type":"scatter","name":"source1_u_a","yaxis":"y2","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]}],{"xaxis":{"title":{"text":"Time in Seconds"}},"template":{"layout":{"coloraxis":{"colorbar":{"ticks":"","outlinewidth":0}},"xaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"hovermode":"closest","paper_bgcolor":"white","geo":{"showlakes":true,"showland":true,"landcolor":"#E5ECF6","bgcolor":"white","subunitcolor":"white","lakecolor":"white"},"colorscale":{"sequential":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]],"diverging":[[0,"#8e0152"],[0.1,"#c51b7d"],[0.2,"#de77ae"],[0.3,"#f1b6da"],[0.4,"#fde0ef"],[0.5,"#f7f7f7"],[0.6,"#e6f5d0"],[0.7,"#b8e186"],[0.8,"#7fbc41"],[0.9,"#4d9221"],[1,"#276419"]],"sequentialminus":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]},"yaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"shapedefaults":{"line":{"color":"#2a3f5f"}},"hoverlabel":{"align":"left"},"mapbox":{"style":"light"},"polar":{"angularaxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","radialaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"autotypenumbers":"strict","font":{"color":"#2a3f5f"},"ternary":{"baxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","caxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"aaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"annotationdefaults":{"arrowhead":0,"arrowwidth":1,"arrowcolor":"#2a3f5f"},"plot_bgcolor":"#E5ECF6","title":{"x":0.05},"scene":{"xaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"zaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"yaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"}},"colorway":["#636efa","#EF553B","#00cc96","#ab63fa","#FFA15A","#19d3f3","#FF6692","#B6E880","#FF97FF","#FECB52"]},"data":{"barpolar":[{"type":"barpolar","marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"carpet":[{"aaxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"},"type":"carpet","baxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"}}],"scatterpolar":[{"type":"scatterpolar","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"parcoords":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"parcoords"}],"scatter":[{"type":"scatter","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2dcontour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2dcontour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattercarpet":[{"type":"scattercarpet","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"mesh3d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"mesh3d"}],"surface":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"surface","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattermapbox":[{"type":"scattermapbox","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergeo":[{"type":"scattergeo","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram":[{"type":"histogram","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"pie":[{"type":"pie","automargin":true}],"choropleth":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"choropleth"}],"heatmapgl":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmapgl","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"bar":[{"type":"bar","error_y":{"color":"#2a3f5f"},"error_x":{"color":"#2a3f5f"},"marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"heatmap":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmap","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contourcarpet":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contourcarpet"}],"table":[{"type":"table","header":{"line":{"color":"white"},"fill":{"color":"#C8D4E3"}},"cells":{"line":{"color":"white"},"fill":{"color":"#EBF0F8"}}}],"scatter3d":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"scatter3d","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergl":[{"type":"scattergl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2d","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scatterternary":[{"type":"scatterternary","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scatterpolargl":[{"type":"scatterpolargl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}]}},"legend":{"yanchor":"bottom","xanchor":"right","y":1.02,"orientation":"h","x":1},"margin":{"l":100,"b":80,"pad":10,"r":80,"t":100},"plot_bgcolor":"#f1f3f7","yaxis":{"title":{"text":"State values"}},"yaxis2":{"titlefont":{"color":"orange"},"overlaying":"y","title":"Action values","side":"right"},"height":550,"width":800},{"showLink":false,"editable":false,"responsive":true,"staticPlot":false,"scrollZoom":true});
});





</script>
```




_______________________________________________________________________________
### High-Level Rendering


```julia
RenderHookResults(hook = hook, 
                    states_to_plot  = [], 
                    actions_to_plot = [],  
                    v_mag_poc       = [1],   # Scaled L₂ norm in αβγ coordinates [V]
                    power_p_inv     = [2],   # Real power [Watts]
                    power_q_inv     = [2],   # Imaginary power [VAi]
                    power_p_poc     = [1],   # Real power [Watts]
                    power_q_poc     = [1],   # Imaginary power [VAi]
                    angles          = [1 2], # Relative angle [degrees]
                    freq            = [1 2], # Angular velocity [Hz]
                    )
```


```@raw html
<div
id = Y1mwz15Dz4hv > </div>
<script>
gd = 'Y1mwz15Dz4hv'
require(['plotly'], function(plotly) {
























plotly.newPlot(gd,[{"mode":"lines","y":[0.0,0.0,0.0,-4.724085663786033e-8,-3.729825730937571e-6,-5.9932561260197614e-5,-0.00044137691359401077,-0.002017181471775627,-0.006684659826880352,-0.017620485613690107,-0.039218480909004004,-0.07674397607621158,-0.13581732885131081,-0.22184658045903125,-0.3395156202103326,-0.49239805894010114,-0.6827221338709214,-0.911269979900952,-1.177379699544645,-1.4790282244512531,-1.812984782520565,-2.175032137042347,-2.560242615454798,-2.9632879968003705,-3.3787529124442317,-3.801423558023502,-4.226529053370929,-4.649922005502895,-5.068194202457949,-5.4787294663629265,-5.8797016893505685,-6.2700278459655205,-6.649288015409135,-7.017623570340522,-7.37562394758511,-7.724210328489368,-8.0645225422736,-8.397813895826088,-8.725357092365545,-9.048363718784223,-9.36791877840937,-9.68493118532387,-10.00010019825234,-10.313897075971614,-10.62656054868507,-10.938104238468888,-11.248333934402512,-11.556872505928093,-11.863190306395303,-12.166638995190562,-12.466486897712581,-12.761954217431384,-13.052246667062068,-13.336586343085875,-13.614238931242738,-13.884536587687137,-14.146896083795667,-14.40083203537746,-14.645965251455692,-14.882026428737372,-15.10885557747752,-15.32639768706887,-15.53469522454719,-15.73387811072087,-15.924151837515435,-16.105784384698232,-16.27909256061504,-16.444428294691672,-16.602157448889084,-16.752637767648434,-16.896237100628753,-17.033325750299483,-17.16425948530971,-17.289368977405832,-17.408956246781756,-17.5232943281014,-17.632627049398266,-17.737168224814106,-17.837101206337515,-17.93257975485053,-18.023730428090726,-18.110656078230953,-18.193439890058443,-18.27214951724044,-18.346841062752034,-18.417562783372194,-18.48435846119937,-18.547270403797263,-18.60634203616157,-18.661620050422627,-18.71315609062001,-18.7610079683849,-18.805240425075638,-18.845925471359124,-18.8831423454853,-18.916977138293777,-18.947522138556728,-18.974874956802854,-18.999137488136665,-19.020414773734835,-19.03881381657483,-19.05444240060373,-19.067407955153353,-19.07781649885151,-19.085771689834623,-19.091374001885455,-19.094720039197544,-19.095901996001466,-19.09500726148395,-19.092118165384953,-19.087311855449975,-19.080660294547172,-19.072230362751068,-19.06208404797823,-19.050278707894954,-19.036867385682655,-19.021899162794252,-19.00541953291536,-18.987470782826914,-18.96809236763024,-18.947321269765855,-18.92519233328674,-18.90173856693113,-18.876991411522916,-18.850980969153355,-18.823736193259606,-18.795285040255017,-18.765654584579394,-18.734871100085485,-18.702960111414114,-18.66994641956395,-18.635854106144308,-18.600706520919154,-18.564526257171536,-18.527335119178282,-18.48915408572828,-18.450003273194795,-18.409901901146185,-18.36886826294368,-18.326919703234672,-18.284072603699556,-18.24034237789578,-18.195743475573387,-18.150289396416717,-18.103992712803013,-18.056865100881076,-18.008917379038586,-17.960159552672295,-17.910600864057244,-17.8602498460857,-17.80911437862296,-17.75720174630571,-17.70451869665727,-17.651071497530715,-17.596865993003163,-17.54190765698962,-17.48620164398259,-17.429752836504576,-17.372565888937814,-17.314645267598767,-17.255995287003056,-17.19662014234901,-17.1365239384031,-17.075710714973592,-17.0141844692683,-16.951949175396877,-16.8890088013859,-16.82536732401244,-16.76102874180353,-16.69599708650859,-16.630276433321413,-16.563870910134,-16.496784706043172,-16.429022079293176,-16.360587364819082,-16.291484981520767,-16.221719439328055,-16.151295346127114,-16.080217414569873,-16.008490468782213,-15.936119450923314,-15.863109427581577,-15.789465595965751,-15.71519328982805,-15.640297985071086,-15.564785304963252,-15.488661024942616,-15.411931076919743,-15.334601553071906,-15.25667870907229,-15.178168966732557,-15.099078916038096,-15.019415316575234,-14.939185098346087,-14.85839536198803,-14.77705337838536,-14.69516658773476,-14.612742598058126,-14.529789183220757,-14.446314280479825,-14.362325987588171,-14.277832559503643,-14.192842404741633,-14.107364081431552,-14.021406293040577,-13.934977883912776,-13.848087834516104,-13.760745256567262,-13.672959387977947,-13.584739587660085,-13.496095330219763,-13.407036200548532,-13.317571888367361,-13.227712182678907,-13.137466966184633,-13.046846209654248,-12.955859966269415,-12.864518365948143,-12.77283160964005,-12.68080996361698,-12.588463753773397,-12.495803359916863,-12.402839210053642,-12.30958177471166,-12.216041561243998,-12.122229108199377,-12.028154979680323,-11.933829759758368,-11.839264046914622,-11.744468448527869,-11.649453575406397,-11.554230036406636,-11.458808433081714,-11.363199354407016,-11.267413371580119,-11.171461032911765,-11.075352858794446,-10.97909933675939,-10.88271091663293,-10.786198005793523,-10.68957096454912,-10.592840101588436,-10.49601566957837,-10.39910786086088,-10.302126803270085,-10.205082556071218,-10.107985106028679,-10.010844363589072,-9.913670159190517,-9.816472239696587,-9.719260264972142,-9.622043804574417,-9.524832334563484,-9.427635234450374,-9.33046178427346,-9.233321161812132,-9.136222439891524,-9.039174583855097,-8.94218644913208,-8.845266778933954,-8.748424202083385,-8.651667230958196,-8.555004259550145,-8.458443561652544,-8.3619932891512,-8.265661470443916,-8.169456008962698,-8.073384681839116,-7.977455138620741,-7.881674900162096,-7.786051357561661,-7.690591771269482,-7.595303270241715,-7.500192851229798,-7.405267378144487,-7.310533581534742,-7.21599805816669,-7.121667270687283,-7.027547547366339,-6.933645081953614,-6.839965933594357,-6.746516026844404,-6.653301151780283,-6.560326964157525,-6.467598985677313,-6.375122604302277,-6.282903074659089,-6.1909455185093805,-6.099254925289756,-6.007836152715527,-5.916693927445834,-5.8258328457966115,-5.735257374556312,-5.644971851814091,-5.5549804878440625,-5.465287366065681,-5.3758964440456225,-5.286811554536996,-5.19803640655897,-5.109574586551901,-5.021429559543044,-4.933604670340941,-4.846103144812414,-4.758928091161403,-4.672082501239396,-4.585569251888247,-4.499391106339722,-4.413550715602703,-4.328050619902193,-4.2428932501044585,-4.158080929248669,-4.07361587398961,-3.9895001961320844,-3.9057359041692337,-3.8223249048179753,-3.739269004571156,-3.6565699112970265,-3.5742292357929557,-3.4922484933861426,-3.410629105528045,-3.329372401417544,-3.248479619579598,-3.1679519095060016,-3.0877903332620065,-3.0079958671074394,-2.9285694030986633,-2.849511750745876,-2.7708236385767577,-2.6925057157911567,-2.614558553833973,-2.536982648030346,-2.4597784191551,-2.3829462150397376,-2.306486312160814,-2.230398917194293,-2.1546841685834863,-2.0793421380929544,-2.004372832374031,-1.9297761944640968,-1.8555521053282007,-1.7817003853523943,-1.7082207958511049,-1.6351130405417678,-1.5623767669970903,-1.4900115681133173,-1.4180169835310608,-1.346392501057295,-1.2751375580843671,-1.2042515429667873,-1.1337337963724963,-1.0635836126537939,-0.9938002411766789,-0.9243828876095144,-0.8553307152808127,-0.7866428464236739,-0.7183183634144186,-0.6503563100676786,-0.582755692823568,-0.5155154819526331,-0.448634612759065,-0.3821119867106688,-0.31594647261925957,-0.25013690776731323,-0.18468209899836552,-0.11958082380905921,-0.05483183140952974,0.009566156218141941,0.07361444325951538,0.13731435912453094,0.20066725742483982,0.26367451499785943,0.3263375309488197,0.3886577257620729,0.45063654032691147,0.5122754350854279,0.5735758891785281,0.6345393995358677,0.6951674800678234,0.7554616608508695,0.8154234873103974,0.875054519478283,0.934356331184631,0.9933305093374543,1.0519786531830277,1.110302373627519,1.168303292511382,1.2259830419365585,1.2833432636214184,1.3403856082584866,1.397111734880582,1.4535233102659788,1.509622008326545,1.5654095095520972,1.6208875004402614,1.6760576729610648,1.7309217240188275,1.785481354930198,1.8397382709464694,1.893694180732183,1.9473507959032617,2.0007098306498676,-419.943946351347,37361.51077794287,60947.65510963726,70695.49821651926,75652.27979054385,78787.11520496305,77779.4511896679,71241.24884790031,62068.69800094557,54197.01178934438,48902.99768992537,45466.72418261698,42990.267185871206,40986.98653165512,39288.919272733074,37899.59285458973,36884.8083393156,36300.48645021222,36162.377756183865,36444.80141774195,37090.03984288027,38019.00470263789,39141.26418951541,40364.79436982571,41605.06969206764,42792.52255000017,43877.21131910822,44829.85929931635,45639.43520183018,46308.50648976555,46848.06642717678,47273.189761408285,47600.155446749326,47844.973833212935,48022.86472198517,48148.16926290042,48234.32219572152,48293.71118931055,48337.40418299625,48374.81372942566,48413.39777563416,48458.48585806805,48513.28320418402,48579.05539108934,48655.45297598104,48740.911567435614,48833.06211481651,48929.10346374494,49026.11088341187,49121.27379583219,49212.06641596211,49296.36001184686,49372.486106668744,49439.259918044656,49495.97327343687,49542.36552327053,49578.57994039782,49605.11065661338,49622.74328710165,49632.49081781483,49635.52640696215,49633.11557640391,49626.55130054605,49617.09587498833,49605.93265526049,49594.12936545491,49582.612971527735,49572.15502376087,49563.36583434773,49556.69593993628,49552.443529213066,49550.766666662836,49551.69914061896,49555.16865045425,49561.01605615739,49569.014541822675,49578.887844768884,49590.32700578661,49603.005350077954,49616.591547663695,49630.760662084875,49645.20314014986,49659.63175351969,49673.786599571635,49687.43835990424,49700.390083447644,49712.47777685951,49723.570062052684,49733.56711636068,49742.3990678609,49750.02398958626,49756.42561717018,49761.610900394175,49765.60747946438,49768.4611535604,49770.23338420258,49770.99885640053,49770.843109080204,49769.860242426614,49768.15071070033,49765.81921060866,49762.972676155645,49759.718390354,49756.162223804175,49752.407010426316,49748.55107178965,49744.686902658475,49740.90003056673,49737.268060947245,49733.859916511574,49730.73527582409,49727.9442118482,49725.52702716912,49723.51427872728,49721.9269812713,49720.77697534625,49720.06744256444,49719.793548339636,49719.94319036269,49720.49782997015,49721.43338319925,49722.721148664146,49724.32875030748,49726.22107447867,49728.36118256842,49730.71118251377,49733.23304480059,49735.88935104602,49738.643965753756,49741.46262431952,49744.31343275004,49747.167276801156,49749.99814029716,49752.783334248095,49755.503640018804,49758.143371208615,49760.690360085435,49763.135875369924,49765.474478900855,49767.70382923748,49769.824440584045,49771.83940556225,49773.75409033496,49775.57581040293,49777.313495081544,49778.977348222266,49780.57851220356,49782.12874157865,49783.64009206559,49785.12462980259,49786.59416499744,49788.06001328301,49789.53278727146,49791.02221999833,49792.537021168086,49794.08476638152,49795.67181884311,49797.30328243134,49798.98298447051,49800.71348607619,49802.49611756392,49804.33103611009,49806.21730263585,49808.15297475236,49810.13521254255,49812.16039397129,49814.22423678996,49816.32192394084,49818.44822965001,49820.59764362365,49822.7644910252,49824.943046190725,49827.1276383435,49829.31274787556,49831.49309207481,49833.66369948239,49835.81997235939,49837.95773702086,49840.07328205683,49842.163384695436,49844.22532577573,49846.256893981794,49848.25638014512,49850.222562550945,49852.15468427935,49854.05242368425,49855.91585915434,49857.745429316565,49859.54188983709,49861.306267942775,49863.03981574042,49864.74396334311,49866.420272734416,49868.07039320944,49869.696019130206,49871.298850627616,49872.88055777052,49874.44274861146,49875.986941408504,49877.51454121373,49879.02682091812,49880.524906747014,49882.00976811074,49883.48221163934,49884.942879159156,49886.392249312645,49887.83064247496,49889.25822858461,49890.67503747899,49892.080971311756,49893.475818621664,49894.859269626664,49896.23093232913,49897.59034903583,49898.93701292234,49900.270384301955,49901.58990629348,49902.89501962043,49904.185176314,49905.45985213333,49906.718557558146,49907.96084724902,49909.18632791065,49910.39466453023,49911.58558499849,49912.758883151866,49913.91442030302,49915.05212535117,49916.17199358494,49917.274084306315,49918.35851741753,49919.42546912165,49920.47516689208,49921.50788386926,49922.52393283906,49923.52365994505,49924.507438277746,49925.475661475655,49926.428737461436,49927.367082422825,49928.29111513501,49929.20125170645,49930.09790081378,49930.98145947793,49931.85230941754,49932.71081400183,49933.55731581143,49934.39213480241,49935.21556705866,49936.02788410567,49936.829332751295,49937.62013541066,49938.40049086792,49939.17057542058,49939.93054435113,49940.680533668376,49941.42066205994,49942.15103299881,49942.87173694778,49943.582853609754,49944.284454174034,49944.9766035137,49945.65936229423,49946.332788958054,49946.99694155529,49947.65187939734,49948.29766451442,49948.93436290395,49949.56204556316,49950.18078930187,49950.79067733851,49951.39179968509,49951.98425333087,49952.56814223804,49953.143577165225,49953.71067533638,49954.26955997465,49954.82035972213,49955.36320796605,49955.8982420933,49956.4256026936,49956.94543273202,49957.457876709734,49957.96307983101,49958.46118719234,49958.95234300867,49959.436689888666,49959.91436816984,49960.385515322036,49960.85026542563,49961.308748728516,49961.7610912849,49962.20741467623,49962.647835813455,49963.0824668173,49963.511414974106,49963.93478276088,49964.352667934414,49964.76516367681,49965.172358791526,49965.57433794047,49965.971181915986,49966.362967938774,49966.74976997423,49967.131659059865,49967.50870363672,49967.880969878366,49968.24852201103,49968.611422620525,49968.96973294018,49969.323513117066,49969.67282245304,49970.017719617914,49970.35826283405,49970.694510030706,49971.02651896841,49971.35434733361,49971.67805280484,49971.99769309108,49972.31332594579,49972.62500915655,49972.932800515126,49973.2367577691,49973.53693855823,49973.83340033908,49974.12620029989,49974.41539526908,49974.7010416199,49974.98319517373,49975.26191110473,49975.537243847684,49975.80924701057,49976.07797329445,49976.34347442133,49976.60580107097,49976.86500282811,49977.12112813988,49977.37422428417,49977.62433734884,49977.87151222153,49978.11579258962,49978.35722095006,49978.59583862809,49978.83168580391,49979.06480154678,49979.29522385498,49979.522989701225,49979.74813508194,49979.970695069875,49980.19070386795,49980.408194864576,49980.62320068924,49980.83575326658,49981.04588386926,49981.25362316864,49981.45900128232,49981.66204781832,49981.862791915584,49982.06126228043,49982.25748721909,49982.45149466513,49982.643312203494,49982.83296708991,49983.020486266156,49983.20589637176,49983.38922375193,49983.570494462336,49983.74973427103,49983.92696865776,49984.10222281138,49984.275521625066,49984.44688969099,49984.61635129328,49984.78393040065,49984.949650659175,49985.113535384626,49985.275607555435,49985.43588980623,49985.59440442227,49985.75117333503,49985.90621811838,49986.05955998685,49986.21121979442,49986.361218035265,49986.50957484552,49986.65631000644,49986.80144294877,49986.944992758516,49987.08697818352,49987.22741764111,49987.366329226614,49987.50373072268,49987.63963960877,49987.77407307144,49987.90704801476,49988.03858107099,49988.16868861118,49988.29738675547,49988.424691383094,49988.55061814301],"type":"scatter","name":"source2_p_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.0,3.3715772654077714e-5,0.0005731203108950692,0.003410296817261127,0.012562302074771244,0.035262907168066154,0.08241232178334935,0.16927150928959084,0.31477387566138226,0.5415110043995401,0.8739694309995288,1.3374263788255787,1.9560103055662617,2.7507098918471837,3.7384138637537174,4.929825314801659,6.329935726110396,7.936711830566662,9.742514698088328,11.73402696862135,13.893780936621724,16.201252265167685,18.633947701489365,21.169128030254388,23.784344165402782,26.459076706353727,29.174880428131015,31.916412063237132,34.67138282369622,37.430812222370704,40.188890139609406,42.942615859842846,45.69158327362936,48.43731966222378,51.182993565788564,53.93274548747409,56.69134643850537,59.46368906414045,62.25445469687289,65.06781089837418,67.90715414266607,70.77498890984137,73.67279245904335,76.60102577770058,79.55912738393653,82.54560755432713,85.55814358564649,88.59371947846641,91.64877879756119,94.71938107707204,97.80136861866498,100.89051576690697,103.9826762191068,107.07390576300936,110.16057054384356,113.23943037070563,116.30770075929328,119.363091889046,122.40382610343039,125.42863545066297,128.43674209001665,131.42782341339935,134.40196610359752,137.3596109478959,140.301492469832,143.22857576143008,146.14199304066844,149.04298301174123,151.93283216707516,154.81281988026882,157.68417412837425,160.54803903176776,163.40544933749376,166.25731221806194,169.10439577463717,171.9473240729131,174.78657789335432,177.62250012753572,180.45530559347225,183.28509400217303,186.11186579416812,188.93553952706313,191.75597021899299,194.57296765555378,197.38631395654107,200.19577989921234,203.00113940190624,205.80218204436358,208.5987232589131,211.39061227175017,214.1777376982693,216.96003095764564,219.73746764619247,222.51006707143088,225.2778902348025,228.0410364875764,230.7996391986802,233.55386067238376,236.30388662861645,239.04992047652848,241.79217762269883,244.5308800099733,247.26625104323486,249.99851104061932,252.72787328827695,255.45454076952515,258.17870358339445,260.90053705873277,263.62020053319765,266.33783675280915,269.0535718313076,271.7675156945788,274.4797629326225,277.1903939733856,279.89947649693124,282.6070670080028,285.3132124933018,288.0179520951025,290.72131874294695,293.4233406943435,296.12404294488636,298.82344847957387,301.52157934504896,304.2184575328785,306.9141056713979,309.60854753097595,312.3018083534296,314.99391502062815,317.68489608173576,320.374781659907,323.063603260641,325.7513935051394,328.43818581022623,331.1240140361579,333.808912121102,336.4929137191407,339.17605185638854,341.85835861598946,344.53986486204826,347.2206000075415,349.9005918308262,352.5798663416374,355.2584476966058,357.9363581613635,360.6136181162605,363.290246100472,365.9662588898484,368.64167160126993,371.31649781921755,373.99074973722077,376.66443830882525,379.33757340306335,382.0101639596219,384.6822181391057,387.35374346566294,390.02474695881546,392.6952352530036,395.3652147026668,398.03469147335926,400.7036716177436,403.3721611373404,406.0401660310263,408.707692330726,411.3747461261689,414.0413335799695,416.7074609348638,419.37313451408545,422.0383607173878,424.7031460129966,427.3674969279741,430.03142003752725,432.6949219538989,435.3580093172124,438.020688786733,440.6829670348711,443.3448507431792,446.00634660091845,448.66746130636034,451.3282015694621,453.9885741181871,456.64858570528247,459.3082431175897,461.967553186175,464.6265227970511,467.28515890309274,469.9434685352083,472.601458814013,475.25913696054334,477.91651030648984,480.57358630340013,483.23037253058226,485.88687670226204,488.54310667284994,491.1990704416041,493.8547761549574,496.51023210904555,499.1654467493098,501.8204286706523,504.4751866151789,507.12972947001964,509.7840662637539,512.438206162867,515.0921584665352,517.7459326022607,520.3995381202849,523.0529846879969,525.7062820844989,528.3594401945219,531.012469002831,533.6653785882884,536.3181791182411,538.9708808428113,541.6234940888215,544.2760292556366,546.928496808073,549.5809072723462,552.2332712306387,554.8855993152929,557.5379022048511,560.1901906178041,562.8424753089205,565.4947670630752,568.1470766909781,570.7994150243048,573.451792910055,576.1042212064982,578.7567107774797,581.4092724879679,584.061917198527,586.7146557610481,589.3674990130453,592.0204577733996,594.6735428366376,597.3267649691965,599.9801349033442,602.6336633330902,605.2873609096243,607.9412382359853,610.5953058632746,613.249574285658,615.9040539362262,618.5587551827555,621.2136883237731,623.868863584029,626.5242911110797,629.1799809713391,631.8359431465108,634.4921875299324,637.1487239237835,639.8055620346498,642.462711472182,645.1201817441937,647.7779822553188,650.4361223036858,653.0946110782053,655.753457656474,658.4126710022372,661.0722599630344,663.7322332684628,666.3925995278667,669.0533672287075,671.714544734715,674.3761402841262,677.0381619883653,679.7006178306243,682.363515664426,685.026863212587,687.6906680657062,690.3549376818602,693.0196793850683,695.6849003649899,698.3506076759703,701.0168082362206,703.6835088282039,706.3507160970297,709.0184365511024,711.6866765612784,714.3554423614522,717.0247400475776,719.6945755785424,722.3649547757077,725.0358833233378,727.7073667691966,730.3794105243832,733.0520198637025,735.7251999268392,738.3989557183934,741.0732921086317,743.7482138339896,746.4237254981155,749.0998315727288,751.7765363983344,754.4538441849099,757.1317590134653,759.8102848363567,762.4894254792307,765.1691846412451,767.8495658966893,770.5305726964416,773.2122083686913,775.8944761204082,778.5773790386272,781.2609200923214,783.945102132564,786.6299278955386,789.3154000023109,792.0015209616824,794.688293170904,797.3757189171622,800.0638003798722,802.7525396307205,805.4419386374818,808.1319992631727,810.8227232690105,813.5141123162568,816.206167966784,818.898891685455,821.5922848416236,824.2863487102401,826.9810844749364,829.6764932277054,832.3725759719872,835.0693336236077,837.7667670130829,840.4648768864987,843.1636639077768,845.8631286599507,848.5632716466894,851.2640932950403,853.9655939551365,856.6677739036056,859.3706333437847,862.0741724088703,864.7783911620909,867.4832895987026,870.1888676482554,872.8951251750739,875.602061980763,878.3096778050926,881.0179723276622,883.7269451701959,886.4365958961298,889.1469240147364,891.8579289800463,894.5696101940985,897.2819670073742,899.9949987207498,902.7087045863107,905.4230838099517,908.1381355507035,910.853858924514,913.5702530032562,916.2873168184071,919.0050493594525,921.723449578006,924.4425163879607,927.1622486648072,929.8826452510503,932.6037049533343,935.3254265454977,938.0478087702884,940.7708503385117,943.4945499322674,946.2189062046087,948.9439177814149,951.6695832620765,954.3959012206278,957.1228702062222,959.8504887456951,962.5787553421046,965.3076684781851,968.0372266161203,970.7674281976372,973.4982716465488,976.2297553685949,978.9618777526691,981.6946371712145,984.428031981879,987.1620605275482,989.8967211373406,992.6320121272522,995.3679318016727,998.1044784530598,1000.8416503634887,1003.5794458039329,1006.3178630376699,1009.0569003179048,1011.7965558901512,1014.5368279926859,1017.2777148565135,1020.0192147067846,1022.7613257629971,1025.5040462389288,1028.2473743443995,1030.9913082849164,1033.7358462624525,1036.4809864761332,1039.2267271221108,1041.973066395075,1044.7200024873473,1047.467533590772,1052.302453555661,-2252.4089630801054,-10372.302701105198,-23008.97320756173,-38770.334724261214,-54547.68999201552,-67201.19656815895,-74199.23800490901,-75278.93231649828,-71568.64272595267,-65120.267968584405,-57759.89462053847,-50735.06924003181,-44711.852018564095,-39839.5858090258,-36091.82781195566,-33278.8254085018,-31256.04481868565,-29866.899776582264,-29013.624485027558,-28620.156462664803,-28629.676879990355,-29003.185554503165,-29687.127904393466,-30632.57086955665,-31766.986112445302,-33020.09284623854,-34313.86698008775,-35581.25855408826,-36767.52347057316,-37833.87717882563,-38760.22767062115,-39538.04719897927,-40171.988818144055,-40671.43757432517,-41051.61961030102,-41328.6751469435,-41519.89858554746,-41642.09466451126,-41710.80685436129,-41740.465119550005,-41743.30561081489,-41730.11497615976,-41709.41992462003,-41688.08751936935,-41670.96974272086,-41661.22492394362,-41660.43675085334,-41668.839879873485,-41685.7373827233,-41709.70301097143,-41739.001044308636,-41771.7271908328,-41806.06603794717,-41840.372204802734,-41873.27655399888,-41903.730572210836,-41931.01793620192,-41954.76135740528,-41974.88099522975,-41991.56526870785,-42005.2098932871,-42016.36934944133,-42025.69794904659,-42033.89781871585,-42041.66957118067,-42049.66830402012,-42058.46894904744,-42068.53968681159,-42080.22608344964,-42093.744337763295,-42109.18252608166,-42126.50877860745,-42145.58420955917,-42166.17961361804,-42187.99467693838,-42210.67785213105,-42233.84674419334,-42257.10663628504,-42280.06777316215,-42302.35965592998,-42323.643110335484,-42343.61947022482,-42362.03715167467,-42378.695839774125,-42393.44808932452,-42406.199116782285,-42416.904534669695,-42425.566943617094,-42432.231312509815,-42436.97979463708,-42439.92609200286,-42441.20964254069,-42440.98988360337,-42439.440631985446,-42436.744878097415,-42433.08993618273,-42428.66316137253,-42423.64815335675,-42418.221522916625,-42412.550173261465,-42406.789078873124,-42401.0795572975,-42395.547971894426,-42390.30487266355,-42385.44450016319,-42381.04464574653,-42377.16680588306,-42373.85660889421,-42371.14447503931,-42369.046478144024,-42367.56538394151,-42366.69182675429,-42366.40560257717,-42366.677040320545,-42367.46842844728,-42368.735464728874,-42370.4287053922,-42372.4949880133,-42374.87880442658,-42377.52360351683,-42380.37300317164,-42383.37189692494,-42386.46744079581,-42389.60991182016,-42392.75343110958,-42395.856548554635,-42398.88268880837,-42401.80046093699,-42404.583837150116,-42407.2122077704,-42409.67032221856,-42411.94812679819,-42414.04051167979,-42415.94697996941,-42417.671252325075,-42419.2208206668,-42420.606464306664,-42421.84174149306,-42422.94246858972,-42423.9261983655,-42424.81170777572,-42425.6185045686,-42426.36636082041,-42427.074880298984,-42427.76310530748,-42428.44916739968,-42429.14998517882,-42429.88101117646,-42430.65602873321,-42431.4869987496,-42432.383955216865,-42433.35494762254,-42434.4060275252,-42435.54127599839,-42436.762868079415,-42438.07116996001,-42439.464864349204,-42440.94109923983,-42442.4956552409,-42444.12312663944,-42445.817111484204,-42447.570406164625,-42449.37520024867,-42451.223267664216,-42453.106150711545,-42455.015333825635,-42456.942404451234,-42458.87919889556,-42460.81793148265,-42462.75130581772,-42464.6726074401,-42466.57577756134,-42468.4554680103,-42470.30707786357,-42472.1267725728,-42473.911486685494,-42475.658911483675,-42477.367469063094,-42479.036274507984,-42480.66508792282,-42482.2542581184,-42483.804659781286,-42485.31762590371,-42486.79487722396,-42488.23845029897,-42489.65062575885,-42491.03385812479,-42492.39070844279,-42493.723780814005,-42495.03566374196,-42496.32887703413,-42497.6058248343,-42498.86875518507,-42500.119726362354,-42501.36058006903,-42502.59292143017,-42503.818105608625,-42505.03723074163,-42506.251136794504,-42507.46040985979,-42508.66539134179,-42509.86619143387,-42511.06270625329,-42512.25463797882,-42513.44151733139,-42514.62272775192,-42515.797530645206,-42516.965091089296,-42518.124503458974,-42519.274816449906,-42520.41505705385,-42521.54425308153,-42522.661453914916,-42523.765749200174,-42524.856285290996,-42525.932279290224,-42526.99303060553,-42528.037929998114,-42529.06646613932,-42530.07822975876,-42531.07291548925,-42532.050321565606,-42533.010347555944,-42533.952990328646,-42534.878338484654,-42535.78656548101,-42536.67792169601,-42537.5527256726,-42538.4113547784,-42539.2542355204,-42540.08183371853,-42540.894644754866,-42541.69318407281,-42542.47797809535,-42543.24955570086,-42544.008440380094,-42544.755143165596,-42545.490156410226,-42546.21394845845,-42546.92695925058,-42547.62959684952,-42548.32223489728,-42549.00521095796,-42549.678825710995,-42550.34334293858,-42550.998990228945,-42551.64596033745,-42552.28441310204,-42552.91447784403,-42553.53625615267,-42554.14982497216,-42554.75523989795,-42555.3525386018,-42555.94174430683,-42556.522869228735,-42557.09591793445,-42557.6608905364,-42558.217785690664,-42558.766603342694,-42559.307347191294,-42559.84002684527,-42560.36465965411,-42560.88127220208,-42561.38990146417,-42561.89059563351,-42562.3834146202,-42562.868430250186,-42563.34572618031,-42563.81539755091,-42564.27755041263,-42564.7323009499,-42565.17977453958,-42565.62010467121,-42566.05343176923,-42566.479901938954,-42566.89966567784,-42567.31287657439,-42567.71969002127,-42568.120261968914,-42568.51474774661,-42568.903300953774,-42569.28607245516,-42569.66320948007,-42570.034854838406,-42570.4011462601,-42570.7622158616,-42571.11818973293,-42571.46918765971,-42571.815322949995,-42572.15670238717,-42572.49342628097,-42572.82558861455,-42573.1532772779,-42573.476574373424,-42573.79555658373,-42574.11029558784,-42574.42085851569,-42574.727308426714,-42575.029704804896,-42575.328104054504,-42575.62255999084,-42575.913124317616,-42576.199847077456,-42576.48277707544,-42576.76196226789,-42577.03745010885,-42577.309287855715,-42577.57752282995,-42577.84220263034,-42578.10337530119,-42578.36108945604,-42578.61539435608,-42578.86633994887,-42579.113976868255,-42579.35835640192,-42579.59953042398,-42579.83755130654,-42580.07247180538,-42580.304344929114,-42580.53322379666,-42580.75916148457,-42580.98221086858,-42581.20242446803,-42581.41985428764,-42581.63455167118,-42581.846567158704,-42582.0559503573,-42582.26274982654,-42582.467012974106,-42582.66878597181,-42582.86811368475,-42583.065039619905,-42583.25960588835,-42583.45185318831,-42583.641820796984,-42583.829546586654,-42584.01506704418,-42584.19841731136,-42584.37963122965,-42584.558741397996,-42584.73577923809,-42584.91077506236,-42585.08375814982,-42585.254756823204,-42585.423798527925,-42585.590909909464,-42585.75611688874,-42585.91944474178,-42586.08091816366,-42586.240561338826,-42586.39839799886,-42586.55445148186,-42586.7087447766,-42586.861300566874,-42587.01214126816,-42587.16128905664,-42587.30876589085,-42587.45459353211,-42587.59879355172,-42587.74138734082,-42587.88239610927,-42588.021840884685,-42588.15974250491,-42588.29612160861,-42588.4309986246,-42588.56439375659,-42588.696326970276,-42588.82681797625,-42588.95588621412,-42589.08355084044,-42589.20983070949,-42589.33474436336,-42589.45831001847,-42589.580545555735,-42589.70146851299,-42589.82109607601,-42589.93944507821,-42590.05653199458,-42590.17237294594,-42590.28698369621,-42590.400379661456,-42590.51257591196,-42590.62358718242,-42590.7334278804,-42590.84211209514,-42590.949653612,-42591.056065922705,-42591.161362239036,-42591.265555506485,-42591.36865841698,-42591.47068342461,-42591.571642756775,-42591.67154842944,-42591.77041225652,-42591.86824586736,-42591.965060712326,-42592.060868078435,-42592.15567909442],"type":"scatter","name":"source1_p_poc","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.0,0.0,3.689678489733429e-10,6.418103847652332e-8,1.442755922520567e-6,1.356277657886112e-5,7.623917068417461e-5,0.00030399880348801146,0.0009498386365694941,0.002474504913540663,0.005605761774685247,0.01137764496543395,0.021143220230505355,0.03655663382649198,0.059523467344366754,0.09212549371199322,0.13653319477914239,0.19492126554015093,0.2693974446141064,0.361947755114875,0.47439673652987613,0.6083806358172777,0.7653323713933835,0.9464773387586523,1.1528381280782516,1.3852454837273267,1.6443530361667755,1.930654205911708,2.244500752613008,2.586122425615488,2.955647073625996,3.3531197922507623,3.778519770769268,4.231773705503735,4.712765466101894,5.221342505192584,5.7573199413123195,6.320483512998974,6.91059217772988,7.527380895786335,8.170563607974083,8.839836347855785,9.534880334478217,10.255364984352983,11.000950836620405,11.771292317289646,12.566040274942974,13.384844111295607,14.227353424898562,15.093219107276216,15.982093959925702,16.893632929245793,17.82749306266721,18.783333260873373,19.76081384803606,20.759595991687796,21.779340997910456,22.819709556231487,23.8803610115399,24.960952748187253,26.06113974368062,27.18057432100788,28.31890610772488,29.47578219284556,30.650847474750226,31.84374518583237,33.05411861182581,34.28161707120491,35.52589640943688,36.786602791519364,38.06337556333505,39.35585519899294,40.66368756395049,41.98652364530121,43.32401775339588,44.675827338581826,46.041614978282816,47.421051028528346,48.81381539795962,50.21959802419687,51.63809854741025,53.069025821364875,54.512097555780755,55.967040043363276,57.43358783144062,58.911483284567765,60.4004760849553,61.90032274197259,63.41078614545225,64.93163515566286,66.4626442089526,68.00359292969456,69.55426575646665,71.1144515991235,72.68394354140531,74.26253859651229,75.8500375162002,77.44624464969635,79.0509678467217,80.66401839797228,82.28521100595567,83.91436377896846,85.55129824128664,87.19583935326351,88.84781553580817,90.5070586945169,92.17340423951508,93.84669109781522,95.52676171574392,97.21346204970362,98.9066415441979,100.60615309666734,102.31185300922087,104.02360092785106,105.74125977017236,107.46469564309409,109.19377775211666,110.92837830409972,112.66837240540501,114.4136379572821,116.16405555025315,117.919508359109,119.67988203995343,121.44506463052458,123.2149464548134,124.98942003276753,126.76837999565069,128.55172300739778,130.339347692115,132.13115456769123,133.92704598533885,135.7269260747516,137.53070069447702,139.33827738699927,141.14956533801774,142.9644753393493,144.78291975489762,146.60481248914851,148.43006895766806,150.25860605914548,152.09034214855495,153.92519701109063,155.76309183657622,157.60394919410408,159.4476930067563,161.29424852627028,163.14354230759423,164.99550218332413,166.85005723803457,168.70713778255634,170.56667532828354,172.42860256160475,174.29285331856082,176.15936255984835,178.02806634626003,179.89890181468104,181.77180715472434,183.64672158608982,185.52358533671574,187.40233962176873,189.28292662350003,191.1652894720036,193.0493722268619,194.93511985968294,196.82247823749066,198.7113941069454,200.60181507935033,202.49368961639024,204.38696701655292,206.281597402173,208.1775317070424,210.07472166453158,211.9731197961628,213.87267940058626,215.7733545429058,217.67510004431298,219.5778714719928,221.48162512927044,223.38631804595116,225.2919079688549,227.19835335251219,229.10561335001017,231.0136478039828,232.92241723771713,234.8318828464116,236.7420064885394,238.6527506773461,240.56407857248132,242.4759539717417,244.38834130294168,246.3012056159446,248.21451257479225,250.12822844999695,252.0423201109388,253.95675501841504,255.87150121732085,257.78652732944977,259.70180254643384,261.6172966227948,263.5329798691167,265.44882314532424,267.3647978540801,269.28087593428194,271.197029854639,273.1132326073566,275.0294577018858,276.9456791587669,278.86187150352936,280.7780097606674,282.6940694476734,284.6100265691245,286.52585761083026,288.4415395340199,290.3570497695852,292.27236621235744,294.18746721543187,296.10233158453434,298.0169385724227,299.9312678733263,301.8452996174183,303.7590143653308,305.6723931027011,307.5854172347571,309.4980685809338,311.4103293695305,313.3221822323948,315.23361019965466,317.1445966944712,319.05512552784,320.9651808934228,322.87474736241313,324.78380987843616,326.6923537524926,328.6003646579314,330.5078286254551,332.41473203816633,334.32106162665093,336.2268044640874,338.13194796139754,340.03647986243385,341.94038823919414,343.84366148708136,345.746288320184,347.6482577665973,349.5495591637922,351.45018215399773,353.35011667963073,355.2493529787491,357.14788158055336,359.045693300913,360.94277923793715,362.8391307675722,364.73473953923275,366.6295974714805,368.5236967477275,370.41702981197847,372.30958936461326,374.2013683582043,376.09235999336545,377.98255771464693,379.8719552064603,381.76054638904446,383.6483254144645,385.5352866626607,387.4214247375167,389.30673446299204,391.19121087926806,393.0748492389466,394.9576450032767,396.83959383843404,398.72069161182753,400.60093438845684,402.48031842729847,404.3588401777369,406.23649627603095,408.1132835418168,409.98919897466396,411.864239750661,413.7384032190448,415.61168689885284,417.48408847564366,419.3556057982334,421.22623687548776,423.09597987313646,424.9648331106374,426.83279505808036,428.69986433314375,430.56603969806054,432.4313200566263,434.29570445128576,436.15919206019174,438.0217821943869,439.88347429496423,441.74426793025464,443.6041627931105,445.4631586981758,447.32125557922416,449.17845348651525,451.03475258420286,452.89015314779425,454.7446555615869,456.5982603162202,458.450968006219,460.3027793275615,462.1536950753273,464.0037161413544,465.85284351192956,467.70107826552305,469.5484215705486,471.39487468317964,473.24043894517195,475.08511578175256,476.9289066995071,478.7718132843362,480.6138371994169,482.4549801832152,484.29524404752243,486.13463067553477,487.9731420199553,489.81078010112094,491.647547005185,493.4834448822965,495.3184759448627,497.1526424657915,498.98594677678466,500.8183912666507,502.6499783796845,504.4807106140256,506.31059052007583,508.13962069894365,509.96780380090786,511.7951425239178,513.6216396120951,515.4472978542983,517.2721200827311,519.0961091715021,520.9192680352735,522.7415996279165,524.5631069412159,526.3837930035513,528.2036608786332,530.0227136642578,531.840954491098,533.6583865215149,535.4750129483836,537.290836993916,539.1058619085891,540.9200909700209,542.7335274818803,544.5461747728285,546.3580361955003,548.1691151254613,549.9794149602568,551.7889391184058,553.5976910384579,555.4056741780712,557.2128920130868,559.0193480366307,560.8250457582425,562.6299887030275,564.4341804108224,566.237624435365,568.0403243434931,569.8422837143717,571.6435061387151,573.4439952180494,575.2437545639667,577.0427877974171,578.8410985480077,580.6386904533092,582.4355671581965,584.2317323141973,586.0271895788392,587.8219426150418,589.6159950905006,591.409350677095,593.2020130503053,594.9939858886412,596.7852728730877,598.5758776866018,600.3658040135062,602.155055539052,603.9436359488359,605.7315489284127,607.5187981627242,609.3053873356959,611.0913201297515,612.8766002253872,614.6612313006934,616.4452170309734,618.228561088361,620.011267141378,621.7933388545445,623.5747798880092,625.3555938971671,627.1357845323121,628.915355438301,630.6943102542058,632.4726526129756,634.2503861410827,1156.1577026921414,-10321.431385755546,-18352.391900479714,-20599.88010005202,-22611.39004040167,-25848.78651509744,-27376.24083943774,-26188.991954190893,-24360.22352831858,-24144.63848613406,-25965.46108452337,-28817.70175801723,-31603.175937726548,-33846.38204882619,-35574.71827331993,-36933.021299204156,-37986.5851726562,-38752.7257106014,-39282.074380983366,-39672.29692743045,-40023.5248610552,-40390.54603774687,-40770.19532544844,-41123.369291363924,-41406.72411768172,-41593.30927028094,-41677.194060422014,-41668.2530989526,-41584.807636959864,-41447.88173203857,-41277.441956018825,-41089.906706110996,-40896.7782856295,-40704.53036146534,-40515.643325073135,-40330.224319321904,-40147.53614291224,-39967.0361331841,-39788.85962523326,-39613.91620988911,-39443.76667455501,-39280.4014704553,-39125.9768815281,-38982.56082237174,-38851.94052236806,-38735.52737904615,-38634.36159614061,-38549.17670073892,-38480.47312328191,-38428.557464581274,-38393.536338000355,-38375.27878376038,-38373.374873783345,-38387.113631036584,-38415.489109228394,-38457.231121258,-38510.84980210368,-38574.68524528516,-38646.95727927585,-38725.81491804851,-38809.38564907761,-38895.82359527881,-38983.354037171826,-39070.31140376428,-39155.169248032864,-39236.56233828627,-39313.30238855698,-39384.38892587871,-39449.01614950366,-39506.575718088796,-39556.65501551605,-39599.03066659993,-39633.65761534603,-39660.654621261616,-39680.28720168407,-39692.94895966778,-39699.14195964567,-39699.456614310504,-39694.55146386859,-39685.133256679124,-39671.93778037421,-39655.711867576756,-39637.19691399494,-39617.11412121181,-39596.151576152435,-39574.95321535666,-39554.10969660476,-39534.15118324027,-39515.542020184046,-39498.677240924466,-39483.880800247134,-39471.40539274653,-39461.43369700955,-39454.080879236804,-39449.398189896616,-39447.37748747162,-39447.95652229807,-39451.024812861,-39456.42994949816,-39463.98416740102,-39473.47104126592,-39484.652165977015,-39497.27369984506,-39511.07265896772,-39525.78286379639,-39541.140452693464,-39556.88889214372,-39572.78342873862,-39588.59494327547,-39604.113181694775,-39619.14935104211,-39633.538081271814,-39647.13876566493,-39659.83630376159,-39671.54128071382,-39682.18962550828,-39691.74179739563,-39700.18155512361,-39707.51436731944,-39713.76552476875,-39718.97801643901,-39723.21023092444,-39726.533543543395,-39729.02984669371,-39730.789077412875,-39731.90679162146,-39732.48182944851,-39732.614110521834,-39732.40259228576,-39731.94341839773,-39731.328278173925,-39730.64299203892,-39729.96633211845,-39729.369081621975,-39728.91333158287,-39728.652008924284,-39728.62862572984,-39728.8772360606,-39729.422583683,-39730.280421680036,-39731.45798311177,-39732.9545806561,-39734.762312475286,-39736.86685137536,-39739.2482946162,-39741.88205243499,-39744.73975442453,-39747.79015430758,-39751.00001531165,-39754.3349602291,-39757.76027228046,-39761.24163503784,-39764.74580185379,-39768.241187435895,-39771.69837636235,-39775.090545414525,-39778.39379857243,-39781.58741534497,-39784.65401477169,-39787.57963890771,-39790.35376088539,-39792.969223713706,-39795.422116838476,-39797.711598133435,-39799.83966943468,-39801.810913975,-39803.63220413064,-39805.31238777972,-39806.86196129902,-39808.29273681906,-39809.617510832846,-39810.8497406303,-39812.00323433127,-39813.09185953915,-39814.12927483936,-39815.12868756406,-39816.10264043248,-39817.06282888422,-39818.01995016311,-39818.98358448672,-39819.96210797467,-39820.96263640496,-39821.99099833581,-39823.051735667934,-39824.14812934299,-39825.28224756643,-39826.4550137179,-39827.66629095771,-39828.91498046016,-39830.19913019029,-39831.51605119073,-39832.86243845134,-39834.23449358892,-39835.628046758684,-39837.03867545382,-39838.461818105396,-39839.892880673084,-39841.32733470795,-39842.76080566532,-39844.18915054003,-39845.60852418776,-39847.015433971166,-39848.40678263367,-39849.77989954412,-39851.13256067372,-39852.462997861,-39853.76989808582,-39855.052393610255,-39856.3100439552,-39857.542810760715,-39858.751026632286,-39859.935359101255,-39861.0967708313,-39862.23647718047,-39863.35590218902,-39864.456634004455,-39865.54038068124,-39866.6089272059,-39867.6640945041,-39868.70770108091,-39869.74152784153,-39870.76728652724,-39871.786592095734,-39872.80093926767,-39873.81168335955,-39874.82002542905,-39875.827001669335,-39876.83347691102,-39877.840142020366,-39878.84751492159,-39879.855944923635,-39880.865619990946,-39881.87657657079,-39882.88871157109,-39883.90179607298,-39884.91549036367,-39885.929359883485,-39886.94289169574,-39887.955511111955,-39888.96659813094,-39889.97550338364,-39890.981563310525,-39891.98411433691,-39892.982505850145,-39893.976111823555,-39894.9643409713,-39895.94664535774,-39896.92252742067,-39897.89154540417,-39898.853317227025,-39899.80752284332,-39900.75390517486,-39901.692269718646,-39902.622482948456,-39903.54446964563,-39904.45820930125,-39905.36373174104,-39906.2611121254,-39907.150465477265,-39908.03194088623,-39908.90571553285,-39909.77198866662,-39910.630975662236,-39911.48290226581,-39912.32799913062,-39913.166496726306,-39913.99862069216,-39914.82458769021,-39915.64460179877,-39916.4588514729,-39917.26750708498,-39918.070719046504,-39918.868616499356,-39919.661306554866,-39920.448874050686,-39921.23138178648,-39922.008871193815,-39922.781363391136,-39923.54886056973,-39924.31134765648,-39925.06879419618,-39925.82115639814,-39926.568379292745,-39927.31039894505,-39928.04714467753,-39928.77854125594,-39929.50451099904,-39930.22497577479,-39930.93985885458,-39931.64908659817,-39932.35258995236,-39933.05030574685,-39933.74217777935,-39934.428157685325,-39935.108205593526,-39935.7822905708,-39936.45039086613,-39937.112493964494,-39937.76859646578,-39938.41870380462,-39939.06282983116,-39939.70099627109,-39940.3332320862,-39940.95957275616,-39941.580059501444,-39942.19473846863,-39942.80365989603,-39943.40687727806,-39944.00444654467,-39944.59642527071,-39945.182871927755,-39945.76384519004,-39946.33940330259,-39946.90960351954,-39947.47450161682,-39948.03415148301,-39948.58860478885,-39949.13791073649,-39949.68211588461,-39950.22126404792,-39950.755396265755,-39951.284550834,-39951.80876339501,-39952.328067077855,-39952.84249268207,-39953.35206889708,-39953.85682254956,-39954.35677887166,-39954.851961782566,-39955.34239417597,-39955.828098207785,-39956.309095577395,-39956.78540779768,-39957.25705644857,-39957.72406341102,-39958.18645107731,-39958.644242535476,-39959.09746172695,-39959.54613357513,-39959.99028408494,-39960.429940413924,-39960.86513091555,-39961.29588515544,-39961.72223390329,-39962.14420910104,-39962.56184381131,-39962.97517214753,-39963.38422918853,-39963.78905088121,-39964.18967393313,-39964.5861356984,-39964.9784740594,-39965.366727306624,-39965.750934019474,-39966.131132950155,-39966.50736291153,-39966.879662672305,-39967.248070859634,-39967.61262587145,-39967.97336579778,-39968.33032835277,-39968.683550818685,-39969.03306999909,-39969.37892218434,-39969.721143126924,-39970.059768027146,-39970.394831527956,-39970.72636771825,-39971.05441014487,-39971.378991830024,-39971.700145295275,-39972.017902590414,-39972.3322953253,-39972.64335470482,-39972.951111565984,-39973.25559641435,-39973.556839461984,-39973.854870663694,-39974.14971975144,-39974.441416266534,-39974.72998959002,-39975.01546896841,-39975.297883537656,-39975.57726234189,-39975.85363434918,-39976.12702846331,-39976.397473532204,-39976.66499835151,-39976.9296316661,-39977.19140216757,-39977.45033848878,-39977.70646919617,-39977.95982277952,-39978.21042764005,-39978.45831207671,-39978.703504271725,-39978.94603227487,-39979.18592398798,-39979.423207148895,-39979.657909315945],"type":"scatter","name":"source2_q_inv","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.0,2.1224701501016327e-7,6.679667341435528e-6,5.0645383476960324e-5,0.00020538624046838555,0.0006178262381644649,0.0015078509375686475,0.0031656112338011924,0.005761691972225522,0.009245041137047591,0.01301969758676605,0.015801332175067462,0.015445163841767418,0.008807337935383908,-0.008016451944891978,-0.039862173657898126,-0.09191843010243561,-0.16996185698896626,-0.27976647473327887,-0.4272923688859578,-0.6183809271894427,-0.8587120003492634,-1.153870902311363,-1.5090209558983059,-1.9292277659254395,-2.41899421270114,-2.9826226694569673,-3.6238321253454018,-4.345964016148226,-5.151828242458539,-6.043691210039626,-7.0233632632847645,-8.092040667751721,-9.250531881327081,-10.499076796998445,-11.837576129690934,-13.265488973775156,-14.781981373547943,-16.385942881431788,-18.076044026060202,-19.850845724990158,-21.708798347349813,-23.648380360607376,-25.668080277964602,-27.7665078728772,-29.942387516024795,-32.19461574355055,-34.52227172418546,-36.924622646973575,-39.401139519113215,-41.951467802699625,-44.57543419040438,-47.27300399897514,-50.044272116262704,-52.88942285387736,-55.80870668337882,-58.80240908881299,-61.87082024388133,-65.01421345616474,-68.23281673935621,-71.5267995235132,-74.8962528957636,-78.34118310548355,-81.86150298140137,-85.45703117139763,-89.12749369490841,-92.87252827943936,-96.69169319706211,-100.58447690332288,-104.55031338853682,-108.58859274368879,-112.69867132748797,-116.8798810263153,-121.13153968803545,-125.4529616423842,-129.84346759848106,-134.3023946069981,-138.82910448252315,-143.42299140068667,-148.08348714943224,-152.8100646637266,-157.60223941196517,-162.45956914967726,-167.3816523437539,-172.36812544695678,-177.41865951914255,-182.53295619799025,-187.71074347406034,-192.95177128545265,-198.2558072274691,-203.6226324402412,-209.0520378006858,-214.54382051044513,-220.09778110644606,-225.7137209791075,-231.39144038117183,-237.13073696622268,-242.93140481599988,-248.79323393660556,-254.7160101685456,-260.69951545081636,-266.74352838007246,-272.84782499282795,-279.0121797135292,-285.2363664022883,-291.52015944984595,-297.8633348679123,-304.26567133231424,-310.72695114436885,-317.2469610825772,-323.8254931264801,-330.4623450401655,-337.1573208105581,-343.9102309409731,-350.72089260562524,-357.5891296750896,-364.51477262598996,-371.49765835067086,-378.5376298842341,-385.63453606664336,-392.78823115777755,-399.9985744225201,-407.26542970133556,-414.58866498079107,-421.96815197659095,-429.4037657384806,-436.89538428636035,-444.44288828267133,-452.0461607453154,-459.7050868031041,-467.41955349421585,-475.18944960636134,-483.0146655562221,-490.8950933050717,-498.83062630611334,-506.8211594796097,-514.8665892101574,-522.9668133620576,-531.1217313073321,-539.3312439625283,-547.5952538298462,-555.9136650391139,-564.2863833874549,-572.7133163743943,-581.1943732304112,-589.7294649367482,-598.3185042372779,-606.9614056404391,-615.6580854121547,-624.4084615601539,-633.2124538094138,-642.069983571423,-650.9809739067266,-659.9453494821039,-668.9630365252619,-678.0339627755411,-687.1580574346685,-696.3352511157084,-705.565475794017,-714.8486647592013,-724.1847525684188,-733.5736750039919,-743.0153690328838,-752.5097727694713,-762.0568254428775,-771.6564673665556,-781.3086399117116,-791.0132854841005,-800.7703475035012,-810.5797703856827,-820.441499526923,-830.3554812904989,-840.3216629945193,-850.3399929010695,-860.4104202065379,-870.5328950323769,-880.707368416109,-890.9337923028119,-901.2121195362553,-911.5423038501714,-921.9242998587414,-932.3580630474154,-942.8435497624289,-953.380717200869,-963.969523399807,-974.6099272245251,-985.3018883581324,-996.0453672885476,-1006.8403252976543,-1017.6867244484002,-1028.5845275733395,-1039.5336982620493,-1050.5342008496923,-1061.5860004047418,-1072.689062717449,-1083.8433542887803,-1095.0488423187496,-1106.3054946965244,-1117.6132799887835,-1128.9721674305583,-1140.3821269146722,-1151.8431289827556,-1163.3551448155613,-1174.9181462241486,-1186.5321056411956,-1198.1969961124016,-1209.9127912882445,-1221.6794654158089,-1233.4969933309449,-1245.365350450213,-1257.2845127634173,-1269.2544568257422,-1281.2751597504107,-1293.346599200968,-1305.4687533840045,-1317.6416010417788,-1329.8651214446343,-1342.139294383971,-1354.4641001645923,-1366.8395195979078,-1379.2655339941064,-1391.742125155504,-1404.269275369006,-1416.8469673990992,-1429.4751844808918,-1442.1539103128448,-1454.8831290499124,-1467.6628252967996,-1480.4929841006312,-1493.3735909446864,-1506.3046317413591,-1519.286092825481,-1532.3179609478284,-1545.4002232687585,-1558.532867351181,-1571.7158811551387,-1584.9492530302996,-1598.2329717107566,-1611.5670263083175,-1624.951406306573,-1638.3861015545826,-1651.8711022614139,-1665.406398989719,-1678.99198265023,-1692.6278444956379,-1706.3139761152286,-1720.0503694287495,-1733.8370166815648,-1747.6739104378787,-1761.5610435765766,-1775.4984092845345,-1789.4860010523983,-1803.5238126681913,-1817.6118382124519,-1831.7500720533933,-1845.938508840838,-1860.1771435020003,-1874.465971235551,-1888.8049875073395,-1903.1941880449833,-1917.6335688329546,-1932.1231261080604,-1946.6628563541665,-1961.2527562977955,-1975.892822903333,-1990.5830533682408,-2005.323445118891,-2020.1139958054407,-2034.954703297742,-2049.8455656808733,-2064.786581251014,-2079.7777485102597,-2094.8190661637364,-2109.9105331141004,-2125.0521484581704,-2140.2439114824792,-2155.485821659535,-2170.777878643743,-2186.12008226731,-2201.512432536187,-2216.9549296271625,-2232.4475738830215,-2247.9903658093135,-2263.5833060709256,-2279.2263954878945,-2294.919635032342,-2310.663025824748,-2326.456569130484,-2342.3002663564316,-2358.1941190476837,-2374.138128884031,-2390.1322976771603,-2406.1766273669705,-2422.2711200183808,-2438.4157778188087,-2454.6106030742235,-2470.8555982073763,-2487.1507657530246,-2503.4961083573353,-2519.891628772272,-2536.3373298556494,-2552.8332145660165,-2569.3792859610053,-2585.975547194807,-2602.6220015145873,-2619.318652259018,-2636.0655028549754,-2652.862556815105,-2669.7098177358353,-2686.607289293732,-2703.5549752452275,-2720.552879421784,-2737.6010057297226,-2754.6993581463407,-2771.8479407189006,-2789.0467575616954,-2806.295812853894,-2823.5951108382656,-2840.9446558182713,-2858.344452155827,-2875.7945042708293,-2893.2948166367264,-2910.84539378132,-2928.446240283001,-2946.0973607693563,-2963.7987599159474,-2981.550442443846,-2999.3524131178406,-3017.2046767451816,-3035.107238174034,-3053.060102291592,-3071.063274021749,-3089.1167583244496,-3107.220560194814,-3125.3746846590966,-3143.579136776344,-3161.833921633902,-3180.13904434938,-3198.494510065164,-3216.9003239513368,-3235.3564912008014,-3253.863017030188,-3272.4199066775727,-3291.0271654021094,-3309.68479848094,-3328.392811210435,-3347.151208902766,-3365.9599968871644,-3384.8191805064857,-3403.7287651169718,-3422.6887560884547,-3441.699158800503,-3460.7599786448322,-3479.871221020456,-3499.032891337361,-3518.244995010674,-3537.507537462392,-3556.8205241223072,-3576.1839604218603,-3595.5978517978238,-3615.0622036906093,-3634.577021541488,-3654.142310793818,-3673.758076891146,-3693.424325277603,-3713.1410613953594,-3732.90829068508,-3752.72601858606,-3772.5942505332187,-3792.5129919584997,-3812.482248289236,-3832.5020249474474,-3852.572327350307,-3872.6931609068747,-3892.8645310221714,-3913.086443090962,-3933.3589025015467,-3953.681914632999,-3974.055484855415,-3994.4796185300556,-4014.954321006686,-4035.4795976255105,-4056.0554537150483,-4076.6818945925124,-4097.358925563027,-4118.086551920179,-4138.86477894254,-4159.693611897101,-4180.5730560375705,-4201.503116601654,-4222.483798815019,-4243.515107885841,-4264.597049009613,-4285.729627365551,-4306.912848115271,-4328.146716407649,-4349.4312373721295,-4370.766416122568,-4392.152257755999,-4413.588767351925,-4435.075949971942,-4460.302415369324,-2014.5523187897732,3502.9010308047245,11053.08202057961,19599.46164604152,28151.818785749037,35963.113546089226,42280.60049639417,46822.542583548086,49608.72471343617,50930.154110042946,51155.297492173806,50656.42324788213,49764.994594710355,48710.8663053537,47640.4000459552,46617.87701110445,45676.26794116801,44835.56025557212,44118.580114664284,43548.61184631203,43141.85503316586,42902.55563664174,42820.958938585114,42873.91322368375,43031.00356370342,43256.69479548446,43517.513836288585,43783.243911213794,44030.224830983534,44241.74342880685,44407.03670918443,44521.76739780954,44585.01545092169,44600.04654315606,44571.54310213133,44506.18577409933,44411.16144897787,44294.15736678048,44163.07250024812,44025.40692561403,43888.54918173859,43758.88376086246,43642.15230253616,43542.76458329012,43464.073608344974,43408.15958998958,43375.983207100566,43367.54319007485,43381.906833905356,43417.51345600224,43472.16180762083,43543.30966227457,43628.10292701913,43723.59313404547,43826.81353792725,43934.88602753369,44045.1101421922,44154.989422455146,44262.310774739446,44365.1384470539,44461.86588121222,44551.1992290671,44632.168261360675,44704.10355177602,44766.61352399987,44819.5605006806,44863.023789890234,44897.27734442329,44922.75239856227,44940.01440967606,44949.73029803025,44952.64439512858,44949.552934852836,44941.28214999103,44928.67035296814,44912.55012238596,44893.7356159718,44873.008855824664,44851.11023157795,44828.72890582852,44806.49611849549,44784.97963881008,44764.67980716139,44746.0272395247,44729.38128197011,44715.03012915178,44703.19153055122,44694.01494502077,44687.58439483996,44683.922374772774,44682.99446066054,44684.714541629415,44688.950617301234,44695.53090166526,44704.25028348243,44714.87686317,44727.15857314904,44740.82965769077,44755.616950207936,44771.24581234945,44787.44564491621,44803.9549075656,44820.52556467294,44836.92693297556,44852.94887696974,44868.40435011385,44883.131268328165,44896.99373467542,44909.88263925147,44921.715671797356,44932.4367958127,44942.01523614599,44950.444042181,44957.738288827386,44963.93298299741,44969.080742106176,44973.249311273,44976.518983168695,44978.97998093376,44980.72986019202,44981.87098092988,44982.50809423087,44982.746082818485,44982.687887735396,44982.432647132206,44982.07406647695,44981.69903328442,44981.38648343812,44981.2065204876,44981.21978427706,44981.47706043074,44982.01911830356,44982.87676135556,44984.07107113722,44985.61382375147,44987.50805602534,44989.7487575696,44992.32366436533,44995.21412959058,44998.396047839284,45001.84080987642,45005.51626635443,45009.38768056892,45013.418652211294,45017.5719961866,45021.81056278829,45026.09798784742,45030.39936383341,45034.68182521286,45038.915043664856,45043.07163088814,45047.12744880352,45051.06182877986,45054.85770321236,45058.50165422195,45061.983885499816,45065.298124344416,45068.44146170433,45071.41413864713,45074.21928798527,45076.86263997629,45079.35220095014,45081.697913516735,45083.91130664206,45086.005143377144,45087.99307342562,45089.88929702191,45091.708245841386,45093.46428583131,45095.17144600802,45096.84317643386,45098.49213770974,45100.130023523234,45101.767416999086,45103.41368087506,45105.076880843204,45106.76374080554,45108.479628264744,45110.22856761669,45112.01327874881,45113.8352380702,45115.694758880076,45117.591087873814,45119.5225145241,45121.48649008358,45123.47975306392,45125.49845814684,45127.53830569524,45129.594669247526,45131.66271863296,45133.73753664398,45135.814227487455,45137.888015559685,45139.954333397596,45142.008897978376,45144.047774823426,45146.067429677874,45148.06476776815,45150.037160922024,45151.98246302301,45153.89901448231,45155.7856365646,45157.64161653387,45159.46668469123,45161.260984443805,45163.02503658905,45164.75969901113,45166.46612298054,45168.14570721209,45169.800050785416,45171.43090596163,45173.04013184683,45174.629649749724,45176.20140099017,45177.75730778266,45179.29923772657,45180.82897230342,45182.34817967282,45183.85839195696,45185.360987079985,45186.85717515612,45188.34798930555,45189.83428072514,45191.3167177458,45192.795788570445,45194.271807331665,45195.744923070175,45197.21513121519,45198.682287130585,45200.14612128623,45201.60625562297,45203.06222067953,45204.51347308959,45205.959413064076,45207.39940151942,45208.83277654041,45210.25886890806,45211.677016466616,45213.08657713712,45214.486940443036,45215.87753743367,45217.25784895431,45218.62741223212,45219.98582578907,45221.33275273128,45222.66792247849,45223.99113103733,45225.3022399331,45226.60117393353,45227.887917713655,45229.162511617695,45230.42504667476,45231.6756590354,45232.91452398591,45234.141849693544,45235.3578708348,45236.56284223794,45237.757032669215,45238.94071887297,45240.11417996301,45241.27769224734,45242.43152455297,45243.57593410507,45244.71116298929,45245.837435225905,45246.954954458,45248.06390225129,45249.16443698697,45250.25669331683,45251.34078215355,45252.416791138196,45253.48478554965,45254.54480958894,45255.59688798984,45256.64102789177,45257.67722091731,45258.7054453946,45259.725668673935,45260.73784947071,45261.74194020478,45262.73788927313,45263.725643225225,45264.70514880433,45265.6763548222,45266.6392138479,45267.59368368942,45268.53972865484,45269.47732059358,45270.40643970315,45271.32707511558,45272.23922526446,45273.142898045415,45274.03811078113,45274.92489001325,45275.80327113173,45276.67329787077,45277.53502168477,45278.38850103205,45279.233800584094,45280.07099038472,45280.900144977735,45281.72134252385,45282.534663927334,45283.340191981566,45284.138010555354,45284.928203834366,45285.71085561602,45286.486048681385,45287.25386424038,45288.01438145961,45288.76767707081,45289.51382506614,45290.25289647395,45290.984959215224,45291.71007803455,45292.4283145047,45293.13972709171,45293.844371278945,45294.54229974116,45295.23356256129,45295.9182074798,45296.5962801686,45297.26782452822,45297.932882987676,45298.59149681529,45299.243706421665,45299.88955165245,45300.52907207223,45301.162307221784,45301.78929685694,45302.41008116405,45303.02470093899,45303.633197744646,45304.235614032455,45304.83199323791,45305.422379841155,45306.00681940383,45306.58535857436,45307.158045070624,45307.72492763587,45308.286055979275,45308.84148069715,45309.39125317513,45309.935425484226,45310.47405026693,45311.00718061503,45311.53486994415,45312.05717187239,45312.57414009686,45313.08582827257,45313.59228990204,45314.09357823212,45314.58974615364,45315.080846116995,45315.56693005885,45316.04804933526,45316.52425467037,45316.9955961174,45317.462123024816,45317.923884021315,45318.38092700583,45318.83329914839,45319.281046902084,45319.724216016024,45320.16285156336,45320.596997970315,45321.02669904764,45321.45199803026,45321.87293761663,45322.28956000966,45322.701906955765,45323.11001978913,45323.51393946814,45323.91370661289,45324.309361541804,45324.700944300064,45325.0884946907,45325.472052294564,45325.85165649545,45326.22734649201,45326.59916131193,45326.96713981821,45327.33132071563,45327.691742550305,45328.048443706466,45328.401462402035,45328.750836679406,45329.09660439653,45329.43880321216,45329.77747057563,45330.11264370982,45330.44435959554,45330.77265495824,45331.097566251876,45331.41912964209,45331.73738099203,45332.05235585235,45332.36408944578],"type":"scatter","name":"source1_q_poc","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.0,0.002690863974277018,0.01158901363789096,0.02991055431859783,0.060515274366018515,0.10596818929847217,0.1686937739701167,0.2509060696216358,0.3546608506370764,0.4817573313576914,0.6337107697555519,0.8116736689955861,1.0163857838953236,1.2481530819679825,1.5068311723846792,1.791866217785353,2.1023273275308365,2.436981455688561,2.794353815679511,3.1728023193593495,3.5705797293188364,3.9858903468091618,4.41694072978418,4.861978035394495,5.319326516314822,5.787411858993028,6.264783660564224,6.7501284286539125,7.242278952336591,7.740217365579367,8.24307362355582,8.750120411144048,9.260763953230612,9.774533157363061,10.291066097484366,10.810096051523102,11.33143659106857,11.854967058996387,12.380618511501986,12.908360609373405,13.438189812194489,13.970118855475336,14.504167854115986,15.040356841570901,15.578699924334643,16.119200854400447,16.661850008173012,17.206622600834272,17.753477992515542,18.302359932069436,18.853197545208133,19.40590692076807,19.960393111261286,20.516552412279054,21.074274774100783,21.63344623278495,22.193951258481544,22.755674941711224,23.31850495637774,23.88233325662369,24.44705748066033,25.012582050864413,25.57881897075959,26.145688332119526,26.713118552779157,27.281046372849154,27.849416641755138,28.418181930911043,28.987302016491352,29.556743273583365,30.126477988316864,30.69648359492723,31.266741879387432,31.8372382031829,32.40796078734914,32.978900072678826,33.55004815586554,34.12139829724805,34.69294449724886,35.26468114162883,35.836602715440485,36.40870358448341,36.98097784056259,37.5534192050253,38.12602098363568,38.69877606493405,39.271676954036266,39.844715833683175,40.41788464486308,40.99117517977647,41.564579180798916,42.138088439946465,42.71169489432224,43.28539071398241,43.85916837951878,44.43302074752186,45.006941102790826,45.58092319686461,46.154961273002385,46.729050078254566,47.30318486365204,47.87736137383767,48.45157582767042,49.02582489143634,49.60010564634568,50.17441555195364,50.748752407064615,51.323114309544366,51.89749961630452,52.47190690454286,53.04633493512608,53.62078261881034,54.19524898580149,54.769733158980095,55.34423433095141,55.91875174493701,56.493284679398144,57.06783243618087,57.642394331892724,58.216969692161754,58.79155784839392,59.36615813662391,59.94076989805481,60.515392480891705,61.09002524310004,61.664667555749105,62.23931880664215,62.81397840397557,63.38864577981525,63.96332039322232,64.53800173290483,65.11268931931147,65.68738270612074,66.2620814811121,66.83678526643152,67.41149371828897,67.98620652614065,68.56092341142303,69.13564412591373,69.71036844979706,70.28509618951495,70.85982717547944,71.43456125971815,72.0092983135191,72.58403822512972,73.15878089756033,73.73352624652722,74.30827419856718,74.88302468934128,75.45777766213965,76.03253306659131,76.60729085757602,77.18205099432922,77.75681343972718,78.33157815973668,78.90634512300926,79.48111430060081,80.0558856657957,80.63065919401477,81.20543486278791,81.7802126517736,82.35499254280774,82.92977451996883,83.50455856964673,84.07934468060526,84.65413284403039,85.22892305355963,85.80371530528771,86.3785095977473,86.95330593186455,87.5281043108895,88.1029047403044,88.67770722771188,89.25251178270582,89.82731841673002,90.40212714292647,90.97693797597857,91.55175093195088,92.12656602813125,92.70138328287523,93.27620271545744,93.85102434593156,94.42584819499844,95.00067428388685,95.57550263424453,96.15033326804145,96.72516620748479,97.30000147494495,97.87483909289246,98.44967908384515,99.02452147032336,99.59936627481433,100.17421351974248,100.7490632274459,101.32391542015755,101.89877011999071,102.473627348926,103.04848712880232,103.62334948130726,104.19821442797003,104.7730819901536,105.34795218904677,105.92282504565664,106.49770058079999,107.07257881509415,107.64745976894767,108.2223434625496,108.79722991585929,109.37211914859499,109.94701118022272,110.52190602994465,111.09680371668806,111.67170425909437,112.24660767550817,112.82151398396725,113.39642320219318,113.97133534758215,114.54625043719753,115.12116848776206,115.6960895156523,116.27101353689243,116.84594056715008,117.42087062173267,117.99580371558429,118.57073986328363,119.14567907904254,119.72062137670507,120.29556676974738,120.87051527127778,121.44546689403762,122.02042165040245,122.5953795523831,123.17034061162784,123.74530483942387,124.32027224669973,124.89524284402769,125.47021664162571,126.0451936493608,126.62017387675063,127.19515733296704,127.77014402683837,128.34513396685261,128.92012716115997,129.49512361757604,130.0701233435847,130.6451263463413,131.22013263267561,131.79514220909522,132.3701550817885,132.94517125662853,133.5201907391755,134.0952135346812,134.67023964809198,135.2452690840525,135.82030184690942,136.3953379407152,136.9703773692316,137.54542013593417,138.12046624401543,138.6955156963892,139.27056849569487,139.84562464430076,140.42068414430858,140.99574699755763,141.57081320562887,142.14588276984858,142.7209556912931,143.29603197079263,143.87111160893525,144.4461946060715,145.02128096231806,145.5963706775617,146.17146375146416,146.74656018346525,147.3216599727875,147.89676311844002,148.47186961922245,149.04697947372895,149.6220926803522,150.19720923728715,150.772329142535,151.34745239390733,151.9225789890293,152.49770892534397,153.0728422001157,153.64797881043447,154.2231187532187,154.79826202521937,155.37340862302398,155.94855854305945,156.52371178159612,157.09886833475113,157.6740281984918,158.2491913686393,158.82435784087195,159.39952761072826,159.97470067361076,160.54987702478905,161.12505665940316,161.70023957246613,162.2754257588683,162.85061521337948,163.4258079306522,164.0010039052253,164.57620313152617,165.15140560387414,165.72661131648337,166.30182026346557,166.87703243883288,167.45224783650067,168.02746645029032,168.60268827393196,169.1779133010668,169.7531415252503,170.32837293995422,170.9036075385693,171.47884531440837,172.0540862607074,172.62933037062984,173.2045776372671,173.7798280536418,174.35508161271025,174.93033830736442,175.5055981304343,176.08086107468964,176.6561271328426,177.2313962975495,177.80666856141346,178.38194391698565,178.9572223567677,179.53250387321367,180.1077884587324,180.683076105688,181.25836680640353,181.8336605531613,182.4089573382055,182.98425715374398,183.55955999194927,184.13486584496107,184.71017470488727,185.28548656380607,185.86080141376732,186.43611924679402,187.011440054884,187.5867638300114,188.16209056412816,188.73742024916498,189.31275287703386,189.88808843962815,190.46342692882482,191.03876833648533,191.61411265445764,192.18945987457587,192.76480998866387,193.3401629885343,193.91551886599098,194.49087761283002,195.0662392208402,195.64160368180518,196.21697098750383,196.79234112971133,197.3677141002008,197.94308989074366,198.51846849311113,199.09384989907477,199.66923410040766,200.2446210888856,200.8200108562875,201.39540339439685,201.97079869500206,202.54619674989786,203.1215975508855,203.69700108977378,204.27240735838063,204.84781634853286,205.42322805206717,205.9986424608313,206.57405956668455,207.149479361498,207.7249018371561,208.30032698555672,208.87575479861232,209.45118526824956,210.026618386411,210.60205414505577,211.17749253615915,211.75293355171397,212.32837718373105,212.90382342423953,213.47927226528768,214.05472369894338,214.63017771729415,215.2056343124485,215.78109347653583,216.3565552017069,216.932019480135,217.50748630401515,218.08295566556558,218.65842755702752,219.23390197066666,219.80937889877208,220.60838789273532,223.7192834951149,231.57414755464404,244.26291506891678,258.7383035180778,270.4142068624807,275.90148848867534,274.37896423530856,267.25021714977146,256.85330332595953,245.4177558674554,234.56867913448403,225.27773787336332,217.9902811400878,212.75422122442208,209.36245802498016,207.4694986803013,206.71471392922308,206.79235118567976,207.48279419549405,208.6392465570746,210.15786465322796,211.95154341361174,213.93086917794994,216.00134374932017,218.06449989054752,220.0272136390574,221.80941289800123,223.3520483325198,224.6209627538818,225.60665789674346,226.32091302355124,226.79044179963424,227.0509357221669,227.14124099086285,227.09975294855636,226.9619588376649,226.75943408987564,226.51946120795424,226.2649703574341,226.014723657523,225.7834311448474,225.58205948726254,225.41806620600582,225.29577320742055,225.21671867186177,225.18008518853665,225.183162698359,225.2218372453965,225.29110343893848,225.3855240542591,225.49964018991213,225.62827076139754,225.7667245828859,225.9109162234054,226.05741578016364,226.2034485378613,226.34685950236258,226.48605760486717,226.61994271220743,226.74782652744636,226.86934936690997,226.98440062445283,227.09304516110808,227.19545900384082,227.29187566462502,227.38254371118927,227.46769613181044,227.54753095295013,227.62220236577542,227.69182067674916,227.75645901889672,227.81616461773507,227.87097248701446,227.92091980600944,227.96605960970214,228.00647267149606,228.0422768976357,228.07363364468617,228.10075086339947,228.12388306960034,228.14332851017625,228.1594239489753,228.17253758725974,228.18306065875655,228.19139816474208,228.19795927678942,228.20314782521402,228.20735333384252,228.21094292970017,228.2142544126208,228.21759064671627,228.2212153501995,228.22535029830138,228.23017388144956,228.23582094654307,228.24238379168486,228.24991417869597,228.25842618931378,228.26789975017894,228.27828464689117,228.2895048579818,228.3014630595127,228.31404516395688,228.3271247809509,228.3405674982521,228.35423490264174,228.36798827389538,228.38169190431765,228.39521601197112,228.40843923001205,228.42125066728534,228.4335515433896,228.44525641024805,228.4562939772534,228.46660756353367,228.4761552048704,228.4849094468609,228.4928568584008,228.49999730101783,228.50634299056,228.5119173876699,228.51675395353095,228.52089480653353,228.52438931461248,228.52729265630018,228.52966438148965,228.53156700034978,228.53306462592994,228.53422169293532,228.5351017718277,228.53576649407177,228.53627460085096,228.53668112413075,228.53703670553264,228.53738705523213,228.53777255006352,228.538227967257,228.53878234779435,228.53945898125752,228.54027550228582,228.541244087344,228.54237173945648,228.54366064786046,228.54510860915968,228.54670949651177,228.54845376359214,228.55032897054852,228.5523203198342,228.55441119065733,228.55658366179009,228.55881901357313,228.5610982011564,228.56340229223764,228.56571286382717,228.56801235381604,228.57028436434777,228.57251391518182,228.5746876463396,228.57679397037444,228.57882317554075,228.58076748199102,228.58262105386095,228.58437997072542,228.586042162418,228.5876073115879,228.58907672865163,228.59045320394802,228.59174084196752,228.5929448824748,228.59407151321125,228.59512767864126,228.5961208889202,228.59705903290623,228.59795019864643,228.59880250432346,228.59962394220327,228.60042223763506,228.6012047246939,228.60197823957856,228.60274903243283,228.60352269782686,228.6043041237406,228.605097458536,228.60590609508023,228.6067326709133,228.60757908312212,228.6084465164061,228.60933548267428,228.61024587043428,228.61117700217818,228.61212769796467,228.6130963434261,228.61408096048825,228.6150792791874,228.61608880907357,228.61710690883476,228.61813085292061,228.61915789410835,228.62018532112103,228.62121051058517,228.6222309727803,228.62324439080442,228.62424865294233,228.6252418781716,228.6262224348852,228.62718895303388,228.62814033000768,228.62907573066857,228.62999458203035,228.63089656314278,228.63178159079115,228.63264980164195,228.63350153149943,228.63433729232156,228.63515774764863,228.6359636870637,228.63675600027753,228.6375356513867,228.6383036538012,228.6390610462886,228.63980887051562,228.64054815041172,228.64127987361152,228.6420049751723,228.6427243237023,228.64343870997064,228.64414883802112,228.64485531875724,228.64555866591644,228.64625929431807,228.64695752022664,228.6476535636479,228.64834755234992,228.6490395273852,228.64972944987952,228.650417208845,228.65110262977822,228.6517854838079,228.6524654971609,228.6531423607372,228.65381573959075,228.65448528213764,228.65515062893297,228.6558114208748,228.6564673067251,228.65711794985089,228.6577630341218,228.65840226891328,228.65903539319422,228.65966217869266,228.66028243215484,228.6608959967282,228.6615027525133,228.66210261634595,228.6626955408734,228.66328151300743,228.66386055183236,228.66443270605788,228.66499805110342,228.66555668590445,228.66610872952413,228.66665431765753,228.6671935991037,228.6677267322783,228.6682538818343,228.66877521544708,228.669290900812,228.6698011029019,228.6703059815085,228.6708056891009,228.67130036901042,228.67179015395433,228.67227516489692,228.6727555102431,228.6732312853524,228.67370257235453,228.674169440248,228.67463194525206,228.6750901313846,228.67554403123856,228.67599366691985,228.67643905111726,228.67688018827135,228.67731707581035,228.6777497054227,228.6781780643389,228.67860213659176,228.67902190424084,228.67943734852824,228.67984845095623,228.6802551942705,228.68065756333422,228.6810555458869,228.68144913318253,228.681838320504,228.68222310755226,228.68260349871565,228.68297950322108,228.68335113517645,228.6837184135103,228.68408136181935,228.68444000813523,228.6847943846201,228.68514452720314,228.68549047517382,228.68583227073628,228.68616995854418,228.68650358522214,228.68683319888663,228.68715884867356,228.68748058428469,228.68779845555716,228.68811251206313,228.68842280274708,228.68872937560292,228.68903227739244,228.6893315534122,228.68962724730255,228.6899194009063,228.69020805416883,228.6904932450832,228.69077500967506,228.69105338202476,228.69132839432274,228.6916000769555,228.6918684586171,228.69213356643988,228.6923954261452,228.69265406220288,228.6929094980021,228.69316175602324,228.69341085801236,228.6936568251507,228.69389967821954,228.69413943775282,228.69437612418267,228.6946097579675,228.69484035970848,228.69506795024677,228.69529255074795,228.69551418276617,228.69573286829495,228.69594862979795,228.69616149022795,228.69637147302842,228.6965786021231,228.6967829018921,228.69698439713943,228.69718311305036,228.69737907514036,228.69757230919893,228.69776284123088,228.69795069739274,228.6981359039274,228.69831848710223,228.69849847314543,228.6986758881866,228.69885075820056,228.69902310895506,228.69919296596456,228.69936035444812,228.6995252992954,228.69968782503653,228.69984795581954,228.70000571539458,228.70016112710172,228.7003142138678,228.70046499820575,228.7006135022199,228.70075974761767,228.70090375572067,228.701045547484,228.70118514351475,228.7013225640938,228.70145782920065,228.7015909585355,228.70172197154596,228.70185088744964,228.70197772526046,228.70210250380939,228.70222524176765,228.7023459576657,228.7024646699121,228.70258139680934,228.70269615656753,228.70280896731583,228.70291984711287,228.70302881395185,228.70313588576738,228.70324108043647,228.70334441578095,228.70344590956492,228.70354557949398,228.70364344320873,228.7037395182823,228.70383382221203,228.7039263724144,228.70401718621676],"type":"scatter","name":"source1_v_mag_poc","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[50.0,50.0,49.9999999999998,49.99999999999651,49.99999999997846,49.99999999991686,49.999999999754216,49.99999999939302,49.99999999868097,49.9999999974027,49.9999999952678,49.9999999919103,49.99999998688751,49.99999997968804,49.99999996974513,49.999999956449834,49.99999993917528,49.99999991729042,49.99999989018638,49.999999857288,49.9999998180735,49.999999772082575,49.99999971892458,49.999999658282555,49.99999958991015,49.999999513631074,49.99999942932871,49.999999336941116,49.99999923644829,49.99999912786395,49.99999901122424,49.99999888657848,49.99999875398117,49.99999861348405,49.999998465131696,49.99999830895639,49.999998144976715,49.99999797319574,49.99999779360175,49.999997606169266,49.999997410861134,49.999997207631324,49.9999969964278,49.999996777195946,49.999996549881565,49.99999631443407,49.9999960708091,49.99999581897091,49.99999555889428,49.9999952905658,49.99999501398484,49.999994729163824,49.99999443612821,49.999994134915944,49.9999938255766,49.9999935081702,49.99999318276585,49.99999284944019,49.9999925082758,49.999992159359586,49.999991802781146,49.99999143863134,49.99999106700092,49.99999068797931,49.99999030165366,49.999989908108006,49.99998950742269,49.99998909967392,49.99998868493359,49.999988263269245,49.99998783474416,49.99998739941763,49.99998695734524,49.99998650857928,49.99998605316919,49.99998559116198,49.99998512260272,49.99998464753498,49.999984166001255,49.99998367804334,49.99998318370268,49.999982683020626,49.99998217603871,49.999981662798774,49.99998114334308,49.999980617714385,49.99998008595595,49.99997954811149,49.99997900422513,49.99997845434128,49.999977898504554,49.99997733675961,49.99997676915104,49.99997619572319,49.99997561652009,49.999975031585265,49.999974440961694,49.99997384469164,49.99997324281665,49.99997263537743,49.999972022413836,49.99997140396484,49.99997078006853,49.9999701507621,49.9999695160819,49.999968876063434,49.99996823074143,49.999967580149864,49.99996692432203,49.999966263290574,49.99996559708757,49.99996492574454,49.99996424929259,49.99996356776234,49.999962881184054,49.999962189587656,49.99996149300273,49.99996079145861,49.999960084984366,49.999959373608824,49.99995865736057,49.999957936268004,49.999957210359284,49.999956479662366,49.99995574420501,49.99995500401475,49.99995425911893,49.99995350954464,49.999952755318766,49.99995199646799,49.99995123301873,49.999950464997184,49.999949692429325,49.99994891534088,49.99994813375735,49.999947347704,49.99994655720587,49.99994576228776,49.999944962974254,49.99994415928973,49.99994335125836,49.99994253890407,49.999941722250654,49.99994090132166,49.99994007614047,49.99993924673028,49.999938413114144,49.99993757531493,49.999936733355334,49.999935887257934,49.99993503704513,49.99993418273919,49.99993332436224,49.999932461936275,49.99993159548314,49.999930725024555,49.99992985058212,49.999928972177294,49.999928089831435,49.99992720356573,49.99992631340128,49.99992541935906,49.99992452145989,49.999923619724505,49.9999227141735,49.99992180482734,49.99992089170638,49.99991997483085,49.999919054220875,49.99991812989642,49.999917201877366,49.999916270183476,49.999915334834355,49.99991439584954,49.9999134532484,49.99991250705024,49.99991155727419,49.99991060393931,49.99990964706453,49.99990868666865,49.999907722770374,49.999906755388274,49.999905784540836,49.9999048102464,49.999903832523216,49.999902851389415,49.999901866863006,49.9999008789619,49.99989988770388,49.999898893106646,49.99989789518775,49.99989689396465,49.9998958894547,49.99989488167513,49.99989387064308,49.99989285637554,49.99989183888942,49.999890818201514,49.9998897943285,49.999888767286954,49.99988773709332,49.99988670376396,49.99988566731512,49.99988462776291,49.99988358512335,49.999882539412354,49.99988149064572,49.999880438839135,49.99987938400818,49.99987832616834,49.99987726533495,49.99987620152329,49.99987513474849,49.999874065025594,49.999872992369525,49.99987191679511,49.99987083831707,49.99986975695001,49.99986867270844,49.99986758560675,49.99986649565924,49.99986540288009,49.99986430728339,49.99986320888311,49.99986210769314,49.99986100372723,49.99985989699909,49.99985878752225,49.9998576753102,49.999856560376294,49.999855442733804,49.99985432239589,49.999853199375615,49.99985207368596,49.99985094533978,49.99984981434985,49.99984868072886,49.999847544489356,49.99984640564385,49.99984526420472,49.99984412018426,49.99984297359466,49.99984182444803,49.999840672756406,49.99983951853168,49.99983836178569,49.999837202530195,49.99983604077683,49.99983487653716,49.99983370982266,49.99983254064471,49.99983136901463,49.99983019494362,49.99982901844282,49.999827839523256,49.99982665819591,49.99982547447165,49.99982428836128,49.99982309987551,49.99982190902498,49.999820715820235,49.99981952027177,49.99981832238998,49.99981712218517,49.9998159196676,49.999814714847425,49.99981350773475,49.9998122983396,49.999811086671905,49.99980987274155,49.999808656558336,49.99980743813198,49.99980621747216,49.999804994588466,49.999803769490406,49.999802542187446,49.99980131268897,49.9998000810043,49.999798847142685,49.999797611113316,49.999796372925324,49.99979513258777,49.99979389010964,49.999792645499895,49.9997913987674,49.99979014992096,49.99978889896934,49.99978764592124,49.999786390785296,49.999785133570086,49.999783874284134,49.99978261293592,49.99978134953383,49.999780084086254,49.99977881660147,49.99977754708773,49.999776275553245,49.999775002006146,49.99977372645453,49.99977244890644,49.99977116936987,49.99976988785276,49.99976860436299,49.999767318908425,49.99976603149685,49.99976474213601,49.999763450833626,49.99976215759733,49.999760862434755,49.99975956535346,49.999758266360956,49.99975696546474,49.99975566267223,49.99975435799083,49.999753051427874,49.999751742990696,49.99975043268654,49.999749120522644,49.999747806506186,49.99974649064432,49.99974517294415,49.99974385341275,49.99974253205714,49.99974120888432,49.99973988390124,49.999738557114824,49.999737228531956,49.99973589815947,49.999734566004186,49.999733232072884,49.99973189637229,49.999730558909114,49.99972921969004,49.99972787872169,49.999726536010684,49.999725191563584,49.99972384538693,49.99972249748724,49.99972114787098,49.999719796544596,49.9997184435145,49.99971708878708,49.99971573236869,49.99971437426565,49.99971301448426,49.99971165303076,49.999710289911405,49.999708925132396,49.99970755869991,49.99970619062009,49.999704820899055,49.99970344954291,49.999702076557725,49.999700701949514,49.99969932572431,49.99969794788808,49.999696568446794,49.99969518740639,49.999693804772754,49.99969242055179,49.999691034749326,49.99968964737121,49.999688258423255,49.99968686791122,49.99968547584087,49.99968408221794,49.99968268704813,49.99968129033712,49.99967989209058,49.999678492314146,49.999677091013425,49.999675688194,49.999674283861445,49.9996728780213,49.999671470679075,49.99967006184029,49.999668651510405,49.99966723969488,49.99966582639914,49.999664411628586,49.99966299538862,49.99966157768461,49.99966015852189,49.99965873790579,49.9996573158416,49.999655892334616,49.99965446739009,49.99965304101327,49.99965161320936,49.99965018398356,49.999648753341056,49.999647321287,49.99964588782654,49.99964445296477,49.99964301670681,49.99964157905773,49.99964014002258,49.9996386996064,49.99963725781422,49.99963581465103,49.999634370121804,49.99963292423152,49.99963146448597,49.99964982265815,49.99970285726653,49.99980166774773,49.99995217049383,50.0001470600161,50.00037145115073,50.000604412216944,50.00083001399288,50.001036928931136,50.00122112164335,50.00138241011162,50.001523145987726,50.001646834523896,50.001756707058505,50.001856069017606,50.00194737648335,50.00203296779746,50.00211458564813,50.00219384515978,50.002272100611066,50.00235054974406,50.00243028859349,50.002512175606206,50.00259694508827,50.00268502749887,50.00277666981938,50.00287184908351,50.00297037529192,50.003071909494956,50.00317602327286,50.00328226043104,50.00339014913978,50.00349925718438,50.003609180019716,50.00371957323098,50.00383013935807,50.00394063849757,50.00405088138705,50.004160725194076,50.00427007123684,50.004378854622765,50.00448704313118,50.00459462668094,50.004701614854206,50.004808028691095,50.00491389675604,50.005019250103125,50.00512411855451,50.00522852870417,50.005332501692514,50.00543605321189,50.00553919285954,50.00564192500273,50.00574424922823,50.00584616142772,50.00594765480293,50.00604872089008,50.00614935062954,50.006249535167434,50.006349266661886,50.00644853878704,50.00654734716128,50.006645689542886,50.00674356589646,50.00684097830366,50.00693793074912,50.0070344288193,50.00713047933077,50.00722608993119,50.007321268693325,50.007416023725966,50.0075103628217,50.00760429315131,50.00769782101636,50.00779095166512,50.007883689170065,50.00797603637002,50.00806799486437,50.008159565060105,50.00825074625613,50.00834153676238,50.008431934041575,50.008521934866984,50.008611535489145,50.008700731802314,50.00878951950773,50.00887789426541,50.008965851833985,50.0090533881935,50.00914049965132,50.00922718292952,50.00931343523365,50.00939925430385,50.009484638448356,50.0095695865617,50.009654098128145,50.0097381732129,50.00982181244228,50.00990501697477,50.00998778846456,50.01007012901906,50.01015204115206,50.0102335277336,50.01031459193814,50.01039523719192,50.0104754671207,50.01055528549855,50.010634696198736,50.01071370314713,50.010792310278866,50.01087052149863,50.01094834064506,50.011025771459444,50.01110281755894,50.01117948241446,50.011255769333125,50.01133168144533,50.01140722169623,50.01148239284141,50.011557197446535,50.01163163789061,50.01170571637245,50.01177943492002,50.01185279540222,50.011925799542546,50.01199844893441,50.012070745057486,50.01214268929468,50.012214282949515,50.01228552726331,50.01235642343192,50.01242697262181,50.01249717598509,50.01256703467332,50.01263654984996,50.012705722701355,50.012774554445976,50.01284304634217,50.012911199694116,50.01297901585623,50.01304649623592,50.01311364229487,50.01318045554886,50.01324693756643,50.01331308996618,50.0133789144133,50.01344441261513,50.013509586315976,50.01357443729156,50.013638967342956,50.0137031782904,50.013767071966946,50.01383065021223,50.013893914866365,50.01395686776421,50.01401951072989,50.014081845571866,50.01414387407848,50.014205598014094,50.0142670191158,50.01432813909077,50.014388959614216,50.01444948232794,50.01450970883954,50.014569640722094,50.014629279514445,50.014688626721906,50.014747683817404,50.014806452243036,50.01486493341186,50.01492312870997,50.014981039498814,50.01503866711755,50.01509601288552,50.015153078104746,50.015209864062435,50.01526637203329,50.0153226032819,50.01537855906478,50.01543424063237,50.01548964923082,50.01554478610353,50.015599652492455,50.015654249639226,50.015708578786054,50.01576264117631,50.015816438055005,50.015869970669,50.015923240267036,50.01597624809958,50.01602899541853,50.01608148347681,50.016133713527765,50.01618568682459,50.01623740461959,50.0162888681634,50.01634007870427,50.0163910374872,50.01644174575323,50.0164922047386,50.01654241567409,50.01659237978434,50.016642098287264,50.01669157239346,50.01674080330581,50.01678979221906,50.01683854031953,50.01688704878493,50.01693531878418,50.01698335147734,50.01703114801568,50.017078709541714,50.01712603718935,50.01717313208409,50.01721999534327,50.017266628076335,50.01731303138517,50.017359206364404,50.01740515410177,50.01745087567848,50.017496372169525,50.01754164464411,50.01758669416592,50.017631521793476,50.01767612858042,50.01772051557581,50.017764683824375,50.01780863436667,50.01785236823938,50.017895886475344,50.01793919010379,50.01798228015037,50.01802515763725,50.01806782358313,50.01811027900328,50.01815252490952,50.01819456231016,50.01823639221,50.018278015610235,50.018319433508374,50.01836064689819,50.01840165676957,50.018442464108475,50.01848306989684,50.018523475112424,50.01856368072881,50.01860368771523,50.018643497036535,50.018683109653146,50.01872252652094,50.01876174859124,50.01880077681076,50.0188396121216,50.01887825546121,50.01891670776236,50.0189549699532,50.01899304295725,50.01903092769339,50.01906862507595,50.019106136014706,50.019143461414934,50.01918060217746,50.01921755919871,50.01925433337079,50.0192909255815,50.019327336714426,50.019363567648995,50.01939961926051,50.01943549242022,50.01947118799539,50.01950670684932,50.019542049841405,50.01957721782719,50.019612211658384,50.01964703218292,50.01968168024496,50.01971615668494,50.019750462339616,50.01978459804201,50.01981856462149,50.01985236290376,50.01988599371087,50.01991945786121,50.01995275616952,50.019985889446914,50.020018858500855,50.020051664135146,50.02008430714995,50.02011678834179,50.02014910850351,50.02018126842434,50.02021326888981,50.020245110681834,50.02027679457865,50.02030832135484,50.020339691781345,50.02037090662547,50.02040196665087,50.020432872617576,50.020463625281984,50.020494225396895,50.02052467371151,50.02055497097145,50.020585117918756,50.020615115291896,50.02064496382585,50.020674664252034,50.02070421729838,50.02073362368933,50.020762884145874,50.020791999385544,50.02082097012244,50.02084979706729,50.020878480927394,50.0209070224067,50.02093542220582,50.02096368102199,50.02099179954917,50.02101977847803,50.02104761849592,50.02107532028696,50.02110288453201,50.02113031190867,50.02115760309136,50.021184758751275,50.02121177955642,50.0212386661716,50.021265419258484,50.02129203947557,50.02131852747819,50.02134488391859,50.02137110944585,50.02139720470597,50.02142317034184,50.02144900699326,50.021474715296954,50.0215002958866,50.021525749392765,50.021551076443046,50.02157627766196,50.02160135367101,50.02162630508871,50.02165113253055,50.021675836609056,50.021700417933786,50.0217248771113,50.02174921474526,50.02177343143637,50.02179752778241,50.02182150437826,50.02184536181589,50.02186910068441,50.021892721570055,50.02191622505619,50.02193961172335,50.02196288214923,50.02198603690874,50.022009076573966,50.022032001714194,50.022054812895966,50.02207751068304,50.02210009563643,50.02212256831443,50.02214492927259,50.02216717906378,50.022189318238134,50.02221134734313,50.02223326692357,50.0222550775216,50.02227677967671,50.022298373925764,50.022319860803016,50.022341240840085,50.022362514566,50.02238368250722,50.02240474518761,50.02242570312848,50.022446556848585,50.022467306864165,50.022487953688895,50.022508497833954,50.022528939808026,50.02254928011729,50.02256951926544,50.02258965775371,50.02260969608088,50.02262963474325,50.02264947423471,50.02266921504674,50.02268885766837,50.022708402586254,50.02272785028465,50.02274720124544,50.02276645594812,50.022785614869846,50.022804678485436,50.022823647267344,50.022842521685725],"type":"scatter","name":"source1_freq","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[50.0,50.0,46.15163652864002,45.03782071076938,43.733350316914056,42.59665147025594,41.830507598972424,41.266602511797515,40.96494442963539,40.783823848232586,40.66309483798052,40.597494611858494,40.58073783373449,40.60629925665651,40.668211019806854,40.75991710471044,40.87491116746678,41.00747371547495,41.152958859892145,41.30783535067036,41.46957403140937,41.6364844195589,41.807536529802064,41.98219681244397,42.16027722894724,42.341803607591245,42.526907936741594,42.71574477993167,42.90843354557341,43.10502035128093,43.30545716891396,43.509591952273084,43.717167258738215,43.92782411303065,44.14110956805553,44.356487117916195,44.573349091438594,44.79103107531439,45.00882758753701,45.226008816522615,45.44183753000392,45.655585615212985,45.86654953055923,46.074064149847686,46.27751464292792,46.47634607737315,46.67007066619298,46.85827256510444,47.040610336475076,47.21681720442363,47.386699362161266,47.550132616466236,47.70705768902445,47.85747450347366,48.00143575509565,48.139040056564106,48.27042490490041,48.39575970163927,48.51523901257493,48.629076225989124,48.73749772723865,48.84073767165908,48.93903340588148,49.0326215558609,49.12173477938377,49.20659915862649,49.28786214622742,49.36677727149642,49.443426558274055,49.51780502240076,49.58997509586651,49.65997835316408,49.727848034464444,49.79361747779296,49.85731767447942,49.91897713677014,49.97862261938442,50.03627961681713,50.091972838642356,50.14572665913969,50.197565516646456,50.24751427568264,50.29559854544908,50.34184493715016,50.38628125045045,50.42893659088732,50.46984142680174,50.509027595189046,50.54652826390541,50.58237785615003,50.61661194305728,50.6492671109292,50.68038081010721,50.70999119234569,50.73813694291794,50.76485711285773,50.79019095589328,50.814177773798605,50.836856773060255,50.85826693493429,50.87844690019323,50.89743486917315,50.91526851715252,50.931984924629134,50.947620521695384,50.96221104543597,50.97579150907702,50.988396181492604,51.00005857561863,51.01081144432586,51.02068678234992,51.02971583296107,51.0379290981677,51.04535635137738,51.0520266515837,51.05796835829509,51.0632091465702,51.06777602166656,51.07169533294061,51.07499278675668,51.07769345826543,51.0798218020015,51.08140166132247,51.0824562767686,51.08300829346553,51.08307976772192,51.08269217299047,51.081866405368146,51.08062278880859,51.07898108021131,51.07696047453674,51.07457961007869,51.07185657400403,51.068808908247036,51.06545361582388,51.06180716761016,51.05788550960514,51.0537040706863,51.049277770842885,51.04462102986333,51.039747776439874,51.034671457646134,51.029405048737146,51.02396106321723,51.01835156312099,51.01258816945137,51.006682072722356,51.00064404355468,50.99448444327995,50.9882132345102,50.981839991636484,50.97537391122637,50.96882382229254,50.962198196412885,50.95550515768521,50.948752492504745,50.941947659156234,50.93509779721671,50.92820973676599,50.92129000740619,50.91434484709209,50.90738021077644,50.90040177887387,50.89341496554909,50.886424926834536,50.87943656858221,50.872454554255,50.865483312562034,50.85852704494132,50.85158973289424,50.84467514517352,50.83778684482788,50.83092819610396,50.82410237120783,50.817312356925896,50.81056096110599,50.80385081899897,50.7971843994598,50.79056401100959,50.78399180775656,50.77746979517727,50.770999835757095,50.764583654490444,50.75822284424116,50.75191887096268,50.745673078779944,50.73948669493284,50.733360834582726,50.72729650548405,50.721294612520886,50.715355962112454,50.7094812664878,50.703671147832324,50.6979261423087,50.692246703954076,50.686633208456136,50.681085956810605,50.67560517886248,50.67019103673404,50.664843628141384,50.659562989602954,50.654349099542,50.64920188128553,50.644121205962534,50.63910689530379,50.63415872434508,50.62927642403723,50.62445968376454,50.61970815377389,50.61502144751685,50.610399143907166,50.60584078949537,50.60134590056281,50.59691396513718,50.59254444493135,50.58823677720779,50.58399037657033,50.57980463668509,50.57567893193282,50.57161261899417,50.56760503836991,50.56365551583796,50.55976336384881,50.55592788286147,50.55214836262123,50.54842408338141,50.54475431707038,50.54113832840592,50.537575375958205,50.534064713163175,50.53060558928799,50.52719725034982,50.52383893998981,50.520529900303444,50.51726937262892,50.5140565982949,50.51089081932907,50.50777127912871,50.504697223094894,50.50166789923116,50.4986825587085,50.495740456397336,50.49284085136794,50.489983007360536,50.48716619322602,50.48438968333847,50.48165275798062,50.47895470370318,50.4762948136591,50.47367238791384,50.47108673373228,50.46853716584374,50.466023006685504,50.46354358662596,50.46109824416822,50.458686326134995,50.456307187835435,50.45396019321507,50.45164471498896,50.44936013475961,50.44710584311955,50.44488123973992,50.44268573344524,50.44051874227538,50.43837969353511,50.43626802383197,50.43418317910293,50.43212461463056,50.43009179504898,50.42808419434057,50.426101295823365,50.4241425921303,50.422207585180054,50.42029578614057,50.41840671538526,50.416539902442686,50.414694885939674,50.412871213538786,50.41106844187013,50.40928613645788,50.40752387164205,50.405781230496046,50.40405780473955,50.40235319464796,50.40066700895796,50.39899886477015,50.39734838744845,50.395715210516684,50.39409897555291,50.39249933208131,50.39091593746211,50.38934845677987,50.387796562729754,50.386259935502935,50.38473826267028,50.383231239065466,50.38173856666686,50.380259954479115,50.37879511841367,50.37734378116951,50.37590567211326,50.374480527159285,50.3730680886498,50.371668105235216,50.37028033175457,50.36890452911643,50.36754046418022,50.36618790963821,50.364846643897884,50.363516450965356,50.36219712032929,50.360888446846054,50.35959023062543,50.35830227691767,50.35702439600142,50.35575640307285,50.354498118136014,50.353249365894314,50.35200997564334,50.35077978116496,50.34955862062283,50.34834633645911,50.34714277529296,50.34594778781996,50.34476122871347,50.34358295652722,50.34241283359933,50.34125072595809,50.340096503229155,50.338950038544176,50.33781120845125,50.336679892826524,50.33555597478784,50.33443934060948,50.33332987963887,50.33222748421453,50.33113204958605,50.33004347383492,50.328961657797684,50.32788650498992,50.32681792153248,50.32575581607864,50.32470009974318,50.32365068603286,50.32260749077844,50.321570432067844,50.32053943018139,50.31951440752815,50.31849528858367,50.31748199982937,50.31647446969319,50.31547262849155,50.31447640837313,50.313485743263435,50.312500568811124,50.311520822335424,50.31054644277514,50.30957737063845,50.3086135479546,50.307654918226326,50.306701426383874,50.30575301873981,50.304809642945564,50.30387124794865,50.30293778395141,50.3020092023706,50.301085455798116,50.30016649796307,50.29925228369461,50.29834276888593,50.29743791045936,50.29653766633229,50.295641995384386,50.29475085742517,50.2938642131633,50.29298202417625,50.29210425288106,50.29123086250595,50.29036181706283,50.28949708132047,50.28863662077888,50.28778040164407,50.28692839080377,50.286080555804105,50.285236864826786,50.284397286666824,50.28356179071167,50.28273034692027,50.281902925803045,50.28107949840311,50.28026003627708,50.27944451147733,50.27863289653455,50.27782516444093,50.27702128863393,50.276221242980625,50.27542500176258,50.27463253966123,50.273843831743804,50.27305885344977,50.272277580577715,50.27149998927279,56.545386577481494,64.56720152514447,68.89856207868235,70.12591008573362,69.83525736107832,68.88354220442956,67.64261454647328,66.1934887171191,64.51550745503509,62.60382591333018,60.47478383353409,58.16971178627031,55.78483532677274,53.47205546286778,51.3943412364205,49.65485741212336,48.3100570280306,47.37267744663195,46.81097503557807,46.565532814001955,46.564363673882255,46.73351471977199,47.005285643989204,47.324168573843025,47.64979142184331,47.95650580901065,48.2303988184105,48.465415819025246,48.66010014347983,48.815613554231994,48.93481995507758,49.02191108346711,49.082145127016545,49.121535737851204,49.14649385085673,49.163451441128,49.178491751950965,49.196990378314865,49.22330716757898,49.26057040646116,49.310594125742895,49.37393081682459,49.45003016382214,49.53745785365888,49.63412842169972,49.737526224553946,49.84490073956978,49.95343698979854,50.060400977360054,50.16326067226958,50.259780097107836,50.348084923146324,50.42670095384061,50.49456864909141,50.55103920505339,50.59585604103128,50.62912507903385,50.65127531337373,50.663011496709075,50.66526122923632,50.65911955906218,50.64579456962107,50.62655665871862,50.60269339833272,50.575470562352955,50.54609942146118,50.51571003635287,50.485330407430865,50.455871356948684,50.42811691840106,50.40271981138852,50.38020129160743,50.360954579830434,50.34525104249866,50.333248450096825,50.32500075560277,50.320468949134984,50.31953258840306,50.322001623493854,50.32762817367489,50.33611796568341,50.347141233384384,50.36034295040136,50.37535233056536,50.39179155656685,50.409283708322135,50.427459865052725,50.44596536081127,50.46446518687737,50.48264854735571,50.50023258724933,50.516965317264216,50.532627762716146,50.54703536500743,50.56003866807901,50.571523328743005,50.58140949774996,50.589650626510405,50.59623176047443,50.60116738489459,50.60449889169168,50.60629173856279,50.606632373017355,50.605624994669284,50.6033882283422,50.60005177807884,50.59575312811317,50.590634351433316,50.584839080197085,50.578509685166516,50.57178470376834,50.56479654847675,50.55766951913487,50.550518134778066,50.543445792713555,50.53654375525267,50.52989045773537,50.523551125435226,50.517577681651616,50.51200892482504,50.50687094888321,50.50217777824322,50.49793218695681,50.49412667034157,50.490744537050475,50.48776108981603,50.485144864005505,50.482858894561204,50.480861983805326,50.47910994488122,50.477556798204624,50.47615590112645,50.474860993987235,50.473627148801775,50.47241160988937,50.471174518793966,50.46987951877477,50.46849423693318,50.46699064463747,50.465345299274006,50.46353947246596,50.461559171734194,50.45939506412076,50.45704231154328,50.45450032859848,50.45177247418883,50.44886568872039,50.44579008872428,50.442558530612274,50.43918615490967,50.43568992174574,50.43208814764822,50.428400052813785,50.424645327042725,50.42084372145866,50.41701467201683,50.413176959658365,50.40934841082569,50.405545640932324,50.40178384230455,50.39807661709674,50.39443585474526,50.390871652674626,50.38739227822039,50.38400416908566,50.38071196911092,50.37751859570781,50.37442533498831,50.371431960406795,50.36853687061899,50.36573724224368,50.36302919327882,50.360407953073185,50.35786803496492,50.355403407974556,50.35300766425898,50.350674179393856,50.34839626293728,50.34616729713366,50.34398086202815,50.34183084567641,50.33971153853854,50.33761771153334,50.33554467759767,50.333488336932604,50.3314452064279,50.329412434026096,50.32738779902313,50.32536969949684,50.32335712820991,50.32134963844955,50.31934730134278,50.31735065622607,50.31536065565316,50.31337860659869,50.31140610935902,50.30944499557149,50.307497266671234,50.30556503398417,50.303650461522906,50.30175571240583,50.29988289967298,50.29803404211738,50.29621102559929,50.29441557016228,50.29264920312717,50.290913238206464,50.28920876055754,50.2875366175823,50.28589741518178,50.284291519090445,50.2827190608452,50.28117994788878,50.279673877266916,50.27820035235262,50.2767587020175,50.275348101670396,50.27396759559436,50.27261612003502,50.27129252652401,50.26999560495915,50.26872410600812,50.267476762451906,50.26625230913656,50.26504950125814,50.2638671307604,50.26270404068136,50.26155913733889,50.260431400296916,50.25931989010344,50.25822375383571,50.25714222852799,50.25607464259432,50.255020415387115,50.25397905505871,50.252950154912,50.25193338844086,50.25092850326978,50.249935314206624,50.248953695621736,50.2479835733621,50.247024916401244,50.24607772841307,50.24514203944423,50.24421789784268,50.243305362581346,50.242404496097016,50.24151535774325,50.24063799793623,50.239772453051785,50.238918741111846,50.23807685827989,50.237246776166614,50.236428439931935,50.23562176715305,50.234826647417165,50.23404294258548,50.23327048766656,50.23250909223002,50.231758542286684,50.23101860255831,50.23028901905806,50.229569521903926,50.22885982828914,50.22815964553649,50.22746867416766,50.2267866109251,50.22611315168934,50.225447994241335,50.22479084082735,50.22414140049094,50.2234993911444,50.222864541359066,50.222236591861886,50.22161529673169,50.22100042429598,50.22039175773416,50.219789095399086,50.219192250872794,50.218601052776954,50.21801534436077,50.217434982892804,50.21685983888369,50.21628979516847,50.21572474587756,50.21516459532511,50.214609256842806,50.21405865158549,50.21351270733416,50.21297135731916,50.21243453908405,50.211902193408825,50.211374263307356,50.210850693112256,50.21033142765678,50.20981641156099,50.20930558862702,50.208798901344785,50.20829629050821,50.20779769493951,50.20730305131653,50.20681229409784,50.20632535553703,50.20584216577844,50.205362653024366,50.20488674376351,50.204414363050375,50.20394543482472,50.20347988226076,50.203017628135825,50.202558595208615,50.202102706598104,50.20164988615462,50.20120005881572,50.20075315094013,50.20030909061442,50.19986780792795,50.1994292352122,50.1989933072423,50.198559961399226,50.19812913779205,50.19770077934024,50.197274831817516,50.19685124385869,50.1964299669319,50.196010955279,50.195594165827366,50.195179558077214,50.19476709396703,50.19435673772221,50.193948455690034,50.19354221616529,50.19313798920986,50.19273574647057,50.19233546099754,50.191937107067,50.19154066001098,50.191146096055974,50.190753392172844,50.19036252593962,50.18997347541803,50.1895862190452,50.18920073554049,50.18881700382783,50.18843500297375,50.18805471214015,50.187676110551145,50.18729917747364,50.18692389220982,50.18655023410119,50.1861781825417,50.18580771699967,50.185438817046354,50.185071462390205,50.18470563291472,50.18434130871912,50.18397847016018,50.18361709789442,50.18325717291889,50.18289867661004,50.18254159075998,50.18218589760859,50.18183157987187,50.18147862076554,50.18112700402357,50.18077671391161,50.18042773523541,50.180080053343794,50.17973365412713,50.17938852401061,50.1790446499434,50.17870201938381,50.17836062028107,50.17802044105365,50.177681470565275,50.177343698099136,50.17700711333028,50.17667170629749,50.17633746737386,50.17600438723855,50.17567245684774,50.17534166740752,50.175012010347146,50.174683477294145,50.17435606005108,50.17402975057439,50.173704540955036,50.173380423401525,50.17305739022505,50.17273543382696,50.17241454668837,50.17209472136182,50.1717759504651,50.171458226676904,50.17114154273413,50.17082589143076,50.170511265618174,50.17019765820665,50.1698850621676],"type":"scatter","name":"source2_freq","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[0.0,0.0,-3.975693351829396e-16,-3.975693351829396e-16,3.975693351829396e-16,3.975693351829396e-16,-1.1927080055488188e-15,3.975693351829396e-16,1.987846675914698e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,1.987846675914698e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,-1.1927080055488188e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,-1.1927080055488188e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,-7.553817368475852e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,4.770832022195275e-15,4.373262687012336e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,5.168401357378215e-15,1.1529510720305249e-14,1.789062008323228e-14,1.789062008323228e-14,1.789062008323228e-14,1.789062008323228e-14,1.789062008323228e-14,1.789062008323228e-14,1.789062008323228e-14,1.789062008323228e-14,1.789062008323228e-14,1.789062008323228e-14,1.789062008323228e-14,1.789062008323228e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,1.789062008323228e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.645484559597742e-14,5.685241493116037e-14,5.764755360152624e-14,5.764755360152624e-14,5.764755360152624e-14,5.764755360152624e-14,5.764755360152624e-14,5.9237830942258e-14,5.9237830942258e-14,5.9237830942258e-14,5.9237830942258e-14,5.9237830942258e-14,5.9237830942258e-14,5.9237830942258e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,4.969616689786745e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,4.3335057534940415e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,3.061283880908635e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.684341886080802e-14,5.367186024969685e-14,5.406942958487979e-14,5.446699892006272e-14,5.52621375904286e-14,5.446699892006272e-14,5.446699892006272e-14,5.446699892006272e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.287672157933097e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,4.969616689786745e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,4.3335057534940415e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.605727626079448e-14,5.684341886080802e-14,6.440623229963622e-14],"type":"scatter","name":"source1_θ","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]},{"mode":"lines","y":[-0.9000000000000001,0.7500000000000003,1.3914926731402885e-14,-0.26553707952359407,-0.4232060038539045,-0.6751456781694096,-0.9596102312587917,-1.2619442227169373,-1.5834632420341812,-1.9102213828460315,-2.243455832041882,-2.5808516661218546,-2.919695731815754,-3.2583581391941014,-3.5954364860168817,-3.9297212206062935,-4.260266577921835,-4.586350559655001,-4.907431848235527,-5.223112751917048,-5.5331020249135285,-5.837183121545566,-6.13518770866053,-6.426975316266229,-6.712418207234815,-6.991391458486827,-7.263767342183885,-7.52941308362451,-7.788191191240644,-8.03996137350888,-8.284583569942772,-8.521921456518708,-8.751846203287819,-8.974240166048256,-9.189000393341047,-9.396041817883544,-9.595300034469,-9.786733609124852,-9.970325831768253,-10.14608589380243,-10.314049446691751,-10.474278562199284,-10.626861111892069,-10.77190962077611,-10.909559659258115,-11.039967848089963,-11.163309563405257,-11.279776424334605,-11.3895736534743,-11.49291738889632,-11.59003202486653,-11.681147644642127,-11.766497599726096,-11.84631627765387,-11.920837088771107,-11.99029069296068,-12.05490347635256,-12.114896280764242,-12.170483379977282,-12.221871691554968,-12.269260207407504,-12.312839623009353,-12.352792142792975,-12.38929143795221,-12.422502732801346,-12.45258299611674,-12.479681215191807,-12.503909065168273,-12.525319414161965,-12.544009928263335,-12.560063751676834,-12.573558251204306,-12.584570996507507,-12.593178322774008,-12.599455236780168,-12.603475708524916,-12.605312728177243,-12.605038343549902,-12.602723699879865,-12.59843905985996,-12.592253808001322,-12.584236441893147,-12.574454550900192,-12.562974784275838,-12.549862811375698,-12.535183276579904,-12.518999751203879,-12.501374684319922,-12.482369354146583,-12.462043821472564,-12.440456886389631,-12.417666049378614,-12.393727477533536,-12.368695976451935,-12.342624968083733,-12.315566474625642,-12.287571108371404,-12.25868806727952,-12.228965135901008,-12.198448691219994,-12.167183712900119,-12.135213797396109,-12.102581175379049,-12.069326731931724,-12.035490028992143,-12.001109329557952,-11.966221623206161,-11.930862652531037,-11.895066940154331,-11.858867816013994,-11.822297444688715,-11.785386852565468,-11.748165954700838,-11.710663581269884,-11.672907503530991,-11.634924459266461,-11.596740177684204,-11.55837940378558,-11.519865922220472,-11.481222580660654,-11.442471312730019,-11.403633160533218,-11.364728296824852,-11.325776046859156,-11.286794909957331,-11.247802580823398,-11.208815970634992,-11.169851227928378,-11.13092375929135,-11.092048249870556,-11.053238683695902,-11.014508363817784,-10.97586993224999,-10.93733538970634,-10.898916115117677,-10.860622884912955,-10.82246589204746,-10.78445476476014,-10.74659858504316,-10.70890590680643,-10.671384773721549,-10.634042736731319,-10.596886871211685,-10.559923793775944,-10.52315967871261,-10.486600274050124,-10.45025091724345,-10.414116550480633,-10.378201735607385,-10.342510668670315,-10.307047194081282,-10.271814818405332,-10.236816723776794,-10.202055780948186,-10.167534561978206,-10.133255352563873,-10.099220164024969,-10.065430744945829,-10.031888592482394,-9.998594963341386,-9.965550884437855,-9.932757163238508,-9.90021439779678,-9.867922986486661,-9.835883137441156,-9.804094877701445,-9.772558062082734,-9.741272381762377,-9.710237372596076,-9.679452423166973,-9.648916782573846,-9.6186295679627,-9.588589771807447,-9.558796268944736,-9.529247823367811,-9.49994309478467,-9.470880644944925,-9.442058943741628,-9.413476375091744,-9.385131242601293,-9.357021775019678,-9.329146131488464,-9.301502406589655,-9.274088635198357,-9.246902797145315,-9.21994282169356,-9.19320659183501,-9.166691948411437,-9.14039669406469,-9.114318597021272,-9.088455394716263,-9.062804797260416,-9.03736449075623,-9.012132140466758,-8.987105393841851,-8.962281883406527,-8.937659229515166,-8.913235042976666,-8.889006927553549,-8.864972482340196,-8.841129304023298,-8.817474989029108,-8.79400713555998,-8.770723345526164,-8.747621226373951,-8.724698392814934,-8.70195246845988,-8.679381087360408,-8.656981895461236,-8.634752551966788,-8.612690730624958,-8.590794120931118,-8.569060429255282,-8.547487379895305,-8.526072716058922,-8.50481420077731,-8.483709617752876,-8.462756772143786,-8.441953491287808,-8.421297625367883,-8.400787048021776,-8.380419656898162,-8.360193374161383,-8.340106146947024,-8.320155947770443,-8.300340774890353,-8.28065865262936,-8.261107631653482,-8.241685789212452,-8.22239122934267,-8.203222083034536,-8.184176508365859,-8.165252690603017,-8.146448842271429,-8.127763203196906,-8.10919404051935,-8.090739648680243,-8.072398349385287,-8.05416849154364,-8.0360484511849,-8.018036631355116,-8.000131461993153,-7.982331399788409,-7.96463492802106,-7.94704055638598,-7.929546820801215,-7.912152283202146,-7.8948555313222695,-7.877655178461474,-7.860549863242677,-7.843538249357763,-7.826619025303622,-7.809790904108916,-7.793052623052611,-7.776402943374672,-7.759840649979872,-7.743364551135233,-7.7269734781617805,-7.710666285121155,-7.694441848497822,-7.6782990668771856,-7.662236860620325,-7.646254171535787,-7.630349962548932,-7.614523217369223,-7.598772940156034,-7.583098155183212,-7.567497906502901,-7.55197125760894,-7.536517291100271,-7.521135108344675,-7.505823829142906,-7.490582591393939,-7.475410550761343,-7.460306880341027,-7.445270770330715,-7.430301427701438,-7.415398075871103,-7.400559954380382,-7.385786318571122,-7.371076439267756,-7.356429602461225,-7.341845108996173,-7.327322274261386,-7.312860427883339,-7.298458913423457,-7.284117088078764,-7.269834322386271,-7.255609999931093,-7.241443517058483,-7.227334282589685,-7.21328171754189,-7.199285254852201,-7.18534433910572,-7.171458426267705,-7.157626983420114,-7.1438494885022195,-7.13012543005549,-7.116454306972822,-7.102835628252091,-7.089268912753896,-7.075753688963677,-7.062289494758238,-7.048875877176416,-7.035512392194126,-7.02219860450376,-7.008934087297856,-6.995718422056916,-6.982551198341634,-6.969432013589073,-6.956360472913448,-6.94333618891063,-6.9303587814670555,-6.917427877572645,-6.904543111137688,-6.891704122813933,-6.878910559819445,-6.866162075767495,-6.853458330499361,-6.840798989920879,-6.828183725842774,-6.81561221582489,-6.803084143023907,-6.790599196044844,-6.778157068796055,-6.765757460347706,-6.753400074793967,-6.741084621118432,-6.728810813062786,-6.716578368999191,-6.704387011805511,-6.6922364687440155,-6.680126471343066,-6.668056755281929,-6.656027060278716,-6.644037129981278,-6.632086711860802,-6.6201755571086,-6.608303420535407,-6.596470060473773,-6.584675238682645,-6.572918720255213,-6.56120027352888,-6.5495196699978395,-6.537876684228295,-6.526271093775916,-6.514702679105791,-6.50317122351472,-6.491676513055591,-6.480218336464159,-6.468796485087791,-6.457410752816621,-6.446060936016475,-6.434746833463835,-6.423468246282945,-6.412224977884787,-6.401016833907649,-6.3898436221600265,-6.378705152564643,-6.367601237104867,-6.356531689772293,-6.345496326516263,-6.334494965194856,-6.323527425527466,-6.312593529048946,-6.301693099065029,-6.290825960609385,-6.279991940402085,-6.2691908668091525,-6.258422569803885,-6.247686880928839,-6.236983633259702,-6.226312661369871,-6.215673801296408,-6.205066890507269,-6.194491767869347,-6.183948273617731,-6.173436249325989,-6.16295553787757,-6.152505983437966,-6.142087431427946,-6.131699728497649,-6.121342722501758,-6.111016262475269,-6.100720198610088,-6.090454382232891,-6.080218665782988,-6.070012902791664,-6.059836947862024,-6.049690656649148,-6.039573885841569,-6.029486493142947,-6.019428337254625,-6.009399277858721,-5.999399175601649,-5.989427892078545,-5.9794852898181015,-5.969571232267653,-5.959685583779344,-5.949828209596277,-5.939998975839387,-5.93019774949488,-5.92042439753925,-5.477728580801027,-4.782688509077166,-4.079727925458469,-3.3796648490136154,-2.693606812007813,-2.040863955753962,-1.4324307981317261,-0.8787062727611396,-0.3898252721961799,0.02594975372411417,0.36140647983085633,0.6113272995613804,0.7753952721443189,0.8597756682714556,0.8760306319271425,0.8372966176881066,0.7580978736445211,0.6526752066958089,0.5333144524109436,0.4099163874095977,0.2897904965886436,0.17770760308219105,0.07623711656612674,-0.013741301064502678,-0.09248479600922214,-0.1609150179022379,-0.22028805978468952,-0.2719946816069136,-0.31744421022125696,-0.357986694160358,-0.3948521712060228,-0.42910147311607205,-0.46159084752566937,-0.4929511027723265,-0.5235804643445807,-0.553650861481012,-0.5831270429732963,-0.6117983516044859,-0.639320076230671,-0.6652602898807554,-0.6891466556284584,-0.7105088632290155,-0.7289138832367551,-0.7439929519366023,-0.755460611224519,-0.7631263958888117,-0.7669001589547895,-0.7667916974163884,-0.7629055939368414,-0.7554321346338486,-0.7446353610628841,-0.7308392972214455,-0.7144132586269523,-0.6957570224662991,-0.6752863882169927,-0.6534196096633267,-0.6305650478359891,-0.6071103958041199,-0.5834137212973035,-0.5597964805925049,-0.5365385249739588,-0.5138749961588241,-0.49199494065964366,-0.4710414204050923,-0.45111290652499036,-0.4322657362136717,-0.4145174270872929,-0.39785063884931027,-0.38221757786089366,-0.3675446519263904,-0.35373720275839315,-0.34068417573438015,-0.32826261372031246,-0.31634189126682416,-0.3047876242059917,-0.293465208224644,-0.2822429539561701,-0.27099480081824034,-0.2596026055846929,-0.2479580129945471,-0.23596392503860147,-0.22353559056087915,-0.21060134042363945,-0.19710299481887952,-0.18299597084077462,-0.16824911947515198,-0.15284432218243837,-0.13677587782857056,-0.12004971058848705,-0.10268242899905058,-0.08470026541054464,-0.06613792424220215,-0.047037366434407854,-0.027446556475160838,-0.007418197099114469,0.012991524798849746,0.03372415934517456,0.054720162397329665,0.07591999449158426,0.0972651351683056,0.118699000316573,0.14016775277609445,0.16162099906506225,0.1830123677410456,0.20429996747859744,0.22544672540047886,0.2464206084821516,0.26719473291219936,0.28774736811659435,0.30806184371396217,0.3281263689517011,0.3479337751664699,0.36748119251418754,0.38676967263136686,0.4058037690306165,0.4245910869229973,0.44314181381736145,0.46146824170223794,0.4795842908950167,0.4975050447773313,0.5152463036537936,0.5328241649059017,0.5502546354915253,0.5675532816926331,0.5847349198641587,0.601813350808705,0.6188011393164211,0.6357094393835234,0.6525478646739775,0.6693244029275652,0.6860453722540263,0.7027154165960595,0.7193375370915777,0.7359131556298912,0.7524422065644533,0.7689232523249028,0.7853536185517082,0.8017295443561049,0.8180463433737661,0.8342985714311782,0.8504801968611373,0.866584769784694,0.8826055870065272,0.8985358495410493,0.9143688101843417,0.9300979089641579,0.9457168947257091,0.9612199315354676,0.9766016890005739,0.9918574159993865,1.0069829976944262,1.0219749960440419,1.0368306743426552,1.0515480065931675,1.0661256727526673,1.080563041086814,1.094860139022421,1.1090176140010874,1.1230366859098544,1.1369190927009973,1.150667030814131,1.1642830919816907,1.1777701979394233,1.1911315344770506,1.2043704861590387,1.217490572920469,1.2304953896053057,1.243388549368754,1.2561736317106114,1.2688541357536312,1.2814334392239342,1.2939147634432795,1.3063011444957844,1.318595410600084,1.3308001655905992,1.3429177783016082,1.3549503775481282,1.3668998523139997,1.3787678566877015,1.3905558190315581,1.4022649548303667,1.4138962826399202,1.4254506425435864,1.4369287165252642,1.4483310501809674,1.4596580752117085,1.4709101321738278,1.4820874930024102,1.4931903828678517,1.5042190009777574,1.515173539990029,1.5260542037589346,1.5368612231929986,1.5475948700606441,1.5582554686339243,1.5688434051138542,1.5793591348300162,1.589803187252023,1.6001761688921328,1.610478764212925,1.6207117346852324,1.6308759161661899,1.64097221478679,1.6510016015524758,1.660965105869145,1.6708638082107814,1.6806988321442382,1.6904713359215953,1.700182503841882,1.709833537571769,1.7194256476000864,1.728960044984003,1.7384379335258147,1.7478605024994058,1.757228920024769,1.7665443271680072,1.7758078328236513,1.7850205094159437,1.7941833894365768,1.803297462818571,1.8123636751294656,1.8213829265523647,1.8303560716104699,1.839283919579877,1.848167235526463,1.8570067418958134,1.8658031205802523,1.8745570153839684,1.8832690348061227,1.8919397550622792,1.9005697232664924,1.909159460699843,1.9177094660957341,1.9262202188779383,1.9346921822935197,1.9431258063901007,1.9515215307938345,1.9598797872527167,1.9682010019169667,1.9764855973360598,1.9847339941595732,1.9929466125356,2.0011238732072507,2.009266198313875,2.0173740119085615,2.0254477402085374,2.0334878115989086,2.0414946564131387,2.049468706516484,2.0574103947201228,2.0653201540548194,2.0731984169331525,2.0810456142295473,2.0888621743059037,2.096648522009941,2.1044050776712595,2.112132256118195,2.1198304657361433,2.127500107585481,2.1351415745944515,2.142755250839587,2.1503415109235644,2.157900719457107,2.165433230649836,2.1729393880112493,2.180419524161467,2.187873960749092,2.1953030084712144,2.2027069671894357,2.2100861261338562,2.2174407641865033,2.22477115023409,2.2320775435798605,2.239360194403906,2.246619344260982,2.2538552266052023,2.261068067331392,2.2682580853229073,2.2754254929971123,2.2825704968398663,2.2896932979214095,2.2967940923873624,2.3038730719190443,2.3109304241586215,2.3179663330957374,2.32498097941292,2.3319745407886505,2.3389471921571,2.3458991059252994,2.3528304521484737,2.3597413986657405,2.366632111198121,2.373502753412251,2.3803534869528256,2.3871844714474313,2.393995864487718,2.400787821590927,2.407560496145415,2.4143140393445988,2.4210486001126332,2.4277643250258647,2.4344613582332397,2.4411398413785217,2.4477999135274695,2.454441711102241,2.4610653678246526,2.467671014670709,2.4742587798367888,2.480828788718932,2.4873811639057157,2.4939160251845647,2.5004334895617237,2.5069336712952617,2.5134166819404875,2.5198826304069475,2.526331623025851,2.532763763626762,2.53917915362215,2.5455778920986343,2.5519600759132133,2.5583257997931614,2.564675156438239,2.571008236623741,2.5773251293031656,2.5836259217094346,2.5899106994531302,2.596179546617432,2.6024325458483153,2.608669778439791,2.614891324413343,2.6210972625912574,2.6272876706635455,2.6334626252483435,2.6396222019458593,2.6457664753855434,2.6518955192674127,2.6580094063971123,2.6641082087154806,2.6701919973228763,2.6762608424989733,2.6823148137181283,2.688353979661443,2.6943784082254134,2.7003881665283442,2.706383320914774,2.712363936958122,2.718330079462979,2.7242818124663275,2.7302191992391194,2.7361423022880076,2.742051183357787,2.747945903434551,2.753826522750167,2.759693100788082,2.765545696290393,2.771384367266499,2.777209171003329,2.7830201640769463,2.788817402365473,2.794600941063746,2.8003708346988385,2.8061271371469547,2.811869901650937,2.817599180838787,2.8233150267427005,2.8290174908185044,2.834706623965151,2.8403824765444066,2.8460450984003987,2.8516945388786947,2.857330846844981,2.862954070703228,2.868564258413046,2.8741614575062213,2.879745715102443,2.8853170779238546,2.8908755923088996,2.8964213042248375,2.9019542592794987,2.907474502731718,2.9129820795009027,2.9184770341758335,2.9239594110221474,2.929429253989054,2.934886606715647,2.940331512535839,2.9457640144832467,2.9511841552950093,2.9565919774154503,2.961987522999337,2.9673708339144724,2.972741951744244,2.9781009177902136,2.9834477730741082,2.9887825583400542,2.9941053140566964,2.9994160804195364,3.0047148973528195],"type":"scatter","name":"source2_θ","x":[0.0,0.0001,0.0002,0.0003,0.0004,0.0005,0.0006,0.0007,0.0008,0.0009,0.001,0.0011,0.0012,0.0013,0.0014,0.0015,0.0016,0.0017,0.0018,0.0019,0.002,0.0021,0.0022,0.0023,0.0024,0.0025,0.0026,0.0027,0.0028,0.0029,0.003,0.0031,0.0032,0.0033,0.0034,0.0035,0.0036,0.0037,0.0038,0.0039,0.004,0.0041,0.0042,0.0043,0.0044,0.0045,0.0046,0.0047,0.0048,0.0049,0.005,0.0051,0.0052,0.0053,0.0054,0.0055,0.0056,0.0057,0.0058,0.0059,0.006,0.0061,0.0062,0.0063,0.0064,0.0065,0.0066,0.0067,0.0068,0.0069,0.007,0.0071,0.0072,0.0073,0.0074,0.0075,0.0076,0.0077,0.0078,0.0079,0.008,0.0081,0.0082,0.0083,0.0084,0.0085,0.0086,0.0087,0.0088,0.0089,0.009,0.0091,0.0092,0.0093,0.0094,0.0095,0.0096,0.0097,0.0098,0.0099,0.01,0.0101,0.0102,0.0103,0.0104,0.0105,0.0106,0.0107,0.0108,0.0109,0.011,0.0111,0.0112,0.0113,0.0114,0.0115,0.0116,0.0117,0.0118,0.0119,0.012,0.0121,0.0122,0.0123,0.0124,0.0125,0.0126,0.0127,0.0128,0.0129,0.013,0.0131,0.0132,0.0133,0.0134,0.0135,0.0136,0.0137,0.0138,0.0139,0.014,0.0141,0.0142,0.0143,0.0144,0.0145,0.0146,0.0147,0.0148,0.0149,0.015,0.0151,0.0152,0.0153,0.0154,0.0155,0.0156,0.0157,0.0158,0.0159,0.016,0.0161,0.0162,0.0163,0.0164,0.0165,0.0166,0.0167,0.0168,0.0169,0.017,0.0171,0.0172,0.0173,0.0174,0.0175,0.0176,0.0177,0.0178,0.0179,0.018,0.0181,0.0182,0.0183,0.0184,0.0185,0.0186,0.0187,0.0188,0.0189,0.019,0.0191,0.0192,0.0193,0.0194,0.0195,0.0196,0.0197,0.0198,0.0199,0.02,0.0201,0.0202,0.0203,0.0204,0.0205,0.0206,0.0207,0.0208,0.0209,0.021,0.0211,0.0212,0.0213,0.0214,0.0215,0.0216,0.0217,0.0218,0.0219,0.022,0.0221,0.0222,0.0223,0.0224,0.0225,0.0226,0.0227,0.0228,0.0229,0.023,0.0231,0.0232,0.0233,0.0234,0.0235,0.0236,0.0237,0.0238,0.0239,0.024,0.0241,0.0242,0.0243,0.0244,0.0245,0.0246,0.0247,0.0248,0.0249,0.025,0.0251,0.0252,0.0253,0.0254,0.0255,0.0256,0.0257,0.0258,0.0259,0.026,0.0261,0.0262,0.0263,0.0264,0.0265,0.0266,0.0267,0.0268,0.0269,0.027,0.0271,0.0272,0.0273,0.0274,0.0275,0.0276,0.0277,0.0278,0.0279,0.028,0.0281,0.0282,0.0283,0.0284,0.0285,0.0286,0.0287,0.0288,0.0289,0.029,0.0291,0.0292,0.0293,0.0294,0.0295,0.0296,0.0297,0.0298,0.0299,0.03,0.0301,0.0302,0.0303,0.0304,0.0305,0.0306,0.0307,0.0308,0.0309,0.031,0.0311,0.0312,0.0313,0.0314,0.0315,0.0316,0.0317,0.0318,0.0319,0.032,0.0321,0.0322,0.0323,0.0324,0.0325,0.0326,0.0327,0.0328,0.0329,0.033,0.0331,0.0332,0.0333,0.0334,0.0335,0.0336,0.0337,0.0338,0.0339,0.034,0.0341,0.0342,0.0343,0.0344,0.0345,0.0346,0.0347,0.0348,0.0349,0.035,0.0351,0.0352,0.0353,0.0354,0.0355,0.0356,0.0357,0.0358,0.0359,0.036,0.0361,0.0362,0.0363,0.0364,0.0365,0.0366,0.0367,0.0368,0.0369,0.037,0.0371,0.0372,0.0373,0.0374,0.0375,0.0376,0.0377,0.0378,0.0379,0.038,0.0381,0.0382,0.0383,0.0384,0.0385,0.0386,0.0387,0.0388,0.0389,0.039,0.0391,0.0392,0.0393,0.0394,0.0395,0.0396,0.0397,0.0398,0.0399,0.04,0.0401,0.0402,0.0403,0.0404,0.0405,0.0406,0.0407,0.0408,0.0409,0.041,0.0411,0.0412,0.0413,0.0414,0.0415,0.0416,0.0417,0.0418,0.0419,0.042,0.0421,0.0422,0.0423,0.0424,0.0425,0.0426,0.0427,0.0428,0.0429,0.043,0.0431,0.0432,0.0433,0.0434,0.0435,0.0436,0.0437,0.0438,0.0439,0.044,0.0441,0.0442,0.0443,0.0444,0.0445,0.0446,0.0447,0.0448,0.0449,0.045,0.0451,0.0452,0.0453,0.0454,0.0455,0.0456,0.0457,0.0458,0.0459,0.046,0.0461,0.0462,0.0463,0.0464,0.0465,0.0466,0.0467,0.0468,0.0469,0.047,0.0471,0.0472,0.0473,0.0474,0.0475,0.0476,0.0477,0.0478,0.0479,0.048,0.0481,0.0482,0.0483,0.0484,0.0485,0.0486,0.0487,0.0488,0.0489,0.049,0.0491,0.0492,0.0493,0.0494,0.0495,0.0496,0.0497,0.0498,0.0499,0.05,0.0501,0.0502,0.0503,0.0504,0.0505,0.0506,0.0507,0.0508,0.0509,0.051,0.0511,0.0512,0.0513,0.0514,0.0515,0.0516,0.0517,0.0518,0.0519,0.052,0.0521,0.0522,0.0523,0.0524,0.0525,0.0526,0.0527,0.0528,0.0529,0.053,0.0531,0.0532,0.0533,0.0534,0.0535,0.0536,0.0537,0.0538,0.0539,0.054,0.0541,0.0542,0.0543,0.0544,0.0545,0.0546,0.0547,0.0548,0.0549,0.055,0.0551,0.0552,0.0553,0.0554,0.0555,0.0556,0.0557,0.0558,0.0559,0.056,0.0561,0.0562,0.0563,0.0564,0.0565,0.0566,0.0567,0.0568,0.0569,0.057,0.0571,0.0572,0.0573,0.0574,0.0575,0.0576,0.0577,0.0578,0.0579,0.058,0.0581,0.0582,0.0583,0.0584,0.0585,0.0586,0.0587,0.0588,0.0589,0.059,0.0591,0.0592,0.0593,0.0594,0.0595,0.0596,0.0597,0.0598,0.0599,0.06,0.0601,0.0602,0.0603,0.0604,0.0605,0.0606,0.0607,0.0608,0.0609,0.061,0.0611,0.0612,0.0613,0.0614,0.0615,0.0616,0.0617,0.0618,0.0619,0.062,0.0621,0.0622,0.0623,0.0624,0.0625,0.0626,0.0627,0.0628,0.0629,0.063,0.0631,0.0632,0.0633,0.0634,0.0635,0.0636,0.0637,0.0638,0.0639,0.064,0.0641,0.0642,0.0643,0.0644,0.0645,0.0646,0.0647,0.0648,0.0649,0.065,0.0651,0.0652,0.0653,0.0654,0.0655,0.0656,0.0657,0.0658,0.0659,0.066,0.0661,0.0662,0.0663,0.0664,0.0665,0.0666,0.0667,0.0668,0.0669,0.067,0.0671,0.0672,0.0673,0.0674,0.0675,0.0676,0.0677,0.0678,0.0679,0.068,0.0681,0.0682,0.0683,0.0684,0.0685,0.0686,0.0687,0.0688,0.0689,0.069,0.0691,0.0692,0.0693,0.0694,0.0695,0.0696,0.0697,0.0698,0.0699,0.07,0.0701,0.0702,0.0703,0.0704,0.0705,0.0706,0.0707,0.0708,0.0709,0.071,0.0711,0.0712,0.0713,0.0714,0.0715,0.0716,0.0717,0.0718,0.0719,0.072,0.0721,0.0722,0.0723,0.0724,0.0725,0.0726,0.0727,0.0728,0.0729,0.073,0.0731,0.0732,0.0733,0.0734,0.0735,0.0736,0.0737,0.0738,0.0739,0.074,0.0741,0.0742,0.0743,0.0744,0.0745,0.0746,0.0747,0.0748,0.0749,0.075,0.0751,0.0752,0.0753,0.0754,0.0755,0.0756,0.0757,0.0758,0.0759,0.076,0.0761,0.0762,0.0763,0.0764,0.0765,0.0766,0.0767,0.0768,0.0769,0.077,0.0771,0.0772,0.0773,0.0774,0.0775,0.0776,0.0777,0.0778,0.0779,0.078,0.0781,0.0782,0.0783,0.0784,0.0785,0.0786,0.0787,0.0788,0.0789,0.079,0.0791,0.0792,0.0793,0.0794,0.0795,0.0796,0.0797,0.0798,0.0799,0.08]}],{"xaxis":{"title":{"text":"Time in Seconds"}},"template":{"layout":{"coloraxis":{"colorbar":{"ticks":"","outlinewidth":0}},"xaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"hovermode":"closest","paper_bgcolor":"white","geo":{"showlakes":true,"showland":true,"landcolor":"#E5ECF6","bgcolor":"white","subunitcolor":"white","lakecolor":"white"},"colorscale":{"sequential":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]],"diverging":[[0,"#8e0152"],[0.1,"#c51b7d"],[0.2,"#de77ae"],[0.3,"#f1b6da"],[0.4,"#fde0ef"],[0.5,"#f7f7f7"],[0.6,"#e6f5d0"],[0.7,"#b8e186"],[0.8,"#7fbc41"],[0.9,"#4d9221"],[1,"#276419"]],"sequentialminus":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]},"yaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"shapedefaults":{"line":{"color":"#2a3f5f"}},"hoverlabel":{"align":"left"},"mapbox":{"style":"light"},"polar":{"angularaxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","radialaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"autotypenumbers":"strict","font":{"color":"#2a3f5f"},"ternary":{"baxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","caxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"aaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"annotationdefaults":{"arrowhead":0,"arrowwidth":1,"arrowcolor":"#2a3f5f"},"plot_bgcolor":"#E5ECF6","title":{"x":0.05},"scene":{"xaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"zaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"yaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"}},"colorway":["#636efa","#EF553B","#00cc96","#ab63fa","#FFA15A","#19d3f3","#FF6692","#B6E880","#FF97FF","#FECB52"]},"data":{"barpolar":[{"type":"barpolar","marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"carpet":[{"aaxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"},"type":"carpet","baxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"}}],"scatterpolar":[{"type":"scatterpolar","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"parcoords":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"parcoords"}],"scatter":[{"type":"scatter","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2dcontour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2dcontour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattercarpet":[{"type":"scattercarpet","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"mesh3d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"mesh3d"}],"surface":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"surface","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattermapbox":[{"type":"scattermapbox","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergeo":[{"type":"scattergeo","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram":[{"type":"histogram","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"pie":[{"type":"pie","automargin":true}],"choropleth":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"choropleth"}],"heatmapgl":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmapgl","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"bar":[{"type":"bar","error_y":{"color":"#2a3f5f"},"error_x":{"color":"#2a3f5f"},"marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"heatmap":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmap","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contourcarpet":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contourcarpet"}],"table":[{"type":"table","header":{"line":{"color":"white"},"fill":{"color":"#C8D4E3"}},"cells":{"line":{"color":"white"},"fill":{"color":"#EBF0F8"}}}],"scatter3d":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"scatter3d","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergl":[{"type":"scattergl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2d","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scatterternary":[{"type":"scatterternary","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scatterpolargl":[{"type":"scatterpolargl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}]}},"legend":{"yanchor":"bottom","xanchor":"right","y":1.02,"orientation":"h","x":1},"margin":{"l":100,"b":80,"pad":10,"r":80,"t":100},"plot_bgcolor":"#f1f3f7","yaxis":{"title":{"text":"State values"}},"yaxis2":{"titlefont":{"color":"orange"},"overlaying":"y","title":"Action values","side":"right"},"height":550,"width":800},{"showLink":false,"editable":false,"responsive":true,"staticPlot":false,"scrollZoom":true});
});





</script>
```




_______________________________________________________________________________
### Analysis

- As expected, the inverter in PQ mode behaves as it should.
- Notice that the frequency of the network is not exactly 50 Hz. This is due to the frequency droop mechanism.
- Neither does the voltage magnitude in steady state equal 230 V. This is due to the voltage droop mechanism.
- While the Luenberger observer allows for fewer sensors to control the inverter, it does deteriorate the dynamic behaviour. 

_______________________________________________________________________________
### References

- Q. -C. Zhong and G. Weiss, "Synchronverters: Inverters That Mimic Synchronous Generators," in IEEE Transactions on Industrial Electronics, vol. 58, no. 4, pp. 1259-1267, April 2011, doi: 10.1109/TIE.2010.2048839.
- Q. -C. Zhong, P. -L. Nguyen, Z. Ma and W. Sheng, "Self-Synchronized Synchronverters: Inverters Without a Dedicated Synchronization Unit," in IEEE Transactions on Power Electronics, vol. 29, no. 2, pp. 617-630, Feb. 2014, doi: 10.1109/TPEL.2013.2258684.
