# NodeConstructor Theory

The NodeConstructor has the goal to convert the physical system, here the grid, into a mathematical model. This mathematical model can then be used to simulate the behavior of the network. This notebook addresses the following points:

- ### Introduction to the NodeConstructor
- ### Representation of the physical grid
- ### Building of the state-space matrices


## Introduction

![Illustration of where the NodeConstructor is located](./assets/OverviewJEG.png)

The NodeConstructor is part of the enviroment (see graphic). Its function is to create the mathematical model of the grid to be simulated. The inputs for the NodeConstructor are the specifications of the grid, which will be discussed in more detail subsequently. Within the NodeConstructor, an ordinary differential equation (ODE) system is created based on the underlying physical models and properties of the individual components. Based on this ODE system, the necessary matrices can be extracted.

In the following, we will discuss how the mathematical model of an electrical power grid can be generated using the NodeConstructor.

## Modelling a physical grid

The following graphic shows a simple example grid:


![Illustration of a simple network and its electrical equivalent circuit diagram.](./assets/ExampleGrid.png)

Two sources (wind turbine on the left and PV arrays on the right), as well as a load (exemplary of a household which is composed of various objects) can be seen in this. In addition, the sources are connected to the load via cables. For the sources, we assume that the voltage is provided by a DC link, so we can assume that the voltage can be modeled by a voltage source and a filter. Loads are divided into active loads, which, based on an internal function, can take power from the network, and passive loads, which are described by their resistive, capacitive and/or inductive components and thus dissipate a certain power. Cable modeling relies on the Pi equivalent circuit like depicted in the fiure below.

If you abstract the above figure and insert the respective components according to the description, you get the following single-phase circuit:

![Abstraction of the previously presented example grid.](./assets/ExampleGridCurcuit.png)

With the help of this example, we will now discuss the derivation of a mathematical model and what information the NodeConstructor will later need to generate this model automatically.

### Fundamentals of electrical engineering

If we annotate the above equivalent circuit, we get the following representation:

![Equivalent curcuit with annotations.](./assets/ExampleGridCurcuit_Annotations_Kirchhof.png)

In this, the components are annotated and the currents are noted with the technical flow direction. For the sake of clarity, the voltages are only indicated by green arrows, but have the same name as the corresponding component to which they are applied.

Based on the components within the equivalent circuit diagram, we can now construct the differential equations. Togehter with Kirchhoff's laws

```math
\begin{align}
    \sum_k^n i_k(t) &= 0,\\
    \sum_k^n u_n(t) &= 0
\end{align}
```

and the component equations

```math
\begin{align}
    u(t) &= R i(t),\\
    i(t) &= C\dot u(t),\\
    u(t) &= L\dot i(t).
\end{align}
```

This gives the following equation system:

```math
\begin{align}
    \text{M1:}\;& 0 = u_{in1}  - u_{R11} - u_{L11} - u_{C11} - u_{RC1}, \tag{1}\\
    \text{M2:}\;& 0 = u_{C11} + u_{RC1} - u_{Cb11},\tag{2}\\
    \text{M3:}\;& 0 = u_{Cb11} - u_{Rb1} - u_{Lb1} - u_{Cb21},\tag{3}\\
    \text{M4:}\;& 0 = u_{RL} - u_{LL}, \tag{4}\\
    \text{M5:}\;& 0 = u_{Cb22} + u_{Rb2} + u_{Lb2} - u_{Cb12}, \tag{5}\\
    \text{M6:}\;& 0 = u_{Cb12} + u_{R22} + u_{L22} - u_{C12} - u_{RC2}, \tag{6}\\
    \text{M7:}\;& 0 = u_{RC2} + u_{C12} + u_{L12} + u_{R12} - u_{in2}, \tag{7}\\
    \text{N1:}\;& 0 = i_{11} - i_{RC1} - i_{Cb11} - i_{Rb1}, \tag{8}\\
    \text{N2:}\;& 0 = i_{Rb1} - i_{Cb21} - i_{CL} - i_{RL} - i_{LL} - i_{Cb22} + i_{Rb2}, \tag{9}\\
    \text{N3:}\;& 0 = i_{22} - i_{Rb2} - i_{Cb21}, \tag{10}\\
    \text{N4:}\;& 0 = i_{12} - i_{RC2} - i_{22}. \tag{11}
\end{align}
```



For the sake of clarity, the time dependencies are omitted here. For node 2 the connected capacitances are summed up, because there are several capacitors connected in parallel, e.g.:

```math
\begin{equation}
    C_{SL} = C_{b1} + C_L + C_{b2}
\end{equation}
```

We then call this capacitor $C_{SL}$ through which the current $i_{CSL}$ flows and the voltage $u_{CSL}$ is applied. Now we insert the component equations into the equations (1) to (11) and rearrange them,  so that an ODE system is obtained:

```math
\begin{align}
    \dot i_{11} &= \frac{1}{L_{11}} \left( u_{in1}  - R_{11} i_{11} - R_{RC1} i_{RC1} - u_{C12} \right), \\
    \dot u_{C11} &= \frac{1}{C_{11}R_{C11}} \left(u_{Cb11} - u_{RC1} \right), \\
    \dot i_{Rb1} &= \frac{1}{L_{b1}} \left( u_{Cb11} - i_{Rb1} R_{b1} \right),\\
    \dot i_{LL} &= \frac{1}{L_{L}} u_{CL}\\
    \dot i_{Rb2} &= \frac{1}{L_{b2}} \left( u_{Cb12} - i_{Rb2} R_{b2} - u_{Cb22}\right),\\
    \dot i_{22} &= \frac{1}{L_{22}} \left( u_{C12}  - i_{22} (R_{C2}+R_{22}) - R_{RC1} i_{RC2} - u_{Cb12} \right), \\
    \dot i_{12} &= \frac{1}{L_{12}} \left( u_{in2}  - i_{12} (R_{12} + R_{RC1}) - R_{RC1} i_{C1} - u_{C12} \right), \\
    \dot u_{Cb11} &= \frac{1}{C_{b1}} \left(i_{11} - i_{Rb1} - \frac{1}{R_{C1}} u_{Cb11} + \frac{1}{R_{C1}} u_{C11}\right),\\
    \dot u_{CSL} &= \frac{1}{C_{SL}} \left( i_{Rb1} + i_{Rb2} - \frac{1}{R_{L}} u_{CSL} - i_{LL} \right),\\
    \dot u_{Cb12} &= \frac{1}{C_{b2}} \left(i_{22} - i_{Rb2}\right),\\
    \dot u_{C12} &= \frac{1}{C_{12}} \left(i_{12} - i_{22}\right).\\
\end{align}
```

These differential equations describe the system completely and could be solved for exampled using numerical solvers like Forward Euler or Runge-Kutta. 

### Forming the system matrices from the ODEs

The differential equations found are now to be converted into the common state-space representation:

```math
\begin{align}
    \dot{\vec{x}}(t) &= \mathbf{A} \vec{x}(t) + \mathbf{B} \vec{u}(t)\\
    \vec{y}(t) &= \mathbf{C} \vec{x}(t) + \mathbf{D} \vec{u}(t)
\end{align}
```

Since we are interested in how our grid evolves over time, the first equation is most relevant to our application. $x(t)$ describes in general the state vector, $\dot x(t)$ the changes of this state vector. The change of the system is described on the one hand by the system dynamics, which is expressed by the matrix $\mathbf{A}$ and on the other hand by the input signals $u(t)$ acting on the system combined with the matrix $\mathbf{B}$.

For our example, we can find the following state vector:

```math
\begin{equation}
    \vec{x}(t) = 
    \begin{pmatrix}
        i_{11} & u_{C11} & u_{Cb11}  & i_{12} & u_{C12} & i_{22} & u_{Cb12} & i_{Rb1} & i_{Rb2} & u_{CSL} & i_{LL}
    \end{pmatrix}^\top
\end{equation}
```

and the following input vector:

```math
\begin{equation}
    \vec{u}(t) = 
    \begin{pmatrix}
        u_{in1} & u_{in2}
    \end{pmatrix}^\top.
\end{equation}
```

It should be noted that the order of the states here is first sources, then cables and finally the loads. 

This leads for the investigated example to the following matrices:

```math
\begin{equation}
    \mathbf{A} = 
    \begin{pmatrix}
        -\frac{R_{11}}{L_{11}} & 0 & -\frac{1}{L_{11}} & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0\\
        0 & -\frac{1}{C_{11} R_{C1}} & \frac{1}{C_{11} R_{C1}} & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0\\
        \frac{1}{C_{b1}} & \frac{1}{C_{b1} R_{C1}} & -\frac{1}{C_{b1} R_{C1}} & 0 & 0 & 0 & 0 & -\frac{1}{C_{b1}} & 0 & 0 & 0\\
        0 & 0 & 0 & -\frac{R_{12} + R_{C2}}{L_{12}} & -\frac{1}{L_{12}} & \frac{R_{C2}}{L_{12}} & 0 & 0 & 0 & 0 & 0\\
        0 & 0 & 0 & \frac{1}{C_{12}} & 0 & -\frac{1}{C_{12}} & 0 & 0 & 0 & 0 & 0\\
        0 & 0 & 0 & \frac{R_{C2}}{L_{22}} & \frac{1}{L_{22}} & -\frac{R_{22} + R_{C2}}{L_{22}} & -\frac{1}{L_{22}} &0&0&0&0\\
        0 & 0 & 0 & 0 & 0 & \frac{1}{C_{b2}} & 0 & 0 & -\frac{1}{C_{b2}} & 0 & 0\\
        0 & 0 & \frac{1}{L_{b1}} & 0 & 0 & 0 & 0 & -\frac{R_{b1}}{L_{b1}} & 0 & -\frac{1}{L_{b1}} & 0\\
        0 & 0 & 0 & 0 & 0 & 0 & \frac{1}{L_{b2}} & 0 & -\frac{R_{b2}}{L_{b1}} & -\frac{1}{L_{b2}} & 0\\
        0 & 0 & 0 & 0 & 0 & 0 & 0 & \frac{1}{C_{SL}} & \frac{1}{C_{SL}} & -\frac{1}{R_{L} C_{SL}} & -\frac{1}{C_{SL}}\\
        0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & \frac{1}{L_{L}} & 0\\
    \end{pmatrix}
\end{equation}
```

and

```math
\begin{equation}
    \mathbf{B} = 
    \begin{pmatrix}
        \frac{1}{L_{11}} & 0 \\
        0 & 0 \\
        0 & 0 \\
        0 & \frac{1}{L_{12}} \\
        0 & 0 \\
        0 & 0 \\
        0 & 0 \\
        0 & 0 \\
        0 & 0 \\
        0 & 0 \\
        0 & 0 \\
    \end{pmatrix}
\end{equation}
```

### Solving the ODE system

Now we will deal with the solving of the ODE. The ODE was defined above in continuous time, but since we are in a digital environment, it first has to be discretised and optain the diverence equations in discret time. Assuming ideal digital-to-analogue (D/A) and analogue-to-digital (A/D) conversion with zero-order hold capability, the following system of equations results:

```math
\begin{align}
    \dot{\vec{x}}[k] &= \mathbf{A}_d \vec{x}[k] + \mathbf{B}_d \vec{u}[k]\\
    \vec{y}[k] &= \mathbf{C}_d \vec{x}[k] + \mathbf{D}_d \vec{u}[k]
\end{align}
```

Note that the matrices $\mathbf{A}_d$ and $\mathbf{B}_d$ are different from the continuous-time counterparts. Further below it will be shown how these can be discretised. There are several tools to solve the difference equation. We use the `ControlSystem` package ([Source](https://juliacontrol.github.io/ControlSystems.jl/latest/man/creating_systems/#ss-Creating-State-Space-Systems)), to be more precise the `lsim()` function. 


```julia
using ControlSystemsBase
using LinearAlgebra
using PlotlyJS
```


```@raw html
<div style="padding: 1em; background-color: #f8d6da; border: 1px solid #f5c6cb; font-weight: bold;">
<p>The WebIO Jupyter extension was not detected. See the
<a href="https://juliagizmos.github.io/WebIO.jl/latest/providers/ijulia/" target="_blank">
    WebIO Jupyter integration documentation
</a>
for more information.
</div>
```



Next we define some the values of the components:


```julia
# Source
R = 1.1e-3
L = 70e-6
R_c = 7e-3
C = 250e-6

# Cable
C_b = 1e-4/2
L_b = 1e-4
R_b = 1e-3

# Load
R_l = 100
C_l = 1e-2
L_l = 1e-2;
```

Now we construct the matrices:


```julia
A = zeros((11,11))
A[1,1] = -R/L
A[1,3] = -1/L
A[2,2] = -1/(C*R_c)
A[2,3] = 1/(C*R_c)
A[3,1] = 1/C_b
A[3,2] = 1/(R_c*C_b)
A[3,3] = -1/(R_c*C_b)
A[3,8] = -1/C_b
A[4,4] = -(R+R_c)/L
A[4,5] = -1/L
A[4,6] = R_c/L
A[5,4] = 1/C
A[5,6] = -1/C
A[6,4] = R_c/L
A[6,5] = 1/L
A[6,6] = -(R+R_c)/L
A[6,7] = -1/L
A[7,6] = 1/C_b
A[7,9] = -1/C_b
A[8,3] = 1/L_b
A[8,8] = -R_b/L_b
A[8,10] = -1/L_b
A[9,7] = 1/L_b
A[9,9] = -R_b/L_b
A[9,10] = -1/L_b

C_SL = C_b + C_l + C_b

A[10,8] = 1/C_SL
A[10,9] = 1/C_SL
A[10,10] = -1/(R_l*C_SL)
A[10,11] = -1/C_SL
A[11,10] = 1/L_l;
```


```julia
B = zeros((11,2))

B[1,1] = 1/L
B[4,2] = 1/L;
```


```julia
C = Diagonal(ones(11));
```


```julia
D = 0;
```

We convert the matrices into the discrete time domain and create a discrete StateSpace object with the help of `ControlSystems`. This object can then represent the dynamics of the system for a given time interval using the function `lsim()`.


```julia
ts = 1e-5
Ad = exp(A*ts)
Bd = A \ (Ad - C) * B
sys_d = StateSpace(Ad, Bd, C, D, ts);
```


```julia
ns = length(A[1,:]) # get num of states
ni = length(B[1,:]) # get num of inputs
t = collect(0:ts:0.1)
x0 = [0.0 for i = 1:ns]
u = [230.0 for i = 1:length(t)]
uu = [u for i = 1:ni ]
uuu = mapreduce(permutedims, vcat, uu);
```

To use `lsim()` you need defined initial states `x0`, a time vector `t` and a input signal `u`. In our case we apply a jump to 230 V to the system.


```julia
xout, _, _, _ = lsim(sys_d,uuu,t,x0=x0);
```


```julia
layout = Layout(xaxis_title="Time in µs", yaxis_title="v_C / V")
p = plot(t, xout[5,:], layout)
```




```@raw html
<div
id = C3JdXRw64DSz > </div>
<script>
gd = 'C3JdXRw64DSz'
require(['plotly'], function(plotly) {
























plotly.newPlot(gd,[{"y":[0.0,0.6560461320016365,2.6148431241647643,5.851447601225117,10.326917551515393,15.98931137419965,22.77498179069205,30.610107928270907,39.41239925965242,49.09289882757174,59.557810602375696,70.71027702590231,82.45203770207716,94.68490852692551,107.31203185376339,120.238861955366,133.37386534464272,146.62893163640067,159.91950672234526,173.16447524102,186.2858328541237,199.20819998131816,211.85823681560396,224.1640242168666,236.0544762184796,247.4588473295783,258.30639171743246,268.5262220410889,278.04740367961864,286.7993059996036,294.7122168899681,301.7182108806466,307.75224560410555,312.7534469865453,316.6665311381229,319.44330111616057,321.0441500924229,321.4394993307555,320.61109995637645,318.55313176116414,315.27304003589654,310.79206226400225,305.1454099035854,298.3820857437462,290.5643336684658,281.7667342606626,272.07497568136864,261.5843438463286,250.39798835042825,238.62503022742115,226.37858399194513,213.7737691768099,200.92578562177223,187.94812215454425,174.95096028431064,162.03982353241446,149.31450963738885,136.86832779818985,124.78764715200079,113.15174665715706,102.03294130379186,91.4969458975796,81.60342626355705,72.40667918384527,63.95637714941234,56.298312331353934,49.47507613305216,43.52661615105447,38.49062104472122,34.4026952190827,31.296298744773182,29.20244284517426,28.149146773103837,28.160677146057278,29.256604991776626,31.45072811249692,34.74991624036397,39.15294329299301,44.64937447073429,51.21857576844843,58.82890969603595,67.4371737946451,76.98832826039411,87.4155461666343,98.64060506501824,110.5746229001004,123.11912500700171,136.167413309254,149.60619450062018,163.31741171856186,177.1802146295149,191.0729964390961,204.8754234405577,218.4703834669718,231.74578397212977,244.5961382004814,256.92388861154257,268.64042983982154,279.66680831439606,289.9340914617747,299.3834153529263,307.96573490687757,315.6413145437824,322.37900878371295,328.1553911234071,332.95379514819894,336.7633339762399,339.5779626963836,341.395643547216,342.217665472743,342.0481588278651,340.89383298551854,338.7639501214279,335.6705333015012,331.6287919852118,326.65773399177846,320.7809206098926,314.02731153125404,306.4321391932758,298.0377483136636,288.89433610223233,279.060531871685,268.60376138138696,257.6003509008902,246.13533817736086,234.30197160323542,222.2008941762934,209.93902452991003,197.62816257122302,185.38336130528074,173.32111851269858,161.55745145667015,150.2059242259545,139.37570033699458,129.16969266113063,119.68287863788926,111.00084129243406,103.19858617718371,96.33967153981304,90.47567444248178,85.64599997043611,81.87802487321404,79.18755178871615,77.5795363862844,77.04903803285279,77.58233552906796,79.15814353182209,81.748862763251,85.32179811211391,89.84028318527324,95.26465750822653,101.55305298136481,108.66195880588859,116.54654821909597,125.1607652638824,134.45718466433485,144.38667189448887,154.89788296217947,165.93665362052516,177.44533512132108,189.3621378406498,201.62054490901886,214.14885531834403,226.86990999092558,239.70104529162822,252.55430691342391,265.3369435724852,277.9521852191376,290.3002952779462,302.2798715720024,313.7893568445094,324.72870788899036,335.0011628664591,344.51503992412205,353.185497081677,360.9361836917617,367.7007176028691,373.4239292674871,378.06282408571707,381.58722674343363,383.98008555707855,385.23743014530703,385.3679913278831,384.3925072083448,382.3427531614039,379.2603452196282,375.19537554232716,370.20494479405295,364.3516590614583,357.7021582709015,350.3257389916219,342.29312725467497,333.67544698545095,324.5434173807992,314.96679871916007,305.01409141290713,294.75247837734025,284.2479867822451,273.56583271124197,262.7709018438614,251.92831154427924,241.10399509548625,230.3652475013936,219.78117435895933,209.4229906653935,199.36412478603177,189.68009372249537,180.4481287094039,171.7465443479362,163.65385920425746,156.24769027955557,149.6034572274303,143.7929439371721,138.88277449137053,134.93286703921072,131.994932447634,130.11108451367053,129.31262503243136,129.61906028402527,131.03739586561832,133.5617447399523,137.17326951271033,141.84046499577317,147.5197718260482,154.1564970714899,161.68600412461018,170.0351224533314,179.12371854413914,188.86636309748184,199.1740265301908,209.955735243951,221.12012489476896,232.57683383303043,244.23768960046579,256.01765334867054,267.83550064694225,279.6142316553723,291.28121828757475,302.76811000631585,314.01053255011794,324.94762451882946,335.5214647985714,345.67644886578273,355.3586738192774,364.51539046311933,373.0945759926539,381.0446730832204,388.3145308651097,394.85357094341344,400.61218794750044,405.54237980298046,409.5985887728256,412.7387210699791,414.9253012103198,416.1267078769456,416.3184314120019,415.484289504674,413.617537404733,410.7218120882203,406.8118560888473,401.91397587190545,396.0662012027906,389.3181253593061,381.73042056997593,373.37403798442676,364.32911603064275,354.6836344410789,344.53186284709125,333.97266205490865,323.1077024574949,312.0396671856172,300.87050740461325,289.6998136408595,278.6233603663544,267.73187163593053,257.1100438576059,246.83584839795262,236.9801223814837,227.6064414845267,218.77125450749517,210.52424676190222,202.90888848835735,195.96311618473533,189.7200892934941,184.20896244378383,179.45561446990962,175.48327966250295,172.3130339132289,169.96409819279825,168.4539336251629,167.79811565067345,168.00998869677767,169.10011664953703,171.07555749897,173.93900211432543,177.6878265691668,182.31311426957703,187.79870796923578,194.12035236784627,201.24498533878003,209.1302300437092,217.72413155709287,226.965170581941,236.78257395707635,247.0969275987609,257.8210830117766,268.8613342967294,280.11882940634325,291.4911679452749,302.8741286580179,314.1634633804261,325.2566909705233,336.05482475035234,346.4639702794325,356.3967366690132,365.77341380129457,374.522879262762,382.58321194145356,389.902003393275,396.43637252021244,402.15270307845464,407.026136328075,411.03986209269334,414.18426005976994,416.4559488889305,417.8568033250111,418.3929989158981,418.0741401636679,416.912521208443,414.9225588300028,412.1204261598084,408.5239026412576,404.152442150034,399.0274475201409,393.1727267464517,386.6150945458064,379.38507337650685,371.5176409548355,363.0529671480555,354.0370820941839,344.5224195625802,334.5681848207197,324.24050434341757,313.612325174664,302.7630440858883,291.7778602264312,280.74685902627306,269.7638489540046,258.9249856318096,248.32722908961333,238.06668901210801,228.23691921934994,218.9272259877117,210.22105498384877,202.19451853690956,194.91511886722705,188.4407140356304,182.81876223102714,178.08586714819575,174.2676332843878,171.37882571779258,169.42381505504477,168.39727545890366,168.2850926423729,169.06543000059483,170.70989508386998,173.18474568685141,176.45207507371072,180.47091924416907,185.19823547397027,190.58971028538514,196.60036603351736,203.18494783810468,210.29808597578958,217.89424235949772,225.9274626496001,234.35096717865562,243.11662360920894,252.17435156493352,261.4715139901305,270.95235146045036,280.5575140089207,290.2237403312116,299.8837267382722,309.4662183262443,318.8963430528684,328.0961963658609,336.9856704083179,345.48350834635824,353.50855173558887,360.9811377322972,367.82459395052985,373.9667723458449,379.3415600144535,383.8903044241438,387.56309336982633,390.31983573686904,392.1310976679499,392.97865953080094,392.85577162300496,391.7671001840791,389.72836931619446,386.76571812489294,382.91480507948586,378.21970261294155,372.73163377421076,366.5076088700842,359.6090231864554,352.10027691315906,344.04747532790066,335.51726130033535,326.575823586309,317.2881136584673,307.71729153683845,297.9244078954735,287.9683163282243,277.9057967665489,267.7918593348913,257.68018802085817,247.62367594822462,237.67499916762023,227.88717397560953,218.31404393380268,209.01064691981443,200.0334194772759,191.44020507232136,183.2900441118627,175.64273613678012,168.55817780397814,162.09549341402018,156.31198712734982,151.26195698278468,146.9954198016867,143.55680254909774,140.98365937791812,139.3054742006341,138.5426061704862,138.70543002653125,139.79371513578434,141.79627666209248,144.69092013618769,148.44468741634708,153.01439829487612,158.34746852606756,164.38297251668274,171.05290797115586,178.28361097853445,185.9972638126413,194.11343440202438,202.5505861718162,211.22749976626176,220.06455386997072,228.98482064891945,237.91494178571364,246.7857631336261,255.53271902476308,264.095970555349,272.42031503904485,280.4548955948748,288.1527499063249,295.47024503023164,302.36645034431325,308.8025030360837,314.7410198419103,320.14560509028405,324.9804986910838,329.2103988909964,332.8004838592507,335.7166440540414,337.92592449633185,339.3971632292487,340.1018000543709,340.01481876243014,339.11577709792647,337.38987209807095,334.82898458467264,331.4326456779619,327.2088703039919,322.17480768445427,316.3571664790307,309.7923822057768,302.52650628533723,294.61480893839763,286.1211015500229,277.1167973103812,267.6797412693739,257.89285176852064,247.8426239808646,237.6175525495595,227.30653374950901,216.99730803225083,206.77500123625867,196.72081729345712,186.91092722737568,177.41558904201403,168.29852127749095,159.6165401765726,151.4194572338044,143.75022107073193,136.64527576125295,130.13509753059148,124.24486369344862,118.99520219164148,114.4029674141734,110.48198825925367,107.24373759891114,104.6978782494364,102.85264890797983,101.71506383374691,101.29091177530354,101.58455214419774,102.59851903991157,104.33295576758066,106.784913313871,109.94755527799806,113.8093185088068,118.35308280979322,123.55540431573031,129.38586544185577,135.80658973887898,142.77196278444123,150.22859077268404,158.11551721595507,166.36470572772274,174.9017838579444,183.6470300753314,192.51657390327216,201.4237685449846,210.28068663266853,218.99968345655893,227.4949684984913,235.68412549374526,243.48952360514744,250.83956749159645,257.669741819081,263.92341569273236,269.55238406860303,274.5171358340591,278.7868512729364,282.33914438376195,285.1595773401146,287.2409846654777,288.5826529163347,289.18940741139687,289.07066052686423,288.23947615787375,286.71170213740766,284.5052168699214,281.6393284843468,278.13435487271016,274.0114015925664,269.29234238507877,263.99999465014724,258.1584702883651,251.79367150342435,244.933892034994,237.61047734833085,229.85849291494716,221.71734811579216,213.2313245747754,204.44996182449245,195.42825990771053,186.2266674801935,176.9108347363549,167.55112246800317,158.2218711618689,149.00044658510134,139.966090143026,131.1986128001585,122.77697998265356,114.77784117457391,107.27406154948139,100.33331374939995,94.01678579037605,88.3780561437991,83.46217956678036,79.30501761673669,75.93283648037071,73.36218235697532,71.60003179752006,70.64420177369931,70.48399248232047,71.1010255814866,72.47023223018994,74.56093938193652,77.3379995594126,80.76290896318963,84.79486124617472,89.39168947109236,94.51065637375441,100.10906266738526,106.14465422590926,112.57582098313672,119.36159264085674,126.46144814058196,133.83496669184237,141.4413573903425,149.23891161297433,157.18442706531175,165.23265433292735,173.33581594611033,181.44324435243365,189.50117899347717,197.45275422736438,205.23819957306372,212.7952622165976,220.05984952145565,226.96687706887926,233.45129616299832,239.44926438348585,244.8994142025036,249.74416835825846,253.93104693266756,257.41391011914817,260.1540825470322,262.12130966207195,263.29450381511293,263.6622470192675,263.2230283237866,261.9852058570791,259.96669619108036,257.1944061282426,253.70343367313237,249.53607522699767,244.74068441251973,239.3704339740739,233.48203560490222,227.13447316285206,220.3878025328006,213.30206650516732,205.93636573422742,198.34811751087238,190.59252323020516,182.72225363147072,174.7873487609539,166.8353177975778,158.9114130077394,151.0590427313873,143.3202809358348,135.73642589146365,128.34855818225,121.19804868231812,114.32697028023914,107.77837284426083,101.59638888788304,95.82614719126535,90.5134827386239,85.70444315487828,81.44460373827923,77.77821455656373,74.74721329757826,72.39014609646524,70.74104494407374,69.82831417124774,69.67367968741462,70.29125305728051,71.68675819950441,73.85696170039309,76.78933879846865,80.46199646879779,84.84386326398169,89.89514325610249,95.56801920319815,101.8075785625387,108.55292578095857,115.7384359311807,123.29509865829651,131.15189786049865,139.23717172731708,147.4798997289264,155.8108677794442,164.16366983538862,172.4755132627085,180.68780593214956,188.7465146235143,196.6022963245675,204.21041577146184,211.53047348061602,218.52597799956575,225.16380366520838,231.41358041268253,237.24706485950261,242.6375418660385,247.55930305609786,251.98724352315213,255.89661043517782,259.2629278891167,262.06211165660466,264.27077597849916,265.866722919987,266.8295936121027,267.1416505764686,266.78865179946797,265.7607707474034,264.0535124450298,261.6685742995178,258.6146016258451,254.9077917609537,250.5723070465602,245.64046548910466,240.15268813282466,234.1571935821945,227.70944209526672,220.87134362133847,213.71025545790266,206.29780526986443,198.70858352951868,191.01875557103028,183.3046470938601,175.64135791123923,168.1014569772112,160.75380733523826,153.66256284552992,146.8863697237504,140.4777955224953,134.4829967518008,128.94162446139694,123.88695541225262,119.34622555409807,115.34113296243972,111.88846966456151,109.00083629309492,106.68739052392506,104.95457992625163,103.8068121807614,103.24702047175803,103.27708896314135,103.89811224068471,105.11047296081155,106.9137331335023,109.30634588557747,112.28520559037818,115.84506432126943,119.97785115129747,124.67193742121849,129.91139537806544,135.67529931367682,141.93711740872942,148.6642389537527,155.81767565251945,163.35196762109007,171.21531489536878,179.34994425889508,187.69270957157687,196.17591212336936,204.7283164604919,213.276327211684,221.74528419427955,230.06082693360915,238.1502760042273,245.94397749155246,253.37655843559656,260.3880452735436,266.92480382946775,272.94026796874215,278.3954341975803,283.2591107184957,287.5079211695358,291.1260748740756,294.10492631131626,296.4423561266159,298.14201384434995,299.2124681264041,299.66631365284763,299.51928433261435,298.789420562444,297.4963337604863,295.6606046562712,293.3033431882107,290.4459278086166,287.10993106324264,283.31722708176386,279.0902656861634,274.4524877777124,269.428848034758,264.04640420224626,258.3349277398104,252.3274885602067,246.06096714500066,239.5764504443052,232.9194734915995,226.14007630615768,219.2926550158613,212.4355967230679,205.63069890179372,198.94238546590867,192.43674249055772,186.1804063339958,180.2393450817672,174.6775803896739,169.5559006114384,164.9306173586292,160.8524162841618,157.3653489726222,154.5060065595655,152.3029074088294,150.77612127816093,149.93714140850253,149.7890044469362,150.326646650144,151.5374740027834,153.40211427700677,155.89531114540688,158.98691464067971,162.6429188220343,166.826496626609,171.498983586178,176.6207662725982,182.15203776774695,188.05339078609094,194.28622886289253,200.81298673580093,207.59716212206922,214.60317194021033,221.79605606911227,229.14106044839622,236.60313824116596,244.14641253949063,251.7336464434153,259.3257661626608,266.88148008669896,274.35703168674496,281.7061169208771,288.8799878906238,295.8277543147417,302.4968834807914,308.83388828296285,314.7851823389374,320.2980715603955,325.32184344020055,329.808909144861,333.71594959312426,337.00501527173594,339.6445306695066,341.6101578493463,342.88547964259584,343.46247093734024,343.3417361337402,342.53250156215904,341.05236295334157,338.9267993341049,336.1884754179222,332.87636411957095,329.0347287585784,324.7120104245536,319.9596695617353,314.83103191470803,309.38018751716777,303.6609874883311,297.72617724240456,291.6266966471141,285.4111681181816,279.1255831145178,272.81318656005976,266.5145479392472,260.26779776090933,254.10899928710708,248.07261833945367,242.19204899473414,236.50015032305282,231.02974914029113,225.81406604671412,220.88702667870828,216.2834268604517,212.03892884564937,208.1898756334562,204.7729209066118,201.82448291238737,199.38004101589436,197.4733031473048,196.13528043422357,195.3933115269189,195.27008315342073,195.78269506415964,196.94181664824862,198.75097916289099,201.2060418796769,204.29486280271243,207.9971953446751,212.2848219336457,217.1219245012705,222.4656807384973,228.26706446793668,234.4718190133224,241.02156453223407,247.85499432268085,254.90911141964767,262.1204555487286,269.42627175765296,276.7655757298009,284.08007669750157,291.3149267021657,298.41927427930926,305.34661098819254,312.05491001559244,318.5065667955795,324.66816164309597,330.5100732754463,336.0059793324695,341.1322852302345,345.8675256281517,350.19178330933516,354.0861683505603,357.53239620153136,360.512496936315,363.00867982591126,365.0033679453748,366.47940727622813,367.42044424237696,367.8114553865309,367.63940349990355,366.8939864622507,365.56843876321693,363.66034149762623,361.17239478424324,358.1131071545536,354.4973594721988,350.34680622500946,345.6900843114339,340.56280834809206,335.0073415965514,329.0723423253899,322.81209623436035,316.2856559121521,309.55581764316594,302.6879737392619,295.74888454420824,288.80541803451547,281.9233063277076,275.1659673393411,268.59343636241016,262.26144666082274,256.22069057079193,250.51628348771527,245.18744295699938,240.26738441417277,235.78342448688605,231.75727272993578,228.2054837329504,225.14003417821715,222.56898400946568,220.49717767463918,218.92694058324676,217.8587275033416,217.29168351966507,217.2240841669519,217.65362911499065,218.57757289643934,219.99268614582243,221.89505112505265,224.27970539699749,227.14015683600167,230.46780123522885,234.2512801540325,238.47582100217355,243.12260344638844,248.1681959325848,253.5841034530575,259.3364627865036,265.3859145508265,271.6876728907105,278.1918039153828,284.8437136111137,291.58483541823165,298.3534975320588,305.085940788683,311.7174502090412,318.18355730416187,324.4212664018525,330.3702567411373,335.9740129747378,341.18083997604606,345.94472329687045,350.2260039834394,353.9918453494874,357.2164792664968,359.8812300472819,361.97432452708745,363.4905069437455,364.43048617147156,364.8002503134474,364.6102892277229,363.87476896954087,362.6107032148861,360.8371654408538,358.5745820584895,355.8441410229944,352.66734299368096,349.0657132836364,345.0606831001306,340.67363845149106,335.92612512391037,330.8401888429354,325.43882162292425,319.7464788046229,313.7896267320964,307.5972786681856,301.2014765191836,294.6376782387999,287.9450152903766,281.1663910300614,274.3483989936006,267.5410494007822,260.79730224567226,254.17241559071405,247.72312759095547,241.5066998210665,235.57985718708457,229.997665671187,224.81239306623556,220.07239949952765,215.82110383229943,212.09606898585793,208.92824404043597,206.34139384537053,204.35173823419115,202.96781320047612,202.19055606105957,202.01360624366103,202.4238034199611,203.40185577080712,204.92314367388545,206.95861842643444,209.4757520415124,212.43949286088602,215.8131827664818,219.55939507934013,223.64065762566526,228.02003263096313,232.66153368955844,237.53036958953024,242.5930147453611,247.8171158735463,253.17125381711995,258.62458759385237,264.14641437898075,269.7056838873886,275.2705082424207,280.8077087719139,286.28243923914886,291.65792090041475,296.8953186998554,301.95378018783896,306.7906497894095,311.3618613276693,315.6225017366689,319.5275292063192,323.0326201004497,326.0951113497213,328.674999049166,330.7359500041305,332.2462811831503,333.17986255172764,333.5169015570143,333.2445724709595,332.3574606278769,330.85779996244696,328.7554917365208,326.067902450646,322.81944914961883,319.04099012609146,314.76904790006694,310.04489885116396,304.91356962304246,299.4227841154581,293.6219063460659,287.5609236346488,281.289511489867,274.85621643255854,268.30778604740294,261.68866718668636,255.04068389982112,248.40289682415454,241.81163596471416,235.3006895282142,228.90162324536138,222.6441978454389,216.55684739079578,210.6671782926112,205.00244816012486,199.58998521344424,194.45751272665615,189.63334866124885,185.14645799481806,181.02634385487534,177.30277297428177,174.00534069323447,171.16289022681408,168.802809692028,166.95023797689913,165.6272165314277,164.85182824450874,164.6373665206003,164.9915773753066,165.91601483771473,167.4055453039509,169.44802996609613,172.02420637652793,175.10778101380018,178.66573486520528,182.6588340403125,187.04232779396182,191.76680756219605,196.77919315126445,202.0238064504701,207.44348925826088,212.98072020796502,218.57868643190133,224.18226846989393,229.7389018550696,235.19928553399365,240.51791544453215,245.65343075385613,250.56876996954412,255.23114387110104,259.6118414597707,263.6858934102803,267.4316243998565,270.8301308323592,273.86472360616585,276.52037653718656,278.7832198010472,280.64011437526773,282.0783381296874,283.0854072202553,283.64904816350554,283.7573268497006,283.39893127883266,282.5635954808283,281.2426434066386,279.42962401337155,277.1210027186623,274.3168701917982,271.02162730974936,267.24460515461465,263.0005811696801,258.31015691387785,253.19996903930323,247.70271284630343,241.8569666471609,235.70681473405517,229.30127650511133,222.69355874893964,215.94015673421936,209.09983714927648,202.23254170487684,195.39825405409835,188.655874398973,182.0621456579491,175.6706723894243,169.53106894630636,163.68826682228988,158.18200318262106,153.04650356582084,148.31036216656625,143.99661346074103,140.12297971123763,136.7022705736218,133.74290403214206,131.2495125960314,129.22359534168683,127.66417516311903,126.56842154449323,125.93220223391322,125.75053220528756,126.01789497194456,126.72841930119962,127.87590324469406,129.45368667074942,131.45438266723744,133.8694867868497,136.68889067243336,139.90033272450245,143.48882282855917,147.43608051214628,151.72002612179455,156.31436268162244,161.18828211469494,166.3063236796806,171.62840510034533,177.11003832681715,182.7027326094279,188.35457807038065,194.01099372034201,199.61561536803814,205.1112915544524,210.44114989121726,215.54969229580837,220.3838757976746,224.89413593533806,229.035312258083,232.7674399588243,236.056377963991,238.87425156610126,241.19969650399523,243.01790081881686,244.32045035455872,245.10499293886204,245.37474460179533,245.13786824153817,244.4067605674706,243.19728667015693,241.52800301111463,239.41940892893447,236.89326397015515,233.9720036305431,230.67827969328007,227.03464362049613,223.06338280813253,218.78651041304227,214.22590039026076,209.40355081976136,204.3419510102317,199.0645216453872,193.59609271633838,187.9633814019093,182.1954315554825,176.32397806025162,170.3837029420255,164.41235559068625,158.4507164521908,152.54239174518278,146.73343570014174,141.07180604509728,135.60666748571154,130.3875662746722,125.46350619661226,120.88196202899678,116.68787047338803,112.9226404758003,109.62322466477917,106.82129133680152,104.54253212580713,102.80613442605338,101.6244401061756,101.00280344437769,100.93965197277187,101.42674451823159,102.44961165058754,103.9881554619309,106.01737852674032,108.50820639123322,111.42836428893222,114.74326715822465,118.41688252080037,122.41252833153003,126.69357238700384,131.2240060378367,135.9688724579494,140.89453818154539,145.96880557588597,151.16087290059474,156.4411571372461,161.7810024062493,167.15230311958962,172.52707571513753,177.87701563489586,183.17307699903762,188.38511115021535,193.4815969676805,198.42949075277681,203.19421683689296,207.73981221224858,212.02922985042713,216.02479641021444,219.68881121932841,222.9842652108374,225.87565133935087,228.32983227462117,230.31692717663262,231.81117731162578,232.7917502860873,233.24344476112154,233.1572615604859,232.5308128970438,231.36854871781364,229.68178753426287,227.48854813237242,224.81318778237045,221.68586151696064,218.141825260119,214.2206126455205,209.9651209043705,205.42064494611637,200.6339005153053,195.65207700030575,190.52195811898568,185.28914444327856,179.99740577869753,174.6881841032691,169.40025948046758,164.1695825300585,159.02926813393012,154.00973653841817,149.138980336022,144.4429293684827,139.94588072474755,135.67095796281168,131.6405626188132,127.87678203064749,124.40172044038827,121.23772508972745,118.40748532988962,115.93399029460707,113.84033903026219,112.14940569676294,110.88337108205694,110.06313975310469,109.70766926615379,109.83324360020468,110.45272705180598,111.57483701627208,113.2034742584644,115.33714742900197,117.96852479878697,121.08414065696826,124.66427682714207,128.68303166241333,133.10858009836332,137.90361932636443,143.0259858627335,148.42942168462272,154.0644600984156,159.87939646049006,165.82130506722638,171.83706166396402,177.87433118337103,183.88248249902102,189.81339605187262,195.6221359594374,201.2674653464038,206.71219176705648,211.92333829837938,216.87214471360622,221.5339116393289,225.88770831405657,229.9159711031704,233.60402494805214,236.9395631794857,239.9121224444286,242.51258881999797,244.73276856299543,246.56505251396476,248.0021971827401,249.03723831070278,249.66354462137085,249.87501096778897,249.6663816164679,249.03368642522267,247.97476560221185,246.48985295367083,244.582183348175,242.25858777150506,239.53003894809962,236.41211208970233,232.9253288220878,229.09535755956938,224.95305027653131,220.5343034209676,215.8797392238082,211.03421243900362,206.04615614777538,200.9667882303056,195.8492070362251,190.74741031252918,185.71527528786504,180.80553976713074,176.06882405971695,171.5527315514386,167.30106184016208,163.35316479159138,159.74345692453858,156.50111356557545,153.64994163502567,151.20842918437938,149.18995935151622,147.60316867065004,146.45242306587238,145.73837970825784,145.45859948730288,145.60817330801237,146.18032584588528,147.16696274117388,148.55913135783118,150.3473709443628,152.52193500432324,155.07287653885427,157.98999514203442,161.262653267745,164.87947689971236,168.82796292272917,173.0940213361773,177.66148475899774,182.51162021735598,187.62267884470202,192.9695178212713,198.523325701018,204.25147738196912,210.11753862458934,216.08143254232115,222.0997722688298,228.12635547649626,234.11280802578847,240.00935620398175,245.7657001732871,251.33195575071858,256.65962777316827,261.70257626457425,266.4179365258113,270.76695611911816,274.7157154233499,278.2357038037016,281.30423019313287,283.9046546741265,286.0264360767612,287.6649992380096,288.8214339535905,289.50204536691496,289.7177821848978,289.48357434929693,288.81761536051357,287.74062617185206,286.27513736635433,284.44482421659194,282.2739253308762,279.7867701210265,277.0074335826553,273.95952922061804,270.6661427886824,267.1499012780828,263.43316372175343,259.5383132968974,255.48812428339286,251.3061729830429,247.01725896051258,242.6478020764019,238.22618179387254,233.7829880999596,229.35115794112548,224.96597709026221,220.6649345228667,216.48742430692332,212.47429828708468,208.66728103341055,205.1082661968045,201.83852016281355,198.89782436603952,196.32359152913318,194.1499932150429,192.4071363133653,191.12032440442803,190.309436440739,189.9884500323889,190.16513008440967,190.84089594545654,192.0108719783856,193.66411797826586,195.7840275778768,198.348875119712,201.3324848288119,204.7049908308059,208.4336528917549,212.48369089322165,216.81910108564898,221.40341907497418,226.2003981827878,231.1745770783171,236.29171712856476,241.51909740183336,246.8256632920812,252.18203288210196,257.56037300227194,262.93416405582445,268.27787869754246,273.56660405517357,278.7756401315131,283.8801081671833,288.85460202226443,293.6729130900055,298.30785502479984,302.73120887518405,306.9138023633927,310.825729409488,314.43670796563697,317.71656623059584,320.63583977997365,323.1664554744223,325.2824725506029,326.9608473500913,328.18218591197774,328.9314482685236,329.198569766289,328.97896801637535,328.27390899167995,327.09071208591433,325.4427813060679,323.34945780921913,320.8356973028481,317.93158397793655,314.67170021741344,311.0943779304371,307.2408626676334,303.1544254038278,298.87945884524026,294.46059523155316,289.9418808604197,285.36603906008366,280.7738482633848,276.20365546342293,271.6910379945659,267.26861866873566,262.96603122155386,258.81002520888387,254.82469234970742,251.03179020813533,247.45113436215777,244.10102706394372,240.99868901582232,238.1606613342669,235.60314702163666,233.34226518781406,231.39419664886006,229.77520608822743,228.5015333471965,227.5891542211794,227.05341896113293,226.9085840977122,227.1672598244889,227.8398006420944,228.93367098461425,230.4528199059403,232.39709946727805,234.7617602018058,237.5370539986245,240.7079700983304,244.25412386412972,248.14981089495578,252.36423124609243,256.8618804207642,261.60309581126177,266.5447398126401,271.64099429050884,276.844235793975,282.1059571372646,287.3776989215093,292.6119543349087,297.7630121636745,302.7877062815626,307.64604478585477,312.3016981547614,316.7223329903676,320.87978570701165,324.7500785212316,328.31328788056015,331.55328263117804,334.4573553987783,337.01577552561,339.22129522137215,341.0686421802517,342.55403171538757,343.67472948425836,344.42869223609006,344.81430890480436,344.83025807444307,344.47549069668514,343.74933932464825,342.6517474504289,341.1836052054358,339.34717109347787,337.1465539286037,334.58822503707665,331.681528276844,328.43915466512186,324.8775494302802,321.01722206910136,316.88293435435133,312.50374697606594,307.9129123164424,303.147608399469,298.2485169309966,293.25925614373193,288.22568647922753,283.19511359485375,278.2154184422527,273.3341479514661,268.5976019748986,264.0499524874073,259.73242958540663,255.68260565458104,251.93380434748954,248.51465497451176,245.4488058773862,242.75480268846326,240.4461294773611,238.53140305804186,237.0147035716094,235.896018240912,235.1717702300465,234.8354010865103,234.87797346806764,235.28876084418084,236.05579260442164,237.1663264030419,238.60722443280363,240.36521638864284,242.42703882654197,244.7794480734406,247.40911140255605,250.30238845252765,253.44502145185527,256.82175836205346,260.4159372783512,264.2090631003406,268.1804084633264,272.3066701502234,276.5617097218029,280.91640303810027,285.3386179073774,289.7933325748963,294.24290049289743,298.64745917600317,302.9654733426541,307.15439537155186,311.17141973850926,314.97430287770896,318.5222161068023,321.77659706843673,324.7019646883182,327.2666639581566,329.4435098555302,331.21030425842116,332.55020556640426,333.45193760186714,333.9098318798091,333.9237051167337,333.4985814976314,332.64427634229463,331.37486404355616,329.70805817284094,327.6645352086295,325.26723526370785,322.54067337601066,319.5102933845523,316.2018932259341,312.64114583339796,308.85323395084515,304.86261040319977,300.69288805297407,296.3668562094058,291.9066130367076,287.3337969137783,282.669894077531,277.9365955322675,273.1561733574107,268.3518453452242,263.5480974116759,258.77093541979013,254.0480418194081,249.40881764132718,244.8842966163215,240.50692519168877,236.31020962029223,232.32823870892952,228.5950978429228,225.14419618053074,222.00753410182597,219.21494182637989,216.79332237781222,214.76593264807505,213.15173516544968,211.96485035274875,211.21413471822837,210.90290477283924,211.0288198032134,211.5839292924558,212.55488314751278,213.92329535487679,215.6662446325662,217.75688943708008,220.16516963639145,222.85856353018826,225.8028668685286,228.96296018832692,232.30353216353384,235.78972967045095,239.38770974055646,243.0650742697207,246.79117497158765,250.53728325361703,254.27662707721782,257.98430404987727,261.6370866097667,265.21314085557185,268.69168504894674,272.05261684134507,275.2761396950961,278.34241870932635,281.2312941419919,283.92207744462246,286.3934497829788,288.62347706327085,290.5897487368116,292.2696404746149,293.6406935689537,294.6810970155806,295.37025202513377,295.6893935356299,295.6222394264261,295.15563577584163,294.2801657887813,292.99069099027116,291.2867958907089,289.1731114479858,286.6594980706785,283.7610753475634,280.498092820723,276.89564357541127,272.983229815658,268.79419654958326,264.36505566487074,259.7347277168637,254.94373242150547,250.03336095663852,245.0448636229864,240.01868518207536,234.99377734144454,230.00701355054352,225.092725732149,220.2823760960092,215.6043701040512,211.0840093506004,206.74357596836686,202.6025335441883,198.67782377193382,194.98423348138897,191.53480349722093,188.34124916290972,185.41436238989505,182.7643667546061,180.4012003718845,178.3347058478948,176.57471231243503,175.13100104430384,174.01315318601982,173.23028512325794,172.7906839025416,172.70136121560907,172.9675496619949,173.5921689354032,174.57529205272562,175.91364262211408,177.6001533771641,179.62361382320182,181.96843096864978,184.6145219474525,187.53735114339995,190.70811752365404,194.09409063377953,197.6590864755459,201.36406765600398,205.16784611549164,209.02786172777547,212.9010063764655,216.74446093810045,220.51651205497888,224.17731668638797,227.68958512594497,231.01915732317664,234.13545273188814,237.0117802452082,239.62550173731017,241.9580499528652,243.99480859265563,245.7248690696562,247.14068421247907,248.23764387442077,249.0136007278382,249.46837631835206,249.6032776349209,249.4206530184067,248.92351326584375,248.11523945347395,246.9993935335502,245.579641451956,243.85979172355243,241.84394545348474,239.53674707593603,236.94371895485887,234.07165778054198,230.92906667948756,227.52659434798377,223.87745146269012,219.99777517367073,215.90691462110442,211.6276140276167,207.18607481751206,202.61188414810124,197.9378038983652,193.19942119968482,188.43466864551124,183.68322901399213,178.9858453303744,174.38356207173862,169.9169270154765,165.62518546155349,161.5454992001999,157.71222161701274,154.15625777661828,150.90453433576215,147.97959891336794,145.3993623590004,143.17699053222324,141.3209450864248,139.8351657091379,138.71937966933058,137.9695186995719,137.57821849504896,137.53537268300215,137.8287111768181,138.4443724721902,139.36744067889887,140.5824208380227,142.07363019849743,143.82548839531634,145.82269560123538,148.05029438694837,150.49361786632173,153.13813335969942,155.9691969261611,158.97173937047842,162.12990844377896,165.426694710446,168.84356980123346,172.360165447072,175.95401980558722,179.60041425093283,183.27231917194243,186.94046165620256,190.57352152566972,194.1384553693674,197.6009413485817,200.92593099074625,204.07828828472262,207.02349145320284,209.72836906789996,212.16183988650508,214.29562504935234,216.10490212005692,217.56887284710263,218.67122034097093,219.40043640833062,219.75000580159102,219.7184408190625,219.30916668429782,218.5302650830625,217.39408978566996,215.9167740953312,214.11765464276516,212.01863955095604,209.64355104587258,207.01747308941168,204.16613354109512,201.1153477783827,197.89054676360888,194.51640745123092,191.01659745475646,187.41363935625074,183.72889328925095,179.98264981527493,176.19431899372177,172.38269623716303,168.56628132780287,164.76362406732798,160.99366859311738,157.27606849316976,153.63144648333673,150.08157548603447,146.64946230601572,143.35932050509535,140.23642524777895,137.30684949718548,134.5970876360721,132.13357901701497,129.9421497685636,128.04739609181112,126.47203601346558,125.23625891777517,124.3571030330217,123.84789034973929,123.71774622936746,123.97122733507743,124.60807666587036,125.62311864937753,127.0063007475024,128.74288118664194,130.8137555928328,133.19590884312652,135.86297267003303,138.7858647717687,141.93348163125532,145.27341511077827,148.77266227256484,152.3982988047023,156.11808885269878,159.90100783795452,163.71765978323688,167.540576499816,171.344392412977,175.105895474013,178.8039611771517,182.41938282496636,185.9346165456209,189.33346388062404,192.6007178089015,195.7217996953009,198.68241476769458,201.46825233460078,204.06475413204464,206.45697008682933,208.6295156232358,210.56663869675427,212.25239832818096,213.67094987515546,214.80692596338568,215.6458962421609,216.1748842347738,216.38291578532676,216.26157116126947,215.80551188853462,215.01295392946534,213.8860608396373,212.4312339568056,210.65928130937934,208.58545254520516,206.22933348497304,203.6146005731161,200.76864218846097,197.7220601418101,194.50807039977005,191.1618268392239,187.7196954100379,184.21850828220414,180.69482826658913,177.18425299207115,173.7207860414277,170.3362986157612,167.06010050452855,163.91863343736233,160.93529358358595,158.13038337547056,155.52118630548665,153.12215222553823,150.9451752749661,148.99994216302304,147.2943253565693,145.83479393614127,144.6268145714137,143.67521624085822,142.98449491249238,142.55903827095398,142.4032555117958,142.52160296207992,142.9185025203627,143.59815630652997,144.56426713188196,145.81968011116237,147.3659656366036,149.2029677581093,151.32834455857343,153.73712824145085,156.42133229590812,159.36963128775506,162.56713563337885,165.99527931264853,169.63183309231295,173.45104974245103,177.4239412507212,181.51868150862813,185.70112169980413,189.93539998656772,194.1846223601926,198.41158793778555,202.57952974000838,206.65284118521146,210.5977592275206,214.38297721216827,217.98016400847177,221.36437062735146,224.51431109413983,227.41251053934246,230.04531996850724,232.40280364064247,234.47851108819336,236.26915123568094,237.77419054014845,238.99540035663907,239.9363806606621,240.6020877426734,240.9983925085897,241.13169363361047,241.00860615605455,240.6357413620205,240.01958825501686,239.1665008245075,238.08278904783472,236.77490541546814,235.24971308668898,233.51481686369002,231.57893427820076,229.45228142351297,227.14694687855462,224.67722722939544,222.05989929116782,219.31440708808495,216.46294580791047,213.5304300934612,210.54433989916896,207.53444341919487,204.53240295708315,201.5712757228373,198.6849270917564,195.90737855003547,193.27211613933986,190.8113875081259,188.55551656051173,186.5322641166835,184.7662609908405,183.27853655665908,182.0861613783297,181.20201706948797,180.63470048270875,180.38856294169418,180.46387883664397,180.85713183987653,181.56140157522273,182.56682906878922,183.8611359497545,185.43017032701144,187.25845164075454,189.3296876034284,191.62723855228964,194.1345080139837,196.83524284158108,199.7137316818152,202.75489647680854,205.94427788417318,209.26792158421733,212.71217811088084,216.26343379422607,219.90779437525805,223.63074563880957,227.4168168602458,231.24927290029885,235.1098594072818,238.97862286953549,242.83382334609928,246.65195279617006,250.40786628605477,254.0750272727966,257.625861966329,261.0322117830497,264.2658674385123,267.2991635736181,270.10560921302783,272.6605270070657,274.94167323459885,276.92981099800187,278.6092109016787,279.96805667871223,280.9987375536496,281.69801438099216,282.0670525081951,282.11132057441773,281.8403607492805,281.2674419156118,280.40911269835476,279.2846757619222,277.9156082099185,276.32495505193793,274.5367234467703,272.57530475688213,270.4649493965985,268.229316138904,265.8911131418735,263.4718427004519,260.9916559016003,258.4693172689864,255.92227344881474,253.36681432874596,250.81830999445597,248.29150287465936,245.80083151722806,243.3607608276447,240.98609336925708,238.692237482743,236.49541046677706,234.41275874024598,232.46238158285675,230.66325047753946,229.03502196718026,227.59774797814478,226.37149343403703,225.37587637428243,224.62955041762748,224.14965302479067,223.95124542447925,224.04677114299932,224.4455597596965,225.15340081130677,226.17220977299024,227.4998039050741,229.1298006847443,231.05164580204715,233.25077158543223,235.70888054895224,238.4043428393347,241.31269101098528,244.40719104555106,247.65946509104322,251.04013919888268,254.51948849578082,258.06805278062836,261.65719745142627,265.25959783980943,268.84962929165687,272.40365045603266,275.9001729611703,279.31991666495827,282.6457556534307,285.8625658110541,288.9569898063726,291.91713946479354,294.7322585232377,297.3923705237758,299.8879370164605,302.20955028763876,304.34768256330926,306.2925101807551,308.0338267613668,309.5610541915512,310.8633545052613,311.92983986482716,312.749872069746,313.3134376923202,313.61158032793276,313.63686780189886,313.38386968913204,312.8496193113323,312.0340345451978,310.9402732991398,309.5750023189857,307.9485619220704,306.0750141313229,303.9720672350125,301.6608757466031,299.16572077489434,296.5135816249773,293.7336147349321,290.8565605385327,287.91410229909275,284.9382032063104,281.96044895187634,279.01142255490754,276.1201364204166,273.3135435782883,270.6161459266591,268.04971230855625,265.63311364549503,263.3822764290537,261.31024993977206,259.4273769303606,257.74155246913756,256.25855145023723,254.9824021547022,253.9157813500072,253.06040583918528,252.4173961394568,251.98759003836997,251.77178702908935,251.7709088896669,251.98606671516265,252.41853026688543,253.0696012737678,253.940397999936,255.03156367795515,256.3429160180109,257.8730586940835,259.6189782821089,261.5756514441166,263.73568714427705,266.0890273452944,268.62272703406654,271.3208306976738,274.1643577074279,277.13140371189814,280.1973593678907,283.3352418515155,286.51612889602006,289.7096798954302,292.88472416419836,296.0098929845034,299.054269784406,301.988031790616,304.78305684090344,307.413470705131,309.8561131636461,312.0909050763313,314.10110353767413,315.8734377001736,317.3981236754678,318.66876278780455,319.68213305447654,320.4378888136529,320.9381876521956,321.1872669881162,321.19099467213175,320.9564186908294,320.49134044853315,319.8039342137106,318.9024322435866,317.7948910141184,316.48904910112867,314.9922818429497,313.31165225379516,311.4540520510447,309.42642140619273,307.23603140690875,304.8908094715184,302.3996852859298,299.7729333785841,297.02248828927486,294.16220943238073,291.2080751453177,288.17828892446954,285.0932853002247,281.9756279574563,278.8497982981794,275.7418783766051,272.67913771318354,269.6895386213717,266.8011790900592,264.0416957245393,261.4376515778291,259.0139347792132,256.7931936306709,254.79533230523066,253.03708852288327,251.5317107416473,250.28874768196349,249.31395764575697,248.60933937409587,248.17328040468465,248.00081334073198,248.08396541171817,248.4121824524588,248.97280516688159,249.751573442139,250.73313364131636,251.9015242671681,253.24061712508606,254.73449402454665,256.3677429870396,258.12566266446504,259.9943689628974,261.96080343160065,264.01264852157834,266.1381600487336,268.3259318378923,270.5646113309208,272.84258771460895,275.1476757162153,277.46681854317796,279.7858324916499,282.0892135667871,284.36002316135347,286.5798656020736,288.7289654185514,290.7863467785011,292.730111950514,294.53781019569203,296.1868834399505,297.6551707059337,298.92144981827875,299.9659925219499,300.7711079970961,301.3216498783177,301.6054632874589,301.6137509965582,301.34134151797156,300.78684648264175,299.9526998797564,298.84507732272425,297.4736991890227,295.8515269620689,293.99436709967665,291.9204010092416,289.6496630068092,287.20349030810763,284.6039700358472,281.87340788500956,279.0338414890805,276.1066187633033,273.11205771487454,270.0691996062999,266.99566218087267,263.9075941818765,260.8197269081872,257.7455133369715,254.69734068254604,251.6867983931141,248.7249807158504,245.8228012347499,242.9912962934759,240.2418949807941,237.5866353365395,235.0383095250304,232.6105247557666,230.31767149392076,228.17479574305946,226.1973776221984,224.40102381074138,222.8010864147491,221.41222515250655,220.24793323598303,219.32004975234207,218.63828259603957,218.20976599894672,218.03867545042468,218.12592035392302,218.4689312562383,219.0615540904394,219.89405882022334,220.95326442298634,222.2227765827229,223.68332906948612,225.31321483794986,227.08878863840357,228.98501961717696,230.97607015969828,233.03587621075027,235.13870454275627,237.25966391846532,239.37514972907852,241.4632053423068,243.50378787389133,245.47893116491963,247.37280413973164,249.17166815290648,250.86374212481599,252.43898894290166,253.8888405270392,255.2058819191047,256.3835166068908,257.41563593720275,258.29631488213164,259.01955463093356,259.57908958549086,259.9682724956929,260.18004688757935,260.20701085588775,260.0415709838497,259.6761798986144,259.10364604943703,258.31749996905654,257.3123977764766,256.08454018830884,254.63208395947558,252.9555225458925,251.0580138813588,248.9456354335178,246.62755003320618,244.11607018622465,241.42661345863394,238.57754682327152,235.58992329178278,232.48711945121164,229.2943874026654,226.03833880905205,222.74638208161724,219.44613599947877,216.1648441462677,212.92881440824453,209.76290641781748,206.69008731769983,203.73107269611228,200.90406518725717,198.22459827153628,195.70548750544177,193.35688603798053,191.1864361081576,189.19950353411517,187.399478241099,185.78812083641384,184.36593328002996,183.13253091776997,182.08699357637434,181.22817503882197,180.55495293446697,180.06640474481114,179.76190004536554,179.6411040410826,179.7038926444024,179.9501845151256,180.37970035558803,180.99166407425483,181.78446396652205,182.75529462520564,183.89980174809327,185.21175227752803,186.68275137117521,188.3020256109868,190.05628871671522,191.92970200395675,193.90393712390897,195.9583434875901,198.07021747863345,200.21516537204621,202.36754707183502,204.5009836085611,206.5889080161132,208.60513690835666,210.52443892056377,212.32307622781056,213.97929660033512,215.47375583919833,216.7898538304435,217.91397168578024,218.83560228285504,219.5473717269335,220.04495455828152,220.3268906512467,220.39431642748417,220.2506269961317,219.90108893350387,219.3524254664249,218.6123967241424,217.68939743171032,216.5920929531471,215.32911203725013,213.90881111134362,212.33912069676046,210.6274797125215,208.78085834621476,206.80586507350478,204.70892857015025,202.49654093896103,200.17554509742104,197.75344652896035,195.2387280320108,192.64114568936034,189.9719850483659,187.2442584106781,184.4728270787398,181.67443624124272,178.8676547002039,176.07271661158754,173.31126756908418,170.60602243645846,167.9803470616853,165.45778013793898,163.06151479550272,160.81386184037686,158.73571777684447,156.84606079749838,155.1614967915117,153.69587516795818,152.459991031849,151.46138615461723,150.7042564598543,150.1894686433296,149.9146833274147,149.87457708346398,150.06115100265413,150.46410949549409,151.07128985752684,151.8691210134364,152.84308885028102,153.97818572049582,155.25932302299648,156.67168818207884,158.20103070826693,159.83386616266466,161.55759153647628,163.36051055145518,165.23177242095105,167.16123241921363,169.1392469352718,171.15641930924704,173.20331547309,175.27017010160753,177.34660453530648,179.4213771354933,181.48218500657705,183.51553326210478,185.5066843654598,187.43969573574327,189.29754900297033,191.06236927766955,192.71572783276983,194.2390169421171,195.61388152550848,196.82268893096358,197.8490158167283,198.67812980417784,199.29744342802968,199.6969189218791,199.86940449444825,199.81088586681656,199.52064079300544,199.0012888717512,198.2587339393472,197.3020014547679,196.14297828454212,194.7960669084059,193.27777006020452,191.60622498697523,189.8007086909868,187.8811366047362,185.86757708479269,183.77980290421337,181.6368986418371,179.4569396328486,177.25675413082197,175.05177574960422,172.8559883457167,170.68196052678482,168.54096219017532,166.4431511591644,164.39781431839006,162.4136448472249,160.49903535458066,158.66236602387,156.91226731678117,155.25783833384295,153.70880450482497,152.275601745324,150.96937838330084,149.80191080981626,148.78543369542436,147.93239047890353,147.2551144207348,146.76545457715116,146.47436437628176,146.3914728891658,146.52466025640288,146.87965898175864,147.45970192146584,148.2652358258486,149.29371632881575,150.53949648297555,151.99381649956285,153.6448975024247,155.47813709477944,157.4763996254559,159.6203894816987,161.88909176377356,164.26026151712713,166.71094047393535,169.21797910078502,171.7585417204829,174.3105735723684,176.85321083722857,179.36711776480445,181.83473893945217,184.2404591976584,186.57066853481302,188.81373425536597,190.95988737338678,193.00103461185398,194.93051105306557,196.74279136754865,198.4331794467719,199.9974970893739,201.43179210156546,202.73208478843614,203.8941694101653,204.91348388511437,205.785057016545,206.50353801467094,207.06330832037,207.45867096414867,207.6841081670563,207.7345938493518,207.60594437236264,207.29518837682346,206.80093512857115,206.12372041834738,205.2663098080472,204.23394083432336,203.0344885790289,201.67854265025213,200.179387899598,198.55288590892715,196.81725916874015,194.99278468705256,193.10140826199017,191.16629459032276,189.21133156363226,187.26060935931116,185.33789614884373,183.46613235881662,181.6669644262356,179.96033694200784,178.36415908268845,176.894057447673,175.56322304469222,174.3823554289357,173.35970214711273,172.50118691956519,171.81061565588436,171.2899456666872,170.93960049852112,170.7588108304556,170.74596093004658,170.8989203182411,171.21534152526795,171.6929070638515,172.32951188042924,173.12337140117955,174.0730496616528,175.17740666391816,176.43546879608203,177.84623062688496,179.40840041538283,181.12010503888243,182.97857256281736,184.9798122199989,187.11831205092165,189.3867738545145,191.77590343993475,194.27427154021382,196.86825728423898,199.54208200494054,202.27793660322465,205.05620092679365,207.85574890881605,210.6543287893573,213.42900384326634,216.15663586524448,218.81439138163572,221.38024928807846,223.83348841885183,226.1551344477061,228.32834745474167,230.33873436889746,232.1745741613987,233.8269479316592,235.2897706722029,236.5597262823203,237.63611207374632,238.52060333412868,239.21695226266235,239.7306385745882,240.06849113524714,240.23830102626414,240.24844641606592,240.10754851032777,239.82417575690474,239.40661048860082,238.86268846644953,238.19971753464696,237.42447704200274,236.54329506547103,235.56219603363937,234.48710732671174,233.32411003708742,232.07971649032353,230.76115548664606,229.37664561499776,227.93563744734493,226.44900691710748,224.92918464394404,223.39020925897495,221.8476967373474,220.31872215161647,218.82161488866842,217.37567298206804,216.00080655897244,214.71712425798083,213.54447963891093,212.50199691111317,211.6075966319906,210.87754230150725,210.32602798609025,209.96482528650583,209.80300521254526,209.84674698477062,210.09924163330103,210.5606937203365,211.22841981172144,212.09703770622698,213.15873613753118,214.40361091410165,215.8200504486557,217.3951515052243,219.11514486504822,220.9658105373651,222.9328631173964,225.0022898664092,227.1606269505555,229.39516287296965,231.69406227582422,234.0464077560961,236.44216189448952,238.872056100928,241.32741690029442,243.7999437072566,246.2814547892458,248.76361985313736,251.2376984233063,253.6943028688152,256.1232036026871,258.51319168758124,260.8520109608879,263.1263680032046,265.3220240173029,267.42396818456,269.416667559695,271.284384290137,273.0115471289586,274.5831610516287,275.9852364539935,277.2052180258298,278.23239303530875,279.05825944554266,279.6768359812553,280.084898884943,280.2821335128925,280.27119294614096,280.05766022205324,279.64991539973005,279.0589132189768,278.29788136343717,277.3819530747265,276.3277508940091,275.15294047552516,273.8757746132053,272.51464778763244,271.08768067090284,269.6123521700965,268.1051938461618,266.58155806021387,265.0554671580699,263.5395466195308,262.0450406021551,260.58190393616627,259.1589606068079,257.7841153030841,256.46460189763263,255.20725089237172,254.01875701277515,252.90592830284874,251.875899252063,250.93629261052084,250.09531750569838,249.36179510574996,248.74510718778538,248.25506734667957,247.90171898903574,247.69506846447305,247.64476546963454,247.75974601869655,248.0478556399741,248.51547190526983,249.16714584886736,250.00528126067184,251.02986927063043,252.23829315895875,253.62521505643403,255.18255230984082,256.8995469806229,258.7629274427463,260.7571565816811,262.86475690118357,265.0666991349987,267.34283792810413,269.67237595264504,272.0343365697535,274.4080249031342,276.7734579641797,279.1117462189324,281.4054116211033,283.63863051342395,285.7973937441254,287.8695806494189,289.8449479912423,291.71503928046394,293.47302393300464,295.11347919115974,296.63213051431853,298.0255680595183,299.29095783522376,300.4257660739539,301.4275143360653,302.2935808855841,303.02106107610274,303.60669600061226,304.0468746795408,304.3377107978975,304.4751906816772,304.45538505481375,304.27471336107175,303.93024626969816,303.42002957767335,302.74341120329814,301.9013524173069,300.89670490948976,299.73443671925827,298.4217923946997,296.96837586599116,295.3861482639915,293.68933708915625,291.8942575229054,290.01905104534706,288.0833506527529,286.10788564095856,284.114041946349,282.1233962574964,280.15724341299875,278.2361369177174,276.37946172486676,274.6050567818688,272.92890231030333,271.3648835170177,269.9246385854858,268.6174945744603,267.45049047461515,266.42848237063356,265.5543216491976,264.82909368981456,264.25240165586206,263.8226780120302,263.53750533221177,263.39392788075605,263.38873635026425,263.518709969741,263.7808028582082,264.1722648476953,264.6906908568539,265.3339970568757,266.1003263138527,266.98788949109854,267.9947529339303,269.11858563979126,270.3563820697452,271.7041781529708,273.15677868877776,274.70751402526184,276.34804260584684,278.0682137909906,279.8560023956808,281.69752278850206,283.57712636247163,285.4775819221086,287.38033425801274,289.2658321227135,291.1139131916216,292.9042305801705,294.61670325007856,296.23197129107626,297.7318366799246,299.09967071642893,300.3207708856015,301.3826523157798,302.2752621687232,302.99110904516283,303.5253036235769,303.8755110568367,304.0418199078788,304.02653639175014,303.8339162017565,303.4698490517283,302.94151311886515,302.2570177186562,301.4250527289108,300.454562498255,299.3544602704368,298.1333966224883,296.7995921885104,295.36074119391515,293.8239882566991,292.1959767385295,290.4829628700957,288.6909861468782,286.82608329012896,284.8945305624116,282.90309754969104,280.8592947612068,278.77159759442554,276.64963035514484,274.50429605240925,272.34784049781103,270.193842681499,268.0571272920889,265.9535993901954,263.9000054179897,261.9136287109124,260.01193126315127,258.2121564972398,256.53091004230106,254.98373691607893,253.58471395909885,252.34607586010958,251.27789166563244,250.38780635705893,249.68085902532874,249.1593855335078,248.8230095197175,248.6687213655612,248.69104055682567,248.88225291041928,249.23271063779782,249.731180340387,250.36522193355663,251.121580278685,251.9865710277399,252.94644286233083,253.9876999003551,255.09737046152506,256.26321149857296,257.4738416479146,258.7187998387072,259.98853051192765,261.27430052445,262.5680565353377,263.8622348971137,265.1495386344606,266.42269785273527,267.6742307871655,268.8962226359471,270.08013832125977,271.216683445058,272.29572505108536,273.306280511517,274.2365791000107,275.0741967911514,275.80626075160035,276.4197160757949,276.9016437753766,277.2396160439293,277.4220725452587,277.4387000351406,277.2807971003301,276.9416062140826,276.416596644728,275.70368394503276,274.8033746826868,273.71882859512573,272.45583428313137,271.02269869414556,269.430054774277,267.6905955744512,265.818746579629,263.83029091129845,261.74196418486366,259.5710370766847,257.33490400677977,255.05069575710195,252.7349323552474,250.40323023986997,248.07007471020697,245.748665107488,243.45083626802156,241.1870557319282,238.9664922005872,236.79714701881122,234.68603720870482,232.6394159703099,230.663014724966,228.7622898054708,226.94265684157125,225.2096967507758,223.56931897547983,222.02787011480245,220.59217924947777,219.2695348830769,218.06759233035825,216.99421436649016,216.05725179821886,215.26427412590917,214.622263447764,214.13728705603302,213.8141656671338,213.65615483308497,213.66465676671277,213.83897859216412,214.17615096718563,214.6708182204317,215.31520775035898,216.09918261768104,217.010377228352,218.03441195755147,219.15517871665762,220.35518601383146,221.61594918292457,222.91840930323232,224.24336301460613,225.57188501556317,226.8857255362535,228.1676664754622,229.40182210678148,230.5738731767165,231.67122668347125,232.68309745788915,233.6005116671158,234.41623631836524,235.12464254870906,235.72151375448672,236.20381227165353,236.56942022931133,236.81687126539086,236.94508996299865,236.9531551324944,236.84010146952102,236.6047717497279,236.24572870412996,235.76123221610314,235.1492836784551,234.40773544891647,233.53445955180112,232.5275662936036,231.38566047560474,230.10812055569477,228.6953845587307,227.14922584306876,225.47300203846476,223.67186156812878,221.75289409948095,219.72521393446667,217.599968613028,215.3902686931196,213.11103859573993,210.77879235973515,208.411341932862,206.02744903612157,203.64643450041106,201.2877611394288,198.97060757822047,196.71345093269318,194.53367580760064,192.4472257717405,190.46831134804154,188.60918573370012,186.8799960887562,185.28871447632272,183.841148599363,182.54102856190858,181.3901621898924,180.38864816920346,179.53513356495958,178.82710031430466,178.261164135196,177.8333690224002,177.5394611198941,177.37512723011287,177.33618546438697,177.41871843448465,177.6191427774697,177.93421251431445,178.36095756978781,178.8965625240055,179.53819412631316,180.2827890965752,181.1268161088409,182.06602747209692,183.0952168062626,184.20799891662534,185.39662710061162,186.65186132716573,187.96289820436326,189.3173705269153,190.7014206350722,192.09984800616806,193.49632763887172,194.8736920800444,196.21426657903837,197.50024401037854,198.7140840319453,199.83891955594424,200.8589530771346,201.75982575533692,202.52894336934938,203.15574528410315,203.63190529735542,203.95145651529418,204.11083607801245,204.10884942577889,203.94655766518974,203.62709526081957,203.155428553913,202.53806832667047,201.78275164991769,200.89810947122174,199.89333675958213,198.77788150649548,197.56116752204827,196.25236383343338,194.86021070675693,193.39290902011828,191.85807609186568,190.26276730483386,188.61355916438157,186.91668598078857,185.1782193569169,183.40427724860874,181.60124767513577,179.77601127902642,177.9361469148506,176.0901052859432,174.24733730231122,172.41836621497905,170.61479556642468,168.84924842656608,167.1352370785755,165.4869660843285,163.91907529746896,162.446332711473,161.08328985593747,159.84391463762523,158.7412179481401,157.78689095166155,156.99096969165308,156.3615425277769,155.90451399081837,155.62343602312802,155.51941439008834,155.59109446876545,155.83472682814164,156.24430920700263,156.8117978671817,157.52737803838374,158.3797804438072,159.35662884277713,160.4448022517919,161.6307950722448,162.90105878109608,164.24231010448102,165.64179262463145,167.0874814586565,168.56822385063325,170.07381206645107,171.59498868658696,173.12338806025707,174.65142112233983,176.1721138017123,177.6789117076834,179.165465541711,180.62541265307377,182.05217028975784,183.43875538380922,184.77764419251886,186.06068287412592,187.2790572283846,188.42332652926865,189.4835227940608,190.4493131595347,191.31021946606648,192.0558858730358,192.67638251719754,193.16253102880952,193.50623625499242,193.70080788456818,193.74125585959698,193.6245444878622,193.34979198570156,192.91840468883134,192.33413824096863,191.60308154751732,190.73356298464398,189.73598209131833,188.62257355023735,187.407113498278,186.10458093156265,184.73078904441272,183.3020026594546,181.83455840146945,180.34450391670757,178.84727126288794,177.35739765651607,175.88830416662117,174.45213982483818,173.05969514518554,171.72038539546497,170.44230032723877,169.23231364284928,168.09624243474266,167.03904433270895,166.0650382679905,165.17813370344584,164.38205293955065,163.68053169728168,163.07748456750346,162.57712402714432,162.1840244421256,161.90312566069977,161.73967427930177,161.69910425158713,161.78686201969776,162.00818458894474,162.36784177105366,162.86985603795898,163.51721493977857,164.3115917659774,165.2530900280044,166.34002641832282,167.56876520132047,168.93361460328066,170.42679281468438,172.03846785144674,173.75687191670903,175.5684872483911,177.45829691942657,179.41009085978558,181.40681465761367,183.4309466112732,185.46488715226428,187.49134421014392,189.49369837202445,191.45633278512804,193.36491460295923,195.20661728639342,196.97027610869307,198.6464726192551,200.22754741697386,201.70754418311427,203.0820913383954,204.34823074246756,205.50420538826384,206.54921992844913,207.48318901027122,208.30648873169372,209.01972605107647,209.62353971308016,210.11844426446552,210.5047261322458,210.78239766244613,210.95121163438296,211.01073525281132,210.9604791649957,210.80007383505614,210.52948280433594,210.1492401228267,209.6606976721483,209.06626729876027,208.36964268042763,207.57598666057993,206.69207136281338,205.7263606599221,204.6890274011308,203.59190105051763,202.4483448898152,201.2730655075429,200.0818607478217,198.8913154464096,197.71845697414724,196.58038469885759,195.49388885695214,194.47507492376027,193.5390093559132,192.69940156134476,191.96833518572691,191.35605938048238,190.87084776304172,190.51892944833608,190.30449299465505,190.2297605503242,190.29512609431237,190.49934860914925,190.83978846571355,191.31267336763037,191.91337899628022,192.6367090757563,193.47715995932194,194.42915600172273,195.48724386097393,196.64623636743826,197.9012995735486,199.24797989582086,200.6821717065817,202.200029143379,203.79782909873452,205.47179516271936,207.21789456718173,209.03162180294456,210.90778346400506,212.84029896876683,214.82203111206198,216.84465894926907,218.89860338103665,220.9730131055396,223.05581547625874,225.13383341085316,227.19296601851218,229.21842723148507,231.19503361854902,233.1075298869596,234.9409384851176,236.68091831084448,238.31411688399294,239.8285004916355,241.21364775227454,242.46099372431635,243.56401401680782,244.51834122649552,245.32180927632768,245.97442469835448,246.47826740927954,246.83732688863898,247.0572827144222,247.14524098237604,247.1094401016416,246.9589407209368,246.70331503333236,246.3523504110776,245.9157812535334,245.40306114996332,244.82318506077328,244.18456833389473,243.49498615082157,242.76157361003405,241.99088328316836,241.18899389922345,240.36166099182554,239.51449803242735,238.65317488969723,237.78361949040334,236.91220836012567,236.0459323029973,235.19252480733624,234.36054276816438,233.5593916923744,232.79929056188047,232.09117481583576,231.44653930177478,230.87722635878998,230.39516725895467,230.01208788462253,229.7391916189616,229.58683386396044,229.56420329767164,229.67902490276737,229.93729894430223,230.34308848907375,230.89836582345714,231.60292535626425,232.4543674310736,233.44815408197968,234.57773432299675,235.83473324224488,237.20919614871266,238.6898764474236,240.2645539301297,241.92036886456847,243.644156711385,245.42276852039166,247.2433630427482,249.09365828900522,250.96213257401826,252.8381678957062,254.7121316470902,256.5753959938142,258.42029758712823,260.24004344994455,262.02857170514415,263.78037816287554,265.49032152485603,267.1534210084324,268.76466048787677,270.31881278106937,271.810296502544,273.2330760237732,274.5806126288263,275.8458720584593,277.0213904517652,278.09939739034394,279.0719915003987,279.9313610452089,280.6700393041046,281.28118242414774,281.7588559593603,282.09831555822416,282.296267264451,282.35109366010863,282.26303356577466,282.0343051434193,281.6691649144158,281.17389826967457,280.5567403528972,279.827729570358,278.9984992469679,278.0820159397553,277.09227548183554,276.0439698304376,274.9521391289483,273.831823997237,272.6977329069601,271.56393858902214,270.44361580795123,269.3488306082332,268.29038840824404,267.27774523176873,266.3189830869432,265.4208471993609,264.58883965244854,263.82736114826446,263.1398902230511,262.5291874572129,261.9975111010167,261.5468301517579,261.179021283342,260.89603712372286,260.7000351396509,260.5934587268748,260.5790648923652,260.6598960048605,260.83919631693584,261.12027715465126,261.506337660336,262.0002506016762,262.60432488653987,263.32005793481363,264.1478918751421,265.08698761155125,266.1350301370048,267.28807709153676,268.54046054259044,269.88474940896367,271.3117769898173,272.81073485016367,274.3693310208246,275.9740072651167,277.610207212974,279.26268461907216,280.9158389967169,282.55406451758347,284.16209741907977,285.72534725919184,287.23019819731354,288.6642680143682,290.0166147359303,291.27788337553505,292.44038833337936,293.49813021087255,294.4467490661709,295.2834192703676,296.00669396598164,296.61630953196914,297.1129622991059,297.4980709423917,297.77353844378956,297.9415272474501,298.00426023757217,297.96385851100604,297.82222468218674,297.58097776672787,297.241442685492,296.80469427294247,296.27165252966324,295.64322289579155,294.9204726967189,294.10483276388817,293.19831167577786,292.20370918002226,291.1248151944404,289.96658135201,288.73525332362885,287.43845405712585,286.08521051233004,284.68591932281765,283.25224992650544,281.79698691711855,280.33381650914356,278.87706491615785,277.4413989645875,276.04150127023314,274.6917336885628,273.4058034394735,272.1964462668327,271.07514022454717,270.05186222401306,269.13489740823155,268.3307088431182,267.64387207025914,267.07707590042696,266.6311876075882,266.30537757554333,266.0972956142451,266.00328874628127,266.0186483899069,266.1378736287611,266.35493672189125,266.66353719615137,267.05733176369364,267.53012887016763,268.0760388198364,268.6895730274873,269.3656888750686,270.0997797479821,270.8876129278696,271.72522096151584,272.60875475373234,273.5343088071007,274.4977306385424,275.4944273571379,276.5191826391113,277.56599687110895,278.6279620761682,279.69718144881006,280.7647410003478,281.8207380748687,282.8543684842849,283.85407088549806,284.8077239481726,285.702888999318,286.5270883312246,287.2681073537107,287.9143073657321,288.4549349894844,288.8804142912878,289.1826083096772,289.35503808682773,289.39304928390584,289.2939189516454,289.0568978955848,288.68318717258427,288.17585042129633,287.5396668005876,286.7809321280421,285.9072182305374,284.92710241637536,283.8498802568564,282.68527546027127,281.44316050452255,280.1333008746734,278.76513427402887,277.3475941215267,275.88898412395463,274.39690785279134,272.8782542134999,271.33923663039315,269.7854808441175,268.22215358518224,266.65412218399143,265.0861335208638,263.52299969505924,261.9697774530019,260.43192877974394,258.9154511043158,257.426967243149,255.973767417503,254.56379831378524,253.20559707075185,251.90817112094265,250.68082782475523,249.5329606554148,248.47380117288964,247.51214803349143,246.65608571310113,245.91270639952174,245.2878485914059,244.78586532194558,244.40943363608554,244.15941505647632,244.03477437328652,244.0325613103244,244.14795659856304,244.3743808843858,244.7036618741481,245.12625232527355,245.63148908103307,246.2078814355013,246.8434158047538,247.52586303658484,248.24307474548698,248.98325580751316,249.735201549981,250.48849014886596,251.23362319727104,251.96211020219533,252.6664957569173,253.34033116568355,253.9780952073235,254.5750713627645,255.12719106058665,255.63085419883632,256.0827392921788,256.4796160153031,256.8181726457171,257.0948699673616,257.3058316318942,257.44677887079575,257.51301492064556,257.4994617003496,257.40074831246704,257.21134798814956,256.9257573128992,256.53870910566184,256.0454083073145,255.44177877344023,254.72470803834744,253.89227696732056,252.94396175130103,251.88079689582312,250.70548965273377,249.42247864633745,248.03793213597854,246.5596842953296,244.99711092249328,243.36094896717128,241.66306701761164,239.91619628810056,238.13363356347378,236.32892889173297,234.51557150097307,232.70668741793963,230.91476158378788,229.15139593471258,227.42711301136393,225.75121228155655,224.13168363053515,222.57517953486354,221.0870444433143,219.67139699696074,218.33125808138976,217.06871545347795,215.88511393979368,214.78125905301988,213.7576213751267,212.81452923530614,211.95233805531544,211.171566198367,210.4729891611388,209.8576863847036,209.3270376994371,208.88266931573327,208.52635217297916,208.25985820999094,208.0847825752601,208.0023418247865,208.0131596513024,208.11705257169382,208.31282822198173,208.59810845859673,208.96918836344256,209.42094055509,209.94677200720207,210.53863798230225,211.1871148391252,211.88153051334106,212.61014855842376,213.36039891746677,214.1191462191726,214.87298447570947,215.60854570617312,216.31280928723237,216.97339877943998,217.57885359604157,218.11886413834097,218.58446085195752,218.96814996606778,219.26399134263247,219.4676167459235,219.57618979447295,219.58831172464016,219.5038797288013,219.323906895634,219.0503145585585,218.6857090601968,218.23315550513118,217.69596097317185,217.077478908487,216.38094502841057,215.6093531837655,214.7653772525555,213.85134248614702,212.86924689304877,211.82083039035064,210.7076867279693,209.53141074020024,208.29377143226645,206.9968998748116,205.64347993783883,204.23692959744272,202.78156090998868,201.2827077506809,199.74681200458338,198.18146099506825,196.59537142778646,194.99831788721082,193.40100680328715,191.81489965705003,190.25199186702437,188.72455615297412,187.24486108681378,185.8248769120807,184.47598147135972,183.2086791855735,182.03234547430188,180.95500782083735,179.98317293093575,179.1217072012071,178.3737751176666,177.74083738174562,177.2227076563682,176.8176639889066,176.52260834750578,176.33326543752736,176.24441016231552,176.25011184916588,176.34398273997286,176.51941827663208,176.7698173898649,177.08877228882028,177.4702190778752,177.90854279863092,178.3986330883362,178.93588942337493,179.51617773115032,180.135742857078,180.7910838229589,181.478800879763,182.19542493312917,182.93724092094283,183.70011709672858,184.47935190081114,185.26954919776227,186.06453117125372,186.8572961744402,187.64002644029068,188.40414788779375,189.14044145700367,189.8392026158506,190.49044305121618,191.0841262243297,191.61042655883745,192.06000063969478,192.42425800582402,192.6956189616667,192.86774732144372,192.9357471104728,192.89631392319637,192.7478337911221,192.49042493425915,192.12592052638067,191.65779345486774,191.09102685254402,190.4319367772988,189.68795568140752,188.86738712976995,187.97914350181864,187.03247908099902,186.03673096561167,185.0010796259469,183.9343397188801,182.84479001779113,181.74004911623996,180.62700103482567,179.51177213563355,178.3997579711893,177.29569601091606,176.20377773857123,175.12779152733188,174.07128608414965,173.03774319582533,172.03074806075998,171.05414567516985,170.1121725500479,169.2095544219427,168.351562512924,167.5440231926146,166.79327847636333,166.10609752212542,165.48954202033397,164.9507909617803,164.49693258147275,164.1347331894034,163.87039401087264,163.7093079945153,163.6558287606282,163.71306344349307,163.88270014993446,164.16487916516937,164.55811496815772,165.05927367914066,165.66360787816194,166.3648479441251,167.1553463137892,168.0262684913734,168.9678223849059,169.96951572140398,171.02042999316998,172.1094986782749,173.2257773951684,174.35869419712617,175.4982693569637,176.6352956749771,177.76147247433835,178.86948891584774,179.95305493834644,181.0068808719945,182.02660943559178,183.00870627704484,183.9503173204155,184.84910283443006,185.7030592505097,186.51034027632116,187.26908874786992,187.97728994623847,188.6326558134029,189.23254770447525,189.77394310702775,190.25344925908632,190.66736393843343,191.01178101783057,191.28273582537275,191.47638305162582,191.58919802709042,191.6181907565409,191.56112121730357,191.41670415371087,191.18479194523934,190.86652507415752,190.46444122125047,189.98253599768918,189.42627067479663,188.8025248791683,188.11949494303153,187.38654129748207,186.61399082925564,185.81290235853973,184.9948052200784,184.17142224807955,183.3543892092708,182.55498285991658,181.78386931569645,181.05088334425642,180.36484757600985,179.73343856420794,179.16310421859234,178.65903451391378,178.22518467214866,177.8643473761591,177.57826813014617,177.36779576521658,177.23305840614756,177.17365405445696,177.1888443621865,177.27774019890018,177.43946824784905,177.67330907088336,177.9787987904751,178.35578865971348,178.80445921431954,179.32528829592172,179.91897486548635,180.58632305067368,181.32809315739988,182.14482830305693,183.036666794664,184.00315130196537,185.04304621377133,186.15417429717655,187.3332829170691,188.5759486620747,189.87652733606356,191.22815400981455,192.6227953033813,194.05135341773834,195.50381879267744,196.9694657744307,198.43708346110583,199.89523207219676,201.33251385507194,202.73784676583122,204.10072898452822,205.41148275455132,206.66146705036863,207.84325012385355,208.95073497636946,209.9792331476419,210.92548478119414,211.78762558621764,212.56510392909155,213.258553718914,213.869630874658,214.4008228675094,214.85524203316555,215.23641398449408,215.54807249377458,215.79397165585215,215.97772502023648,216.10267975234703,216.1718318389258,216.18778599904155,216.1527614250071,216.06864189129217,215.93706627087462,215.75955321940256,215.5376518483171,215.27310871099812,214.96804045002136,214.62510105028096,214.24763283344322,213.83979110415748,213.4066336771609,212.95416830786402,212.48935322326076,212.02004839118487,211.5549177461511,211.10328517410983,210.67494951103237,210.2799660028332,209.92840349183393,209.63008794246727,209.39434372549223,209.22974430315966,209.14388358551034,209.1431782793517,209.2327100755249,209.41611459338617,209.6955217239604,210.07154950314813,210.54335103384554,211.10871139623046,211.76418907224343,212.50529428806576,213.32669495642347,214.2224396670289,215.18618649099272,216.21142626729483,217.291689529508,218.420727281308,219.59265738303415,220.80207028561924,222.04409013767264,223.31438977514662,224.60916064964616,225.92504122679975,227.25900965987665,228.6082484964965,229.96999070532573,231.34135733497783,232.7191975853634,234.09994195824697,235.47947846510158,236.85306064307386,238.2152544288236,239.55992885486413,240.8802931736396,242.16898050512197,243.41817557624208,244.61978170796456,245.76562003522014,246.84765213049047,247.85821583922157,248.79026329639905,249.6375898231376,250.39504271488855,251.05869981326032,251.62600915624256,252.09588285434705,252.468740546898,252.74650023908822,252.93251687969925,253.0314715790148,253.04921675502,252.99258461006747,252.8691680708346,252.68708458274486,252.4547338719713,252.18056093763255,251.87283510769046,251.53945500842624,251.18778781216142,250.82454922015947,250.45572840837508,250.08655973050818,249.721540463722,249.36449142916743,249.01865505073997,248.68682344997774,248.3714876153151,248.07499761118152,247.79972326216242,247.54820478654227,247.3232834590077,247.1282035213511,246.9666781719183,246.84291446374132,246.76159422195892,246.72781053270336,246.7469618291742,246.82460797555981,246.96629490044234,247.17735614474427,247.46270106927824,247.8266003404003,248.27247963253848,248.80273223583504,249.41856044803478,250.1198543035189,250.90511441749717,251.77142359173172,252.7144694502889,253.72861787302867,254.8070345005419,255.9418492263161,257.1243564921463,258.3452424697786,259.59482893482385,260.8633228836117,262.1410607481116,263.4187364373482,264.68760335524445,265.93964196534426,267.16768631725705,268.3655051208066,269.5278353377428,270.6503687325754,271.7296942543426,272.76320138307636,273.74895155055015,274.6855263319727,275.5718622215486,276.40708239300153,277.1903358767784,277.9206540585374,278.5968333477585,279.217351337704,279.7803218599971,280.2834921307008,280.7242828069196,281.0998693494963,281.407300746486,281.6436495175307,281.80618510387933,281.8925613486513,281.90100786126754,281.83051468730315,281.68099989066945,281.4534503900579,281.1500276382497,280.7741314266858,280.3304171506964,279.82476417630147,279.2641953876367,278.6567504380227,278.01131755041513,277.3374307937239,276.6450414922481,275.94427371648516,275.2451745881042,274.5574703696093,273.890338988386,273.252208782782,272.65059190013307,272.0919589945334,271.5816597589837,271.12389149277897,270.721715472417,270.3771184896237,270.09111466838135,269.8638806909246,269.69491595307176,269.5832180148524,269.52746307206183,269.52618108001735,269.57791561562124,269.6813595423107,269.8354589916171,270.0394800173019,270.2930344152964,270.5960635215144,270.9487811772031,271.35157936207105,271.8049021168589,272.3090951972112,272.8642403234886,273.4699838412799,274.12537003445317,274.82868921370823,275.57735004452127,276.3677844131916,277.1953915190423,278.05452590866014,278.93853193729353,279.8398247695966,280.7500156406934,281.66007681386446,282.5605396122659,283.44171717641694,284.29394229674483,285.1078098591413,285.87441316351834,286.5855636451484,287.2339843319915,287.8134686662705,288.3189980377243,288.7468134296565,289.09443886019875,289.3606566914869,289.545437254547,289.649827474437,289.67580516282476,289.62610727138923,289.50404158570603,289.31329202519953,289.05772786620776,288.74122681605843,288.3675209584805,287.94007321392786,287.4619901862297,286.93597519373293,286.3643230187537,285.74895557351215,285.09149539718874,284.39337178777936,283.65595254532235,282.88069285605576,282.0692918563137,281.22384693263996,280.3469958660769,279.4420375112351,278.51302278401664,277.56480925894857,276.6030745678883,275.6342859472451,274.66562558781163,273.7048737788734,272.76025408391797,271.8402468212876,270.953378842751,270.10799891602346,269.3120488548076,268.57284085786694,267.89685130005614,267.2895404741583,266.7552065513164,266.29688037362246,265.9162657009248,265.6137273085117,265.3883269885047,265.23790516708084,265.1592036332757,265.14802289767437,265.1994060627734,265.30783987521806,265.4674629052461,265.6722705967304,265.9163072610973,266.19383593117715,266.4994783010372,266.82831868501853,267.17596794286317,267.53858553140776,267.91286013999536,268.29595162550805,268.68539906494516,269.0790015783538,269.47468004643724,269.87032887829304,270.2636675216717,270.6521014230959,271.0326016378036,271.4016112868553,271.75498561396495,272.0879705839445,272.3952228835398,272.6708719434852,272.9086223158375,273.10189253328974,273.2439845633891,273.32827625637293,273.34842786101876,273.29859281835115,273.1736226840982,272.969256196825,272.682283191639,272.31067522420494,271.85367635631644,271.3118494802636,270.6870757248311,269.98250677866855,269.20247226811455,268.35234651711653,267.4383809833446,266.4675103054053,265.44714112679355,264.38493361990663,263.28858587989197,262.1656310818121,261.0232565113973,259.8681523317454,258.70639630213736,257.5433787077825,256.3837695939876,255.2315281388666,254.08995176387532,252.96176048828696,251.84921019169605,250.7542269538973,249.67855357179246,248.6238987636998,247.5920794926326,246.5851472750141,245.60549026615107,244.65590427915382,243.73962762731858,242.86033668947957,242.02210127724203,241.2293011172227,240.48650693252344,239.79833160107614,239.16925857901194,238.60345611447514,238.1045866705513,237.67562137771785,237.3186692247085,237.034830077033,236.82407951631768,236.68519197774947,236.61570680576628,236.61193974609745,236.66904015444456,236.78109194466762,236.94125413944704,237.14193493600902,237.37499155961765,237.63194693286988,237.90421340342053,238.1833134866455,238.4610878068918,238.7298811485504,238.9826987173358,239.21332629997255,239.41640991336683,239.58749265239237,239.72300866766992,239.8202364149718,239.87721540075793,239.89263249535566,239.8656854004691,239.795931962629,239.68313466208426,239.52710974526912,239.327590102374,239.08411014022047,238.79591961052583,238.4619316936399,238.08070869618962,237.6504866007068,237.1692375184674,236.63476695905752,236.04484085456122,235.39733556711948,234.69040275627756,233.92264005859892,233.0932580855541,232.2022343007219,231.25044489159305,230.23976677632032,229.17314332852325,228.0546091887877,226.88927156692893,225.68324761847754,224.44355968876795,223.17799234344733,221.89491703461624,220.60309188680986,219.3114453417498,218.02885321022472,216.7639190018536,215.52476722252155,214.3188586548843,213.15283550515048,212.0324027688263,210.96225031939267,209.94601815360036,208.98630504325183,208.08471865976935,207.24196316811197,206.4579584381359,205.73198348998537,205.06283565449917,204.44899624802625,203.88879336769625,203.38055271670618,202.92272815213605,202.51400486766917,202.15336971500554,201.84014504596283,201.5739845225763,201.35483148556156,201.18284257896576,201.0582812893692,200.9813877680967,200.95223267428136,200.97056373297445,201.03565419522738,201.14616238996206,201.30001106999063,201.49429430196943,201.72521828198785,201.9880807462898,202.27729167971182,202.58643590554766,202.90837598039533,203.2353917289601,203.55935084584942,203.87190336374138,204.1646915248983,204.42956576170306,204.65879713507283,204.84527671582634,204.9826930161982,205.0656796542403,205.08992690673023,205.05225259968574,204.95062980628828,204.7841709648421,204.55307018301016,204.25850754723018,203.9025211016267,203.48785370357336,203.01778312396212,202.4959444799152,201.92615433000768,201.31224551594218,200.65792111449,199.96663470861927,199.24150265899794,198.4852522378384,197.70020747283374,196.88831244628363,196.0511897144626,195.19022956454089,194.30670411353745,193.4018988660216,192.47725335756942,191.53450197133864,190.57580595388387,189.60386807668232,188.62202226967895,187.63429184627594,186.64541157769716,185.66081077256803,184.68655657528782,183.72925880639713,182.79593971901224,181.89387392995428,181.0304054035937,180.21274963569806,179.44779003706407,178.7418779075704,178.10064529977186,177.52883950171955,177.0301868507224,176.60729217604313,176.26157843242072,175.99326911797377,175.80141397162953,175.68395632559378,175.63783845736347,175.65913944847458,175.74323850842532,175.8849955416811,176.078939983701,176.31945864640315,176.60097350788052,176.91810104387602,177.2657857934626,177.63940231996114,178.03482149233676,178.44843897878619,178.87716590918325,179.318383717723,179.76986711345901,180.22968084279282,180.69605731438332,181.16726317980846,181.6414635492888,182.11659264051477,182.59023930432943,183.05955506277422,183.52119107569524,183.9712688861905,184.40538796526266,184.81867107856883,185.20584643832012,185.56136358962084,185.87953811884546,186.15471866004924,186.38146839878692,186.5547523984301,186.6701216484098,186.72388477958106,186.71325890727894,186.63649202123204,186.4929506938042,186.28316855399154,186.00885288729344,185.6728487717765,185.27906224202772,184.83234597736293,184.33835283482574,183.80336409689627,183.23410049856935,182.63752487728237,182.02064561256535,181.39032987454092,180.75313509042022,180.11516599891326,179.48196324935827,178.85842779014484,178.24878336994195,177.65657744637107,177.0847187665686,176.5355479589307,176.01093575589425,175.5124020436099,175.04124787989355,174.5986919914987,174.1860030877966,173.80461961865254,173.45624934351378,173.14294222680675,172.86713167058974,172.63164085971366,172.4396529341709,172.29464571538816,172.2002936917107,172.16034180881792,172.1784572161596,172.25806640593672,172.40218607868258,172.6132565314959,172.89298636661303,173.24221685858365,173.66081342124647,174.1475903273124,174.7002732204075,175.3155021056324,175.9888755063218,176.71503443553647,177.48778285676093,178.3002395017456,179.14501436804852,180.01440201454477,180.90058297207815,181.79582422954746,182.69266986086782,183.58411341914382,184.46374471099045,185.32586492405025,186.16556574277857,186.9787699642711,187.76223311898508,188.51350760646665,189.23087276897593,189.913236046579,190.56001179644377,191.1709854422306,191.7461712908694,192.28567257959932,192.78955208526762,193.25772095385148,193.68985232751953,194.085324917112,194.44319996575396,194.7622331647178,195.04092111539785,195.27757998538692,195.47045218497024,195.6178352891175,195.71822613386018,195.77047209334577,195.77392104378532,195.728561469704,195.63514456954067,195.4952810511898,195.31150653037372,195.08731099225832,194.82712956895497,194.5362938291551,194.22094476990608,193.88791064061965,193.54455451481581,193.19859806294545,192.85792918997092,192.53040202140096,192.22363810875274,191.94483766012868,191.70060908752015,191.49682422676577,191.33850527790662,191.2297479014736,191.1736830742429,191.17247835178358,191.22737720659563,191.33877321270953,191.5063141288092,191.72902948124963,192.00547414059488,192.3338796772708,192.7123050093591,193.13877803136182,193.611420526332,194.12854968158487,194.68875089562084,195.29091820801162,195.93426051751646,196.61827367948626,197.34268048983023,198.10734236824354,198.91214815272147,199.7568867262249,200.64111114564875,201.56400248283666,202.52424168868282,203.51989744878622,204.54833723039212,205.60616756553785,206.68920813472485,207.7925024864407,208.9103663408354,210.03647247946353,211.16396931898595,212.28562850460304,213.3940153308571,214.4816745831447,215.54132355625714,216.56604359024678,217.54946149063605,218.4859126680775,219.37057871765714,220.199593414003,220.97011265982107,221.68034571117178,222.32954691987126,222.91796918225685,223.44678216309154,223.9179600762344,224.33414526137645,224.69849492357508,225.01451914184412,225.2859185673388,225.5164301056881,225.70968831977657,225.8691093293684,225.99780267397327,226.09851501508405,226.1736077677498,226.22506886388697,226.25455696045285,226.26347461421642,226.2530653457133,226.22452819218154,226.17914237236837,226.1183941063595,226.04409748138676,225.9585015378199,225.8643764532684,225.76507278986873,225.66454918325314,225.5673655171344,225.4786404574317,225.40397411833808,225.3493385000371,225.32094007591488,225.32506042483985,225.36788202171462,225.45530715304446,225.59277836877138,225.78510889352802,226.03633099879505,226.3495695038702,226.726946371223,227.16952085307705,227.6772679092325,228.249095741348,228.8829013733934,229.57566135103372,230.3235529304218,231.1220996671962,231.96633417456363,232.85097005367663,233.77057464886806,234.71973436172058,235.6932047658988,236.6860386713512,237.6936865433114,238.71206522168984,239.7375926284442,240.7671880024275,241.79823906552468,242.82853930260566,243.85620013759785,244.87954412500008,245.896986280515,246.90691129382014,247.90755456875493,248.89689481123975,249.87256524537494,250.83178951748536,251.7713470004467,252.68757060764494,253.57637845194694,254.43333883361464,255.253766209715,256.03284408273436,256.76576923813144,257.44791053894437,258.0749746147737,258.6431703085896,259.1493636931896,259.5912158429804,259.96729632708644,260.27716653633047,260.5214284100504,260.701735813635,260.82076764668363,260.88216363992876,260.89042562913335,260.85078878126694,260.76906870512255,260.6514915299671,260.5045148226481,260.3346475957079,260.14827761753656,259.9515137735982,259.75005037039375,259.54905906676913,259.35311262430645,259.1661429685351,258.99143423493064,258.8316496334948,258.68888919949063,258.5647738979536,258.460550199362,258.377208213543,258.3156058118475,258.27659091752014,258.2611143132918,258.270325893789,258.3056482471575,258.36882273427887,258.46192477694365,258.58734678659937,258.7477489718454,258.94598006026,259.18497166349675,259.467611514474,259.7966020326763,260.1743115631882,260.6026261399068,261.08280971641915,261.6153804838508,262.20001017000203,262.8354521256796,263.5195026086934,264.24899804595714,265.0198492741882,265.82711192233995,266.6650902995683,267.52747048467717,268.4074768631277,269.29804520094206,270.1920045409562,271.0822597972,271.96196692844796,272.8246929912722,273.66455418374943,274.47632615045205,275.2555222664854,275.9984372772251,276.70215545421786,277.3645242432364,277.9840961330259,278.5600430716028,279.09204911806836,279.58018807131543,280.02479350853304,280.426328960864,280.7852658366463,281.10197618164466,281.3766464690124,281.6092173875455,281.7993531103393,281.94644185617557,282.04962779131426,282.1078725532036,282.1200430033093,282.0850203222555,282.00182432569727,281.86974596908425,281.6884804719361,281.45825335607304,281.1799319649932,280.8551156993902,280.4861992320867,280.07640430094784,279.62977725110807,279.1511512261409,278.64607370152765,278.12070181945654,277.5816696295872,277.0359327802289,276.49059736345333,275.9527404357202,275.42923017018353,274.9265536255384,274.4506597378023,274.00682437578047,273.59954318826453,273.2324565694414,272.90830945169876,272.6289468863254,272.39534458354973,272.20767184694745,272.06538274358604,271.9673299828606,271.91189490440235,271.8971262537188,271.92088008991766,271.9809532392095,272.0752031759539,272.20164805402385,272.3585417799243,272.5444204530324,272.7581181213228,272.9987515270118,273.26567525439316,273.55841034995797,273.8765509752047,274.21965489674943,274.58712455084236,274.97808599100205,275.39127320800475,275.82492509126286,276.27670169110104,276.74362547441285,277.22205199223475,277.70767286356363,278.1955523045192,278.68019668338553,279.1556548516669,279.6156453796082,280.05370539652597,280.4633545764879,280.838266979075,281.1724429966504,281.4603735977924,281.6971893946645,281.8787877825728,282.0019324653565,282.0643210351706,282.0646178492025,282.00245115712886,281.87837519280583,281.6937996605139,281.4508906311474,281.1524482351142,280.8017676257615,280.4024904341694,279.95845430531074,279.4735480788118,278.95157975753835,278.3961636172002,277.81063169264996,277.1979734911517,276.5608062031277,275.9013759903418,275.2215892193802,274.5230708648872,273.8072458187964,273.07543758698796,272.3289778992932,271.5693201526246,270.798149381925,270.0174816212474,269.22974606859214,268.4378443740149,267.6451825832216,266.85567272456984,266.07370264953266,265.3040744399794,264.5519133905815,263.82255117219285,263.12138819823514,262.453741376486,261.82468427217924,261.23888719048745,260.70046478170417,260.2128384754104,259.77862037572197,259.3995242331935,259.07630780275457,258.808749368921,258.5956595487368,258.4349277561152,258.3236010179565,258.25799126030194,258.23380581308027,258.2462947853289,258.2904081953997,258.360955341608,258.45275888781134,258.5607965149743,258.68032373313406,258.8069725184452,258.93682178026654,259.06643720265936,259.1928796620049,259.3136831105122,259.4268044455112,259.5305493711286,259.62347952442155,259.7043071161922,259.77178397661953,259.82459216442163,259.86124318194237,259.87999234401735,259.8787740022261,259.85516217316786,259.80635972074015,259.7292176719737,259.6202845866628,259.47588424035,259.2922183057339,259.06548931228855,258.7920380019285,258.4684883412943,258.0918929440533,257.65987152657027,257.17073527470984,256.6235906256046,256.01841693371995,255.3561137455299,254.6385148865111,253.86836819121947,253.04928139749558,252.1856363916656,251.28247554660658,250.3453652582066,249.38024288759772,248.39325409980842,247.39058801408055,246.3783176256693,245.3622526219898,244.34781101587035,243.33991499272759,242.3429150709287,241.36054517378312,240.39590958632868,239.45150110553485,238.52924807581698,237.63058651726584,236.75655227909567,235.90788715173792,235.08515219943294,234.2888412655159,233.5194876702441,232.77775756154102,232.06452416844581,231.38091830307536,230.72835180120643,230.1085121128605,229.52332787198904,228.97490690297735,228.46544967516073,227.9971426128186,227.57203683341663,227.19191875958143,226.8581795841695,226.57169073435423,226.33269227069195,226.14070058115135,225.99444081778043,225.89180832232577,225.82986185910096,225.80484989297284,225.81226949968223,225.8469558606826,225.9031987602694,225.97488114869682,226.05563373151097,226.1389987496271,226.21859566781643,226.28828141409912,226.34229811252507,226.37540191085998,226.38296748798973,226.36106408175564,226.30650034064212,226.216836895777,226.09036718991536,225.9260687019801,225.72352818599958,225.48284582508674,225.20452421813607,224.88934881683068,224.5382667776982,224.15227117121717,223.73229709963553,223.27913553878082,222.7933696763725,222.27533722587359,221.72512071949083,221.14256620477235,220.52732916992792,219.8789449880547,219.19691978097453,218.48083643275262,217.7304695930117,216.9459029479368,216.1276418322214,215.2767144191887,214.39475525064864,213.48406572553446,212.5476473123608,211.5892046253008,210.61311703546036,209.62437909779172,208.62851167606067,207.63144715976716,206.63939350905272,205.65868296641958,204.69561208005052,203.75628015069003,202.84643331814928,201.97132123920255,201.13557268983706,200.34309548385147,199.5970048857846,198.89958327238753,198.2522722374793,197.65569672116683,197.10971915991766,196.6135201810268,196.1657010798627,195.76440228720747,195.40743130994943,195.09239324812657,194.81681697360415,194.57827040025415,194.37445896318047,194.20330241816856,194.06298631903277,193.95198596362638,193.86906214302968,193.8132296012071,193.7837006311903,193.77980761850046,193.8009095201151,193.84628817580077,193.91504094025817,194.00597636812324,194.1175195666019,194.2476333585042,194.39376059643232,194.55279187891216,194.72106159849304,194.89437376951207,195.06805751703214,195.23705054049665,195.39600737807692,195.5394279684734,195.66180090573826,195.7577549672509,195.82221200809448,195.8505341823752,195.83865868062122,195.78321375056674,195.681610666404,195.53210748236162,195.33384178942913,195.08683121737087,194.7919420087887,194.45082755571696,194.06584025107517,193.63992129168534,193.17647411026348,192.67922785741865,192.15209776341,191.59904926295357,191.02397246303886,190.43057289046533,189.82228350733308,189.20220177961187,188.57305418999195,187.93718907519337,187.2965971194806,186.65295733183655,186.0077049531811,185.36211655468028,184.71740666071733,184.0748296090018,183.43578007794798,182.8018857824088,182.1750862590852,181.55769241008127,180.9524225076308,180.3624116303935,179.79119293513443,179.24265069158557,178.7209465426438,178.23042191606268,177.77548083041532,177.36045843805312,176.98948147311725,176.66632727981437,176.39428825823472,176.17604837310293,176.01357783464314,175.90805120782744,175.8597930808032,175.86825408361395,175.93201856393347,176.04884367443466,176.21572808708905,176.4290071028904,176.68446964606875,176.97749158585302,177.3031790700574,177.65651512195888,178.0325026668135,178.42629742045912,178.83332467498,179.24937492303857,179.67067442505365,180.09392818048835,180.51633424460547,180.9355698573579,181.3497513415754,181.75737110523224,182.1572162753146,182.5482744364514,182.92963259687986,183.3003758239494,183.659491965045,184.00578849952865,184.33782687310983,184.65387868500287,184.9519068823826,185.2295737303678,185.48427584230961,185.7132050528591,185.91343247441287,186.0820117721558,186.2160965929264,186.31306624662108,186.3706532101088,186.3870658307901,186.36109976055766,186.292232143412,186.1806933860427,186.02751241859497,185.8345326466978,185.60439723845548,185.34050390657805,185.04693085739845,184.7283370067359,184.38984083336044,184.0368832883815,183.6750809491167,183.310076059291,182.94739021072266,182.59228818968938,182.2496579466379,181.923911780854,181.61891270791162,181.3379286562716,181.0836156896156,180.85802994927587,180.66266653496487,180.49852216891554,180.36617728999101,180.26589226251983,180.1977117092703,180.16157062371022,180.15739590123985,180.18519725238087,180.24514210512302,180.33761003413468,180.46322342171487,180.62285239721248,180.81759354707697,181.0487233602673,181.3176287956115,181.62571865377194,181.97432053827848,182.36456903942144,182.79729132670042,183.27289656067674,183.79127542129248,184.35171560258374,184.9528383662836,185.5925602183865,186.26808252681136,186.9759104997173,187.7119014656271,188.47134091528906,189.24904335828447,190.0394737879839,190.83688450206554,191.63546124678223,192.4294721818729,193.21341302375728,193.98214192439755,194.73099817124867,195.455899622233,196.15341487522767,196.82080745734876,197.4560507382055,198.05781374928645,198.6254195517603,199.15877916132882,199.65830524047624,200.1248107433677,200.5593983965708,200.96334728409397,201.33800285876342,201.68467642190006,202.00455951529833,202.29865778596408,202.56774776217514,202.81235867885306,203.03278007993435,203.22909448002653,203.40123296350643,203.54905031049276,203.67241513379344,203.771309647299,203.84593310993577,203.89680273066412,203.92484589287534,203.93147795725787,203.91866060993823,203.8889367005083,203.84543871134417,203.79186935284267,203.73245421811077,203.6718678800123,203.61513619755405,203.56751884503646,203.53437712107728,203.52103288101452,203.53262492423787,203.57396933220147,203.64943008466068,203.76280578982025,203.91723757422244,204.11514213149715,204.35817268005064,204.6472091933883,204.98237781513092,205.36309792929475,205.78815399982142,206.2557880918938,206.7638080032777,207.30970521737322,207.89077647736445,208.50424269338035,209.14735913547935,209.81751142090326,210.51229264435457,211.2295580801019,211.9674551471198,212.72442770635658,213.4991951797917,214.29070836926064,215.09808513594928,215.9205302114892,216.75724429036288,217.60732815471886,218.46968787515604,219.3429470992923,220.22537208572626,221.11481448302223,222.00867592621222,222.90389737574512,223.79697481610341,224.68400153211286,225.56073576373646,226.4226911792469,227.26524637353575,228.08376855765815,228.87374581217898,229.63092177218718,230.35142642259268,231.0318968186298,231.669582001487,232.26242712914396,232.80913284892438,233.30918714870842,233.76286827534594,234.1712187314011,234.53599178113822,234.85957324005358,235.14488251974902,235.39525788971568,235.6143316487479,235.80590133389597,235.97380321226132,236.12179409531328,236.25344699843404,236.37206536797197,236.480619556883,236.58170800324453,236.67754421833703,236.7699692936111,236.86048826218877,236.95032737271222,237.04050821834468,237.13193376989543,237.2254807360479,237.32209234806103,237.42286565783022,237.5291277469045,237.64249585337592,237.7649173011308,237.89868621536445,238.04643527147593,238.2111020848734,238.39587123516395,238.604094256681,238.83919114784086,239.10453799013413,239.40334606892984,239.73853841039198,240.11262986400496,240.52761675685616,240.9848817289801,241.48511865015487,242.0282815542494,242.61356035755486,243.23938481321184,243.90345676309778,244.60280935354731,245.33389055439508,246.0926671308004,246.8747442256835,247.67549496851424,248.49019407172042,249.31414923215112,250.14282432863882,250.9719488879845,251.7976090548288,252.61631630539304,253.42505133796038,254.22128189145081,255.00295461830657,255.76846249709814,256.516590542684,257.24644369097706,257.9573616430834,258.64882610267017,259.3203661973904,259.9714679219368,260.6014931749745,261.20961339939356,261.7947620051631,262.3556087007267,262.8905576381827,263.39776995401127,263.87520993076174,264.32071268749877,264.73207009744533,265.1071305939718,265.44390771562655,265.7406917001288,265.9961581948631,266.20946822047097,266.38035390139623,266.50918514329544,266.59701335759667,266.6455894603742,266.6573546475298,266.63540380474114,266.583422778213,266.5056020398665,266.4065304602877,266.2910738933275,266.16424402638285,266.03106342136533,265.89643283904013,265.7650067954884,265.641082851864,265.52850941116196,265.43061582672556,265.3501674675845,265.28934709618727,265.2497625621142,265.2324794715151,265.2380762263564,265.26671770592253,265.31824294358637,265.39226148215573,265.4882527054732,265.60566236182444,265.74399071958044,265.90286731475766,266.0821080358056,266.2817513009652,266.502071264305,266.7435672754714,267.00693014709583,267.2929870824783,267.60262831605195,267.93671955690195,268.29600514649485,268.6810074024161,269.0919278906373,269.5285563351779,269.9901925375084,270.4755860560383,270.98289752089755,271.50968437654547,272.0529126109846,272.60899470986715,273.1738527354419,273.74300414389273,274.3116667872294,274.87487855834235,275.42762638153243,275.96497876510284,276.4822159426348,276.974951745102,277.4392417608076,277.8716730324592,278.2694314742762,278.6303443180666,278.95289615695987,279.23621848326763,279.4800539436646,279.6846977915745,279.8509201381215,279.9798735312455,280.0729910797451,280.1318807499654,280.1582215769238,280.15366734381547,280.11976280476364,280.057876781366,279.9691554938433,279.85449834399293,279.71455711029955,279.54975821191846,279.36034641665765,279.14644717612845,278.90814373172543,278.6455643027732,278.35897408686105,278.0488665026829,277.7160481034056,277.3617118838929,276.98749428257617,276.59551200815764,276.1883758592209,275.76917989686035,275.3414656142004,274.90916205495,274.4765040974441,274.0479322747845,273.6279784853185,273.221142709304,272.8317663473563,272.4639080079877,272.1212274841106,271.80688327629633,271.52344836371907,271.27284802627184,271.05632243002486,270.87441545979146,270.72698998093784,270.613268405348,270.53189619145667,270.4810247899892,270.4584096132856,270.46151790492297,270.4876409536347,270.53400495305345,270.59787496315437,270.67664687151256,270.7679229586339,270.86956760393434,270.97974077790997,271.09690819298163,271.2198282656439,271.3475173084475,271.4791955553855,271.61421766617406,271.751992199031,271.89189514362806,272.03318293445216,272.17491040275746,272.31585887081746,272.45447905881673,272.58885269063336,272.7166756912548,272.83526471792226,272.9415875198134,273.03231634269486,273.10390235266476,273.1526679121221,273.174912561368,273.1670277922565,273.1256151864066,273.0476022567504,272.93035039008083,272.77174963691016,272.5702957153793,272.32514545599435,272.0361479689518,271.703850010942,271.32947530150796,270.91487882418784,270.4624783781053,269.9751667576591,269.456208873985,268.90912884367816,268.3375925213662,267.7452911202764,267.13583144092104,266.51263781966566,265.87887023828864,265.2373621384189,264.5905804084541,263.94060881199283,263.28915486906,262.6375789505297,261.98694316776835,261.3380765953353,260.6916525091652,260.04827270060764,259.40855357031506,258.77320863356414,258.1431222835558,257.51941014987335,256.90346212943103,256.29696511768515,255.7019035783751,255.12053730200134,254.55535695274185,254.00901922472565,253.48426455719533,252.9838213346609,252.51030127121268,252.06609120694614,251.65324680097137,251.27339357602403,250.92764045550368,250.61651035098888,250.3398915370888,250.09701253352327,249.886442054414,249.7061143422067,249.55337894298881,249.4250727667018,249.31761117281937,249.22709388663299,249.1494208317152,249.08041249748376,249.0159292708915,248.95198425748916,248.88484449388017,248.81111609070524,248.72780970900914,248.6323838176153,248.52276435004967,248.3973406151994,248.25493855156614,248.09477358609007,247.9163864039761,247.719565801325,247.50426343275447,247.27050564854144,247.0183077206556,246.7475955794868,246.45813973302137,246.1495053418946,245.82102151434168,245.47177181282643,245.1006067857825,244.7061781157109,244.28699277334965,243.8414844501439,243.3680985667307,242.86538637500885,242.33210312677653,241.76730500170157,241.17043948649166,240.5414241758618,239.88070950992122,239.18932174369687,238.46888342205347,237.72160975637453,236.95028050954895,236.1581882300449,235.3490648698958,234.52698991314392,233.69628407393486,232.8613933486077,232.0267686862227,231.19674675226415,230.37543719009105,229.5666214380726,228.77366755563116,227.99946468046477,227.24637972562715,226.51623778197919,225.81032647853766,225.12942333374667,224.47384396822713,223.84350800458424,223.23801860659498,222.6567509537078,222.09894454099717,221.56379406008645,221.05053375941327,220.5585105943409,220.0872421367058,219.6364560842023,219.20610924625814,218.79638502984517,218.40766964536272,218.04050843613783,217.6955448427793,217.37344548720404,217.07481564931152,216.80010997047182,216.5495435226029,216.32300841351287,216.1200008565487,215.93956312799983,215.7802440954547,215.64008106303942,215.51660459413748,215.40686679548423,215.307492339913,215.21475033151307,215.12464403780226,215.03301458503682,214.93565398339481,214.82842235648297,214.7073640200386,214.56881709970446,214.40951169490134,214.22665216845408,214.01797993979028,213.7818141415321,213.517068613905,213.22324490050644,212.90040211070348,212.54910566579025,212.17035798824898,211.7655150720527,211.33619354159003,210.8841732331297,210.41130049420698,209.91939728521115,209.41018079037238,208.88519762277673,208.3457758732967,207.7929972507845,207.2276904436401,206.65044565997127,206.06164913688116,205.46153531042341,204.85025336429302,204.2279440785711,203.5948223216768,202.95126019945994,202.29786581235948,201.63555277803863,200.96559614203417,200.2896709982976,199.6098710383689,198.92870529526942,198.24907249111854,197.57421357606992,196.9076441986659,196.2530699137526,195.61428785784562,194.99507935490476,194.3990984194712,193.82976137262864,193.2901427662134,192.78288252308906,192.3101086610636,191.87337920327,191.4736459283741,191.11124152922318,190.7858905849377,190.49674356946426,190.2424319806186,190.02114163672627,189.83070030676765,189.66867515991808,189.53247507643925,189.41945267653858,189.3270010060468,189.2526401625496,189.19408973421363,189.14932372451636,189.11660560691269,189.0945022428163,189.08187654668694,189.07785993293047,189.08180667023893,189.09323324265478,189.11174662199863,189.13696595120624,189.16844249152274,189.2055827800539,189.2475797731777,189.29335632497498,189.3415246907265,189.39036488844042,189.4378237414638,189.48153531564876,189.51886231363338,189.54695685705636,189.56283803442312,189.56348267320962,189.54592505793406,189.5073608000887,189.4452497981815,189.35741322086943,189.24211970373773,189.09815645741492,188.92488171510647,188.7222558633288,188.49084965301927,188.23182902452692,187.94691724019654,187.6383361420958,187.3087293816007,186.96107134874532,186.5985662169323,186.224541977411,185.84234454447474,185.45523695640753,185.06630838235097,184.67839708849368,184.29403074795874,183.91538653816198,183.54427240682935,183.18212975918289,182.8300566832836,182.4888497476797,182.1590614320039,181.8410694378551,181.53515351655471,181.24157507389452,180.96065468915842,180.69284282255035,180.43877937410943,180.1993383774311,179.9756549301232,179.76913243659624,179.58142931611917,179.4144254532505,179.2701697795468,179.15081141630768,179.0585177232276,178.99538333900313,178.96333482812872,178.9640358354153,178.9987976809199,179.06850010112694,179.17352636902058,179.3137163306625,179.48834001516005,179.69609345428694,179.93511724060352,180.20303721632752,180.49702557852564,180.81387966693853,181.15011482181112,181.5020670060194,181.86600041435378,182.23821506724715,182.6151494178623,182.99347328820244,183.37016697638012,183.74258311549536,184.10848877582976,184.46608633838642,184.81401277486293,185.151318088993,185.4774247483147,185.7920709077855,186.09524104665084,186.3870882650206,186.66785288443145,186.93778214693467,187.19705570261985,187.4457212219995,187.68364388642226,187.91047272775631,188.12562584948515,188.32829551477863,188.51747298832805,188.69199192574206,188.85058807471498,188.99197214026208,189.11491192011607,189.21831927523368,189.30133719283336,189.3634221417771,189.40441711562642,189.42461119727932,189.4247821379511,189.40621928808025,189.37072520449567,189.32059533542588,189.25857629651162,189.18780433896765,189.11172661875386,189.0340087506071,188.958432827556,188.88879056892642,188.8287765028937,188.78188608090838,188.75132336151285,188.7399224039854,188.75008580405068,188.7837429217414,188.84232934155168,188.92678802007865,189.0375914730323,189.17478328929815,189.338036290037,189.526723825942,189.74000006843733,189.9768847336094,190.23634750225006,190.51738747411207,190.81910331478474,191.14075030207204,191.48178122606492,191.84186900351182,192.22090988460127,192.6190072050948,193.0364367121641,193.47359551160346,193.93093759360494,194.40889964612532,194.9078214194517,195.42786523370506,195.96893930539522,196.53062940526155,197.11214295600908,197.71226905626293,198.32935710864294,198.96131577757865,199.6056329561326,200.25941633537792,200.91945310153523,201.58228629092764,202.24430446283932,202.90184065097503,203.551276061896,204.189143729205,204.81222731885123,205.41765051435593,206.00295287876267,206.56614876821052,207.10576672479,207.62086775927148,208.11104199615147,208.5763842389455,209.0174500663077,209.43519503470088,209.83090039096032,210.20608934541025,210.5624383899313,210.9016883434558,211.22555976008255,211.5356770454705,211.83350511120426,212.12030168204194,212.397087495356,212.66463564202036,212.9234802457738,213.17394361917724,213.41618002443488,213.65023325967928,213.87610453351817,214.09382652248405,214.30353915730385,214.50556257286408,214.70046278914953,214.88910605926574,215.07269840650804,215.25280764452697,215.43136609242805,215.61065321197736,215.79325845371358,215.98202564714023,216.17998125234075,216.39024965486007,216.61595938723687,216.86014466239325,217.12564687989206,217.41502080137735,217.73044988440867,218.07367482513638,218.44593871268393,218.8479513754247,219.27987454440336,219.74132842165776,220.23141917510807,220.7487858423953,221.2916641675141,221.85796406595364,222.44535675913215,223.05136717087737,223.6734669604127,224.30916358857255,224.95608107501636,225.61202858979576,226.2750537067903,226.9434779929567,227.61591357129748,228.29126032612328,228.96868446232833,229.64758013120903,230.32751674132265,231.0081753365433,231.68927800487896,232.3705146504382,233.05147159802362,233.73156639821497,234.40999286604745,235.08567983627742,235.75726638183198,236.4230953581908,237.08122615167986,237.72946647577726,238.36542203089763,238.98656187378722,239.59029648356704,240.1740648077377,240.73542606010724,241.27215175052518,241.78231336893865,242.26436132677756,242.71719116774048,243.14019367646293,243.53328630526062,243.89692426495714,244.23209063757363,244.5402659137469,244.82337838184537,245.08373774568372,245.3239551742436,245.54685364711918,245.75537291947788,245.95247366659385,246.14104536863408,246.32382226181198,246.50331122503002,246.6817348160442,246.8609918526651,247.04263699570154,247.22787978062163,247.41760251724514,247.61239548504662,247.81260694772175,248.01840474175864,248.2298455999335,248.44694798273923,248.6697640287616,248.8984463070618,249.133305356551,249.37485451293534,249.62383922566838,249.88124891831785,250.14831040089322,250.4264628516149,250.71731539517603,251.02258926158947,251.34404736384926,251.68341483890103,252.04229461762876,252.42208239823262,252.82388547716388,253.2484497383447,253.6960987225892,254.1666881146078,254.65957822511518,255.173626149927,255.70719830315625,256.2582029989697,256.8241417490613,257.4021770034251,257.9892132389783,258.581987636796,259.17716611851375,259.7714402600581,260.3616205790956,260.9447219022347,261.51803694777,262.0791948868501,262.62620243734636,263.15746595881507,263.67179400586684,264.1683808092212,264.64677213585526,265.10681588070764,265.5486005154638,265.97238512487166,266.37852516654374,266.7673982754941,267.13933439051374,267.49455420871817,267.8331194918902,268.1548980793699,268.4595456427286,268.74650529067384,269.0150251475282,269.2641930371782,269.49298645885574,269.70033519130686,269.88519315257525,270.04661561113693,270.18383751848694,270.29634863070305,270.3839612125039,270.4468664654165,270.485676373771,270.5014483895639,270.49569124186263,270.4703511132266,270.42777842470986,270.3706764602631,270.30203398921583,270.22504486364204,270.14301823322893,270.05928349916456,269.97709439572463,269.89953662990774,269.82944332368317,269.76932210005396,269.7212970544555,269.6870680884553,269.66788919331776,269.66456630324024,269.6774743424712,269.7065921189783,269.7515528204138,269.81170709242076,269.8861950650017,269.9740232714331,270.07414219756635,270.1855202175892,270.30720991390774,270.4384032305323,270.5784725470143,270.7269955496677,270.8837626766898,271.04876687630383,271.22217639140496,271.4042922190373,271.59549273914985,271.7961687199364,272.0066524491752,272.2271450838282,272.45764643570914,272.6978913126531,272.9472962174284,273.204919687223,273.4694388620776,273.7391440374714,274.01195202798715,274.28543819413767,274.5568860140188,274.82335216617844,275.08174427780034,275.32890782541375,275.5617181886689,275.777173576649,275.97248448534106,276.1451555073963,276.2930556926277,276.4144742300822,276.50815896008197,276.5733360888617,276.6097104242699,276.6174464291887,276.59713134895236,276.5497225602517,276.47648206557545,276.378901678982,276.25862288390385,276.11735556952874,275.95679985784477,275.7785750189441,275.5841590494599,275.3748418807272,275.1516944217584,274.9155547677394,274.6670319636546,274.4065267544509,274.1342678282973,273.85036121660204,273.5548497978005,273.24777929910795,272.92926683042066,272.59956783616104,272.259137421813,271.9086822983617,271.5492000747445,271.1820032901068,270.80872637938813,270.4313146654337,270.0519954212492,269.673231997316,269.297662911213,268.92802860283007,268.56708922597613,268.2175373411969,267.88190966922434,267.56250214422425,267.26129236665076,266.9798732042891,266.7194007452348,266.4805590965032,266.2635436836621,266.06806378397425,265.89336406605037,265.73826396281885,265.60121282086305,265.4803579936731,265.37362241922574,265.27878777586767,265.19357906756807,265.1157464624141,265.04314039764495,264.973776360122,264.90588633174497,264.8379546241066,264.7687366767991,264.6972603144237,264.62280990006883,264.5448947382946,264.4632039204734,264.3775505257015,264.28780865374245,264.1938471432033,264.0954639988178,263.9923255075709,263.88391376699525,263.7694858939711,263.6480475528547,263.5183426709587,263.378860337929,263.2278589592614,263.063406801806,262.88343717961595,262.6858157286694,262.4684165508509,262.2292035062381,261.96631262458686,261.67813150818864,261.36337171429204,261.0211304300939,260.65093827007655,260.25279070731807,259.82716146185703,259.37499706802413,258.8976927820961,258.3970509228335,257.8752236120511,257.3346426549605,256.7779399305871,256.2078621185543,255.62718384650796,255.03862338938515,254.44476488558186,253.84799066495236,253.25042672922254,252.65390371615806,252.05993485208117,251.46971149654894,250.8841159554651,250.3037503331962,249.72897935778298,249.15998438976231,248.5968252521524,248.0395061265621,247.4880415684335,246.94251871283367,246.40315196965625,245.87032693129729,245.3446308142338,244.82686749676196,244.31805605928903,243.81941263648557,243.33231630470254,242.85826060520097,242.39879309775017,241.95544600823615,241.5296615428524,241.12271576340737,240.73564503626042,240.36917897506754,240.0236834995641,239.69911714448187,239.39500309977933,239.11041867965398,238.84400304412128,238.59398307879243,238.3582164236616,238.13424977782822,237.91938983904197,237.7107836046445,237.5055042966589,237.30063890238065,237.0933732567374,236.8810707369994,236.66134098631017,236.43209561162217,236.19158848608095,235.9384390897514,235.67163820367904,235.39053618452007,235.09481494268715,234.78444557941106,234.459634363967,234.12076031384964,233.7683080478119,233.4027997934908,233.0247304371767,232.63450930328406,232.23241195619906,231.8185447485432,231.39282412773184,230.95497189473625,230.5045267284035,230.0408713920134,229.5632741732268,229.0709423197993,228.56308456293252,228.0389793032192,227.49804469803368,226.93990675166296,226.36446157778144,225.77192827457213,225.1628893118139,224.5383159526569,223.8995769880634,223.24842990958678,222.5869945420471,221.91771005500058,221.24327712388416,220.56658777393685,219.8906460727513,219.21848330767855,218.55307156787273,217.8972397328708,217.25359574612744,216.6244587294739,216.01180399001817,215.41722331101812,214.84190213727638,214.28661440445111,213.75173486528908,213.2372678809025,212.74289081787762,212.26800946521044,211.81182229647285,211.37338998260552,210.95170633082532,210.5457667968686,210.1546308921332,209.77747517459176,209.41363405289653,209.06262631809867,208.72416611055712,208.39815788897673,208.08467584874572,207.7839290914453,207.49621463157195,207.22186099889768,206.96116572012775,206.71433031400358,206.48139659131306,206.26218800714395,206.05625956943712,205.86285937820813,205.6809042762039,205.5089713653344,205.34530632199392,205.18784857134565,205.03427250146848,204.8820430589394,204.72848331184048,204.5708509338218,204.4064200867646,204.23256488462133,204.0468405225473,203.84705825859845,203.63135073462843,203.39822460293087,203.14659806057165,202.87582165109123,202.58568153382197,202.27638530067944,201.9485312931747,201.6030631931981,201.24121238725678,200.86443119772434,200.4743205056623,200.0725555358267,199.6608136232129,199.2407076298786,198.81372833920946,198.38119864062637,197.94424165834354,197.503764208113,197.06045612658184,196.61480515319946,196.16712620004995,195.7176030649529,195.26633996863524,194.81341976310173,194.35896529348094,193.9032002186575,193.44650561611655,192.98946891246644,192.53292208163893,192.0779666166906,191.62598347878367,191.17862702165664,190.73780273997147,190.30562955073654,189.88438814339074,189.47645768275646,189.0842437806682,188.7101011335431,188.35625452921974,188.0247220409821,187.71724414373912,187.43522221111715,187.1796693970894,186.9511762954117,186.74989303632762,186.57552866095463,186.42736775207467,186.30430344053636,186.20488509377378,186.1273782692135,186.06983391804096,186.03016338474745,186.00621548764755,185.99585189829727,185.99701716650617,186.0078000552327,186.02648333903412,186.05157985490948,186.08185334174158,186.11632342471304,186.15425495089178,186.19513271701643,186.23862340658118,186.28452722997835,186.33272230333094,186.38310518047837,186.4355311486838,186.48975790194694,186.5453960163819,186.60186928012342,186.65838739539083,186.71393290092936,186.76726339436604,186.81692930654526,186.86130663719774,186.89864324766532,186.9271165646341,186.94489991791716,186.9502342484488,186.9415016054276,186.91729672070738,186.87649301105813,186.81829961165892,186.7423064740555,186.6485151464772,186.53735356343867,186.40967396806096,186.26673393307706,186.1101612912283,185.94190458879595,185.76417139566047,185.57935740422877,185.3899696968533,185.1985478339568,185.0075864991005,184.81946332866107,184.6363752586553,184.4602862553371,184.2928886841503,184.1355798457177,183.9894544059382,183.8553126120125,183.73368336107242,183.62486041620608,183.52894938635103,183.44592253710758,183.3756781075417,183.3181005937464,183.27311843421666,183.24075569596837,183.22117470485972,183.21470707034527,183.22187119701806,183.24337511880046,183.28010429727343,183.3330948509018,183.4034934833338,183.49250611432268,183.60133784741797,183.7311274011985,183.8828794595134,184.05739854360266,184.25522796723308,184.47659720716237,184.7213806171816,184.98906985569752,185.27876171337243,185.58916225435306,185.91860736235589,186.26509895417328,186.62635533116125,186.99987342517213,187.38300009631627,187.77300918695087,188.16718075284044,188.56287879306893,188.95762388995794,189.3491574440301,189.73549463242972,190.11496380881002,190.48623076756465,190.84830707847772,191.20054251842598,191.54260244182052,191.87443169851937,192.1962073870364,192.5082832867533,192.81112921673156,193.1052687998548,193.3912191577147,193.66943592186522,193.9402666285736,194.20391508409924,194.4604186713985,194.70963984964592,194.951272313128,195.184861467337,195.40983809011632,195.62556331610463,195.83138245191918,196.02668463071743,196.21096497424523,196.38388576617035,196.54533316110937,196.69546615828924,196.83475494660618,196.96400625924215,197.0843740332319,197.19735441811312,197.30476497893918,197.4087087506596,197.51152458080696,197.61572590488132,197.723930696913,197.83878579525134,197.9628890968766,198.09871322725715,198.24853422114947,198.41436849661713,198.597920983245,198.80054669813015,199.02322737941952,199.26656402265726,199.53078535989258,199.81577151734092,200.12109132624127,200.44605108294311,200.78975199336188,201.1511531227243,201.5291364252352,201.9225703623718,202.33036873593463,202.75154165612915,203.18523601996876,203.6307634669046,204.0876144754226,204.55545802954785,205.03412707736499,205.5235907830024,206.02391529799814,206.53521540946122,207.05759992776953,207.59111402915872,208.1356819499462,208.69105342978818,209.2567571212535,209.8320638315746,210.41596195813176,211.00714684847733,211.60402509156916,212.20473396728812,212.80717548708876,213.40906369135502,214.00798316885928,214.6014561672837,215.18701520225613,215.7622777699054,216.32501964036624,216.87324326362543,217.40523805178773,217.9196297011027,218.41541626210818,218.89198932859162,219.34913946123154,219.78704575095617,220.20625021918733,220.60761850606332,220.99228897433014,221.361612920832,221.71708901029857,222.0602953057322,222.39282235301243,222.7162106804577,223.03189580254522,223.34116338543342,223.64511666336443,223.94465751957497,224.24048189869728,224.533089439146,224.82280644463077,225.10982059414013,225.39422515729228,225.67606996999615,225.95541606070395,226.23239061920637,226.50723897821382,226.78037043390248,227.0523950568734,227.32414912240372,227.59670739329943,227.87138118829952,228.14970192705124,228.43339061935808,228.72431452091098,229.0244328704484,229.33573421789254,229.66016831826792,229.99957587769936,230.35561957912208,230.72971977905092,231.12299805464983,231.53623140341907,231.96981937567543,232.4237657798063,232.89767587535877,233.39076919708373,233.90190737424157,234.42963556424556,234.97223544674267,235.52778715836945,236.09423711889883,236.66946842839113,237.25137041605672,237.8379039996035,238.4271597643931,239.01740608129597,239.60712512893613,240.1950353411495,240.78009952941363,241.3615186944065,241.93871230058764,242.5112865029393,243.07899244832674,243.64167729224354,244.199230947989,244.75153180004594,245.29839465555477,245.83952407534187,246.3744759261462,246.90262954446354,247.42317232368114,247.93509786047446,248.4372180594067,248.9281888351336,249.4065483095519,249.8707657156794,250.31929862708404,250.75065566248705,251.16346149444783,251.55652083572002,251.9288780951987,252.27986958652232,252.60916552648058,252.91679955900844,253.20318415779042,253.46911096422076,253.7157358710444,253.94454942622713,254.1573338667318,254.35610875971304,254.5430677945412,254.72050970371595,254.89076657171375,255.05613290374325,255.2187987656624,255.38079007541378,255.54391873760486,255.70974478687643,255.87955207016918,256.05433828630805,256.2348194507997,256.4214481036614,256.61444386768807,256.81383433087444,257.019503702747,257.2312463073185,257.4488217452507,257.6720084963289,257.9006528434095,258.1347102744724,258.37427694552326,258.619609341181,258.8711309224982,259.12942526840425,259.3952159603008,259.6693341896004,259.9526757469058,260.24614964332886,260.55062108797597,260.8668518759349,261.1954414107585,261.5367715854477,261.8909585761868,262.2578142723476,262.63681959176387,263.0271113367132,263.4274835643582,263.83640371114666,264.25204296211376,264.6723196321798,265.09495366508344,265.51752979095437,265.9375664450427,266.3525872605301,266.76019182258267,267.15812241526504,267.54432370510574,267.9169926737397,268.2746166177483,268.6159976500396,268.9402628315974,269.24685979886283,269.53553849196226,269.8063202938748,270.0594565237708,270.2953787562709,270.514643834925,270.7178766923821,270.90571416912405,271.0787529335077,271.2375043532845,271.38235876637543,271.5135610679365,271.6311988998314,271.7352040308991,271.8253667887325,271.90136268443587,271.96278969890045,272.00921410796985,272.0402222455656,272.0554752635788,272.0547637630211,272.03805915228236,272.00555873639973,271.95772184828746,271.895294782871,271.8193228644945,271.73114863704643,271.63239588053773,271.52493989001607,271.4108651644796,271.2924123076613,271.171916505063,271.0517403830622,270.9342043532967,270.8215176833755,270.71571350613783,270.6185907857525,270.53166591010785,270.45613609344286,270.3928561763895,270.34232973368603,270.30471467807604,270.2798428199313,270.26725214383794,270.2662299320489,270.27586433303895,270.2951015688778,270.322805718535,270.3578179185683,270.399011892505,270.44534295151703,270.49588798918967,270.54987450199087,270.6066972774515,270.66592207178155,270.7272763117938,270.79062756500247,270.855951189301,270.92328916459076,270.9927025921048,271.06422069756854,271.1377893734065,271.2132223325241,271.290157819853,271.3680235445152,271.4460120700314,271.5230683548241,271.5978904989444,271.66894405886745,271.73448957674265,271.79262227114185,271.84132218977464,271.87851256468485,271.9021236664528,271.910159149134,271.90076172804964,271.8722750459454,271.823298758794,271.7527342014781,271.6598184586208,271.5441452426969,271.4056716402156,271.24471049291975,271.06190889787393,270.8582140007411,270.63482788465933,270.3931538903746,270.1347371141283,269.8612020971209,269.5741908305839,269.27530414822905,268.96604936623555,268.6477966714447,268.32174627022306,267.9889077194102,267.65009219809383,267.3059177795751,266.9568270635859,266.603115866323,266.2449710749211,265.88251528455066,265.51585547592396,265.14513277742685,264.77057030041885,264.39251614124447,264.011478903504,263.6281534953677,263.24343547798435,262.85842285487655,262.4744048663193,262.09283805161306,261.71531052916043,261.34349608323294,260.9791002037163,260.6238006716564,260.27918559563744,259.94669196546886,259.6275477918087,259.3227207429949,259.0328758816384,258.75834465947617,258.49910677270657,258.2547858404705,258.02465917958654,257.80768124495944,257.60251962383575,257.40760184865184,257.22117075992725,257.04134573508907,256.86618682267965,256.69375869822505,256.52219139423096,256.34973495010775,256.17480546787993,255.99602052823633,255.81222249406156,255.62248887489469,255.4261296118258,255.2226718321719,255.0118332807369,254.79348622517404,254.56761412609907,254.33426373246812,254.09349549015326,253.84533522493564,253.58972997668982,253.32651062404503,253.05536356097952,252.77581318868928,252.48721639402044,252.1887695310153,251.87952773921543,251.5584357572582,251.22436875854444,250.87618118058245,250.51276107025447,250.13308714736198,249.7362856152519,249.32168372931062,248.88885727268058,248.43766937656704,247.96829854499842,247.48125427889482,246.97737931392587,246.4578381586334,245.924092308662,245.37786318380418,244.82108445203616,244.25584593671735,243.68433172205087,243.1087553560957,242.53129518545452,241.95403283443076,241.37889766530134,240.80761973480833,240.2416933120403,239.6823524682303,239.13055961875804,238.58700722483178,238.05213218213544,237.52614177168564,237.00904945825414,236.50071832466577,236.000909551876,235.50933311450078,235.02569777169435,234.54975749832045,234.0813517174985,233.62043705116858,233.16710878135592,232.72161078607837,232.2843333501132,231.85579891852964,231.4366365247895,231.02754625033916,230.62925562604198,230.24247033839356,229.86782193139325,229.50581538093238,229.15677945277343,228.82082263576098,228.4977971750028,228.18727332929038,227.8885254638792,227.6005309906986,227.3219825143743,227.05131286823612,226.78673206491115,226.52627457557062,226.26785482220535,226.0093283460337,225.7485558241827,225.4834669612108,225.21212128933334,224.93276307086236,224.64386779992807,224.344178232048,224.0327284069398,223.70885474404852,223.37219394949574,223.02266814316218,222.66045826047633,222.28596737160004,221.89977606054995,221.50259239249684,221.09519924918456,220.67840191728274,220.25297876740956,219.81963766538234,219.37898042260946,218.9314771373171,218.47745172668175,218.01707933110663,217.55039561852504,217.07731736318405,216.59767305427835,216.1112417376478,215.61779783756094,215.11715936935946,214.60923675510563,214.0940794037925,213.57191731781043,213.04319523297843,212.508597177678,211.9690598276679,211.42577361129628,210.88017115480218,210.3339033159818,209.78880370223465,209.24684317202454,208.7100763455725,208.18058257361386,207.66040411021166,207.15148439142217,206.65560932793642,206.1743543763785,205.70903986823646,205.260696662375,204.8300436682929,204.41747818981293,204.02307939386827,203.64662455006953,203.28761704808016,202.9453246146697,202.61882565099276,202.30706121907363,202.0088899448107,201.72314298638247,201.44867624755886,201.18441719312455,200.92940393873232,200.68281472318543,200.44398640426942,200.21242122161905,199.9877817099386,199.7698742894889,199.55862267426602,199.35403278948996,199.15615134997807,198.96502059558577,198.78063189113158,198.60288096506514,198.4315274801546,198.26616140498913,198.106178298962,197.95076515436736,197.7988978819635,197.64935091009875,197.5007187246123,197.3514485409394,197.19988270451373,197.04430889189305,196.88301576047934,196.71435139153547,196.53678170530782,196.34894600717743,196.14970695095923,195.9381924731557,195.71382764620967,195.47635489906068,195.22584163353386,194.96267489507287,194.6875434032879,194.401407878094,194.10546117812868,193.80108026954667,193.4897724392582,193.17311843680835,192.85271535959095,192.53012208045854,192.20680985618398,191.88412045837478,191.56323375116864,191.2451461244558,190.93066060469073,190.62038883846165,190.31476450983183,190.0140671443298,189.71845470221157,189.4280029000394,189.142748846927,188.8627363583506,188.588060227789,188.31890679879677,188.05558838399133,187.79856941198577,187.54848263081317,187.30613423324604,187.0724973673281,186.84869412292662,186.635966709182,186.4356391255581,186.24907114989355,186.07760689304845,185.92252047902932,185.78496158558002,185.66590361384607,185.56609714501315,185.48603109249393,185.42590358324267,185.38560412062873,185.36470801904903,185.3624834863793,185.37791109641796,185.40971477264026,185.45640282912274,185.5163171142944,185.5876879041321,185.66869191416478,185.75751065840745,185.85238638524118,185.95167296491613,186.05387938297667,186.15770389363044,186.26205738581046,186.36607508615955,186.46911633739617,186.57075281530115,186.67074615019388,186.76901646775767,186.86560383078807,186.96062492359252,187.05422755569288,187.14654565909876,187.23765740885452,187.3275489124106,187.41608559953883,187.50299301769485,187.5878482205376,187.67008235706726,187.74899445614955,187.8237757883385,187.8935436061643,187.95738254598,188.01439154632862,188.06373382212257,188.10468724736234,188.13669245150388,188.15939602850014,188.17268648814786,188.17672093468204,188.1719409190641,188.15907645505033,188.13913778656394,188.1133951139847,188.08334709740546,188.05067952389956,188.0172160237915,187.984863121952,187.95555219345323,187.93118104388373,187.91355784552817,187.9043500310618,187.90504048333284,187.9168929771037,187.9409283464151,187.9779122948178,188.0283551642224,188.09252336298243,188.1704615571773,188.26202418198486,188.36691436099355,188.48472795463573,188.61500021356636,188.75725240132863,188.91103577857757,189.07597050669804,189.25177732287352,189.43830024582368,189.63551906991884,189.84355096882194,190.06264112845295,190.29314293134865,190.53548878867673,190.79015323203183,191.0576103075505,191.33828763719302,191.63251970929602,191.94050302231213,192.26225562873847,192.5975834145838,192.94605511437894,193.30698762022178,193.67944261873663,194.0622350095325,194.45395295308577,194.85298879658012,195.2575795644028,195.66585520503747,196.07589238381294,196.4857713223178,196.89363302527318,197.2977342125518,197.69649738895626,198.08855373125175,198.4727768378145,198.84830585195365,199.2145570111378,199.5712232627479,199.91826219189397,200.25587309675942,200.58446459118704,200.9046145845319,201.2170248608268,201.52247273385115,201.8217623787538,202.1156784282866,202.4049442734517,202.6901872325382,202.9719123638866,203.25048621690826,203.52613126835527,203.79893120548402,204.0688466254385,204.33574015207762,204.59940945745808,204.85962624226957,205.11617890042675,205.3689163850148,205.61779071676207,205.86289563613036,206.1044990928191,206.3430675818337,206.57928075673323,206.81403525613524,207.04843724252876,207.28378374343086,207.5215334729669,207.76326836627766,208.01064755074384,208.2653558811587,208.52904945958826,208.80330072939222,209.08954576792306,209.3890363019522,209.70279873914222,210.03160215994947,210.37593676540928,210.73600375075011,211.11171700009353,211.50271640355624,211.9083920157884,212.32791773461466,212.76029270803235,213.2043883017109,213.65899819671628,214.12288905188308,214.59484916383374,215.07373268985904,215.5584972574936,216.04823315581692,216.54218276758343,217.03974943370457,217.54049551429821,218.044129993347,218.55048653633065,219.05949342251063,219.5711372089855,220.08542231941092,220.60232896884773,221.12177192618853,221.6435625723928,222.16737653870652,222.69272891337903,223.2189586037212,223.74522295382013,224.27050317242288,224.7936205491043,225.31326286010537,225.82801981844494,226.3364259347532,226.83700875136316,227.32834011371318,227.8090879657239,228.27806610884085,228.73427945038202,229.1769624812444,229.60560905475333,230.01999197019325,230.42017137385145,230.80649155070196,231.17956626199106,231.540253357666,231.88961992800253,232.22889972814048,232.5594449882186,232.88267499084304,233.2000239429229,233.51289068313545,233.82259264897226,234.13032628486832,234.43713571825398,234.74389108227817,235.05127734605546,235.35979395284505,235.66976499300267,235.98135908194152,236.29461760283223,236.60948953606194,236.9258707553891,237.24364544207896,237.56272716490832,237.88309720103874,238.20483782877398,238.5281585997819,238.85341398079927,239.18111122294022,239.51190784580658,239.84659868578132,240.18609302359408,240.5313828457266,240.88350377911604,241.24349064343167,241.6123298686673,241.99091121194581,242.3799812664649,242.78010118426826,243.19161083666904,243.6146013219706,244.04889731598055,244.49405026828768,244.94934290202625,245.41380490529852,245.88623913821922,246.3652571500453,246.8493223335811,247.33679866325892,247.8260026886975,248.31525630129832,248.80293776539503,249.28752860879334,249.76765419461407,250.24211613513404,250.70991514092998,251.1702634023242,251.6225861484697,252.0665125936766,252.5018570312493,252.92859134341595,253.3468106353828,253.75669404927635,254.15846305172064,254.55233960478245,254.9385066181566,255.31707294162194,255.68804489848992,256.05130599675977,256.40660600428794,256.75356006127515,257.0916579548445,257.42028312532614,257.73874044131264,258.04629129862013,258.3421941923361,258.6257486024979,258.89633983886785,259.1534824190234,259.3968596103522,259.6263569476221,259.8420878341716,260.0444097309483,260.2339299129689,260.41150030244125,260.5782014441654,260.7353162428748,260.88429460502067,261.0267105918964,261.1642140727772,261.2984791460808,261.4311517589611,261.56379899287344,261.69786239268683,261.83461750469746,261.9751414656152,262.1202900672438,262.2706852320371,262.4267132986039,262.58853396157747,262.7560991660098,262.9291807509626,263.10740519654115,263.2902934761207,263.47730376916127,263.66787466245427,263.8614664652937,264.05759838676624,264.2558795644078,264.45603227998123,264.65790613169486,264.86148242985973,265.06686861850307,265.2742830705858,265.48403113051245,265.69647375689635,265.9119905258544,266.1309390690587,266.35361322486693,266.58020226437947,266.8107535128331,267.04514052250477,267.28303867485016,267.5239097115294,267.76699623604736,268.01132671392645,268.255730956507,268.49886553000795,268.73924801579017,268.97529858677126,269.2053869827174,269.4278826836085,269.6412059102131,269.8438770332456,270.0345620494415,270.2121119805123,270.3755943588429,270.52431536586,270.6578316642069,270.7759514883863,270.87872510308,270.9664252755348,271.03951891026657,271.0986314348123,271.14450588151067,271.1779588639238,271.19983578452144,271.2109676256787,271.2121315683816,271.2040174581973,271.1872018080955,271.16213061012945,271.1291117448959,271.0883172545776,271.0397952097962,270.9834903806373,270.9192724453735,270.8469700616365,270.76640880551554,270.6774507710053,270.580033526535,270.47420615194966,270.3601602268452,270.2382539017985,270.10902754407334,269.9732098899196,269.83171413349504,269.68562391162203,269.53616967630234,269.38469645525026,269.23262445804124,269.08140436787164,268.9324694463749,268.787186756403,268.6468098658834,268.51243533175426,268.3849650799227,268.2650765047367,268.15320172488674,268.0495169721201,267.9539425786872,267.8661534956323,267.7855997446928,267.71153570920706,267.6430567298826,267.57914111232117,267.51869539338094,267.4606005661148,267.403756935606,267.3471253716966,267.2897629342077,267.2308511605653,267.16971570810085,267.105836512424,267.0388481344561,266.96853049506325,266.89479071010743,266.8176372131726,266.7371477633795,266.6534332601144,266.5665995082356,266.4767091848504,266.3837462465474,266.2875848847692,266.18796489381634,266.0844749736379,265.9765450661359,265.86344834147184,265.7443129351722,265.61814301477096,265.4838482537234,265.3402803366561,265.18627473750433,265.0206957207806,264.8424823314129,264.65069306998817,264.4445470014622,264.2234592137977,263.9870688198144,263.7352580663278,263.4681615601053,263.1861651170427,262.88989426338895,262.58019293860406,262.2580934414477,261.9247790985873,261.58154149568156,261.22973437588485,260.8707264663743,260.50585553192064,260.136385873908,259.7634712979993,259.38812527425466,259.0111996257834,258.63337262638936,258.2551468881146,257.8768569023147,257.498685590035,257.120688745572,256.7428258458639,256.3649953692074,255.98707253687155,255.60894727236442,255.2305601714828,254.85193439174043,254.47320159610928,254.09462041118923,253.71658626674105,253.33963195050993,252.96441871482767,252.59171828316047,252.22238659865508,251.8573306068716,251.4974697476199,251.143694126036,250.79682152522827,250.45755550188161,250.1264467679501,249.80385990795233,249.48994722040456,249.18463111719322,248.8875960846992,248.5982907276731,248.3159399063967,248.03956646615484,247.76802157212836,247.50002222778713,247.23419419332052,246.96911825134643,246.7033776040048,246.43560413693254,246.16452135390807,245.88898196741212,245.60799841534174,245.32076494760346,245.02667036866654,244.72530101016682,244.41643401576212,244.10002152207085,243.77616678847934,243.44509374043176,243.10711172408656,242.7625775076533,242.41185669406642,242.0552867243417,241.69314355044318,241.32561384624555,240.9527743164385,240.57457927236052,240.19085719143087,239.80131648694044,239.40556021326825,239.00310894452508,238.59343061780157,238.17597574910596,237.75021613084712,237.3156849200743,236.87201593711916,236.41898001952143,235.95651641485955,235.4847574411145,235.0040449816714,234.5149377963218,234.01820909803166,233.51483434332104,233.00596968585666,232.49292202211117,231.97711198969645,231.4600316407449,230.94319878566557,230.4281101725748,229.91619572590966,229.40877601124043,228.90702492516058,228.4119393382258,227.9243170594771,227.4447440619568,226.97359143227098,226.51102200856576,226.0570061762999,225.6113458257138,225.17370506314816,224.74364593173036,224.32066715285433,223.90424376088728,223.4938654766916,223.08907175201284,222.68948161178753,222.29481671452444,221.90491642653242,221.51974414403816,221.13938457502275,220.76403218472785,220.3939714893412,220.02955032609665,219.67114761160252,219.31913740350834,218.97385128771367,218.63554121336588,218.30434488563913,217.98025570239267,217.66309899177023,217.35251598563013,217.04795656527014,216.74868136208943,216.4537733103729,216.16215825768214,215.87263376611148,215.58390480959412,215.29462471088593,215.00343938568335,214.70903278494552,214.41017135906432,214.10574541253814,213.79480537285517,213.47659125424238,213.15055394232414,212.8163673415065,212.47393089133323,212.1233624469184,211.76498200636075,211.39928722938907,211.02692210239263,210.64864044413434,210.26526619628956,209.87765259064813,209.4866423228698,209.09303078934727,208.6975342630864,208.30076460624898,207.903211755906,207.5052347947201,207.10706195212606,206.70879939865614,206.31044822176128,205.91192853071698,205.5131092541044,205.11384188605442,204.7139962230148,204.31349602245194,203.9123525143618,203.51069380570607,203.10878843113034,202.70706160932681,202.30610314708713,201.90666637235995,201.50965795043058,201.11611891888637,200.72719774208187,200.3441166100086,199.96813256759484,199.60049533960753,199.24240389889462,198.89496390214723,198.55914808367964,198.2357615556843,197.92541372044346,197.62849816875885,197.345181536686,197.0754018405147,196.81887633139308,196.57511843097936,196.3434628530059,196.1230976063792,195.91310123465257,195.71248339202353,195.5202267005427,195.33532778488515,195.1568354421312,194.9838840711178,194.81572075009174,194.6517246986088,194.4914182715318,194.33446908788548,194.18068337130353,194.02999104696832,193.88242357774155,193.73808590669373,193.5971241843503,193.45969118044718,193.32591140031315,193.1958479389591,193.069473010992,192.94664389666704,192.8270857542253,192.7103823814425,192.59597558437784,192.48317335100606,192.37116655593408,192.25905346449957,192.14587088426367,192.0306304512903,191.91235825657083,191.79013582929343,191.66314040827857,191.53068245528667,191.39223849291568,191.24747757871117,191.0962800439788,190.9387475140813,190.77520366647417,190.6061856504666,190.432426564327,190.254829836321,190.0744367631305,189.89238880071758,189.70988646137573,189.5281468330964,189.3483617951659,189.17165895426405,188.9990671710412,188.831488296448,188.66967640325686,188.51422539883964,188.3655654613575,188.22396827637243,188.08956058888398,187.96234515110982,187.84222776168062,187.72904877737082,187.62261725040767,187.5227457146081,187.4292836187359,187.34214748670004,187.26134606711136,187.18699900953806,187.11934795700282,187.058759355253,187.00571872745041,186.96081662458496,186.92472691283604,186.89817847558712,186.88192176796684,186.87669194672333,186.88317049298004,186.90194733973414,186.93348550487872,186.97809011460484,187.0358834873114,187.10678764582025,187.19051525154353,187.28656952780625,187.39425328303923,187.51268668183403,187.64083296700724,187.7775309320241,187.92153260138576,188.0715443146365,188.2262692410617,188.384449285601,188.54490438533145,188.70656733791841,188.86851254113154,189.02997734333525,189.19037509192952,189.34929939971838,189.5065196054524,189.66196786045293,189.81571870448653,189.9679623784218,190.1189734386644,190.26907647245042,190.41861085169265,190.56789649914228,190.7172025724539,190.8667208030909,191.01654496690423,191.16665762547075,191.31692487974735,191.46709944125743,191.6168318738696,191.76568941504934,191.91318137261888,192.058789733249,192.20200333102326,192.34235372358137,192.47945082010594,192.61301630519216,192.74291300545866,192.86916854624687,192.991991933436,193.111782055227,193.22912751191188,193.34479762655823,193.45972494295535,193.5749799554224,193.69173921529975,193.8112483002686,193.93478139729962,194.06359942425655,194.19890868996106,194.34182206403173,194.4933244976234,194.65424451119267,194.82523295743258,195.00674999269887,195.19906076821383,195.4022399051445,195.6161843686572,195.84063392866395,196.07519801145347,196.319387426663,196.57264921467439,196.83440271283885,197.1040748924354,197.38113307416197,197.66511328511066,197.95564276625643,198.25245546340733,198.55539971918225,198.864437808594,199.1796374038292,199.50115549151377,199.82921567512818,200.16408015478208,200.50601796740858,200.85527127733613,201.21202161930634,201.576358007463,201.94824873406003,202.3275184953458,202.7138322090549,203.10668654249164,203.50540977042033,203.9091701488233,204.316992546719,204.72778264687116,205.14035762992262,205.5534819159638,205.96590627057824,206.37640840292406,206.78383310075898,207.1871299657783,207.5853869312046,207.9778579560194,208.36398358542388,208.74340342962384,209.1159600239095,209.48169397079928,209.84083070690656,210.19375966015667,210.54100694502202,210.883203064727,211.22104733326054,211.55527088395888,211.88660018753163,212.21572295798114,212.54325818244044,212.86973177819598,213.19555906944007,213.52103490405787,213.84633181663503,214.17150620967854,214.4965120936118,214.82122152008608,215.1454504838497,215.4689887746591,215.7916320478729,216.11321426155695,216.43363860542678,216.7529051238812,217.07113340764633,217.38857898706723,217.7056423912495,218.0228702236716,218.34094802626842,218.66068513823237,218.98299218011522,219.30885218600991,219.63928674610588,219.97531879096144,220.3179338333512,220.66804157386298,221.02643976782474,221.3937821440825,221.77055196622487,222.15704254450148,222.55334565665848,222.95934843647487,223.3747388607654,223.79901953132452,224.23152903032815,224.671469747828,225.11794075788518,225.56997407227917,226.02657244037064,226.4867467987838,226.94955150831365,227.41411564589157,227.87966883955048,228.34556043242839,228.81127112218482,229.27641662578932,229.74074334525545,230.2041164351871,230.66650107567798,231.12793811307236,231.58851552758566,232.04833740532314,232.50749222145384,232.96602227451692,233.4238960474218,233.88098511213653,234.33704695074795,234.7917147484373,235.2444948406426,235.69477208692535,236.14182301935557,236.58483619588839,237.02293880113444,237.45522819849072,237.8808068665518,238.29881896328,238.70848666351583,239.1091444142003,239.50026934723948,239.88150627729348,240.25268598113826,240.61383579257736,240.96518193456873,241.30714342806405,241.64031784325368,241.96545957127356,242.28345167150087,242.59527267175255,242.90195994929036,243.20457148657866,243.50414786870914,243.8016763656484,244.09805882338148,244.3940848799992,244.69041173682837,244.98755136590844,245.2858656418229,245.5855694687323,245.88674155430434,246.18934208301997,246.49323618292286,246.79822178097032,247.1040602182924,247.41050785948315,247.71734688658333,248.02441352080027,248.33162206035806,248.63898335341614,248.9466166285176,249.25475396549948,249.56373708820814,249.87400657574338,250.18608399944955,250.50054787690775,250.81800467158402,251.13905633983418,251.4642661213188,251.794124374338,252.12901626860543,252.46919306394713,252.81474852870969,253.16560179541125,253.52148762674975,253.8819546894292,254.24637202598942,254.61394349716943,254.98372956108437,255.35467538180964,255.72564393830132,256.09545255165307,256.4629110776104,256.8268599308664,257.18620612202506,257.53995559637815,257.8872403600265,258.2273391530148,258.5596907666325,258.88389948515396,259.19973254097374,259.5071098849751,259.8060869694673,260.09683159852614,260.3795962014512,260.6546871136908,260.9224325941606,261.18315136098823,261.4371233867416,261.68456456148004,261.9256066146167,262.1602833961619,262.38852426979315,262.6101549827465,262.8249059711645,263.0324276556539,263.2323119016477,263.4241184827054,263.60740510979326,263.7817593903097,263.9468309676145,264.1023620710189,264.2482147786912,264.38439345794353,264.5110610904103,264.62854850099353,264.73735587285717,264.83814632719367,264.93173175525465,265.01905148987345,265.1011447735531,265.1791183008492,265.254110367539,265.32725333467636,265.3996362029834,265.4722690876169,265.5460512855415,265.62174444270664,265.69995206569683,265.78110629627156,265.86546249429983,265.95310177453484,266.04394123634336,266.1377512340955,266.2341786798254,266.33277506748703,266.4330276753373,266.53439225169313,266.63632542726157,266.7383151275033,266.8399073792645,266.9407281107069,267.0404988214601,267.139045336209,267.23629923184296,267.33229192598094,267.4271418122001,267.5210352037525,267.6142021831703,267.7068887323259,267.7993267216992,267.89170345810965,267.98413252070185,268.0766275539466,268.16908053706135,268.26124581940337,268.35273091311575,268.4429946832405,268.53135319005054,268.61699303867925,268.6989916981609,268.7763438862919,268.84799279741236,268.9128646939579,268.9699052028407,269.0181155637456,269.05658707316394,269.0845320555738,269.10130986700426,269.1064466871929,269.09964817152587,269.08080439649314,269.04998692356014,269.00743820576747,268.9535539481996,268.88885938759944,268.8139807594446,268.72961345705073,268.63648854452924,268.53533935565764,268.4268698904421,268.31172661159957,268.19047505029783,268.0635823645268,267.93140666851485,267.7941935848065,267.6520800812931,267.5051052641129,267.3532274244639,267.1963463025626,267.034329252774,266.86703978494825,266.6943668291285,266.5162530307077,266.33272043283716,266.14389203976657,265.95000797158576,265.75143520612744,265.5486702425729,265.3423343957864,265.13316182093837,264.92198075407015,264.7096888156698,264.4972235421281,264.2855295674773,264.0755240616003,263.86806213151516,263.66390390412687,263.46368493138516,263.26789139613226,263.07684135756557,262.89067297167696,262.7093402700973,262.532616698838,262.360106226375,262.1912614487917,262.0254077681358,261.86177241701716,261.6995168634863,261.5377709677362,261.37566718448994,261.21237311597,261.0471208193423,260.87923145414095,260.7081341097583,260.5333779671614,260.35463730603124,260.17170924972413,259.9845045258282,259.7930318894067,259.5973771899655,259.39767834427806,259.1940976907885,258.98679333627297,258.77589115479043,258.5614590602124,258.34348504891756,258.12186030521303,257.8963683894904,257.6666812024077,257.4323620547264,257.1928757909343,256.9476055355083,256.6958752735382,256.4369771611796,256.1702022026395,255.89487274269325,255.610375116934,255.3161907816836,255.01192431282152,254.69732681421476,254.37231350406148,254.03697453941206,253.69157848017508,253.3365681662586,252.9725491657326,252.60027132800508,252.22060432440202,251.83450836112212,251.44300149058728,251.0471251142607,250.64790935444978,250.24633997017995,249.8433284031803,249.43968636915366,249.03610616587298,248.63314756609432,248.2312318157772,247.83064288488202,247.4315357386408,247.03395103151456,246.63783529317362,246.24306539312596,245.84947585279343,245.456887432188,245.0651353601713,244.6740956054949,244.2837076988014,243.89399280745707,243.50506602530618,243.11714215442913,242.7305346091283,242.34564744505846,241.96296088904967,241.58301109814514,241.20636519114646,240.8335928559912,240.46523602792328,240.10177824659587,239.74361532901554,239.39102893807097,239.04416448626475,238.7030145984929,238.3674090775668,238.03701198627294,237.71132609734389,237.3897045866851,237.07136947517716,236.75543597964904,236.44094163213936,236.12687878402176,235.81222894054892,235.4959972807214,235.17724571174705,234.85512288692516,234.52888977631133,234.1979396125207,233.86181132725687,233.52019593223073,233.17293566335175,232.82001608033315,232.46155167566067,232.09776587835844,231.72896662189098,231.35551886712184,230.97781561904299,230.59624904240894,230.21118326314067,229.82293034052856,229.43173071536012,229.03773919070755,228.64101719839164,228.2415317611325,227.8391611961905,227.43370724006763,227.02491292484595,226.612485223327,226.19612121876878,225.77553635941442,225.3504932384335,224.92082930262671,224.48648194023107,224.04750952694857,223.60410721316177,223.15661650352644,222.70552799870816,222.25147702129388,221.79523221544173,221.33767757342702,220.87978868290696,220.42260428867831,219.96719450629965,219.51462719950055,219.0659341297343,218.6220784995591,218.183925441117,217.75221685054817,217.3275517466419,216.91037304904592,216.50096134274278,216.09943583842966,215.7057623715496,215.31976792514538,214.94116083206706,214.56955552749224,214.204500497977,213.84550791994974,213.49208340655437,213.14375429073758,212.80009496381456,212.4607479574784,212.12543969434333,211.79399012505874,211.46631580332306,211.14242630619404,210.8224142671903,210.50643963487622,210.19470908178752,209.88745175130828,209.58489272967557,209.287225756219,208.9945867306574,208.70702953940483,208.42450560534232,208.14684837357981,207.87376368943865,207.60482671768628,207.33948571005223,207.07707256916927,206.81681980011874,206.55788310439286,206.29936857292392,206.0403631904632,205.77996718578856,205.5173266602679,205.25166490651569,204.98231089032433,204.708723509502,204.43051045535225,204.14744077520015,203.85945055336262,203.5666414766285,203.26927241047815,202.967744465168,202.66258035807164,202.35439916330566,202.0438877669171,201.73177050403692,201.41877853529465,201.10562051938462,200.79295605716246,200.48137322446632,200.17137128443656,199.8633493874454,199.55760174276594,199.25431939787185,198.95359840707533,198.65545382968168,198.35983868687492,198.06666674243968,197.7758377689834,197.48726382917138,197.20089504749384,196.91674337501757,196.6349029559566,196.3555658850855,196.07903238950678,195.80571476420334,195.5361347225984,195.27091417358213,195.01075978676613,194.75644203970947,194.5087697371251,194.26856123708455,194.0366138001591,193.81367258478392,193.60040084028353,193.39735279647127,193.2049506182671,193.0234665921787,192.85301144928118,192.6935294201941,192.5448002776094,192.40644826889647,192.27795749358077,192.15869295615926,192.04792624063103,191.94486452402103,191.84868148390592,191.75854856785628,191.6736650851975,191.59328565380392,191.5167436828509,191.4434697886699,191.3730043136471,191.30500343310138,191.23923867571347,191.17559003147016,191.114033158975,191.05462151367487,190.99746448387222,190.94270282849226,190.8904828486827,190.84093078729234,190.79412893295623,190.75009480992375,190.70876466601334,190.66998223810876,190.63349348973136,190.5989476934404,190.5659048890717,190.5338494050884,190.50220880265334,190.4703773076488,190.43774255022186,190.40371424741699,190.3677533517347,190.32940015294392,190.2882998640629,190.24422434284057,190.19708879104422,190.14696252542336,190.09407321324403,190.03880429617774,189.98168567182108,189.9233780442174,189.86465167562562,189.80636055474696,189.74941322697103,189.6947416980037,189.6432699149306,189.5958833436223,189.55340109773977,189.51655193589028,189.48595523707144,189.4621078013084,189.44537701617318,189.43600059674307,189.43409276395641,189.43965639218268,189.45260034858993,189.47276098064555,189.499926497677,189.53386284876024,189.57433962970336,189.62115456005142,189.67415515630358,189.7332563852921,189.79845330353538,189.86982796259375,189.94755017254715,190.03187204923026,190.12311660824022,190.22166099223304,190.3279152106376,190.44229751727056,190.56520773848976,190.69699998265497,190.8379562045581,190.98826206388696,191.14798640648993,191.3170655169737,191.49529305041196,191.68231626224616,191.8776388339049,192.08063025404545,192.29054137925183,192.50652548104233,192.7276638046664,192.95299443405796,193.1815430884752,193.41235437855016,193.64452202772867,193.87721662046644,194.10970956808802,194.3413921800774,194.57178898240463,194.8005647219937,195.02752482203184,195.252609389482,195.47588120631968,195.69750844244984,195.91774309502662,196.1368963721385,196.3553123875035,196.5733416092969,196.79131550665954,197.00952376193533,197.22819526937775,197.44748392980298,197.66745998666258,197.88810734621183,198.10932699883233,198.33094632727799,198.5527337680819,198.77441800138,198.99571059726168,199.2163308564151,199.43603145928674,199.65462348768344,199.87199940835788,200.08815270826994,200.30319304066992,200.51735597101103,200.73100668994118,200.94463737267134,201.1588581935495,201.374382334447,201.59200563834656,201.81258183897273,202.036994528829,202.2661271995309,202.50083279091334,202.7419042137196,202.9900472633045,203.24585722121896,203.50980025411693,203.78220047496055,204.06323324277767,204.3529249593352,204.65115929062264,204.95768941522041,205.27215559742447,205.59410711622382,205.92302736572302,206.25836078950454,206.59954022860163,206.94601325437682,207.29726612395422,207.65284413330355,208.0123673442726,208.3755409161399,208.742159566171,209.11210600171617,209.48534349183925,209.86190306221084,210.24186608645445,210.62534329511516,211.01245141693826,211.40328879606778,211.79791138638043,212.1963105074661,212.59839365666866,213.00396951286976,213.41273804868436,213.8242863999471,214.23809083886226,214.65352487582825,215.06987319158034,215.48635079286262,215.90212650749376,216.3163497029503,216.72817893857902,217.13681115436484,217.54150996442456,217.9416316628797,218.3366476614129,218.72616225579497,219.10992485368186,219.48783607575376,219.85994745219938,220.22645476048402,220.58768537139673,220.94408027184747,221.29617169910884,221.64455753836376,221.98987379229598,222.33276651998574,222.6738646580559,223.0137550791933,223.3529611150062,223.6919255785267,224.03099907667757,224.37043411762946,224.71038520720575,225.05091480868467,225.39200472828887,225.73357220070037,226.07548970016964,226.41760730620783,226.75977631871075,227.10187275263704,227.4438193503706,227.78560483015227,228.12729923718084,228.46906447233133,228.81115933092275,229.15393867704788,229.49784669223692,229.84340445421557,230.19119240559104,230.54182854750186,230.89594342516668,231.2541531486971,231.61703180425462,231.9850846517683,232.35872347382312,232.73824533762087,233.12381586346174,233.5154578677868,233.9130459782199,234.31630751637067,234.72482962713005,235.13807231733256,235.5553867684106,235.9760380225404,236.39923092353527,236.82413803368763,237.24992815414257,237.67579405397615,238.1009780630018,238.5247943027001,238.9466465122146,239.36604066241776,239.7825918281966,240.1960250927728,240.60617057211897,241.01295295590114,241.41637624773526,241.81650463692836,242.21344063328422,242.60730173561853,242.99819697615368,243.38620468330367,243.77135273459936,244.15360243345958,244.53283694552073,244.908854982766,245.28137013980339,245.6500159812836,246.0143566686884,246.37390261490043,246.72813038183438,247.07650580431903,247.41850914454943,247.7536609651733,248.0815473615428,248.4018432173563,248.7143322416143,249.01892270362666,249.31565799845433,249.60472143637767,249.88643494313152,250.1612516674301,250.429742802571,250.6925792234848,250.95050880414416,251.20433049906066,251.4548664353127,251.70293335975353,251.94931481469155,252.1947353730187,252.43983815278244,252.68516665744224,252.93115176067633,253.17810438555722,253.4262141313011,253.67555379214963,253.92608940839895,254.17769520505112,254.4301725238623,254.6832716526887,254.9367153124536,255.19022248418398,255.44353125034172,255.69641938640999,255.94872156701848,256.2003422389613,256.4512634512994,256.70154720778515,256.9513322045876,257.2008251210569,257.450286927012,257.700014941214,257.9503216080627,258.20151114090737,258.4538553011975,258.7075696366478,258.9627914858414,259.21956097205583,259.4778060599918,259.73733254332416,259.99781957932805,260.2588211024266,260.51977314622314,260.78000679920757,261.038766228763,261.2952309464964,261.54854126892684,261.7978257625751,262.042229360096,262.28094079956804,262.5132180741561,262.73841068220366,262.9559776329662,263.16550038198864,263.36669013110145,263.5593892175,263.7435666192533,263.9193079050554,264.08680023842004,264.2463132961309,264.39817716463205,264.5427584256218,264.6804357258006,264.81157614127795,264.9365135938927,265.05553045757154,265.16884331422426,265.27659359002456,265.3788435363494,265.475577729208,265.5667099619778,265.65209511441014,265.7315453115117,265.8048494531809,265.8717950116055,265.93219086802594,265.9858899001211,266.03281003911195,266.0729525912877,266.10641675805186,266.13340948445256,266.15425000821847,266.1693687568998,266.1793005353598,266.1846722441653,266.1861856557721,266.18459603481284,266.1806876077208,266.1752470539213,266.16903629718536,266.16276591592634,266.1570704631534,266.1524868919663,266.1494371258588,266.1482156028957,266.14898237002654,266.1517620216065,266.1564484793353,266.16281531477523,266.1705310357863,266.17917850916956,266.1882774864647,266.19730904883147,266.20574069802353,266.213050798162,266.2187511183501,266.22240633664137,266.22364953555086,266.22219293919187,266.2178334005946,266.2104524312028,266.2000108580287,266.18653848205423,266.17011937908114,266.15087371731266,266.1289371523092,266.10443898983584,266.0774803737575,266.04811375607864,266.0163248395156,265.98201805318365,265.94500643591346,265.9050065691823,265.8616389348697,265.81443378603433,265.76284232671424,265.7062527146986,265.6440101440929,265.57544004576033,265.4998732748397,265.4166720443368,265.325255318046,265.22512239715763,265.1158735217551,264.9972264564548,264.86902823096176,264.73126145078396,264.5840448680084,264.42762819256706,264.26238141567757,264.08877919395957,263.9073810907147,263.7188086771427,263.5237206501625,263.32278721708906,263.1166650259085,262.9059738819333,262.6912763893142,262.4730614947099,262.25173269890223,262.0276014514671,261.8008859669509,261.57171541290614,261.34013913571494,261.10614032428356,260.8696532783302,260.6305832593048,260.38882776780986,260.1442980186992,259.89693937755794,259.6467495803253,259.3937936782914,259.1382148271933,258.880240262334,258.6201820597876,258.358432563258,258.09545464232576,257.831767225627,257.5679268073639,257.30450584408004,257.0420691293863,256.7811493483212,256.5222230643167,256.2656883775984,256.0118454150218,255.76088067193575,255.5128560339579,255.26770307059329,255.02522292574872,254.78509184630929,254.5468721038515,254.3100277911719,254.07394472884346,253.83795351027902,253.60135455751032,253.36344396215395,253.1235388518473,252.88100105348275,252.63525791912494,252.38581933352586,252.13229012556843,251.87437734906146,251.6118921683083,251.34474636661955,251.07294377663052,250.79656719523905,250.51576157941528,250.23071450988903,249.94163504787073,249.6487321883677,249.35219412830148,249.05216951785673,248.74875175199978,248.4419671918484,248.13176799140246,247.8180299553743,247.50055558164598,247.17908216156118,246.8532945375716,246.52284186508408,246.1873575069667,245.84648101653605,245.49988104702223,245.1472779686802,244.7884649818703,244.42332658520272,244.0518533885347,243.67415244439601,243.29045249869355,242.90110382043832,242.50657254722444,242.10742976381707,241.7043358008283,241.29802048502987,240.88926027962881,240.47885341106382,240.06759418036688,239.65624769662912,239.24552624562762,238.836068419556,238.42842198867424,238.02303130019877,237.6202297541563,237.22023764257227,236.823165361042,236.42902172493066,236.0377268606222,235.6491289090249,235.26302358608348,234.8791755033182,234.49734006773366,234.11728475907267,233.73880862434342,233.36175893249253,232.98604409053382,232.61164212795703,232.23860429797,231.86705360932066,231.49717837713305,231.1292211509262,230.76346362858618,230.4002083833733,230.03975840560292,229.68239558229288,229.328359300438,228.97782635939592,228.63089331515215,228.28756225727594,227.94773084456372,227.61118720685343,227.27761006971602,226.94657418880942,226.61756090573004,226.28997337158765,225.96315574213136,225.63641544172427,225.30904743356479,224.9803593286131,224.64969612114697,224.31646335702823,223.98014762062547,223.64033336369118,223.29671528713132,222.9491057146959,222.597436654235,222.24175651389498,221.8822217132925,221.51908368908803,221.1526720269829,220.7833746458988,220.4116161049294,220.0378351921165,219.6624629816801,219.28590251171684,218.90851113949833,218.5305864814526,218.15235664765424,217.77397524662638,217.3955213778031,217.01700455973102,216.63837427615624,216.25953357352023,215.88035592511895,215.50070440070547,215.1204520548798,214.7395023797633,214.35780866073705,213.97539112875228,213.59235091601073,213.20887998769433,212.82526643218904,212.44189473489007,212.0592409234699,211.6778627416729,211.2983852701671,210.92148265305462,210.54785679475643,210.17821405336844,209.81324106480702,209.45358088147128,209.09981059718828,208.75242155750433,208.41180312476737,208.07823078768502,207.75185918454918,207.4327203596004,207.12072730610123,206.81568258146277,206.51729152319365,206.22517936286582,205.93891134074315,205.65801477627556,205.38200195692912,205.11039267453378,204.84273526611412,204.57862510344114,204.3177196176383,204.05974913453304,203.80452302307586,203.55193091109857,203.3019389867064,203.05458166582522,202.80994915309913,202.56817164153293,202.32940107458072,202.09379152348018,201.86147930580103,201.63256398462474,201.40709134084986,201.1850393063352,200.9663076884914,200.75071231579315,200.5379839990895,200.32777244786348,200.1196550171664,199.91314990364262,199.70773317136315,199.50285878256005,199.2979806455703,199.0925755808813,198.88616605195662,198.67834151334378,198.46877729412265,198.2572500566171,198.04364904209643,197.82798252796013,197.61037916352407,197.3910841113504,197.17045018453663,196.9489244237768,196.72703078814482,196.50534982846307,196.28449636170515,196.06509626141104,195.84776351761477,195.63307869839474,195.42156986506768,195.21369685855328,195.0098396926599,194.8102915705729,194.61525679520815,194.42485358523166,194.23912155000545,194.05803333204852,193.8815097075908,193.70943725583138,193.54168757496896,193.37813694482966,193.2186853160065,193.06327354481283,192.91189788989567,192.7646209349232,192.62157829449188,192.482980687138,192.34911120813672,192.22031789255627,192.0970019122526,191.9796019858465,191.86857578587092,191.76437929144552,191.6674451494211,191.57816116592923,191.49685005057822,191.42375147713093,191.35900741043707,191.3026514856273,191.25460302066972,191.21466600803677,191.1825331777239,191.15779496542189,191.1399529697526,191.12843725414652,191.12262665405274,191.12187109885335,191.1255148579396,191.13291957704462,191.14348598630463,191.1566732347519,191.17201493312237,191.18913116127442,191.20773590897704,191.2276396581867,191.24874706864026,191.27104998336927,191.2946162132668,191.31957477745237,191.3460984575452,191.37438465963845,191.40463566074786,191.43703934252673,191.47175148283088,191.50888058706363,191.54847610087134,191.59052066113662,191.63492682308367,191.68153845921918,191.73013677352887,191.78045062510282,191.83217062226595,191.88496624358697,191.9385050765709,191.99247314709095,192.04659524886043,192.10065417582487,192.15450781160695,192.2081031363778,192.2614863671989,192.31480864483632,192.36832690799085,192.42239984290413,192.47747904948352,192.53409581121966,192.59284408240703,192.65436050072864,192.71930238599637,192.78832478881841,192.86205770085724,192.94108452870194,193.0259228667631,193.1170084845409,193.2146832764418,193.3191877167993,193.4306581296615,193.54912883443814,193.67453897763934,193.80674362078176,193.9455284376014,194.09062719128713,194.2417410239976,194.3985585036094,194.5607753409263,194.72811271602427,194.90033323356388,195.0772536594267,195.25875376787837,195.4447808402807,195.63534959199336,195.8305375511171,196.03047615813858,196.23533808645638,196.44532148806596,196.66063203565113,196.88146375321455,197.10797969585468,197.34029355178026,197.5784531955107,197.8224271227597,198.07209454991656,198.32723977205686,198.58755115293928,198.8526248800191,199.12197336971457,199.39503796596222,199.67120535114975,199.94982689452033,200.2302400092244,200.5117904833788,200.79385469828097,201.0758606510189,201.35730675886842,201.63777753590654,201.91695539225677,202.19462800494495,202.47069093608638,202.74514541720742,203.01809146523172,203.28971673317184,203.56028171449063,203.8301021032074,204.09952925263096,204.3689297668228,204.63866529586747,204.90907358692118,205.1804517689319,205.4530427238594,205.72702522781694,206.00250834076505,206.27953029401203,206.55806188289102,206.83801413029806,207.1192497579875,207.40159779861887,207.68487051322398,207.96888165481698,208.25346504577675,208.53849241819313,208.82388950352006,209.1096494486403,209.39584277503,209.68262327877758,209.97022948224753,210.2589814820084,210.54927328000826,210.8415609231739,211.13634699816384,211.43416222132095,211.7355450188342,212.0410201006358,212.35107708795653,212.66615025580882,212.986600397899,213.31269971541622,213.64462047835522,213.98242801649573,214.32607837689125,214.67542074721632,215.03020450197053,215.39009049397765,215.75466599905295,216.12346253830242,216.4959756597699,216.87168566644732,217.25007823586057,217.63066388963014,218.01299533879072,218.3966818486134,218.78139992894586,219.16689985409747,219.55300773961713,219.93962314019646,220.32671237094686,220.71429798096344,221.10244501152883,221.4912448408919,221.88079754441696,222.2711937764418,222.66249720438515,223.05472849529426,223.4478517717653,223.84176432241674,224.23629017880188,224.6311779649396,225.02610319833005,225.4206749843513,225.8144468117401,226.20693093772306,226.59761565878256,226.98598460709496,227.37153710154104,227.75380852167612,228.13238966632318,228.50694410586962,228.87722263642254,229.2430740895294,229.60445193559,229.9614163327435,230.31413150474833,230.6628585691001,231.00794416795756,231.3498054673149,231.6889122732101,232.0257671580595,232.36088458799892,232.6947700883544,233.02790047672644,233.36070613213974,233.6935561575541,234.02674713764793,234.36049600236043,234.6949372892897,235.03012486609566,235.36603793971625,235.70259095473529,236.03964678038523,236.37703241500867,236.71455630723858,237.05202631141256,237.38926726502797,237.72613719985816,238.0625412743646,238.39844263920097,238.73386961334103,239.0689187469258,239.4037535678611,239.73859904088104,240.0737319980533,240.4094680164463,240.7461454105313,241.08410716379473,241.42368173775657,241.76516376117843,242.10879561430667,242.45475088191097,242.80312055678675,243.15390273711668,243.50699638385757,243.86219949739794,244.21921184693178,244.5776421530999,244.93701939661827,245.29680771471868,245.65642416425632,246.0152584848063,246.37269389453084,246.72812790124107,247.0809921134752,247.43077009137062,247.77701238169215,248.11934803005823,248.45749204836522,248.79124852692883,249.1205093078628,249.44524836680847,249.7655122722667,250.08140729386542,250.39308390238944,250.70071953638833,251.00450059582099,251.30460465808784,251.60118389420109,251.89435059376015,252.1841655905334,252.4706302220055,252.7536822646023,253.033196071559,253.3089869138561,253.580819298224,253.84841882180672,254.11148693193317,254.36971780159655,254.6228164149673,254.87051688865023,255.11260003703538,255.3489092248851,255.57936363542393,255.80396821318703,256.0228197108776,256.2361084695927,256.4441157815217,256.64720691212915,256.8458200829797,257.0404519250585,257.231640094788,257.4199438914426,257.6059238176945,257.79012107923927,257.9730380220607,258.1551204568072,258.33674272158214,258.5181961923583,258.6996817716813,258.8813066806864,259.0630856574648,259.2449464381037,259.426739177108,259.60824926285045,259.7892128116541,259.96933399002506,260.14830322535965,260.32581532573164,260.50158654111203,260.67536966087397,260.84696635232336,261.01623609643616,261.18310126195837,261.3475480677854,261.509623405085,261.6694277133181,261.82710431644364,261.9828258160382,262.13677829683076,262.28914421893154,262.4400849436099,262.5897238620583,262.7381310680045,262.88531043681814,263.03118985001015,263.175615141247,263.3183481467281,263.4590690291317,263.59738282151045,263.73282991717167,263.8649000252382,263.9930489300298,264.11671724512337,264.23535024762185,264.34841782026473,264.4554335215687,264.5559718475326,264.6496828403067,264.7363033347768,264.81566430614623,264.88769398131325,264.9524165936891,265.0099468837691,265.0604806646313,265.10428197130597,265.14166748522837,265.1729890608316,265.19861527371063,265.21891295395915,265.2342296620219,265.2448780081395,265.25112261322425,265.25317036428714,265.2511644389893,265.24518237092866,265.23523821058404,265.22128861779663,265.2032425117892,265.1809737149904,265.15433586729995,265.12317876623547,265.08736521197847,265.046787408659,265.00138199569915,264.9511428544663,264.8961309520926,264.83648063995213,264.772402010778,264.7041791259528,264.6321641422315,264.5567675836458,264.47844520824583,264.39768209999005,264.3149747640392,264.23081211121354,264.1456562788158,264.05992424720796,263.9739711738242,263.88807628069856,263.8024320025633,263.71713693688406,263.6321929435071,263.54750653001423,263.46289444046033,263.37809315129505,263.29277178008255,263.20654774044885,263.11900433943794,263.0297094182547,262.9382340891148,262.84417062204375,262.74714858583445,262.6468484442568,262.543011946871,262.4354488261268,262.3240395097468,262.20873376933196,262.08954544155154,261.96654356590966,261.8398404720286,261.709577509786,261.5759092391097,261.4389869763391,261.29894262665863,261.15587371553363,261.0098304672569,260.8608056690862,260.70872791081854,260.55345860982874,260.3947930300614,260.23246529079466,260.06615714835107,259.8955101324809,259.72014043943335,259.53965583519204,259.3536737127545,259.1618393824145,258.96384365716546,258.7594388274475,258.54845219883725,258.33079648876287,258.10647653752403,257.87559197647045,257.6383357023613,257.394988220916,257.1459081332225,256.89151923500395,256.63229487054736,256.3687403214337,256.10137410790435,255.83070913272212,255.55723460107484,255.2813996053147,255.00359917253363,254.72416344095868,254.4433504648773,254.16134295601367,253.87824906204392,253.59410707112502,253.3088937260065,253.0225356432337,252.73492317195218,252.4459259012757,252.15540894161293,251.86324906806865,251.5693498249426,251.27365474879934,250.9761579704475,250.6769115979987,250.37602945648433,250.07368695516297,249.77011706136028,249.46560256853135,249.16046504525346,248.8550510306074,248.54971619050608,248.24480826117306,247.94064967430867,247.6375207799319,247.33564455633277,247.03517362340577,246.73618025969157,246.43864997079822,246.1424789754892,245.84747577509427,245.5533667625006,245.25980561975032,244.96638605901262,244.67265729054463,244.37814146210644,244.0823522143894,243.78481344151558,243.48507733742662,243.18274084836946,242.87745973659457,242.56895958631995,242.25704324334868,241.94159436604568,241.6225769679089,241.3000310401143,240.97406454528453,240.64484226074165,240.3125721108999,239.97748975583838,239.63984228986337,239.29987194554496,238.9578006931926,238.6138165733811,238.26806250377362,237.92062816622968,237.57154541318596,237.22078744236674,236.8682717860094,236.5138669556577,236.15740238694727,235.7986811510094,235.43749474947742,235.0736391964333,234.70693151893477,234.3372257817644,233.96442776309385,233.58850747474742,233.20950883016584,232.8275559092745,232.4428554446038,232.0556953480219,231.666439302128,231.27551764400948,230.88341496106102,230.49065498886378,230.097783540861,229.7053503014534,229.31389037289142,228.92390647892017,228.53585269392127,228.1501204871192,227.76702775145057,227.386811332236,227.00962338994023,226.63553173352355,226.26452405651838,225.89651580767995,225.53136124230508,225.16886703876506,224.8088077358509,224.45094215686186,224.0950299406035,223.74084729990804,223.38820117484838,223.03694103796224,222.68696773774187,222.33823892755254,221.99077081152987,221.64463613716535,221.29995856579924,220.95690374651966,220.61566759581726,220.27646243550154,219.93950175697714,219.60498445496728,219.27307940423754,218.94391123724373,218.61754811973785,218.29399221842732,217.97317341519982,217.6549466535288,217.33909311331058,217.02532521047897,216.71329521775684,216.4026071133069,216.09283109474003,215.78352005578273,215.47422721920617,215.16452405778233,214.85401761827134,214.54236639261256,214.22929395403867,213.91459968989685,213.59816611156134,213.27996239718172,212.9600440159599,212.63854848311172,212.31568749218036,211.99173585564066,211.6670178461665,211.34189166115007,211.01673282531038,210.69191739576098,210.3678058382096,210.04472840189445,209.72297273664856,209.40277437261156,209.08431052808092,208.76769753198207,208.45299195386607,208.14019533638518,207.8292622332392,207.52011107969483,207.21263727216726,206.90672771589422,206.60227602156036,206.29919749691092,205.99744308972447,205.69701149341256,205.39795872307946,205.1004046030105,204.80453576926482,204.51060497477977,204.21892667951136,203.92986910440294,203.64384311511347,203.36128846964243,203.08265810446025,202.80840123914325,202.53894614431582,202.2746834385354,202.0159507555455,201.7630195553387,201.51608474428645,201.2752576268626,201.04056254168293,200.8119373465704,200.58923772095685,200.37224505938843,200.1606775473539,199.9542038496045,199.7524587099495,199.55505966698988,199.36162403727207,199.17178530859977,198.98520812211873,198.8016011003375,198.6207268952851,198.44240898035127,198.26653488313406,198.0930557456986,197.92198229310378,197.75337748068335,197.58734626544697,197.4240230979488,197.26355785017208,197.10610097618866,196.9517887413598,196.80072935065098,196.6529907575575,196.50859084480123,196.36749054113787,196.22959028199483,196.09473004348484,195.96269298897278,195.83321257480827,195.70598277718068,195.5806709349854,195.4569325628691,195.3344273816145,195.2128357453353,195.09187462011778,194.97131228806091,194.85098101311985,194.7307870074726,194.6107171739584,194.49084226431918,194.3713162759385,194.2523721019569,194.1343136410249,194.0175047535511,193.90235561171184,193.78930712236598,193.67881419851904,193.57132871111935,193.4672829658673,193.36707451876885,193.27105307104387,193.17951007259344,193.09267151941333,193.01069426168496,192.9336659546295,192.8616085932606,192.79448538491252,192.7322105396449,192.67466140740927,192.62169227008474,192.57314901243478,192.52888385304692,192.48876931659044,192.4527106721868,192.42065614699453,192.3926043447941,192.36860845003852,192.34877697057382,192.33327095793837,192.32229783306144,192.316102127377,192.31495361528772,192.31913345485515,192.32891906216688,192.3445685153544,192.36630531301552,192.39430429732684,192.42867949518416,192.46947453425483,192.5166561599282,192.57011122064736,192.6296473112891,192.69499707642308,192.76582598723192,192.8417432274266,192.92231516395321,193.00708074596162,193.0955680772797,193.18731134859573,193.28186729874065,193.37883040070918,193.47784603594582,193.57862102635505,193.68093103189491,193.78462448521387,193.88962291501258,193.99591769722446,194.10356345790453,194.21266852425506,194.32338297148996,194.43588493538408,194.55036594702335,194.66701609299324,194.78600980863797,194.9074930739311,195.0315727029839,195.1583083034601,195.28770733723883,195.41972354621478,195.55425882596677,195.6911684446901,195.83026932501033,195.97135094151437,196.1141882456912,196.25855591986772,196.40424318842062,196.5510683818736,196.69889245915832,196.8476307448046,196.9972622284319,197.1478358988808,197.29947373810953,197.4523701725829,197.6067879632694,197.76305069994783,197.92153224166105,198.08264360364282,198.24681792361673,198.4144942401155,198.58610087720564,198.76203925053312,198.94266888788502,199.12829439466407,199.31915499410397,199.5154171390018,199.71717053318136,199.924427725131,200.13712725246953,200.35514013368442,200.57827933246398,200.80631166882046,201.03897152793616,201.2759756286458,201.51703806329408,201.76188481192315,202.010266966747,202.26197197584523,202.51683232407697,202.77473120859136,203.03560492864395,203.29944188616633,203.5662782753839,203.83619071718385,204.109286257592,204.38569029104156,204.66553308071218,204.94893562424224,205.23599564960585,205.52677452100536,205.82128578853684,206.11948603057652,206.42126851876833,206.72646008841934,207.0348214297451,207.34605083651996,207.65979126764782,207.97564040345242,208.29316322117904,208.61190648152345,208.93141441688738,209.2512448477778,209.57098492966335,209.89026574991553,210.20877505227486,210.52626746154706,210.84257170896439,211.15759451215004,211.47132093487568,211.78381123185866,212.09519436330564,212.40565853342028,212.71543925782962,213.0248055890083,213.33404521986273,213.64344923897545,213.95329732388558,214.2638441305914,214.57530756975547,214.8878595564813,215.2016196865016,215.51665213424266,215.83296589579118,216.15051832135265,216.46922170672033,216.78895255074863,217.10956294435746,217.43089344358336,217.75278670053711,218.07510108592592,218.39772353712087,218.72058090655258,219.04364916432812,219.3669599222033,219.6906038874599,220.0147310173918,220.33954731948612,220.66530841985008,220.99231019367483,221.3208769075762,221.651347456339,221.98406037897112,222.31933840569704,222.65747331510326,222.99871186765316,223.34324352891025,223.69119060581184,224.04260129694106,224.39744600934526,224.75561712780578,225.11693224627527,225.48114069465925,225.84793302643016,226.21695298245643,226.58781132170967,226.9601008166477,227.33341165486056,227.70734647188115,228.08153426371476,228.45564249032773,228.82938677971003,229.20253777099268,229.57492478766147,229.94643620014352,230.31701651212003,230.68666037765738,231.05540391760573,231.42331384524806,231.79047502550534,232.15697717315746,232.52290143931222,232.8883076394947,233.2532228410676,233.6176319541424,233.98147086259823,234.3446224959674,234.70691608592486,235.06812968119027,235.42799582074514,235.7862100965058,236.1424421818352,236.49634876962745,236.84758776011572,237.1958329694548,237.54078859913278,237.88220271498804,238.21987903259165,238.55368639048393,238.88356540979507,239.20953198201335,239.5316773886048,239.85016502837382,240.16522390190363,240.47713916807788,240.7862402369301,241.09288698813106,241.39745479880733,241.700319123205,242.00184038694368,242.30234993924657,242.6021377486493,242.90144243435722,243.200444101573,243.49926030126198,243.797945270731,244.0964924396682,244.39484001591674,244.69287930514702,244.99046527710576,245.287428775636,245.58358968618376,245.87877032735497,246.17280832467986,246.46556825546122,246.75695142176264,247.04690321060633,247.33541763090173,247.62253876863474,247.90835906741236,248.19301451187005,248.4766769577591,248.759544006007,249.04182695063443,249.32373743516973,249.60547352363244,249.8872059265265,250.1690651177955,250.45113003558106,250.73341898015872,251.01588321076017,251.29840360497397,251.580790587265,251.86278736508382,252.14407634082866,252.42428840446223,252.7030146634013,252.97982004110762,253.2542580801332,253.5258862241729,253.79428083011896,254.05905117638028,254.31985178699279,254.57639247947677,254.82844566333907,255.07585055932591,255.31851416949686,255.55640899654009,255.78956767867857,256.01807486526076,256.2420567994036,256.4616691905136,256.67708404509955,256.88847617458947,257.0960101113017,257.2998281377068,257.5000400710856,257.69671534899766,257.88987783574765,258.0795036228904,258.26552193549384,258.4478190888747,258.6262452766103,258.80062381841475,258.9707623639354,259.1364654426221,259.2975476760956,259.45384693174236,259.6052366965665,259.75163698859734,259.8930231973787,260.0294323513559,260.16096644274916,260.2877925928115,260.4101400042413,260.5282938143752,260.6425861239092,260.7533846228652,260.86107936066026,260.96606830392744,261.068742389133,261.1694708037545,261.26858721844667,261.36637764385415,261.46307050210174,261.5588293889007,261.65374886362827,261.7478534489281,261.84109985648166,261.9333822902324,262.0245405211251,262.11437028655297,262.2026354504989,262.28908127290794,262.3734480836365,262.45548464012575,262.53496046950704,262.6116765539803,262.68547380989946,262.7562389311938,262.82390731021064,262.88846290626714,262.94993509588915,263.0083927002833,263.0639355365656,263.1166839717326,263.1667670653779,263.21430996312904,263.2594212437383,263.3021809265866,263.3426298128614,263.38076076463415,263.4165124251141,263.44976575580984,263.4803436188714,263.5080134732514,263.5324930898807,263.5534590323522,263.57055750396444,263.58341703699676,263.59166240231457,263.594929051933,263.592877377413,263.5852060744723,263.57166394854846,263.5520595749329,263.5262683363562,263.4942364948363,263.4559821061934,263.41159274697736,263.3612201862133,263.305072289882,263.2434025863025,263.17649803829295,263.1046656570407,263.0282186484185,262.94746280217555,262.8626838170567,262.77413620142033,262.6820343022088,262.5865458997614,262.4877886680439,262.38582964666017,262.2806877105591,262.1723388640842,262.0607240362754,261.9457589219693,261.82734530517644,261.7053832230791,261.5797832848691,261.450478451801,261.31743461366113,261.18065936079876,261.0402084465495,260.8961895572682,260.74876314982987,260.59814027187394,260.44457744008866,260.2883688080747,260.12983599960006,259.9693161077975,259.8071484595415,259.6436608117009,259.4791556786776,259.31389748690106,259.14810121195416,258.98192307985977,258.8154538096337,258.6487147449765,258.48165707568216,258.31416419168295,258.1460570528492,257.977102304057,257.807022725622,257.6355094912297,257.46223561508975,257.2868699118862,257.10909077016817,256.928599053272,256.7451294909252,256.55846000673154,256.36841853743766,256.1748870333922,255.97780247891686,255.7771549285682,255.5729827122442,255.36536511055058,255.15441293403967,254.9402575489539,254.72303897227573,254.5028937060403,254.27994299257742,254.0542821480471,253.82597157263712,253.59502994523382,253.3614299930466,253.12509708874288,252.88591077639506,252.6437091708985,252.39829602172122,252.14945008895816,251.89693635518145,251.64051849702673,251.37997197104892,251.11509703275294,250.84573100771712,250.57175916937794,250.2931236474655,250.00982989063468,249.7219503313226,249.42962504377076,249.13305934001127,248.83251840540183,248.52831922682068,248.22082020500247,247.91040896050427,247.59748893430478,247.28246544431082,246.9657318849721,246.64765674747034,246.32857209311516,246.0087640350266,245.68846567701468,245.36785282936168,245.0470426757572,244.72609541155498,244.40501871887756,244.083774796978,243.76228953435475,243.4404632992847,243.11818274340152,242.7953329629497,242.471809346958,242.14752846155153,241.82243737385434,241.4965209045371,241.16980641049594,240.8423658324789,240.51431488967845,240.1858094566426,239.8570393092589,239.5282195680915,239.19958029157212,238.87135477194843,238.5437671582164,238.21702006877626,237.89128286023555,237.56668118745603,237.24328842524366,236.9211194274059,236.60012697920612,236.2802011607925,235.96117168921106,235.6428131529672,235.32485290380342,235.00698123320504,234.6888633432836,234.37015252925917,234.0505039285694,233.72958816290128,233.40710420563445,233.08279084795268,232.75643621005372,232.42788484562126,232.09704211266728,231.76387562546628,231.42841375316885,231.0907412829179,230.7509925109428,230.4093421565856,230.06599460467038,229.7211720653295,229.3751022930168,229.02800652527125,228.68008828592156,228.33152364772855,227.9824534686156,227.6329780079223,227.28315420022446,226.93299572099144,226.58247582823364,226.23153281522488,225.880077769188,225.52800420686722,225.17519905666998,224.82155438385067,224.4669792138517,224.11141080157532,223.75482472140095,223.39724321276145,223.03874130589566,222.67945036728634,222.31955883825688,221.95931008621372,221.59899743844403,221.23895661533996,220.87955591573777,220.52118462463665,220.16424020676214,219.80911491340365,219.4561824613561,219.1057854400144,218.75822406586727,218.41374683475874,218.07254352494948,217.73474088331497,217.40040118928167,217.0695237435229,216.7420491787241,216.41786634568842,216.0968213971938,215.77872858114046,215.46338216939338,215.1505688937311,214.84008023828056,214.53172394983687,214.22533417284956,213.92077969216655,213.61796986981412,213.31685798670298,213.0174418396221,212.71976159087367,212.4238950146956,212.12995042350664,211.83805768071494,211.54835780887643,211.26099177703824,210.97608909523862,210.69375685506716,210.41406983236328,210.1370622127693,209.86272141587483,209.59098438356517,209.32173656870302,209.05481371822242,208.7900063975491,208.52706705869,208.26571931992987,208.0056690078691,207.74661641866138,207.48826918967137,207.2303551387963,206.9726344281452,206.7149104416702,206.45703883096203,206.19893427636993,205.94057462702494,205.68200221714162,205.4233223001647,205.16469868938208,204.90634683587984,204.64852470475648,204.39152192157985,204.13564774739663,203.88121849772438,203.62854504593665,203.3779210430444,203.12961244457432],"type":"scatter","x":[0.0,1.0e-5,2.0e-5,3.0e-5,4.0e-5,5.0e-5,6.0e-5,7.0e-5,8.0e-5,9.0e-5,0.0001,0.00011,0.00012,0.00013,0.00014,0.00015,0.00016,0.00017,0.00018,0.00019,0.0002,0.00021,0.00022,0.00023,0.00024,0.00025,0.00026,0.00027,0.00028,0.00029,0.0003,0.00031,0.00032,0.00033,0.00034,0.00035,0.00036,0.00037,0.00038,0.00039,0.0004,0.00041,0.00042,0.00043,0.00044,0.00045,0.00046,0.00047,0.00048,0.00049,0.0005,0.00051,0.00052,0.00053,0.00054,0.00055,0.00056,0.00057,0.00058,0.00059,0.0006,0.00061,0.00062,0.00063,0.00064,0.00065,0.00066,0.00067,0.00068,0.00069,0.0007,0.00071,0.00072,0.00073,0.00074,0.00075,0.00076,0.00077,0.00078,0.00079,0.0008,0.00081,0.00082,0.00083,0.00084,0.00085,0.00086,0.00087,0.00088,0.00089,0.0009,0.00091,0.00092,0.00093,0.00094,0.00095,0.00096,0.00097,0.00098,0.00099,0.001,0.00101,0.00102,0.00103,0.00104,0.00105,0.00106,0.00107,0.00108,0.00109,0.0011,0.00111,0.00112,0.00113,0.00114,0.00115,0.00116,0.00117,0.00118,0.00119,0.0012,0.00121,0.00122,0.00123,0.00124,0.00125,0.00126,0.00127,0.00128,0.00129,0.0013,0.00131,0.00132,0.00133,0.00134,0.00135,0.00136,0.00137,0.00138,0.00139,0.0014,0.00141,0.00142,0.00143,0.00144,0.00145,0.00146,0.00147,0.00148,0.00149,0.0015,0.00151,0.00152,0.00153,0.00154,0.00155,0.00156,0.00157,0.00158,0.00159,0.0016,0.00161,0.00162,0.00163,0.00164,0.00165,0.00166,0.00167,0.00168,0.00169,0.0017,0.00171,0.00172,0.00173,0.00174,0.00175,0.00176,0.00177,0.00178,0.00179,0.0018,0.00181,0.00182,0.00183,0.00184,0.00185,0.00186,0.00187,0.00188,0.00189,0.0019,0.00191,0.00192,0.00193,0.00194,0.00195,0.00196,0.00197,0.00198,0.00199,0.002,0.00201,0.00202,0.00203,0.00204,0.00205,0.00206,0.00207,0.00208,0.00209,0.0021,0.00211,0.00212,0.00213,0.00214,0.00215,0.00216,0.00217,0.00218,0.00219,0.0022,0.00221,0.00222,0.00223,0.00224,0.00225,0.00226,0.00227,0.00228,0.00229,0.0023,0.00231,0.00232,0.00233,0.00234,0.00235,0.00236,0.00237,0.00238,0.00239,0.0024,0.00241,0.00242,0.00243,0.00244,0.00245,0.00246,0.00247,0.00248,0.00249,0.0025,0.00251,0.00252,0.00253,0.00254,0.00255,0.00256,0.00257,0.00258,0.00259,0.0026,0.00261,0.00262,0.00263,0.00264,0.00265,0.00266,0.00267,0.00268,0.00269,0.0027,0.00271,0.00272,0.00273,0.00274,0.00275,0.00276,0.00277,0.00278,0.00279,0.0028,0.00281,0.00282,0.00283,0.00284,0.00285,0.00286,0.00287,0.00288,0.00289,0.0029,0.00291,0.00292,0.00293,0.00294,0.00295,0.00296,0.00297,0.00298,0.00299,0.003,0.00301,0.00302,0.00303,0.00304,0.00305,0.00306,0.00307,0.00308,0.00309,0.0031,0.00311,0.00312,0.00313,0.00314,0.00315,0.00316,0.00317,0.00318,0.00319,0.0032,0.00321,0.00322,0.00323,0.00324,0.00325,0.00326,0.00327,0.00328,0.00329,0.0033,0.00331,0.00332,0.00333,0.00334,0.00335,0.00336,0.00337,0.00338,0.00339,0.0034,0.00341,0.00342,0.00343,0.00344,0.00345,0.00346,0.00347,0.00348,0.00349,0.0035,0.00351,0.00352,0.00353,0.00354,0.00355,0.00356,0.00357,0.00358,0.00359,0.0036,0.00361,0.00362,0.00363,0.00364,0.00365,0.00366,0.00367,0.00368,0.00369,0.0037,0.00371,0.00372,0.00373,0.00374,0.00375,0.00376,0.00377,0.00378,0.00379,0.0038,0.00381,0.00382,0.00383,0.00384,0.00385,0.00386,0.00387,0.00388,0.00389,0.0039,0.00391,0.00392,0.00393,0.00394,0.00395,0.00396,0.00397,0.00398,0.00399,0.004,0.00401,0.00402,0.00403,0.00404,0.00405,0.00406,0.00407,0.00408,0.00409,0.0041,0.00411,0.00412,0.00413,0.00414,0.00415,0.00416,0.00417,0.00418,0.00419,0.0042,0.00421,0.00422,0.00423,0.00424,0.00425,0.00426,0.00427,0.00428,0.00429,0.0043,0.00431,0.00432,0.00433,0.00434,0.00435,0.00436,0.00437,0.00438,0.00439,0.0044,0.00441,0.00442,0.00443,0.00444,0.00445,0.00446,0.00447,0.00448,0.00449,0.0045,0.00451,0.00452,0.00453,0.00454,0.00455,0.00456,0.00457,0.00458,0.00459,0.0046,0.00461,0.00462,0.00463,0.00464,0.00465,0.00466,0.00467,0.00468,0.00469,0.0047,0.00471,0.00472,0.00473,0.00474,0.00475,0.00476,0.00477,0.00478,0.00479,0.0048,0.00481,0.00482,0.00483,0.00484,0.00485,0.00486,0.00487,0.00488,0.00489,0.0049,0.00491,0.00492,0.00493,0.00494,0.00495,0.00496,0.00497,0.00498,0.00499,0.005,0.00501,0.00502,0.00503,0.00504,0.00505,0.00506,0.00507,0.00508,0.00509,0.0051,0.00511,0.00512,0.00513,0.00514,0.00515,0.00516,0.00517,0.00518,0.00519,0.0052,0.00521,0.00522,0.00523,0.00524,0.00525,0.00526,0.00527,0.00528,0.00529,0.0053,0.00531,0.00532,0.00533,0.00534,0.00535,0.00536,0.00537,0.00538,0.00539,0.0054,0.00541,0.00542,0.00543,0.00544,0.00545,0.00546,0.00547,0.00548,0.00549,0.0055,0.00551,0.00552,0.00553,0.00554,0.00555,0.00556,0.00557,0.00558,0.00559,0.0056,0.00561,0.00562,0.00563,0.00564,0.00565,0.00566,0.00567,0.00568,0.00569,0.0057,0.00571,0.00572,0.00573,0.00574,0.00575,0.00576,0.00577,0.00578,0.00579,0.0058,0.00581,0.00582,0.00583,0.00584,0.00585,0.00586,0.00587,0.00588,0.00589,0.0059,0.00591,0.00592,0.00593,0.00594,0.00595,0.00596,0.00597,0.00598,0.00599,0.006,0.00601,0.00602,0.00603,0.00604,0.00605,0.00606,0.00607,0.00608,0.00609,0.0061,0.00611,0.00612,0.00613,0.00614,0.00615,0.00616,0.00617,0.00618,0.00619,0.0062,0.00621,0.00622,0.00623,0.00624,0.00625,0.00626,0.00627,0.00628,0.00629,0.0063,0.00631,0.00632,0.00633,0.00634,0.00635,0.00636,0.00637,0.00638,0.00639,0.0064,0.00641,0.00642,0.00643,0.00644,0.00645,0.00646,0.00647,0.00648,0.00649,0.0065,0.00651,0.00652,0.00653,0.00654,0.00655,0.00656,0.00657,0.00658,0.00659,0.0066,0.00661,0.00662,0.00663,0.00664,0.00665,0.00666,0.00667,0.00668,0.00669,0.0067,0.00671,0.00672,0.00673,0.00674,0.00675,0.00676,0.00677,0.00678,0.00679,0.0068,0.00681,0.00682,0.00683,0.00684,0.00685,0.00686,0.00687,0.00688,0.00689,0.0069,0.00691,0.00692,0.00693,0.00694,0.00695,0.00696,0.00697,0.00698,0.00699,0.007,0.00701,0.00702,0.00703,0.00704,0.00705,0.00706,0.00707,0.00708,0.00709,0.0071,0.00711,0.00712,0.00713,0.00714,0.00715,0.00716,0.00717,0.00718,0.00719,0.0072,0.00721,0.00722,0.00723,0.00724,0.00725,0.00726,0.00727,0.00728,0.00729,0.0073,0.00731,0.00732,0.00733,0.00734,0.00735,0.00736,0.00737,0.00738,0.00739,0.0074,0.00741,0.00742,0.00743,0.00744,0.00745,0.00746,0.00747,0.00748,0.00749,0.0075,0.00751,0.00752,0.00753,0.00754,0.00755,0.00756,0.00757,0.00758,0.00759,0.0076,0.00761,0.00762,0.00763,0.00764,0.00765,0.00766,0.00767,0.00768,0.00769,0.0077,0.00771,0.00772,0.00773,0.00774,0.00775,0.00776,0.00777,0.00778,0.00779,0.0078,0.00781,0.00782,0.00783,0.00784,0.00785,0.00786,0.00787,0.00788,0.00789,0.0079,0.00791,0.00792,0.00793,0.00794,0.00795,0.00796,0.00797,0.00798,0.00799,0.008,0.00801,0.00802,0.00803,0.00804,0.00805,0.00806,0.00807,0.00808,0.00809,0.0081,0.00811,0.00812,0.00813,0.00814,0.00815,0.00816,0.00817,0.00818,0.00819,0.0082,0.00821,0.00822,0.00823,0.00824,0.00825,0.00826,0.00827,0.00828,0.00829,0.0083,0.00831,0.00832,0.00833,0.00834,0.00835,0.00836,0.00837,0.00838,0.00839,0.0084,0.00841,0.00842,0.00843,0.00844,0.00845,0.00846,0.00847,0.00848,0.00849,0.0085,0.00851,0.00852,0.00853,0.00854,0.00855,0.00856,0.00857,0.00858,0.00859,0.0086,0.00861,0.00862,0.00863,0.00864,0.00865,0.00866,0.00867,0.00868,0.00869,0.0087,0.00871,0.00872,0.00873,0.00874,0.00875,0.00876,0.00877,0.00878,0.00879,0.0088,0.00881,0.00882,0.00883,0.00884,0.00885,0.00886,0.00887,0.00888,0.00889,0.0089,0.00891,0.00892,0.00893,0.00894,0.00895,0.00896,0.00897,0.00898,0.00899,0.009,0.00901,0.00902,0.00903,0.00904,0.00905,0.00906,0.00907,0.00908,0.00909,0.0091,0.00911,0.00912,0.00913,0.00914,0.00915,0.00916,0.00917,0.00918,0.00919,0.0092,0.00921,0.00922,0.00923,0.00924,0.00925,0.00926,0.00927,0.00928,0.00929,0.0093,0.00931,0.00932,0.00933,0.00934,0.00935,0.00936,0.00937,0.00938,0.00939,0.0094,0.00941,0.00942,0.00943,0.00944,0.00945,0.00946,0.00947,0.00948,0.00949,0.0095,0.00951,0.00952,0.00953,0.00954,0.00955,0.00956,0.00957,0.00958,0.00959,0.0096,0.00961,0.00962,0.00963,0.00964,0.00965,0.00966,0.00967,0.00968,0.00969,0.0097,0.00971,0.00972,0.00973,0.00974,0.00975,0.00976,0.00977,0.00978,0.00979,0.0098,0.00981,0.00982,0.00983,0.00984,0.00985,0.00986,0.00987,0.00988,0.00989,0.0099,0.00991,0.00992,0.00993,0.00994,0.00995,0.00996,0.00997,0.00998,0.00999,0.01,0.01001,0.01002,0.01003,0.01004,0.01005,0.01006,0.01007,0.01008,0.01009,0.0101,0.01011,0.01012,0.01013,0.01014,0.01015,0.01016,0.01017,0.01018,0.01019,0.0102,0.01021,0.01022,0.01023,0.01024,0.01025,0.01026,0.01027,0.01028,0.01029,0.0103,0.01031,0.01032,0.01033,0.01034,0.01035,0.01036,0.01037,0.01038,0.01039,0.0104,0.01041,0.01042,0.01043,0.01044,0.01045,0.01046,0.01047,0.01048,0.01049,0.0105,0.01051,0.01052,0.01053,0.01054,0.01055,0.01056,0.01057,0.01058,0.01059,0.0106,0.01061,0.01062,0.01063,0.01064,0.01065,0.01066,0.01067,0.01068,0.01069,0.0107,0.01071,0.01072,0.01073,0.01074,0.01075,0.01076,0.01077,0.01078,0.01079,0.0108,0.01081,0.01082,0.01083,0.01084,0.01085,0.01086,0.01087,0.01088,0.01089,0.0109,0.01091,0.01092,0.01093,0.01094,0.01095,0.01096,0.01097,0.01098,0.01099,0.011,0.01101,0.01102,0.01103,0.01104,0.01105,0.01106,0.01107,0.01108,0.01109,0.0111,0.01111,0.01112,0.01113,0.01114,0.01115,0.01116,0.01117,0.01118,0.01119,0.0112,0.01121,0.01122,0.01123,0.01124,0.01125,0.01126,0.01127,0.01128,0.01129,0.0113,0.01131,0.01132,0.01133,0.01134,0.01135,0.01136,0.01137,0.01138,0.01139,0.0114,0.01141,0.01142,0.01143,0.01144,0.01145,0.01146,0.01147,0.01148,0.01149,0.0115,0.01151,0.01152,0.01153,0.01154,0.01155,0.01156,0.01157,0.01158,0.01159,0.0116,0.01161,0.01162,0.01163,0.01164,0.01165,0.01166,0.01167,0.01168,0.01169,0.0117,0.01171,0.01172,0.01173,0.01174,0.01175,0.01176,0.01177,0.01178,0.01179,0.0118,0.01181,0.01182,0.01183,0.01184,0.01185,0.01186,0.01187,0.01188,0.01189,0.0119,0.01191,0.01192,0.01193,0.01194,0.01195,0.01196,0.01197,0.01198,0.01199,0.012,0.01201,0.01202,0.01203,0.01204,0.01205,0.01206,0.01207,0.01208,0.01209,0.0121,0.01211,0.01212,0.01213,0.01214,0.01215,0.01216,0.01217,0.01218,0.01219,0.0122,0.01221,0.01222,0.01223,0.01224,0.01225,0.01226,0.01227,0.01228,0.01229,0.0123,0.01231,0.01232,0.01233,0.01234,0.01235,0.01236,0.01237,0.01238,0.01239,0.0124,0.01241,0.01242,0.01243,0.01244,0.01245,0.01246,0.01247,0.01248,0.01249,0.0125,0.01251,0.01252,0.01253,0.01254,0.01255,0.01256,0.01257,0.01258,0.01259,0.0126,0.01261,0.01262,0.01263,0.01264,0.01265,0.01266,0.01267,0.01268,0.01269,0.0127,0.01271,0.01272,0.01273,0.01274,0.01275,0.01276,0.01277,0.01278,0.01279,0.0128,0.01281,0.01282,0.01283,0.01284,0.01285,0.01286,0.01287,0.01288,0.01289,0.0129,0.01291,0.01292,0.01293,0.01294,0.01295,0.01296,0.01297,0.01298,0.01299,0.013,0.01301,0.01302,0.01303,0.01304,0.01305,0.01306,0.01307,0.01308,0.01309,0.0131,0.01311,0.01312,0.01313,0.01314,0.01315,0.01316,0.01317,0.01318,0.01319,0.0132,0.01321,0.01322,0.01323,0.01324,0.01325,0.01326,0.01327,0.01328,0.01329,0.0133,0.01331,0.01332,0.01333,0.01334,0.01335,0.01336,0.01337,0.01338,0.01339,0.0134,0.01341,0.01342,0.01343,0.01344,0.01345,0.01346,0.01347,0.01348,0.01349,0.0135,0.01351,0.01352,0.01353,0.01354,0.01355,0.01356,0.01357,0.01358,0.01359,0.0136,0.01361,0.01362,0.01363,0.01364,0.01365,0.01366,0.01367,0.01368,0.01369,0.0137,0.01371,0.01372,0.01373,0.01374,0.01375,0.01376,0.01377,0.01378,0.01379,0.0138,0.01381,0.01382,0.01383,0.01384,0.01385,0.01386,0.01387,0.01388,0.01389,0.0139,0.01391,0.01392,0.01393,0.01394,0.01395,0.01396,0.01397,0.01398,0.01399,0.014,0.01401,0.01402,0.01403,0.01404,0.01405,0.01406,0.01407,0.01408,0.01409,0.0141,0.01411,0.01412,0.01413,0.01414,0.01415,0.01416,0.01417,0.01418,0.01419,0.0142,0.01421,0.01422,0.01423,0.01424,0.01425,0.01426,0.01427,0.01428,0.01429,0.0143,0.01431,0.01432,0.01433,0.01434,0.01435,0.01436,0.01437,0.01438,0.01439,0.0144,0.01441,0.01442,0.01443,0.01444,0.01445,0.01446,0.01447,0.01448,0.01449,0.0145,0.01451,0.01452,0.01453,0.01454,0.01455,0.01456,0.01457,0.01458,0.01459,0.0146,0.01461,0.01462,0.01463,0.01464,0.01465,0.01466,0.01467,0.01468,0.01469,0.0147,0.01471,0.01472,0.01473,0.01474,0.01475,0.01476,0.01477,0.01478,0.01479,0.0148,0.01481,0.01482,0.01483,0.01484,0.01485,0.01486,0.01487,0.01488,0.01489,0.0149,0.01491,0.01492,0.01493,0.01494,0.01495,0.01496,0.01497,0.01498,0.01499,0.015,0.01501,0.01502,0.01503,0.01504,0.01505,0.01506,0.01507,0.01508,0.01509,0.0151,0.01511,0.01512,0.01513,0.01514,0.01515,0.01516,0.01517,0.01518,0.01519,0.0152,0.01521,0.01522,0.01523,0.01524,0.01525,0.01526,0.01527,0.01528,0.01529,0.0153,0.01531,0.01532,0.01533,0.01534,0.01535,0.01536,0.01537,0.01538,0.01539,0.0154,0.01541,0.01542,0.01543,0.01544,0.01545,0.01546,0.01547,0.01548,0.01549,0.0155,0.01551,0.01552,0.01553,0.01554,0.01555,0.01556,0.01557,0.01558,0.01559,0.0156,0.01561,0.01562,0.01563,0.01564,0.01565,0.01566,0.01567,0.01568,0.01569,0.0157,0.01571,0.01572,0.01573,0.01574,0.01575,0.01576,0.01577,0.01578,0.01579,0.0158,0.01581,0.01582,0.01583,0.01584,0.01585,0.01586,0.01587,0.01588,0.01589,0.0159,0.01591,0.01592,0.01593,0.01594,0.01595,0.01596,0.01597,0.01598,0.01599,0.016,0.01601,0.01602,0.01603,0.01604,0.01605,0.01606,0.01607,0.01608,0.01609,0.0161,0.01611,0.01612,0.01613,0.01614,0.01615,0.01616,0.01617,0.01618,0.01619,0.0162,0.01621,0.01622,0.01623,0.01624,0.01625,0.01626,0.01627,0.01628,0.01629,0.0163,0.01631,0.01632,0.01633,0.01634,0.01635,0.01636,0.01637,0.01638,0.01639,0.0164,0.01641,0.01642,0.01643,0.01644,0.01645,0.01646,0.01647,0.01648,0.01649,0.0165,0.01651,0.01652,0.01653,0.01654,0.01655,0.01656,0.01657,0.01658,0.01659,0.0166,0.01661,0.01662,0.01663,0.01664,0.01665,0.01666,0.01667,0.01668,0.01669,0.0167,0.01671,0.01672,0.01673,0.01674,0.01675,0.01676,0.01677,0.01678,0.01679,0.0168,0.01681,0.01682,0.01683,0.01684,0.01685,0.01686,0.01687,0.01688,0.01689,0.0169,0.01691,0.01692,0.01693,0.01694,0.01695,0.01696,0.01697,0.01698,0.01699,0.017,0.01701,0.01702,0.01703,0.01704,0.01705,0.01706,0.01707,0.01708,0.01709,0.0171,0.01711,0.01712,0.01713,0.01714,0.01715,0.01716,0.01717,0.01718,0.01719,0.0172,0.01721,0.01722,0.01723,0.01724,0.01725,0.01726,0.01727,0.01728,0.01729,0.0173,0.01731,0.01732,0.01733,0.01734,0.01735,0.01736,0.01737,0.01738,0.01739,0.0174,0.01741,0.01742,0.01743,0.01744,0.01745,0.01746,0.01747,0.01748,0.01749,0.0175,0.01751,0.01752,0.01753,0.01754,0.01755,0.01756,0.01757,0.01758,0.01759,0.0176,0.01761,0.01762,0.01763,0.01764,0.01765,0.01766,0.01767,0.01768,0.01769,0.0177,0.01771,0.01772,0.01773,0.01774,0.01775,0.01776,0.01777,0.01778,0.01779,0.0178,0.01781,0.01782,0.01783,0.01784,0.01785,0.01786,0.01787,0.01788,0.01789,0.0179,0.01791,0.01792,0.01793,0.01794,0.01795,0.01796,0.01797,0.01798,0.01799,0.018,0.01801,0.01802,0.01803,0.01804,0.01805,0.01806,0.01807,0.01808,0.01809,0.0181,0.01811,0.01812,0.01813,0.01814,0.01815,0.01816,0.01817,0.01818,0.01819,0.0182,0.01821,0.01822,0.01823,0.01824,0.01825,0.01826,0.01827,0.01828,0.01829,0.0183,0.01831,0.01832,0.01833,0.01834,0.01835,0.01836,0.01837,0.01838,0.01839,0.0184,0.01841,0.01842,0.01843,0.01844,0.01845,0.01846,0.01847,0.01848,0.01849,0.0185,0.01851,0.01852,0.01853,0.01854,0.01855,0.01856,0.01857,0.01858,0.01859,0.0186,0.01861,0.01862,0.01863,0.01864,0.01865,0.01866,0.01867,0.01868,0.01869,0.0187,0.01871,0.01872,0.01873,0.01874,0.01875,0.01876,0.01877,0.01878,0.01879,0.0188,0.01881,0.01882,0.01883,0.01884,0.01885,0.01886,0.01887,0.01888,0.01889,0.0189,0.01891,0.01892,0.01893,0.01894,0.01895,0.01896,0.01897,0.01898,0.01899,0.019,0.01901,0.01902,0.01903,0.01904,0.01905,0.01906,0.01907,0.01908,0.01909,0.0191,0.01911,0.01912,0.01913,0.01914,0.01915,0.01916,0.01917,0.01918,0.01919,0.0192,0.01921,0.01922,0.01923,0.01924,0.01925,0.01926,0.01927,0.01928,0.01929,0.0193,0.01931,0.01932,0.01933,0.01934,0.01935,0.01936,0.01937,0.01938,0.01939,0.0194,0.01941,0.01942,0.01943,0.01944,0.01945,0.01946,0.01947,0.01948,0.01949,0.0195,0.01951,0.01952,0.01953,0.01954,0.01955,0.01956,0.01957,0.01958,0.01959,0.0196,0.01961,0.01962,0.01963,0.01964,0.01965,0.01966,0.01967,0.01968,0.01969,0.0197,0.01971,0.01972,0.01973,0.01974,0.01975,0.01976,0.01977,0.01978,0.01979,0.0198,0.01981,0.01982,0.01983,0.01984,0.01985,0.01986,0.01987,0.01988,0.01989,0.0199,0.01991,0.01992,0.01993,0.01994,0.01995,0.01996,0.01997,0.01998,0.01999,0.02,0.02001,0.02002,0.02003,0.02004,0.02005,0.02006,0.02007,0.02008,0.02009,0.0201,0.02011,0.02012,0.02013,0.02014,0.02015,0.02016,0.02017,0.02018,0.02019,0.0202,0.02021,0.02022,0.02023,0.02024,0.02025,0.02026,0.02027,0.02028,0.02029,0.0203,0.02031,0.02032,0.02033,0.02034,0.02035,0.02036,0.02037,0.02038,0.02039,0.0204,0.02041,0.02042,0.02043,0.02044,0.02045,0.02046,0.02047,0.02048,0.02049,0.0205,0.02051,0.02052,0.02053,0.02054,0.02055,0.02056,0.02057,0.02058,0.02059,0.0206,0.02061,0.02062,0.02063,0.02064,0.02065,0.02066,0.02067,0.02068,0.02069,0.0207,0.02071,0.02072,0.02073,0.02074,0.02075,0.02076,0.02077,0.02078,0.02079,0.0208,0.02081,0.02082,0.02083,0.02084,0.02085,0.02086,0.02087,0.02088,0.02089,0.0209,0.02091,0.02092,0.02093,0.02094,0.02095,0.02096,0.02097,0.02098,0.02099,0.021,0.02101,0.02102,0.02103,0.02104,0.02105,0.02106,0.02107,0.02108,0.02109,0.0211,0.02111,0.02112,0.02113,0.02114,0.02115,0.02116,0.02117,0.02118,0.02119,0.0212,0.02121,0.02122,0.02123,0.02124,0.02125,0.02126,0.02127,0.02128,0.02129,0.0213,0.02131,0.02132,0.02133,0.02134,0.02135,0.02136,0.02137,0.02138,0.02139,0.0214,0.02141,0.02142,0.02143,0.02144,0.02145,0.02146,0.02147,0.02148,0.02149,0.0215,0.02151,0.02152,0.02153,0.02154,0.02155,0.02156,0.02157,0.02158,0.02159,0.0216,0.02161,0.02162,0.02163,0.02164,0.02165,0.02166,0.02167,0.02168,0.02169,0.0217,0.02171,0.02172,0.02173,0.02174,0.02175,0.02176,0.02177,0.02178,0.02179,0.0218,0.02181,0.02182,0.02183,0.02184,0.02185,0.02186,0.02187,0.02188,0.02189,0.0219,0.02191,0.02192,0.02193,0.02194,0.02195,0.02196,0.02197,0.02198,0.02199,0.022,0.02201,0.02202,0.02203,0.02204,0.02205,0.02206,0.02207,0.02208,0.02209,0.0221,0.02211,0.02212,0.02213,0.02214,0.02215,0.02216,0.02217,0.02218,0.02219,0.0222,0.02221,0.02222,0.02223,0.02224,0.02225,0.02226,0.02227,0.02228,0.02229,0.0223,0.02231,0.02232,0.02233,0.02234,0.02235,0.02236,0.02237,0.02238,0.02239,0.0224,0.02241,0.02242,0.02243,0.02244,0.02245,0.02246,0.02247,0.02248,0.02249,0.0225,0.02251,0.02252,0.02253,0.02254,0.02255,0.02256,0.02257,0.02258,0.02259,0.0226,0.02261,0.02262,0.02263,0.02264,0.02265,0.02266,0.02267,0.02268,0.02269,0.0227,0.02271,0.02272,0.02273,0.02274,0.02275,0.02276,0.02277,0.02278,0.02279,0.0228,0.02281,0.02282,0.02283,0.02284,0.02285,0.02286,0.02287,0.02288,0.02289,0.0229,0.02291,0.02292,0.02293,0.02294,0.02295,0.02296,0.02297,0.02298,0.02299,0.023,0.02301,0.02302,0.02303,0.02304,0.02305,0.02306,0.02307,0.02308,0.02309,0.0231,0.02311,0.02312,0.02313,0.02314,0.02315,0.02316,0.02317,0.02318,0.02319,0.0232,0.02321,0.02322,0.02323,0.02324,0.02325,0.02326,0.02327,0.02328,0.02329,0.0233,0.02331,0.02332,0.02333,0.02334,0.02335,0.02336,0.02337,0.02338,0.02339,0.0234,0.02341,0.02342,0.02343,0.02344,0.02345,0.02346,0.02347,0.02348,0.02349,0.0235,0.02351,0.02352,0.02353,0.02354,0.02355,0.02356,0.02357,0.02358,0.02359,0.0236,0.02361,0.02362,0.02363,0.02364,0.02365,0.02366,0.02367,0.02368,0.02369,0.0237,0.02371,0.02372,0.02373,0.02374,0.02375,0.02376,0.02377,0.02378,0.02379,0.0238,0.02381,0.02382,0.02383,0.02384,0.02385,0.02386,0.02387,0.02388,0.02389,0.0239,0.02391,0.02392,0.02393,0.02394,0.02395,0.02396,0.02397,0.02398,0.02399,0.024,0.02401,0.02402,0.02403,0.02404,0.02405,0.02406,0.02407,0.02408,0.02409,0.0241,0.02411,0.02412,0.02413,0.02414,0.02415,0.02416,0.02417,0.02418,0.02419,0.0242,0.02421,0.02422,0.02423,0.02424,0.02425,0.02426,0.02427,0.02428,0.02429,0.0243,0.02431,0.02432,0.02433,0.02434,0.02435,0.02436,0.02437,0.02438,0.02439,0.0244,0.02441,0.02442,0.02443,0.02444,0.02445,0.02446,0.02447,0.02448,0.02449,0.0245,0.02451,0.02452,0.02453,0.02454,0.02455,0.02456,0.02457,0.02458,0.02459,0.0246,0.02461,0.02462,0.02463,0.02464,0.02465,0.02466,0.02467,0.02468,0.02469,0.0247,0.02471,0.02472,0.02473,0.02474,0.02475,0.02476,0.02477,0.02478,0.02479,0.0248,0.02481,0.02482,0.02483,0.02484,0.02485,0.02486,0.02487,0.02488,0.02489,0.0249,0.02491,0.02492,0.02493,0.02494,0.02495,0.02496,0.02497,0.02498,0.02499,0.025,0.02501,0.02502,0.02503,0.02504,0.02505,0.02506,0.02507,0.02508,0.02509,0.0251,0.02511,0.02512,0.02513,0.02514,0.02515,0.02516,0.02517,0.02518,0.02519,0.0252,0.02521,0.02522,0.02523,0.02524,0.02525,0.02526,0.02527,0.02528,0.02529,0.0253,0.02531,0.02532,0.02533,0.02534,0.02535,0.02536,0.02537,0.02538,0.02539,0.0254,0.02541,0.02542,0.02543,0.02544,0.02545,0.02546,0.02547,0.02548,0.02549,0.0255,0.02551,0.02552,0.02553,0.02554,0.02555,0.02556,0.02557,0.02558,0.02559,0.0256,0.02561,0.02562,0.02563,0.02564,0.02565,0.02566,0.02567,0.02568,0.02569,0.0257,0.02571,0.02572,0.02573,0.02574,0.02575,0.02576,0.02577,0.02578,0.02579,0.0258,0.02581,0.02582,0.02583,0.02584,0.02585,0.02586,0.02587,0.02588,0.02589,0.0259,0.02591,0.02592,0.02593,0.02594,0.02595,0.02596,0.02597,0.02598,0.02599,0.026,0.02601,0.02602,0.02603,0.02604,0.02605,0.02606,0.02607,0.02608,0.02609,0.0261,0.02611,0.02612,0.02613,0.02614,0.02615,0.02616,0.02617,0.02618,0.02619,0.0262,0.02621,0.02622,0.02623,0.02624,0.02625,0.02626,0.02627,0.02628,0.02629,0.0263,0.02631,0.02632,0.02633,0.02634,0.02635,0.02636,0.02637,0.02638,0.02639,0.0264,0.02641,0.02642,0.02643,0.02644,0.02645,0.02646,0.02647,0.02648,0.02649,0.0265,0.02651,0.02652,0.02653,0.02654,0.02655,0.02656,0.02657,0.02658,0.02659,0.0266,0.02661,0.02662,0.02663,0.02664,0.02665,0.02666,0.02667,0.02668,0.02669,0.0267,0.02671,0.02672,0.02673,0.02674,0.02675,0.02676,0.02677,0.02678,0.02679,0.0268,0.02681,0.02682,0.02683,0.02684,0.02685,0.02686,0.02687,0.02688,0.02689,0.0269,0.02691,0.02692,0.02693,0.02694,0.02695,0.02696,0.02697,0.02698,0.02699,0.027,0.02701,0.02702,0.02703,0.02704,0.02705,0.02706,0.02707,0.02708,0.02709,0.0271,0.02711,0.02712,0.02713,0.02714,0.02715,0.02716,0.02717,0.02718,0.02719,0.0272,0.02721,0.02722,0.02723,0.02724,0.02725,0.02726,0.02727,0.02728,0.02729,0.0273,0.02731,0.02732,0.02733,0.02734,0.02735,0.02736,0.02737,0.02738,0.02739,0.0274,0.02741,0.02742,0.02743,0.02744,0.02745,0.02746,0.02747,0.02748,0.02749,0.0275,0.02751,0.02752,0.02753,0.02754,0.02755,0.02756,0.02757,0.02758,0.02759,0.0276,0.02761,0.02762,0.02763,0.02764,0.02765,0.02766,0.02767,0.02768,0.02769,0.0277,0.02771,0.02772,0.02773,0.02774,0.02775,0.02776,0.02777,0.02778,0.02779,0.0278,0.02781,0.02782,0.02783,0.02784,0.02785,0.02786,0.02787,0.02788,0.02789,0.0279,0.02791,0.02792,0.02793,0.02794,0.02795,0.02796,0.02797,0.02798,0.02799,0.028,0.02801,0.02802,0.02803,0.02804,0.02805,0.02806,0.02807,0.02808,0.02809,0.0281,0.02811,0.02812,0.02813,0.02814,0.02815,0.02816,0.02817,0.02818,0.02819,0.0282,0.02821,0.02822,0.02823,0.02824,0.02825,0.02826,0.02827,0.02828,0.02829,0.0283,0.02831,0.02832,0.02833,0.02834,0.02835,0.02836,0.02837,0.02838,0.02839,0.0284,0.02841,0.02842,0.02843,0.02844,0.02845,0.02846,0.02847,0.02848,0.02849,0.0285,0.02851,0.02852,0.02853,0.02854,0.02855,0.02856,0.02857,0.02858,0.02859,0.0286,0.02861,0.02862,0.02863,0.02864,0.02865,0.02866,0.02867,0.02868,0.02869,0.0287,0.02871,0.02872,0.02873,0.02874,0.02875,0.02876,0.02877,0.02878,0.02879,0.0288,0.02881,0.02882,0.02883,0.02884,0.02885,0.02886,0.02887,0.02888,0.02889,0.0289,0.02891,0.02892,0.02893,0.02894,0.02895,0.02896,0.02897,0.02898,0.02899,0.029,0.02901,0.02902,0.02903,0.02904,0.02905,0.02906,0.02907,0.02908,0.02909,0.0291,0.02911,0.02912,0.02913,0.02914,0.02915,0.02916,0.02917,0.02918,0.02919,0.0292,0.02921,0.02922,0.02923,0.02924,0.02925,0.02926,0.02927,0.02928,0.02929,0.0293,0.02931,0.02932,0.02933,0.02934,0.02935,0.02936,0.02937,0.02938,0.02939,0.0294,0.02941,0.02942,0.02943,0.02944,0.02945,0.02946,0.02947,0.02948,0.02949,0.0295,0.02951,0.02952,0.02953,0.02954,0.02955,0.02956,0.02957,0.02958,0.02959,0.0296,0.02961,0.02962,0.02963,0.02964,0.02965,0.02966,0.02967,0.02968,0.02969,0.0297,0.02971,0.02972,0.02973,0.02974,0.02975,0.02976,0.02977,0.02978,0.02979,0.0298,0.02981,0.02982,0.02983,0.02984,0.02985,0.02986,0.02987,0.02988,0.02989,0.0299,0.02991,0.02992,0.02993,0.02994,0.02995,0.02996,0.02997,0.02998,0.02999,0.03,0.03001,0.03002,0.03003,0.03004,0.03005,0.03006,0.03007,0.03008,0.03009,0.0301,0.03011,0.03012,0.03013,0.03014,0.03015,0.03016,0.03017,0.03018,0.03019,0.0302,0.03021,0.03022,0.03023,0.03024,0.03025,0.03026,0.03027,0.03028,0.03029,0.0303,0.03031,0.03032,0.03033,0.03034,0.03035,0.03036,0.03037,0.03038,0.03039,0.0304,0.03041,0.03042,0.03043,0.03044,0.03045,0.03046,0.03047,0.03048,0.03049,0.0305,0.03051,0.03052,0.03053,0.03054,0.03055,0.03056,0.03057,0.03058,0.03059,0.0306,0.03061,0.03062,0.03063,0.03064,0.03065,0.03066,0.03067,0.03068,0.03069,0.0307,0.03071,0.03072,0.03073,0.03074,0.03075,0.03076,0.03077,0.03078,0.03079,0.0308,0.03081,0.03082,0.03083,0.03084,0.03085,0.03086,0.03087,0.03088,0.03089,0.0309,0.03091,0.03092,0.03093,0.03094,0.03095,0.03096,0.03097,0.03098,0.03099,0.031,0.03101,0.03102,0.03103,0.03104,0.03105,0.03106,0.03107,0.03108,0.03109,0.0311,0.03111,0.03112,0.03113,0.03114,0.03115,0.03116,0.03117,0.03118,0.03119,0.0312,0.03121,0.03122,0.03123,0.03124,0.03125,0.03126,0.03127,0.03128,0.03129,0.0313,0.03131,0.03132,0.03133,0.03134,0.03135,0.03136,0.03137,0.03138,0.03139,0.0314,0.03141,0.03142,0.03143,0.03144,0.03145,0.03146,0.03147,0.03148,0.03149,0.0315,0.03151,0.03152,0.03153,0.03154,0.03155,0.03156,0.03157,0.03158,0.03159,0.0316,0.03161,0.03162,0.03163,0.03164,0.03165,0.03166,0.03167,0.03168,0.03169,0.0317,0.03171,0.03172,0.03173,0.03174,0.03175,0.03176,0.03177,0.03178,0.03179,0.0318,0.03181,0.03182,0.03183,0.03184,0.03185,0.03186,0.03187,0.03188,0.03189,0.0319,0.03191,0.03192,0.03193,0.03194,0.03195,0.03196,0.03197,0.03198,0.03199,0.032,0.03201,0.03202,0.03203,0.03204,0.03205,0.03206,0.03207,0.03208,0.03209,0.0321,0.03211,0.03212,0.03213,0.03214,0.03215,0.03216,0.03217,0.03218,0.03219,0.0322,0.03221,0.03222,0.03223,0.03224,0.03225,0.03226,0.03227,0.03228,0.03229,0.0323,0.03231,0.03232,0.03233,0.03234,0.03235,0.03236,0.03237,0.03238,0.03239,0.0324,0.03241,0.03242,0.03243,0.03244,0.03245,0.03246,0.03247,0.03248,0.03249,0.0325,0.03251,0.03252,0.03253,0.03254,0.03255,0.03256,0.03257,0.03258,0.03259,0.0326,0.03261,0.03262,0.03263,0.03264,0.03265,0.03266,0.03267,0.03268,0.03269,0.0327,0.03271,0.03272,0.03273,0.03274,0.03275,0.03276,0.03277,0.03278,0.03279,0.0328,0.03281,0.03282,0.03283,0.03284,0.03285,0.03286,0.03287,0.03288,0.03289,0.0329,0.03291,0.03292,0.03293,0.03294,0.03295,0.03296,0.03297,0.03298,0.03299,0.033,0.03301,0.03302,0.03303,0.03304,0.03305,0.03306,0.03307,0.03308,0.03309,0.0331,0.03311,0.03312,0.03313,0.03314,0.03315,0.03316,0.03317,0.03318,0.03319,0.0332,0.03321,0.03322,0.03323,0.03324,0.03325,0.03326,0.03327,0.03328,0.03329,0.0333,0.03331,0.03332,0.03333,0.03334,0.03335,0.03336,0.03337,0.03338,0.03339,0.0334,0.03341,0.03342,0.03343,0.03344,0.03345,0.03346,0.03347,0.03348,0.03349,0.0335,0.03351,0.03352,0.03353,0.03354,0.03355,0.03356,0.03357,0.03358,0.03359,0.0336,0.03361,0.03362,0.03363,0.03364,0.03365,0.03366,0.03367,0.03368,0.03369,0.0337,0.03371,0.03372,0.03373,0.03374,0.03375,0.03376,0.03377,0.03378,0.03379,0.0338,0.03381,0.03382,0.03383,0.03384,0.03385,0.03386,0.03387,0.03388,0.03389,0.0339,0.03391,0.03392,0.03393,0.03394,0.03395,0.03396,0.03397,0.03398,0.03399,0.034,0.03401,0.03402,0.03403,0.03404,0.03405,0.03406,0.03407,0.03408,0.03409,0.0341,0.03411,0.03412,0.03413,0.03414,0.03415,0.03416,0.03417,0.03418,0.03419,0.0342,0.03421,0.03422,0.03423,0.03424,0.03425,0.03426,0.03427,0.03428,0.03429,0.0343,0.03431,0.03432,0.03433,0.03434,0.03435,0.03436,0.03437,0.03438,0.03439,0.0344,0.03441,0.03442,0.03443,0.03444,0.03445,0.03446,0.03447,0.03448,0.03449,0.0345,0.03451,0.03452,0.03453,0.03454,0.03455,0.03456,0.03457,0.03458,0.03459,0.0346,0.03461,0.03462,0.03463,0.03464,0.03465,0.03466,0.03467,0.03468,0.03469,0.0347,0.03471,0.03472,0.03473,0.03474,0.03475,0.03476,0.03477,0.03478,0.03479,0.0348,0.03481,0.03482,0.03483,0.03484,0.03485,0.03486,0.03487,0.03488,0.03489,0.0349,0.03491,0.03492,0.03493,0.03494,0.03495,0.03496,0.03497,0.03498,0.03499,0.035,0.03501,0.03502,0.03503,0.03504,0.03505,0.03506,0.03507,0.03508,0.03509,0.0351,0.03511,0.03512,0.03513,0.03514,0.03515,0.03516,0.03517,0.03518,0.03519,0.0352,0.03521,0.03522,0.03523,0.03524,0.03525,0.03526,0.03527,0.03528,0.03529,0.0353,0.03531,0.03532,0.03533,0.03534,0.03535,0.03536,0.03537,0.03538,0.03539,0.0354,0.03541,0.03542,0.03543,0.03544,0.03545,0.03546,0.03547,0.03548,0.03549,0.0355,0.03551,0.03552,0.03553,0.03554,0.03555,0.03556,0.03557,0.03558,0.03559,0.0356,0.03561,0.03562,0.03563,0.03564,0.03565,0.03566,0.03567,0.03568,0.03569,0.0357,0.03571,0.03572,0.03573,0.03574,0.03575,0.03576,0.03577,0.03578,0.03579,0.0358,0.03581,0.03582,0.03583,0.03584,0.03585,0.03586,0.03587,0.03588,0.03589,0.0359,0.03591,0.03592,0.03593,0.03594,0.03595,0.03596,0.03597,0.03598,0.03599,0.036,0.03601,0.03602,0.03603,0.03604,0.03605,0.03606,0.03607,0.03608,0.03609,0.0361,0.03611,0.03612,0.03613,0.03614,0.03615,0.03616,0.03617,0.03618,0.03619,0.0362,0.03621,0.03622,0.03623,0.03624,0.03625,0.03626,0.03627,0.03628,0.03629,0.0363,0.03631,0.03632,0.03633,0.03634,0.03635,0.03636,0.03637,0.03638,0.03639,0.0364,0.03641,0.03642,0.03643,0.03644,0.03645,0.03646,0.03647,0.03648,0.03649,0.0365,0.03651,0.03652,0.03653,0.03654,0.03655,0.03656,0.03657,0.03658,0.03659,0.0366,0.03661,0.03662,0.03663,0.03664,0.03665,0.03666,0.03667,0.03668,0.03669,0.0367,0.03671,0.03672,0.03673,0.03674,0.03675,0.03676,0.03677,0.03678,0.03679,0.0368,0.03681,0.03682,0.03683,0.03684,0.03685,0.03686,0.03687,0.03688,0.03689,0.0369,0.03691,0.03692,0.03693,0.03694,0.03695,0.03696,0.03697,0.03698,0.03699,0.037,0.03701,0.03702,0.03703,0.03704,0.03705,0.03706,0.03707,0.03708,0.03709,0.0371,0.03711,0.03712,0.03713,0.03714,0.03715,0.03716,0.03717,0.03718,0.03719,0.0372,0.03721,0.03722,0.03723,0.03724,0.03725,0.03726,0.03727,0.03728,0.03729,0.0373,0.03731,0.03732,0.03733,0.03734,0.03735,0.03736,0.03737,0.03738,0.03739,0.0374,0.03741,0.03742,0.03743,0.03744,0.03745,0.03746,0.03747,0.03748,0.03749,0.0375,0.03751,0.03752,0.03753,0.03754,0.03755,0.03756,0.03757,0.03758,0.03759,0.0376,0.03761,0.03762,0.03763,0.03764,0.03765,0.03766,0.03767,0.03768,0.03769,0.0377,0.03771,0.03772,0.03773,0.03774,0.03775,0.03776,0.03777,0.03778,0.03779,0.0378,0.03781,0.03782,0.03783,0.03784,0.03785,0.03786,0.03787,0.03788,0.03789,0.0379,0.03791,0.03792,0.03793,0.03794,0.03795,0.03796,0.03797,0.03798,0.03799,0.038,0.03801,0.03802,0.03803,0.03804,0.03805,0.03806,0.03807,0.03808,0.03809,0.0381,0.03811,0.03812,0.03813,0.03814,0.03815,0.03816,0.03817,0.03818,0.03819,0.0382,0.03821,0.03822,0.03823,0.03824,0.03825,0.03826,0.03827,0.03828,0.03829,0.0383,0.03831,0.03832,0.03833,0.03834,0.03835,0.03836,0.03837,0.03838,0.03839,0.0384,0.03841,0.03842,0.03843,0.03844,0.03845,0.03846,0.03847,0.03848,0.03849,0.0385,0.03851,0.03852,0.03853,0.03854,0.03855,0.03856,0.03857,0.03858,0.03859,0.0386,0.03861,0.03862,0.03863,0.03864,0.03865,0.03866,0.03867,0.03868,0.03869,0.0387,0.03871,0.03872,0.03873,0.03874,0.03875,0.03876,0.03877,0.03878,0.03879,0.0388,0.03881,0.03882,0.03883,0.03884,0.03885,0.03886,0.03887,0.03888,0.03889,0.0389,0.03891,0.03892,0.03893,0.03894,0.03895,0.03896,0.03897,0.03898,0.03899,0.039,0.03901,0.03902,0.03903,0.03904,0.03905,0.03906,0.03907,0.03908,0.03909,0.0391,0.03911,0.03912,0.03913,0.03914,0.03915,0.03916,0.03917,0.03918,0.03919,0.0392,0.03921,0.03922,0.03923,0.03924,0.03925,0.03926,0.03927,0.03928,0.03929,0.0393,0.03931,0.03932,0.03933,0.03934,0.03935,0.03936,0.03937,0.03938,0.03939,0.0394,0.03941,0.03942,0.03943,0.03944,0.03945,0.03946,0.03947,0.03948,0.03949,0.0395,0.03951,0.03952,0.03953,0.03954,0.03955,0.03956,0.03957,0.03958,0.03959,0.0396,0.03961,0.03962,0.03963,0.03964,0.03965,0.03966,0.03967,0.03968,0.03969,0.0397,0.03971,0.03972,0.03973,0.03974,0.03975,0.03976,0.03977,0.03978,0.03979,0.0398,0.03981,0.03982,0.03983,0.03984,0.03985,0.03986,0.03987,0.03988,0.03989,0.0399,0.03991,0.03992,0.03993,0.03994,0.03995,0.03996,0.03997,0.03998,0.03999,0.04,0.04001,0.04002,0.04003,0.04004,0.04005,0.04006,0.04007,0.04008,0.04009,0.0401,0.04011,0.04012,0.04013,0.04014,0.04015,0.04016,0.04017,0.04018,0.04019,0.0402,0.04021,0.04022,0.04023,0.04024,0.04025,0.04026,0.04027,0.04028,0.04029,0.0403,0.04031,0.04032,0.04033,0.04034,0.04035,0.04036,0.04037,0.04038,0.04039,0.0404,0.04041,0.04042,0.04043,0.04044,0.04045,0.04046,0.04047,0.04048,0.04049,0.0405,0.04051,0.04052,0.04053,0.04054,0.04055,0.04056,0.04057,0.04058,0.04059,0.0406,0.04061,0.04062,0.04063,0.04064,0.04065,0.04066,0.04067,0.04068,0.04069,0.0407,0.04071,0.04072,0.04073,0.04074,0.04075,0.04076,0.04077,0.04078,0.04079,0.0408,0.04081,0.04082,0.04083,0.04084,0.04085,0.04086,0.04087,0.04088,0.04089,0.0409,0.04091,0.04092,0.04093,0.04094,0.04095,0.04096,0.04097,0.04098,0.04099,0.041,0.04101,0.04102,0.04103,0.04104,0.04105,0.04106,0.04107,0.04108,0.04109,0.0411,0.04111,0.04112,0.04113,0.04114,0.04115,0.04116,0.04117,0.04118,0.04119,0.0412,0.04121,0.04122,0.04123,0.04124,0.04125,0.04126,0.04127,0.04128,0.04129,0.0413,0.04131,0.04132,0.04133,0.04134,0.04135,0.04136,0.04137,0.04138,0.04139,0.0414,0.04141,0.04142,0.04143,0.04144,0.04145,0.04146,0.04147,0.04148,0.04149,0.0415,0.04151,0.04152,0.04153,0.04154,0.04155,0.04156,0.04157,0.04158,0.04159,0.0416,0.04161,0.04162,0.04163,0.04164,0.04165,0.04166,0.04167,0.04168,0.04169,0.0417,0.04171,0.04172,0.04173,0.04174,0.04175,0.04176,0.04177,0.04178,0.04179,0.0418,0.04181,0.04182,0.04183,0.04184,0.04185,0.04186,0.04187,0.04188,0.04189,0.0419,0.04191,0.04192,0.04193,0.04194,0.04195,0.04196,0.04197,0.04198,0.04199,0.042,0.04201,0.04202,0.04203,0.04204,0.04205,0.04206,0.04207,0.04208,0.04209,0.0421,0.04211,0.04212,0.04213,0.04214,0.04215,0.04216,0.04217,0.04218,0.04219,0.0422,0.04221,0.04222,0.04223,0.04224,0.04225,0.04226,0.04227,0.04228,0.04229,0.0423,0.04231,0.04232,0.04233,0.04234,0.04235,0.04236,0.04237,0.04238,0.04239,0.0424,0.04241,0.04242,0.04243,0.04244,0.04245,0.04246,0.04247,0.04248,0.04249,0.0425,0.04251,0.04252,0.04253,0.04254,0.04255,0.04256,0.04257,0.04258,0.04259,0.0426,0.04261,0.04262,0.04263,0.04264,0.04265,0.04266,0.04267,0.04268,0.04269,0.0427,0.04271,0.04272,0.04273,0.04274,0.04275,0.04276,0.04277,0.04278,0.04279,0.0428,0.04281,0.04282,0.04283,0.04284,0.04285,0.04286,0.04287,0.04288,0.04289,0.0429,0.04291,0.04292,0.04293,0.04294,0.04295,0.04296,0.04297,0.04298,0.04299,0.043,0.04301,0.04302,0.04303,0.04304,0.04305,0.04306,0.04307,0.04308,0.04309,0.0431,0.04311,0.04312,0.04313,0.04314,0.04315,0.04316,0.04317,0.04318,0.04319,0.0432,0.04321,0.04322,0.04323,0.04324,0.04325,0.04326,0.04327,0.04328,0.04329,0.0433,0.04331,0.04332,0.04333,0.04334,0.04335,0.04336,0.04337,0.04338,0.04339,0.0434,0.04341,0.04342,0.04343,0.04344,0.04345,0.04346,0.04347,0.04348,0.04349,0.0435,0.04351,0.04352,0.04353,0.04354,0.04355,0.04356,0.04357,0.04358,0.04359,0.0436,0.04361,0.04362,0.04363,0.04364,0.04365,0.04366,0.04367,0.04368,0.04369,0.0437,0.04371,0.04372,0.04373,0.04374,0.04375,0.04376,0.04377,0.04378,0.04379,0.0438,0.04381,0.04382,0.04383,0.04384,0.04385,0.04386,0.04387,0.04388,0.04389,0.0439,0.04391,0.04392,0.04393,0.04394,0.04395,0.04396,0.04397,0.04398,0.04399,0.044,0.04401,0.04402,0.04403,0.04404,0.04405,0.04406,0.04407,0.04408,0.04409,0.0441,0.04411,0.04412,0.04413,0.04414,0.04415,0.04416,0.04417,0.04418,0.04419,0.0442,0.04421,0.04422,0.04423,0.04424,0.04425,0.04426,0.04427,0.04428,0.04429,0.0443,0.04431,0.04432,0.04433,0.04434,0.04435,0.04436,0.04437,0.04438,0.04439,0.0444,0.04441,0.04442,0.04443,0.04444,0.04445,0.04446,0.04447,0.04448,0.04449,0.0445,0.04451,0.04452,0.04453,0.04454,0.04455,0.04456,0.04457,0.04458,0.04459,0.0446,0.04461,0.04462,0.04463,0.04464,0.04465,0.04466,0.04467,0.04468,0.04469,0.0447,0.04471,0.04472,0.04473,0.04474,0.04475,0.04476,0.04477,0.04478,0.04479,0.0448,0.04481,0.04482,0.04483,0.04484,0.04485,0.04486,0.04487,0.04488,0.04489,0.0449,0.04491,0.04492,0.04493,0.04494,0.04495,0.04496,0.04497,0.04498,0.04499,0.045,0.04501,0.04502,0.04503,0.04504,0.04505,0.04506,0.04507,0.04508,0.04509,0.0451,0.04511,0.04512,0.04513,0.04514,0.04515,0.04516,0.04517,0.04518,0.04519,0.0452,0.04521,0.04522,0.04523,0.04524,0.04525,0.04526,0.04527,0.04528,0.04529,0.0453,0.04531,0.04532,0.04533,0.04534,0.04535,0.04536,0.04537,0.04538,0.04539,0.0454,0.04541,0.04542,0.04543,0.04544,0.04545,0.04546,0.04547,0.04548,0.04549,0.0455,0.04551,0.04552,0.04553,0.04554,0.04555,0.04556,0.04557,0.04558,0.04559,0.0456,0.04561,0.04562,0.04563,0.04564,0.04565,0.04566,0.04567,0.04568,0.04569,0.0457,0.04571,0.04572,0.04573,0.04574,0.04575,0.04576,0.04577,0.04578,0.04579,0.0458,0.04581,0.04582,0.04583,0.04584,0.04585,0.04586,0.04587,0.04588,0.04589,0.0459,0.04591,0.04592,0.04593,0.04594,0.04595,0.04596,0.04597,0.04598,0.04599,0.046,0.04601,0.04602,0.04603,0.04604,0.04605,0.04606,0.04607,0.04608,0.04609,0.0461,0.04611,0.04612,0.04613,0.04614,0.04615,0.04616,0.04617,0.04618,0.04619,0.0462,0.04621,0.04622,0.04623,0.04624,0.04625,0.04626,0.04627,0.04628,0.04629,0.0463,0.04631,0.04632,0.04633,0.04634,0.04635,0.04636,0.04637,0.04638,0.04639,0.0464,0.04641,0.04642,0.04643,0.04644,0.04645,0.04646,0.04647,0.04648,0.04649,0.0465,0.04651,0.04652,0.04653,0.04654,0.04655,0.04656,0.04657,0.04658,0.04659,0.0466,0.04661,0.04662,0.04663,0.04664,0.04665,0.04666,0.04667,0.04668,0.04669,0.0467,0.04671,0.04672,0.04673,0.04674,0.04675,0.04676,0.04677,0.04678,0.04679,0.0468,0.04681,0.04682,0.04683,0.04684,0.04685,0.04686,0.04687,0.04688,0.04689,0.0469,0.04691,0.04692,0.04693,0.04694,0.04695,0.04696,0.04697,0.04698,0.04699,0.047,0.04701,0.04702,0.04703,0.04704,0.04705,0.04706,0.04707,0.04708,0.04709,0.0471,0.04711,0.04712,0.04713,0.04714,0.04715,0.04716,0.04717,0.04718,0.04719,0.0472,0.04721,0.04722,0.04723,0.04724,0.04725,0.04726,0.04727,0.04728,0.04729,0.0473,0.04731,0.04732,0.04733,0.04734,0.04735,0.04736,0.04737,0.04738,0.04739,0.0474,0.04741,0.04742,0.04743,0.04744,0.04745,0.04746,0.04747,0.04748,0.04749,0.0475,0.04751,0.04752,0.04753,0.04754,0.04755,0.04756,0.04757,0.04758,0.04759,0.0476,0.04761,0.04762,0.04763,0.04764,0.04765,0.04766,0.04767,0.04768,0.04769,0.0477,0.04771,0.04772,0.04773,0.04774,0.04775,0.04776,0.04777,0.04778,0.04779,0.0478,0.04781,0.04782,0.04783,0.04784,0.04785,0.04786,0.04787,0.04788,0.04789,0.0479,0.04791,0.04792,0.04793,0.04794,0.04795,0.04796,0.04797,0.04798,0.04799,0.048,0.04801,0.04802,0.04803,0.04804,0.04805,0.04806,0.04807,0.04808,0.04809,0.0481,0.04811,0.04812,0.04813,0.04814,0.04815,0.04816,0.04817,0.04818,0.04819,0.0482,0.04821,0.04822,0.04823,0.04824,0.04825,0.04826,0.04827,0.04828,0.04829,0.0483,0.04831,0.04832,0.04833,0.04834,0.04835,0.04836,0.04837,0.04838,0.04839,0.0484,0.04841,0.04842,0.04843,0.04844,0.04845,0.04846,0.04847,0.04848,0.04849,0.0485,0.04851,0.04852,0.04853,0.04854,0.04855,0.04856,0.04857,0.04858,0.04859,0.0486,0.04861,0.04862,0.04863,0.04864,0.04865,0.04866,0.04867,0.04868,0.04869,0.0487,0.04871,0.04872,0.04873,0.04874,0.04875,0.04876,0.04877,0.04878,0.04879,0.0488,0.04881,0.04882,0.04883,0.04884,0.04885,0.04886,0.04887,0.04888,0.04889,0.0489,0.04891,0.04892,0.04893,0.04894,0.04895,0.04896,0.04897,0.04898,0.04899,0.049,0.04901,0.04902,0.04903,0.04904,0.04905,0.04906,0.04907,0.04908,0.04909,0.0491,0.04911,0.04912,0.04913,0.04914,0.04915,0.04916,0.04917,0.04918,0.04919,0.0492,0.04921,0.04922,0.04923,0.04924,0.04925,0.04926,0.04927,0.04928,0.04929,0.0493,0.04931,0.04932,0.04933,0.04934,0.04935,0.04936,0.04937,0.04938,0.04939,0.0494,0.04941,0.04942,0.04943,0.04944,0.04945,0.04946,0.04947,0.04948,0.04949,0.0495,0.04951,0.04952,0.04953,0.04954,0.04955,0.04956,0.04957,0.04958,0.04959,0.0496,0.04961,0.04962,0.04963,0.04964,0.04965,0.04966,0.04967,0.04968,0.04969,0.0497,0.04971,0.04972,0.04973,0.04974,0.04975,0.04976,0.04977,0.04978,0.04979,0.0498,0.04981,0.04982,0.04983,0.04984,0.04985,0.04986,0.04987,0.04988,0.04989,0.0499,0.04991,0.04992,0.04993,0.04994,0.04995,0.04996,0.04997,0.04998,0.04999,0.05,0.05001,0.05002,0.05003,0.05004,0.05005,0.05006,0.05007,0.05008,0.05009,0.0501,0.05011,0.05012,0.05013,0.05014,0.05015,0.05016,0.05017,0.05018,0.05019,0.0502,0.05021,0.05022,0.05023,0.05024,0.05025,0.05026,0.05027,0.05028,0.05029,0.0503,0.05031,0.05032,0.05033,0.05034,0.05035,0.05036,0.05037,0.05038,0.05039,0.0504,0.05041,0.05042,0.05043,0.05044,0.05045,0.05046,0.05047,0.05048,0.05049,0.0505,0.05051,0.05052,0.05053,0.05054,0.05055,0.05056,0.05057,0.05058,0.05059,0.0506,0.05061,0.05062,0.05063,0.05064,0.05065,0.05066,0.05067,0.05068,0.05069,0.0507,0.05071,0.05072,0.05073,0.05074,0.05075,0.05076,0.05077,0.05078,0.05079,0.0508,0.05081,0.05082,0.05083,0.05084,0.05085,0.05086,0.05087,0.05088,0.05089,0.0509,0.05091,0.05092,0.05093,0.05094,0.05095,0.05096,0.05097,0.05098,0.05099,0.051,0.05101,0.05102,0.05103,0.05104,0.05105,0.05106,0.05107,0.05108,0.05109,0.0511,0.05111,0.05112,0.05113,0.05114,0.05115,0.05116,0.05117,0.05118,0.05119,0.0512,0.05121,0.05122,0.05123,0.05124,0.05125,0.05126,0.05127,0.05128,0.05129,0.0513,0.05131,0.05132,0.05133,0.05134,0.05135,0.05136,0.05137,0.05138,0.05139,0.0514,0.05141,0.05142,0.05143,0.05144,0.05145,0.05146,0.05147,0.05148,0.05149,0.0515,0.05151,0.05152,0.05153,0.05154,0.05155,0.05156,0.05157,0.05158,0.05159,0.0516,0.05161,0.05162,0.05163,0.05164,0.05165,0.05166,0.05167,0.05168,0.05169,0.0517,0.05171,0.05172,0.05173,0.05174,0.05175,0.05176,0.05177,0.05178,0.05179,0.0518,0.05181,0.05182,0.05183,0.05184,0.05185,0.05186,0.05187,0.05188,0.05189,0.0519,0.05191,0.05192,0.05193,0.05194,0.05195,0.05196,0.05197,0.05198,0.05199,0.052,0.05201,0.05202,0.05203,0.05204,0.05205,0.05206,0.05207,0.05208,0.05209,0.0521,0.05211,0.05212,0.05213,0.05214,0.05215,0.05216,0.05217,0.05218,0.05219,0.0522,0.05221,0.05222,0.05223,0.05224,0.05225,0.05226,0.05227,0.05228,0.05229,0.0523,0.05231,0.05232,0.05233,0.05234,0.05235,0.05236,0.05237,0.05238,0.05239,0.0524,0.05241,0.05242,0.05243,0.05244,0.05245,0.05246,0.05247,0.05248,0.05249,0.0525,0.05251,0.05252,0.05253,0.05254,0.05255,0.05256,0.05257,0.05258,0.05259,0.0526,0.05261,0.05262,0.05263,0.05264,0.05265,0.05266,0.05267,0.05268,0.05269,0.0527,0.05271,0.05272,0.05273,0.05274,0.05275,0.05276,0.05277,0.05278,0.05279,0.0528,0.05281,0.05282,0.05283,0.05284,0.05285,0.05286,0.05287,0.05288,0.05289,0.0529,0.05291,0.05292,0.05293,0.05294,0.05295,0.05296,0.05297,0.05298,0.05299,0.053,0.05301,0.05302,0.05303,0.05304,0.05305,0.05306,0.05307,0.05308,0.05309,0.0531,0.05311,0.05312,0.05313,0.05314,0.05315,0.05316,0.05317,0.05318,0.05319,0.0532,0.05321,0.05322,0.05323,0.05324,0.05325,0.05326,0.05327,0.05328,0.05329,0.0533,0.05331,0.05332,0.05333,0.05334,0.05335,0.05336,0.05337,0.05338,0.05339,0.0534,0.05341,0.05342,0.05343,0.05344,0.05345,0.05346,0.05347,0.05348,0.05349,0.0535,0.05351,0.05352,0.05353,0.05354,0.05355,0.05356,0.05357,0.05358,0.05359,0.0536,0.05361,0.05362,0.05363,0.05364,0.05365,0.05366,0.05367,0.05368,0.05369,0.0537,0.05371,0.05372,0.05373,0.05374,0.05375,0.05376,0.05377,0.05378,0.05379,0.0538,0.05381,0.05382,0.05383,0.05384,0.05385,0.05386,0.05387,0.05388,0.05389,0.0539,0.05391,0.05392,0.05393,0.05394,0.05395,0.05396,0.05397,0.05398,0.05399,0.054,0.05401,0.05402,0.05403,0.05404,0.05405,0.05406,0.05407,0.05408,0.05409,0.0541,0.05411,0.05412,0.05413,0.05414,0.05415,0.05416,0.05417,0.05418,0.05419,0.0542,0.05421,0.05422,0.05423,0.05424,0.05425,0.05426,0.05427,0.05428,0.05429,0.0543,0.05431,0.05432,0.05433,0.05434,0.05435,0.05436,0.05437,0.05438,0.05439,0.0544,0.05441,0.05442,0.05443,0.05444,0.05445,0.05446,0.05447,0.05448,0.05449,0.0545,0.05451,0.05452,0.05453,0.05454,0.05455,0.05456,0.05457,0.05458,0.05459,0.0546,0.05461,0.05462,0.05463,0.05464,0.05465,0.05466,0.05467,0.05468,0.05469,0.0547,0.05471,0.05472,0.05473,0.05474,0.05475,0.05476,0.05477,0.05478,0.05479,0.0548,0.05481,0.05482,0.05483,0.05484,0.05485,0.05486,0.05487,0.05488,0.05489,0.0549,0.05491,0.05492,0.05493,0.05494,0.05495,0.05496,0.05497,0.05498,0.05499,0.055,0.05501,0.05502,0.05503,0.05504,0.05505,0.05506,0.05507,0.05508,0.05509,0.0551,0.05511,0.05512,0.05513,0.05514,0.05515,0.05516,0.05517,0.05518,0.05519,0.0552,0.05521,0.05522,0.05523,0.05524,0.05525,0.05526,0.05527,0.05528,0.05529,0.0553,0.05531,0.05532,0.05533,0.05534,0.05535,0.05536,0.05537,0.05538,0.05539,0.0554,0.05541,0.05542,0.05543,0.05544,0.05545,0.05546,0.05547,0.05548,0.05549,0.0555,0.05551,0.05552,0.05553,0.05554,0.05555,0.05556,0.05557,0.05558,0.05559,0.0556,0.05561,0.05562,0.05563,0.05564,0.05565,0.05566,0.05567,0.05568,0.05569,0.0557,0.05571,0.05572,0.05573,0.05574,0.05575,0.05576,0.05577,0.05578,0.05579,0.0558,0.05581,0.05582,0.05583,0.05584,0.05585,0.05586,0.05587,0.05588,0.05589,0.0559,0.05591,0.05592,0.05593,0.05594,0.05595,0.05596,0.05597,0.05598,0.05599,0.056,0.05601,0.05602,0.05603,0.05604,0.05605,0.05606,0.05607,0.05608,0.05609,0.0561,0.05611,0.05612,0.05613,0.05614,0.05615,0.05616,0.05617,0.05618,0.05619,0.0562,0.05621,0.05622,0.05623,0.05624,0.05625,0.05626,0.05627,0.05628,0.05629,0.0563,0.05631,0.05632,0.05633,0.05634,0.05635,0.05636,0.05637,0.05638,0.05639,0.0564,0.05641,0.05642,0.05643,0.05644,0.05645,0.05646,0.05647,0.05648,0.05649,0.0565,0.05651,0.05652,0.05653,0.05654,0.05655,0.05656,0.05657,0.05658,0.05659,0.0566,0.05661,0.05662,0.05663,0.05664,0.05665,0.05666,0.05667,0.05668,0.05669,0.0567,0.05671,0.05672,0.05673,0.05674,0.05675,0.05676,0.05677,0.05678,0.05679,0.0568,0.05681,0.05682,0.05683,0.05684,0.05685,0.05686,0.05687,0.05688,0.05689,0.0569,0.05691,0.05692,0.05693,0.05694,0.05695,0.05696,0.05697,0.05698,0.05699,0.057,0.05701,0.05702,0.05703,0.05704,0.05705,0.05706,0.05707,0.05708,0.05709,0.0571,0.05711,0.05712,0.05713,0.05714,0.05715,0.05716,0.05717,0.05718,0.05719,0.0572,0.05721,0.05722,0.05723,0.05724,0.05725,0.05726,0.05727,0.05728,0.05729,0.0573,0.05731,0.05732,0.05733,0.05734,0.05735,0.05736,0.05737,0.05738,0.05739,0.0574,0.05741,0.05742,0.05743,0.05744,0.05745,0.05746,0.05747,0.05748,0.05749,0.0575,0.05751,0.05752,0.05753,0.05754,0.05755,0.05756,0.05757,0.05758,0.05759,0.0576,0.05761,0.05762,0.05763,0.05764,0.05765,0.05766,0.05767,0.05768,0.05769,0.0577,0.05771,0.05772,0.05773,0.05774,0.05775,0.05776,0.05777,0.05778,0.05779,0.0578,0.05781,0.05782,0.05783,0.05784,0.05785,0.05786,0.05787,0.05788,0.05789,0.0579,0.05791,0.05792,0.05793,0.05794,0.05795,0.05796,0.05797,0.05798,0.05799,0.058,0.05801,0.05802,0.05803,0.05804,0.05805,0.05806,0.05807,0.05808,0.05809,0.0581,0.05811,0.05812,0.05813,0.05814,0.05815,0.05816,0.05817,0.05818,0.05819,0.0582,0.05821,0.05822,0.05823,0.05824,0.05825,0.05826,0.05827,0.05828,0.05829,0.0583,0.05831,0.05832,0.05833,0.05834,0.05835,0.05836,0.05837,0.05838,0.05839,0.0584,0.05841,0.05842,0.05843,0.05844,0.05845,0.05846,0.05847,0.05848,0.05849,0.0585,0.05851,0.05852,0.05853,0.05854,0.05855,0.05856,0.05857,0.05858,0.05859,0.0586,0.05861,0.05862,0.05863,0.05864,0.05865,0.05866,0.05867,0.05868,0.05869,0.0587,0.05871,0.05872,0.05873,0.05874,0.05875,0.05876,0.05877,0.05878,0.05879,0.0588,0.05881,0.05882,0.05883,0.05884,0.05885,0.05886,0.05887,0.05888,0.05889,0.0589,0.05891,0.05892,0.05893,0.05894,0.05895,0.05896,0.05897,0.05898,0.05899,0.059,0.05901,0.05902,0.05903,0.05904,0.05905,0.05906,0.05907,0.05908,0.05909,0.0591,0.05911,0.05912,0.05913,0.05914,0.05915,0.05916,0.05917,0.05918,0.05919,0.0592,0.05921,0.05922,0.05923,0.05924,0.05925,0.05926,0.05927,0.05928,0.05929,0.0593,0.05931,0.05932,0.05933,0.05934,0.05935,0.05936,0.05937,0.05938,0.05939,0.0594,0.05941,0.05942,0.05943,0.05944,0.05945,0.05946,0.05947,0.05948,0.05949,0.0595,0.05951,0.05952,0.05953,0.05954,0.05955,0.05956,0.05957,0.05958,0.05959,0.0596,0.05961,0.05962,0.05963,0.05964,0.05965,0.05966,0.05967,0.05968,0.05969,0.0597,0.05971,0.05972,0.05973,0.05974,0.05975,0.05976,0.05977,0.05978,0.05979,0.0598,0.05981,0.05982,0.05983,0.05984,0.05985,0.05986,0.05987,0.05988,0.05989,0.0599,0.05991,0.05992,0.05993,0.05994,0.05995,0.05996,0.05997,0.05998,0.05999,0.06,0.06001,0.06002,0.06003,0.06004,0.06005,0.06006,0.06007,0.06008,0.06009,0.0601,0.06011,0.06012,0.06013,0.06014,0.06015,0.06016,0.06017,0.06018,0.06019,0.0602,0.06021,0.06022,0.06023,0.06024,0.06025,0.06026,0.06027,0.06028,0.06029,0.0603,0.06031,0.06032,0.06033,0.06034,0.06035,0.06036,0.06037,0.06038,0.06039,0.0604,0.06041,0.06042,0.06043,0.06044,0.06045,0.06046,0.06047,0.06048,0.06049,0.0605,0.06051,0.06052,0.06053,0.06054,0.06055,0.06056,0.06057,0.06058,0.06059,0.0606,0.06061,0.06062,0.06063,0.06064,0.06065,0.06066,0.06067,0.06068,0.06069,0.0607,0.06071,0.06072,0.06073,0.06074,0.06075,0.06076,0.06077,0.06078,0.06079,0.0608,0.06081,0.06082,0.06083,0.06084,0.06085,0.06086,0.06087,0.06088,0.06089,0.0609,0.06091,0.06092,0.06093,0.06094,0.06095,0.06096,0.06097,0.06098,0.06099,0.061,0.06101,0.06102,0.06103,0.06104,0.06105,0.06106,0.06107,0.06108,0.06109,0.0611,0.06111,0.06112,0.06113,0.06114,0.06115,0.06116,0.06117,0.06118,0.06119,0.0612,0.06121,0.06122,0.06123,0.06124,0.06125,0.06126,0.06127,0.06128,0.06129,0.0613,0.06131,0.06132,0.06133,0.06134,0.06135,0.06136,0.06137,0.06138,0.06139,0.0614,0.06141,0.06142,0.06143,0.06144,0.06145,0.06146,0.06147,0.06148,0.06149,0.0615,0.06151,0.06152,0.06153,0.06154,0.06155,0.06156,0.06157,0.06158,0.06159,0.0616,0.06161,0.06162,0.06163,0.06164,0.06165,0.06166,0.06167,0.06168,0.06169,0.0617,0.06171,0.06172,0.06173,0.06174,0.06175,0.06176,0.06177,0.06178,0.06179,0.0618,0.06181,0.06182,0.06183,0.06184,0.06185,0.06186,0.06187,0.06188,0.06189,0.0619,0.06191,0.06192,0.06193,0.06194,0.06195,0.06196,0.06197,0.06198,0.06199,0.062,0.06201,0.06202,0.06203,0.06204,0.06205,0.06206,0.06207,0.06208,0.06209,0.0621,0.06211,0.06212,0.06213,0.06214,0.06215,0.06216,0.06217,0.06218,0.06219,0.0622,0.06221,0.06222,0.06223,0.06224,0.06225,0.06226,0.06227,0.06228,0.06229,0.0623,0.06231,0.06232,0.06233,0.06234,0.06235,0.06236,0.06237,0.06238,0.06239,0.0624,0.06241,0.06242,0.06243,0.06244,0.06245,0.06246,0.06247,0.06248,0.06249,0.0625,0.06251,0.06252,0.06253,0.06254,0.06255,0.06256,0.06257,0.06258,0.06259,0.0626,0.06261,0.06262,0.06263,0.06264,0.06265,0.06266,0.06267,0.06268,0.06269,0.0627,0.06271,0.06272,0.06273,0.06274,0.06275,0.06276,0.06277,0.06278,0.06279,0.0628,0.06281,0.06282,0.06283,0.06284,0.06285,0.06286,0.06287,0.06288,0.06289,0.0629,0.06291,0.06292,0.06293,0.06294,0.06295,0.06296,0.06297,0.06298,0.06299,0.063,0.06301,0.06302,0.06303,0.06304,0.06305,0.06306,0.06307,0.06308,0.06309,0.0631,0.06311,0.06312,0.06313,0.06314,0.06315,0.06316,0.06317,0.06318,0.06319,0.0632,0.06321,0.06322,0.06323,0.06324,0.06325,0.06326,0.06327,0.06328,0.06329,0.0633,0.06331,0.06332,0.06333,0.06334,0.06335,0.06336,0.06337,0.06338,0.06339,0.0634,0.06341,0.06342,0.06343,0.06344,0.06345,0.06346,0.06347,0.06348,0.06349,0.0635,0.06351,0.06352,0.06353,0.06354,0.06355,0.06356,0.06357,0.06358,0.06359,0.0636,0.06361,0.06362,0.06363,0.06364,0.06365,0.06366,0.06367,0.06368,0.06369,0.0637,0.06371,0.06372,0.06373,0.06374,0.06375,0.06376,0.06377,0.06378,0.06379,0.0638,0.06381,0.06382,0.06383,0.06384,0.06385,0.06386,0.06387,0.06388,0.06389,0.0639,0.06391,0.06392,0.06393,0.06394,0.06395,0.06396,0.06397,0.06398,0.06399,0.064,0.06401,0.06402,0.06403,0.06404,0.06405,0.06406,0.06407,0.06408,0.06409,0.0641,0.06411,0.06412,0.06413,0.06414,0.06415,0.06416,0.06417,0.06418,0.06419,0.0642,0.06421,0.06422,0.06423,0.06424,0.06425,0.06426,0.06427,0.06428,0.06429,0.0643,0.06431,0.06432,0.06433,0.06434,0.06435,0.06436,0.06437,0.06438,0.06439,0.0644,0.06441,0.06442,0.06443,0.06444,0.06445,0.06446,0.06447,0.06448,0.06449,0.0645,0.06451,0.06452,0.06453,0.06454,0.06455,0.06456,0.06457,0.06458,0.06459,0.0646,0.06461,0.06462,0.06463,0.06464,0.06465,0.06466,0.06467,0.06468,0.06469,0.0647,0.06471,0.06472,0.06473,0.06474,0.06475,0.06476,0.06477,0.06478,0.06479,0.0648,0.06481,0.06482,0.06483,0.06484,0.06485,0.06486,0.06487,0.06488,0.06489,0.0649,0.06491,0.06492,0.06493,0.06494,0.06495,0.06496,0.06497,0.06498,0.06499,0.065,0.06501,0.06502,0.06503,0.06504,0.06505,0.06506,0.06507,0.06508,0.06509,0.0651,0.06511,0.06512,0.06513,0.06514,0.06515,0.06516,0.06517,0.06518,0.06519,0.0652,0.06521,0.06522,0.06523,0.06524,0.06525,0.06526,0.06527,0.06528,0.06529,0.0653,0.06531,0.06532,0.06533,0.06534,0.06535,0.06536,0.06537,0.06538,0.06539,0.0654,0.06541,0.06542,0.06543,0.06544,0.06545,0.06546,0.06547,0.06548,0.06549,0.0655,0.06551,0.06552,0.06553,0.06554,0.06555,0.06556,0.06557,0.06558,0.06559,0.0656,0.06561,0.06562,0.06563,0.06564,0.06565,0.06566,0.06567,0.06568,0.06569,0.0657,0.06571,0.06572,0.06573,0.06574,0.06575,0.06576,0.06577,0.06578,0.06579,0.0658,0.06581,0.06582,0.06583,0.06584,0.06585,0.06586,0.06587,0.06588,0.06589,0.0659,0.06591,0.06592,0.06593,0.06594,0.06595,0.06596,0.06597,0.06598,0.06599,0.066,0.06601,0.06602,0.06603,0.06604,0.06605,0.06606,0.06607,0.06608,0.06609,0.0661,0.06611,0.06612,0.06613,0.06614,0.06615,0.06616,0.06617,0.06618,0.06619,0.0662,0.06621,0.06622,0.06623,0.06624,0.06625,0.06626,0.06627,0.06628,0.06629,0.0663,0.06631,0.06632,0.06633,0.06634,0.06635,0.06636,0.06637,0.06638,0.06639,0.0664,0.06641,0.06642,0.06643,0.06644,0.06645,0.06646,0.06647,0.06648,0.06649,0.0665,0.06651,0.06652,0.06653,0.06654,0.06655,0.06656,0.06657,0.06658,0.06659,0.0666,0.06661,0.06662,0.06663,0.06664,0.06665,0.06666,0.06667,0.06668,0.06669,0.0667,0.06671,0.06672,0.06673,0.06674,0.06675,0.06676,0.06677,0.06678,0.06679,0.0668,0.06681,0.06682,0.06683,0.06684,0.06685,0.06686,0.06687,0.06688,0.06689,0.0669,0.06691,0.06692,0.06693,0.06694,0.06695,0.06696,0.06697,0.06698,0.06699,0.067,0.06701,0.06702,0.06703,0.06704,0.06705,0.06706,0.06707,0.06708,0.06709,0.0671,0.06711,0.06712,0.06713,0.06714,0.06715,0.06716,0.06717,0.06718,0.06719,0.0672,0.06721,0.06722,0.06723,0.06724,0.06725,0.06726,0.06727,0.06728,0.06729,0.0673,0.06731,0.06732,0.06733,0.06734,0.06735,0.06736,0.06737,0.06738,0.06739,0.0674,0.06741,0.06742,0.06743,0.06744,0.06745,0.06746,0.06747,0.06748,0.06749,0.0675,0.06751,0.06752,0.06753,0.06754,0.06755,0.06756,0.06757,0.06758,0.06759,0.0676,0.06761,0.06762,0.06763,0.06764,0.06765,0.06766,0.06767,0.06768,0.06769,0.0677,0.06771,0.06772,0.06773,0.06774,0.06775,0.06776,0.06777,0.06778,0.06779,0.0678,0.06781,0.06782,0.06783,0.06784,0.06785,0.06786,0.06787,0.06788,0.06789,0.0679,0.06791,0.06792,0.06793,0.06794,0.06795,0.06796,0.06797,0.06798,0.06799,0.068,0.06801,0.06802,0.06803,0.06804,0.06805,0.06806,0.06807,0.06808,0.06809,0.0681,0.06811,0.06812,0.06813,0.06814,0.06815,0.06816,0.06817,0.06818,0.06819,0.0682,0.06821,0.06822,0.06823,0.06824,0.06825,0.06826,0.06827,0.06828,0.06829,0.0683,0.06831,0.06832,0.06833,0.06834,0.06835,0.06836,0.06837,0.06838,0.06839,0.0684,0.06841,0.06842,0.06843,0.06844,0.06845,0.06846,0.06847,0.06848,0.06849,0.0685,0.06851,0.06852,0.06853,0.06854,0.06855,0.06856,0.06857,0.06858,0.06859,0.0686,0.06861,0.06862,0.06863,0.06864,0.06865,0.06866,0.06867,0.06868,0.06869,0.0687,0.06871,0.06872,0.06873,0.06874,0.06875,0.06876,0.06877,0.06878,0.06879,0.0688,0.06881,0.06882,0.06883,0.06884,0.06885,0.06886,0.06887,0.06888,0.06889,0.0689,0.06891,0.06892,0.06893,0.06894,0.06895,0.06896,0.06897,0.06898,0.06899,0.069,0.06901,0.06902,0.06903,0.06904,0.06905,0.06906,0.06907,0.06908,0.06909,0.0691,0.06911,0.06912,0.06913,0.06914,0.06915,0.06916,0.06917,0.06918,0.06919,0.0692,0.06921,0.06922,0.06923,0.06924,0.06925,0.06926,0.06927,0.06928,0.06929,0.0693,0.06931,0.06932,0.06933,0.06934,0.06935,0.06936,0.06937,0.06938,0.06939,0.0694,0.06941,0.06942,0.06943,0.06944,0.06945,0.06946,0.06947,0.06948,0.06949,0.0695,0.06951,0.06952,0.06953,0.06954,0.06955,0.06956,0.06957,0.06958,0.06959,0.0696,0.06961,0.06962,0.06963,0.06964,0.06965,0.06966,0.06967,0.06968,0.06969,0.0697,0.06971,0.06972,0.06973,0.06974,0.06975,0.06976,0.06977,0.06978,0.06979,0.0698,0.06981,0.06982,0.06983,0.06984,0.06985,0.06986,0.06987,0.06988,0.06989,0.0699,0.06991,0.06992,0.06993,0.06994,0.06995,0.06996,0.06997,0.06998,0.06999,0.07,0.07001,0.07002,0.07003,0.07004,0.07005,0.07006,0.07007,0.07008,0.07009,0.0701,0.07011,0.07012,0.07013,0.07014,0.07015,0.07016,0.07017,0.07018,0.07019,0.0702,0.07021,0.07022,0.07023,0.07024,0.07025,0.07026,0.07027,0.07028,0.07029,0.0703,0.07031,0.07032,0.07033,0.07034,0.07035,0.07036,0.07037,0.07038,0.07039,0.0704,0.07041,0.07042,0.07043,0.07044,0.07045,0.07046,0.07047,0.07048,0.07049,0.0705,0.07051,0.07052,0.07053,0.07054,0.07055,0.07056,0.07057,0.07058,0.07059,0.0706,0.07061,0.07062,0.07063,0.07064,0.07065,0.07066,0.07067,0.07068,0.07069,0.0707,0.07071,0.07072,0.07073,0.07074,0.07075,0.07076,0.07077,0.07078,0.07079,0.0708,0.07081,0.07082,0.07083,0.07084,0.07085,0.07086,0.07087,0.07088,0.07089,0.0709,0.07091,0.07092,0.07093,0.07094,0.07095,0.07096,0.07097,0.07098,0.07099,0.071,0.07101,0.07102,0.07103,0.07104,0.07105,0.07106,0.07107,0.07108,0.07109,0.0711,0.07111,0.07112,0.07113,0.07114,0.07115,0.07116,0.07117,0.07118,0.07119,0.0712,0.07121,0.07122,0.07123,0.07124,0.07125,0.07126,0.07127,0.07128,0.07129,0.0713,0.07131,0.07132,0.07133,0.07134,0.07135,0.07136,0.07137,0.07138,0.07139,0.0714,0.07141,0.07142,0.07143,0.07144,0.07145,0.07146,0.07147,0.07148,0.07149,0.0715,0.07151,0.07152,0.07153,0.07154,0.07155,0.07156,0.07157,0.07158,0.07159,0.0716,0.07161,0.07162,0.07163,0.07164,0.07165,0.07166,0.07167,0.07168,0.07169,0.0717,0.07171,0.07172,0.07173,0.07174,0.07175,0.07176,0.07177,0.07178,0.07179,0.0718,0.07181,0.07182,0.07183,0.07184,0.07185,0.07186,0.07187,0.07188,0.07189,0.0719,0.07191,0.07192,0.07193,0.07194,0.07195,0.07196,0.07197,0.07198,0.07199,0.072,0.07201,0.07202,0.07203,0.07204,0.07205,0.07206,0.07207,0.07208,0.07209,0.0721,0.07211,0.07212,0.07213,0.07214,0.07215,0.07216,0.07217,0.07218,0.07219,0.0722,0.07221,0.07222,0.07223,0.07224,0.07225,0.07226,0.07227,0.07228,0.07229,0.0723,0.07231,0.07232,0.07233,0.07234,0.07235,0.07236,0.07237,0.07238,0.07239,0.0724,0.07241,0.07242,0.07243,0.07244,0.07245,0.07246,0.07247,0.07248,0.07249,0.0725,0.07251,0.07252,0.07253,0.07254,0.07255,0.07256,0.07257,0.07258,0.07259,0.0726,0.07261,0.07262,0.07263,0.07264,0.07265,0.07266,0.07267,0.07268,0.07269,0.0727,0.07271,0.07272,0.07273,0.07274,0.07275,0.07276,0.07277,0.07278,0.07279,0.0728,0.07281,0.07282,0.07283,0.07284,0.07285,0.07286,0.07287,0.07288,0.07289,0.0729,0.07291,0.07292,0.07293,0.07294,0.07295,0.07296,0.07297,0.07298,0.07299,0.073,0.07301,0.07302,0.07303,0.07304,0.07305,0.07306,0.07307,0.07308,0.07309,0.0731,0.07311,0.07312,0.07313,0.07314,0.07315,0.07316,0.07317,0.07318,0.07319,0.0732,0.07321,0.07322,0.07323,0.07324,0.07325,0.07326,0.07327,0.07328,0.07329,0.0733,0.07331,0.07332,0.07333,0.07334,0.07335,0.07336,0.07337,0.07338,0.07339,0.0734,0.07341,0.07342,0.07343,0.07344,0.07345,0.07346,0.07347,0.07348,0.07349,0.0735,0.07351,0.07352,0.07353,0.07354,0.07355,0.07356,0.07357,0.07358,0.07359,0.0736,0.07361,0.07362,0.07363,0.07364,0.07365,0.07366,0.07367,0.07368,0.07369,0.0737,0.07371,0.07372,0.07373,0.07374,0.07375,0.07376,0.07377,0.07378,0.07379,0.0738,0.07381,0.07382,0.07383,0.07384,0.07385,0.07386,0.07387,0.07388,0.07389,0.0739,0.07391,0.07392,0.07393,0.07394,0.07395,0.07396,0.07397,0.07398,0.07399,0.074,0.07401,0.07402,0.07403,0.07404,0.07405,0.07406,0.07407,0.07408,0.07409,0.0741,0.07411,0.07412,0.07413,0.07414,0.07415,0.07416,0.07417,0.07418,0.07419,0.0742,0.07421,0.07422,0.07423,0.07424,0.07425,0.07426,0.07427,0.07428,0.07429,0.0743,0.07431,0.07432,0.07433,0.07434,0.07435,0.07436,0.07437,0.07438,0.07439,0.0744,0.07441,0.07442,0.07443,0.07444,0.07445,0.07446,0.07447,0.07448,0.07449,0.0745,0.07451,0.07452,0.07453,0.07454,0.07455,0.07456,0.07457,0.07458,0.07459,0.0746,0.07461,0.07462,0.07463,0.07464,0.07465,0.07466,0.07467,0.07468,0.07469,0.0747,0.07471,0.07472,0.07473,0.07474,0.07475,0.07476,0.07477,0.07478,0.07479,0.0748,0.07481,0.07482,0.07483,0.07484,0.07485,0.07486,0.07487,0.07488,0.07489,0.0749,0.07491,0.07492,0.07493,0.07494,0.07495,0.07496,0.07497,0.07498,0.07499,0.075,0.07501,0.07502,0.07503,0.07504,0.07505,0.07506,0.07507,0.07508,0.07509,0.0751,0.07511,0.07512,0.07513,0.07514,0.07515,0.07516,0.07517,0.07518,0.07519,0.0752,0.07521,0.07522,0.07523,0.07524,0.07525,0.07526,0.07527,0.07528,0.07529,0.0753,0.07531,0.07532,0.07533,0.07534,0.07535,0.07536,0.07537,0.07538,0.07539,0.0754,0.07541,0.07542,0.07543,0.07544,0.07545,0.07546,0.07547,0.07548,0.07549,0.0755,0.07551,0.07552,0.07553,0.07554,0.07555,0.07556,0.07557,0.07558,0.07559,0.0756,0.07561,0.07562,0.07563,0.07564,0.07565,0.07566,0.07567,0.07568,0.07569,0.0757,0.07571,0.07572,0.07573,0.07574,0.07575,0.07576,0.07577,0.07578,0.07579,0.0758,0.07581,0.07582,0.07583,0.07584,0.07585,0.07586,0.07587,0.07588,0.07589,0.0759,0.07591,0.07592,0.07593,0.07594,0.07595,0.07596,0.07597,0.07598,0.07599,0.076,0.07601,0.07602,0.07603,0.07604,0.07605,0.07606,0.07607,0.07608,0.07609,0.0761,0.07611,0.07612,0.07613,0.07614,0.07615,0.07616,0.07617,0.07618,0.07619,0.0762,0.07621,0.07622,0.07623,0.07624,0.07625,0.07626,0.07627,0.07628,0.07629,0.0763,0.07631,0.07632,0.07633,0.07634,0.07635,0.07636,0.07637,0.07638,0.07639,0.0764,0.07641,0.07642,0.07643,0.07644,0.07645,0.07646,0.07647,0.07648,0.07649,0.0765,0.07651,0.07652,0.07653,0.07654,0.07655,0.07656,0.07657,0.07658,0.07659,0.0766,0.07661,0.07662,0.07663,0.07664,0.07665,0.07666,0.07667,0.07668,0.07669,0.0767,0.07671,0.07672,0.07673,0.07674,0.07675,0.07676,0.07677,0.07678,0.07679,0.0768,0.07681,0.07682,0.07683,0.07684,0.07685,0.07686,0.07687,0.07688,0.07689,0.0769,0.07691,0.07692,0.07693,0.07694,0.07695,0.07696,0.07697,0.07698,0.07699,0.077,0.07701,0.07702,0.07703,0.07704,0.07705,0.07706,0.07707,0.07708,0.07709,0.0771,0.07711,0.07712,0.07713,0.07714,0.07715,0.07716,0.07717,0.07718,0.07719,0.0772,0.07721,0.07722,0.07723,0.07724,0.07725,0.07726,0.07727,0.07728,0.07729,0.0773,0.07731,0.07732,0.07733,0.07734,0.07735,0.07736,0.07737,0.07738,0.07739,0.0774,0.07741,0.07742,0.07743,0.07744,0.07745,0.07746,0.07747,0.07748,0.07749,0.0775,0.07751,0.07752,0.07753,0.07754,0.07755,0.07756,0.07757,0.07758,0.07759,0.0776,0.07761,0.07762,0.07763,0.07764,0.07765,0.07766,0.07767,0.07768,0.07769,0.0777,0.07771,0.07772,0.07773,0.07774,0.07775,0.07776,0.07777,0.07778,0.07779,0.0778,0.07781,0.07782,0.07783,0.07784,0.07785,0.07786,0.07787,0.07788,0.07789,0.0779,0.07791,0.07792,0.07793,0.07794,0.07795,0.07796,0.07797,0.07798,0.07799,0.078,0.07801,0.07802,0.07803,0.07804,0.07805,0.07806,0.07807,0.07808,0.07809,0.0781,0.07811,0.07812,0.07813,0.07814,0.07815,0.07816,0.07817,0.07818,0.07819,0.0782,0.07821,0.07822,0.07823,0.07824,0.07825,0.07826,0.07827,0.07828,0.07829,0.0783,0.07831,0.07832,0.07833,0.07834,0.07835,0.07836,0.07837,0.07838,0.07839,0.0784,0.07841,0.07842,0.07843,0.07844,0.07845,0.07846,0.07847,0.07848,0.07849,0.0785,0.07851,0.07852,0.07853,0.07854,0.07855,0.07856,0.07857,0.07858,0.07859,0.0786,0.07861,0.07862,0.07863,0.07864,0.07865,0.07866,0.07867,0.07868,0.07869,0.0787,0.07871,0.07872,0.07873,0.07874,0.07875,0.07876,0.07877,0.07878,0.07879,0.0788,0.07881,0.07882,0.07883,0.07884,0.07885,0.07886,0.07887,0.07888,0.07889,0.0789,0.07891,0.07892,0.07893,0.07894,0.07895,0.07896,0.07897,0.07898,0.07899,0.079,0.07901,0.07902,0.07903,0.07904,0.07905,0.07906,0.07907,0.07908,0.07909,0.0791,0.07911,0.07912,0.07913,0.07914,0.07915,0.07916,0.07917,0.07918,0.07919,0.0792,0.07921,0.07922,0.07923,0.07924,0.07925,0.07926,0.07927,0.07928,0.07929,0.0793,0.07931,0.07932,0.07933,0.07934,0.07935,0.07936,0.07937,0.07938,0.07939,0.0794,0.07941,0.07942,0.07943,0.07944,0.07945,0.07946,0.07947,0.07948,0.07949,0.0795,0.07951,0.07952,0.07953,0.07954,0.07955,0.07956,0.07957,0.07958,0.07959,0.0796,0.07961,0.07962,0.07963,0.07964,0.07965,0.07966,0.07967,0.07968,0.07969,0.0797,0.07971,0.07972,0.07973,0.07974,0.07975,0.07976,0.07977,0.07978,0.07979,0.0798,0.07981,0.07982,0.07983,0.07984,0.07985,0.07986,0.07987,0.07988,0.07989,0.0799,0.07991,0.07992,0.07993,0.07994,0.07995,0.07996,0.07997,0.07998,0.07999,0.08,0.08001,0.08002,0.08003,0.08004,0.08005,0.08006,0.08007,0.08008,0.08009,0.0801,0.08011,0.08012,0.08013,0.08014,0.08015,0.08016,0.08017,0.08018,0.08019,0.0802,0.08021,0.08022,0.08023,0.08024,0.08025,0.08026,0.08027,0.08028,0.08029,0.0803,0.08031,0.08032,0.08033,0.08034,0.08035,0.08036,0.08037,0.08038,0.08039,0.0804,0.08041,0.08042,0.08043,0.08044,0.08045,0.08046,0.08047,0.08048,0.08049,0.0805,0.08051,0.08052,0.08053,0.08054,0.08055,0.08056,0.08057,0.08058,0.08059,0.0806,0.08061,0.08062,0.08063,0.08064,0.08065,0.08066,0.08067,0.08068,0.08069,0.0807,0.08071,0.08072,0.08073,0.08074,0.08075,0.08076,0.08077,0.08078,0.08079,0.0808,0.08081,0.08082,0.08083,0.08084,0.08085,0.08086,0.08087,0.08088,0.08089,0.0809,0.08091,0.08092,0.08093,0.08094,0.08095,0.08096,0.08097,0.08098,0.08099,0.081,0.08101,0.08102,0.08103,0.08104,0.08105,0.08106,0.08107,0.08108,0.08109,0.0811,0.08111,0.08112,0.08113,0.08114,0.08115,0.08116,0.08117,0.08118,0.08119,0.0812,0.08121,0.08122,0.08123,0.08124,0.08125,0.08126,0.08127,0.08128,0.08129,0.0813,0.08131,0.08132,0.08133,0.08134,0.08135,0.08136,0.08137,0.08138,0.08139,0.0814,0.08141,0.08142,0.08143,0.08144,0.08145,0.08146,0.08147,0.08148,0.08149,0.0815,0.08151,0.08152,0.08153,0.08154,0.08155,0.08156,0.08157,0.08158,0.08159,0.0816,0.08161,0.08162,0.08163,0.08164,0.08165,0.08166,0.08167,0.08168,0.08169,0.0817,0.08171,0.08172,0.08173,0.08174,0.08175,0.08176,0.08177,0.08178,0.08179,0.0818,0.08181,0.08182,0.08183,0.08184,0.08185,0.08186,0.08187,0.08188,0.08189,0.0819,0.08191,0.08192,0.08193,0.08194,0.08195,0.08196,0.08197,0.08198,0.08199,0.082,0.08201,0.08202,0.08203,0.08204,0.08205,0.08206,0.08207,0.08208,0.08209,0.0821,0.08211,0.08212,0.08213,0.08214,0.08215,0.08216,0.08217,0.08218,0.08219,0.0822,0.08221,0.08222,0.08223,0.08224,0.08225,0.08226,0.08227,0.08228,0.08229,0.0823,0.08231,0.08232,0.08233,0.08234,0.08235,0.08236,0.08237,0.08238,0.08239,0.0824,0.08241,0.08242,0.08243,0.08244,0.08245,0.08246,0.08247,0.08248,0.08249,0.0825,0.08251,0.08252,0.08253,0.08254,0.08255,0.08256,0.08257,0.08258,0.08259,0.0826,0.08261,0.08262,0.08263,0.08264,0.08265,0.08266,0.08267,0.08268,0.08269,0.0827,0.08271,0.08272,0.08273,0.08274,0.08275,0.08276,0.08277,0.08278,0.08279,0.0828,0.08281,0.08282,0.08283,0.08284,0.08285,0.08286,0.08287,0.08288,0.08289,0.0829,0.08291,0.08292,0.08293,0.08294,0.08295,0.08296,0.08297,0.08298,0.08299,0.083,0.08301,0.08302,0.08303,0.08304,0.08305,0.08306,0.08307,0.08308,0.08309,0.0831,0.08311,0.08312,0.08313,0.08314,0.08315,0.08316,0.08317,0.08318,0.08319,0.0832,0.08321,0.08322,0.08323,0.08324,0.08325,0.08326,0.08327,0.08328,0.08329,0.0833,0.08331,0.08332,0.08333,0.08334,0.08335,0.08336,0.08337,0.08338,0.08339,0.0834,0.08341,0.08342,0.08343,0.08344,0.08345,0.08346,0.08347,0.08348,0.08349,0.0835,0.08351,0.08352,0.08353,0.08354,0.08355,0.08356,0.08357,0.08358,0.08359,0.0836,0.08361,0.08362,0.08363,0.08364,0.08365,0.08366,0.08367,0.08368,0.08369,0.0837,0.08371,0.08372,0.08373,0.08374,0.08375,0.08376,0.08377,0.08378,0.08379,0.0838,0.08381,0.08382,0.08383,0.08384,0.08385,0.08386,0.08387,0.08388,0.08389,0.0839,0.08391,0.08392,0.08393,0.08394,0.08395,0.08396,0.08397,0.08398,0.08399,0.084,0.08401,0.08402,0.08403,0.08404,0.08405,0.08406,0.08407,0.08408,0.08409,0.0841,0.08411,0.08412,0.08413,0.08414,0.08415,0.08416,0.08417,0.08418,0.08419,0.0842,0.08421,0.08422,0.08423,0.08424,0.08425,0.08426,0.08427,0.08428,0.08429,0.0843,0.08431,0.08432,0.08433,0.08434,0.08435,0.08436,0.08437,0.08438,0.08439,0.0844,0.08441,0.08442,0.08443,0.08444,0.08445,0.08446,0.08447,0.08448,0.08449,0.0845,0.08451,0.08452,0.08453,0.08454,0.08455,0.08456,0.08457,0.08458,0.08459,0.0846,0.08461,0.08462,0.08463,0.08464,0.08465,0.08466,0.08467,0.08468,0.08469,0.0847,0.08471,0.08472,0.08473,0.08474,0.08475,0.08476,0.08477,0.08478,0.08479,0.0848,0.08481,0.08482,0.08483,0.08484,0.08485,0.08486,0.08487,0.08488,0.08489,0.0849,0.08491,0.08492,0.08493,0.08494,0.08495,0.08496,0.08497,0.08498,0.08499,0.085,0.08501,0.08502,0.08503,0.08504,0.08505,0.08506,0.08507,0.08508,0.08509,0.0851,0.08511,0.08512,0.08513,0.08514,0.08515,0.08516,0.08517,0.08518,0.08519,0.0852,0.08521,0.08522,0.08523,0.08524,0.08525,0.08526,0.08527,0.08528,0.08529,0.0853,0.08531,0.08532,0.08533,0.08534,0.08535,0.08536,0.08537,0.08538,0.08539,0.0854,0.08541,0.08542,0.08543,0.08544,0.08545,0.08546,0.08547,0.08548,0.08549,0.0855,0.08551,0.08552,0.08553,0.08554,0.08555,0.08556,0.08557,0.08558,0.08559,0.0856,0.08561,0.08562,0.08563,0.08564,0.08565,0.08566,0.08567,0.08568,0.08569,0.0857,0.08571,0.08572,0.08573,0.08574,0.08575,0.08576,0.08577,0.08578,0.08579,0.0858,0.08581,0.08582,0.08583,0.08584,0.08585,0.08586,0.08587,0.08588,0.08589,0.0859,0.08591,0.08592,0.08593,0.08594,0.08595,0.08596,0.08597,0.08598,0.08599,0.086,0.08601,0.08602,0.08603,0.08604,0.08605,0.08606,0.08607,0.08608,0.08609,0.0861,0.08611,0.08612,0.08613,0.08614,0.08615,0.08616,0.08617,0.08618,0.08619,0.0862,0.08621,0.08622,0.08623,0.08624,0.08625,0.08626,0.08627,0.08628,0.08629,0.0863,0.08631,0.08632,0.08633,0.08634,0.08635,0.08636,0.08637,0.08638,0.08639,0.0864,0.08641,0.08642,0.08643,0.08644,0.08645,0.08646,0.08647,0.08648,0.08649,0.0865,0.08651,0.08652,0.08653,0.08654,0.08655,0.08656,0.08657,0.08658,0.08659,0.0866,0.08661,0.08662,0.08663,0.08664,0.08665,0.08666,0.08667,0.08668,0.08669,0.0867,0.08671,0.08672,0.08673,0.08674,0.08675,0.08676,0.08677,0.08678,0.08679,0.0868,0.08681,0.08682,0.08683,0.08684,0.08685,0.08686,0.08687,0.08688,0.08689,0.0869,0.08691,0.08692,0.08693,0.08694,0.08695,0.08696,0.08697,0.08698,0.08699,0.087,0.08701,0.08702,0.08703,0.08704,0.08705,0.08706,0.08707,0.08708,0.08709,0.0871,0.08711,0.08712,0.08713,0.08714,0.08715,0.08716,0.08717,0.08718,0.08719,0.0872,0.08721,0.08722,0.08723,0.08724,0.08725,0.08726,0.08727,0.08728,0.08729,0.0873,0.08731,0.08732,0.08733,0.08734,0.08735,0.08736,0.08737,0.08738,0.08739,0.0874,0.08741,0.08742,0.08743,0.08744,0.08745,0.08746,0.08747,0.08748,0.08749,0.0875,0.08751,0.08752,0.08753,0.08754,0.08755,0.08756,0.08757,0.08758,0.08759,0.0876,0.08761,0.08762,0.08763,0.08764,0.08765,0.08766,0.08767,0.08768,0.08769,0.0877,0.08771,0.08772,0.08773,0.08774,0.08775,0.08776,0.08777,0.08778,0.08779,0.0878,0.08781,0.08782,0.08783,0.08784,0.08785,0.08786,0.08787,0.08788,0.08789,0.0879,0.08791,0.08792,0.08793,0.08794,0.08795,0.08796,0.08797,0.08798,0.08799,0.088,0.08801,0.08802,0.08803,0.08804,0.08805,0.08806,0.08807,0.08808,0.08809,0.0881,0.08811,0.08812,0.08813,0.08814,0.08815,0.08816,0.08817,0.08818,0.08819,0.0882,0.08821,0.08822,0.08823,0.08824,0.08825,0.08826,0.08827,0.08828,0.08829,0.0883,0.08831,0.08832,0.08833,0.08834,0.08835,0.08836,0.08837,0.08838,0.08839,0.0884,0.08841,0.08842,0.08843,0.08844,0.08845,0.08846,0.08847,0.08848,0.08849,0.0885,0.08851,0.08852,0.08853,0.08854,0.08855,0.08856,0.08857,0.08858,0.08859,0.0886,0.08861,0.08862,0.08863,0.08864,0.08865,0.08866,0.08867,0.08868,0.08869,0.0887,0.08871,0.08872,0.08873,0.08874,0.08875,0.08876,0.08877,0.08878,0.08879,0.0888,0.08881,0.08882,0.08883,0.08884,0.08885,0.08886,0.08887,0.08888,0.08889,0.0889,0.08891,0.08892,0.08893,0.08894,0.08895,0.08896,0.08897,0.08898,0.08899,0.089,0.08901,0.08902,0.08903,0.08904,0.08905,0.08906,0.08907,0.08908,0.08909,0.0891,0.08911,0.08912,0.08913,0.08914,0.08915,0.08916,0.08917,0.08918,0.08919,0.0892,0.08921,0.08922,0.08923,0.08924,0.08925,0.08926,0.08927,0.08928,0.08929,0.0893,0.08931,0.08932,0.08933,0.08934,0.08935,0.08936,0.08937,0.08938,0.08939,0.0894,0.08941,0.08942,0.08943,0.08944,0.08945,0.08946,0.08947,0.08948,0.08949,0.0895,0.08951,0.08952,0.08953,0.08954,0.08955,0.08956,0.08957,0.08958,0.08959,0.0896,0.08961,0.08962,0.08963,0.08964,0.08965,0.08966,0.08967,0.08968,0.08969,0.0897,0.08971,0.08972,0.08973,0.08974,0.08975,0.08976,0.08977,0.08978,0.08979,0.0898,0.08981,0.08982,0.08983,0.08984,0.08985,0.08986,0.08987,0.08988,0.08989,0.0899,0.08991,0.08992,0.08993,0.08994,0.08995,0.08996,0.08997,0.08998,0.08999,0.09,0.09001,0.09002,0.09003,0.09004,0.09005,0.09006,0.09007,0.09008,0.09009,0.0901,0.09011,0.09012,0.09013,0.09014,0.09015,0.09016,0.09017,0.09018,0.09019,0.0902,0.09021,0.09022,0.09023,0.09024,0.09025,0.09026,0.09027,0.09028,0.09029,0.0903,0.09031,0.09032,0.09033,0.09034,0.09035,0.09036,0.09037,0.09038,0.09039,0.0904,0.09041,0.09042,0.09043,0.09044,0.09045,0.09046,0.09047,0.09048,0.09049,0.0905,0.09051,0.09052,0.09053,0.09054,0.09055,0.09056,0.09057,0.09058,0.09059,0.0906,0.09061,0.09062,0.09063,0.09064,0.09065,0.09066,0.09067,0.09068,0.09069,0.0907,0.09071,0.09072,0.09073,0.09074,0.09075,0.09076,0.09077,0.09078,0.09079,0.0908,0.09081,0.09082,0.09083,0.09084,0.09085,0.09086,0.09087,0.09088,0.09089,0.0909,0.09091,0.09092,0.09093,0.09094,0.09095,0.09096,0.09097,0.09098,0.09099,0.091,0.09101,0.09102,0.09103,0.09104,0.09105,0.09106,0.09107,0.09108,0.09109,0.0911,0.09111,0.09112,0.09113,0.09114,0.09115,0.09116,0.09117,0.09118,0.09119,0.0912,0.09121,0.09122,0.09123,0.09124,0.09125,0.09126,0.09127,0.09128,0.09129,0.0913,0.09131,0.09132,0.09133,0.09134,0.09135,0.09136,0.09137,0.09138,0.09139,0.0914,0.09141,0.09142,0.09143,0.09144,0.09145,0.09146,0.09147,0.09148,0.09149,0.0915,0.09151,0.09152,0.09153,0.09154,0.09155,0.09156,0.09157,0.09158,0.09159,0.0916,0.09161,0.09162,0.09163,0.09164,0.09165,0.09166,0.09167,0.09168,0.09169,0.0917,0.09171,0.09172,0.09173,0.09174,0.09175,0.09176,0.09177,0.09178,0.09179,0.0918,0.09181,0.09182,0.09183,0.09184,0.09185,0.09186,0.09187,0.09188,0.09189,0.0919,0.09191,0.09192,0.09193,0.09194,0.09195,0.09196,0.09197,0.09198,0.09199,0.092,0.09201,0.09202,0.09203,0.09204,0.09205,0.09206,0.09207,0.09208,0.09209,0.0921,0.09211,0.09212,0.09213,0.09214,0.09215,0.09216,0.09217,0.09218,0.09219,0.0922,0.09221,0.09222,0.09223,0.09224,0.09225,0.09226,0.09227,0.09228,0.09229,0.0923,0.09231,0.09232,0.09233,0.09234,0.09235,0.09236,0.09237,0.09238,0.09239,0.0924,0.09241,0.09242,0.09243,0.09244,0.09245,0.09246,0.09247,0.09248,0.09249,0.0925,0.09251,0.09252,0.09253,0.09254,0.09255,0.09256,0.09257,0.09258,0.09259,0.0926,0.09261,0.09262,0.09263,0.09264,0.09265,0.09266,0.09267,0.09268,0.09269,0.0927,0.09271,0.09272,0.09273,0.09274,0.09275,0.09276,0.09277,0.09278,0.09279,0.0928,0.09281,0.09282,0.09283,0.09284,0.09285,0.09286,0.09287,0.09288,0.09289,0.0929,0.09291,0.09292,0.09293,0.09294,0.09295,0.09296,0.09297,0.09298,0.09299,0.093,0.09301,0.09302,0.09303,0.09304,0.09305,0.09306,0.09307,0.09308,0.09309,0.0931,0.09311,0.09312,0.09313,0.09314,0.09315,0.09316,0.09317,0.09318,0.09319,0.0932,0.09321,0.09322,0.09323,0.09324,0.09325,0.09326,0.09327,0.09328,0.09329,0.0933,0.09331,0.09332,0.09333,0.09334,0.09335,0.09336,0.09337,0.09338,0.09339,0.0934,0.09341,0.09342,0.09343,0.09344,0.09345,0.09346,0.09347,0.09348,0.09349,0.0935,0.09351,0.09352,0.09353,0.09354,0.09355,0.09356,0.09357,0.09358,0.09359,0.0936,0.09361,0.09362,0.09363,0.09364,0.09365,0.09366,0.09367,0.09368,0.09369,0.0937,0.09371,0.09372,0.09373,0.09374,0.09375,0.09376,0.09377,0.09378,0.09379,0.0938,0.09381,0.09382,0.09383,0.09384,0.09385,0.09386,0.09387,0.09388,0.09389,0.0939,0.09391,0.09392,0.09393,0.09394,0.09395,0.09396,0.09397,0.09398,0.09399,0.094,0.09401,0.09402,0.09403,0.09404,0.09405,0.09406,0.09407,0.09408,0.09409,0.0941,0.09411,0.09412,0.09413,0.09414,0.09415,0.09416,0.09417,0.09418,0.09419,0.0942,0.09421,0.09422,0.09423,0.09424,0.09425,0.09426,0.09427,0.09428,0.09429,0.0943,0.09431,0.09432,0.09433,0.09434,0.09435,0.09436,0.09437,0.09438,0.09439,0.0944,0.09441,0.09442,0.09443,0.09444,0.09445,0.09446,0.09447,0.09448,0.09449,0.0945,0.09451,0.09452,0.09453,0.09454,0.09455,0.09456,0.09457,0.09458,0.09459,0.0946,0.09461,0.09462,0.09463,0.09464,0.09465,0.09466,0.09467,0.09468,0.09469,0.0947,0.09471,0.09472,0.09473,0.09474,0.09475,0.09476,0.09477,0.09478,0.09479,0.0948,0.09481,0.09482,0.09483,0.09484,0.09485,0.09486,0.09487,0.09488,0.09489,0.0949,0.09491,0.09492,0.09493,0.09494,0.09495,0.09496,0.09497,0.09498,0.09499,0.095,0.09501,0.09502,0.09503,0.09504,0.09505,0.09506,0.09507,0.09508,0.09509,0.0951,0.09511,0.09512,0.09513,0.09514,0.09515,0.09516,0.09517,0.09518,0.09519,0.0952,0.09521,0.09522,0.09523,0.09524,0.09525,0.09526,0.09527,0.09528,0.09529,0.0953,0.09531,0.09532,0.09533,0.09534,0.09535,0.09536,0.09537,0.09538,0.09539,0.0954,0.09541,0.09542,0.09543,0.09544,0.09545,0.09546,0.09547,0.09548,0.09549,0.0955,0.09551,0.09552,0.09553,0.09554,0.09555,0.09556,0.09557,0.09558,0.09559,0.0956,0.09561,0.09562,0.09563,0.09564,0.09565,0.09566,0.09567,0.09568,0.09569,0.0957,0.09571,0.09572,0.09573,0.09574,0.09575,0.09576,0.09577,0.09578,0.09579,0.0958,0.09581,0.09582,0.09583,0.09584,0.09585,0.09586,0.09587,0.09588,0.09589,0.0959,0.09591,0.09592,0.09593,0.09594,0.09595,0.09596,0.09597,0.09598,0.09599,0.096,0.09601,0.09602,0.09603,0.09604,0.09605,0.09606,0.09607,0.09608,0.09609,0.0961,0.09611,0.09612,0.09613,0.09614,0.09615,0.09616,0.09617,0.09618,0.09619,0.0962,0.09621,0.09622,0.09623,0.09624,0.09625,0.09626,0.09627,0.09628,0.09629,0.0963,0.09631,0.09632,0.09633,0.09634,0.09635,0.09636,0.09637,0.09638,0.09639,0.0964,0.09641,0.09642,0.09643,0.09644,0.09645,0.09646,0.09647,0.09648,0.09649,0.0965,0.09651,0.09652,0.09653,0.09654,0.09655,0.09656,0.09657,0.09658,0.09659,0.0966,0.09661,0.09662,0.09663,0.09664,0.09665,0.09666,0.09667,0.09668,0.09669,0.0967,0.09671,0.09672,0.09673,0.09674,0.09675,0.09676,0.09677,0.09678,0.09679,0.0968,0.09681,0.09682,0.09683,0.09684,0.09685,0.09686,0.09687,0.09688,0.09689,0.0969,0.09691,0.09692,0.09693,0.09694,0.09695,0.09696,0.09697,0.09698,0.09699,0.097,0.09701,0.09702,0.09703,0.09704,0.09705,0.09706,0.09707,0.09708,0.09709,0.0971,0.09711,0.09712,0.09713,0.09714,0.09715,0.09716,0.09717,0.09718,0.09719,0.0972,0.09721,0.09722,0.09723,0.09724,0.09725,0.09726,0.09727,0.09728,0.09729,0.0973,0.09731,0.09732,0.09733,0.09734,0.09735,0.09736,0.09737,0.09738,0.09739,0.0974,0.09741,0.09742,0.09743,0.09744,0.09745,0.09746,0.09747,0.09748,0.09749,0.0975,0.09751,0.09752,0.09753,0.09754,0.09755,0.09756,0.09757,0.09758,0.09759,0.0976,0.09761,0.09762,0.09763,0.09764,0.09765,0.09766,0.09767,0.09768,0.09769,0.0977,0.09771,0.09772,0.09773,0.09774,0.09775,0.09776,0.09777,0.09778,0.09779,0.0978,0.09781,0.09782,0.09783,0.09784,0.09785,0.09786,0.09787,0.09788,0.09789,0.0979,0.09791,0.09792,0.09793,0.09794,0.09795,0.09796,0.09797,0.09798,0.09799,0.098,0.09801,0.09802,0.09803,0.09804,0.09805,0.09806,0.09807,0.09808,0.09809,0.0981,0.09811,0.09812,0.09813,0.09814,0.09815,0.09816,0.09817,0.09818,0.09819,0.0982,0.09821,0.09822,0.09823,0.09824,0.09825,0.09826,0.09827,0.09828,0.09829,0.0983,0.09831,0.09832,0.09833,0.09834,0.09835,0.09836,0.09837,0.09838,0.09839,0.0984,0.09841,0.09842,0.09843,0.09844,0.09845,0.09846,0.09847,0.09848,0.09849,0.0985,0.09851,0.09852,0.09853,0.09854,0.09855,0.09856,0.09857,0.09858,0.09859,0.0986,0.09861,0.09862,0.09863,0.09864,0.09865,0.09866,0.09867,0.09868,0.09869,0.0987,0.09871,0.09872,0.09873,0.09874,0.09875,0.09876,0.09877,0.09878,0.09879,0.0988,0.09881,0.09882,0.09883,0.09884,0.09885,0.09886,0.09887,0.09888,0.09889,0.0989,0.09891,0.09892,0.09893,0.09894,0.09895,0.09896,0.09897,0.09898,0.09899,0.099,0.09901,0.09902,0.09903,0.09904,0.09905,0.09906,0.09907,0.09908,0.09909,0.0991,0.09911,0.09912,0.09913,0.09914,0.09915,0.09916,0.09917,0.09918,0.09919,0.0992,0.09921,0.09922,0.09923,0.09924,0.09925,0.09926,0.09927,0.09928,0.09929,0.0993,0.09931,0.09932,0.09933,0.09934,0.09935,0.09936,0.09937,0.09938,0.09939,0.0994,0.09941,0.09942,0.09943,0.09944,0.09945,0.09946,0.09947,0.09948,0.09949,0.0995,0.09951,0.09952,0.09953,0.09954,0.09955,0.09956,0.09957,0.09958,0.09959,0.0996,0.09961,0.09962,0.09963,0.09964,0.09965,0.09966,0.09967,0.09968,0.09969,0.0997,0.09971,0.09972,0.09973,0.09974,0.09975,0.09976,0.09977,0.09978,0.09979,0.0998,0.09981,0.09982,0.09983,0.09984,0.09985,0.09986,0.09987,0.09988,0.09989,0.0999,0.09991,0.09992,0.09993,0.09994,0.09995,0.09996,0.09997,0.09998,0.09999,0.1]}],{"xaxis":{"title":{"text":"Time in µs"}},"template":{"layout":{"coloraxis":{"colorbar":{"ticks":"","outlinewidth":0}},"xaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"hovermode":"closest","paper_bgcolor":"white","geo":{"showlakes":true,"showland":true,"landcolor":"#E5ECF6","bgcolor":"white","subunitcolor":"white","lakecolor":"white"},"colorscale":{"sequential":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]],"diverging":[[0,"#8e0152"],[0.1,"#c51b7d"],[0.2,"#de77ae"],[0.3,"#f1b6da"],[0.4,"#fde0ef"],[0.5,"#f7f7f7"],[0.6,"#e6f5d0"],[0.7,"#b8e186"],[0.8,"#7fbc41"],[0.9,"#4d9221"],[1,"#276419"]],"sequentialminus":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]},"yaxis":{"gridcolor":"white","zerolinewidth":2,"title":{"standoff":15},"ticks":"","zerolinecolor":"white","automargin":true,"linecolor":"white"},"shapedefaults":{"line":{"color":"#2a3f5f"}},"hoverlabel":{"align":"left"},"mapbox":{"style":"light"},"polar":{"angularaxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","radialaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"autotypenumbers":"strict","font":{"color":"#2a3f5f"},"ternary":{"baxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"bgcolor":"#E5ECF6","caxis":{"gridcolor":"white","ticks":"","linecolor":"white"},"aaxis":{"gridcolor":"white","ticks":"","linecolor":"white"}},"annotationdefaults":{"arrowhead":0,"arrowwidth":1,"arrowcolor":"#2a3f5f"},"plot_bgcolor":"#E5ECF6","title":{"x":0.05},"scene":{"xaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"zaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"},"yaxis":{"gridcolor":"white","gridwidth":2,"backgroundcolor":"#E5ECF6","ticks":"","showbackground":true,"zerolinecolor":"white","linecolor":"white"}},"colorway":["#636efa","#EF553B","#00cc96","#ab63fa","#FFA15A","#19d3f3","#FF6692","#B6E880","#FF97FF","#FECB52"]},"data":{"barpolar":[{"type":"barpolar","marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"carpet":[{"aaxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"},"type":"carpet","baxis":{"gridcolor":"white","endlinecolor":"#2a3f5f","minorgridcolor":"white","startlinecolor":"#2a3f5f","linecolor":"white"}}],"scatterpolar":[{"type":"scatterpolar","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"parcoords":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"parcoords"}],"scatter":[{"type":"scatter","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2dcontour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2dcontour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contour":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contour","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattercarpet":[{"type":"scattercarpet","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"mesh3d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"mesh3d"}],"surface":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"surface","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scattermapbox":[{"type":"scattermapbox","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergeo":[{"type":"scattergeo","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram":[{"type":"histogram","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"pie":[{"type":"pie","automargin":true}],"choropleth":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"choropleth"}],"heatmapgl":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmapgl","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"bar":[{"type":"bar","error_y":{"color":"#2a3f5f"},"error_x":{"color":"#2a3f5f"},"marker":{"line":{"color":"#E5ECF6","width":0.5}}}],"heatmap":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"heatmap","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"contourcarpet":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"contourcarpet"}],"table":[{"type":"table","header":{"line":{"color":"white"},"fill":{"color":"#C8D4E3"}},"cells":{"line":{"color":"white"},"fill":{"color":"#EBF0F8"}}}],"scatter3d":[{"line":{"colorbar":{"ticks":"","outlinewidth":0}},"type":"scatter3d","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scattergl":[{"type":"scattergl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"histogram2d":[{"colorbar":{"ticks":"","outlinewidth":0},"type":"histogram2d","colorscale":[[0.0,"#0d0887"],[0.1111111111111111,"#46039f"],[0.2222222222222222,"#7201a8"],[0.3333333333333333,"#9c179e"],[0.4444444444444444,"#bd3786"],[0.5555555555555556,"#d8576b"],[0.6666666666666666,"#ed7953"],[0.7777777777777778,"#fb9f3a"],[0.8888888888888888,"#fdca26"],[1.0,"#f0f921"]]}],"scatterternary":[{"type":"scatterternary","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}],"scatterpolargl":[{"type":"scatterpolargl","marker":{"colorbar":{"ticks":"","outlinewidth":0}}}]}},"margin":{"l":50,"b":50,"r":50,"t":60},"yaxis":{"title":{"text":"v_C \/ V"}}},{"showLink":false,"editable":false,"responsive":true,"staticPlot":false,"scrollZoom":true});
});





</script>
```






```julia

```
